/*
 * aceInternal.h
 *
 * Internal header file for the ACE simulated-annealing-based
 * array folder/compactor package.
 */

/* define this to create the paranoid library for application development
#define PARANOID
*/

/* include packages that ace uses					*/
#include "port.h"
#include "copyright.h"
#include "utility.h"
#include "errtrap.h"

/* if PARANOID is defined, turn on ace's paranoid comipile flag		*/
#ifdef PARANOID
#ifndef ACE_PARANOD
#define ACE_PARANOID
#endif /* ACE_PARANOID */
#endif /* PARANOID */

#ifdef ACE_PARANOID
#include "st.h"
#endif

#include "ace.h"

/* debugging and optimization for applications without UCFs
#define ACE_NO_UCF
*/

/* compile in the tracing code						*/
#define ACE_COMP_TRACE

/* standard macros							*/
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

/* User Constraint Function (UCF) facility structures.			*/
typedef struct aceUCF {
    double (* cnstr_fn)				/* constraint function	*/
	ARGS((int count, struct aceObject *(array[])));
    int (* check_fn)				/* constraint checking fn */
	ARGS((int count, struct aceObject *(array[])));
    double lastRet;				/* last return value	*/
    double restoreSave;				/* store last return value */
    unsigned eventMask;				/* event mask for this inst */
    short called;				/* bool set if called	*/
    int argCount;				/* number of args	*/
    aceObject **argObjs;			/* obj args to the user fn */
} aceUCF;

typedef struct aceUCFEvent {
    struct aceUCFEvent *next;			/* next in queue	*/
    int index;					/* index of instance	*/
} aceUCFEvent;

/* internal structures to describe the matrix.				*/
typedef struct aceRange {
    int lower, upper;				/* range of indices	*/
} aceRange;

typedef struct aceMaster {
    aceObject *objNext;				/* link in obj list	*/
    aceObjType objType;				/* type of object	*/
    aceId id;					/* id of master		*/
    char *name;					/* single char          */
    char *path;					/* full path (if appl.) */
    aceCoord cellCap;				/* capacity used by inst */
} aceMaster;

typedef struct aceCell {
    /* fields included in aceApCell					*/
    aceObject *objNext;				/* link in obj list	*/
    aceObjType objType;				/* type of object	*/
    aceId id;					/* unique id number	*/
    aceCoord loc;				/* current location	*/
    aceId masterId;				/* cell's master	*/
    /* internal ace fields						*/
    aceMaster *master;				/* master of this inst.	*/
    struct aceNet *rowNet, *colNet;		/* row/col nets		*/
} aceCell;

typedef struct aceNet {
    /* fields included in aceApNet					*/
    aceObject *objNext;				/* link in obj list	*/
    aceObjType objType;				/* type of object	*/
    aceId id;					/* unique id number	*/
    unsigned termMask;				/* term net bits	*/
    short isRow;				/* bool, set if row net	*/
    int rowcol;					/* row or col location	*/
    aceRange range;				/* span of rowcol	*/
    struct aceNet *next;			/* next net at rowcol	*/
    /* internal ace fields						*/
    int cellCount;				/* number of cells	*/
    aceCell **cells;				/* array of ptrs to cells */
    int dependCount;				/* how many user ucf's	*/
    int *ucfIndices;				/* which ucf's		*/
    aceRange save0, save1;			/* for fast move undoing */
    /*** equivalent net info */
} aceNet;

typedef struct acePlane {
    /* fields included in aceApPlane					*/
    aceObject *objNext;				/* link in obj list	*/
    aceObjType objType;				/* type of object	*/
    aceId id;					/* unique id number	*/
    aceRange range;				/* RCs used in the plane */
    short isRow;				/* plane orientation	*/
    /* internal ace fields						*/
    struct acePlane *next;			/* linked list		*/
    int emptyC;					/* count of empty RCs	*/
    int deadC;					/* count of dead RCs	*/
    int netExitC;				/* count of exiting nets */
    int netCap;					/* max net capacity	*/
    int cellCap;				/* max cell capacity	*/
    int saveUpper;				/* save old range.upper	*/
} acePlane;

/*
 * default values for netCap and the cellCaps
 */
#define ACE_DEFAULT_NET_CAPACITY	1
#define ACE_DEFAULT_CELL_CAPACITY	1

typedef struct aceRowCol {
    /* fields included in aceApRowCol					*/
    aceObject *objNext;				/* link in obj list	*/
    aceObjType objType;				/* type of object	*/
    aceId id;					/* unique id number	*/
    int rowcol;					/* index of this rowcol	*/
    short isRow;				/* bool, orientation	*/
    int count;					/* number of nets	*/
    /* internal ace fields						*/
    int termNetC;				/* number of term nets	*/
    short used;					/* bool, is rowcol active */
    aceNet *netlist;				/* list of nets contained */
    acePlane *plane;				/* the plane contain this RC */
} aceRowCol;

typedef struct aceMapEl {			/* array map element	*/
    aceCoord netCap;				/* net capacity used	*/
    aceCoord cellCap;				/* cell capacity used	*/
} aceMapEl;

typedef struct aceMatrix {
    aceObject *objNext;				/* link in obj list	*/
    aceObjType objType;				/* type of object	*/
    aceId id;					/* unique id number	*/
    unsigned status;				/* status info		*/
    unsigned mode;				/* mode info		*/
    aceCoord size;				/* number of rows/cols	*/
    aceRowCol *rowRCs, *colRCs;			/* edge pointer arrays	*/
    acePlane *rowPlanes, *colPlanes;		/* planes		*/
    int termNetC;				/* term net count	*/
    aceNet **termNets;				/* ptrs to term nets	*/
    aceMapEl **map;				/* topographical map	*/
} aceMatrix;

/* matrix status (MS) bits						*/
#define ACE_MS_STARTED	(1<<0)			/* if not set, first ace call */
#define ACE_MS_CREATED	(1<<1)			/* matrix has been created */
#define ACE_MS_UCF	(1<<2)			/* matrix has UCFs	*/

/* matrix mode (MM) bits						*/
#define ACE_MM_AUTO_RESIZE	(1<<0)		/* automatic enlarge	*/

typedef struct aceMove {
    int index;					/* what type of move	*/
    aceNet *net0, *net1;			/* which nets		*/
    int displacement;				/* how far		*/
    acePlane *plane;				/* which plane (if needed) */
    short isRow;				/* bool, set if row net	*/
    int mustAccept;				/* set if move must be acc */
} aceMove;

typedef struct aceComCount {			/* count of cost components */
    aceCoord usedRC;				/* used rowcol count	*/
    aceCoord usedBias;				/* used rc bais count	*/
    aceCoord netLen;				/* net length count	*/
    aceCoord netCap;				/* net overlap count	*/
    aceCoord cellCap;				/* cell overlap count	*/
} aceComCount;

typedef enum aceErrCode {
    ACE_NO_ERR,
    ACE_BAD_ID,
    ACE_BAD_TYPE,
    ACE_BAD_ARG,
    ACE_NO_MEM,
    ACE_OUT_OF_BOUNDS,
    ACE_GENERAL_ERR
} aceErrCode;

typedef struct aceGenState {
    aceObject *obj;
    int index1;
    int index2;
} aceGenState;

typedef struct aceStateStats {
    int numGen;
    int numAcc;
    int numRej;
    double startCost;
    long cpuTime;
} aceStateStats;

typedef struct aceAnnealStats {
    int numStates;
    aceStateStats thisState;
    aceStateStats allStates;
} aceAnnealStats;

typedef struct aceOneMoveStats {
    int accCnt, rejCnt;
    double accDC, rejDC;
} aceOneMoveStats;

typedef struct aceMoveSetStats {
    aceOneMoveStats move[ ACE_MOVE_COUNT ];
} aceMoveSetStats;

typedef struct aceMoveStats {
    aceMoveSetStats thisState;
    aceMoveSetStats allStates;
} aceMoveStats;

/* the global structure used for globals inside ace			*/
typedef struct aceGlobals {
    aceMatrix matrix;
    aceMove move;
    double temp;
    double cost;
    double delC;
    double range;
    aceComCount comc;
    aceComCount comdc;
    aceAnnealParams aparams;
    aceMoveStats mstat;
    aceAnnealStats astat;
    int rcC;
    int planeC;
    int netC;
    int cellC;
    double ucfCost;
    double ucfNewCost;
    aceUCF *ucfArray;
    int ucfCount;
    aceUCFEvent *ucfEventQ;
    aceUCFEvent *ucfCalledQ;
    unsigned traceMask;
    FILE *traceFp;
    char *errName;
    int errDepth;
#ifdef ACE_PARANOID
    st_table *idTbl;
#else
    aceObject *objList;
#endif /* ACE_PARANOID */
} aceGlobals;

aceGlobals AceG;

/*  the internal routines in ACE					*/
/* addObject.c								*/
extern aceId		ace_add_cell();
extern aceId		ace_add_master();
extern aceId		ace_create_matrix();
extern void		ace_insert_cell();
extern char		*ace_object_type_string();

/* allocate.c								*/
extern aceCell		*ace_new_cell();
extern aceCell		**ace_new_cell_ptrs();
extern aceMapEl		**ace_new_map();
extern aceMaster	*ace_new_master();
extern aceNet		*ace_new_net();
extern aceObject	*ace_new_object();
extern acePlane		*ace_new_plane();
extern aceRowCol	*ace_new_rowcols();
extern int		*ace_new_ucf_indices();

/* anneal.c								*/
extern void		ace_anneal();

/* cleanUp.c								*/
extern void		ace_clean_up();

/* cost.c								*/
extern void		ace_build_map();
extern double		ace_cost();
extern void		ace_init_a_params();
extern void		ace_prep_matrix();
extern void		ace_shrink_nets();

/* debug.c								*/
extern void		ace_check_netcounts();
extern void		ace_check_nets();
extern void		ace_dump_net_cap_map();
extern void		ace_verify_matrix();
extern void		ace_write_matrix();

/* defaults.c								*/
extern void		ace_get_defaults();
extern aceDefaults	*ace_get_default_ptr();
extern void		ace_set_defaults();
extern void		ace_set_own_defaults();

/* deltaCost.c								*/
extern double		ace_delta_cost();

/* dispNet.c								*/
extern double		ace_cost_of_count();
extern double		ace_disp_net_move();
extern void		ace_displace_net();
extern void		ace_replace_net();
extern double		ace_swap_net_move();

/* error.c								*/
extern void		ace_error_handler();

/* flipTerm.c								*/
extern double		ace_flip_term_net();
extern void		ace_unflip_term_net();

/* genMove.c								*/
extern void		ace_gen_move();
extern void		ace_init_gen_move();
extern int		ace_move_asm_disp();
extern int		ace_move_asm_flip_term();
extern int		ace_move_asm_rm_overlap();
extern int		ace_move_asm_swap();
extern void		ace_move_first_rci();
extern double		ace_move_prob_disp();
extern double		ace_move_prob_flip_term();
extern double		ace_move_prob_rm_overlap();
extern double		ace_move_prob_swap();
extern int		ace_random();
extern void		ace_update_gen_move();

/* generate.c								*/
extern void		ace_free_gen_status();
extern int		ace_gen_matrix_planes();
extern int		ace_gen_net_cells();
extern int		ace_gen_rowcol_nets();
extern int		ace_gen_plane_rowcols();

/* id.c									*/
extern aceId		ace_add_object();
extern void		ace_free_an_object();
extern void		ace_free_objects();
extern aceObject	*ace_get_by_id();

/* move.c								*/
extern void		ace_accept_move();
extern void		ace_reject_move();

/* output.c								*/

/* rmOverlap.c								*/
extern int		ace_col_nc_in_row();
extern int		ace_is_overlap();
extern void		ace_mv_map_col();
extern void		ace_mv_map_row();
extern void		ace_rm_overlap();
extern double		ace_rm_overlap_move();
extern void		ace_rmo_free();
extern void		ace_rmo_pack_matrix();
extern void		ace_rmo_pack_plane();
extern int		ace_row_nc_in_col();
extern void		ace_set_rowcols();
extern void		ace_zero_map_col();
extern void		ace_zero_map_row();

/* schedule.c								*/
extern void		ace_setup_scheduler();

/* start.c								*/
extern void		ace_check_ap_structs();
extern void		ace_start();

/* trace.c								*/
#ifdef ACE_COMP_TRACE
extern void		ace_trace();
extern void		ace_init_trace_time();
#endif /* ACE_COMP_TRACE */

/* ucf.c								*/
extern int		ace_add_ucf();
extern void		ace_ucf_check_cnstr();
extern void		ace_make_depend();
extern void		ace_ucf_efreeq();
extern aceUCFEvent	*ace_ucf_egetq();
extern void		ace_ucf_eputq();
extern void		ace_ucf_free_array();
extern void		ace_ucf_free_e();
extern aceUCFEvent	*ace_ucf_new_e();

/* some macros for data structure access				*/
#define ACE_ROW_RC( h )  & (AceG.matrix.rowRCs[h])
#define ACE_COL_RC( v )  & (AceG.matrix.colRCs[v])

#define ACE_GET_RCNET( rc, i, net)					\
	for( net = rc->netlist; net != NIL(aceNet); net = net->next){	\
	    if(( i >= net->range.lower) && ( i <= net->range.upper)){	\
		break;							\
	    }								\
	}

/* macro for fast rounding of positive doubles				*/
#define ACE_RND( x ) ((int)( x + 0.5 ))

/*
 * marco for calling the trace function
 */
#define ACE_TRACE( mask )						\
	if( mask & AceG.traceMask ){					\
	    ace_trace( (unsigned) mask );				\
	}

/*
 * In the paranoid version of ace_get_object, if ace_get_by_id returns,
 * then the id must be OK.
 */
#ifdef ACE_PARANOID
#define ace_get_object( id)						\
	(ace_get_by_id( id))
#else
#define ace_get_object( id)						\
	((aceObject *) id)
#endif

#define ace_new_object( obj ) ((aceObject *)obj)

/*
 * error handler macros
 */
#define ACE_START( name )						\
	if( AceG.errDepth++ == 0){					\
	    AceG.errName = name;					\
	    errPushHandler( ace_error_handler);				\
	}								\
	if( !( AceG.matrix.status & ACE_MS_STARTED)){			\
	    ace_start();						\
	}

#define ACE_END()							\
	if( --AceG.errDepth == 0) errPopHandler();

/*
 * UCF macros
 */
#define ACE_UCF_EVENT( mask, iArray, count, i, e)			\
	for( i = count - 1; i >= 0; --i){				\
	    if( AceG.ucfArray[ iArray[ i ] ].eventMask & (mask)){	\
		e = ace_ucf_new_e();					\
		e->index = iArray[ i ];					\
		ace_ucf_eputq( e, &AceG.ucfEventQ);			\
	    }								\
	}
