/*
 * allocate.c
 *
 * Does most allocation of new data structure space.
 * Reallocs are done in other places.
 * All memory allocated is initaialized to zero.
 */

#include "aceInternal.h"

/*
 * ace_new_map
 *
 * Allocate a new cell/net map.  This function can be called multiple
 * times (in the case that the matrix needs to be resize).
 * The map is a pointer to an array of pointers to arrays of map elements.
 * It is referenced as map[ row ][ col ].
 */
aceMapEl **
ace_new_map( os, ns, oldmap)
aceCoord *os, *ns;				/* old size, new size	*/
aceMapEl **oldmap;
{
    aceMapEl **map;
    int i;

    /* make the horizontal (aceMap *) array				*/
    if( ns->row > os->row){
	map = REALLOC( aceMapEl *, AceG.matrix.map, ns->row);
	if( map == NIL(aceMapEl *)){
	    errRaise( ACE_PKG_NAME, ACE_NO_MEM,
			"no memory for reallocing the cell/net map");
	}
    } else {
	map = oldmap;
    }

    if( ns->col > os->col){
	for( i = ns->row - 1; i >= 0; --i){
	    if( i < os->row){
		map[i] = REALLOC( aceMapEl, map[i], ns->col);
	    } else {
		map[i] = ALLOC( aceMapEl, ns->col);
	    }
	}
    }

    return( map);
}

/*
 * ace_new_rowcols
 *
 * allocate an array of aceRowCols.
 */
aceRowCol *
ace_new_rowcols( size )
int size;
{
    aceRowCol *rc;

    rc = ALLOC( aceRowCol, size);
    if( rc == NIL(aceRowCol)){
	errRaise( ACE_PKG_NAME, ACE_NO_MEM,
		"no memory for new aceRowCol array");
    }
    (void)memset((char *)rc, 0, sizeof(aceRowCol)*size);
    return( rc);
}

/*
 * ace_new_cell_ptrs
 *
 * This routine gets a new array of pointers to cells. One such array is
 * included in each aceNet. Since it isn't known when the net is created
 * how many cells there will eventually be, it is expected that this
 * array may be realloced occasionally. The initial size of 5 is used
 * to reduce the realloc cost for small nets while not wasting too much memory.
 */
aceCell **
ace_new_cell_ptrs()
{
    aceCell **cellp;

    cellp = ALLOC( aceCell *, 5);
    if( cellp == NIL(aceCell *)){
	errRaise( ACE_PKG_NAME, ACE_NO_MEM,
		"no memory for new aceCell * array");
    }
    (void)memset( (char *)cellp, 0, sizeof( aceCell *) * 5);
    return( cellp);
}

/*
 * Allocate an initial array of ucf indices (ints).
 * As above, start with some (arbitrary) initial size and realloc
 * as necessay
 */
int *
ace_new_ucf_indices()
{
    int *ip;

    ip = ALLOC( int, 2);
    if( ip == NIL(int)){
	errRaise( ACE_PKG_NAME, ACE_NO_MEM,
		"no memory for new UCF index array");
    }
    (void)memset((char *)ip, 0, sizeof( int) * 2);
    return( ip);
}


/*
 * ace_new_plane
 *
 * allocate a plane struct.
 */
acePlane *
ace_new_plane()
{
    acePlane *m;

    m = ALLOC( acePlane, 1);
    if( m == NIL(acePlane)){
	errRaise( ACE_PKG_NAME, ACE_NO_MEM,
		"no memory for new acePlane");
    }
    (void)memset((char *)m, 0, sizeof(acePlane));
    return(m);
}

/*
 * ace_new_master
 *
 * allocate a master struct.
 */
aceMaster *
ace_new_master()
{
    aceMaster *m;

    m = ALLOC( aceMaster, 1);
    if( m == NIL(aceMaster)){
	errRaise( ACE_PKG_NAME, ACE_NO_MEM,
		"no memory for new aceMaster");
    }
    (void)memset((char *)m, 0, sizeof(aceMaster));
    return(m);
}

/*
 * ace_new_cell
 *
 * allocate a cell struct.
 */
aceCell *
ace_new_cell()
{
    aceCell *m;

    m = ALLOC( aceCell, 1);
    if( m == NIL(aceCell)){
	errRaise( ACE_PKG_NAME, ACE_NO_MEM,
		"no memory for new aceCell");
    }
    (void)memset((char *)m, 0, sizeof(aceCell));
    return(m);
}

/*
 * ace_new_net
 *
 * allocate a net struct.
 */
aceNet *
ace_new_net()
{
    aceNet *m;

    m = ALLOC( aceNet, 1);
    if( m == NIL(aceNet)){
	errRaise( ACE_PKG_NAME, ACE_NO_MEM,
		"no memory for new aceNet");
    }
    (void)memset((char *)m, 0, sizeof(aceNet));
    return(m);
}
