/*
 * anneal.c
 *
 * This file contains the annealing utility routines,
 * and most importantly, the annealing outer and inner loops.
 */

#include "aceInternal.h"

/* for checking that the delta cost is properly evaluated		*/
#ifdef ACE_PARANOID
#define COST_PARANOID
#endif

/* 
 * ace_anneal
 *
 * Run the simulated annealing engine.
 */
void
ace_anneal( fastFlag )
unsigned fastFlag;
{
    aceDefaults *def;
    int (* stop_inner)();
#ifdef ACE_COMP_TRACE
    int i;
#endif /* ACE_COMP_TRACE */

#ifdef COST_PARANOID
    double pcDiff;
#endif /* COST_PARANOID */

    def = ace_get_default_ptr();
    stop_inner = def->stopInner;

    /* initialize cost parameters					*/
    ace_init_a_params();

    /* Check and prepare the matrix data strucure			*/
    ace_prep_matrix();

    /* Construct the initial cell/net map				*/
    ace_build_map();

    /* initialize the and cost						*/
    AceG.cost = ace_cost();
    AceG.ucfCost = AceG.ucfNewCost;
    AceG.temp = def->initTemp( AceG.cost, fastFlag);
    ace_init_gen_move();

#ifdef ACE_COMP_TRACE
    /* initialize the state statistics struct				*/
    AceG.astat.allStates.startCost = AceG.cost;
    if( AceG.traceMask & ACE_TRACE_CPU_TIME){
	ace_init_trace_time();
    }
#endif /* ACE_COMP_TRACE */

    /* initialize the range limiter					*/
    if( AceG.matrix.size.row > AceG.matrix.size.col){
	AceG.range = AceG.matrix.size.row;
    } else {
	AceG.range = AceG.matrix.size.col;
    }

#ifdef ACE_COMP_TRACE
    /* trace initial state						*/
    ACE_TRACE( ACE_TRACEF_INITIAL);
#endif /* ACE_COMP_TRACE */

    if( def->initStopInner != ( void (*)()) 0){
	def->initStopInner( AceG.matrix.size.row, AceG.matrix.size.col,
		AceG.planeC, AceG.netC, AceG.cellC,
		AceG.temp, AceG.cost, fastFlag);
    }

    while( !( def->stopOuter( AceG.temp, AceG.cost, fastFlag))){
	/* prepare bookkeeping stuff for new temp state			*/
	AceG.astat.thisState.numGen = AceG.astat.thisState.numAcc
		= AceG.astat.thisState.numRej = 0;
#ifdef ACE_COMP_TRACE
	AceG.astat.thisState.startCost = AceG.cost;
	(void)memset((char *) &( AceG.mstat.thisState), 0,
			sizeof( aceMoveSetStats));
	++AceG.astat.numStates;
#endif /* ACE_COMP_TRACE */

	while( !( stop_inner( AceG.astat.thisState.numGen,
			      AceG.astat.thisState.numAcc, fastFlag))){
	    ace_gen_move();
	    ++AceG.astat.thisState.numGen;

	    AceG.delC = ace_delta_cost();

	    /* The mysterious constant here is (2 ** 31) - 1 (which is
	     * the max value of a long on a machine with 32 bit longs
	     * and a twos complement representation).
	     *** Should be the same as INT_MAX, yes?
	     */
	    if(( AceG.delC <= 0.0) ||
	         AceG.move.mustAccept ||
		 (((double) random())/(double)2147483647 <
		   exp( -AceG.delC / AceG.temp))){

		/* accept the move					*/
		++AceG.astat.thisState.numAcc;
#ifdef ACE_COMP_TRACE
		++AceG.mstat.thisState.move[ AceG.move.index ].accCnt;
		AceG.mstat.thisState.move[ AceG.move.index ].accDC += AceG.delC;
#endif /* ACE_COMP_TRACE */
		ace_accept_move();
		AceG.cost += AceG.delC;

#ifdef ACE_COMP_TRACE
		/* trace the move					*/
		ACE_TRACE( ACE_TRACEF_ACC_MOVE);
#endif /* ACE_COMP_TRACE */
	    } else {
		/* reject the move					*/
		++AceG.astat.thisState.numRej;
#ifdef ACE_COMP_TRACE
		++AceG.mstat.thisState.move[ AceG.move.index ].rejCnt;
		AceG.mstat.thisState.move[ AceG.move.index ].rejDC += AceG.delC;
#endif /* ACE_COMP_TRACE */
		ace_reject_move();
#ifdef COST_PARANOID
		if(( pcDiff = AceG.cost - ace_cost()) > 0.01){
		    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
			"move not properly rejected: %.4f", pcDiff);
		}
#endif /* COST_PARANOID */

#ifdef ACE_COMP_TRACE
		/* trace the move					*/
		ACE_TRACE( ACE_TRACEF_REJ_MOVE);
#endif /* ACE_COMP_TRACE */
	    }
	}

#ifdef ACE_COMP_TRACE
	/* sum up statistics						*/
	AceG.astat.allStates.numGen += AceG.astat.thisState.numGen;
	AceG.astat.allStates.numAcc += AceG.astat.thisState.numAcc;
	AceG.astat.allStates.numRej += AceG.astat.thisState.numRej;

	for( i = ACE_MOVE_COUNT - 1; i >= 0; --i){
	    AceG.mstat.allStates.move[ i ].accCnt +=
		AceG.mstat.thisState.move[ i ].accCnt;
	    AceG.mstat.allStates.move[ i ].rejCnt +=
		AceG.mstat.thisState.move[ i ].rejCnt;
	    AceG.mstat.allStates.move[ i ].accDC +=
		AceG.mstat.thisState.move[ i ].accDC;
	    AceG.mstat.allStates.move[ i ].rejDC +=
		AceG.mstat.thisState.move[ i ].rejDC;
	}

	/* do tracing							*/
	ACE_TRACE( ACE_TRACEF_STATE );
#endif /* ACE_COMP_TRACE */

	/* update temp for next state					*/
	AceG.temp = def->updateTemp( AceG.temp, AceG.cost, fastFlag);
	ace_update_gen_move();
	if( AceG.range < 2.0){
	    AceG.range = 1.0;
	} else {
	    AceG.range *= 0.8;
	}
    }

#ifdef ACE_COMP_TRACE
    ACE_TRACE( ACE_TRACEF_FINAL );
#endif /* ACE_COMP_TRACE */
}
