/*
 * apAddPlanes.c
 *
 * Application interface routines to set up the plane structure of the matrix.
 */

#include "aceInternal.h"


/*
 * aceCreateMatrix
 *
 * Create the full size matrix and return the id of the matrix.
 * xsize and ysize are the number of colums and rows, respectively,
 * and correspond to up/down and left/right dimensions.
 * Can only be called once per annealing process.
 */
aceId
aceCreateMatrix( rows, cols)
int rows, cols;			/* number of rows and columns		*/
{
    aceId id, ace_create_matrix();

    ACE_START( "aceCreateMatrix");

    if( AceG.matrix.status & ACE_MS_CREATED){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR, "matrix already created");
    }
    if( rows <= 0){
	errRaise( ACE_PKG_NAME, ACE_OUT_OF_BOUNDS,
		"`rows' non-positive");
    }
    if( cols <= 0){
	errRaise( ACE_PKG_NAME, ACE_OUT_OF_BOUNDS,
		"`cols' non-positive");
    }

#ifdef ACE_PARANOID
    /* setup the hash table						*/
    AceG.idTbl = st_init_table( st_numcmp, st_numhash);
#endif

    id = ace_create_matrix( rows, cols);

    ACE_END();

    return( id);
}

/*
 * aceSplitPlane
 *
 * Split the plane into two planes at the specified break point
 * within the plane.  plane[breakpt] becomes new_plane[0].
 * New plane inherits cellCap and netCap of splitted plane.
 */
/*ARGSUSED*/
void
aceSplitPlane( id, breakpt)
aceId id;			/* id of plane				*/
int breakpt;			/* break point (in matrix coords)	*/
{
    acePlane *plane, *np;
    aceObject *oPlane;
    int i;

    ACE_START( "aceSplitPlane");

    /* check arguments							*/
    if(( oPlane = ace_get_object( id )) == NIL(aceObject)){
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"`id' is corrupt");
    } else if( oPlane->type != ACE_OBJ_PLANE){
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"id refers to object of type %s, not type %s",
		ace_object_type_string( oPlane->type),
		ace_object_type_string( ACE_OBJ_PLANE));
    } else {
	plane = (acePlane *)( oPlane);
	if(( breakpt < plane->range.lower) || ( breakpt > plane->range.upper)){
	    errRaise( ACE_PKG_NAME, ACE_OUT_OF_BOUNDS,
		"`breakpt' is %d, must be in range [%d, %d]", breakpt,
		plane->range.lower, plane->range.upper);
	}
    }

    /* get a new plane							*/
    np = ace_new_plane();

    /* insert it as an object						*/
    oPlane = ace_new_object( np );
    oPlane->type = ACE_OBJ_PLANE;
    np->id = ace_add_object( oPlane);

    /* update fields of new and old planes				*/
    np->cellCap = plane->cellCap;
    np->netCap = plane->netCap;
    np->isRow = plane->isRow;
    np->range.upper = plane->range.upper;
    np->range.lower = breakpt;
    plane->range.upper = breakpt - 1;

    /* update rowcol fields						*/
    if( plane->isRow){
	/* its a row plane						*/
	for( i = np->range.lower; i <= np->range.upper; ++i){
	    AceG.matrix.rowRCs[ i ].plane = np;
	}
    } else {
	for( i = np->range.lower; i <= np->range.upper; ++i){
	    AceG.matrix.colRCs[ i ].plane = np;
	}
    }

    /* link new plane into plane list					*/
    np->next = plane->next;
    plane->next = np;

    /* update global statistics						*/
    ++AceG.planeC;

    ACE_END();
}

/*
 * aceSetNetCapacity
 *
 * Set the capacity for each direction.
 * This is done on a per-plane basis.
 */
void
aceSetNetCapacity( id, capacity)
aceId id;			/* id of plane				*/
int capacity;			/* row and column net capacity		*/
{
    acePlane *plane;

    ACE_START( "aceSetNetCapacity");

    /* simple sanity check						*/
    if( capacity < 0){
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"`capacity' argument was %d, but it must be non-negative",
		capacity);
    }

    /* update the plane							*/
    plane = (acePlane *) ace_get_object( id);
    plane->netCap = capacity;

    ACE_END();
}

/*
 * aceSetCellCapacity
 *
 * Set the capacity for each direction.
 * This is done on a per-plane basis.
 */
void
aceSetCellCapacity( id, capacity)
aceId id;			/* id of plane				*/
int capacity;			/* row and column cell capacity		*/
{
    acePlane *plane;

    ACE_START( "aceSetCellCapacity");

    /* simple sanity check						*/
    if( capacity < 0){
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"`capacity' argument was %d, but it must be non-negative",
		capacity);
    }

    /* update the matrix						*/
    plane = (acePlane *) ace_get_object( id);
    if( plane->objType != ACE_OBJ_PLANE){
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"id refers to object of type %s, not type %s",
		ace_object_type_string( plane->objType),
		ace_object_type_string( ACE_OBJ_PLANE));
    }

    plane->cellCap = capacity;

    ACE_END();
}
