/*
 * apGenerate.c
 *
 * Routines to be used to generate the contents of ace objects.
 * So far, only generating the nets of a plane and the cells of a net
 * are planned.
 */

#include "aceInternal.h"

/*
 * aceInitGenerator
 *
 * Initialize a generator structure.
 */
void
aceInitGenContents( id, gen)
aceId id;		/* id of container object */
aceGenerator *gen;	/* ptr to application's aceGenerator struct */
{
    aceGenState *state;

    ACE_START( "aceInitGenContents");

    state = ALLOC( aceGenState, 1);
    if( state == NIL(aceGenState)){
	errRaise( ACE_PKG_NAME, ACE_NO_MEM,
		"no memory for generator state structure");
    }

    state->obj = ace_get_object( id );
    state->index1 = state->index2 = -1;
    gen->state = (char *)state;

    switch( state->obj->type ){
	case ACE_OBJ_MATRIX:
	    gen->gen_fn = ace_gen_matrix_planes;
	    gen->free_fn = ace_free_gen_status;
	    break;
	case ACE_OBJ_PLANE:
	    gen->gen_fn = ace_gen_plane_rowcols;
	    gen->free_fn = ace_free_gen_status;
	    break;
	case ACE_OBJ_ROWCOL:
	    gen->gen_fn = ace_gen_rowcol_nets;
	    gen->free_fn = ace_free_gen_status;
	    break;
	case ACE_OBJ_NET:
	    gen->gen_fn = ace_gen_net_cells;
	    gen->free_fn = ace_free_gen_status;
	    break;
	default:
	    errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"cannot generate contents of an ace %s",
		ace_object_type_string( state->obj->type ));
	    break;
    }

    ACE_END();
}

/*
 * aceGenFirstContainer
 *
 * Generate the object(s) containing the object refered to by id.
 * Store the result(s) in *rowId and/or *colId.
 * Only the immediate container is generated.
 */
void
aceGenFirstContainer( id, rowObj, colObj )
aceId id;			/* id of contained object */
aceObject **rowObj, **colObj;	/* places to store the generation results */
{
    aceObject *obj, *contObj;
    short prefRow;

    ACE_START( "aceGenFirstContainer");

    /* check args							*/
    if( id == ace_null_id){
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG, "id is null");
    } else if(( obj = ace_get_object( id )) == NIL(aceObject)){
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG, "id is corrupt");
    }

    if( obj->type == ACE_OBJ_CELL){
	if(( rowObj == NIL(aceObject *)) || ( colObj == NIL(aceObject *))){
	    errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
	   "either rowObj or colObj is NIL. Need both non-NIL for cell object");
	}
    } else {
	if(( rowObj == NIL(aceObject *)) && ( colObj == NIL(aceObject *))){
	    errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"both rowObj or colObj are NIL");
	}
    }

    /* generate the container						*/
    if( obj->type == ACE_OBJ_CELL){
	*rowObj = (aceObject *)(((aceCell *)obj)->rowNet);
	*colObj = (aceObject *)(((aceCell *)obj)->colNet);
    } else {
	switch( obj->type ){
	    case ACE_OBJ_NET:
		if(((aceNet *)obj)->isRow){
		    prefRow = TRUE;
		    contObj = (aceObject *)
			    &(AceG.matrix.rowRCs[ ((aceNet *)obj)->rowcol ]);
		} else {
		    prefRow = FALSE;
		    contObj = (aceObject *)
			    &(AceG.matrix.colRCs[ ((aceNet *)obj)->rowcol ]);
		}
		break;
	    case ACE_OBJ_ROWCOL:
		if(((aceRowCol *)obj)->isRow){
		    prefRow = TRUE;
		} else {
		    prefRow = FALSE;
		}
		contObj = (aceObject *) (((aceRowCol *)obj)->plane);
		break;
	    case ACE_OBJ_PLANE:
		prefRow = TRUE;
		contObj = (aceObject *) &(AceG.matrix);
		break;
	    default:
		errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		    "can't generate the container of a %s type object",
		    ace_object_type_string( obj->type ));
		break;
	}

	/* store the value in the appropriate place			*/
	if( prefRow ){
	    if( rowObj != NIL(aceObject *)){
		*rowObj = contObj;
		if( colObj != NIL(aceObject *)){
		    (*colObj)->id = ace_null_id;
		}
	    } else {
		*colObj = contObj;
	    }
	} else {
	    if( colObj != NIL(aceObject *)){
		*colObj = contObj;
		if( rowObj != NIL(aceObject *)){
		    (*rowObj)->id = ace_null_id;
		}
	    } else {
		*rowObj = contObj;
	    }
	}
    }

    ACE_END();
}

/*
 * aceFreeGenerator
 *
 * Routine to free the generator and the state of a generator
 * before all elements have been generated.
 */
void
aceFreeGenerator( gen )
aceGenerator *gen;	/* free state memory of this generator */
{
    ACE_START( "aceFreeGenerator");
    if( gen != NIL(aceGenerator)){
	gen->free_fn( gen );
    } else {
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG, "gen is NIL");
    }
    ACE_END();
}
