/* 
 * apId.c
 *
 * Application aceId interaction (eg, get by ids, get ids, etc).
 */

#include "aceInternal.h"

/*
 * aceGetCellNets
 *
 * Get the horizontal and vertical net ids associated with a cell.
 * If hId or vId is nil, then it won't be returned.
 */
/*ARGSUSED*/
void
aceGetCellNets( cellId, rowId, colId)
aceId cellId;			/* id of cell				*/
aceId *rowId, *colId;		/* places to return net ids		*/
{
    aceCell *cell;
    aceObject *oCell;
    char *errMsg = NIL(char);

    ACE_START( "aceGetCellNets")

    /* check args							*/
    if(( oCell = ace_get_object( cellId)) == NIL(aceObject)){
	errMsg = "cellId is a corrupt id";
    } else if( oCell->type != ACE_OBJ_CELL){
	errMsg = "cellId isn't the id of a cell";
    } else if(( rowId == NIL(aceId)) && ( colId == NIL(aceId))){
	errMsg = "both rowId and colId are NIL";
    }
    if( errMsg != NIL(char)){
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG, errMsg);
    }

    cell = (aceCell *)( oCell);
    if( rowId != NIL(aceId)){
	*rowId = cell->rowNet->id;
    }
    if( colId != NIL(aceId)){
	*colId = cell->colNet->id;
    }

    ACE_END();
}

/*
 * aceGetNetsAtLoc
 *
 * Get the horizontal and vertical net ids at the given location.
 * If hId or vId is nil, then it won't be returned.
 * If there is more that one net of the same orientation at the location,
 * it is an error because of ambiguity.
 */
/*ARGSUSED*/
void
aceGetNetsAtLoc( row, col, rowId, colId)
int row, col;			/* location in the matrix		*/
aceId *rowId, *colId;		/* places to return net ids		*/
{
    aceNet *net, *theNet;

    ACE_START( "aceGetNetsAtLoc")

    /* check args							*/
    if(( row >= AceG.matrix.size.row ) || ( col >= AceG.matrix.size.col) ||
       ( row < 0) || ( col < 0 )){
	errRaise( ACE_PKG_NAME, ACE_OUT_OF_BOUNDS,
		"location [%d, %d] is out of bounds", row, col);
    }
    if(( rowId == NIL(aceId)) && ( colId == NIL(aceId))){
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG, "both rowId and colId are NIL");
    }

    /* get the row net							*/
    net = AceG.matrix.rowRCs[ row ].netlist;
    if( net == NIL(aceNet)){
	if( rowId != NIL(aceId)){
	    *rowId = ace_null_id;
	}
    } else {
	theNet = NIL(aceNet);
	for(; net != NIL(aceNet); net = net->next){
	    if(( net->range.lower <= col) && ( net->range.upper >= col)){
		if( theNet == NIL(aceNet)){
		    theNet = net;
		} else if( rowId != NIL(aceId)){
		    /* more than one net at the location		*/
		    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
			"more than one row net at location [ %d, %d ]",
			row, col);
		}
	    }
	}
	if( rowId != NIL(aceId)){
	    *rowId = theNet->id;
	}
    }

    /* get the col net							*/
    net = AceG.matrix.colRCs[ col ].netlist;
    if( net == NIL(aceNet)){
	if( colId != NIL(aceId)){
	    *colId = ace_null_id;
	}
    } else {
	theNet = NIL(aceNet);
	for(; net != NIL(aceNet); net = net->next){
	    if(( net->range.lower <= row) && ( net->range.upper >= row)){
		if( theNet == NIL(aceNet)){
		    theNet = net;
		} else if( colId != NIL(aceId)){
		    /* more than one net at the location		*/
		    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
			"more than one col net at location [ %d, %d ]",
			row, col);
		}
	    }
	}
	if( colId != NIL(aceId)){
	    *colId = theNet->id;
	}
    }

    ACE_END();
}

/*
 * aceGetPlanesAtLoc
 *
 * Get the horizontal and vertical plane ids associated with a location.
 * If hId or vId is nil, then it won't be returned.
 */
void
aceGetPlanesAtLoc( row, col, rowId, colId)
int row, col;			/* location in full matrix coords	*/
aceId *rowId, *colId;		/* places to return plane ids		*/
{
    ACE_START( "aceGetPlaneIds")

    if(( row >= AceG.matrix.size.row ) || ( col >= AceG.matrix.size.col) ||
       ( row < 0) || ( col < 0 )){
	errRaise( ACE_PKG_NAME, ACE_OUT_OF_BOUNDS,
		"location [%d, %d] is out of bounds", row, col);
    }

    *rowId = AceG.matrix.rowRCs[ row ].plane->id;
    *colId = AceG.matrix.colRCs[ col ].plane->id;

    ACE_END();
}

/*
 * aceGetById
 *
 * Get the thing with the given Id.  It is expected that the user knows
 * what sort of object this is (plane, net, cell), because it is expected
 * that this will be used in a UCF, and when the UCF call is set up,
 * the argument types will always be in certain order (as the arguments
 * a normal procedure call are) and will be interpretted in the same
 * order in the function.
 * The address of the object is cast to a (char *) and returned.
 */
aceObject *
aceGetById( id )
aceId id;		/* id of any object */
{
#ifdef ACE_PARANOID
    aceObject *obj;

    ACE_START( "aceGetById")

    obj = ace_get_by_id( id );

    ACE_END();

    return( obj );
#else
    /* the low overhead, fast way					*/
    return( (aceObject *) id);
#endif
}
