/*
 * cost.c
 *
 * Contains the cost function for determining the cost of the current state
 * of the matrix in the simulated annealing process.
 */

#include "aceInternal.h"

/*** debugging */
static aceComCount lcc;
static double lucfCost;

/*
 * ace_cost
 *
 * This function deterimines the total cost of the current configuration
 * of the matrix.
 *
 * Components of Total Cost:
 * (1) non-empty rows and columns.
 * (2) empty rowcols at the top and right of plane.
 * (3) the `true' length of each net [ disreguarding empty RCs ].
 * (4) net overlap.
 * (5) bias of used rowcols in each plane towards lower side.
 * (6) return value of each user constraint function.
 */
double
ace_cost()
{
    double cost = 0.0;
    aceNet *net;
    acePlane *plane;
    int row, col;		/* index var				*/
    int diff;			/* net overlap count			*/
    aceMapEl **map;		/* pointer to the map			*/

    /*** debugging */
    lcc = AceG.comc;
    lucfCost = AceG.ucfCost;

    /* initialize cost counting varaibles				*/
    AceG.comc.usedRC.row = AceG.comc.usedRC.col = 0;
    AceG.comc.netLen.row = AceG.comc.netLen.col = 0;
    AceG.comc.usedBias.row = AceG.comc.usedBias.col = 0;
    AceG.comc.netCap.row = AceG.comc.netCap.col = 0;
    AceG.comc.cellCap.row = AceG.comc.cellCap.col = 0;

    /* non-empty horizontal rowcols and net length			*/
    for( row = AceG.matrix.size.row - 1; row >= 0; --row){
	if( AceG.matrix.rowRCs[ row ].count != 0){
	    ++AceG.comc.usedRC.row;
	    for( net = AceG.matrix.rowRCs[row].netlist; net != NIL(aceNet);
		    net = net->next){
		AceG.comc.netLen.row += net->range.upper - net->range.lower + 1;
	    }
	}
    }
    /* bias towards lower side of plane					*/
    for( plane = AceG.matrix.rowPlanes; plane != NIL(acePlane);
		plane = plane->next){
	AceG.comc.usedBias.row += plane->range.upper - plane->range.lower + 1;
	/* subtract out all net length [ that will be added in ] that actually
	 * crosses empty space.
	 */
	AceG.comc.netLen.col -=(plane->emptyC + plane->deadC) * plane->netExitC;
    }

    /* non-empty vertical rowcol and net length				*/
    for( col = AceG.matrix.size.col - 1; col >= 0; --col){
	if( AceG.matrix.colRCs[col].count != 0){
	    ++AceG.comc.usedRC.col;
	    for( net = AceG.matrix.colRCs[col].netlist; net != NIL(aceNet);
		    net = net->next){
		AceG.comc.netLen.col += net->range.upper - net->range.lower + 1;
	    }
	}
    }
    /* bias towards lower side of plane					*/
    for( plane = AceG.matrix.colPlanes; plane != NIL(acePlane);
		plane = plane->next){
	AceG.comc.usedBias.col += plane->range.upper - plane->range.lower + 1;
	AceG.comc.netLen.row -=(plane->emptyC + plane->deadC) * plane->netExitC;
    }

    cost +=   AceG.comc.usedRC.row * AceG.aparams.usedRC.row 
	    + AceG.comc.usedRC.col * AceG.aparams.usedRC.col 
	    + AceG.comc.netLen.row * AceG.aparams.netLength.row
	    + AceG.comc.netLen.col * AceG.aparams.netLength.col
	    + AceG.comc.usedBias.row * AceG.aparams.usedBias.row
	    + AceG.comc.usedBias.col * AceG.aparams.usedBias.col;

    /* Net overlaps. The algorithm is based on the cell/net map and
     * the given capacity.  If a capacity is exceded, a penalty is assesed.
     */
    map = AceG.matrix.map;
    for( row = AceG.matrix.size.row - 1; row >= 0; --row){
	if( !AceG.matrix.rowRCs[ row ].used) continue;
	for( col = AceG.matrix.size.col - 1; col >= 0; --col){
	    if(  !AceG.matrix.colRCs[ col ].used) continue;
	    /* get the row plane at this row				*/
	    plane = AceG.matrix.rowRCs[ row ].plane;
	    if(( diff = map[ row ][ col ].netCap.row - plane->netCap) > 0){
		AceG.comc.netCap.row += diff;
	    }
	    if(( diff = map[ row ][ col ].cellCap.row - plane->cellCap) > 0){
		AceG.comc.cellCap.row += diff;
	    }

	    /* get the column plane at this col				*/
	    plane = AceG.matrix.colRCs[ col ].plane;
	    if(( diff = map[ row] [ col ].netCap.col - plane->netCap) > 0){
		AceG.comc.netCap.col += diff;
	    }
	    if(( diff = map[ row ][ col ].cellCap.col - plane->cellCap) > 0){
		AceG.comc.cellCap.col += diff;
	    }
	}
    }
    cost += AceG.comc.netCap.row * AceG.aparams.netCapacity.row +
	    AceG.comc.netCap.col * AceG.aparams.netCapacity.col +
	    AceG.comc.cellCap.row * AceG.aparams.cellCapacity.row +
	    AceG.comc.cellCap.col * AceG.aparams.cellCapacity.col;

    /* user constraint functions					*/
#ifndef ACE_NO_UCF
    AceG.ucfNewCost = -cost;
    for( row = AceG.ucfCount - 1; row >= 0; --row){
	AceG.ucfArray[ row ].lastRet =
		AceG.ucfArray[ row ].cnstr_fn( AceG.ucfArray[ row ].argCount,
				        AceG.ucfArray[ row ].argObjs);
	cost += AceG.ucfArray[ row ].lastRet;
    }
    AceG.ucfNewCost += cost;
#endif /* ACE_NO_UCF */

    return( cost );
}

/*
 * ace_build_map
 *
 * Build the Cell/Net map based on the current matrix.  This
 * will be called once before annealing starts, and any subsequent time
 * a matrix resize is necessary.  This routine assumes that the matrix
 * size will only be increase, never decreased. It also assumes that each
 * cell is part of a single net in each direction, so during the cell linking
 * phase, only one set of nets (rows or cols) need be traversed in order
 * to see every cell.
 */
void
ace_build_map()
{
    aceMapEl **map;
    static aceCoord allocSize;
    static int firstCall = TRUE;
    int row, col, c;
    aceNet *net;
    aceCell *cell;

    if( firstCall){
	firstCall = FALSE;
	allocSize.row = allocSize.col = 0;
    }

    /* alloc or realloc the map						*/
    map = ace_new_map( &allocSize, &( AceG.matrix.size), AceG.matrix.map);
    allocSize.row = MAX( allocSize.row, AceG.matrix.size.row);
    allocSize.col = MAX( allocSize.col, AceG.matrix.size.col);

    /* zero out any old data						*/
    for( row = AceG.matrix.size.row - 1; row >= 0; --row){
	(void)memset((char *)map[ row ], 0,
			sizeof( aceMapEl) * AceG.matrix.size.col);
    }

    /* look at each horizontal net					*/
    for( row = AceG.matrix.size.row - 1; row >= 0; --row){
	for( net = AceG.matrix.rowRCs[ row ].netlist; net != NIL(aceNet);
		net = net->next){

	    /* increment the net count at each location it crosses	*/
	    for( col = net->range.lower; col <= net->range.upper; ++col){
		++map[ row ][ col ].netCap.row;
	    }

	    /* add cell capacity for each cell (in both directions)	*/
	    for( c = net->cellCount - 1; c >= 0; --c){
		cell = net->cells[c];
		map[ row ][ cell->loc.col ].cellCap.row +=
					cell->master->cellCap.row;
		map[ row ][ cell->loc.col ].cellCap.col +=
					cell->master->cellCap.col;
	    }
	}
    }

    /* traverse col nets, but don't look at cells			*/
    for( col = AceG.matrix.size.col - 1; col >= 0; --col){
	for( net = AceG.matrix.colRCs[ col ].netlist; net != NIL(aceNet);
		net = net->next){

	    /* increment the net count at each location it crosses	*/
	    for( row = net->range.lower; row <= net->range.upper; ++row){
		++map[ row ][ col ].netCap.col;
	    }
	}
    }

    AceG.matrix.map = map;
}

/*
 * ace_prep_matrix
 *
 * Do one-time-only initialization on the users matrix. This involves:
 *	1) shrinking all net ranges from the user-specified size to the
 *	   minimum range that inculdes all cells in the net.
 *	2) calculating the number of nets exiting the upper edge of
 *	   all planes.
 */
void
ace_prep_matrix()
{
    aceNet *net;
    acePlane *plane;
    int index;

    ace_shrink_nets( TRUE, NIL(int));
    ace_shrink_nets( FALSE, NIL(int));

    /* for each row [ horizontal ] net, look at each col planes [ that contain
     * vertical nets ] and see if it crosses range.upper.
     */
    for( index = AceG.matrix.size.row - 1; index >= 0; --index){
	for( net = AceG.matrix.rowRCs[ index ].netlist; net != NIL(aceNet);
		net = net->next){
	    for( plane = AceG.matrix.colPlanes; plane != NIL(acePlane);
			plane = plane->next){
		if(( net->range.lower <= plane->range.upper) &&
		   (( net->range.upper > plane->range.upper) ||
		    ( net->termMask && !(net->termMask & ACE_TERM_LOWER_VE)))){
		    ++plane->netExitC;
		}
	    }
	}
    }

    /* for each col [ vertical ] net, look at each row planes [ that contain
     * horizontal nets ] and see if it crosses range.upper.
     */
    for( index = AceG.matrix.size.col - 1; index >= 0; --index){
	for( net = AceG.matrix.colRCs[ index ].netlist; net != NIL(aceNet);
		net = net->next){
	    for( plane = AceG.matrix.rowPlanes; plane != NIL(acePlane);
			plane = plane->next){
		if(( net->range.lower <= plane->range.upper) &&
		   (( net->range.upper > plane->range.upper) ||
		    ( net->termMask && !(net->termMask & ACE_TERM_LOWER_VE)))){
		    ++plane->netExitC;
		}
	    }
	}
    }

    /* set the used field for each RC					*/
    for( plane = AceG.matrix.rowPlanes; plane != NIL(acePlane);
	 plane = plane->next){
	for( index = plane->range.lower; index <= plane->range.upper; ++index){
	    AceG.matrix.rowRCs[ index ].used = TRUE;
	}
    }
    for( plane = AceG.matrix.colPlanes; plane != NIL(acePlane);
	 plane = plane->next){
	for( index = plane->range.lower; index <= plane->range.upper; ++index){
	    AceG.matrix.colRCs[ index ].used = TRUE;
	}
    }
}

/*
 * ace_shrink_nets
 *
 * Shrink nets down to their minimum size.
 * If dLen not NIL, do delta net length calc and DLEN events.
 * The dLen calculations cover change in net length due to movement
 * of each endpoint in its own plane.
 */
void
ace_shrink_nets( isRow, dLen)
short isRow;
int *dLen;
{
    int rci, i; /* rowcol index, other index				*/
    aceNet *net;
    aceRange or; /* original range					*/
#ifndef ACE_NO_UCF
    aceUCFEvent *ucfe;
#endif /* ACE_NO_UCF */

    if( isRow ){
	/* for each horizontal RC					*/
	for( rci = AceG.matrix.size.row - 1; rci >= 0; --rci){
	    /* for each net in the RC					*/
	    for( net = AceG.matrix.rowRCs[ rci ].netlist; net != NIL(aceNet);
		    net = net->next){
		/* find true range					*/
		or = net->range;
		net->range.upper = -1;
		net->range.lower = INT_MAX;
		for( i = net->cellCount - 1; i >= 0; --i){
		    if( net->cells[ i ]->loc.col > net->range.upper){
			net->range.upper = net->cells[ i ]->loc.col;
		    }
		    if( net->cells[ i ]->loc.col < net->range.lower){
			net->range.lower = net->cells[ i ]->loc.col;
		    }
		}
		/* if its a term net, set its virtual range		*/
		if( net->termMask ){
		    if( net->termMask & ACE_TERM_LOWER_VE){
			net->range.lower = 0;
		    } else {
			net->range.upper = AceG.matrix.size.col - 1;
		    }
		}
		if( dLen != NIL(int)){
		    if( net->range.upper - net->range.lower -
			    or.upper + or.lower){
			*dLen += net->range.upper - net->range.lower -
				or.upper + or.lower;
#ifndef ACE_NO_UCF
			ACE_UCF_EVENT( ACE_UCFE_NET_DLEN, net->ucfIndices,
				net->dependCount, i, ucfe);
#endif /* ACE_NO_UCF */
		    }

		    for( i = net->range.lower; i < or.lower; ++i){
			++AceG.matrix.map[ rci ][ i ].netCap.row;
		    }
		    for( i = or.lower; i < net->range.lower; ++i){
			--AceG.matrix.map[ rci ][ i ].netCap.row;
		    }

		    for( i = net->range.upper; i > or.upper; --i){
			++AceG.matrix.map[ rci ][ i ].netCap.row;
		    }
		    for( i = or.upper; i > net->range.upper; --i){
			--AceG.matrix.map[ rci ][ i ].netCap.row;
		    }
		}
	    }
	}
    } else {

	/* for each vertical RC						*/
	for( rci = AceG.matrix.size.col - 1; rci >= 0; --rci){
	    /* for each net in the RC					*/
	    for( net = AceG.matrix.colRCs[ rci ].netlist; net != NIL(aceNet);
		    net = net->next){
		/* find true range					*/
		or = net->range;
		net->range.upper = -1;
		net->range.lower = INT_MAX;
		for( i = net->cellCount - 1; i >= 0; --i){
		    if( net->cells[ i ]->loc.row > net->range.upper){
			net->range.upper = net->cells[ i ]->loc.row;
		    }
		    if( net->cells[ i ]->loc.row < net->range.lower){
			net->range.lower = net->cells[ i ]->loc.row;
		    }
		}
		/* if its a term net, set its virtual range		*/
		if( net->termMask ){
		    if( net->termMask & ACE_TERM_LOWER_VE){
			net->range.lower = 0;
		    } else {
			net->range.upper = AceG.matrix.size.row - 1;
		    }
		}
		if( dLen != NIL(int)){
		    if( or.upper - or.lower -
			    net->range.upper + net->range.lower){
			*dLen += net->range.upper - net->range.lower -
				or.upper + or.lower;
#ifndef ACE_NO_UCF
			ACE_UCF_EVENT( ACE_UCFE_NET_DLEN, net->ucfIndices,
				net->dependCount, i, ucfe);
#endif /* ACE_NO_UCF */
		    }

		    for( i = net->range.lower; i < or.lower; ++i){
			++AceG.matrix.map[ i ][ rci ].netCap.col;
		    }
		    for( i = or.lower; i < net->range.lower; ++i){
			--AceG.matrix.map[ i ][ rci ].netCap.col;
		    }

		    for( i = net->range.upper; i > or.upper; --i){
			++AceG.matrix.map[ i ][ rci ].netCap.col;
		    }
		    for( i = or.upper; i > net->range.upper; --i){
			--AceG.matrix.map[ i ][ rci ].netCap.col;
		    }
		}
	    }
	}
    }
}

/*
 * ace_init_a_params
 *
 * Set the annealing parameters to there initial values.
 */
void
ace_init_a_params()
{
    AceG.aparams = ( ace_get_default_ptr() )->cost;
}
