/*
 * debug.c
 *
 * This file contains the debugging routines for checking
 * the constancy of the matrixl state.
 */

#include "aceInternal.h"

/*
 * ace_check_nets
 *
 * Make sure the nets are correct.
 */
void
ace_check_nets()
{
    int rci, netc;
    aceNet *net;

    for( rci = AceG.matrix.size.row - 1; rci >= 0; --rci){
	netc = 0;
	for( net = AceG.matrix.rowRCs[ rci ].netlist; net != NIL(aceNet);
			net = net->next){
	    ++netc;
	    if( net->rowcol != rci){
		(void)fprintf( stderr, "row %d contains net in row %d\n",
			rci, net->rowcol);
	    }
	}
	if( netc != AceG.matrix.rowRCs[ rci ].count){
	    (void)fprintf( stderr, "row %d contains %d nets but count is %d\n",
		    netc, AceG.matrix.rowRCs[ rci ].count);
	}
    }
    for( rci = AceG.matrix.size.col - 1; rci >= 0; --rci){
	netc = 0;
	for( net = AceG.matrix.colRCs[ rci ].netlist; net != NIL(aceNet);
			net = net->next){
	    ++netc;
	    if( net->rowcol != rci){
		(void)fprintf( stderr, "col %d contains net in col %d\n",
			rci, net->rowcol);
	    }
	}
	if( netc != AceG.matrix.colRCs[ rci ].count){
	    (void)fprintf( stderr, "col %d contains %d nets but count is %d\n",
		    netc, AceG.matrix.colRCs[ rci ].count);
	}
    }
    (void)fflush( stderr );
}

/*
 * ace_dump_net_cap_map
 */
void
ace_dump_net_cap_map()
{
    int row, col;
    (void)printf( "Row Net Cap Map\n");
    for( row = AceG.matrix.size.row - 1; row >= 0; --row){
	for( col = 0; col < AceG.matrix.size.col; ++col){
	    (void)printf( "%2d", AceG.matrix.map[ row ][ col ].netCap.row);
	}
	(void)printf( "\n");
    }
    (void)printf( "\n\nCol Net Cap Map\n");
    for( row = AceG.matrix.size.row - 1; row >= 0; --row){
	for( col = 0; col < AceG.matrix.size.col; ++col){
	    (void)printf( "%2d", AceG.matrix.map[ row ][ col ].netCap.col);
	}
	(void)printf( "\n");
    }
}

/*
 * ace_verify_matrix
 *
 * Do some final checking to look for cell and net overlap that exceeds
 * capacity.
 */
void
ace_verify_matrix()
{
    int row, col;
    acePlane *p;
    aceMapEl **map;

    map = AceG.matrix.map;
    for( row = 0; row < AceG.matrix.size.row; ++row ){
	p = AceG.matrix.rowRCs[ row ].plane;
	for( col = 0; col < AceG.matrix.size.col; ++col){
	    if( map[ row ][ col ].netCap.row > p->netCap){
		(void)fprintf( stderr, "row %d: at %d: illegal net overlap\n",
			row, col);
		break;
	    }
	}
    }
    for( col = 0; col < AceG.matrix.size.col; ++col ){
	p = AceG.matrix.colRCs[ col ].plane;
	for( row = 0; row < AceG.matrix.size.row; ++row){
	    if( map[ row ][ col ].netCap.col > p->netCap){
		(void)fprintf( stderr, "col %d: at %d: illegal net overlap\n",
			col, row);
		break;
	    }
	}
    }
    for( row = 0; row < AceG.matrix.size.row; ++row ){
	for( col = 0; col < AceG.matrix.size.col; ++col){
	    p = AceG.matrix.rowRCs[ row ].plane;
	    if( map[ row ][ col ].cellCap.row > p->cellCap){
		(void)fprintf( stderr, "at [%d, %d]: illegal row cellCap\n",
			row, col);
	    }
	    p = AceG.matrix.colRCs[ col ].plane;
	    if( map[ row ][ col ].cellCap.col > p->cellCap){
		(void)fprintf( stderr, "at [%d, %d]: illegal col cellCap\n",
			row, col);
	    }
	}
    }
    (void)fflush( stderr );
}

/*
 * ace_check_netcounts
 *
 * Little garbage debugging routine.
 */
void
ace_check_netcounts()
{
    int rci, count;
    aceNet *net;

    for( rci = AceG.matrix.size.row - 1; rci >= 0; --rci){
	count = 0;
	for( net = AceG.matrix.rowRCs[rci].netlist; net != NIL(aceNet);
		net = net->next){
	    ++count;
	}
	if( count != AceG.matrix.rowRCs[rci].count){
	    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
			"ace_check_netcounts: inconsistent state");
	}
    }
    for( rci = AceG.matrix.size.col - 1; rci >= 0; --rci){
	count = 0;
	for( net = AceG.matrix.colRCs[rci].netlist; net != NIL(aceNet);
		net = net->next){
	    ++count;
	}
	if( count != AceG.matrix.colRCs[rci].count){
	    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
			"ace_check_netcounts: inconsistent state");
	}
    }
}

/*
 * ace_write_matrix
 *
 * Textually output the current matrix.
 * This version assumes there is no overlap in nets or cells.
 */
void
ace_write_matrix( fd, printNets)
FILE *fd;
int printNets;
{
    int row, col, i, off;
    int rows = 0, cols = 0;
    aceRowCol *rc;
    aceNet *net;
    char *name;

    for( col = 0; col < AceG.matrix.size.col; ++col){
	if( AceG.matrix.colRCs[ col ].count != 0){
	    ++cols;
	}
    }
    for( row = 0; row < AceG.matrix.size.row; ++row){
	if( AceG.matrix.rowRCs[ row ].count != 0){
	    ++rows;
	}
    }

    if( printNets){
	/* write general stats						*/
	(void)fprintf( fd, "MATRIX: %d rows by %d cols\n",
		    AceG.matrix.size.row, AceG.matrix.size.col);
	/* write nets							*/
	(void)fprintf( fd, "VERTICAL NETS\n");
	for( col = 0; col < AceG.matrix.size.col; ++col){
	    rc = ACE_COL_RC( col);
	    if( rc->count != 0){
		(void)fprintf( fd, "col %d: %d net(s)\n", col, rc->count);
	    }
	    for( net = rc->netlist; net != NIL(aceNet); net = net->next){
		(void)fprintf( fd, "\t%d to %d: %d cell(s)",
		    net->range.lower, net->range.upper, net->cellCount);
		if( net->termMask ){
		    (void)fprintf( fd, ": TERM NET\n");
		} else {
		    (void)fprintf( fd, "\n");
		}
	    }
	}
	(void)fprintf( fd, "HORIZONTAL NETS\n");
	for( row = 0; row < AceG.matrix.size.row; ++row){
	    rc = ACE_ROW_RC( row);
	    if( rc->count != 0){
		(void)fprintf( fd, "row %d: %d net(s)\n", row, rc->count);
	    }
	    for( net = rc->netlist; net != NIL(aceNet); net = net->next){
		(void)fprintf( fd, "\t%d to %d: %d cell(s)",
		    net->range.lower, net->range.upper, net->cellCount);
		if( net->termMask ){
		    (void)fprintf( fd, ": TERM NET\n");
		} else {
		    (void)fprintf( fd, "\n");
		}
	    }
	}
    }

    /* write cellmap							*/
    (void)fprintf( fd, "Actual size: %d rows by %d cols\n", rows, cols);
    --rows; /* off-by-one adjustment					*/
    for( row = AceG.matrix.size.row - 1; row >= 0; --row){
	rc = ACE_ROW_RC( row);
	if( rc->count == 0) continue;
	if( rows % 10 == 0){
	    (void)fprintf( fd, "-- ");
	} else if( rows % 5 == 0){
	    (void)fprintf( fd, " - ");
	} else {
	    (void)fprintf( fd, "   ");
	}
	--rows;
	off = 0;
	for( col = 0; col < AceG.matrix.size.col; ++col){
	    if( !AceG.matrix.colRCs[ col ].used) ++off;
	    if( AceG.matrix.colRCs[ col ].count == 0) continue;
	    name = NIL(char);
	    for( net = rc->netlist; net != NIL(aceNet); net = net->next){
		if(( col >= net->range.lower) && (col <= net->range.upper)){
		    for( i = 0; i < net->cellCount; ++i){
			if(( net->cells[i]->loc.col + off) == col ){
			    name = net->cells[i]->master->name;
			    break;
			}
		    }
		    if( name == NIL(char)){
			break;
		    }
		}
	    }
	    if( name == NIL(char)){
		(void)fprintf( fd, ".");
	    } else {
		(void)fprintf( fd, "%s", name);
	    }
	}
	(void)fprintf( fd, "\n");
    }
    (void)fprintf( fd, "   ");
    for( col = 0; col < cols; ++col){
	if( col % 5 == 0){
	    (void)fprintf( fd, "|");
	} else {
	    (void)fprintf( fd, " ");
	}
    }
    (void)fprintf( fd, "\n");
    ace_verify_matrix();
}
