/*
 * deltaCost.c
 *
 * Evaluate the cost of accepting a proposed move.
 * The the state of the matrix is actually changed to the proposed
 * next state so the UCF functions can properly evaluate the
 * incremental constraint cost.
 * The delta cost routine is basically a move dispatcher, executing the
 * appropriate move, evaluating the change in cost of user constraints,
 * and returning the total change in cost.
 */

#include "aceInternal.h"

/* paranoid checking of the deltaCost function				*/
#ifdef ACE_PARANOID
#define COST_PARANOID
#endif

/*
 * ace_delta_cost
 *
 * Return the change in cost from this state to the next state if
 * the proposed move were used to create the next state.
 */
double
ace_delta_cost()
{
    double delta_cost;
#ifndef ACE_NO_UCF
    aceUCFEvent *e;
    aceUCF *ucf;
#endif /* ACE_NO_UCF */

#ifdef COST_PARANOID
    double true_delta_cost;
    static int errCount = 0;
#endif /* COST_PARANOID */

    /* initialize delta cost counting varaibles			*/
    AceG.comdc.usedRC.row = AceG.comdc.usedRC.col = 0;
    AceG.comdc.netLen.row = AceG.comdc.netLen.col = 0;
    AceG.comdc.usedBias.row = AceG.comdc.usedBias.col = 0;
    AceG.comdc.netCap.row = AceG.comdc.netCap.col = 0;
    AceG.comdc.cellCap.row = AceG.comdc.cellCap.col = 0;

    if( AceG.move.index == ACE_MOVE_SWAP){
	delta_cost = ace_swap_net_move();
    } else if( AceG.move.index == ACE_MOVE_DISPLACE){
	delta_cost = ace_disp_net_move();
    } else if( AceG.move.index == ACE_MOVE_FLIP_TERM){
	delta_cost = ace_flip_term_net( AceG.move.net0);
    } else if( AceG.move.index == ACE_MOVE_RM_OVERLAP){
	delta_cost = ace_rm_overlap_move( AceG.move.plane);
    }

#ifndef ACE_NO_UCF
    AceG.ucfNewCost -= delta_cost;
    /* process UCF events						*/
    while(( e = ace_ucf_egetq( &AceG.ucfEventQ)) != NIL(aceUCFEvent)){
	if( AceG.ucfArray[ e->index ].called ){
	    ace_ucf_free_e( e );
	} else {
	    ucf = &( AceG.ucfArray[ e->index]);
	    /* save in case this move is rejected			*/
	    ucf->restoreSave = ucf->lastRet;
	    /* call UCF							*/
	    ucf->lastRet = ucf->cnstr_fn( ucf->argCount, ucf->argObjs);
	    /* mark it as called for this move				*/
	    ucf->called = TRUE;
	    /* update delta cost					*/
	    delta_cost += ucf->lastRet - ucf->restoreSave;
	    /* put the event on the `called Q' for accepting or rejecting
	     * this move
	     */
	    ace_ucf_eputq( e, &AceG.ucfCalledQ);
	}
    }
    AceG.ucfNewCost += delta_cost;
#endif /* ACE_NO_UCF */

#ifdef COST_PARANOID
    /* if paranoid, do full cost and delta cost for each move and make sure
     * they match ( within a small tolerance).
     */
    true_delta_cost = ace_cost() - AceG.cost;
    if( errCount < 50){
	if( fabs( delta_cost - true_delta_cost) > 0.01){
	    (void)fprintf( AceG.traceFp, "WARNING: delta cost error: %.4f\n",
		    delta_cost - true_delta_cost);
	    /*** debug */
	    ace_dump_net_cap_map();
	    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
		"delta cost error: off by %.4f", delta_cost - true_delta_cost);
	    ++errCount;
	}
	if( errCount == 50){
	    (void)fprintf( AceG.traceFp, "NO MORE DELTA COST ERROR MESSAGES\n");
	}
    }
    delta_cost = true_delta_cost;
#endif /* COST_PARANOID */

    return( delta_cost);
}
