/*
 * dispNet.c
 *
 * Execute and evaluate the change in cost for the displace net
 * and swap net moves.
 * Also, provide a procedure to undo the displace net and swap nets moves.
 */

#include "aceInternal.h"

/*
 * ace_disp_net_move
 *
 * Return the change in cost for the proposed displace move.
 */
double
ace_disp_net_move()
{
    /* In the displace move, check plane resize.			*/
    ace_displace_net( AceG.move.net0, AceG.move.displacement, 0, TRUE);

    return( ace_cost_of_count());
}

/*
 * ace_swap_net_move
 *
 * Return the change in cost for the proposed swap move.
 * The swap is done as two displaces with no plane resize checking.
 */
double
ace_swap_net_move()
{
    /* In the swap move, plane size won't change.			*/
    ace_displace_net( AceG.move.net0, AceG.move.displacement, 0, FALSE);
    ace_displace_net( AceG.move.net1, -AceG.move.displacement, 1, FALSE);

    return( ace_cost_of_count());
}

/*
 * ace_cost_of_count
 *
 * Calcultate the cost based on the count of instaces
 * of each type of cost.
 */
double
ace_cost_of_count()
{
    return(
	AceG.aparams.usedRC.row * (double)AceG.comdc.usedRC.row +
	AceG.aparams.usedRC.col * (double)AceG.comdc.usedRC.col +
	AceG.aparams.netLength.row * (double)AceG.comdc.netLen.row +
	AceG.aparams.netLength.col * (double)AceG.comdc.netLen.col +
	AceG.aparams.usedBias.row * (double)AceG.comdc.usedBias.row +
	AceG.aparams.usedBias.col * (double)AceG.comdc.usedBias.col +
	AceG.aparams.netCapacity.row * (double)AceG.comdc.netCap.row +
	AceG.aparams.netCapacity.col * (double)AceG.comdc.netCap.col +
	AceG.aparams.cellCapacity.row * (double)AceG.comdc.cellCap.row +
	AceG.aparams.cellCapacity.col * (double)AceG.comdc.cellCap.col
    );
}

/*
 * Utility bit definitions used only in ace_displace_net().
 * Hopefully they will reduce the number of variables necessary in
 * the procedure, and possibly speed things up a bit.
 */
#define DCS_ROW_NET		(1 << 0) /* net orientation */
#define DCS_NET_UB		(1 << 1) /* cell is on net upper boundary */
#define DCS_NET_LB		(1 << 2) /* cell is on net lower boundary */
#define DCS_NET_B		(DCS_NET_UB | DCS_NET_LB) /* any boundary */

/*
 * ace_displace_net
 *
 * Displace a net by the given displacement.
 * Arguments are assumed to be correct (i.e., the new location is assumed
 * to be in the matrix, as well as in the same plane, etc.).
 *
 ***Opt: depending on the orien of the displaced net, call an optimized
 *       cost_of_count() that only multiplies the counts that could
 *       have possibly changed [ may not really save that much ].
 */
void
ace_displace_net( net, displacement, whichSave, doPlaneSize)
aceNet *net;		/* the net being displaced			*/
int displacement;	/* the displacement				*/
int whichSave;		/* which pnet range save to use			*/
int doPlaneSize;	/* boolean, set if this is a displace move	*/
{
    register unsigned status;		/* various uses			*/
    int i, index;			/* loop indecies		*/
    int newloc;				/* cell loc coord that changed	*/
    aceMapEl **map;			/* matrix capacity map		*/
    aceCoord loc;			/* location being used		*/
    aceMaster *master;			/* master of cell		*/
    aceNet *pnet;			/* perpendicular net		*/
    aceRange range;			/* range of shrinkage of pnet	*/
    aceRowCol *rc;			/* pointer to an RC		*/
    acePlane *plane;			/* pointer to a plane		*/
    int pbound;				/* upper bound of move plane	*/
    int temp;				/* temp space for storing ints	*/
#ifndef ACE_NO_UCF
    aceUCFEvent *ucfe;			/* process UCF events		*/
    int ucfi;				/* index for UCF depend arrays	*/
#endif /* ACE_NO_UCF */

    map = AceG.matrix.map;
    if( net->isRow){
	status = DCS_ROW_NET;
	pbound = AceG.matrix.rowRCs[ net->rowcol ].plane->range.upper;
    } else {
	status = 0;
	pbound = AceG.matrix.colRCs[ net->rowcol ].plane->range.upper;
    }

    /* For each cell in the net, move it to its new location.
     * Check cell capacities at new and old locations for changing cost.
     * Check each cells perpendicular net for a change in length,
     * and evaluate all associtated delta costs.
     */
    for( i = net->cellCount - 1; i >= 0; --i){
	status &= ~(DCS_NET_B);
	loc = net->cells[ i ]->loc;
	master = net->cells[ i ]->master;
	if( status & DCS_ROW_NET ){
	    pnet = net->cells[ i ]->colNet;

	    /* check if cell is on net boundary				*/
	    if( pnet->range.upper == loc.row){
		status |= DCS_NET_UB;
	    }
	    if( pnet->range.lower == loc.row){
		status |= DCS_NET_LB;
	    }
	} else {
	    pnet = net->cells[ i ]->rowNet;

	    /* check if cell is on net boundary				*/
	    if( pnet->range.upper == loc.col){
		status |= DCS_NET_UB;
	    }
	    if( pnet->range.lower == loc.col){
		status |= DCS_NET_LB;
	    }
	}

	/* generate UCF event						*/
#ifndef ACE_NO_UCF
	ACE_UCF_EVENT( ACE_UCFE_NET_CELL_MOVE, pnet->ucfIndices,
			pnet->dependCount, ucfi, ucfe);
#endif /* ACE_NO_UCF */

	/* decrease map cell capacity at original location		*/
	if(( temp = map[ loc.row ][ loc.col ].cellCap.row -
		AceG.matrix.rowRCs[ loc.row ].plane->cellCap) > 0){
	    AceG.comdc.cellCap.row -= MIN( temp, master->cellCap.row);
	}
	map[ loc.row ][ loc.col ].cellCap.row
		-= master->cellCap.row;

	if(( temp = map[ loc.row ][ loc.col ].cellCap.col -
		AceG.matrix.colRCs[ loc.col ].plane->cellCap) > 0){
	    AceG.comdc.cellCap.col -= MIN( temp, master->cellCap.col);
	}
	map[ loc.row ][ loc.col ].cellCap.col
		-= master->cellCap.col;
	
	/* change location						*/
	if( status & DCS_ROW_NET){
	    newloc = net->cells[ i ]->loc.row = loc.row =
			loc.row + displacement;
	} else {
	    newloc = net->cells[ i ]->loc.col = loc.col =
			loc.col + displacement;
	}

	/* increase cell capacity at new location			*/
	if(( temp = map[ loc.row ][ loc.col ].cellCap.row +
		master->cellCap.row -
		AceG.matrix.rowRCs[ loc.row ].plane->cellCap) > 0){
	    AceG.comdc.cellCap.row += MIN( temp, master->cellCap.row);
	}
	map[ loc.row ][ loc.col ].cellCap.row
		+= master->cellCap.row;

	if(( temp = map[ loc.row ][ loc.col ].cellCap.col +
		master->cellCap.col -
		AceG.matrix.colRCs[ loc.col ].plane->cellCap) > 0){
	    AceG.comdc.cellCap.col += MIN( temp, master->cellCap.col);
	}
	map[ loc.row ][ loc.col ].cellCap.col
		+= master->cellCap.col;

	/* save original range before it gets updated			*/
	if( whichSave == 0){
	    pnet->save0 = pnet->range;
	} else {
	    pnet->save1 = pnet->range;
	}

	/* If cell was on the boundary, the net could have shrunk.
	 * Here, range is keeping track of the deleted portion of pnet
	 */
	if(( status & DCS_NET_LB) && (displacement > 0)){
	    /* net lower boundary shrunk				*/
	    range.lower = pnet->range.lower;

	    /* if this is a term net, use its virtual range		*/
	    if( pnet->termMask & ACE_TERM_LOWER_VE){
		range.upper = pnet->range.lower; /* should be 0		*/
	    } else {
		/* find new minmum index				*/
		range.upper = INT_MAX;
		if( status & DCS_ROW_NET){
		    for( index = pnet->cellCount - 1; index >= 0; --index){
			if( range.upper > pnet->cells[ index ]->loc.row){
			    range.upper = pnet->cells[ index ]->loc.row;
			}
		    }
		} else {
		    for( index = pnet->cellCount - 1; index >= 0; --index){
			if( range.upper > pnet->cells[ index ]->loc.col){
			    range.upper = pnet->cells[ index ]->loc.col;
			}
		    }
		}
	    }
	    /* update pnet and shrink range to not include endpoint	*/
	    pnet->range.lower = range.upper;
	    --range.upper;
	} else if(( status & DCS_NET_UB) && (displacement < 0)){
	    /* net upper boundary shrunk				*/
	    range.upper = pnet->range.upper;

	    /* check to see if its a virtual range			*/
	    if( pnet->termMask && !( pnet->termMask & ACE_TERM_LOWER_VE)){
		range.lower = pnet->range.upper;
	    } else {
		/* find new maximum index				*/
		range.lower = -1;
		if( status & DCS_ROW_NET){
		    for( index = pnet->cellCount - 1; index >= 0; --index){
			if( range.lower < pnet->cells[index]->loc.row){
			    range.lower = pnet->cells[index]->loc.row;
			}
		    }
		} else {
		    for( index = pnet->cellCount - 1; index >= 0; --index){
			if( range.lower < pnet->cells[index]->loc.col){
			    range.lower = pnet->cells[index]->loc.col;
			}
		    }
		}
	    }
	    /* update pnet and shrink range to not include endpoint	*/
	    pnet->range.upper = range.lower;
	    ++range.lower;
	} else {
	    range.lower = 0;
	    range.upper = -1;
	}

	/* now, deleted range is known, so delete it			*/
	if( status & DCS_ROW_NET){
	    /* set plane here cuz used below				*/
	    plane = AceG.matrix.colRCs[ loc.col ].plane;
	    for( index = range.lower; index <= range.upper; ++index){
		if(( index <= pbound) &&
		   ( map[ index ][ pnet->rowcol ].netCap.col > plane->netCap)){
		    --AceG.comdc.netCap.col;
		}
		--( map[ index ][ pnet->rowcol ].netCap.col);
	    }
	} else {
	    /* set plane here cuz used below				*/
	    plane = AceG.matrix.rowRCs[ loc.row ].plane;
	    for( index = range.lower; index <= range.upper; ++index){
		if(( index <= pbound) &&
		  ( map[ pnet->rowcol ][ index ].netCap.row > plane->netCap)){
		    --AceG.comdc.netCap.row;
		}
		--( map[ pnet->rowcol ][ index ].netCap.row);
	    }
	}

	/* see if the new cell loc is outside of range			*/
	range.upper = -1;
	range.lower = 0;
	if( newloc > pnet->range.upper){
	    range.upper = newloc;
	    range.lower = pnet->range.upper + 1;
	    pnet->range.upper = newloc;
	} else if( newloc < pnet->range.lower){
	    range.lower = newloc;
	    range.upper = pnet->range.lower - 1;
	    pnet->range.lower = newloc;
	}

	/* Add to the net capacity along the new extent of the net,
	 * and calculate the change in cost for both the change in net length
	 * and the change in net capacity.
	 */
	if( status & DCS_ROW_NET){
	    if( whichSave == 0){
		/* DeltaLength = LENf - LENi
		 *             = (Uf - Lf + 1 ) - (Ui - Li + 1)
		 *             = Uf - Lf - Ui + Uf
		 */
		index = pnet->range.upper - pnet->range.lower -
			pnet->save0.upper + pnet->save0.lower;
	    } else {
		index = pnet->range.upper - pnet->range.lower -
			pnet->save1.upper + pnet->save1.lower;
	    }
	    AceG.comdc.netLen.col += index;
#ifndef ACE_NO_UCF
	    if( index ){
		ACE_UCF_EVENT( ACE_UCFE_NET_DLEN, pnet->ucfIndices,
			   pnet->dependCount, ucfi, ucfe);
	    }
#endif /* ACE_NO_UCF */

	    for( index = range.lower; index <= range.upper; ++index){
		if(( index <= pbound) &&
		  ( map[ index ][ pnet->rowcol ].netCap.col >= plane->netCap)){
		    ++AceG.comdc.netCap.col;
		}
		++( map[ index ][ pnet->rowcol ].netCap.col);
	    }

	} else {
	    if( whichSave == 0){
		/* (Uf - Lf ) - (Uf - Li)				*/
		index = pnet->range.upper - pnet->range.lower -
			pnet->save0.upper + pnet->save0.lower;
	    } else {
		index = pnet->range.upper - pnet->range.lower -
			pnet->save1.upper + pnet->save1.lower;
	    }
	    AceG.comdc.netLen.row += index;
#ifndef ACE_NO_UCF
	    if( index ){
		ACE_UCF_EVENT( ACE_UCFE_NET_DLEN, pnet->ucfIndices,
			   pnet->dependCount, ucfi, ucfe);
	    }
#endif /* ACE_NO_UCF */

	    for( index = range.lower; index <= range.upper; ++index){
		if(( index <= pbound) &&
		   ( map[ pnet->rowcol ][ index ].netCap.row >= plane->netCap)){
		    ++AceG.comdc.netCap.row;
		}
		++( map[ pnet->rowcol ][ index ].netCap.row);
	    }
	}

    } /* end moving cells						*/

    /* next step: update the map net count entries			*/
    if( status & DCS_ROW_NET){
	/* set this now cuz we need it later anyway			*/
	rc = &( AceG.matrix.rowRCs[ net->rowcol ]);
	/* get net's plane						*/
	plane = rc->plane;
	temp = plane->netCap;
	loc.row = net->rowcol;
	loc.col = loc.row + displacement;

	for( i = net->range.lower; i <= net->range.upper; ++i){
	    /* decrease net count along old rowcol			*/
	    if( AceG.matrix.colRCs[ i ].used &&
	       ( map[ loc.row ][ i ].netCap.row > temp)){
		--AceG.comdc.netCap.row;
	    }
	    --(map[ loc.row ][ i ].netCap.row);

	    /* increase net count along new rowcol			*/
	    if( AceG.matrix.colRCs[ i ].used &&
	       (map[ loc.col ][ i ].netCap.row >= temp)){
		++AceG.comdc.netCap.row;
	    }
	    ++(map[ loc.col ][ i ].netCap.row);
	}
    } else {
	rc = &( AceG.matrix.colRCs[ net->rowcol ]);
	/* get net's plane						*/
	plane = rc->plane;
	temp = plane->netCap;
	loc.row = net->rowcol;
	loc.col = loc.row + displacement;

	for( i = net->range.lower; i <= net->range.upper; ++i){
	    /* decrease net count along old rowcol			*/
	    if( AceG.matrix.rowRCs[ i ].used &&
	       ( map[ i ][ loc.row ].netCap.col > temp)){
		--AceG.comdc.netCap.col;
	    }
	    --(map[ i ][ loc.row ].netCap.col);

	    /* increase net count along new rowcol			*/
	    if( AceG.matrix.rowRCs[ i ].used &&
	       (map[ i ][ loc.col ].netCap.col >= temp)){
		++AceG.comdc.netCap.col;
	    }
	    ++(map[ i ][ loc.col ].netCap.col);
	}
    }
#ifndef ACE_NO_UCF
    ACE_UCF_EVENT( ACE_UCFE_NET_MOVE, net->ucfIndices, net->dependCount,
		   ucfi, ucfe);
#endif /* ACE_NO_UCF */

    /* next step: move the net to the new rowcol			*/
    /* unlink the net from the RC netlist				*/
    if( rc->netlist == net){
	rc->netlist = net->next;
    } else {
	for( pnet = rc->netlist; pnet->next != net; pnet = pnet->next) ;
	pnet->next = net->next;
    }

    /* decrement the netcount at the rowcol				*/
    if( rc->count == 1){
	if( status & DCS_ROW_NET){
	    --AceG.comdc.usedRC.row;
	} else {
	    --AceG.comdc.usedRC.col;
	}
    }
    --( rc->count);
    if( net->termMask){
	--( rc->termNetC);
    }

    /* get the new rc and adjust the rowcol number of the net		*/
    rc += displacement;
    net->rowcol += displacement;

    /* link into net list						*/
    net->next = rc->netlist;
    rc->netlist = net;

    /* increment netcount						*/
    if( rc->count == 0){
	if( status & DCS_ROW_NET){
	    ++AceG.comdc.usedRC.row;
	} else {
	    ++AceG.comdc.usedRC.col;
	}
    }
    ++( rc->count);
    if( net->termMask){
	++( rc->termNetC);
    }

    /* calc new max used rowcol [ ie, plane size ]			*/
    if( doPlaneSize ){
	plane->saveUpper = plane->range.upper;
	if( status & DCS_ROW_NET){
	    if( net->rowcol > plane->range.upper){
		/* Add in the netCap due to overlapping nets in the newly
		 * non-empty area.  This includes overlapping termNets leaving
		 * the upper edge and any other overlapping nets leaving
		 * a particular plane.
		 */
		for( i = AceG.matrix.size.col - 1; i >= 0; --i){
		    if( AceG.matrix.colRCs[ i ].count <= 1) continue;
		    /* else traverse the newly non-empty rows		*/
		    for( index = plane->range.upper + 1; index <= net->rowcol;
			    ++index){
			/* if there is net overlap, add it in	*/
			if(( temp = map[ index ][ i ].netCap.col -
				AceG.matrix.colRCs[ i ].plane->netCap) > 0){
			    AceG.comdc.netCap.col += temp;
			}
		    }
		}
		/* set the new used flags				*/
		for( index = plane->range.upper + 1; index <= net->rowcol;
			    ++index){
		    AceG.matrix.rowRCs[ index ].used = TRUE;
		}

		/* increased max used rowcol				*/
		AceG.comdc.usedBias.row = net->rowcol - plane->range.upper;
		plane->range.upper = net->rowcol;
		plane->emptyC -= AceG.comdc.usedBias.row;
	    } else if((( net->rowcol - displacement) == plane->range.upper) &&
		      ( displacement < 0)){
		/* could have decreased, so look at rc of former max used */
		rc -= displacement;
		while( rc->count == 0){
		    /* Looks dangerous, but guarenteed theres at least one
		     * rc with a net below here. Obviously, the the new net loc
		     * isn't larger than the former plane.range.upper, and
		     * since the nets former rowcol == former plane.range.upper,
		     * its current loc must be less than the former p.r.upper.
		     *
		     * Intuitively, we're starting at the outer edge, and
		     * working inward, so we're bound to hit something.
		     */
		    rc->used = FALSE;
		    --rc;
		    --AceG.comdc.usedBias.row;
		    --plane->range.upper;
		}
		/* get rid of any net overlap in the newly-empty col */
		for( i = AceG.matrix.size.col - 1; i >= 0; --i){
		    if( AceG.matrix.colRCs[ i ].count <= 
			AceG.matrix.colRCs[ i ].plane->netCap) continue;
		    /* else traverse the newly empty rows		*/
		    for( index = plane->range.upper - AceG.comdc.usedBias.row;
			 index > plane->range.upper; --index){
			/* if there is net overlap, add it in	*/
			if(( temp = map[ index ][ i ].netCap.col -
				AceG.matrix.colRCs[ i ].plane->netCap) > 0){
			    AceG.comdc.netCap.col -= temp;
			}
		    }
		}
		plane->emptyC -= AceG.comdc.usedBias.row;
	    }

	    /* if AceG.comdc.usedBias non-zero, net length changed		*/
	    AceG.comdc.netLen.col += AceG.comdc.usedBias.row * plane->netExitC;
	} else {
	    if( net->rowcol > plane->range.upper){
		/* increased max used rowcol				*/
		for( i = AceG.matrix.size.row - 1; i >= 0; --i){
		    if( AceG.matrix.rowRCs[ i ].count <= 1) continue;
		    /* else traverse the newly non-empty rows		*/
		    for( index = plane->range.upper + 1; index <= net->rowcol;
			    ++index){
			/* if there is net overlap, add it in	*/
			if(( temp = map[ i ][ index ].netCap.row -
				AceG.matrix.rowRCs[ i ].plane->netCap) > 0){
			    AceG.comdc.netCap.row += temp;
			}
		    }
		}
		/* set the new used flags				*/
		for( index = plane->range.upper + 1; index <= net->rowcol;
			    ++index){
		    AceG.matrix.colRCs[ index ].used = TRUE;
		}

		/* do the cost stuff					*/
		AceG.comdc.usedBias.col = net->rowcol - plane->range.upper;
		plane->range.upper = net->rowcol;
		plane->emptyC -= AceG.comdc.usedBias.col;
	    } else if((( net->rowcol - displacement) == plane->range.upper) &&
		      ( displacement < 0)){
		/* could have decreased, so look at rc of former maxUsed */
		rc -= displacement;
		while( rc->count == 0){
		    rc->used = FALSE;
		    --rc;
		    --AceG.comdc.usedBias.col;
		    --plane->range.upper;
		}
		/* get rid of any net overlap in the newly-empty col */
		for( i = AceG.matrix.size.row - 1; i >= 0; --i){
		    if( AceG.matrix.rowRCs[ i ].count <=
			AceG.matrix.rowRCs[ i ].plane->netCap) continue;
		    /* else traverse the newly empty rows		*/
		    for( index = plane->range.upper - AceG.comdc.usedBias.col;
			 index > plane->range.upper; --index){
			/* if there is net overlap, add it in	*/
			if(( temp = map[ i ][ index ].netCap.row -
				AceG.matrix.rowRCs[ i ].plane->netCap) > 0){
			    AceG.comdc.netCap.row -= temp;
			}
		    }
		}
		plane->emptyC -= AceG.comdc.usedBias.col;
	    }

	    /* if AceG.comdc.usedBias non-zero, net length changed		*/
	    AceG.comdc.netLen.row += AceG.comdc.usedBias.col * plane->netExitC;
	}
    }
}

/*
 * ace_replace_net
 *
 * This undoes the work of ace_displace_net() without worrying about cost.
 * ace_displace_net() stores some info that makes undoing the move
 * fairly trivial.
 */
void
ace_replace_net( net, displacement, whichRestore, doPlaneSize)
aceNet *net;
int displacement;
int whichRestore;
int doPlaneSize; /* set if its a displacement move and size could change */
{
    int i, index;			/* loop indecies		*/
    aceMapEl **map;			/* the map			*/
    aceCoord loc;
    aceMaster *master;
    aceNet *pnet;
    acePlane *plane;
    aceRowCol *rc;
    int netInRow;
    aceRange oldRange;			/* range before proposed move	*/

    map = AceG.matrix.map;
    netInRow = net->isRow;

    /* replace each cell						*/
    for( i = net->cellCount - 1; i >= 0; --i){
	loc = net->cells[ i ]->loc;
	master = net->cells[ i ]->master;
	if( netInRow ){
	    pnet = net->cells[ i ]->colNet;
	} else {
	    pnet = net->cells[ i ]->rowNet;
	}

	/* adjust cell capacities					*/
	map[ loc.row ][ loc.col ].cellCap.row
		-= master->cellCap.row;
	map[ loc.row ][ loc.col ].cellCap.col
		-= master->cellCap.col;
	
	/* adjust cell loc						*/
	if( netInRow){
	    net->cells[ i ]->loc.row = loc.row = loc.row + displacement;
	} else {
	    net->cells[ i ]->loc.col = loc.col = loc.col + displacement;
	}

	/* adjust cell capacities					*/
	map[ loc.row ][ loc.col ].cellCap.row
		+= master->cellCap.row;
	map[ loc.row ][ loc.col ].cellCap.col
		+= master->cellCap.col;

	/* update appropriate net capacities				*/
	if( whichRestore == 0){
	    oldRange = pnet->save0;
	} else {
	    oldRange = pnet->save1;
	}

	if( pnet->range.lower > oldRange.lower){
	    /* expand this net down to its former size			*/
	    if( netInRow){
		for(index = pnet->range.lower - 1; index >= oldRange.lower;
			--index){
		    ++( map[ index ][ pnet->rowcol ].netCap.col);
		}
	    } else {
		for(index = pnet->range.lower - 1; index >= oldRange.lower;
			--index){
		    ++( map[ pnet->rowcol ][ index ].netCap.row);
		}
	    }
	} else if( pnet->range.lower < oldRange.lower){
	    /* shrink this net up to its former size			*/
	    if( netInRow){
		for(index = pnet->range.lower; index < oldRange.lower;
			++index){
		    --( map[ index ][ pnet->rowcol ].netCap.col);
		}
	    } else {
		for(index = pnet->range.lower; index < oldRange.lower;
			++index){
		    --( map[ pnet->rowcol ][ index ].netCap.row);
		}
	    }
	}

	if( pnet->range.upper < oldRange.upper){
	    /* expand this net up to its former size			*/
	    if( netInRow){
		for(index = pnet->range.upper + 1; index <= oldRange.upper;
			++index){
		    ++( map[ index ][ pnet->rowcol ].netCap.col);
		}
	    } else {
		for(index = pnet->range.upper + 1; index <= oldRange.upper;
			++index){
		    ++( map[ pnet->rowcol ][ index ].netCap.row);
		}
	    }
	} else if( pnet->range.upper > oldRange.upper){
	    /* shrink this net down to its former size			*/
	    if( netInRow){
		for(index = pnet->range.upper; index > oldRange.upper;
			--index){
		    --( map[ index ][ pnet->rowcol ].netCap.col);
		}
	    } else {
		for(index = pnet->range.upper; index > oldRange.upper;
			--index){
		    --( map[ pnet->rowcol ][ index ].netCap.row);
		}
	    }
	}

	/* restore pnet's range						*/
	pnet->range = oldRange;
    } /* end restoring cells and perpendicualar nets			*/

    /* restore map net count entries					*/
    loc.row = net->rowcol;
    loc.col = loc.row + displacement;
    if( netInRow){
	for( i = net->range.lower; i <= net->range.upper; ++i){
	    --( map[ loc.row ][ i ].netCap.row);
	    ++( map[ loc.col ][ i ].netCap.row);
	}
    } else {
	for( i = net->range.lower; i <= net->range.upper; ++i){
	    --( map[ i ][ loc.row ].netCap.col);
	    ++( map[ i ][ loc.col ].netCap.col);
	}
    }

    /* move net back to its RC						*/
    if( netInRow){
	rc = &( AceG.matrix.rowRCs[ net->rowcol ]);
    } else {
	rc = &( AceG.matrix.colRCs[ net->rowcol ]);
    }

    /* unlink net from RC netlist					*/
    if( rc->netlist == net){
	rc->netlist = net->next;
    } else {
	for( pnet = rc->netlist; pnet->next != net; pnet = pnet->next) ;
	pnet->next = net->next;
    }

    /* decrement net count at rc					*/
    --( rc->count);
    if( net->termMask){
	--( rc->termNetC);
    }

    /* put net into oringal rc						*/
    rc += displacement;
    net->rowcol += displacement;

    net->next = rc->netlist;
    rc->netlist = net;

    ++( rc->count);
    if( net->termMask){
	 ++( rc->termNetC);
    }

    if( doPlaneSize ){
	/* get the plane						*/
	plane =  rc->plane;
	if( plane->saveUpper != plane->range.upper){
	    /* get the starting rc					*/
	    if( netInRow){
		rc = &( AceG.matrix.rowRCs[ plane->saveUpper ]);
	    } else {
		rc = &( AceG.matrix.colRCs[ plane->saveUpper ]);
	    }
	    if( plane->saveUpper > plane->range.upper){
		for( index = plane->saveUpper; index >= plane->range.upper;
			--index){
		    rc->used = TRUE;
		    --rc;
		}
	    } else {
		++rc;
		for( index = plane->saveUpper + 1; index <= plane->range.upper;
			++index){
		    rc->used = FALSE;
		    ++rc;
		}
	    }
	    plane->emptyC += plane->range.upper - plane->saveUpper;
	    plane->range.upper = plane->saveUpper;
	}
    }
}
