/*
 * flipTerm.c
 *
 * Do the flip term net move.
 */

#include "aceInternal.h"

/*
 * ace_flip_term_net
 *
 * Flip the terminal on this net to the other side of the matrix.
 * Calculate the change in cost due to net overlap, net length,
 * and UCFs.
 */
double
ace_flip_term_net( net )
aceNet *net;
{
    aceRange cellRange, range;
    acePlane *plane;
    aceMapEl **map;
    int i, netCapCount = 0, dlen;
#ifndef ACE_NO_UCFS
    int ucfi;
    aceUCFEvent *ucfe;
#endif /* ACE_NO_UCFS */

    map = AceG.matrix.map;
    /* find the range of cells in this net (not counting veritual ext.)	*/
    cellRange.lower = INT_MAX;
    cellRange.upper = -1;
    if( net->isRow){
	for( i = net->cellCount - 1; i >= 0; --i){
	    if( net->cells[ i ]->loc.col > cellRange.upper){
		cellRange.upper = net->cells[ i ]->loc.col;
	    }
	    if( net->cells[ i ]->loc.col < cellRange.lower){
		cellRange.lower = net->cells[ i ]->loc.col;
	    }
	}
    } else {
	for( i = net->cellCount - 1; i >= 0; --i){
	    if( net->cells[ i ]->loc.row > cellRange.upper){
		cellRange.upper = net->cells[ i ]->loc.row;
	    }
	    if( net->cells[ i ]->loc.row < cellRange.lower){
		cellRange.lower = net->cells[ i ]->loc.row;
	    }
	}
    }

    /* remove net capacity of current virtual extension (VE)		*/
    if( net->termMask & ACE_TERM_LOWER_VE){
	range.lower = net->range.lower;
	range.upper = cellRange.lower - 1;
	if( cellRange.upper != net->range.upper){ /* never seems to happen */
	    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
		    "ace_flip_term_net: inconsistent term net range");
	}

	/* toggle the VE bits						*/
	net->termMask &= ~ACE_TERM_LOWER_VE;
    } else {
	range.upper = net->range.upper; /* which is the upper VE value	*/
	range.lower = cellRange.upper + 1;
	if( cellRange.lower != net->range.lower){
	    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
		    "ace_flip_term_net: inconsistent term net range");
	}

	/* toggle the VE bits						*/
	net->termMask |= ACE_TERM_LOWER_VE;
    }

    /* While evaluating delta cost, be careful not to penalize overlap
     * in empty/dead space of perpendicuar planes;
     */
    if( net->isRow){
	plane = AceG.matrix.rowRCs[ net->rowcol ].plane;
	for( i = range.lower; i <= range.upper; ++i){
	    if( AceG.matrix.colRCs[ i ].used &&
	       ( map[ net->rowcol ][ i ].netCap.row > plane->netCap)){
		--netCapCount;
	    }
	    --map[ net->rowcol ][ i ].netCap.row;
	}
    } else {
	plane = AceG.matrix.colRCs[ net->rowcol ].plane;
	for( i = range.lower; i <= range.upper; ++i){
	    if( AceG.matrix.rowRCs[ i ].used &&
	       ( map[ i ][ net->rowcol ].netCap.col > plane->netCap)){
		--netCapCount;
	    }
	    --map[ i ][ net->rowcol ].netCap.col;
	}
    }

    /* add net capacity of new virtual extension (VE bit set to new val) */
    if( net->termMask & ACE_TERM_LOWER_VE){
	range.lower = 0;
	range.upper = cellRange.lower - 1;
    } else {
	range.upper = net->isRow ?
			AceG.matrix.size.col - 1 : AceG.matrix.size.row - 1;
	range.lower = cellRange.upper + 1;
    }
    if( net->isRow){
	for( i = range.lower; i <= range.upper; ++i){
	    if(( i <= AceG.matrix.colRCs[ i ].plane->range.upper) &&
	       ( map[ net->rowcol ][ i ].netCap.row >= plane->netCap)){
		++netCapCount;
	    }
	    ++map[ net->rowcol ][ i ].netCap.row;
	}
    } else {
	for( i = range.lower; i <= range.upper; ++i){
	    if(( i <= AceG.matrix.rowRCs[ i ].plane->range.upper) &&
	       ( map[ i ][ net->rowcol ].netCap.col >= plane->netCap)){
		++netCapCount;
	    }
	    ++map[ i ][ net->rowcol ].netCap.col;
	}
    }

    /* Calculate change in net length. Must look at each perp plane and throw
     * away the dead and empty RCs to get the true net length. Also, must
     * maintain the netExitC count for each plane.
     */
    plane = net->isRow ? AceG.matrix.colPlanes : AceG.matrix.rowPlanes;
    dlen = 0;
    for( ; plane != NIL(acePlane); plane = plane->next){
	if( cellRange.lower > plane->range.upper){
	    if( net->termMask & ACE_TERM_LOWER_VE){
		/* now newly crossing this plane			*/
		dlen += plane->range.upper - plane->range.lower + 1;
		++plane->netExitC;
	    } else {
		/* no longer cross this plane				*/
		dlen -= plane->range.upper - plane->range.lower + 1;
		--plane->netExitC;
	    }
	} else if( cellRange.lower >= plane->range.lower){
	    /* lower end of cell range in this plane			*/
	    if( net->termMask & ACE_TERM_LOWER_VE){
		/* lower piece of plane now added to the range		*/
		dlen += cellRange.lower - plane->range.lower;
	    } else {
		/* lower piece of this plane no longer in range		*/
		dlen -= cellRange.lower - plane->range.lower;
	    }
	} /* otherwise cellRange.lower is in a lower plane		*/

	if( cellRange.upper < plane->range.lower){
	    if( net->termMask & ACE_TERM_LOWER_VE){
		/* no longer cross this plane				*/
		dlen -= plane->range.upper - plane->range.lower + 1;
		--plane->netExitC;
	    } else {
		/* now newly crossing this plane			*/
		dlen += plane->range.upper - plane->range.lower + 1;
		++plane->netExitC;
	    }
	} else if( cellRange.upper <= plane->range.upper){
	    /* upper end of cell range in this plane			*/
	    if( net->termMask & ACE_TERM_LOWER_VE){
		/* upper piece of this plane no longer in range		*/
		dlen -= plane->range.upper - cellRange.upper;
		--plane->netExitC;
	    } else {
		/* upper piece of plane now added to the range		*/
		dlen += plane->range.upper - cellRange.upper;
		++plane->netExitC;
	    }
	} /* otherwise cellRange.upper is in a higher plane		*/
    }

    /* save and update the net's range					*/
    net->save0 = net->range;
    net->save1 = cellRange; /* a little hacky, but as long as there's space...*/
    if( net->termMask & ACE_TERM_LOWER_VE){
	net->range.lower = range.lower;
	net->range.upper = cellRange.upper;
    } else {
	net->range.lower = cellRange.lower;
	net->range.upper = range.upper;
    }

    /* generate DLEN event and NET_CELL_MOVE events			*/
    if( dlen ){
	ACE_UCF_EVENT( ACE_UCFE_NET_DLEN, net->ucfIndices, net->dependCount,
		ucfi, ucfe);
    }
    ACE_UCF_EVENT( ACE_UCFE_NET_CELL_MOVE | ACE_UCFE_FLIP_TERM,
		net->ucfIndices, net->dependCount, ucfi, ucfe);

    if( net->isRow){
	AceG.comdc.netLen.row += dlen;
	AceG.comdc.netCap.row += netCapCount;
	return( AceG.aparams.netLength.row * (double)dlen +
		AceG.aparams.netCapacity.row * (double)netCapCount );
    } else {
	AceG.comdc.netLen.col += dlen;
	AceG.comdc.netCap.col += netCapCount;
	return( AceG.aparams.netLength.col * (double)dlen +
		AceG.aparams.netCapacity.col * (double)netCapCount );
    }
}

/*
 * ace_unflip_term_net
 *
 * Undo a terminal net flip move.
 */
void
ace_unflip_term_net( net )
aceNet *net;
{
    aceRange range;
    aceMapEl **map;
    acePlane *plane;
    int i;

    map = AceG.matrix.map;

    /* remove net capacity that was added in the flip			*/
    if( net->termMask & ACE_TERM_LOWER_VE){
	range.lower = net->range.lower;
	range.upper = net->save1.lower - 1;

	/* toggle the VE bits						*/
	net->termMask &= ~ACE_TERM_LOWER_VE;
    } else {
	range.upper = net->range.upper;
	range.lower = net->save1.upper + 1;

	/* toggle the VE bits						*/
	net->termMask |= ACE_TERM_LOWER_VE;
    }
    if( net->isRow){
	for( i = range.lower; i <= range.upper; ++i){
	    --map[ net->rowcol ][ i ].netCap.row;
	}
    } else {
	for( i = range.lower; i <= range.upper; ++i){
	    --map[ i ][ net->rowcol ].netCap.col;
	}
    }

    /* add back net capacity of old virtual extension			*/
    if( net->termMask & ACE_TERM_LOWER_VE){
	range.lower = 0;
	range.upper = net->save1.lower - 1;
    } else {
	if( net->isRow){
	    range.upper = AceG.matrix.size.col - 1;
	} else {
	    range.upper = AceG.matrix.size.row - 1;
	}
	range.lower = net->save1.upper + 1;
    }
    if( net->isRow){
	for( i = range.lower; i <= range.upper; ++i){
	    ++map[ net->rowcol ][ i ].netCap.row;
	}
    } else {
	for( i = range.lower; i <= range.upper; ++i){
	    ++map[ i ][ net->rowcol ].netCap.col;
	}
    }

    /* calculate change in net length					*/
    if( net->isRow){
	plane = AceG.matrix.colPlanes;
    } else {
	plane = AceG.matrix.rowPlanes;
    }
    for( ; plane != NIL(acePlane); plane = plane->next){
	if( net->save1.lower > plane->range.upper){
	    if( net->termMask & ACE_TERM_LOWER_VE){
		/* now newly crossing this plane			*/
		++plane->netExitC;
	    } else {
		/* no longer cross this plane				*/
		--plane->netExitC;
	    }
	}

	if( net->save1.upper < plane->range.lower){
	    if( net->termMask & ACE_TERM_LOWER_VE){
		/* no longer cross this plane				*/
		--plane->netExitC;
	    } else {
		/* now newly crossing this plane			*/
		++plane->netExitC;
	    }
	} else if( net->save1.upper <= plane->range.upper){
	    /* upper end of cell range in this plane			*/
	    if( net->termMask & ACE_TERM_LOWER_VE){
		/* upper piece of this plane no longer in range		*/
		--plane->netExitC;
	    } else {
		/* upper piece of plane now added to the range		*/
		++plane->netExitC;
	    }
	} /* otherwise cellRange.upper is in a higher plane		*/
    }

    /* replace the net's range						*/
    net->range = net->save0;
}
