/*
 * genMove.c
 *
 * Generate the move that will produce the next proposed state.
 */

#include "aceInternal.h"

#define ACE_MOVE_GEN_MAX 1023
#define ACE_MOVE_GEN_MASK 0x3ff

/* move generation probability vector					*/
static int genVec[ ACE_MOVE_COUNT ];

/* array of funcions that calculate the probability of a move occuring	*/
static double (* probFnVec[ ACE_MOVE_COUNT ])();

/* array of funcions that assemble move parameters			*/
static int (* asmFnVec[ ACE_MOVE_COUNT ])();

/*
 * ace_gen_move
 *
 * Generate the next move.
 * This move will be used blindly by the routines that execute the move,
 * so the move assembler routines do the checking for legality and sanity
 * of the move.
 */
void
ace_gen_move()
{
    int i, moveIndex;

    AceG.move.mustAccept = FALSE;
    while( TRUE ){
	/* Pick the type of move to try					*/
	moveIndex = random() & ACE_MOVE_GEN_MASK;
	for( i = 0; i < ACE_MOVE_COUNT; ++i){
	    if( moveIndex >= genVec[ i ] ){
		AceG.move.index = i;
		break;
	    }
	}
	if( asmFnVec[ i ]() ){
	    /* if the asmbler routine succeded, return, else try again	*/
	    break;
	}
    }
}

/*
 * ace_move_asm_flip_term
 *
 * Assemble a flip-terminal-net move.
 */
int
ace_move_asm_flip_term()
{
    int i, genCount;

    if( AceG.matrix.termNetC < 1) return( FALSE );

    for( genCount = 0; genCount < 4; ++genCount){
	i = ace_random( 0, AceG.matrix.termNetC - 1);

	if( AceG.matrix.termNets[ i ]->termMask & ACE_TERM_TWO_STATE){
	    AceG.move.index = ACE_MOVE_FLIP_TERM;
	    AceG.move.net0 = AceG.matrix.termNets[ i ];
	    return( TRUE );
	}
    }
    return( FALSE );
}

/*
 * ace_move_asm_swap
 *
 * Assemble a swap net move.
 */
int
ace_move_asm_swap()
{
    int index, rangeLimit, tries;
    aceRowCol *rc, *rcArray;
    acePlane *plane;
    aceNet *net;
    aceRange limit;

    rangeLimit = (int) AceG.range;
    while( TRUE ){
	/* get first rowcol						*/
	ace_move_first_rci( &rcArray, &rc, &plane);

	net = rc->netlist;
	if( net == NIL(aceNet)){
	    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
		"ace_move_asm_swap: internal inconsistent state");
	}
	for( index = ace_random( 0, rc->count - 1); index > 0; --index){
	    net = net->next;
	}
	if( net == NIL(aceNet)){
	    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
		"ace_move_asm_swap: internal inconsistent state");
	}

	/* get the plane to find limits of rowcol selection		*/
	plane = rc->plane;

	/* find boundaries of move					*/
	limit.upper = MIN( net->rowcol + rangeLimit,
			   plane->range.upper + plane->emptyC);
	limit.lower = MAX( net->rowcol - rangeLimit, plane->range.lower);

	/* find a legal and useful destination rowcol			*/
	if( limit.lower == limit.upper){
	    /* can't swap this net (cuz 1 row in plane, etc)		*/
	    return( FALSE);
	}
	tries = 0;
	while( TRUE ){
	    index = ace_random( limit.lower, limit.upper);
	    if( ++tries > rangeLimit + 1) break;
	    if( index != net->rowcol){
		/* doint a swap, need rc with a net			*/
		rc = &( rcArray[ index ]);
		if( rc->count != 0){
		    break;
		}
	    }
	}
	if( tries > rangeLimit + 1) continue; /* try with new first rc	*/

	/* pick one net at this rc for the second net of the swap	*/
	AceG.move.index = ACE_MOVE_SWAP;
	AceG.move.net0 = net;
	AceG.move.displacement = index - net->rowcol;

	/* find swap net						*/
	net = rc->netlist;
	if( net == NIL(aceNet)){
	    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
		"ace_move_asm_swap: internal inconsistent state");
	}
	for( index = ace_random( 0, rc->count - 1); index > 0; --index){
	    net = net->next;
	}
	if( net == NIL(aceNet)){
	    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
		"ace_move_asm_swap: internal inconsistent state");
	}
	AceG.move.net1 = net;
	break;
    }
    return( TRUE );
}

/*
 * ace_move_asm_disp
 *
 * Assemble a displace-net move.
 */
int
ace_move_asm_disp()
{
    int index, rangeLimit;
    aceRowCol *rc, *rcArray;
    acePlane *plane;
    aceNet *net;
    aceRange limit;

    rangeLimit = (int) AceG.range;
    while( TRUE ){
	/* get first rowcol						*/
	ace_move_first_rci( &rcArray, &rc, &plane);

	net = rc->netlist;
	if( net == NIL(aceNet)){
	    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
		"ace_move_asm_disp: internal inconsistent state");
	}
	for( index = ace_random( 0, rc->count - 1); index > 0; --index){
	    net = net->next;
	}
	if( net == NIL(aceNet)){
	    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
		"ace_move_asm_disp: internal inconsistent state");
	}

	/* get the plane to find limits of rowcol selection		*/
	plane = rc->plane;

	/* find boundaries of move					*/
	limit.upper = MIN( net->rowcol + rangeLimit,
			   plane->range.upper + plane->emptyC);
	limit.lower = MAX( net->rowcol - rangeLimit, plane->range.lower);

	/* find a legal and useful destination rowcol			*/
	if( limit.lower == limit.upper){
	    /* can't displace this net (cuz 1 row plane, etc)		*/
	    return( FALSE);
	}
	while( TRUE ){
	    index = ace_random( limit.lower, limit.upper);
	    if( index != net->rowcol){
		break;
	    }
	}

	/* pick one net at this rc for the second net of the swap	*/
	AceG.move.index = ACE_MOVE_DISPLACE;
	AceG.move.net0 = net;
	AceG.move.displacement = index - net->rowcol;
	break;
    }
    return( TRUE );
}

/*
 * ace_move_first_rci
 *
 * Get the rowcol for a find either the net of a displacement or the
 * first net of a swap pair.  Each rowcol has an equal probability
 * of being picked, so moves should be spread out equally over the matrix.
 */
void
ace_move_first_rci( rcArray, rc, plane)
aceRowCol **rcArray;
aceRowCol **rc;
acePlane **plane;
{
    int index;

    do {
	index = ace_random( 0, AceG.rcC - 1);
	if( index < AceG.matrix.size.row){
	    /* use row RCs						*/
	    *rcArray = AceG.matrix.rowRCs;
	    *rc = &( AceG.matrix.rowRCs[ index ]);
	    *plane = AceG.matrix.rowPlanes;
	} else {
	    /* use col RCs						*/
	    index -= AceG.matrix.size.row;
	    *rcArray = AceG.matrix.colRCs;
	    *rc = &( AceG.matrix.colRCs[ index ]);
	    *plane = AceG.matrix.colPlanes;
	}
    } while( (*rc)->count == 0);
}

/*
 * ace_move_asm_rm_overlap
 *
 * Generate a remove-overlap move.
 * (Which is actually a pack-plane move right now).
 */
int
ace_move_asm_rm_overlap()
{
    int index;
    acePlane *p;

    AceG.move.index = ACE_MOVE_RM_OVERLAP;
    AceG.move.isRow = TRUE;
    AceG.move.mustAccept = TRUE;

    index = ace_random( 0, AceG.planeC - 1);
    p = AceG.matrix.rowPlanes;
    for( --index; index >= 0; --index){
	p = p->next;
	if( p == NIL(acePlane)) break;
    }
    if( p == NIL(acePlane)){
	AceG.move.isRow = FALSE;
	p = AceG.matrix.colPlanes;
	for( --index; ( index >= 0) && ( p != NIL(acePlane)); --index){
	    p = p->next;
	}
    }
    if( p == NIL(acePlane)){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
		"ace_move_asm_rm_overlap: plane is NIL");
    }
    AceG.move.plane = p;

    return( TRUE);
}

/*
 * ace_random
 *
 * return a psuedo-random integer in the given range.
 */
int
ace_random( lower, upper)
{
    if( lower == upper){
	return( lower);
    }

#ifdef ACE_PARANOID
    if( lower > upper){
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG, "ace_random: lower %d, upper %d",
		lower, upper);
    }
#endif /* ACE_PARANOID */

    return(( random() % ( upper - lower + 1)) + lower);
}

static double rmoStartTemp;

/*
 * ace_init_gen_move
 *
 * Initialize the structures used by the move generator.
 */
void
ace_init_gen_move()
{
    aceDefaults *def;
    int i;

    /* init the array of funcions that assemble move parameters		*/
    asmFnVec[ 0 ] = ace_move_asm_swap;
    asmFnVec[ 1 ] = ace_move_asm_disp;
    asmFnVec[ 2 ] = ace_move_asm_flip_term;
    asmFnVec[ 3 ] = ace_move_asm_rm_overlap;

    /* init the array of funcions generate probabilities of moves	*/
    def = ace_get_default_ptr();
    for( i = 0; i < ACE_MOVE_COUNT; ++i){
	probFnVec[ i ]  = def->probFnVec[ i ];
    }

    /* init the move propability vector					*/
    ace_update_gen_move();

    /* setup mechanism to vary probabilities of move gen as temp changes */
    rmoStartTemp = AceG.temp / 4.0;
}

/*
 * ace_update_gen_move
 *
 * Update the move generation probabilities.
 */
void
ace_update_gen_move()
{
    double sum = 0, prob;
    int i;

    for( i = 0; i < ACE_MOVE_COUNT; ++i){
	prob = probFnVec[ i ]( AceG.temp );
	sum += prob;
	if( prob > 0.0){
	    if( i != 0){
		genVec[ i ] = genVec[ i - 1 ] -
				ACE_RND((double)ACE_MOVE_GEN_MAX * prob);
	    } else {
		genVec[ 0 ] = ACE_RND((double)ACE_MOVE_GEN_MAX * ( 1.0 - prob));
	    }
	} else {
	    if( i != 0){
		genVec[ i ] = genVec[ i - 1 ];
	    } else {
		genVec[ i ] = ACE_MOVE_GEN_MAX + 1;
	    }
	}
    }
    /* avoid floating point off-by-one rounding quantization errors	*/
    genVec[ ACE_MOVE_COUNT - 1 ] = 0;

    if( fabs(sum - 1.0) > 0.005){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR, "%s (it was %.3f)",
	    "total move generation probability not close enough to 1.0", sum);
    }
}

/* used by the ace_move_prob_xxx routines				*/
#define TEMP_BRK_PNT_PCT	0.02

#define START_SWAP_PCT		0.667
#define START_DISP_PCT		0.167
#define START_FLIP_PCT		0.166
#define START_RMO_PCT		0.0

#define END_SWAP_PCT		0.80
#define END_DISP_PCT		0.19
#define END_FLIP_PCT		0.0
#define END_RMO_PCT		0.01

/*
 * ace_move_prob_swap
 *
 * For a given temperature, return the probability of a swap move being
 * generated.
 */
double
ace_move_prob_swap( temp )
double temp;
{
    static double tBreak = -1.0;

    if( tBreak < -0.5){
	tBreak = temp * TEMP_BRK_PNT_PCT;
    }
    if( temp > tBreak){
	return( START_SWAP_PCT);
    } else {
	return(END_SWAP_PCT + ( START_SWAP_PCT - END_SWAP_PCT) / tBreak * temp);
    }
}

/*
 * ace_move_prob_disp
 *
 * For a given temperature, return the probability of a displacement move being
 * generated.
 */
double
ace_move_prob_disp( temp )
double temp;
{
    static double tBreak = -1.0;

    if( tBreak < -0.5){
	tBreak = temp * TEMP_BRK_PNT_PCT;
    }
    if( temp > tBreak){
	return( START_DISP_PCT);
    } else {
	return(END_DISP_PCT + ( START_DISP_PCT - END_DISP_PCT) / tBreak * temp);
    }
}

/*
 * ace_move_prob_flip_term
 *
 * For a given temperature, return the probability of a flip term net move being
 * generated.
 */
double
ace_move_prob_flip_term( temp )
double temp;
{
    static double tBreak = -1.0;

    if( tBreak < -0.5){
	tBreak = temp * TEMP_BRK_PNT_PCT;
    }
    if( temp > tBreak){
	return( START_FLIP_PCT);
    } else {
	return(END_FLIP_PCT + ( START_FLIP_PCT - END_FLIP_PCT) / tBreak * temp);
    }
}

/*
 * ace_move_prob_rm_overlap
 *
 * For a given temperature, return the probability of a remove-overlap move
 * being generated.
 */
double
ace_move_prob_rm_overlap( temp )
double temp;
{
    static double tBreak = -1.0;

    if( tBreak < -0.5){
	tBreak = temp * TEMP_BRK_PNT_PCT;
    }
    if( temp > tBreak){
	return( START_RMO_PCT);
    } else {
	return(END_RMO_PCT + ( START_RMO_PCT - END_RMO_PCT) / tBreak * temp);
    }
}
