/*
 * id.c
 *
 * This file contains routines to implement aceIds.
 */

#include "aceInternal.h"

/*
 * ace_add_object
 *
 * This function "registers" the object and assigns a unique id to the object.
 *
 * In this first implementation, all objects are added to a list of objects
 * and the only use of the list is to free all objects when the free
 * matrix function is called. The unique aceId is the address to the
 * object structure.
 */
aceId
ace_add_object( obj )
aceObject *obj;
{
#ifdef ACE_PARANOID
    if( st_insert( AceG.idTbl, (char *) obj, (char *)obj) == 1){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR, "id already in table");
    }
#else
    obj->next = AceG.objList;
    AceG.objList = obj;
#endif

    return( (aceId) obj);
}

/*
 * ace_free_objects
 *
 * Step through the object list, and free each object.
 *
 * Currently, objects are masters, cells, nets, planes, and one matrix.
 */
void
ace_free_objects()
{
    aceObject *obj, *fobj;

#ifdef ACE_PARANOID
    st_generator *gen;

    st_foreach_item( AceG.idTbl, gen,((char **)&obj), ((char **)&fobj)){
	if( st_delete( AceG.idTbl, (char **)&obj, (char **)&fobj) != 1){
	    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
			"st package couldn't delete item from id table");
	} else {
	    ace_free_an_object( fobj);
	}
    }
    st_free_table( AceG.idTbl);
    AceG.idTbl = NIL(st_table);

#else

    obj = AceG.objList;
    while( obj != NIL(aceObject)){
	fobj = obj;
	obj = obj->next;
	ace_free_an_object( fobj);
    }

#endif
}

/* 
 * ace_free_an_object
 *
 * Free the appropriate fields of each piece of the obj->object,
 * then free obj->object, the free obj.
 */
void
ace_free_an_object( obj)
aceObject *obj;
{
    switch( obj->type){
	case ACE_OBJ_MASTER:
	    FREE( ((aceMaster *)obj)->name);
	    FREE( ((aceMaster *)obj)->path);
	    break;
	case ACE_OBJ_NET:
	    if( ((aceNet *)obj)->cellCount > 0){
		FREE( ((aceNet *)obj)->cells);
	    }
	    if( ((aceNet *)obj)->dependCount > 0){
		FREE( ((aceNet *)obj)->ucfIndices);
	    }
	    break;
	case ACE_OBJ_CELL:
	case ACE_OBJ_PLANE:
	    /* no fields to free					*/
	    break;
	case ACE_OBJ_ROWCOL: /* malloced as a large array, don't want to free */
	case ACE_OBJ_MATRIX: /* not actually malloced			*/
	    return;
	    break;
	case ACE_OBJ_UNKNOWN:
	default:
	    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
		    "unknown object type 0x%x", obj->type);
	    break;
    }
    FREE( obj);
}

/*
 * ace_get_by_id
 *
 * If we're doing paranoid ids, aceGetById and the ace_get_object macro
 * call this routine to check the id.
 */
/*ARGSUSED*/
aceObject *
ace_get_by_id( id )
aceId id;
{
#ifdef ACE_PARANOID
    aceObject *obj;

    /* check to see that its been added					*/
    if(( st_lookup( AceG.idTbl, (char *)id, (char **)&obj) == 0)
	    || ( id != (aceId) obj)){
	obj = NIL(aceObject);
    }
    if( obj == NIL(aceObject)){
	errRaise( ACE_PKG_NAME, ACE_BAD_ID, "corrupted id");
    }
    return( obj );
#endif
}
