/*
 * rmOverlap.c
 *
 * Routines here implement the occasional move used to periodically
 * reduce/remove net ovelap in the matrix.
 */

#include "aceInternal.h"

static int dLen;
static int dUBias;
static int dPNetCap;
static short pmShifted; /* set if pack_matrix shifted something		*/

static aceMove saveMove;
static double delC;

static int ace_find_empty();
static int saveEI, saveRci;

/*
 * ace_rm_overlap_move
 *
 * Despite the name, this routine actually gets rid of empty rowcols
 * in the matrix interior, pushing them out to the outer edge.
 */
double
ace_rm_overlap_move( plane )
acePlane *plane;
{
    aceCoord loc;
#ifndef ACE_NO_UCF
    aceNet *tnet;
    aceUCFEvent *ucfe;
    int ucfi;
    int i;
#endif /* ACE_NO_UCF */

    dLen = dUBias = dPNetCap = 0;
    pmShifted = FALSE;
    delC = 0.0;

    /* move empty RCs to outside and resize the matrix			*/
    if( !ace_is_overlap( plane->isRow, plane->range.lower,
			plane->range.upper, &loc)){
	ace_rmo_pack_plane( plane, plane->isRow);
	if( pmShifted){
	    ace_set_rowcols( plane->isRow, 0, plane->isRow ?
		    AceG.matrix.size.row - 1 : AceG.matrix.size.col - 1);
	    ace_shrink_nets( !plane->isRow, &dLen);
	}
	/* move empty to dead						*/
	if( plane->emptyC > (int)AceG.range){
	    plane->deadC++;
	    plane->emptyC--;
	}
    } else {
	ace_rm_overlap(plane, plane->isRow, &loc);
    }

#ifndef ACE_NO_UCF
    /* only thing missing is a good way to do DLEN events for termNets
     * with an upper extention.
     * First try, if any dUBias, gen event for every termNet.
     */
    if( pmShifted ){
	for( i = AceG.matrix.termNetC - 1; i >= 0; --i){
	    /* if same orien, generate event				*/
	    if(( AceG.matrix.termNets[ i ]->isRow && plane->isRow) ||
	       ( !AceG.matrix.termNets[ i ]->isRow && !plane->isRow)){
		tnet = AceG.matrix.termNets[ i ];
		ACE_UCF_EVENT( ACE_UCFE_NET_DLEN, tnet->ucfIndices,
			    tnet->dependCount, ucfi, ucfe);
	    }
	}
    }
#endif /* ACE_NO_UCF */

    if( plane->isRow ){
	AceG.comdc.usedBias.row += dUBias;
	AceG.comdc.netLen.col += dLen;
	AceG.comdc.netCap.col += dPNetCap;
	return( delC +
		AceG.aparams.usedBias.row * (double)dUBias +
		AceG.aparams.netLength.col * (double)dLen +
		AceG.aparams.netCapacity.col * (double)dPNetCap );
    } else {
	AceG.comdc.usedBias.col += dUBias;
	AceG.comdc.netLen.row += dLen;
	AceG.comdc.netCap.row += dPNetCap;
	return( delC +
		AceG.aparams.usedBias.col * (double)dUBias +
		AceG.aparams.netLength.row * (double)dLen +
		AceG.aparams.netCapacity.row * (double)dPNetCap );
    }
}

/*
 * ace_rm_overlap
 *
 * Remove an instance of overlap.
 * If there are empty or dead rowcols, move an empty rowcol next to the
 * rowcol with overlap, otherwise, force a displacement of one net involved
 * in the overlap.
 */
void
ace_rm_overlap( plane, isRow, loc)
acePlane *plane;
short isRow;
aceCoord *loc;
{
    aceNet *movenet, *net;
    aceRowCol *rca;
    int rci, fi;
    int index, emptyInd;
#ifndef ACE_NO_UCF
    int ucfi;
    aceUCFEvent *ucfe;
#endif /* ACE_NO_UCF */

    if( isRow){
	rca = AceG.matrix.rowRCs;
	rci = loc->row;
	fi = loc->col;
    } else {
	rca = AceG.matrix.colRCs;
	rci = loc->col;
	fi = loc->row;
    }
    if( !ace_find_empty( plane, isRow, rci, &emptyInd)){
	/* there are no empty rowcol in the plane			*/
	movenet = NIL(aceNet);
	for( net = rca[ rci ].netlist; net != NIL(aceNet); net = net->next){
	if(( net->range.upper >= fi) && ( net->range.lower <= fi)){
		movenet = net;
		break;
	    }
	}
	if( movenet == NIL(aceNet)){
	    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
		    "ace_rm_overlap: cannot find net at overlap location");
	}
	saveMove = AceG.move;
	AceG.move.net0 = movenet;
	if( movenet->rowcol == plane->range.lower){
	    AceG.move.displacement = 1;
	} else if( movenet->rowcol == plane->range.upper){
	    AceG.move.displacement = -1;
	} else {
	    AceG.move.displacement = (random() & 0x1) ? 1 : -1;
	}
	AceG.move.isRow = isRow;
	delC = ace_disp_net_move();
	AceG.move = saveMove;
    } else {
	delC = 0.0;
	pmShifted = TRUE;
	/* empty out an RC and put the net there			*/
	if( emptyInd == -1){
	    if( plane->emptyC == 0){
		/* move a dead into plane range				*/
		plane->deadC--;
	    } else {
		/* move a empty into plane range			*/
		plane->emptyC--;
	    }
	    emptyInd = ++plane->range.upper;
	    rca[ emptyInd ].used = TRUE;

	    /* decrement the empty rc count				*/
	    ++dUBias;
	    dLen += plane->netExitC;
	} else {
	    /* count up net cap being overwritten			*/
	    if( isRow){
		dPNetCap -= ace_col_nc_in_row( emptyInd );
	    } else {
		dPNetCap -= ace_row_nc_in_col( emptyInd );
	    }
	}
	saveEI = emptyInd;
	saveRci = rci;
	if( abs( emptyInd - rci ) == 1){
	    /* don't do anything cuz empty is already next to overlap	*/
	    dPNetCap = 0; /* could have been set above			*/
	    return;
	} else if( rci < emptyInd){
	    for( index = emptyInd - 1; index > rci; --index){
		rca[ index + 1 ].count = rca[ index ].count;
		rca[ index + 1 ].termNetC = rca[ index ].termNetC;
		rca[ index + 1 ].netlist = rca[ index ].netlist;
#ifndef ACE_NO_UCF
		/* do ucf events					*/
		for( net = rca[ index ].netlist; net != NIL(aceNet);
			net = net->next){
		    ACE_UCF_EVENT( ACE_UCFE_NET_MOVE, net->ucfIndices,
				net->dependCount, ucfi, ucfe);
		}
#endif /* ACE_NO_UCF */
		/* update the map					*/
		if( isRow ){
		    ace_mv_map_row( index, index + 1);
		} else {
		    ace_mv_map_col( index, index + 1);
		}
	    }
	    /* make rci + 1 look empty					*/
	    rca[ rci + 1 ].count = 0;
	    rca[ rci + 1 ].termNetC = 0;
	    rca[ rci + 1 ].netlist = NIL(aceNet);
	    if( isRow ){
		ace_zero_map_row( rci + 1);
	    } else {
		ace_zero_map_col( rci + 1);
	    }
	} else {
	    for( index = emptyInd + 1; index < rci; ++index){
		rca[ index - 1 ].count = rca[ index ].count;
		rca[ index - 1 ].termNetC = rca[ index ].termNetC;
		rca[ index - 1 ].netlist = rca[ index ].netlist;
		/* update the map					*/
		if( isRow ){
		    ace_mv_map_row( index, index - 1);
		} else {
		    ace_mv_map_col( index, index - 1);
		}
	    }
	    /* make rci + 1 look empty					*/
	    rca[ rci - 1 ].count = 0;
	    rca[ rci - 1 ].termNetC = 0;
	    rca[ rci - 1 ].netlist = NIL(aceNet);
	    if( isRow ){
		ace_zero_map_row( rci - 1);
	    } else {
		ace_zero_map_col( rci - 1);
	    }
	}

	/* update the locations of nets and cells			*/
	if( rci < emptyInd){
	    ace_set_rowcols( isRow, rci + 2, emptyInd);
	} else {
	    ace_set_rowcols( isRow, emptyInd, rci - 2);
	}
	/* update perpendicular ranges					*/
	/*** sp. case: nets ending betwn rci & plane.r.up are effected?	*/
	ace_shrink_nets( !isRow, &dLen);
	if( rci < emptyInd){
	    if( isRow){
		dPNetCap += ace_col_nc_in_row( rci + 1);
	    } else {
		dPNetCap += ace_row_nc_in_col( rci + 1);
	    }
	} else {
	    if( isRow){
		dPNetCap += ace_col_nc_in_row( rci - 1);
	    } else {
		dPNetCap += ace_row_nc_in_col( rci - 1);
	    }
	}
	/*** could displace a net in the overlap here			*/
    }
}

/*
 * ace_rmo_pack_plane
 *
 * Remove all internal empty RCs from this plane and put them out into
 * empty space.  Then, if there's enough empty space, move some empty
 * space into dead space.
 */
void
ace_rmo_pack_plane( p, isRow)
acePlane *p;		/* plane to pack				*/
short isRow;		/* bool, is this horizontal orien or not	*/
{
    aceRowCol *rca; /* the RC array					*/
    int ei, fi; /* empty index, full index				*/
    int rci; /* rowcol index variable while eliminating overlap		*/

    rca = isRow ? AceG.matrix.rowRCs : AceG.matrix.colRCs;
    /* find first empty rowcol in this plane				*/
    for( ei = p->range.lower; ( ei <= p->range.upper) && rca[ ei ].count; ++ei);

    if( ei <= p->range.upper){
	if( isRow){
	    dPNetCap -= ace_col_nc_in_row( ei );
	} else {
	    dPNetCap -= ace_row_nc_in_col( ei );
	}
    }
    /* shift everything towards the lower side				*/ 
    fi = ei;
    while( TRUE ){
	/* locate next non-empty rowcol					*/
	for( ++fi; fi <= p->range.upper; ++fi){
	    if( rca[ fi ].count){
		break;
	    } else {
		if( isRow ){
		    dPNetCap -= ace_col_nc_in_row( fi );
		} else {
		    dPNetCap -= ace_row_nc_in_col( fi );
		}
	    }
	}

	/* if there's no more, done					*/
	if( fi > p->range.upper){
	    break;
	}

	/* copy fi to ei						*/
	rca[ ei ].count = rca[ fi ].count;
	rca[ ei ].termNetC = rca[ fi ].termNetC;
	rca[ ei ].netlist = rca[ fi ].netlist;

	/* If ei has any perpendicular net overlap violations, eliminate
	 * them from the count, then copy map row or col fi to ei.
	 */
	if( isRow ){
	    ace_mv_map_row( fi, ei);
	} else {
	    ace_mv_map_col( fi, ei);
	}

	/* make fi look empty						*/
	rca[ fi ].count = 0;
	rca[ fi ].termNetC = 0;
	rca[ fi ].netlist = NIL(aceNet);

	/* find next empty rc						*/
	for( ++ei; rca[ ei ].count != 0; ++ei) ;
    }

    /* ei points one past the rightmost end of a group of full RCs	*/
    if( ei <= p->range.upper){
	pmShifted = TRUE;
	for( rci = p->range.upper; rci >= ei; --rci){
	    rca[ rci ].used = FALSE;
	    if( isRow ){
		ace_zero_map_row( rci );
	    } else {
		ace_zero_map_col( rci );
	    }
	}
	dUBias -= p->range.upper - ei + 1;
	dLen -= (p->range.upper - ei + 1) * p->netExitC;
	p->emptyC += p->range.upper - ei + 1;
	p->range.upper -= p->range.upper - ei + 1;
    }

    /*** resize plane (emptyC stuff)
     * Can't put any into deadC with this version cuz there's no rmo move
     * that can move things from dead to empty in case its necessary to
     * get rid of overlap.
    if( p->emptyC > 3){
	++p->deadC;
	--p->emptyC;
    }
    ***/
}

/*
 * ace_col_nc_in_row
 *
 * Sum up the net overlaps of vertical nets in this row.
 */
int
ace_col_nc_in_row( rc )
int rc;
{
    int i, diff, sum = 0;

    for( i = AceG.matrix.size.col - 1; i >= 0; --i){
	if( !AceG.matrix.colRCs[ i ].used) continue;
	if(( diff = AceG.matrix.map[ rc ][ i ].netCap.col -
		AceG.matrix.colRCs[ i ].plane->netCap) > 0){
	    sum += diff;
	}
    }
    return( sum );
}

/*
 * ace_row_nc_in_col
 *
 * Sum up the net overlaps of vertical nets in this row.
 */
int
ace_row_nc_in_col( rc )
int rc;
{
    int i, diff, sum = 0;

    for( i = AceG.matrix.size.row - 1; i >= 0; --i){
	if( !AceG.matrix.rowRCs[ i ].used) continue;
	if(( diff = AceG.matrix.map[ i ][ rc ].netCap.row -
		AceG.matrix.rowRCs[ i ].plane->netCap) > 0){
	    sum += diff;
	}
    }
    return( sum );
}

/*
 * ace_mv_map_row
 *
 * Move a row of the map from src to dst.
 */
void
ace_mv_map_row( src, dst)
int src, dst;
{
    int i;

    for( i = AceG.matrix.size.col - 1; i >= 0; --i){
	AceG.matrix.map[ dst ][ i ].cellCap =
			AceG.matrix.map[ src ][ i ].cellCap;
	AceG.matrix.map[ dst ][ i ].netCap.row =
			AceG.matrix.map[ src ][ i ].netCap.row;
    }
}

/*
 * ace_mv_map_col
 *
 * Move a col of the map from src to dst.
 */
void
ace_mv_map_col( src, dst)
int src, dst;
{
    int i;

    for( i = AceG.matrix.size.row - 1; i >= 0; --i){
	AceG.matrix.map[ i ][ dst ].cellCap =
			AceG.matrix.map[ i ][ src ].cellCap;
	AceG.matrix.map[ i ][ dst ].netCap.col =
			AceG.matrix.map[ i ][ src ].netCap.col;
    }
}

/*
 * ace_zero_map_row
 *
 * Zero out the row entries in the given row.
 */
void
ace_zero_map_row( rci )
int rci;
{
    int i;

    for( i = AceG.matrix.size.col - 1; i >= 0; --i){
	AceG.matrix.map[ rci ][ i ].cellCap.row = 0;
	AceG.matrix.map[ rci ][ i ].cellCap.col = 0;
	AceG.matrix.map[ rci ][ i ].netCap.row = 0;
    }
}

/*
 * ace_zero_map_col
 *
 * Zero out the col entries in the given col.
 */
void
ace_zero_map_col( rci )
int rci;
{
    int i;

    for( i = AceG.matrix.size.row - 1; i >= 0; --i){
	AceG.matrix.map[ i ][ rci ].cellCap.row = 0;
	AceG.matrix.map[ i ][ rci ].cellCap.col = 0;
	AceG.matrix.map[ i ][ rci ].netCap.col = 0;
    }
}

/*
 * ace_set_rowcols
 *
 * Set the net rowcol field and cell loc fields to be the same as the
 * rowcol that contains the net in its net list.
 */
void
ace_set_rowcols( isRow, lower, upper)
short isRow;
int lower, upper;
{
    int rci, ci; /* rowcol index, cell index				*/
    aceNet *net;
#ifndef ACE_NO_UCF
    int ucfi;
    aceUCFEvent *ucfe;
#endif /* ACE_NO_UCF */

    if( isRow ){
	/* for each horizontal RC					*/
	for( rci = lower; rci <= upper; ++rci){
	    /* for each net in the RC					*/
	    for( net = AceG.matrix.rowRCs[ rci ].netlist; net != NIL(aceNet);
		    net = net->next){
		/* assign the rowcol index to the net and its cells	*/
#ifndef ACE_NO_UCF
		if( net->rowcol != rci ){
		    ACE_UCF_EVENT( ACE_UCFE_NET_MOVE, net->ucfIndices,
			    net->dependCount, ucfi, ucfe);
		}
#endif /* ACE_NO_UCF */
		net->rowcol = rci;
		for( ci = net->cellCount - 1; ci >= 0; --ci){
		    net->cells[ ci ]->loc.row = rci;
		}
	    }
	}
    } else {
	/* for each vertical RC						*/
	for( rci = lower; rci <= upper; ++rci){
	    /* for each net in the RC					*/
	    for( net = AceG.matrix.colRCs[ rci ].netlist; net != NIL(aceNet);
		    net = net->next){
		/* assign the rowcol index to the net and its cells	*/
#ifndef ACE_NO_UCF
		if( net->rowcol != rci ){
		    ACE_UCF_EVENT( ACE_UCFE_NET_MOVE, net->ucfIndices,
			    net->dependCount, ucfi, ucfe);
		}
#endif /* ACE_NO_UCF */
		net->rowcol = rci;
		for( ci = net->cellCount - 1; ci >= 0; --ci){
		    net->cells[ ci ]->loc.col = rci;
		}
	    }
	}
    }
}

/*
 * ace_is_overlap
 *
 * Retrun true if any overlap (cell or net) is found in the range
 * and return the location of the overlap.
 */
int
ace_is_overlap( isRow, lower, upper, loc )
short isRow;
int upper, lower;
aceCoord *loc;
{
    int row, col;

    if( isRow ){
	for( row = lower; row < upper; ++row){
	    for( col = AceG.matrix.size.col - 1; col >= 0; --col){
		if( AceG.matrix.map[ row ][ col ].netCap.row >
		    AceG.matrix.rowRCs[ row ].plane->netCap){
		    loc->row = row;
		    loc->col = col;
		    return( TRUE );
		}
		if( AceG.matrix.map[ row ][ col ].cellCap.row >
		    AceG.matrix.rowRCs[ row ].plane->cellCap){
		    loc->row = row;
		    loc->col = col;
		    return( TRUE );
		}
	    }
	}
    } else {
	for( row = AceG.matrix.size.row - 1; row >= 0; --row){
	    for( col = lower; col < upper; ++col){
		if( AceG.matrix.map[ row ][ col ].netCap.col >
		    AceG.matrix.colRCs[ col ].plane->netCap){
		    loc->row = row;
		    loc->col = col;
		    return( TRUE );
		}
		if( AceG.matrix.map[ row ][ col ].cellCap.col >
		    AceG.matrix.colRCs[ col ].plane->cellCap){
		    loc->row = row;
		    loc->col = col;
		    return( TRUE );
		}
	    }
	}
    }
    return( FALSE );
}

/*
 * ace_find_empty
 *
 * Find the best empty RC in the plane for shifting over to near an overlap.
 * Check lower direction first cuz its better to shift in that direction.
 */
static int
ace_find_empty( plane, isRow, rci, index)
acePlane *plane;
short isRow;
int rci;
int *index;
{
    int i;

    if( isRow){
	for( i = rci - 1; i >= plane->range.lower; --i){
	    if( AceG.matrix.rowRCs[ i ].count == 0){
		break;
	    }
	}
    } else {
	for( i = rci - 1; i >= plane->range.lower; --i){
	    if( AceG.matrix.colRCs[ i ].count == 0){
		break;
	    }
	}
    }
    if( i >= plane->range.lower){
	*index = i;
	return( TRUE );
    }
    if( isRow){
	for( i = rci + 1; i <= plane->range.upper; ++i){
	    if( AceG.matrix.rowRCs[ i ].count == 0){
		break;
	    }
	}
    } else {
	for( i = rci + 1; i <= plane->range.upper; ++i){
	    if( AceG.matrix.colRCs[ i ].count == 0){
		break;
	    }
	}
    }
    if( i <= plane->range.upper){
	*index = i;
	return( TRUE );
    }
    *index = -1;
    return( plane->emptyC + plane->deadC);
}
