/*
 * schedule.c
 *
 * These contain the default set of annealing temperature scheduling routines.
 * aceInternal.h IS NOT included intentionally, because the the interface
 * to these routines should be such that the application can define its own.
 *
 * The standard routines are:
 *	void ace_init_stop_inner( int rows, int cols, int planes, int nets,
 *		int cells, double temp, double cost, int fastFlag){}
 *	int ace_stop_inner( int gen, int accept, int fastFlag){}
 *	int ace_stop_outer( double temp, double cost, int fastFlag){}
 *	double ace_init_temperature( double cost, int fastFlag){}
 *	double ace_update_temperature( double temperature, double cost,
 *					int fastFlag){}
 */

#include "port.h"
#include "utility.h"
#include "ace.h"

static void ace_init_stop_inner();
static int ace_stop_inner();
static int ace_stop_outer();
static double ace_init_temp();
static double ace_update_temp();

/*
 * ace_setup_scheduler
 *
 * Add the default routines to the defaults structure during the first call
 * to ace.
 */
void
ace_setup_scheduler( def )
aceDefaults *def;
{
    def->initStopInner = ace_init_stop_inner;
    def->stopInner = ace_stop_inner;
    def->stopOuter = ace_stop_outer;
    def->initTemp = ace_init_temp;
    def->updateTemp = ace_update_temp;
}

/* annealing state info							*/
static double tempBreakPoint;
static int maxAccept;
static int maxStates;

/*
 * ace_init_stop_inner
 *
 * Initialize the inner loop stopping parameters.
 */
/*ARGSUSED*/
static void
ace_init_stop_inner( rows, cols, planes, nets, cells, temp, cost, fast)
int rows, cols, planes, nets, cells;
double temp, cost;
int fast;
{
    if( fast ){
	maxStates = cells * 5;
	maxAccept = cells * 2;
    } else {
	maxStates = cells * 10;
	maxAccept = cells * 4;
    }
}

/*
 * ace_stop_inner
 *
 * Criterion for stopping the inner loop of the annealing engine
 * and moving the the next temperature state.
 */
/*ARGSUSED*/
static int
ace_stop_inner( gen, accept, fast)
int gen, accept, fast;
{
    if(( accept >= maxAccept) || ( gen >= maxStates)){
	return( 1 );
    }
    return( 0 );
}

/*
 * ace_stop_outer
 *
 * Criterion for stopping the annealing engine.
 */
/*ARGSUSED*/
static int
ace_stop_outer( temp, cost, fast)
double temp, cost;
int fast;
{
    static double cost1 = -1.0;
    static double cost2 = -2.0;
    static double cost3 = -3.0;

    if( fast ){
	/* In fast mode, cost2 and cost3 are either 1 or negetive, 1 meaning
	 * that that state's cost was within .5% of the next states cost.
	 */
	if( fabs((cost - cost1) / cost1) < 0.005){
	    if(( cost2 >= 0.0) && (cost3 >= 0.0)){
		return( 1 );
	    }
	    cost3 = cost2;
	    cost2 = 1.0;
	} else {
	    cost3 = cost2;
	    cost2 = -1.0;
	}
	cost1 = cost;
    } else {
	if(( cost == 0.0) ||
	   ( temp < 0.05) ||
	   (( cost == cost1) && ( cost == cost2) && ( cost == cost3))){
	    return( 1 );
	}
	cost3 = cost2;
	cost2 = cost1;
	cost1 = cost;
    }
    return( 0 );
}

/*
 * ace_update_temp
 *
 * Calculate next temp.
 */
/*ARGSUSED*/
static double
ace_update_temp( temp, cost, fast)
double temp, cost;
int fast;
{
    if( temp > tempBreakPoint){
	temp *= 0.8;
    } else {
	temp *= 0.9;
    }
    return( temp );
}

/*
 * ace_init_temp
 *
 * calulate some initial temperature
 */
/*ARGSUSED*/
static double
ace_init_temp( cost, fast )
double cost;
int fast;
{
    double temp;

    temp = cost;
    if( fast ){
	tempBreakPoint = temp * 0.005;
    } else {
	tempBreakPoint = temp * 0.01;
    }
    return( temp );
}
