/*
 * ucf.c
 *
 * UCF internal support routines and data structures.
 */

#include "aceInternal.h"

/* global UCF block							*/
static int AceUCFSize = 0;

/*
 * ace_ucf_check_cnstr
 *
 * Call all the registered checking functions, and do something if the
 * constraints are found to be unsatisfied.
 */
void
ace_ucf_check_cnstr()
{
    int i;

    for( i = AceG.ucfCount - 1; i >= 0; --i){
	if( AceG.ucfArray[ i ].check_fn != (int (*)()) 0 ){
	    if( AceG.ucfArray[ i ].check_fn( AceG.ucfArray[ i ].argCount,
					     AceG.ucfArray[ i ].argObjs )){
		/*** debug */
		ace_dump_net_cap_map();
		errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
			"ace_ucf_check_cnstr: %s %d %s",
			"User constraint function", i, "was not satisfied");
	    }
	}
    }
}

/*
 * ace_add_ucf
 *
 * Add user specified function to the UCF table. This may realloc the
 * UCF internal array periodically to keep it big enough.
 */
int
ace_add_ucf( cnstr_fn, check_fn, emask, argCount, argIds)
double (*cnstr_fn)();
int (*check_fn)();
unsigned emask;
int argCount;
aceId argIds[];
{
    int i;

    if( AceUCFSize == 0){
	AceG.ucfCount = 1;
	AceUCFSize = 8;
	if(( AceG.ucfArray = ALLOC( aceUCF, AceUCFSize)) == NIL(aceUCF)){
	    errRaise( ACE_PKG_NAME, ACE_NO_MEM, "no memory to alloc UCF array");
	}
    } else if( ++AceG.ucfCount > AceUCFSize){
	AceUCFSize <<= 1;
	if(( AceG.ucfArray = REALLOC( aceUCF, AceG.ucfArray, AceUCFSize))
		== NIL(aceUCF)){
	    errRaise( ACE_PKG_NAME, ACE_NO_MEM,
			"no memory to realloc UCF array");
	}
    }

    /* fill in the new params into UCFCount				*/
    (void)memset( (char *)&(AceG.ucfArray[ AceG.ucfCount - 1 ]), 0,
		sizeof(aceUCF));
    AceG.ucfArray[ AceG.ucfCount - 1 ].cnstr_fn = cnstr_fn;
    AceG.ucfArray[ AceG.ucfCount - 1 ].check_fn = check_fn;
    AceG.ucfArray[ AceG.ucfCount - 1 ].eventMask = emask;
    AceG.ucfArray[ AceG.ucfCount - 1 ].argCount = argCount;
    AceG.ucfArray[ AceG.ucfCount - 1 ].argObjs = ALLOC( aceObject *, argCount);
    for( i = 0; i < argCount; ++i){
	AceG.ucfArray[ AceG.ucfCount - 1 ].argObjs[i] =
						ace_get_object( argIds[i]);
    }

    return( AceG.ucfCount - 1);
}

/*
 * ace_make_depend
 *
 * Semantically similar to a make dependancy, this functions stores
 * the elements that, when moved, trigger a call this UCF instance.
 * For fast access during annealing (ie to evaualte a proposed move),
 * the the dependancy is stored with the element, not with the UCF instance.
 * At this point, count is known to be non-negitive and depIds is non-NIL.
 *
 * Note: it is important to store the index of the UCF instance, not the
 *   address, because the UCF array can be realloced.
 */
void
ace_make_depend( count, depIds, ucfIndex)
int count;
aceId *depIds;
int ucfIndex;
{
    int i;
    aceObject *obj;
    aceNet *net;

    for( i = 0; i < count; ++i){
	obj = ace_get_object( depIds[ i ]);
	switch( obj->type){
	    case ACE_OBJ_NET:
		net = (aceNet *)obj;
		/* if no dependancies yet, alloc small array of them	*/
		if( net->dependCount == 0){
		    net->ucfIndices = ace_new_ucf_indices();
		} else {
		    /* else, call realloc to make sure array is big enough */
		    if(( net->ucfIndices =
			   REALLOC( int, net->ucfIndices, net->dependCount + 1))
			   == NIL(int)){
			errRaise( ACE_PKG_NAME, ACE_NO_MEM,
				"no memory to realloc net ucfIndices array");
		    }
		}
		/* mark this UCF index as depending on this net		*/
		net->ucfIndices[ (net->dependCount)++ ] = ucfIndex;
		break;
	    default:
		errRaise( ACE_PKG_NAME, ACE_BAD_TYPE,
			"not supporting constraint dependencies on %s objects",
			ace_object_type_string( obj->type));
		break;
	}
    }
}

/*
 * ace_ucf_free_array
 *
 * Free the argument storage memory and the array itself.
 */
void
ace_ucf_free_array()
{
    int i;

    for( i = AceG.ucfCount - 1; i >= 0; --i){
	if( AceG.ucfArray[ i ].argCount > 0){
	    FREE( AceG.ucfArray[i].argObjs);
	}
    }
    if( AceG.ucfCount > 0){
	FREE( AceG.ucfArray);
    }
}

/*
 * ace_ucf_eputq
 *
 * Add a UCF event instance to the event queue.
 */
void
ace_ucf_eputq( event, q)
aceUCFEvent *event, **q;
{
    event->next = *q;
    *q = event;
}

/*
 * ace_ucf_egetq
 *
 * Get a UCF event instance from the event queue.
 */
aceUCFEvent *
ace_ucf_egetq( q)
aceUCFEvent **q;
{
    aceUCFEvent *e;

    if( *q == NIL(aceUCFEvent)){
	return( NIL(aceUCFEvent));
    } else {
	e = *q;
	*q = e->next;
	return( e );
    }
}

static aceUCFEvent *AceUCFFreeQ = NIL(aceUCFEvent);

/*
 * ace_ucf_new_e
 *
 * If there's a leftover event, return it, otherwise, malloc a new one.
 */
aceUCFEvent *
ace_ucf_new_e()
{
    aceUCFEvent *e;

    if( AceUCFFreeQ == NIL(aceUCFEvent)){
	e = ALLOC( aceUCFEvent, 1);
	if( e == NIL(aceUCFEvent)){
	    errRaise( ACE_PKG_NAME, ACE_NO_MEM,
		"no memory for new UCF event record");
	}
    } else {
	e = AceUCFFreeQ;
	AceUCFFreeQ = AceUCFFreeQ->next;
    }
    return( e );
}

/*
 * ace_ucf_free_e
 *
 * `Free' the event, which means add it to the free queue.
 */
void
ace_ucf_free_e( e )
aceUCFEvent *e;
{
    e->next = AceUCFFreeQ;
    AceUCFFreeQ = e;
}

/*
 * ace_ucf_efreeq
 *
 * Do a real free on all the events structs in the ucf free queue.
 */
void
ace_ucf_efreeq()
{
    aceUCFEvent *be_free;

    while( AceUCFFreeQ != NIL(aceUCFEvent)){
	be_free = AceUCFFreeQ;
	AceUCFFreeQ = AceUCFFreeQ->next;
	FREE( be_free);
    }
    AceUCFFreeQ = NIL(aceUCFEvent);
}
