/*
 *  FILE: checkArea.c
 *
 *  DESCRIPTION: routines realted to area of a cell,
 *		whether it is soft or hard
 *
 *
 *  CREATION DATE: xx/xx/xx		AUTHOR:
 *
 *  FUNCTIONS:
 *  alba_checkAreaInformation(InstanceP)
 *  SoftCellCheck(CellInterfaceP)
 *  FloorplanBagUpdate(SoftCellP, w, h, minAR, maxAR)
 *  SoftCellCheckConsistency(SoftCell)
 */
#include "copyright.h"
#include "port.h"
#include "alba.h"
#include "vov.h"

#define MIN_AR MinHeightOverWidth.contents.prop.value.real
#define MAX_AR MaxHeightOverWidth.contents.prop.value.real


/*
 *  alba_checkAreaInformation(Instance)
 *  This routine checks that the interface of the master exists
 *  and that it contains information about the area of the cell
 *  If the interface is not found it is created  using vulcan
 *  If no area information is found, the user
 *  will be required to provide the area estimation interactively
 */
void
alba_checkAreaInformation(InstanceP)
     octObject *InstanceP;
{
    octObject 	MasterInterface;
    double 	inputDouble();
    char		command[80];
    int		status;

    if (verbose) printf("Checking area information of %s:%s\n",
			InstanceP->contents.instance.master,
			InstanceP->contents.instance.view);
    MasterInterface.type = OCT_FACET;
    MasterInterface.contents.facet.cell = InstanceP->contents.instance.master;
    MasterInterface.contents.facet.view = InstanceP->contents.instance.view;
    MasterInterface.contents.facet.version = InstanceP->contents.instance.version;
    MasterInterface.contents.facet.facet = "interface";
    MasterInterface.contents.facet.mode = "r";

    if (octOpenFacet(&MasterInterface) < OCT_OK) {
	(void) sprintf(buf,"the interface of %s:%s doesn't exist",
		InstanceP->contents.instance.master,
		InstanceP->contents.instance.view);
	Warning(buf);
	/*
	 *  try to use vulcan on the cell
	 */
	(void) sprintf(command,"vulcan  %s:%s",
		MasterInterface.contents.facet.cell,MasterInterface.contents.facet.view);
	status = system (command);
	if (status == 127 ) FatalError("Vulcan failed to do the interface");
	/*
	 *  try again to read the MasterInterface
	 */
	if (octOpenFacet (&MasterInterface) < OCT_OK)  FatalError("giving up\n");
    }
    VOVinputFacet( &MasterInterface );
    /*
     *  Now the MasterInterface is open, check its geometry
     *  Check first if it is a Soft Cell
     */
    (void) SoftCellCheck(&MasterInterface);
}


int
SoftCellCheck(CellInterfaceP)
    octObject *CellInterfaceP;
{
    short		FloorplanBagExists=FALSE;
    double	inputDouble();
    octObject     TargetBox;
  
    /*
     *  Two cases: if the FloorplanBag is there, give the option to
     *    change the values of the properties in the bag.
     *    Otherwise check that there is area information on the layer PLACE
     *    If there isn't, Create the FloorplanBag and add the information
     */
    if (octGetByName(CellInterfaceP,&FloorplanBag) != OCT_NOT_FOUND) {
	FloorplanBagExists = TRUE;
	if (octGenFirstContent(&FloorplanBag,OCT_BOX_MASK,&TargetBox) != OCT_OK)
	  Warning("obsolete cell");
	if (octGetByName(&FloorplanBag,&MinHeightOverWidth)!=OCT_OK) {
	    Warning("getting minAR");
	}
	if (octGetByName(&FloorplanBag,&MaxHeightOverWidth)!=OCT_OK) {
	    Warning("getting maxAR");
	}
    }

    /*
     * If the FloorplanBag exists, make sure the information in it is consistent
     * in both the contents and the interface
     */
    if (FloorplanBagExists == TRUE) {
	(void) SoftCellCheckConsistency(CellInterfaceP);
    }
}






int
SoftCellCheckConsistency(InterfaceP)
     octObject *InterfaceP;	/* its a facet */
{
    octObject Geo, TargetBox;
    octObject SoftCellContents;
    octGenerator genGeo;

    if (InterfaceP->type != OCT_FACET) FatalError("1");
    if (strcmp(InterfaceP->contents.facet.facet,"interface")!=0) {
	Warning("Facet should be \"interface\"");
	return(0);
    }
    SoftCellContents.type = OCT_FACET;
    SoftCellContents.contents.facet.cell = InterfaceP->contents.facet.cell;
    SoftCellContents.contents.facet.view = InterfaceP->contents.facet.view;
    SoftCellContents.contents.facet.version = InterfaceP->contents.facet.version;
    SoftCellContents.contents.facet.facet = "contents";
    SoftCellContents.contents.facet.mode = "a";

    InterfaceP->contents.facet.mode = "a";
    if (octOpenFacet(InterfaceP) < OCT_OK) FatalError("opening softcell");
    if (octOpenFacet(&SoftCellContents) < OCT_OK) FatalError("opening softcell cont");
  
    if (octGetByName(InterfaceP, &FloorplanBag) == OCT_NOT_FOUND) {
	printf("The cell is not SOFT!!\n");
    } else {
	if (octGenFirstContent(&FloorplanBag,OCT_BOX_MASK,&TargetBox)!=OCT_OK)
	  FatalError("get target box");

	OCT_ASSERT(octGetOrCreate(InterfaceP, &LayerPlace));
	/* delete all geometries attached to the layer, except TargetBox*/
	OCT_ASSERT(octInitGenContents(&LayerPlace, OCT_GEO_MASK, &genGeo));
	while (octGenerate(&genGeo, &Geo) == OCT_OK) {
	    if (!octIdsEqual(Geo.objectId, TargetBox.objectId)) OCT_ASSERT(octDelete(&Geo));
	}
	if (octIsAttached(&LayerPlace,&TargetBox)!=OCT_OK) {
	    if (verbose) printf("Attaching box (%d,%d) (%d,%d) on layer %s in %s:%s:%s\n",
				TargetBox.contents.box.lowerLeft.x,
				TargetBox.contents.box.lowerLeft.y,
				TargetBox.contents.box.upperRight.x,
				TargetBox.contents.box.upperRight.y,
				LayerPlace.contents.layer.name,
				InterfaceP->contents.facet.cell,
				InterfaceP->contents.facet.view,
				InterfaceP->contents.facet.facet);
	    OCT_ASSERT(octAttach(&LayerPlace,&TargetBox));
	}
	if (octFlushFacet(InterfaceP) != OCT_OK) FatalError("Flushing");
	/*
	 * do the same for the contents facet
	 */
	OCT_ASSERT(octGetOrCreate(&SoftCellContents, &LayerPlace));
	/* delete all geometries attached to the layer */
	OCT_ASSERT(octInitGenContents(&LayerPlace, OCT_GEO_MASK, &genGeo));
	while (octGenerate(&genGeo, &Geo) == OCT_OK) {
	    OCT_ASSERT(octDelete(&Geo));
	}
	if (octCreate(&LayerPlace,&TargetBox)!=OCT_OK)  FatalError("creating box");
	if (octFlushFacet(&SoftCellContents) != OCT_OK) FatalError("Flushing");
    }
    return(1);			/* OK */
}

