/*   symmetrize.c
 *
 *   Authors:      Antonella Bellettini,
 *                 Edoardo Charbon,
 *                 Enrico Malavasi
 *   Ver. 2.00    -    August 1991
 *
 *   Purpose: symmetric ajustment for analog circuits.
 */
#include "port.h"
#include "oct.h" 
#include "oh.h"
#include "options.h"
#include "utility.h"
#include "copyright.h"
#include "tap.h"
#include "th.h"

extern short verbose;
#define  VERBOSE(fct)  (verbose) ? fct : 0
#define  SWAP(type, a, b)   {type swapx; swapx = a; a = b; b = swapx;}

typedef int _axisDir;
#define   AXIS_HORIZONTAL   3
#define   AXIS_VERTICAL     1

#define   VERT_SYMMETRY_BAG    "YSYMMETRIES"
#define   HORIZ_SYMMETRY_BAG   "XSYMMETRIES"
#define   SYMMETRY_AXIS_BAG    "SYM"
#define   COUPLE_BAG           "COUPLE"
#define   POSITION_PROP        "POSITION"

typedef struct _symPair {
    octObject InstCell1;
    octObject InstCell2;
    unsigned char single;       /*  single/double cell */
} _symPair;


static octCoord
  getAxisPosition(axis)
octObject *axis;
{
    octObject prop;
    OH_ASSERT(ohGetByPropName(axis, &prop, POSITION_PROP));
    return ((octCoord)  prop.contents.prop.value.integer);
}
static void
  setAxisPosition(axis, pos)
octObject *axis;
octCoord   pos;
{
    octObject prop;
    OH_ASSERT(ohGetByPropName(axis, &prop, POSITION_PROP));
    if( prop.contents.prop.value.integer != pos) {
	prop.contents.prop.value.integer = pos;
	octModify( &prop );
    }
}

static int
  getLambda()
{
    static int lambda_value = 0;
    if( lambda_value == 0)   lambda_value = thLambda();
    return lambda_value;
}

static octCoord
  lambdaAlign(pos)
octCoord pos;
{
    pos /= getLambda();
    pos *= getLambda();
    return  pos;
}

#define  ALLOC_STEP  300

static int
  SymmetryFinder( SymBag, axisDir, symmPair )
octObject   *SymBag;
_axisDir     axisDir;
_symPair **symmPair;
/*
 * This function fills the elements of symPair[] vector according to the 
 * symmetry information contained in the given facet.
 */
{
    octObject    Instance;     /* Instances */
    octObject    CoupleBag;
    octGenerator CoupleGener, InstGener;
    int          j;
    int          aligned;
    int          cellNum   = 0;
    int          symPairNo = 0;
    octCoord     axisPosition;
    _symPair    *symPair = NULL;
     int         symmAllocated = 0;

    axisPosition =  getAxisPosition(SymBag);
    OH_ASSERT(octInitGenContents(SymBag, OCT_BAG_MASK, &CoupleGener));
    while(octGenerate(&CoupleGener, &CoupleBag) == OCT_OK) {
	if(symPairNo >= symmAllocated) {
	    symmAllocated += ALLOC_STEP;
	    symPair = REALLOC(_symPair, symPair, symmAllocated);
	}
	octInitGenContents(&CoupleBag, OCT_INSTANCE_MASK, &InstGener);
	if(octGenerate(&InstGener, &(symPair[symPairNo].InstCell1)) == OCT_OK) {
	    if(octGenerate(&InstGener, &(symPair[symPairNo].InstCell2)) == OCT_OK) {
		symPair[symPairNo].single = 0;
		cellNum += 2;
		/* Cell1 must be at the extreme left in the couple for vertical symmetry */ 
		if (axisDir == AXIS_VERTICAL) { 
		    if (symPair[symPairNo].InstCell2.contents.instance.transform.translation.x <
			symPair[symPairNo].InstCell1.contents.instance.transform.translation.x) {
			Instance = symPair[symPairNo].InstCell2;
			symPair[symPairNo].InstCell2 = symPair[symPairNo].InstCell1;
			symPair[symPairNo].InstCell1 = Instance;
		    }
		}
		
		/* Cell1 must be at the top in the couple for horizontal symmetry */ 
		if (axisDir == AXIS_HORIZONTAL) { 
		    if (symPair[symPairNo].InstCell2.contents.instance.transform.translation.y >
			symPair[symPairNo].InstCell1.contents.instance.transform.translation.y) {
			Instance = symPair[symPairNo].InstCell2;
			symPair[symPairNo].InstCell2 = symPair[symPairNo].InstCell1;
			symPair[symPairNo].InstCell1 = Instance;
		    }
		}
	    } else {
		symPair[symPairNo].single = 1;
		cellNum++;
	    }
	    symPairNo++;
	}
	octFreeGenerator(&InstGener);
    }
    octFreeGenerator(&CoupleGener);
    VERBOSE(printf("--> %d symmetric cells found\n",cellNum));

    (*symmPair) = symPair;  /* No more reallocations required  */

    /*  Put the axis in lambda-aligned position. Check if it is compatible
     *  with the single-cells. 
     */
    axisPosition = lambdaAlign(axisPosition);
    VERBOSE( printf("--> Checking LAMBDA-alignment of cells located on axis\n"));
    aligned = 1;
    for(j=0 ; j<symPairNo ; j++) {
	if(symPair[j].single) {
	    octBox  bb;
	    octCoord dx;
	    octBB(&(symPair[j].InstCell1), &bb);
	    dx = (axisDir == AXIS_VERTICAL) ? ohBoxWidth(&bb) : ohBoxHeight(&bb);
	    if(dx % getLambda()) {
		VERBOSE( printf( "Warning: size of instance %s is not multiple of LAMBDA\n",
			symPair[j].InstCell1.contents.instance.name));
		continue;
	    }
	    if(dx % (getLambda()*2)) {
		aligned = 0;
	    } else if(! aligned) {
		VERBOSE ( printf("Warning: instances on the symmetry axis cannot be LAMBDA-aligned\n"));
		return symPairNo;   /*  I don't know what to do with alignment */
	    }
	}
    }
    if(! aligned) 	axisPosition += (getLambda()/2);
    setAxisPosition( SymBag, axisPosition );

    return symPairNo;
}
    
static void
  SymmetryEnforcer( facet, SymBag, axisDir, symPairNo, symPair )
octObject   *facet;
octObject   *SymBag;
_axisDir    axisDir;
int         symPairNo;
_symPair *symPair;
/*
 * Place all cell pairs according to symmetries
 */
{
    octBox          bb, bb1,bb2;     /* Bounding boxes.*/
    octObject      *cell1,*cell2;
    octCoord        Trx, Try;    /* Translation */
    struct octTransform  transform;
    octCoord        dist, dx;
    int             i;
    octCoord  axisPosition;

    axisPosition =  getAxisPosition(SymBag);
    /*
     * Symmetric placement of the cells
     */

    transform.generalTransform[0][0] = 0;
    transform.generalTransform[0][1] = 0;
    transform.generalTransform[1][0] = 0;
    transform.generalTransform[1][1] = 0;
    transform.translation.x = 0;
    transform.translation.y = 0;

    for(i=0 ; i<symPairNo ; i++) {
	cell1=(&(symPair[i].InstCell1));
	octBB(cell1, &bb1);

/*
 * If there is only one cell in the structure symPair[i], cell1 must be put 
 * symmetrically on the symmetry axis.
 */
	if (symPair[i].single) { 
	    if (axisDir == AXIS_VERTICAL) {
		Trx = axisPosition - (bb1.upperRight.x+bb1.lowerLeft.x)/2;
		cell1->contents.instance.transform.translation.x += Trx;
	    }
	    if (axisDir == AXIS_HORIZONTAL) {
		Trx = axisPosition - (bb1.upperRight.y+bb1.lowerLeft.y)/2;
		cell1->contents.instance.transform.translation.y += Trx;
	    }
	    octModify(cell1);
	    
	} else {   /*  regular pair  */
	    cell2=(&(symPair[i].InstCell2));
	    octBB(cell2, &bb);

	    cell2->contents.instance.transform.transformType =
	      cell1->contents.instance.transform.transformType;
	    transform.transformType = (axisDir == AXIS_VERTICAL) ? OCT_MIRROR_X : OCT_MIRROR_Y;
	    octTransformAndModifyGeo(cell2, &transform);
	    octBB(cell2, &bb2);
	    cell2->contents.instance.transform.translation.x += (bb.lowerLeft.x-bb2.lowerLeft.x);
	    cell2->contents.instance.transform.translation.y += (bb.upperRight.y-bb2.upperRight.y);
	    octModify(cell2);
	    octBB(cell2, &bb2);
/*
 * Assign the right position to cells. If the couple is all on the right
 * or the left of the axis, the distance between cells is
 * maintained (only if it is greater than the minimal allowed distance).
 * The cell at the left (cell1) is then translated to the half of this
 * distance on the left hand side of the axis. (same for horizontal sym.
 * Where "left" and "right" are replaced with "top" and "bottom")      
 */
	    if(axisDir == AXIS_VERTICAL) {

		if(bb1.lowerLeft.x > bb2.lowerLeft.x) {
		    SWAP ( octBox, bb1, bb2);
		    SWAP ( octObject *, cell1, cell2);
		}
		
		dist = bb2.lowerLeft.x - bb1.upperRight.x;
		while((dist % (getLambda()*2)) != 0)  dist++;
		dist -= (bb2.lowerLeft.x - bb1.upperRight.x);
		if( dist ) {
		    bb2.lowerLeft.x  += dist;
		    bb2.upperRight.x += dist;
		    cell2->contents.instance.transform.translation.x += dist;
		}

		Trx = axisPosition - (bb2.upperRight.x + bb1.lowerLeft.x)/2;
		cell1->contents.instance.transform.translation.x += Trx;
		cell2->contents.instance.transform.translation.x += Trx;
		
		Try = (bb2.lowerLeft.y - bb1.lowerLeft.y)/2;
		cell1->contents.instance.transform.translation.y += Try;
		cell2->contents.instance.transform.translation.y -= Try;
		cell1->contents.instance.transform.translation.y =
		  lambdaAlign(cell1->contents.instance.transform.translation.y);
		cell2->contents.instance.transform.translation.y =
		  lambdaAlign(cell2->contents.instance.transform.translation.y);
		
	    } else {

		if(bb1.lowerLeft.y > bb2.lowerLeft.y) {
		    SWAP ( octBox, bb1, bb2);
		    SWAP ( octObject *, cell1, cell2);
		}

		dist = bb2.lowerLeft.y - bb1.upperRight.y;
		while((dist % (getLambda()*2)) != 0)  dist++;
		dist -= (bb2.lowerLeft.y - bb1.upperRight.y);
		if( dist ) {
		    bb2.lowerLeft.y  += dist;
		    bb2.upperRight.y += dist;
		    cell2->contents.instance.transform.translation.y += dist;
		}

		Trx = axisPosition - (bb2.upperRight.y + bb1.lowerLeft.y)/2;
		cell1->contents.instance.transform.translation.y += Trx;
		cell2->contents.instance.transform.translation.y += Trx;
		
		Try = (bb2.lowerLeft.x - bb1.lowerLeft.x)/2;
		cell1->contents.instance.transform.translation.x += Try;
		cell2->contents.instance.transform.translation.x -= Try;
		cell1->contents.instance.transform.translation.x =
		  lambdaAlign(cell1->contents.instance.transform.translation.x);
		cell2->contents.instance.transform.translation.x =
		  lambdaAlign(cell2->contents.instance.transform.translation.x);
	    }

	    octModify(cell1);
	    octModify(cell2);
	}
    }
}

static void
  readAndAdjustSymmetry(facet, symmBagName)
octObject *facet;
char      *symmBagName;
/* This functiion reads symmetry constraints from oct and adjusts geometries */
{ 
    octObject     SymmetryBag, SymBag; /*SYMMETRIES bag.*/
    octGenerator  SymGener;
    _axisDir      axisDir;
    _symPair     *symPair;
    int           symPairNo;   /*  No. of symmetric pairs */
    
    axisDir = (strcmp(symmBagName, HORIZ_SYMMETRY_BAG)) ? AXIS_VERTICAL : AXIS_HORIZONTAL;

    if(ohGetByBagName(facet, &SymmetryBag, symmBagName) == OCT_OK) {
	octInitGenContents(&SymmetryBag, OCT_BAG_MASK, &SymGener);
	while(octGenerate(&SymGener, &SymBag) == OCT_OK) {
	    if(! strcmp((SymBag.contents.bag.name), SYMMETRY_AXIS_BAG)) {

		/*  2.)  Find the structure of the symmetry  */
		symPairNo = SymmetryFinder( &SymBag, axisDir, &symPair);

		/*  3.)  Fix the structure if necessary  */
		SymmetryEnforcer(facet, &SymBag, axisDir, symPairNo, symPair);
	    }
	} 
	octFreeGenerator(&SymGener);
    } else {
	VERBOSE ( printf("Warning: No %s bag found.\n", symmBagName));
    }

    if(symPairNo > 0) {
	FREE( symPair );
    }
}

void
  symmetrize( facet )
octObject *facet;
{
    thTechnology( facet );
 
    /* read symmetry constraints from oct and modify geometries */
    
    VERBOSE(printf("--> Processing vertical symmetry axes ...\n"));
    readAndAdjustSymmetry(facet, VERT_SYMMETRY_BAG);

    VERBOSE(printf("--> Processing horizontal symmetry axes ...\n"));
    readAndAdjustSymmetry(facet, HORIZ_SYMMETRY_BAG);
}
