/*
 * annealPla.c
 *
 * Call routines to initialize ace, add constraints, and anneal the matrix,
 * and move ace's output into anchor's layout generation data structures.
 */

#include "copyright.h"
#include "port.h"
#include "anchor.h"

extern s_options Options;
extern int plaCheckCF();

/* pla input								*/
extern s_signal *InSigs, *OutSigs, *ProdSigs;
extern int InSigCount, OutSigCount, ProdCount;
extern char **PlaTT;
extern int CellCount;

/* layout generation state information (ace output)			*/
extern int AndCols, OrCols, Rows;
extern s_input **Inputs;
extern s_output **Outputs;
char **Core, **allocCore();

/* ace stuff used locally						*/
annealMasters Masters;
aceId matrixId;
aceId rowPlaneId, andPlaneId, orPlaneId;

/*
 * annealPla
 *
 * Do the appropriate operations to prepare the pla(5CAD)-style input
 * for generating layout. The steps are
 *   1) initialize ace with this particular annealing problem, and
 *         set up constraints of adjacency, order and folding.
 *   2) run the annealing.
 *   3) verify constraints.
 *   4) extract the output from ace.
 */
annealPla()
{
    /* add masters and create the matrix				*/
    initAce();

    /* add cells and get netIds						*/
    addPla();

    /* add any constraints to ace					*/
    addConstraints();
    setCFParams();

    /* let ace go to work						*/
    aceAnnealMatrix( Options.fastAnneal );

    /* get the final state						*/
    getAceOutput();

    aceFreeMatrix();
}

/*
 * initAce
 *
 * Start up ace. Add masters and create the matrix.
 */
initAce()
{
    int row, col;

    /* setup tracing							*/
    aceSetTrace( Options.traceLevel, stdout);

    /* figure out how many inputs and ouputs there will really be	*/
    AndCols = 0;
    for( col = 0; col < InSigCount*2; ++col){
	/* if signal is used, count it					*/
	if( InSigs[ col ].used){
	    ++AndCols;
	}
    }
    OrCols = 0;
    for( col = 0; col < OutSigCount; ++col){
	/* if signal is used, count it					*/
	if( OutSigs[ col ].used){
	    ++OrCols;
	}
    }
    Rows = 0;
    for( row = 0; row < ProdCount; ++row){
	/* if product term is used, count it				*/
	if( ProdSigs[ col ].used){
	    ++Rows;
	}
    }

    /* create the matrix						*/
    matrixId = aceCreateMatrix( Rows, AndCols + OrCols);

    /* split the matrix into two vertical planes			*/
    aceGetPlanesAtLoc( 0, 0, &rowPlaneId, &andPlaneId);
    aceSplitPlane( andPlaneId, AndCols);

    /* get or plane id							*/
    aceGetPlanesAtLoc( 0, AndCols, &rowPlaneId, &orPlaneId);

    /* add masters							*/
    Masters.and = aceAddMaster( "s", "", 1, 1);
    Masters.or = aceAddMaster( "p", "", 1, 1);
}

/*
 * addPla
 *
 * Add all the cells to the matrix and store netIds in the XxxSig structs.
 */
addPla()
{
    int row, col;
    int rowoff, coloff;
    aceId rowNetId, colNetId;
    unsigned termMask;

    CellCount = 0;
    /* add and-plane cells						*/
    rowoff = 0;
    for( row = 0; row < ProdCount; ++row){
	if( ProdSigs[ row ].used){
	    /* add all and-plane cells in this row			*/
	    coloff = 0;
	    for( col = 0; col < InSigCount*2; ++col){
		if( InSigs[ col ].used){
		    if( PlaTT[ row ][ col ] == '1'){
			aceAddCell( Masters.and, row - rowoff, col - coloff);
			++CellCount;
		    }
		} else {
		    ++coloff;
		}
	    }
	    /* add all or-plane cells in this row			*/
	    coloff = 0;
	    for( col = 0; col < OutSigCount; ++col){
		if( OutSigs[ col ].used){
		    if( PlaTT[ row ][ InSigCount*2 + col ] == '1'){
			aceAddCell( Masters.or, row - rowoff,
						AndCols + col - coloff);
			++CellCount;
		    }
		} else {
		    ++coloff;
		}
	    }
	    /* get the product term net id				*/
	    aceGetNetsAtLoc( row - rowoff, 0, &rowNetId, &colNetId);
	    ProdSigs[ row ].id = rowNetId;
	} else {
	    ++rowoff;
	    ProdSigs[ row ].id = ace_null_id;
	}
    }

    /* get the col net ids and set up the folding type			*/
    if( Options.fold){
	termMask = ACE_TERM_TWO_STATE;
    } else {
	termMask = 0;
    }
    coloff = 0;
    for( col = 0; col < InSigCount*2; ++col){
	if( InSigs[ col ].used ){
	    aceGetNetsAtLoc( 0, col - coloff, &rowNetId, &colNetId);
	    InSigs[ col ].id = colNetId;
	    aceRegisterTermNet( colNetId, termMask);
	} else {
	    ++coloff;
	    InSigs[ col ].id = ace_null_id;
	}
    }
    coloff = 0;
    for( col = 0; col < OutSigCount; ++col){
	if( OutSigs[ col ].used ){
	    aceGetNetsAtLoc( 0, AndCols + col - coloff, &rowNetId, &colNetId);
	    OutSigs[ col ].id = colNetId;
	    aceRegisterTermNet( colNetId, termMask);
	} else {
	    ++coloff;
	    OutSigs[ col ].id = ace_null_id;
	}
    }
}

/*
 * addConstraints
 *
 * Set the cost function weights and add the necessary constraints to create
 * a legal pla, plus the constraits of order and folding and/or critical nets
 * for this particular pla.
 */
addConstraints()
{
    double ( *fn)();
    double plaFoldCF(), plaAdjCF();
    int i;
    aceId args[ 2 ];
    aceDefaults *defs;

    /* setup cost function parameters					*/
    defs = aceGetDefaults();

    defs->cost.usedRC.row = 0.0;
    defs->cost.usedBias.row = 0.0;
    defs->cost.usedRC.col = 2.0 * Rows;
    defs->cost.usedBias.col = 3.0 * Rows;
    defs->cost.netLength.row = defs->cost.netLength.col = 5.0;
    defs->cost.netCapacity.row =
		defs->cost.netLength.col * (double)(AndCols + OrCols) / 2.0;
    defs->cost.cellCapacity.row = 1.0 * (AndCols+OrCols);
    defs->cost.netCapacity.col = ( defs->cost.netLength.col + 5.0) * Rows;
    defs->cost.cellCapacity.col = 7.0 * Rows;

    aceSetDefaults( defs );

    if( Options.fold ){
	/* need the more complicated constraint functions for folding	*/
	fn = plaFoldCF;
    } else {
	fn = plaAdjCF;
    }

    /* register the constraints						*/
    for( i = InSigCount - 1; i >= 0; --i){
	if( InSigs[ i*2 ].used && InSigs[ i*2 + 1 ].used){
	    args[ 0 ] = InSigs[ i*2 ].id;
	    args[ 1 ] = InSigs[ i*2 + 1 ].id;
	    aceRegisterUCF( fn, plaCheckCF,
			ACE_UCFE_NET_MOVE | ACE_UCFE_FLIP_TERM,
			2, args, 2, args);
	}
    }
}

/*
 * getAceOutput
 *
 * Extract the final state from ace.
 */
getAceOutput()
{
    aceApNet *net;
    aceApCell *cell;
    aceApPlane *plane;
    aceObject *obj;
    int i;
    s_input *in1, *in2, *getInputByName();
    s_output *out, *newOutput();
    aceGenerator gen;

    plane = (aceApPlane *)aceGetById( rowPlaneId );
    if( Rows != ( plane->upper + 1)){
	/*
	errRaise( optProgName, 1, "number of rows changed");
	*/
	printf("WARNING: number of rows changed\n");
    }

    plane = (aceApPlane *)aceGetById( andPlaneId );
    AndCols = plane->upper + 1;
    plane = (aceApPlane *)aceGetById( orPlaneId );
    OrCols = plane->upper + 1 - AndCols;

    Inputs = ALLOC( s_input *, AndCols);
    memset((char *)Inputs, 0, sizeof(s_input *) * AndCols);
    Outputs = ALLOC( s_output *, OrCols);
    memset((char *)Outputs, 0, sizeof(s_output *) * OrCols);

    Core = allocCore( Rows, AndCols + OrCols);

    for( i = InSigCount - 1; i >= 0; --i){
	in1 = in2 = NIL(s_input);
	if( InSigs[ i*2 ].id != ace_null_id){
	    net = (aceApNet *)aceGetById( InSigs[ i*2 ].id );
	    in1 = getInputByName( InSigs[ i*2 ].name );
	    in1->col = net->rowcol;
	    in1->top = net->upper;
	    in1->bottom = net->lower;
	    in1->phase = POS;
	    addInput( in1 );
	} else {
	    /* create the struct so all in->comp fields are non-NIL	*/
	    in1 = getInputByName( InSigs[ i*2 ].name );
	    in1->col = in1->bottom = in1->top = -1;
	}
	if( InSigs[ i*2 + 1 ].id != ace_null_id){
	    net = (aceApNet *)aceGetById( InSigs[ i*2 + 1 ].id );
	    in2 = getInputByName( InSigs[ i*2 + 1 ].name );
	    in2->col = net->rowcol;
	    in2->top = net->upper;
	    in2->bottom = net->lower;
	    in2->phase = NEG;
	    addInput( in2 );
	} else {
	    /* create the struct so all in->comp fields are non-NIL	*/
	    in2 = getInputByName( InSigs[ i*2 + 1 ].name );
	    in2->col = in2->bottom = in2->top = -1;
	}
	if(( in1 != NIL(s_input)) && ( in2 != NIL(s_input))){
	    in1->comp = in2;
	    in2->comp = in1;
	}
    }

    for( i = OutSigCount - 1; i >= 0; --i){
	if( OutSigs[ i ].id != ace_null_id){
	    net = (aceApNet *)aceGetById( OutSigs[ i ].id );
	    out = newOutput( OutSigs[ i ].name);
	    out->col = net->rowcol;
	    out->top = net->upper;
	    out->bottom = net->lower;
	    addOutput( out );
	}
    }

    for( i = Rows - 1; i >= 0; --i){
	/* generate each cell						*/
	if( ProdSigs[ i ].id != ace_null_id){
	    aceInitGenContents( ProdSigs[ i ].id, &gen );
	    while( aceGenerate( &gen, &obj)){
		cell = (struct aceApCell *) obj;
		if( cell->masterId == Masters.and){
		    Core[ cell->loc.row ][ cell->loc.col ] = SUM;
		} else if( cell->masterId == Masters.or){
		    Core[ cell->loc.row ][ cell->loc.col ] = PRODUCT;
		} else {
		    errRaise( optProgName, 1, "master id doesn't match");
		}
	    }
	}
    }
}

/*
 * addInput
 *
 * add an input to the Inputs array
 */
addInput( in )
s_input *in;
{
    s_input *inp;

    if( Inputs[ in->col ] == NIL(s_input) ||
      ( Inputs[ in->col ]->bottom > in->bottom)){
	in->next = Inputs[ in->col ];
	Inputs[ in->col ] = in;
    } else {
	for( inp = Inputs[ in->col ]; inp != NIL(s_input); inp = inp->next){
	    if(( inp->next == NIL(s_input)) ||
	       ( inp->next->bottom > in->bottom)){
		break;
	    }
	}
	/* put between inp and inp->next				*/
	in->next = inp->next;
	inp->next = in;
    }
}

/*
 * addOutput
 *
 * add an output to the Outputs array
 */
addOutput( out )
s_output *out;
{
    s_output *outp;

    if( Outputs[ out->col - AndCols ] == NIL(s_output) ||
      ( Outputs[ out->col - AndCols ]->bottom > out->bottom)){
	out->next = Outputs[ out->col - AndCols ];
	Outputs[ out->col - AndCols ] = out;
    } else {
	for( outp = Outputs[ out->col - AndCols ]; outp != NIL(s_output);
		outp = outp->next){
	    if(( outp->next == NIL(s_output)) ||
	       ( outp->next->bottom > out->bottom)){
		break;
	    }
	}
	/* put between inp and inp->next				*/
	out->next = outp->next;
	outp->next = out;
    }
}
