/*
 * plaOut.c
 *
 * Print out a pla(5CAD) format description of the PLA for which layout is
 * being generated.  This is primarily useful for detecting errors in
 * the 'sanitize matrix' algorithms used to clean up the input personality
 * matrix.
 */

#include "copyright.h"
#include "port.h"
#include "anchor.h"
#include "vov.h"

extern s_options Options;

static inC = 0, outC = 0;
static s_input **inArray = NIL(s_input *);
static s_output **outArray = NIL(s_output *);

extern octObject oFacet;
static char filename[ 256 ];

/*
 * printPLA
 *
 * write out the pla(5CAD) file.  Read the the comparison pla(5CAD) file
 * (the argument to the '-v' option) and print out inputs and outputs
 * in the same order.
 */
printPLA( rows, orC, core, outputs)
int rows, orC;
char **core;
s_output **outputs;
{
    int row, col;
    s_input *in, *getInputByName();
    s_output *out;
    FILE *fd;

    /* open fd in append mode (in case this routine called multiple times
     * in one execution).
     */
    sprintf( filename, "%s_%s.pla", oFacet.contents.facet.cell,
				oFacet.contents.facet.view);
    VOVoutput( VOV_UNIX_FILE, filename );
    if(( fd = fopen( filename, "a")) == NIL(FILE)){
	fprintf( stderr,
		"%s: couldn't open \"%s\" PLA verification output file.\n",
		optProgName, filename);
	return;
    }

    /* if first call, parse header info of reference pla		*/
    if( inC == 0){
	if( !readPLAheader( outputs, orC)){
	    fprintf( stderr,
		"couldn't parse \"%s\" for verification.\n",
		Options.plaVerifyFile);
	    inC = 0;
	    return;
	}
    }

    /* header stuff */
    fprintf( fd, ".i %d\n.o %d\n", inC, outC);
    fprintf( fd, ".ilb");
    for( col = 0; col < inC; ++col){
	fprintf( fd, " %s", inArray[ col ]->name);
    }
    fprintf( fd, "\n");

    fprintf( fd, ".ob");
    for( col = 0; col < outC; ++col){
	fprintf( fd, " %s", outArray[ col ]->name);
    }
    fprintf( fd, "\n");

    fprintf( fd, ".p %d\n", rows);

    /* truth table type data */
    for( row = 0; row < rows; ++row){
	/* travese inputs in same order as before */
	for( col = 0; col < inC; ++col){
	    in = inArray[ col ];
	    /* look at this input */
	    if((( in->top >= row) && (in->bottom <= row)) ||
	       (( in->top <= row) && (in->bottom >= row))){
		/* right range */
		if( core[ row ][ in->col ] == SUM){
		    fprintf( fd, "1");
		    continue;
		}
	    }
	    if(( in->comp != NIL(s_input)) && 
	       (( in->comp->top >= row) && (in->comp->bottom <= row)) ||
	       (( in->comp->top <= row) && (in->comp->bottom >= row))){
		if( core[ row ][ in->comp->col ] == SUM){
		    fprintf( fd, "0");
		    continue;
		}
	    }
	    fprintf( fd, "-");
	}

	fprintf( fd, " ");
	/* now do outputs, again in same order */
	for( col = 0; col < outC; ++col){
	    out = outArray[ col ];
	    if((( out->top >= row) && (out->bottom <= row)) ||
	       (( out->top <= row) && (out->bottom >= row))){
		if( core[ row ][ out->col ] == PRODUCT){
		    fprintf( fd, "1");
		    continue;
		}
	    }
	    fprintf( fd, "0");
	}
	fprintf( fd, "\n");
    }

    /* end */
    fprintf( fd, ".e\n");

    /* close the file							*/
    fclose( fd );
}

/*
 * readPLAheader
 *
 * Read in the number of inputs, the number of ouptuts,
 * and the actual names of the ins and outs
 */
int
readPLAheader( outputs, orC)
s_output **outputs;
int orC;
{
    FILE *fd;
    char token[ NAMESIZE ];
    int index, c;
    s_input *in, *getInputByName();
    s_output *out;

    if(( fd = VOVfopen( Options.plaVerifyFile, "r")) == NIL(FILE)){
	fprintf( stderr,
		"%s: couldn't open \"%s\" PLA verification file: ", 
		optProgName, Options.plaVerifyFile);
	return( FALSE );
    }

    while( TRUE ){
	fscanf( fd, "%s", token);
	if( !strcmp( token, ".i")){
	    if( fscanf( fd, " %d", &inC) != 1){
		fprintf( stderr, "%s: couldn't scan number of inputs: ",
			optProgName);
		return( FALSE );
	    }
	} else if( !strcmp( token, ".o")){
	    if( fscanf( fd, " %d", &outC) != 1){
		fprintf( stderr, "%s: couldn't scan number of outputs: ",
			optProgName);
		return( FALSE );
	    }
	} else if( !strcmp( token, ".ilb")){
	    if( inC == 0){
		fprintf( stderr, "%s: .ilb before .i: ", optProgName);
		return( FALSE );
	    }
	    inArray = ALLOC( s_input *, inC);
	    for( index = 0; index < inC; ++index){
		if( fscanf( fd, "%s", token) != 1){
		    fprintf( stderr, "%s: couldn't read an input name: ",
			optProgName);
		    return( FALSE );
		}
		if(( in = getInputByName( token)) == NIL(s_input)){
		    fprintf( stderr,
		    "%s: input named \"%s\" in \"%s\" not in input personality matrix. Aborting verification.\n",
		    optProgName, token, Options.plaVerifyFile);
		    return FALSE;
		}
		inArray[ index ] = in;
	    }
	    if( index != inC){
		fprintf( stderr, "%s: number of inputs (%d) does not match number of input names (%d): ", optProgName, inC, index);
		return( FALSE );
	    }
	} else if( !strcmp( token, ".ob")){
	    if( outC == 0){
		fprintf( stderr, "%s: outC is still 0: ", optProgName);
		return( FALSE );
	    }
	    outArray = ALLOC( s_output *, outC);
	    for( index = 0; index < outC; ++index){
		if( fscanf( fd, "%s", token) != 1){
		    fprintf( stderr, "%s: couldn't read an output name: ",
			optProgName);
		    return( FALSE );
		}
		for( c = 0; c < orC; ++c){
		    for( out = outputs[ c ]; out != NIL(s_output);
			    out = out->next){
			if( !strcmp( token, out->name)){
			    break;
			}
		    }
		    if( out != NIL(s_output)){
			outArray[ index ] = out;
			break;
		    }
		}
		if( out == NIL(s_output)){
		    fprintf( stderr,
		    "%s: output named \"%s\" in \"%s\" not in input personality matrix. Quiting verification.\n",
		    optProgName, token, Options.plaVerifyFile);
		    return( FALSE );
		}
	    }
	    if( index != outC){
		fprintf( stderr, "%s: number of outputs (%d) does not match number of output names (%d): ",
			optProgName, outC, index);
		return( FALSE );
	    }
	} else if( token[ 0 ] == '.'){
	    /* found some other directive				*/
	    while(( c = getc( fd)) != '\n') ;
	} else {
	    fprintf( stderr, "%s: unknown token \"%s\": ", optProgName, token);
	    return( FALSE );
	}

	/* if found everything, leave					*/
	if(( inArray != NIL(s_input *)) && (outArray != NIL(s_output *))){
	    break;
	}
    }
    return( TRUE );
}
