/*
 * plaTermCF.c
 *
 * Anchor's adjacent net constraint function tailored for anchor's
 * handling of pla inputs.
 */

#include "copyright.h"
#include "port.h"
#include "anchor.h"

/* imported to calculate dynamic values for cost function parameters	*/
extern double my_rint();
extern s_options Options;
extern int AndCols, OrCols, Rows, CellCount;

/* penalty constants 							*/
static double FixedCost;
static double DistCost;
static double OpSidesCost;

/*
 * setCFParams
 *
 * Calculate the cost function parameters.
 */
setCFParams()
{
    aceDefaults *def;
    int aveCellPerCol;

    aveCellPerCol = (int) my_rint( (double)CellCount / (double)(AndCols+OrCols));

    /* get the anchor-updated default structure				*/
    def = aceGetDefaults();

    /* The fixed cost is the cost for not being adjacent.		*/
    FixedCost = (double)aveCellPerCol * def->cost.netLength.row;
    FixedCost += def->cost.usedRC.col + def->cost.usedBias.col;
    /* average cost of cell overlap for a displacement			*/
    FixedCost += def->cost.cellCapacity.col *
		(double)((aveCellPerCol*(aveCellPerCol+1)))/(double)(2*Rows);
    FixedCost *= 0.85;

    /* The distance cost is for each column that the nets are separated	*/
    DistCost = ((double)Rows * def->cost.netCapacity.col) / 5.0;

    /* The opposite sides cost is to insure that both nets are on the same
     * side of the core.
     */
    OpSidesCost = ((double)Rows *
		  (def->cost.netCapacity.row + def->cost.netLength.row)) / 1.5;

    FixedCost = my_rint( FixedCost);
    DistCost = my_rint( DistCost );
    OpSidesCost = my_rint( OpSidesCost );
    if( Options.debug ){
	printf( "Fixed: %.1f; Dist: %.1f; OpSides %.1f\n",
	    FixedCost, DistCost, OpSidesCost);
    }
}

/*
 * plaFoldCF
 *
 * A penalty function for keeping two nets adjacent and folded to the
 * same side.
 */
/*ARGSUSED*/
double
plaFoldCF( count, array)
int count;
aceObject *(array[]);
{
    int dist, opSides = 0;

    /* calc distance between the two nets				*/
    dist = ((aceApNet *)array[0])->rowcol - ((aceApNet *)array[1])->rowcol;

    if( dist < 0){
	dist = -dist;
    }

    /* cost if terms not on same side of matrix				*/
    if(( (((aceApNet *)array[ 0 ])->termMask & ACE_TERM_LOWER_VE) &&
	 ! (((aceApNet *)array[ 1 ])->termMask & ACE_TERM_LOWER_VE )) ||
       ( (! (((aceApNet *)array[ 0 ])->termMask & ACE_TERM_LOWER_VE)) &&
	 (((aceApNet *)array[ 1 ])->termMask & ACE_TERM_LOWER_VE )) ){
	
	opSides = 1;
    }

    if( dist > 1){
	return( FixedCost + dist * DistCost + (double)opSides * OpSidesCost);
    } else {
	return( (double)opSides * OpSidesCost);
    }
}

/*
 * plaAdjCF
 *
 * A penalty function for keeping two nets adjacent. Assumes there is
 * no folding going on.
 */
/*ARGSUSED*/
double
plaAdjCF( count, array)
int count;
aceObject *(array[]);
{
    int dist;

    /* calc distance between the two nets				*/
    dist = ((aceApNet *)array[0])->rowcol - ((aceApNet *)array[1])->rowcol;

    if( dist < 0){
	dist = -dist;
    }

    if( dist == 1) return( 0.0);
    if( dist == 0) return( FixedCost );

    return( FixedCost + dist * DistCost);
}

/*
 * plaCheckCF
 *
 * A penalty function for keeping two nets adjacent and folded to the
 * same side.
 */
/*ARGSUSED*/
int
plaCheckCF( count, array)
int count;
aceObject *(array[]);
{
    int dist, opSides = 0;

#ifdef UNDEFINED
    /* the undefined version was for when anchor was unable to recover
     * on its own from unsatisfied constraints.
     * Code is now in place to guarentee results, so this function is now
     * trivial
     */
    /* calc distance between the two nets				*/
    dist = ((aceApNet *)array[0])->rowcol - ((aceApNet *)array[1])->rowcol;

    if( dist < 0){
	dist = -dist;
    }

    if( dist != 1) return( TRUE );

    /* cost if terms not on same side of matrix				*/
    if(( (((aceApNet *)array[ 0 ])->termMask & ACE_TERM_LOWER_VE) &&
	 ! (((aceApNet *)array[ 1 ])->termMask & ACE_TERM_LOWER_VE )) ||
       ( (! (((aceApNet *)array[ 0 ])->termMask & ACE_TERM_LOWER_VE)) &&
	 (((aceApNet *)array[ 1 ])->termMask & ACE_TERM_LOWER_VE )) ){
	
	return( TRUE );
    }

    return( FALSE );
#endif /* UNDEFINED */
    return ( FALSE );
}


/* rint() is not portable (e.g. it does not work on sequent). so here it is */
double my_rint( x )
    double x;
{
    int k = x;			/* Convert to closest integer */
    x = k;			/*  */
    return x;			/* return a double */
}
