/*
 * readPla.c
 *
 * Read a pla(5CAD) input file.
 * Set up some global data structures to use to set up ace.
 */

#include "copyright.h"
#include "port.h"
#include "anchor.h"
#include "vov.h"

extern s_options Options;

extern s_signal *InSigs;
extern int InSigCount;
extern s_signal *OutSigs;
extern int OutSigCount;
extern s_signal *ProdSigs;
extern int ProdCount;
extern char **PlaTT;

/*
 * readPla
 *
 * Read in the pla. Allocate and fill the name fields of the XxxSig structures.
 */
readPla( fname )
char *fname;		/* name of the input pla file			*/
{
    FILE *fp;

    if(( fp = VOVfopen( fname, "r")) == NIL(FILE)){
	errRaise( optProgName, 1,
		"readPla: could not open input file \"%s\"", fname);
    }
    rewind( fp );

    InSigCount = OutSigCount = ProdCount = 0;

    /* read in the input and output signal names			*/
    readPlaSignals( fp );

    /* read the pla truth table information				*/
    readPlaTT( fp );

    fclose( fp );
}

/*
 * readPlaTT
 *
 * Read in the pla's truth table.
 */
readPlaTT( fp )
FILE *fp;
{
    int c, row, col;
    char **allocCore();

    PlaTT = allocCore( ProdCount, InSigCount*2 + OutSigCount);

    /* Read each line of the truth table.  There should be one non-blank
     * for each input and output in each line, and one line for each
     * product term.  As per pla(5CAD), 'X', 'x' and '-' all mean no
     * connection, as well as '0' in the or plane.
     */
    for( row = 0; row < ProdCount; ++row){
	for( col = 0; col < InSigCount; ++col){
	    /* find InSigCount non-blanks on this line			*/
	    while((( c = getc( fp )) == ' ') || ( c == '\t')) ;

	    /* process one character					*/
	    if( c == '1'){
		/* connection to positive input signal			*/
		ProdSigs[ row ].used = TRUE;
		InSigs[ col*2 ].used = TRUE;
		PlaTT[ row ][ col*2 ] = '1';
		PlaTT[ row ][ col*2 + 1 ] = '-';
	    } else if( c == '0'){
		/* connection to compement of input signal		*/
		ProdSigs[ row ].used = TRUE;
		InSigs[ col*2 + 1 ].used = TRUE;
		PlaTT[ row ][ col*2 + 1 ] = '1';
		PlaTT[ row ][ col*2 ] = '-';
	    } else if(( c == '-') || ( c == 'x') || ( c == 'X')){
		/* no connection					*/
		PlaTT[ row ][ col*2 ] = '-';
		PlaTT[ row ][ col*2 + 1 ] = '-';
	    } else if( c == '\n'){
		errRaise( optProgName, 1, "%s line %d %s",
		    "readPlaTT: unexpected newline in truth table", row,
		    "from the start of the truth table");
	    } else {
		errRaise( optProgName, 1, "%s (%c) %s line %d %s",
		    "readPlaTT: illegal character", c,
		    "in truth table", row,
		    "from the start of the truth table");
	    }
	}
	/* do same for output terms					*/
	for( col = 0; col < OutSigCount; ++col){
	    /* find OutSigCount non-blanks on this line			*/
	    while((( c = getc( fp )) == ' ') || ( c == '\t')) ;

	    /* process one character					*/
	    if( c == '1'){
		/* connection to positive input signal			*/
		ProdSigs[ row ].used = TRUE;
		OutSigs[ col ].used = TRUE;
		PlaTT[ row ][ InSigCount*2 + col ] = '1';
	    } else if(( c == '0') || ( c == '-') || ( c == 'x') || ( c == 'X')){
		/* no connection					*/
		PlaTT[ row ][ InSigCount*2 + col ] = '0';
	    } else if( c == '\n'){
		errRaise( optProgName, 1, "%s line %d %s",
		    "readPlaTT: unexpected newline in truth table", row,
		    "from the start of the truth table");
	    } else {
		errRaise( optProgName, 1, "%s (%c) %s line %d %s",
		    "readPlaTT: illegal character", c,
		    "in truth table", row,
		    "from the start of the truth table");
	    }
	}

	/* read to newline						*/
	while( getc( fp ) != '\n') ;
    }
}

/*
 * readPlaSignals
 *
 * Read in the number of inputs, the number of ouptuts,
 * and the actual names of the ins and outs
 */
readPlaSignals( fp )
FILE *fp;
{
    char token[ NAMESIZE ], *formSigNameComp();
    int index, c;

    token[ 0 ] = '\0';

    while( TRUE ){
	if( fscanf( fp, "%s", token) != 1){
	    errRaise( optProgName, 1,
		"readPlaSignals: expected a control, couldn't scan one");
	}
	if( !strcmp( token, ".i")){
	    if( fscanf( fp, " %d", &InSigCount) != 1){
		errRaise( optProgName, 1,
		    "readPlaSignals: couldn't scan integer after .i control");
	    }
	} else if( !strcmp( token, ".o")){
	    if( fscanf( fp, " %d", &OutSigCount) != 1){
		errRaise( optProgName, 1,
		    "readPlaSignals: couldn't scan integer after .o control");
	    }
	} else if( !strcmp( token, ".p")){
	    if( fscanf( fp, " %d", &ProdCount) != 1){
		errRaise( optProgName, 1,
		    "readPlaSignals: couldn't scan integer after .p control");
	    }
	    ProdSigs = ALLOC( s_signal, ProdCount);
	    if( ProdSigs == NIL(s_signal)){
		errRaise( optProgName, 1,
		    "readPlaSignals: no memory for product term signal array");
	    }
	    memset((char *)ProdSigs, 0, sizeof(s_signal)*ProdCount);
	} else if( !strcmp( token, ".ilb")){
	    if( InSigCount == 0){
		errRaise( optProgName, 1, "readPlaSignals: .ilb before .i");
	    }
	    /* double size so there's room for complements later	*/
	    InSigs = ALLOC( s_signal, InSigCount * 2);
	    if( InSigs == NIL(s_signal)){
		errRaise( optProgName, 1,
		    "readPlaSignals: no memory for input signal array");
	    }
	    memset((char *)InSigs, 0, sizeof(s_signal)*InSigCount);
	    for( index = 0; index < InSigCount; ++index){
		if( fscanf( fp, "%s", token) != 1){
		    errRaise( optProgName, 1,
			"readPlaSignals: couldn't read an input name");
		}
		InSigs[ index*2 ].name = util_strsav( token );
		InSigs[ index*2 + 1].name =
				formSigNameComp( InSigs[ index*2 ].name);
	    }
	    if( index != InSigCount){
		errRaise( optProgName, 1, "%s (%d) %s (%d)",
			"readPlaSignals: number of inputs", InSigCount,
			"does not match number of input names", index);
	    }
	} else if( !strcmp( token, ".ob")){
	    if( OutSigCount == 0){
		errRaise( optProgName, 1, "readPlaSignals: .ob before .o");
	    }
	    OutSigs = ALLOC( s_signal, OutSigCount);
	    if( OutSigs == NIL(s_signal)){
		errRaise( optProgName, 1,
		    "readPlaSignals: no memory for output signal array");
	    }
	    memset((char *)OutSigs, 0, sizeof(s_signal)*OutSigCount);
	    for( index = 0; index < OutSigCount; ++index){
		if( fscanf( fp, "%s", token) != 1){
		    errRaise( optProgName, 1,
			"readPlaSignals: couldn't read an output name");
		}
		OutSigs[ index ].name = util_strsav( token );
		OutSigs[ index ].used = FALSE;
	    }
	    if( index != OutSigCount){
		errRaise( optProgName, 1, "%s (%d) %s (%d)",
			"readPlaSignals: number of outputs", OutSigCount,
			"does not match number of output names", index);
	    }
	} else if( token[ 0 ] == '.'){
	    /* found some other control					*/
	    while( getc( fp) != '\n') ;
	} else {
	    errRaise( optProgName, 1,
		"readPlaSignals: found unkown token \"%s\"", token);
	}

	/* if found everything, leave					*/
	if(( InSigs != NIL(s_signal)) && ( OutSigs != NIL(s_signal)) &&
	   ( ProdCount != 0)){
	    break;
	}
    }

    /* strip off any ignored control directives before truth table	*/
    /* if pointing at white space now, read to next newline		*/
    /*** This might trash the first line of the TT */
    if((( c = getc( fp)) == ' ') || ( c == '\t')){
	while( getc( fp) != '\n') ;
    }
    while(( c = getc( fp )) == '.' ){
	while( getc( fp ) != '\n') ;
    }
    ungetc( c, fp);
}

/*
 * formSigNameComp
 *
 * Do string hacking to come up with a "standard" name for the
 * named signal's complement.
 */
char *
formSigNameComp( name )
char *name;
{
    int len;
    char *cname;

    len = strlen( name );
    if( name[ len - 1] == '*'){
	cname = util_strsav( name );
	cname[ len - 1 ] = '\0';
    } else {
	cname = ALLOC( char, len + 2);
	strncpy( cname, name, len);
	cname[ len ] = '*';
	cname[ len + 1 ] = '\0';
    }
    return( cname );
}

/*
 * dumpReadPla
 *
 * Dump the pla just read in.
 */
dumpReadPla()
{
    int row, col;

    if( InSigCount) printf( ".i %d\n", InSigCount);
    if( OutSigCount) printf( ".o %d\n", OutSigCount);
    if( InSigCount) printf( ".ilb");
    for( row = 0; row < InSigCount; ++row){
	printf( " %s", InSigs[ row*2 ].name);
    }
    if( InSigCount) printf( "\n");
    if( OutSigCount) printf( ".ob");
    for( row = 0; row < OutSigCount; ++row){
	printf( " %s", OutSigs[ row ].name);
    }
    if( OutSigCount) printf( "\n");
    printf( ".p %d\n", ProdCount);

    for( row = 0; row < ProdCount; ++row){
	for( col = 0; col < InSigCount; ++col){
	    if( PlaTT[ row ][ col*2 ] == '-'){
		if( PlaTT[ row ][ col*2 + 1 ] == '-'){
		    printf( "-");
		} else {
		    if( PlaTT[ row ][ col*2 + 1 ] != '1'){
			errRaise( optProgName, 1, "dumpReadPla: bad data");
		    }
		    printf( "0");
		}
	    } else {
		if(( PlaTT[ row ][ col*2 ] != '1') ||
		   ( PlaTT[ row ][ col*2 + 1] != '-')){
		    errRaise( optProgName, 1, "dumpReadPla: bad data");
		}
		printf( "1");
	    }
	}
	printf( " ");
	for( col = 0; col < OutSigCount; ++col){
	    if(( PlaTT[ row ][ InSigCount*2 + col ] == '1') ||
	       ( PlaTT[ row ][ InSigCount*2 + col ] == '0')){
		printf( "%c", PlaTT[ row ][ InSigCount*2 + col ]);
	    } else {
		errRaise( optProgName, 1, "dumpReadPla: bad data");
	    }
	}
	printf( "\n");
    }
    printf( ".e");
}
