/*
 * sanitizeMatrix.c
 *
 * This file is, with one exception, devoted to fixing the input
 * personality matrix created by Genie so it is possible to create
 * a PLA with the desired logic behavior.  The major fixes are:
 *
 * 1) Moving complementary inputs to adjacent rows in the core.  Genie
 *    occasionally allows an input to end up folded to the opposite side
 *    of the core as its compement.  This is fixed by adding a new row
 *    in the AND plane and copying the signal's column over to the new
 *    location next to the complement.  This function is purely a Genie
 *    bug fix.
 *
 * 2) Shifting around the output column order.  This is an area-saving
 *    optomization and should remain with the program no matter what
 *    the error-status of the input.
 */

#include "copyright.h"
#include "port.h"
#include "anchor.h"

extern s_options Options;

/* The data that we actually use to build the PLA			*/
extern char **Core;
extern int Rows, AndCols, OrCols;
extern s_input **Inputs;
extern s_output **Outputs;
extern int OddEvenBreak;

static char **tempCore;

/* 
 * sanitizeMatrix
 *
 * 1) If complemetary signals are not adjacent, add new columns and force
 *      them next to each other so a PLA can be created.
 *
 * 2) Arrange the output order by fold type so that ground and signal rows
 *      in the OR plane can be shared.
 */
sanitizeMatrix()
{
    tempCore = Core;
    Core = NIL(char *);

    sanitizeAndPlane();

    rearrangeOrPlane();
}

/*
 * sanitizeAndPlane
 *
 * This routine makes sure that all signals with complements are adjacent
 * to there compliments.
 */
sanitizeAndPlane()
{
    int		newCols;

    /* if illegal inputs, move them to legal locations			*/
    newCols = rearrangeInputs();

    /* set Core params and malloc memory for Core			*/
    createCore(Rows, AndCols + newCols, OrCols);

    /* copy tempCore to Core with Inputs[] shifts, and shift Outputs	*/
    copyToCore( newCols );
}

/*
 * rearrangeInputs
 *
 * Step through Inputs[], and rearrange them if they are in illegal
 * configurations.  Illegal configurations are cases where an input
 * and its complement are not in adjacent columns and/or do not enter
 * the core from the same side of the core.
 */
int
rearrangeInputs()
{
    s_input	*in, *inComp;
    int		newCols, i;

    /* set up								*/
    newCols = 0;
    for( i = 0; i < AndCols - 1; ++i){
	in = Inputs[i + newCols];
	if(in == NIL(s_input)) {
	    continue; /* no inputs at this col, goto next column	*/
	}
	/* check top input of current col				*/
	if((( inComp = in->comp) == NIL(s_input)) || ( inComp->col == -1) ||
	   ( inComp->col < in->col)){
	    ; /* no comp or comp behind input, goto bottom input	*/
	} else if( illegalComp( in, inComp)){
	    ++newCols;
	    moveComplement(in, newCols);
	    continue; /* bottom input will be OK			*/
	}

	/* check bottom input at this col				*/
	if((in = in->next) == NIL(s_input)) {
	    continue; /* no bottom input				*/
	}
	if((( inComp = in->comp) == NIL(s_input)) || ( inComp->col == -1) ||
	   ( inComp->col < in->col)){
	    ; /* no comp or comp behind input, goto bottom input	*/
	} else if( illegalComp( in, inComp)){
	    ++newCols;
	    moveComplement(in, newCols);
	}
    }

    return ( newCols );
}

/*
 * illegalComp
 *
 * Return true if input and comp are in illegal configuration
 */
int
illegalComp( in, comp)
s_input *in, *comp;
{
    if( in->col != comp->col - 1){
	return( TRUE );
    }
    if((( in->bottom == 0) && (comp->bottom == 0)) ||
       (( in->top == Rows - 1) && (comp->top == Rows - 1))){
	return( FALSE );
    }
    return( TRUE );
}

/*
 * moveComplement
 *
 * In the Inputs array, move the complement of in next to in 
 * taking in to acount the newCols offset.
 */
moveComplement(in, newCols)
s_input		*in;
int		newCols;
{
    int		i;
    s_input	*inComp, *tIn;

    /* make Inputs[] big enough						*/
    Inputs = REALLOC(s_input *, Inputs, AndCols + newCols);

    /* shift Inputs[] over from max to in->col + newCols 		*/
    for( i = AndCols + newCols - 2; i >= in->col + newCols; --i){
	Inputs[i+1] = Inputs[i];
    }

    /* take in->comp out of it's list and put it in new, empty col	*/
    inComp = in->comp;
    if(( tIn = Inputs[inComp->col + newCols]) == inComp){
	Inputs[inComp->col + newCols] = inComp->next;
    } else if( tIn->next == inComp){
	tIn->next = NIL(s_input);
    } else {
	errRaise( optProgName, -1,
		"More than 2 inputs at col %d.", inComp->col);
    }

    /* Assign the comp to its new row and make sure its the only one there. */
    Inputs[in->col+newCols] = inComp;
    inComp->next = NIL(s_input);
}

/*
 * copyToCore
 *
 * Copy the core data from the tempCore the Core.
 * This is made a little complex because some new columns have been
 * created.  The inputs in the Inputs[] have (possibly) been shifted, but
 * they still reference their old sections of the tempCore.
 */
copyToCore( newCols )
int	newCols; /* number of new colums (for shifting Outputs[])	*/
{
    int		coreCol, givenCol, row, endRow;
    s_input	*in;

    /* init everything to DOT (just to be safe)				*/
    for( row = 0; row < Rows; ++row){
	for( coreCol = 0; coreCol < AndCols+OrCols; ++coreCol){
	    Core[row][coreCol] = DOT;
	}
    }
    /* copy over And plane by inputs					*/
    for( coreCol = 0; coreCol < AndCols ; ++coreCol){
	for( in = Inputs[coreCol]; in != NIL(s_input); in = in->next){
	    if( in->col == -1) continue;
	    givenCol = in->col;
	    endRow = in->top;
	    for(row = in->bottom; row <= endRow; ++row){
		Core[row][coreCol] = tempCore[row][givenCol];
	    }
	    in->col = coreCol;
	}
    }
    /* make inputs alone in col cover all of col (top to bottom)	*/
    for( coreCol = 0; coreCol < AndCols; ++coreCol){
	in = Inputs[ coreCol ];
	if(( in == NIL(s_input)) || ( in->col == -1)) continue;
	if( in->next == NIL(s_input)){
	    in->top = Rows - 1;
	    in->bottom = 0;
	}
    }

    /* copy over Or plane directly (with offset)			*/
    for( coreCol = AndCols; coreCol < AndCols+OrCols; ++coreCol){
	for( row = 0; row < Rows; ++row){
	    Core[row][coreCol] = tempCore[row][givenCol];
	}
    }
}

/*
 * createCore
 *
 * set the Core paramaters, malloc Core
 */
createCore(rows, andCols, orCols)
int	rows, andCols, orCols;
{
    char **allocCore();

    Rows = rows;
    AndCols = andCols;
    OrCols = orCols;

    Core = allocCore(rows, andCols + orCols);
}

/* 
 * rearrangeOrPlane
 *
 * Arrange the or columns by fold type.  There are three types:
 *  1) Don't Care: This is when there are 0 or 1 outputs in the column,
 *       or there are 2 outputs, but they are separated (top of one to bottom
 *       of the other) by 1 or more empty rows (ie not adjacent). 
 *  2) Odd: This is when there are 2 outputs in the column that are
 *       adjacent (close folding), and the upper output ends on an odd row.
 *       All output columns of this type go on the left side of the OR plane.
 *  3) Even: 2 outputs in the columns closely folded with the upper one
 *       ending on an even column.  These go on the right.
 *
 * The Don't Cares go in between the Odds and Evens.
 */
rearrangeOrPlane()
{
    enum outputType	*outType;
    int			numDc, numEven, numOdd;
    int			dcIndex, oddIndex, evenIndex;
    int			row, col;
    s_output		*output, *output2, **outArray;
    int			slop, lookahead;

    /* classify the folded outputs as ODD, EVEN, or DC			*/
    outType = ALLOC( enum outputType, OrCols );
    numDc = numEven = numOdd = 0;

    for( col = 0; col < OrCols; col++ ){
	/* if ( not folded || not closely folded )			*/
	if(((output = Outputs[col]) == NIL(s_output))
		|| ((output2 = output->next) == NIL(s_output))
		|| (output->top != output2->bottom - 1)){
	    outType[ col ] = DONT_CARE;
	    numDc++;
	} else if( output->top % 2){
	    outType[ col ] = ODD;
	    numOdd++;
	} else {
	    outType[ col ] = EVEN;
	    numEven++;
	}
    }
    /* reorder the output columns					*/
    dcIndex = evenIndex = oddIndex = 0;
    outArray = Outputs;
    Outputs = ALLOC( s_output *, OrCols );
    for( col = 0; col < OrCols; col++ ){
	switch( outType[ col ] ){
	    case ODD:
		Outputs[oddIndex++] = outArray[col];
		break;
	    case DONT_CARE:
		Outputs[numOdd + dcIndex++] = outArray[col];
		break;
	    case EVEN:
		Outputs[numOdd + numDc + evenIndex++] = outArray[col];
		break;
	}
    }
    /* correct the personality matrix by replacing tempCore[row][col]	*/
    /* with Core[ row ][ output->col ] for each output, fix output->col	*/
    for( col = 0; col < OrCols; ++col ){
	if((output = Outputs[col]) == NIL( s_output)){
	    continue;
	}
	for( row = 0; row < Rows; ++row ){
	    Core[ row ][ col + AndCols ] = tempCore[ row ][ output->col ];
	}
	for( ; output != NIL ( s_output ); output = output->next ){
	    output->col = col + AndCols;
	}
    }

    /*** this may not be correct */
    /* find a good place to break between odd and even outputs		*/
    OddEvenBreak = -1;
    slop = ( OrCols - 2 * ( Options.orGndLineSep/2 ))%Options.orGndLineSep;
    if( Options.debug > 0 ){
	printf("od=%d,dc=%d,ev=%d,sl=%d\n", numOdd, numDc, numEven, slop);
    }
    for( col = Options.orGndLineSep/2 ; col < numOdd + numDc; 
	 col += Options.orGndLineSep ){
	for( lookahead = 0; lookahead <= slop; lookahead++ ){
	    if( col + lookahead >= numOdd ){
		OddEvenBreak = col + lookahead;
		break;
	    }
	}
	if( OddEvenBreak >= 0 )break;
    }
    /* not much thought given to this (ie, could be improved, possibly)	*/
    if( OddEvenBreak == -1 ){
	if ( numOdd > 0){
	    OddEvenBreak = numOdd; /* between Odd and Dc		*/
	} else if( numDc > 0){
	    OddEvenBreak = numDc; /* between Dc and Even		*/
	} else {
	    /* do all OR ground on even rows				*/ 
	    OddEvenBreak = 0;
	}
    }

    FREE( outType);
}
