/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CDSOct.c

Interface of channel definition and ordering to OCT database.

CDSOctCopyAndOpen	Copy oct facet and open new oct facet
CDSOctInput		Init atlas internal data structures = object polygons
CDSOctOutput		Put in channel tree + gwire graph and close oct facet
CDSOctGetFTerm		Get fterm of connector instance
CDSOctPurge		Purge oct channel definition structure.

Author(s):		Mitsuru Igusa
Revision history:	$Log:	CDSOct.c,v $
 * Revision 1.1  90/02/07  16:12:23  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.2  89/02/13  14:57:29  octtools
 * added better checks on open facet failures
 * 
 * Revision 1.1  89/01/27  22:36:51  octtools
 * Initial revision
 * 
 * Revision 1.7  88/05/07  19:49:03  mitch
 * Using octOpenMaster instead of MIOctOpenMasterOfInst.
 * 
 * Revision 1.6  88/04/15  16:13:40  mitch
 * Fix updating of oct facet to check if there is already a channel           
 * definition tree.  Also check for old corner bags.    
 * Fix CDSOctPurge was leaving one cornerEdge for each instance.
 * 
 * Revision 1.5  88/04/15  14:10:18  mitch
 * Put in channel path for used external channels.
 * 
 * Revision 1.4  88/04/08  11:30:46  mitch
 * Moved call to octBegin to main procedure.
 * 
 * Revision 1.3  88/04/05  16:17:42  mitch
 * Fixed bug in initializing CDSMaxNumCh used in printing the sequence file.
 * 
 * Revision 1.2  88/04/05  12:31:13  mitch
 * Made all operations off of one oct facet instead of several.
 * Divorced CDSPrintSeqFile from CDSOctOutput routine.
 * 
 * Revision 1.1  88/03/31  11:39:06  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "miutil.h"
#include "CDSInc.h"
#include "vov.h"

/* Get oct input (boundaries of all instances)
----------------------------------------------------------------------------*/
CDSOctCopyAndOpen(cellName, oldView, newView, octMain)
    String	cellName;
    String	oldView, newView;
    octObject	*octMain;
{
    octObject	octOldMain, octProp;

    MI_MOD_START("CDSOctCopyAndOpen");

    if (ohOpenFacet(&octOldMain, cellName, oldView, "contents", "r") < OCT_OK) {
	octError("can not open the input facet");
	exit(-1);
    }
    OH_ASSERT(octCloseFacet(&octOldMain));
    /* Copy facet */
    octOldMain.contents.facet.facet = "interface";
    *octMain = octOldMain;
    octMain->contents.facet.view = newView;
    MSG(1, ("Copying interface facet <%s> view=<%s> to view=<%s>\n",
	    cellName, oldView, newView));
    if (octCopyFacet(octMain, &octOldMain) < OCT_OK) {
	octError("can copy the input interface to the output facet");
	exit(-1);
    }
    VOVinputFacet( &octOldMain );
    VOVoutputFacet( octMain );
    octOldMain.contents.facet.facet = octMain->contents.facet.facet = "contents";
    MSG(1, ("Copying contents facet <%s> view=<%s> to view=<%s>\n",
	    cellName, oldView, newView));
    if (octCopyFacet(octMain, &octOldMain) < OCT_OK) {
	octError("can copy the input to the output facet");
	exit(-1);
    }
    VOVinputFacet( &octOldMain );
    VOVoutputFacet( octMain );
    /* Open new contents facet */
    octMain->contents.facet.mode = "a";
    if (octOpenFacet(octMain) < OCT_OK) {
	octError("can not open the output facet");
	exit(-1);
    }
    OH_ASSERT(ohGetOrCreatePropStr(octMain, &octProp, "EDITSTYLE", "SYMBOLIC"));
    OH_ASSERT(ohGetOrCreatePropStr(octMain, &octProp, "VIEWTYPE" , "SYMBOLIC"));
    OH_ASSERT(ohGetOrCreatePropStr(octMain, &octProp, "CELLCLASS", "MODULE"));

    MI_MOD_END();
}				/* ............................. End of procedure CDSOctCopyAndOpen */
/* Get oct input (boundaries of all instances)
----------------------------------------------------------------------------*/
CDSOctInput(octMain)

octObject	*octMain;
{
#define		MAX_OCT_PTS 1000
octObject	octInst;
octObject	octInstMaster;
octObject	octInstBag;
octObject	octCInstBag, *octCInstParent;
octObject	octFTerm;
octObject	octInstTypeProp;
octObject	blkLayer, blkGeom, blkPriority;
struct octBox	blkBox;
octGenerator	genInst, genBlkGeom;
Boolean		fTermExistsFlag = FALSE;
int		CDSOctGetFTerm();
Boolean		orientFlg;
int32		numOctPts;
struct octPoint	octPts[MAX_OCT_PTS];
short		o, p;
CDSLocDef	*bdy;
CDSBlkDef	*b;

MI_MOD_START("CDSOctInput");
					/* Get blockages first */
CDSBlk = NULL;
if (ohGetByLayerName(octMain,&blkLayer,"PLACE")==OCT_OK &&
    CDSParm.useBlkFlg) {
    OH_ASSERT(octInitGenContents(&blkLayer,OCT_GEO_MASK,&genBlkGeom));
    while (octGenerate(&genBlkGeom, &blkGeom)==OCT_OK) {
	OH_ASSERT(octBB(&blkGeom, &blkBox));

	b = ALLOC_TYPE(CDSBlkDef);	/* Singly linked list */
	b->next = CDSBlk;
	CDSBlk = b;
					/* Set the box dimensions */
	b->xLo = blkBox.lowerLeft .x;
	b->xHi = blkBox.upperRight.x;
	b->yLo = blkBox.lowerLeft .y;
	b->yHi = blkBox.upperRight.y;
					/* Set the priority */
	b->priority = 1;
	if (ohGetByPropName(&blkGeom,&blkPriority,"priority")==OCT_OK) {
	    if (blkPriority.contents.prop.type!=OCT_INTEGER)
		MIerror("priority should be of type integer\n");
	    else
		b->priority = blkPriority.contents.prop.value.integer;
	}
	b->priority = MIN(b->priority, MAXINT-1);

	MSG(-1, ("Channel blockage at (%d %d) (%d %d) priority=%d\n",
	    b->xLo, b->yLo, b->xHi, b->yHi, b->priority));
    }
}

OH_ASSERT(ohGetByBagName(octMain, &octInstBag, "INSTANCES"));
if (ohGetByBagName(octMain, &octCInstBag, "CONNECTORS")==OCT_OK)
then octCInstParent = &octCInstBag;
else octCInstParent = NULL;
					/* Count number of instances */
octInitGenContents(&octInstBag, OCT_INSTANCE_MASK, &genInst);
for (CDSHead.numObj=1; octGenerate(&genInst, &octInst)==OCT_OK;
     CDSHead.numObj++);
if (octCInstParent) {
    octInitGenContents(octCInstParent, OCT_INSTANCE_MASK, &genInst);
    while (octGenerate(&genInst, &octInst)==OCT_OK) {
	if (CDSOctGetFTerm(&octInst, &octFTerm)) {
	    CDSHead.numObj++;
	    fTermExistsFlag = TRUE;
	}
    }
}
MSG(1, ("Number of instances = %d\n", CDSHead.numObj-1));
CDSObj = ALLOC_ARRAY(CDSHead.numObj, CDSObjDef);

octInitGenContents(&octInstBag, OCT_INSTANCE_MASK, &genInst);
for (o=0; octGenerate(&genInst, &octInst)==OCT_OK; o++) {
					/* Open instance master */
    OH_ASSERT(ohOpenMaster(&octInstMaster, &octInst, "interface", "r"));
    VOVinputFacet( &octInstMaster );
    numOctPts = MAX_OCT_PTS;		/* Get bounding box or polygon */
    MIOctGetInstBdy(&octInstMaster, &octInst, octPts, &numOctPts, &orientFlg);

    bdy = ALLOC_ARRAY(numOctPts, CDSLocDef);
    for (p=0; p<numOctPts; p++) {	/* ASSUME counter clockwise !! */
	bdy[p][CDS_X] = (int)octPts[p].x;
	bdy[p][CDS_Y] = (int)octPts[p].y;
    }

    CDSObj[o].dbRef	= STR_COPY(octInst.contents.instance.name);
    CDSObj[o].dbRefId  =  octInst.objectId;
    CDSObj[o].numBdy	= numOctPts;
    CDSObj[o].bdy	= bdy;
    CDSObj[o].bdyOrient	= (orientFlg ? CDS_CW : CDS_CCW);
    CDSObj[o].ioPadFlag = FALSE;
    if (ohGetByPropName(&octInst, &octInstTypeProp, "PLACEMENT_CLASS")==OCT_OK)
	if (!strcmp(octInstTypeProp.contents.prop.value.string,"PAD"))
	    CDSObj[o].ioPadFlag = TRUE;
    if (ohGetByPropName(&octInst, &octInstTypeProp, "CELLTYPE")==OCT_OK)
	if (!strcmp(octInstTypeProp.contents.prop.value.string,"PAD"))
	    CDSObj[o].ioPadFlag = TRUE;

    MSG(2, ("obj %d: <%s> numBdy=%d ioPadFlg=%c\n",
	o, CDSObj[o].dbRef, CDSObj[o].numBdy, CDSObj[o].ioPadFlag ? 'y' : 'n'));
}

if (!fTermExistsFlag) goto CDSOctInputEnd;
					/* Do the formal terminals */

octInitGenContents(octCInstParent, OCT_INSTANCE_MASK, &genInst);
while (octGenerate(&genInst, &octInst)==OCT_OK) {
    if (!CDSOctGetFTerm(&octInst, &octFTerm)) continue;

    MSG(0, ("Including fterm %s of inst %s\n",
	octInst.contents.instance.name, octFTerm.contents.term.name));
    OH_ASSERT(ohOpenMaster(&octInstMaster, &octInst, "interface", "r"));
    VOVinputFacet( &octInstMaster );
    numOctPts = MAX_OCT_PTS;		/* Get bounding box or polygon */
    MIOctGetInstBdy(&octInstMaster, &octInst, octPts, &numOctPts, &orientFlg);

    bdy = ALLOC_ARRAY(numOctPts, CDSLocDef);
    for (p=0; p<numOctPts; p++) {	/* ASSUME counter clockwise !! */
	bdy[p][CDS_X] = (int)octPts[p].x;
	bdy[p][CDS_Y] = (int)octPts[p].y;
    }

    CDSObj[o].dbRef	= STR_COPY(octInst.contents.instance.name);
    CDSObj[o].dbRefId  =  octInst.objectId;
    CDSObj[o].numBdy	= numOctPts;
    CDSObj[o].bdy	= bdy;
    CDSObj[o].bdyOrient	= (orientFlg ? CDS_CW : CDS_CCW);
    CDSObj[o].ioPadFlag = FALSE;
    if (ohGetByPropName(&octInst, &octInstTypeProp, "PLACEMENT_CLASS")==OCT_OK)
	if (!strcmp(octInstTypeProp.contents.prop.value.string,"PAD"))
	    CDSObj[o].ioPadFlag = TRUE;
    if (ohGetByPropName(&octInst, &octInstTypeProp, "CELLTYPE")==OCT_OK)
	if (!strcmp(octInstTypeProp.contents.prop.value.string,"PAD"))
	    CDSObj[o].ioPadFlag = TRUE;
    if (!CDSObj[o].ioPadFlag)
	MSG(-1, ("Instance %s has a formal terminal but is not an io pad\n",
	    octInst.contents.instance.name));

    MSG(2, ("obj %d: <%s> numBdy=%d\n", o, CDSObj[o].dbRef, CDSObj[o].numBdy));

    o++;
}

CDSOctInputEnd:
MI_MOD_END();
} /* ............................. End of procedure CDSOctInput */
/* Get fTerm of connector instance (if fterm found then return 1 else ret 0)
----------------------------------------------------------------------------*/
int
CDSOctGetFTerm(octInst, octFTerm)

octObject	*octInst;
octObject	*octFTerm;
{
octObject	octTerm;
octGenerator	genTerm;
octGenerator	genFTerm;
Boolean		fTermInstFlag = FALSE;

MI_MOD_START("CDSOctGetFTerm");

octInitGenContents(octInst, OCT_TERM_MASK, &genTerm);
while (octGenerate(&genTerm, &octTerm)==OCT_OK) {
    OH_ASSERT(octInitGenContainers(&octTerm, OCT_TERM_MASK, &genFTerm));
    if (octGenerate(&genFTerm, octFTerm)==OCT_OK) {
	if (octIdIsNull(octFTerm->contents.term.instanceId)) {
	    fTermInstFlag = TRUE;
	    break;
	}
    }
    MIerror("Expecting a formal term on aterm %s of inst %s\n",
	octTerm.contents.term.name, octInst->contents.instance.name);
}

MI_MOD_END();
if (fTermInstFlag) return(1);
else		   return(0);
} /* ............................. End of procedure CDSOctGetFTerm */
/* Create channel definition view of oct data
----------------------------------------------------------------------------*/
CDSOctAddEmptyRooms(octMain)

octObject	*octMain;
{
    int		o;
    int		numEmptyRoom = 0;
    char		*techName;
    octObject	emptyMaster, emptyMasterInterface;
    octObject	octInst, octProp, octLayer, octGeom, instBag;

    MI_MOD_START("CDSOctAddEmptyRooms");

    for (o=0; o<CDSHead.numObj; o++)
      if (!CDSObj[o].dbRef) break;
    if (o==CDSHead.numObj)	/* No empty rooms to add */
      goto CDSOctAddEmptyRoomsEnd;
    /* Open or create emptyMaster */
    if (ohOpenFacet(&emptyMaster,"emptyRoom","symbolic","contents","a")==OCT_NEW_FACET) {
	VOVoutputFacet( &emptyMaster );
	OH_ASSERT(ohGetByPropName(octMain, &octProp, "TECHNOLOGY"));
	techName = octProp.contents.prop.value.string;
	ohGetOrCreatePropStr(&emptyMaster, &octProp, "TECHNOLOGY", techName);
	ohGetOrCreatePropStr(&emptyMaster, &octProp, "EDITSTYLE", "SYMBOLIC");
	ohGetOrCreatePropStr(&emptyMaster, &octProp, "VIEWTYPE" , "SYMBOLIC");
	ohGetOrCreatePropStr(&emptyMaster, &octProp, "CELLCLASS", "MODULE");
	ohGetOrCreateLayer(&emptyMaster, &octLayer, "PLACE");
	ohCreateBox(&octLayer, &octGeom, 0, 0, CDSEmptyRoomDimX, CDSEmptyRoomDimY);
	octCloseFacet(&emptyMaster);
	emptyMasterInterface = emptyMaster;
	emptyMasterInterface.contents.facet.facet = "interface";
	OH_ASSERT(octCopyFacet(&emptyMasterInterface, &emptyMaster));
    }
    VOVinputFacet( &emptyMaster );

    OH_ASSERT(ohGetOrCreateBag(octMain, &instBag, "INSTANCES"));

    /* Create emptyRoom instances */
    for (o=0; o<CDSHead.numObj; o++) {
	if (CDSObj[o].dbRef) continue;

	sprintf(MIBuf,"emptyRoom%d", numEmptyRoom);
	CDSObj[o].dbRef = STR_COPY(MIBuf);
	MSG(-1, ("Adding empty room %s at (%d %d)\n",
		 CDSObj[o].dbRef, CDSObj[o].bdy[0][CDS_X], CDSObj[o].bdy[0][CDS_Y]));

	OH_ASSERT(ohCreateInstance(octMain, &octInst, CDSObj[o].dbRef,
				   emptyMaster.contents.facet.cell, emptyMaster.contents.facet.view,
				   CDSObj[o].bdy[0][CDS_X], CDSObj[o].bdy[0][CDS_Y], OCT_NO_TRANSFORM));
	OH_ASSERT(octAttach(&instBag, &octInst));

	numEmptyRoom++;
    }

 CDSOctAddEmptyRoomsEnd:
    MI_MOD_END();
}				/* ............................. End of procedure CDSOctAddEmptyRooms */
/* Create channel definition view of oct data
----------------------------------------------------------------------------*/
CDSOctOutput(octMain, showChanFlg)

octObject	*octMain;
Boolean		showChanFlg;
{
octObject	octParent;
octObject	octChan, octInst, octChOrd;
octObject	octPath;
octObject	octGraph, octGNode, octGEdge, octGEdgeChan, octGProp;
octObject	octCorner, octCEdge, octCEdgeChan, octCProp;
octObject	octECProp;
octObject	octLayer;		/* Stuff to show slices in oct */
octObject	octChild;
octGenerator	genChild;

int		numOctPts;
struct octPoint	octPts[MAX_OCT_PTS];
octId		*treeOctId;

int		stkRef;
short		k, rpd, tId, bdyIdx;
short		s, o, c;
CDSTreeDef	*t;
CDSGNodeDef	*g, *gn;

MI_MOD_START("CDSOctOutput");

if (*debugFlag>=0)
    CDSChkFTermAttachedToInstanceBag("Before CDSOctOutput", octMain);

#ifdef CDS_PRUNE_TREE
{					/* Prune tree */
CDSTreeDef	*t, *tp;

stkRef = MIStackPtr;
tp = NULL;
MI_STACK_PUSH(CDSTree);
while (MIStackPtr!=stkRef) {
    MI_STACK_POP(CDSTreeDef*, t);
    if (!t) continue;
    if (t->type!=CDS_TREE_ELEMENT) {	/* Not a leaf cell */
	MI_STACK_PUSH(t->child[CDS_LO]);
	MI_STACK_PUSH(t->child[CDS_HI]);
    }
    if (t->type==CDS_TREE_EXT_DONE &&	/* External channel to be pruned */
	!t->useExtFlag) {
	if (t->child[CDS_LO]) tc = t->child[CDS_LO];
	else		      tc = t->child[CDS_HI];
	if (!tp) {
	    CDSTree = tc;
	    tc->parentId = CDS_NULL_TREEID;
	}
	else {
	    DASSERT(t->parentId==tp->id);
	    if (tp->child[CDS_LO]) tp->child[CDS_LO] = tc;
	    else                   tp->child[CDS_HI] = tc;
	    tc->parentId = tp->id;
	}
	FREE_TYPE(t, CDSTreeDef);
	continue;
    }
    tp = t;
}
}
#endif

treeOctId = ALLOC_ARRAY(CDSHead.numTree, octId);
stkRef = MIStackPtr;
MI_STACK_PUSH(CDSTree);

while (MIStackPtr!=stkRef) {		/* Traverse tree just to create chan */
    MI_STACK_POP(CDSTreeDef*, t);
    if (!t) continue;
    MSG(1, ("Tree %d: type=%d\n", CDS_ID(t), t->type));
    DASSERT(t->type!=CDS_TREE_NOT_DONE);

    if (t->parentId==CDS_NULL_TREEID) {
	if (ohGetByBagName(octMain, &octChan, "CDChanDef")==OCT_OK)
	    OH_ASSERT(octDelete(&octChan));
	OH_ASSERT(ohCreateBag(octMain, &octChan, "CDChanDef"));
    }
    else {
	OH_ASSERT(ohGetById(&octParent, treeOctId[t->parentId]));
	OH_ASSERT(ohCreateBag(&octParent, &octChan, "CDChanDef"));
    }
    treeOctId[t->id] = octChan.objectId;

    MSG(2, ("On oct channel id=%s\n", octFormatId(octChan.objectId)));
    if (t->type!=CDS_TREE_ELEMENT) {	/* Not a leaf cell */
	MI_STACK_PUSH(t->child[CDS_LO]);
	MI_STACK_PUSH(t->child[CDS_HI]);
    }
    if (t->type==CDS_TREE_EXT_DONE) {
	if (!t->useExtFlag) {		/* External channel - no routing */
	    OH_ASSERT(ohCreatePropInt(&octChan, &octECProp,
		"CDOuterChannel", TRUE));
	}
	else {				/* External channel - routing allowed */
	    MSG(-1, ("Use %s external channel\n",
		(t->id==CDS_W_EXT_TREEID ? "left" :
		(t->id==CDS_E_EXT_TREEID ? "right" :
		(t->id==CDS_S_EXT_TREEID ? "bottom" : "top"))) ));
	}
    }
}

CDSMaxNumCh = 0;
stkRef = MIStackPtr;
MI_STACK_PUSH(CDSTree);

while (MIStackPtr!=stkRef) {		/* Traverse tree again */
    MI_STACK_POP(CDSTreeDef*, t);
    if (!t) continue;
    MSG(1, ("Tree %d: type=%d order=%d\n", CDS_ID(t), t->type, t->order));
    DASSERT(t->type!=CDS_TREE_NOT_DONE);

    OH_ASSERT(ohGetById(&octChan, treeOctId[t->id]));

    MSG(2, ("On oct channel id=%s\n", octFormatId(octChan.objectId)));
    if (t->type!=CDS_TREE_ELEMENT) {	/* Not a leaf cell */
	MI_STACK_PUSH(t->child[CDS_LO]);
	MI_STACK_PUSH(t->child[CDS_HI]);

	OH_ASSERT(ohCreatePropInt(&octChan, &octChOrd,
	    "CDChanOrder", t->order));
	if (t->type!=CDS_TREE_EXT_DONE || t->useExtFlag)
	    CDSMaxNumCh = MAX(CDSMaxNumCh, t->order);

	if (t->type>0 || t->useExtFlag) { /* Init split path */
	    if (t->useExtFlag)
	    then numOctPts = 2;
	    else numOctPts = (int)t->type;
	    URANGE(numOctPts, MAX_OCT_PTS);
	    for (s=0; s<numOctPts; s++) {
		octPts[s].x = (octCoord)t->slice[s][CDS_X];
		octPts[s].y = (octCoord)t->slice[s][CDS_Y];
		MSG(3, ("Slice %2d: (%d %d)\n", s, octPts[s]));
	    }
					/* Attach path to channel */
	    OH_ASSERT(ohCreatePath(&octChan, &octPath, 3));
	    OH_ASSERT(octPutPoints(&octPath, numOctPts, octPts));
	    if (showChanFlg) {
		OH_ASSERT(ohGetOrCreateLayer(octMain, &octLayer,
		    (t->type==2 ? "MET1" : "POLY")));
		OH_ASSERT(octAttach(&octLayer, &octPath));
	    }
	}
    }
    else {				/* Leaf cell: */
					/* Root of gw polygon around inst */
	OH_ASSERT(ohCreateBag(&octChan, &octGraph, "CDGraph"));
					/* First point in polygon */
	octGNode.type = OCT_POINT;	
	octGNode.contents.point.x = t->node->loc[CDS_X];
	octGNode.contents.point.y = t->node->loc[CDS_Y];
	OH_ASSERT(octCreate(&octGraph, &octGNode));
					/* Init polygon around leaf cell */
	numOctPts = 0;			/*   clockwise for now */
	for (g=t->node, rpd=CDS_E; TRUE; g=gn, rpd=CDS_REVDIR(k), numOctPts++) {
	    for (k=CDS_NEXTDIR(rpd,CDS_CCW); !g->bch[k].node;
		 k=CDS_NEXTDIR(k,CDS_CCW));
	    gn = g->bch[k].node;
	    DASSERT(g->bch[k].objNum==t->objNum);

	    MSG(3, ("Polygon %2d: (%d %d)\n",
		numOctPts, gn->loc[CDS_X], gn->loc[CDS_Y]));

	    tId = g->bch[k].treeId;
	    DASSERT(tId!=CDS_NULL_TREEID);
	    OH_ASSERT(ohGetById(&octGEdgeChan, treeOctId[tId]));
	    OH_ASSERT(ohCreateBag(&octGEdgeChan, &octGEdge, "CDGraphEdge"));
	    OH_ASSERT(octAttach(&octGEdge, &octGNode));
					/* TODO: Nonzero edge width property */
	    OH_ASSERT(ohCreatePropInt(&octGEdge, &octGProp, "CDEdgeWidth", 0));

	    if (gn==t->node) break;
					/* Next node point in polygon */
	    octGNode.contents.point.x = (octCoord)gn->loc[CDS_X];
	    octGNode.contents.point.y = (octCoord)gn->loc[CDS_Y];
	    OH_ASSERT(octCreate(&octGEdge, &octGNode));
	}

	o = t->objNum;
	if (o>=0) {			/* Attach instance to leaf channel */
	    MSG(1, ("Put corners on inst %s\n", CDSObj[o].dbRef));
	    OH_ASSERT(ohGetByInstName(octMain, &octInst,
		(String)CDSObj[o].dbRef));
	    OH_ASSERT(octAttach(&octChan, &octInst));
					/* Attach corner info to inst */
	    if (ohGetByBagName(&octInst, &octCorner, "CDCorners")==OCT_OK)
		OH_ASSERT(octDelete(&octCorner));
	    OH_ASSERT(ohCreateBag(&octInst, &octCorner, "CDCorners"));
					/* Each corner */
	    for (c=0; c<CDSObj[o].numCorn; c++) {
		if (CDSObj[o].bdyOrient==CDS_CW)
		then bdyIdx = CDSObj[o].numBdy-CDSObj[o].corn[c].bdyIdx-1;
		else bdyIdx = CDSObj[o].corn[c].bdyIdx;
		MSG(2, ("Corner %d = bdy %d\n", c, bdyIdx));
		OH_ASSERT(ohCreatePropInt((c==0 ? &octCorner : &octCEdge),
		    &octCProp, "CDCorner", bdyIdx));
					/* Attach to channel */
		tId = CDSObj[o].corn[c].treeId;
		DASSERT(tId!=CDS_NULL_TREEID);
		OH_ASSERT(ohGetById(&octCEdgeChan, treeOctId[tId]));
		OH_ASSERT(ohCreateBag(&octCEdgeChan, &octCEdge, "CDCornerEdge"));
		OH_ASSERT(octAttach(&octCEdge, &octCProp));
	    }
	}
    }
}

FREE_ARRAY(CDSHead.numTree, treeOctId, octId);

/* Wipe out stuff on PLACE layer */
if (ohGetByLayerName(octMain, &octLayer,"PLACE")==OCT_OK) {
    octInitGenContents(&octLayer, OCT_ALL_MASK, &genChild);
    while (octGenerate(&genChild, &octChild)==OCT_OK) {
	OH_ASSERT(octDelete(&octChild));
    }
}

if (*debugFlag>=0)
    CDSChkFTermAttachedToInstanceBag("After CDSOctOutput", octMain);

MI_MOD_END();
} /* ............................. End of procedure CDSOctOutput */
/* Purge oct channel definition stuff
----------------------------------------------------------------------------*/
int
CDSOctPurge(octMain)

octObject	*octMain;
{
octObject	octChan;
octObject	octSubChan;
octObject	octPath;
octObject	octECProp;
octObject	octChOrd;
octObject	octInst;
octObject	octSubStuff;
octGenerator	genSubChan;
octGenerator	genSubStuff;

octObject	octGraph;		/* Oct objects for polygon */
octObject	octGEdge;
octObject	octGPoint;

octObject	octCorn;		/* Oct objects for corners */
octObject	octCEdge;
octObject	octCHead;

int		stkRef;
Boolean		leafCellFlg;
int		rc = 0;

MI_MOD_START("CDSOctPurge");

if (ohGetByBagName(octMain, &octChan, "CDChanDef")!=OCT_OK) {
    rc = 4;
    goto CDSOctPurgeEnd;
}

stkRef = MIStackPtr;
MI_STACK_PUSH(octChan.objectId);

while (MIStackPtr!=stkRef) {		/* Traverse channel tree */
    MI_STACK_POP(octId, octChan.objectId);

    octInitGenContents(&octChan, OCT_BAG_MASK, &genSubChan);
    for (leafCellFlg=TRUE; octGenerate(&genSubChan, &octSubChan)==OCT_OK; ) {
	if (strcmp(octSubChan.contents.bag.name, "CDChanDef")) continue;
	MI_STACK_PUSH(octSubChan.objectId);
	leafCellFlg = FALSE;
    }

    if (!leafCellFlg) {
	if (ohGetByPropName(&octChan, &octECProp, "CDOuterChannel")==OCT_OK) {
	    OH_ASSERT(octDelete(&octECProp));
	}
	else {
	    OH_ASSERT(octGenFirstContent(&octChan, OCT_PATH_MASK, &octPath));
	    OH_ASSERT(octDelete(&octPath));
	}
	OH_ASSERT(ohGetByPropName(&octChan, &octChOrd, "CDChanOrder"));
	OH_ASSERT(octDelete(&octChOrd));
	OH_ASSERT(octDelete(&octChan));
    }
    else {
					/* Get rid of corner stuff */
	OH_ASSERT(octGenFirstContent(&octChan, OCT_INSTANCE_MASK, &octInst));
	OH_ASSERT(ohGetByBagName(&octInst, &octCHead, "CDCorners"));
	OH_ASSERT(ohGetByPropName(&octCHead, &octCorn, "CDCorner"));
	OH_ASSERT(octDelete(&octCHead));
	while (TRUE) {			/* Go through rest of corner chain */
	    OH_ASSERT(octGenFirstContainer(&octCorn, OCT_BAG_MASK, &octCEdge));
	    OH_ASSERT(octDelete(&octCorn));
	    if (octGenFirstContent(&octCEdge, OCT_PROP_MASK, &octCorn)!=OCT_OK)
		break;
	    OH_ASSERT(octDelete(&octCEdge));
	}
	OH_ASSERT(octDelete(&octCEdge));
					/* Get rid of graph stuff */
	if (ohGetByBagName(&octChan, &octGraph, "CDGraph")==OCT_OK) {
	    OH_ASSERT(octGenFirstContent(&octGraph, OCT_POINT_MASK, &octGPoint));
	    OH_ASSERT(octDelete(&octGraph));
	    while (!octIdIsNull(octGPoint.objectId)) {
		OH_ASSERT(octGenFirstContainer(&octGPoint, OCT_BAG_MASK, &octGEdge));
		DASSERT(!strcmp(octGEdge.contents.bag.name, "CDGraphEdge"));
		OH_ASSERT(octDelete(&octGPoint));
		octGPoint.objectId = oct_null_id;
		octInitGenContents(&octGEdge, OCT_ALL_MASK, &genSubStuff);
		while (octGenerate(&genSubStuff, &octSubStuff)==OCT_OK) {
		    if (octSubStuff.type==OCT_POINT)
			octGPoint = octSubStuff;
		    else OH_ASSERT(octDelete(&octSubStuff));
		}
		OH_ASSERT(octDelete(&octGEdge));
	    }
	}

	OH_ASSERT(octDelete(&octChan));
    }
}

CDSOctPurgeEnd:
MI_MOD_END();
return(rc);
} /* ............................. End of procedure CDSOctPurge */
