// This is a C++ source file.

// Use the BOSS macros to describe the layout of a shifter cell.
// You should be reasonably familiar with OCT and the OCT Symbolic
// policy to understand fully this description.

// The description is constructive and procedural, one piece at a time.
// It helps to draw a colored stick diagram to understand the description.
// 
#include "String.h"
#include "boss.h"		
#include "utility.h"
#include "symbolic.h"
#include "oh.h"
#include "errtrap.h"
#include "options.h"
#include "th.h"

void testSide( int q, BossDirection d, int& m, BossDirection& dir ) 
{
    if ( q < m ) {
	dir=d;
	m=q;
    }
}

BossDirection 	side( octObject& term, octBox& bb )
{
    octBox tBB;
    ohTerminalBB( &term, &tBB );
    BossDirection dir = -1;
    int mindist = 5 * thLambda();
    
    testSide(tBB.lowerLeft.x-bb.lowerLeft.x,BOSS_LEFT,mindist,dir);
    testSide(tBB.lowerLeft.y-bb.lowerLeft.y,BOSS_DOWN,mindist,dir );
    testSide(bb.upperRight.x - tBB.upperRight.x,BOSS_RIGHT,mindist,dir);
    testSide(bb.upperRight.y - tBB.upperRight.y,BOSS_UP,mindist,dir);

    octObject master, inst, ft, at, path;

    ohGetById(&inst,term.contents.term.instanceId );
    ohOpenMaster( &master,&inst,"contents", "r" );
    
    OH_ASSERT(ohGetByTermName( &master, &ft, term.contents.term.name ));
    if ( octGenFirstContent( &ft,OCT_TERM_MASK,&at ) == OCT_OK ) {
      if ( octGenFirstContainer(&at,OCT_PATH_MASK,&path) == OCT_OK ) {
	struct octPoint *points;
	int32 num;
	ohGetPoints( &path, &num, &points );
	if ( points[0].x == points[1].x ) {
	  /* Path is vertical, allow only vertical exit */
	  if ( dir != BOSS_UP && dir != BOSS_DOWN ) dir = -1;
	} else if ( points[0].x == points[1].x ) {
	  /* Path is horizontal, allow only horizontal exit */
	  if ( dir != BOSS_LEFT && dir != BOSS_RIGHT ) dir = -1;
	}
      }
    }
    

    return dir;			/* No side. */
}


void bossArray(char* arrayCell, char* arrayView, 
	       char *elementCell, char* elementView, 
	       int rows, int cols,
	       int flip_row, int flip_col,
	       int centerTerms )
{
    BossFacet arrayFacet = BossFacet( arrayCell, arrayView); 	
    thTechnology( &arrayFacet );

    BossInstArray array( rows, cols );
    array.populate_all( arrayFacet, elementCell, elementView, "c" );

    if ( centerTerms ) {
	int offset = -2 * thLambda();
	array.offsetBB_all( offset, offset, offset, offset );
    }
    if ( flip_row ) {
	for ( int i = 0; i < rows ; i++ ) {
	    if ( i & 0x1 ) {
		array.transform_row( OCT_MIRROR_Y, i );
	    }
	}
    }
    if ( flip_col ) {
	for ( int j = 0; j < cols ; j++ ) {
	    if ( j & 0x1 ) {
		array.transform_col( OCT_MIRROR_X, j );
	    }
	}
    }
    array.place_all();
    
    cout << "\nConnecting all cells by abutment...\n";

    for (int i = 0 ; i < rows ; i++ ) {
	for ( int j = 0 ; j < cols ; j++ ) {
	    BossInstance& inst = array.elmt( i, j );
	    symReconnect( &inst, 0 , 0 );
	}	
    }
    
    /* Promote the terminals */
    octBox arrayBB;
    octBB( &arrayFacet, &arrayBB );

    octGenerator genTerm;
    BossInstance& inst = array.elmt( 0, 0 ); /* Get one of the instances. */
    octObject term;
    octInitGenContents( &inst, OCT_TERM_MASK, &genTerm );
    while ( octGenerate( &genTerm, &term )==OCT_OK ) {
	String name = term.contents.term.name;
	String ftermName;
	switch ( side( term, arrayBB ) ) {
	case BOSS_LEFT:
	    ftermName = "L." + name;
	    cout.form( "Promoting %s along LEFT\n", (char*)ftermName );
	    array.promote_col( arrayFacet, 0, (char*)name, (char*)ftermName );
	    break;
	case BOSS_DOWN:
	    ftermName = "B." + name;
	    cout.form( "Promoting %s along BOTTOM\n", (char*)ftermName );
	    array.promote_row( arrayFacet, 0, (char*)name, (char*)ftermName );
	    break;
	default:
	    ;
	}
    }


    inst = array.elmt( rows-1, cols-1 ); /* Get one of the instances. */
    octInitGenContents( &inst, OCT_TERM_MASK, &genTerm );
    while ( octGenerate( &genTerm, &term )==OCT_OK ) {
	String name = term.contents.term.name;
	String ftermName;
	
	switch ( side( term, arrayBB ) ) {
	case BOSS_RIGHT:
	    ftermName = "R." + name;
	    cout.form( "Promoting %s along RIGHT\n", (char*)ftermName );
	    array.promote_col( arrayFacet, cols-1, (char*)name, (char*)ftermName );
	    break;
	case BOSS_UP:
	    ftermName = "T." + name;
	    cout.form( "Promoting %s along TOP\n", (char*)ftermName );
	    array.promote_row( arrayFacet, rows-1, (char*)name, (char*)ftermName );
	    break;
	default:
	    ;
	}
    }



    arrayFacet.createInterface();
}


optionStruct optionList[] = {
    { "o", "cell:view",    "Specify output cell and view"},
    { "r", "number",       "rows" },
    { "c", "number",       "columns" },
    { "e", "cell:view",    "Element of the array" },
    { "R", 0,              "Flip every other row" },
    { "C", 0,              "Flip every other column" },
    { "T", 0,              "Center terminals" },
    { "d", "",             "Enables debug"},
    { 0 , 0 , 0 }
};


main( int argc, char** argv )
{
    int rows = 0;
    int cols = 0;
    int flip_row = 0;
    int flip_col = 0;
    int centerTerms = 0;
    extern int optind;
    extern char* optarg;
    int c;
    char *arrayName = 0;
    char *elementName = 0;
    int debug = 0;

    while ( ( c = optGetOption( argc, argv ) ) != EOF) {
	switch( c ) {
	case 'r': rows = atoi( optarg ); break;
	case 'c': cols = atoi( optarg ); break;
	case 'R': flip_row = 1; break;
	case 'C': flip_col = 1; break;
	case 'T': centerTerms = 1; break;
	case 'd': debug = 1; break;
	case 'o': arrayName = util_strsav( optarg ); break;
	case 'e': elementName = util_strsav( optarg ); break;
	default:
	    optUsage();
	}
    }
    if ( argc <= 1 ) optUsage();

    // Required: first call in main().
    bossBegin( argc, argv );

    char* arrayCell = arrayName;
    char* arrayView = strchr( arrayName, ':' );
    if ( arrayView ) {
	*arrayView++ = '\0';
    } else {
	arrayView = "symbolic";
    }

    char* elementCell = elementName;
    char* elementView = strchr( elementName, ':' );
    if ( elementView ) {
	*elementView++ = '\0';
    } else {
	elementView = "symbolic";
    }

    bossArray( arrayCell, arrayView, elementCell, elementView, rows, cols, flip_row, flip_col, centerTerms );

    // Do not call any other BOSS function in main():
    // call only bossBegin() and bossEnd().
    // 
    // bossEnd() does not return. It is like calling exit().
    // This is required: it should be the last call in main().
    bossEnd( 0 );
}

