#include "electra.h"
int signal_must_skip = 2;
/*
 * context
 *
 * This routine allocates an instance for every non DOT charactor in
 * the Matrix.  The instance is a function of the charactors type
 * ( see parseInfile ) and if the charactor is of type TRAN, the
 * charactor's value and context.
 *
 * Legal charactor types are:
 *
 *   DOT:     Nothing at this point - signals and networks are passed through.
 *   TERM:    The output of a primary network. 
 *   TRAN:    A transistor in a network. The ascii value of the charactor
 *            representing the transistor is mapped into two instances, a
 *            NMOS transistor and a PMOS transistor in the CharArray[].  Any
 *            charactor which is not mapped into a non TRAN type ( anything
 *            except the chars .oOsSpP ) is mapped into a TRAN.  This 
 *            mechanism will allow about 121 combinations of NMOS and PMOS
 *            transistors - or a choice of 11 NMOS and 11 PMOS sizes.  The
 *            context of a transistor is its placement in the network ( top
 *            and bottom transistors ) and the type of network ( series or
 *            parallel ).  It is used to determine if the transistor has
 *            contacts on the top and/or bottom.  See newTran().
 */
makeNetworks()
{
    s_charEntry *name;
    char *sName, *pName;
    int sContext, pContext;
    int firstTran, lastTran, prevTran;
    int skip, oddCol;
    mkId sTrans, pTrans;
    s_network *network;
    int row, col;
    int oddTran;
    octObject *newTerm();
    octObject *newTran();
    octObject term;
    octBox area;
    s_signal *signal;
    int runNum;
    enum octTransformType sTransform, pTransform, p0Transform, p1Transform;
    int *vddCol, *vssCol;

    vddCol = ALLOC( int, NumRows );
    vssCol = ALLOC( int, NumRows );
    for( row = NumRows - 1; row >= 0; row-- ){
	vddCol[row] = FALSE;
	vssCol[row] = FALSE;
    }

    for ( col = 0; col < NumCols; col++ ){

	oddCol = col%2;

	for ( network = NetworkArray[col]; network != NULL;
		network = network->next){
	    firstTran = network->firstTran;
	    lastTran = network->lastTran;

	    area.lowerLeft.x = area.upperRight.x = col;

	    /* set the metal masks					*/
	    for( row = network->bottom ; row <= network->top; row++ ){
		Matrix[row][col].metal1Mask += 1;
	    }
	    if( oddCol ){
		for( row = firstTran; row <= lastTran; row++ ){
		    Matrix[row][col].metal1Mask = 3;
		}
	    } else {
		for( row = firstTran; row >= lastTran; row-- ){
		    Matrix[row][col].metal1Mask = 3;
		}
	    }

	    /* set up the series and parallel instance transformations	*/
	    if( network->type == NAND ){
		sTrans = NTrans;
		pTrans = PTrans;
		if( oddCol ){
		    sTransform = OCT_MIRROR_Y;
		    p0Transform = OCT_ROT180;
		    p1Transform = OCT_MIRROR_X;
		} else {
		    sTransform = OCT_MIRROR_X;
		    p0Transform = OCT_NO_TRANSFORM;
		    p1Transform = OCT_MIRROR_Y;
		}
	    } else {
		sTrans = PTrans;
		pTrans = NTrans;
		if( oddCol ){
		    sTransform = OCT_ROT180;
		    p0Transform = OCT_MIRROR_Y;
		    p1Transform = OCT_NO_TRANSFORM;
		} else {
		    sTransform = OCT_NO_TRANSFORM;
		    p0Transform = OCT_MIRROR_X;
		    p1Transform = OCT_ROT180;
		}
	    }

	    /* Traverse the network from top to bottom ( if !oddCol )	*/
	    /* or from bottom to top ( if oddCol ). Create and attach	*/
	    /* trans as necessary.  					*/

	    oddTran = FALSE;
	    pTransform = p0Transform;

	    /* the first trans always SOURCE_CONTS			*/
	    sContext =  DRAIN_CONTS;
	    pContext = DRAIN_CONTS|SOURCE_CONTS;

	    for( prevTran = firstTran;; prevTran = row ){
		skip = FALSE;

		/* the last series transistor always has SOURCE_CONTS	*/
		/* the last parallel transistor always has a contact	*/
		if( prevTran == lastTran ){
		    sContext |= SOURCE_CONTS;
		    if ( oddTran ) pContext |= DRAIN_CONTS;
		    else           pContext |= SOURCE_CONTS;
		    
		}

		/* go to the next transistor - check for skips		*/
		/* a skip occurs if more than signal_must_skip rows	*/
		/* are passed or the network passes over a signal 	*/
		/* polyrun						*/
		if( oddCol ){
		    for ( row = prevTran + 1; row <= lastTran ; row++ ){
			if( Matrix[row][col].type  == TRAN ) break;
			if( skip ) continue;
			signal = Matrix[row][col].signal;
			if( signal == NIL( s_signal ))continue;

			/* always skip if the signal is all POLY	*/
			if( signal->numRuns == 0 ){
			    skip = TRUE;
			}

			/* skip if the network covers a polyrun		*/ 
			for( runNum = 0; runNum < signal->numRuns; runNum++ ){
			    if( signal->polyrun[ runNum ].left > col )break;
			    if( signal->polyrun[ runNum ].right >= col ){
				skip = TRUE;
				break;
			    }
			}
		    }
		    if( row > prevTran + signal_must_skip ) skip = TRUE;
		} else {
		    for ( row = prevTran - 1; row >= lastTran ; row-- ){
			if( Matrix[row][col].type  == TRAN )break;
			if( skip ) continue;
			signal = Matrix[row][col].signal;
			if( signal == NIL( s_signal ))continue;

			/* always skip if the signal is all POLY	*/
			if( signal->numRuns == 0 ){
			    skip = TRUE;
			}

			/* skip if the network covers a polyrun		*/ 
			for( runNum = 0; runNum < signal->numRuns; runNum++ ){
			    if( signal->polyrun[ runNum ].left > col )break;
			    if( signal->polyrun[ runNum ].right >= col ){
				skip = TRUE;
				break;
			    }
			}
		    }
		    if( row < prevTran - signal_must_skip ) skip = TRUE;
		}

		/* if skip, the series trans jump up to metal and the	*/
		/* parallel trans get contacted to power or the output	*/
		if( skip ){
		    sContext |= SKIP_CONTS;
		    if( oddTran ) pContext |= DRAIN_CONTS;
		    else          pContext |= SOURCE_CONTS;
		}

		/* allocate the transistor at the prevTran position	*/
		name = &charTable[(int)Matrix[prevTran][col].letter];
		if( name->set == FALSE )
		    charTableError( Matrix[prevTran][col].letter );
		if( network->type == NAND ){
		    sName = name->nmos;
		    pName = name->pmos;
		} else {
		    sName = name->pmos;
		    pName = name->nmos;
		}
		area.lowerLeft.y = 3 * prevTran + 2;
		area.upperRight.y = area.lowerLeft.y + 2;

		mkAttach(sTrans, &area, 
			            newTran( sName, sContext, sTransform));
		mkAttach(pTrans, &area, 
				    newTran( pName, pContext, pTransform));


		/* mark the vssCol and vddCol to create power contacts	*/
		if(( sContext&SOURCE_CONTS ) == SOURCE_CONTS ){
		    if( network->type == NAND ) vssCol[prevTran] = TRUE;
		    else                        vddCol[prevTran] = TRUE;
		    
		}
		if(( pContext&SOURCE_CONTS ) == SOURCE_CONTS ){
		    if( network->type == NAND ) vddCol[prevTran] = TRUE;
		    else                        vssCol[prevTran] = TRUE;
		}

		/* create the output col connector if the parallel 	*/
		/* tran has DRAIN_CONTS					*/
		if(( pContext&DRAIN_CONTS ) == DRAIN_CONTS ){
		    if( oddCol^oddTran ){
			area.upperRight.y = area.lowerLeft.y;
		    } else {
			area.lowerLeft.y = area.upperRight.y;
		    }
		    if( network->type == NAND ){
			mkAttach( NandOutput, &area, newTerm( M1TERM ));
		    } else {
			mkAttach( NorOutput, &area, newTerm( M1TERM ));
		    }
		}

		/* stop at the last transistor				*/
		if( prevTran == lastTran )break;

		/* if skip, the series trans have to come back down 	*/
		/* metal						*/
		if( skip ) sContext = DRAIN_CONTS;
		else 	   sContext = NO_CONTS;
		
		if( oddTran ){
		    pContext = SOURCE_CONTS;
		    if( skip ) pContext |= DRAIN_CONTS;
		    pTransform = p0Transform;
		    oddTran = FALSE;
		} else {
		    pContext = DRAIN_CONTS;
		    if( skip )
			pContext |= SOURCE_CONTS;
		    pTransform = p1Transform;
		    oddTran = TRUE;
		}
	    }
	}

	/* now create the power columns 				*/
	if(( col%2 == 0 )||( col == NumCols - 1 )){
	    for( row = NumRows - 1; row >= 0; row-- ){
		if( vssCol[row] == TRUE ){
		    vssCol[row] = FALSE;
		    term = *newTerm( POWERTERM );
		    area.lowerLeft.y = area.upperRight.y = 3 * row + 2;
		    mkAttach(VssArray, &area, &term );
		    area.lowerLeft.y += 2;
		    area.upperRight.y += 2;
		    mkAttach(VssArray, &area, &term );
		}
	    }
	}
	if(( col%2 == 1 )||( col == NumCols - 1 )){
	    for( row = NumRows - 1; row >= 0; row-- ){
		if( vddCol[row] == TRUE ){
		    vddCol[row] = FALSE;
		    term = *newTerm( POWERTERM );
		    area.lowerLeft.y = area.upperRight.y = 3 * row + 2;
		    mkAttach(VddArray, &area, &term );
		    area.lowerLeft.y += 2;
		    area.upperRight.y += 2;
		    mkAttach(VddArray, &area, &term );
		}
	    }
	}
    }
    FREE( vddCol );
    FREE( vssCol );
}
