#include "electra.h"
/*
 * parseIn
 *
 * parseIn reads the Infile to find the personality matrix and
 *   the topology of the array. It allocates Matrix and all signals
 *   ( horizontal signals) and networks ( vertical signals ).  
 *   The elements of Matrix are pointed to the correct signal and
 *   network.
 *
 * The expected form of the charactor array is:

 "any string" NumRows NumCols

 * Followed by a personality Matrix[NumRows][NumCols] of char.
 * It is an error for the personality matrix to be bigger than
 *   NumRows x NumCols.
 * Followed by a set of any of these:

 ( row "row specification")
 ( column "column specification" )
 ( comment "comment string which cannot contain )" )


 * The row and column specification are given in the headers
 *    to the parseRow and parseColumn procedures.
 * 
 */
parseIn()
{
    s_matrix *matrixPtr;
    FILE *fopen(), *infile;
    int row, col;
    int c;
    char  keyword[ 80 ];
    /*
     * open the file named InfileName
     * if InfileName == NULL use stdin
     */
     if ((infile = fopen( InfileName , "r" )) == NULL){
	 fprintf(stderr, "Fatal error - could not open file named %s\n",
			 InfileName);
	 exit(-1);
     }
	
    /*
     * The first keyword in the file should be an arbitrary string 
     * followed by two integers which are the size of the charactor array.
     */
    if ( fscanf( infile, "%*s %d %d", &NumRows, &NumCols ) != 2 ){
	parseError("The first line should be: new numrows numcols\n");
    } else {
	Matrix = ALLOC( s_matrix *, NumRows );
	for ( row = NumRows - 1 ; row >= 0 ; row-- ) {
	    matrixPtr = Matrix[ row ] = ALLOC( s_matrix, NumCols );
	    for ( col = NumCols - 1 ; col >= 0 ; col-- ) {
		matrixPtr->signal = NIL( s_signal );
		matrixPtr->network = NIL( s_network );
		matrixPtr->metal1Mask = 0;
		matrixPtr->metal2Mask = 0;
		matrixPtr++;
	    }
	}
    }
    /*
     * Try to read count lines of NumCols charactors.  
     * If NumCols charactors cannot be read - the matrix is too narrow. 
     * If any charactors can be read before the newline after NumCols
     *    charactors has been read - the matrix is too wide.
     * If NumRows lines cannot be read - the matrix is not deep enough.
     * If another line can be read after NumRows lines have been read -
     *    the matrix is too deep.
     */
    for ( row = NumRows - 1 ; row >= 0 ; row-- ) {
	/*
	 * eat any white space at the begining of the line
	 */
        while ((( c = getc(infile)) == ' ')||(c == '\t')||(c == '\n'));
        if ( c == '(' ) parseError("charactor matrix not deep enough");
        (void)ungetc( c , infile );
        /*
         * read the line of the charactor array
         */
        for ( col = 0 ; col < NumCols; col++ ){
	    if (( c = Matrix[ row ][ col ].letter = getc( infile )) == EOF){
	        parseError( "unexpected EOF while reading charactor matrix");
	    } else if ( c == '\n' ){
	        parseError("charactor matrix not wide enough");
	        ( void )ungetc( c , infile );
	    } 
        }
        /*
         * read to the end if the line
         */
	while ((c = getc(infile)) != '\n'){
	    switch ( c ){
	        case EOF: parseError(
			  "unexpected EOF while reading charactor matrix");
	    case ' ':
	    case '\t': break;
	    default: parseError("charactor matrix too wide");
		while(( c = getc(infile)) != '\n'){
		    if ( c == EOF ){
			 parseError(
			 "unexpected EOF while reading charactor matrix");
		    }
		}
	    }
	}
    }
    /*
     * test for extra rows - error if the next non white char is not a 'c'
     */
    while ((( c = getc(infile)) == ' ')||(c == '\t')||(c == '\n'));
    if ( c != '('/*)*/){
	parseError("charactor matrix is too deep");
    } else {
	(void)ungetc( c , infile );
    }
    /*
     * determine the type field for each element in the matrix
     */
    for ( row = NumRows - 1 ; row >= 0 ; row-- ) {
	for( col = NumCols - 1; col >= 0 ; col-- ){
	    switch ( Matrix[row][col].letter ){
	        case '.':
		    Matrix[row][col].type = DOT;
		    break;
	        case 'p':
	        case 's':
		    Matrix[row][col].type = TERM;
		    break;
	        default:
		    Matrix[row][col].type = TRAN;
		    break;
	    }
	}
    }
    /*
     * while not EOF 
     *   read a "("
     *   read the keyword
     *   call the correct procedure
     *   read a ")"
     * Error conditions:
     *   no opening or closing parenthesis
     *   nested declarations
     */  
    while ( openParen( infile ) != EOF ){
	(void)fscanf( infile ,"%s", keyword );
	if ( strcmp( keyword , "row" ) == 0 ){
	    parseRow( infile );
	} else if ( strncmp( keyword , "col", 3 ) == 0 ){
	    parseColumn( infile );
	} else if ( strncmp( keyword , "com", 3 ) == 0 ){
	    while(( c = getc( infile )) !=/*(*/ ')'){
	        if ( c == EOF ){
		    parseError("no closing parenthesis after comment\n");
		}
	    }
	    /* put the closing parenthesis back on infile */
	    (void)ungetc( c , infile );
	} else {
	    parseError( strcat( "unknown keyword: ", keyword ));
	}
	/*
	* close the parenthesis
        */
	while (( c = getc( infile )) !=/*(*/ ')') {
	    if ( c == EOF ) {
		parseError(strcat("unexpected EOF after keyword ", keyword ));
	    } else if ( c == '('/*)*/ ){
		parseError( strcat("no closing parenthesis after keyword ", 
				    keyword ));
		( void )ungetc( c , infile );
	    }
	}
    }
}
/*
 * openParen
 *
 * read until the first open parenthesis in infile
 */
openParen( infile )
FILE *infile;
{
  int c;
  while (( c = getc( infile )) != '(' /*)*/){
    if( c == EOF ){
      return ( EOF );
    }
  }
  return ( TRUE );
}

/*
 * parseRow
 *
 * This procedure parses a row declaration to derive the signal names 
 * and their locations.
 * The format for a row is:

 rownumber signalname [ signal location ][signalname [ signal location ]...

 * The signalnames are strings and the signal locations are integers.
 * Unnamed signals are denoted by an explicit empty string, ie "".
 * For example:

 ( row 3 foo ) 

 * specifies that the one signal in row three is named foo.
 * 

 (row 2 foo 0 foobar 2 "" 5 )

 * specifies that in row 2, a signal named foo starts in column 0,
 * a signal named foobar starts in column 2, and an unnamed signal
 * starts in column 5.
 *
 * The signal position is optional iff a row only contains one signal.
 */
parseRow( infile )
FILE *infile;
{
    int rownum;
    char signame[80];
    int sigpos;
    int finished = FALSE;
    s_signal *allocSignal();
    s_net *allocNet();
    /*
     * read the rownum and possibly the first signal name and position
     */
    switch( fscanf(infile,/*(*/ " %d %[^ \t)] %d", &rownum, signame, &sigpos)){
	case EOF:
	    parseError( "unexpected EOF in row declaration");
	    return;
	case 0:
	    parseError( 
	    "at least the row number must be given in a row declaration");
	    return;
	case 1:
	    /* empty row - ignore it */
	    return;
	case 2:
	    /*
	     * Got rownum and signam but not sigpos.
	     * This is a specification error, 
	     * but recoverable by setting sigpos = 0
	     */
	    sigpos = 0;
    }

    /*** until genie changes */
    rownum = NumRows - 1 - rownum;

    if (( rownum >= NumRows ) || ( rownum < 0)){
	parseError("row number is not in range of 0 to number of rows - 1");
	return;
    }
    while( finished == FALSE ) {
	/*
	 * this loop is repeated until no more signals are given
	 */
	if (( sigpos >= NumCols ) || ( sigpos < 0)){
	    parseError(strcat( strcat("starting position for signal ", signame),
			       "is out of range"));
	    return;
	}
	Matrix[ rownum ][ sigpos ].signal = allocSignal( rownum );
	Matrix[ rownum ][ sigpos ].signal->row = rownum;
	Matrix[ rownum ][ sigpos ].signal->left = sigpos;
	if ( strcmp( signame , "\"\"" ) != 0 ){
	    Matrix[ rownum ][ sigpos ].signal->net = allocNet();
	    Matrix[ rownum ][ sigpos ].signal->net->name = util_strsav( signame );
	} else {
	    Matrix[ rownum ][ sigpos ].signal->net = NULL;
	}
	switch( fscanf(infile,/*(*/ " %[^ \t)] %d", signame, &sigpos)){
	    case EOF:
		parseError( "unexpected EOF in row declaration");
		return;
	    case 0:
		/*
		 * finished with this row
		 */
		return;
	    case 1:
		/*
		 * Only the signame was given.  This is unrecoverable because
		 * the folding information is not given 
		 * If the user wishes to continue this signal can be ignored
		 */
		parseError(strcat("no starting position was given for signal ",
				   signame));
		finished = TRUE;
	}
    }
}

/*
 * parseColumn
 *
 * this procedure parses a column declaration to derive the network locations.
 * The format for a column is:

 colnumber location [ location ]...

 * The network locations are integers.
 * For example:

 (column 2 0 2 5 )

 * specifies that in column 2, a network starts in column 0,
 * a network starts in column 2, and a network starts in column 5.
 */
parseColumn( infile )
FILE *infile;
{
    int colnum;
    int netpos;
    s_network *allocNetwork();
    /*
     * read the colnum
     */
    switch ( fscanf(infile," %d %d", &colnum , &netpos )){
	case EOF:
	    parseError( "unexpected EOF in row declaration");
	    return;
	case 0:
	case 1:
	    parseError("I can't do anything with an empty column declaration");
	    return;
    }
    while ( TRUE ){
	if (( colnum >= NumCols ) || ( colnum < 0)){
	    parseError("Column number is out of range\n");
	    return;
	}
	/*** until genie changes */
	netpos = NumRows - 1 - netpos;

	if (( netpos >= NumRows ) || ( netpos < 0)){
	    parseError( "Network position is out of range\n");
	    return;
	}
	Matrix[netpos][colnum].network = allocNetwork(colnum);
	Matrix[netpos][colnum].network->col = colnum;
	Matrix[netpos][colnum].network->top = netpos;
	switch (fscanf(infile," %d", &netpos)){
	    case EOF:
		parseError("unexpected EOF in column declaration");
		return;
	    case 0:
		return;
	}
    }
}
