#include "electra.h"
#include "vulcan.h"

/*
 * createTran
 *
 * this program creates the transistors for the electra cell library.  The
 * connectors used by electra come from the standard primitive library.
 *
 * electraTran takes two command line options:
 *
 * -c celllib specifies where the transistor goes ( the celllib )
 * -{n,p} N specifies NFET or PFET transistor of width N
 */
/*
 * design rules - should eventually come from TAP
 */
#define GATE_OVERLAP 40
#define GATE_LENGTH 40
#define CONTACTED_DIFF_WIDTH 100
#define UNCONTACTED_DIFF_WIDTH 20
#define CUT_WIDTH 40
#define CUT_PITCH 80
#define M1_CUT_OVERLAP 20
#define CONTACT_WIDTH 80
#define MAX_NAKED_DS 40
/*
 * creates a transistor
 */
createTran( oInst, type, contact_mask, width, length, technology )
octObject *oInst;
enum fetType type;
unsigned contact_mask;
int width;
int length;
char *technology;
{
    octObject oFacet;
    octObject oInterface;

    openTranFacet( &oFacet, oInst, technology );
    createTranFacet( &oFacet, type, width, length, contact_mask );
    vulcanize( &oFacet, &oInterface );
    addMosProps( &oInterface, type, width );
    octCloseFacet( &oInterface );
    octCloseFacet( &oFacet );
}

/*
 * open oFacet as cell:physical:contents for write
 *
 * add the VIEWTYPE, TECHNOLOGY, and EDITSTYLE props
 */
openTranFacet( oFacet, oInst, technology )
octObject *oFacet;
octObject *oInst;
char *technology;
{
    octObject oProp;

    oFacet->type = OCT_FACET;
    oFacet->contents.facet.cell = oInst->contents.instance.master;
    oFacet->contents.facet.view = oInst->contents.instance.view;
    oFacet->contents.facet.facet = "contents";
    oFacet->contents.facet.version = oInst->contents.instance.version;
    oFacet->contents.facet.mode = "w";

    if( octOpenFacet( oFacet ) < OCT_OK ){
	fprintf( stderr, "Could not open %s: %s\n",
		 ohFormatName( oFacet ), octErrorString());
	exit( 1 );
    }

    oProp.type = OCT_PROP;
    oProp.contents.prop.type = OCT_STRING;

    oProp.contents.prop.name = "VIEWTYPE";
    oProp.contents.prop.value.string = "PHYSICAL";
    octCreate( oFacet, &oProp );

    oProp.contents.prop.name = "EDITSTYLE";
    octCreate( oFacet, &oProp );

    oProp.contents.prop.name = "TECHNOLOGY";
    oProp.contents.prop.value.string = technology;
    octCreate( oFacet, &oProp );
}

/*
 * addMosProps
 *
 * add the CELLTYPE, MOSFETTYPE, MOSFETWIDTH, and MOSFETLENGTH props
 */
addMosProps( oFacet, type, width )
octObject *oFacet;	/* the tran facet			*/
enum fetType type;  		/* NFET or PFET 			*/
int width;		/* width in microns			*/
{
    octObject oProp;

    oProp.type = OCT_PROP;
    oProp.contents.prop.type = OCT_STRING;

    oProp.contents.prop.name = "CELLTYPE";
    oProp.contents.prop.value.string = "MOSFET";
    octCreate( oFacet, &oProp );

    oProp.contents.prop.name = "MOSFETTYPE";
    if( type == NFET ) oProp.contents.prop.value.string = "NENH";
    else               oProp.contents.prop.value.string = "PENH";
    octCreate( oFacet, &oProp );

    oProp.contents.prop.type = OCT_REAL;
    oProp.contents.prop.name = "MOSFETWIDTH";
    oProp.contents.prop.value.real = (double)width * 1e-6;
    octCreate( oFacet, &oProp );

    oProp.contents.prop.name = "MOSFETLENGTH";
    oProp.contents.prop.value.real = 2e-6;
    octCreate( oFacet, &oProp );
}

createTranFacet( oFacet, type, width, length, contact_mask )
octObject *oFacet;
enum fetType type;
int width;
int length;
unsigned contact_mask;
{
    octObject oTerm;
    octObject oDbox, oMbox, oPbox, oCbox, oWbox;
    octObject oPlayer, oDlayer, oClayer, oWlayer, oMlayer;
    int offset;	/* used to center CUTs in metal			*/

    /* convert width to oct units 				*/
    width *= 20;
    /* create the layers					*/
    createLayer( oFacet, &oPlayer, "POLY" );

    oDlayer.type = OCT_LAYER;
    if( type == NFET ) createLayer( oFacet, &oDlayer, "NDIF" );
    else               createLayer( oFacet, &oDlayer, "PDIF" );

    if(contact_mask&(SOURCE_CONTS|DRAIN_CONTS|SKIP_CONTS)){
	oClayer.type = OCT_LAYER;
	if( type == NFET ) createLayer( oFacet, &oClayer, "COND" );
	else               createLayer( oFacet, &oClayer, "COPD" );
	createLayer( oFacet, &oMlayer, "MET1" );
    }
    if( type == PFET ){
	createLayer( oFacet, &oWlayer, "NWEL" );
    }

    /*									*/
    /* SOURCE terminal and implementations				*/
    /*									*/

    createFormalTerm( oFacet, &oTerm, "SOURCE" );

    oDbox.type = OCT_BOX;
    oDbox.contents.box.lowerLeft.x = GATE_OVERLAP;
    oDbox.contents.box.lowerLeft.y = 0;
    oDbox.contents.box.upperRight.x = oDbox.contents.box.lowerLeft.x + width;

    if( contact_mask&(SOURCE_CONTS|SKIP_CONTS) )
	oDbox.contents.box.upperRight.y = CONTACTED_DIFF_WIDTH;
    else                         
	oDbox.contents.box.upperRight.y = UNCONTACTED_DIFF_WIDTH;
    octCreate( &oDlayer, &oDbox );
    octAttach( &oTerm, &oDbox );

    if( contact_mask&(SOURCE_CONTS|SKIP_CONTS)){
	oMbox = oDbox;
	oMbox.contents.box.upperRight.y = CONTACT_WIDTH;
	if( contact_mask&SOURCE_CONTS ){
	    oMbox.contents.box.lowerLeft.x = 
		MIN( oMbox.contents.box.upperRight.x - CONTACT_WIDTH,
		     oMbox.contents.box.lowerLeft.x + MAX_NAKED_DS );
	    /* quantizing the metal to box to CONTACT_WIDTH units	*/
	    oMbox.contents.box.upperRight.x -= 
		( oMbox.contents.box.upperRight.x - 
		  oMbox.contents.box.lowerLeft.x ) % CONTACT_WIDTH;
	} else {
	    /* for better compaction ?
	    oMbox.contents.box.upperRight.x = 
		MAX( oMbox.contents.box.upperRight.x - MAX_NAKED_DS,
		     oMbox.contents.box.lowerLeft.x + CONTACT_WIDTH );
	    */
	    oMbox.contents.box.upperRight.x = 
		 oMbox.contents.box.lowerLeft.x + CONTACT_WIDTH;
	}
	octCreate( &oMlayer, &oMbox );
	octAttach( &oTerm, &oMbox );

	oCbox.type = OCT_BOX;

	oCbox.contents.box.lowerLeft.y = M1_CUT_OVERLAP;
	oCbox.contents.box.upperRight.y = 
	    oMbox.contents.box.upperRight.y - M1_CUT_OVERLAP;

	offset = (oMbox.contents.box.upperRight.x - 
		  oMbox.contents.box.lowerLeft.x)%CUT_PITCH / 2;
	for( oCbox.contents.box.lowerLeft.x = 
		 oMbox.contents.box.lowerLeft.x + M1_CUT_OVERLAP + offset;
	     oCbox.contents.box.lowerLeft.x <= 
		 oMbox.contents.box.upperRight.x - M1_CUT_OVERLAP - CUT_WIDTH;
	     oCbox.contents.box.lowerLeft.x += CUT_PITCH ){
	    oCbox.contents.box.upperRight.x = 
		oCbox.contents.box.lowerLeft.x + CUT_WIDTH;
	    octCreate( &oClayer, &oCbox );
	    octAttach( &oTerm, &oCbox );
	}
    }

    /*									*/
    /* GATE terminal and implementations				*/
    /*									*/

    createFormalTerm( oFacet, &oTerm, "GATE" );

    /* the gate poly							*/
    oPbox.type = OCT_BOX;
    oPbox.contents.box.lowerLeft.x = 0;
    oPbox.contents.box.lowerLeft.y = oDbox.contents.box.upperRight.y;
    oPbox.contents.box.upperRight.x = 
	oDbox.contents.box.upperRight.x + GATE_OVERLAP;
    oPbox.contents.box.upperRight.y = 
	oPbox.contents.box.lowerLeft.y + GATE_LENGTH;

    octCreate( &oPlayer, &oPbox );
    octAttach( &oTerm, &oPbox );

    /* the diffusion underneath the gate				*/
    oDbox.contents.box.lowerLeft.y = oPbox.contents.box.lowerLeft.y;
    oDbox.contents.box.upperRight.y = oPbox.contents.box.upperRight.y;
    octCreate( &oDlayer, &oDbox );

    /*									*/
    /* DRAIN terminal and implementations				*/
    /*									*/

    createFormalTerm( oFacet, &oTerm, "DRAIN" );

    oDbox.type = OCT_BOX;
    oDbox.contents.box.lowerLeft.y = oDbox.contents.box.upperRight.y;
    if( contact_mask&(DRAIN_CONTS))
	oDbox.contents.box.upperRight.y += CONTACTED_DIFF_WIDTH;
    else            
	oDbox.contents.box.upperRight.y += UNCONTACTED_DIFF_WIDTH;
    octCreate( &oDlayer, &oDbox );
    octAttach( &oTerm, &oDbox );

    if( contact_mask&(DRAIN_CONTS)){
	oMbox = oDbox;
	oMbox.contents.box.lowerLeft.y = 
	    oMbox.contents.box.upperRight.y - CONTACT_WIDTH;
	/*  or better compaction?
	oMbox.contents.box.upperRight.x = 
	    MAX( oDbox.contents.box.upperRight.x - MAX_NAKED_DS,
		 oDbox.contents.box.lowerLeft.x + CONTACT_WIDTH );
	*/
	oMbox.contents.box.upperRight.x = 
	     oDbox.contents.box.lowerLeft.x + CONTACT_WIDTH;
	octCreate( &oMlayer, &oMbox );
	octAttach( &oTerm, &oMbox );

	oCbox.type = OCT_BOX;
	oCbox.contents.box.lowerLeft.y = 
	    oMbox.contents.box.lowerLeft.y + M1_CUT_OVERLAP;
	oCbox.contents.box.upperRight.y = 
	    oCbox.contents.box.lowerLeft.y + CUT_WIDTH;
	offset = (oMbox.contents.box.upperRight.x - 
		  oMbox.contents.box.lowerLeft.x)%CUT_PITCH / 2;
	for( oCbox.contents.box.lowerLeft.x = 
		 oMbox.contents.box.lowerLeft.x + M1_CUT_OVERLAP + offset;
	     oCbox.contents.box.lowerLeft.x <= 
		 oMbox.contents.box.upperRight.x - M1_CUT_OVERLAP - CUT_WIDTH;
	     oCbox.contents.box.lowerLeft.x += CUT_PITCH ){
	    oCbox.contents.box.upperRight.x = 
		oCbox.contents.box.lowerLeft.x + CUT_WIDTH;
	    octCreate( &oClayer, &oCbox );
	    octAttach( &oTerm, &oCbox );
	}
    }

    /*									*/
    /* WELL terminal and implementations				*/
    /*									*/

    if( type == PFET ){
	createFormalTerm( oFacet, &oTerm, "WELL" );

	oWbox = oDbox;
	oWbox.contents.box.lowerLeft.x = GATE_OVERLAP;
	oWbox.contents.box.lowerLeft.y = 0;
	octCreate( &oWlayer, &oWbox );
	octAttach( &oTerm, &oWbox );
    }
}
createLayer( oFacet, oLayer, name )
octObject *oFacet;
octObject *oLayer;
char *name;
{
    oLayer->type = OCT_LAYER;
    oLayer->contents.layer.name = name;
    octCreate( oFacet, oLayer );
}
createFormalTerm( oFacet, oTerm, name )
octObject *oFacet;
octObject *oTerm;
char *name;
{
    oTerm->type = OCT_TERM;
    oTerm->contents.term.name = name;
    oTerm->contents.term.instanceId = oct_null_id;
    oTerm->contents.term.width = 1;
    octCreate( oFacet, oTerm );
}
vulcanize( oFacet, oInterface )
octObject *oFacet;
octObject *oInterface;
{
    vulcanMessageType message;

    *oInterface = *oFacet;
    oInterface->contents.facet.facet = "interface";
    octOpenFacet( oInterface );
    message = vulcan( oFacet, oInterface );
    if(( message == VULCAN_SEVERE )||(message == VULCAN_FATAL )){
	fprintf(stderr,"Vulcan failed\n");
	exit(-1);
    }
}
