/*      estcap.c
 *
 *      estCap - A tool for capacitance estimation/extraction on unrouted layout.
 *      Author: Enrico Malavasi
 *      Ver. 1.10   -   November 1991
 */
#include "port.h"
#include "oct.h"
#include "utility.h"
#include "options.h"
#include "oh.h"
#include "tap.h"
#include "th.h"
#include "pep.h"
#include "vov.h"

static unsigned int   verbose_flag;
#define   VERBOSE(fct)    (verbose_flag) ? fct : 0

#define   VERSION_NUMBER   "1.10"
#define   SWAP(t,a,b)       {t x; x=a; a=b; b=x; }
#define   semiPerimeter(bb)   ((ohBoxWidth(bb)+ohBoxHeight(bb))*thCoordSize())

extern double pepModParallel();
extern double pepModCrossing();

typedef   double  Farad;

typedef    struct  _gsNet {
    octBox   bb;
    char    *name;
    unsigned char  implemented;
} _gsNet;
typedef    struct  _gsCap {
    Farad    cMin;
    Farad    cMax;
    unsigned char  extracted;
} _gsCap;
#define   GS_ESTIMATED_CAP     0
#define   GS_EXTRACTED_CAP     1

optionStruct optionList[] = {
    {"o", "out_file",           "Output file (def. stdout)."},
    {"t", "mincap",             "Min (threshold) cap (in fF) (def: 0.01)."},
    {"b", "net1,net2",          "Extract/estimate the cap between the two given nets only."},
    {"e", "",                   "Extract caps for routed nets, no estimates."},
    {"s", "filename",           "Print extracted/max-estimated caps in spice format."},
    {"u", "",                   "Estimate caps for unplaced layout."},
    {"E", "",                   "Extract caps for routed nets, estimate the others."},
    {"f", "techfile",           "Read local tech. file overriding default technology description."},
    {"h", "",                   "Print this message."},
    {"v", "",                   "Verbose mode."},
    { OPT_RARG,	"cell[:view]",	"View defaults to symbolic."},
    { 0, 0, 0 }
};

static  _gsNet   *gsNet = NULL;
static  int       gsNetNum = 0;
static  _gsCap  **gsCap;
static  _gsCap   *gndCap;

static FILE *fout;

static    Farad        maxBulkCap;
static    Farad        minBulkCap;
static    Farad        maxParallelCap;
static    Farad        maxCrossOverCap;
static    Farad        thresholdCap;

static unsigned int  extract_estimate;
#define  GS_EXTRACT   0x0001
#define  GS_ESTIMATE  0x0002
#define  GS_UNPLACED  0x0004

static int
  isNetImplemented(net)
octObject *net;
/* Returns TRUE if a physical implementation esists for the given net */
{
    octObject path;
    return (octGenFirstContent(net, OCT_PATH_MASK, &path) == OCT_OK);
}

static void
  gsExit(s)
int s;
/*  Clean exit */
{
    int  i;

    /* Free the allocated memory  */
    if(gsNetNum > 0) {
	for(i=1 ; i<gsNetNum ; i++) {
	    FREE(gsCap[i]);
	}
	FREE(gsCap);
	FREE(gsNet);
    }

    if((fout != stdout) && (fout != NULL))   fclose(fout);
    VOVend( s );
}

static void
  gcTopBottomLayers(facet, topLayer, bottomLayer)
octObject *facet;
octObject *topLayer;
octObject *bottomLayer;
/*  Find out the lowest and highest layer among the ones available for routing */
{
    octGenerator  lyrGen;
    octObject     layer;
    double        bottomHeight = (1.0);
    double        topHeight = (-1.0);
    double        height;
    
    OH_ASSERT(octInitGenContents(facet, OCT_LAYER_MASK, &lyrGen));
    while(octGenerate(&lyrGen, &layer) == OCT_OK) {
	if(strcmp(thGetLayerClass(&layer), "ROUTING") == 0) {
	    height = thGetLayerPropReal(&layer, "HEIGHT");
	    if(height < bottomHeight) {
		bottomHeight = height;
		*bottomLayer = layer;
	    }
	    if(height > topHeight) {
		topHeight = height;
		*topLayer = layer;
	    }
	}
    }
    octFreeGenerator(&lyrGen);
    if(topHeight < 0) {
	fprintf(stderr, "No routing layers in this facet\n");
	gsExit(1);
    }
}

static void
  gsParallelCaps(n1, n2, bb)
int       n1;
int       n2;
octBox   *bb;       /*  Intersection bounding box  */
/*  Parallel contribution to cross-coupling */
{
    double  len;

    len = semiPerimeter(bb);
    if(n1 < n2)  SWAP(int, n1, n2);
    gsCap[n1][n2].cMax += (len * maxParallelCap);
}

static void
  gsCrossOverCaps(n1, n2, bb)
int       n1;
int       n2;
octBox   *bb;       /*  Intersection bounding box  */
/*  Crossover contribution to cross-coupling */
{
    double   CrossOverNo;
    double   ratio;
    octBox   nbb;

    CrossOverNo = 3e-15;
    VERBOSE(fprintf(stderr, "bb = (%d,%d)-(%d,%d) - ", bb->lowerLeft.x, bb->lowerLeft.y, bb->upperRight.x, bb->upperRight.y));
    ratio = ohBoxArea(bb);
    ratio *= ratio;
    nbb = gsNet[n1].bb;    ratio /=  ohBoxArea(&nbb);
    VERBOSE(fprintf(stderr, "bb1 = (%d,%d)-(%d,%d) - ", nbb.lowerLeft.x, nbb.lowerLeft.y, nbb.upperRight.x, nbb.upperRight.y));
    nbb = gsNet[n2].bb;    ratio /=  ohBoxArea(&nbb);
    VERBOSE(fprintf(stderr, "bb2 = (%d,%d)-(%d,%d) - ", nbb.lowerLeft.x, nbb.lowerLeft.y, nbb.upperRight.x, nbb.upperRight.y));
    VERBOSE(fprintf(stderr, "Ratio: %g\n", ratio));
    if(n1 < n2)  SWAP(int, n1, n2);
    gsCap[n1][n2].cMax += (CrossOverNo * ratio);
}

#define  gsPrompt      fprintf(stderr, ".")
#define  gsEndPrompt   fprintf(stderr, "\n")

static void
  gsEstimateNetCap(n1, n2)
int n1;
int n2;
/*  Estimate cross-coupling cap between nets n1 and n2 */
{
    octBox intBb;

    intBb.lowerLeft.x  = MAX(gsNet[n1].bb.lowerLeft.x, gsNet[n2].bb.lowerLeft.x);
    intBb.lowerLeft.y  = MAX(gsNet[n1].bb.lowerLeft.y, gsNet[n2].bb.lowerLeft.y);
    intBb.upperRight.x = MIN(gsNet[n1].bb.upperRight.x, gsNet[n2].bb.upperRight.x);
    intBb.upperRight.y = MIN(gsNet[n1].bb.upperRight.y, gsNet[n2].bb.upperRight.y);
	    
    if((ohBoxWidth(&intBb) > 0) && (ohBoxHeight(&intBb) > 0)) {
	VERBOSE(fprintf(stderr, "Max cap betw. %s and %s: ", gsNet[n1].name, gsNet[n2].name));
	gsParallelCaps(n1, n2, &intBb);  /*  Run-along capacitances */
	VERBOSE(fprintf(stderr, "%12g F (parallel) - ", gsCap[n1][n2].cMax));
	gsCrossOverCaps(n1, n2, &intBb);   /*  CrossOver capacitances  */
	VERBOSE(fprintf(stderr, "%12g F (crossover)\n", gsCap[n1][n2].cMax));
    }
}

static void
  gsExtractNetCap(n1, n2, net1, net2, MARGIN)
int n1;
int n2;
octObject *net1;
octObject *net2;
octCoord   MARGIN;  /*  Distance at which net interaction is considered null */
/*  Extract cross-coupling cap between nets n1 and n2 */
{
    octBox intBb;

    intBb.lowerLeft.x  = MAX(gsNet[n1].bb.lowerLeft.x, gsNet[n2].bb.lowerLeft.x);
    intBb.lowerLeft.y  = MAX(gsNet[n1].bb.lowerLeft.y, gsNet[n2].bb.lowerLeft.y);
    intBb.upperRight.x = MIN(gsNet[n1].bb.upperRight.x, gsNet[n2].bb.upperRight.x);
    intBb.upperRight.y = MIN(gsNet[n1].bb.upperRight.y, gsNet[n2].bb.upperRight.y);
	    
    if((ohBoxWidth(&intBb)+MARGIN > 0) && (ohBoxHeight(&intBb)+MARGIN > 0)) {

	VERBOSE( fprintf(stderr, "Computing cross-cap between %s and %s\n",
			 gsNet[n1].name, gsNet[n2].name) );
	gsCap[n1][n2].cMax = pepCrossCapacitance(net1, net2);
		
    }
}

static void
  gsEstimate(FacetName, localTechFile, netName1, netName2)
char *FacetName;
char *localTechFile;
char *netName1;
char *netName2;
/*  Extract/estimate all capacitances */
{
    octGenerator NetGener;          /*  Net generator  */
    octObject    oNet;            /*  Oct nets  */
    octObject    facet, ifacet;            /*  Facet   */
    octObject    topLayer, bottomLayer;
    octObject    net1, net2;
    octBox       bb;
    octCoord     bminw, tminw;
    double       coordSize, eziSdrooc;
    int          i, j;
    char         name[128];
    octCoord   MARGIN;  /*  Distance at which net interaction is considered null */

    /*  Open the given facet     */
    ohUnpackDefaults(&facet, "r", ":symbolic:contents");
    ohUnpackFacetName(&facet, FacetName);
    if(octOpenFacet(&facet) < OCT_OK) {
	fprintf(stderr, "FATAL ERROR: facet %s:%s:contents doesn't exist\n",
		facet.contents.facet.cell,
		facet.contents.facet.view);
	gsExit(1);
    }
    sprintf(name, "%s:%s:contents", facet.contents.facet.cell, facet.contents.facet.view);
    VOVinput ( VOV_OCT_FACET , name );

    if(ohOpenInterface(&ifacet, &facet, "r") < OCT_OK) {
	fprintf(stderr, "FATAL ERROR: facet %s:%s:interface doesn't exist\n",
		facet.contents.facet.cell,
		facet.contents.facet.view);
	gsExit(1);
    }
    sprintf(name, "%s:%s:interface",facet.contents.facet.cell, facet.contents.facet.view);
    VOVinput ( VOV_OCT_FACET , name );

    thTechnology(&facet);
    if(localTechFile != NULL) {
	VERBOSE(fprintf(stderr, "Reading technology file %s\n", localTechFile));
	thReadFile( localTechFile );
	VOVinput ( VOV_UNIX_FILE , localTechFile );
    }

    /*  Determine the top and bottom layer  */
    gcTopBottomLayers(&ifacet, &topLayer, &bottomLayer);

    coordSize = thCoordSize();
    eziSdrooc = 1.0 / coordSize;
    bminw = tapGetMinWidth( &bottomLayer );
    tminw = tapGetMinWidth( &topLayer );
    maxBulkCap = thGetAreaCap(&bottomLayer) * bminw * coordSize + thGetPerimCap(&bottomLayer);
    minBulkCap = thGetAreaCap(&topLayer) * tminw * coordSize + thGetPerimCap(&topLayer);
    maxParallelCap  = pepModParallel(tapGetMinSpacing(&bottomLayer, &bottomLayer), bminw, bminw, (octCoord) eziSdrooc, &bottomLayer);
    maxCrossOverCap = pepModCrossing(bminw, tminw, &bottomLayer, &topLayer);

    /*  Generate the net array   */
    OH_ASSERT(octInitGenContents(&facet, OCT_NET_MASK, &NetGener));
    gsNet = ALLOC(_gsNet, 1);
    if(netName1 == NULL) {
	while(octGenerate(&NetGener, &oNet) == OCT_OK) {
	    if((oNet.contents.net.name[0] == '\0') || (oNet.contents.net.name[0] == ' ')) {
		fprintf(stderr, "WARNING: Net with no name - ignored\n");
		continue;
	    }
	    if(ohBB(&oNet, &bb) == OCT_OK)  {
		
		gsNet = REALLOC(_gsNet, gsNet, gsNetNum+1);
		gsNet[gsNetNum].name = util_strsav(oNet.contents.net.name);
		gsNet[gsNetNum].implemented = isNetImplemented(&oNet);
		gsNet[gsNetNum++].bb = bb;
		gsPrompt;
		VERBOSE(fprintf(stderr, "Net %s bb: (%d,%d)-(%d,%d)\n",
				gsNet[gsNetNum-1].name,
				gsNet[gsNetNum-1].bb.lowerLeft.x,
				gsNet[gsNetNum-1].bb.lowerLeft.y,
				gsNet[gsNetNum-1].bb.upperRight.x,
				gsNet[gsNetNum-1].bb.upperRight.y));
	    }
	}
    } else {
	OH_ASSERT( ohGetByNetName(&facet, &oNet, netName1) );
	if(ohBB(&oNet, &bb) == OCT_OK)  {
	    
	    gsNet = REALLOC(_gsNet, gsNet, gsNetNum+1);
	    gsNet[gsNetNum].name = util_strsav(oNet.contents.net.name);
	    gsNet[gsNetNum].implemented = isNetImplemented(&oNet);
	    gsNet[gsNetNum++].bb = bb;
	    gsPrompt;
	    VERBOSE(fprintf(stderr, "Net %s bb: (%d,%d)-(%d,%d)\n",
			    gsNet[gsNetNum-1].name,
			    gsNet[gsNetNum-1].bb.lowerLeft.x,
			    gsNet[gsNetNum-1].bb.lowerLeft.y,
			    gsNet[gsNetNum-1].bb.upperRight.x,
			    gsNet[gsNetNum-1].bb.upperRight.y));
	}

	ohGetByNetName(&facet, &oNet, netName2);
	if(ohBB(&oNet, &bb) == OCT_OK)  {
	    
	    gsNet = REALLOC(_gsNet, gsNet, gsNetNum+1);
	    gsNet[gsNetNum].name = util_strsav(oNet.contents.net.name);
	    gsNet[gsNetNum].implemented = isNetImplemented(&oNet);
	    gsNet[gsNetNum++].bb = bb;
	    gsPrompt;
	    VERBOSE(fprintf(stderr, "Net %s bb: (%d,%d)-(%d,%d)\n",
			    gsNet[gsNetNum-1].name,
			    gsNet[gsNetNum-1].bb.lowerLeft.x,
			    gsNet[gsNetNum-1].bb.lowerLeft.y,
			    gsNet[gsNetNum-1].bb.upperRight.x,
			    gsNet[gsNetNum-1].bb.upperRight.y));
	}
    }
    octFreeGenerator(&NetGener);
    gsEndPrompt;
    
    if(gsNetNum == 0) {
	fprintf(stderr, "No nets found in the facet\n");
	gsNetNum = 0;     /* Don't delete this line */
	gsExit(0);
    }	
    
    /*  Sort the net set   */
    for(i=0 ; i<gsNetNum ; i++) {
	for(j=i+1 ; j<gsNetNum ; j++) {
	    if(strcasecmp(gsNet[i].name, gsNet[j].name) > 0) {
		SWAP(_gsNet, gsNet[i], gsNet[j]);
	    }
	}
    }
    
    /*  Allocate and initialize the gsCap and gndCap array    */
    gsCap  = ALLOC(_gsCap *, gsNetNum);
    gndCap = ALLOC(_gsCap,   gsNetNum);
    for(i=0 ; i<gsNetNum ; i++) {
	if( i > 0 )  gsCap[i] = ALLOC(_gsCap, i);
	gndCap[i].cMin = 0.0;
	gndCap[i].cMax = 0.0;
	gndCap[i].extracted = GS_ESTIMATED_CAP;
	for(j=0 ; j<i ; j++) {            /*  Always j < i  */
	    gsCap[i][j].cMin = 0.0;
	    gsCap[i][j].cMax = 0.0;
	    gsCap[i][j].extracted = GS_ESTIMATED_CAP;
	}
    }
    if( extract_estimate == GS_EXTRACT ) {
	for(i=0 ; i<gsNetNum ; i++) {
	    gndCap[i].extracted = GS_EXTRACTED_CAP;
	    if( gsNet[i].implemented) {
		for(j=0 ; j<i ; j++) {            /*  Always j < i  */
		    if( gsNet[j].implemented) {
			gsCap[i][j].extracted = GS_EXTRACTED_CAP;
		    }
		}
	    }
	}
    }
	
    if( extract_estimate ==  GS_UNPLACED ) {
	octGenerator Gener;
	octObject    inst;
	octBox       bb, sBB;
	Farad        cap;
	
	sBB.lowerLeft.x = sBB.upperRight.x = 0;
	sBB.lowerLeft.y = sBB.upperRight.y = 0;
	OH_ASSERT(octInitGenContents(&facet, OCT_INSTANCE_MASK, &Gener));
	while(octGenerate(&Gener, &inst) == OCT_OK) {
	    ohBB(&inst, &bb);
	    sBB.upperRight.x += (ohBoxWidth( &bb ) * ohBoxHeight( &bb ));
	}
	sBB.upperRight.x = sqrt( (double) (sBB.upperRight.x * 2.0) );
	sBB.upperRight.y = sBB.upperRight.x;
	octFreeGenerator( &Gener );
	cap = (sBB.upperRight.x + sBB.upperRight.y) * coordSize * maxBulkCap;
	for(i=0 ; i<gsNetNum ; i++) {
	    gndCap[ i ].cMax = cap;
	}
    } else {
	double  len;

	MARGIN = 10.0e-6 / coordSize;    /*  10.0 um margin  */
	for(i=0 ; i<gsNetNum ; i++) {
	    ohGetByNetName(&facet, &net1, gsNet[i].name);
	    if(gndCap[i].extracted == GS_EXTRACTED_CAP) {
		gndCap[ i ].cMax = pepCapacitance( &net1 );
	    } else {
		len = semiPerimeter(&(gsNet[ i ].bb));
		gndCap[ i ].cMax = 2 * len * minBulkCap;
		gndCap[ i ].cMin = len * maxBulkCap;
	    }
	    for(j=0 ; j<i ; j++) {
		ohGetByNetName(&facet, &net2, gsNet[j].name);
		if(gsCap[i][j].extracted == GS_EXTRACTED_CAP) {
		    gsExtractNetCap(i, j, &net1, &net2, MARGIN);
		} else {
		    gsEstimateNetCap(i, j);
		}
	    }
	}
    }
	  
    octCloseFacet(&facet);
    octCloseFacet(&ifacet);
}

static void
  gsPrintOut( spiceFile )
char *spiceFile;
{
    FILE *fspi = NULL;
    int i, j, realCapNo;

    /*  Print output ------------------------------ */
    for(i=1,realCapNo=0 ; i<gsNetNum ; i++) {
	for(j=0 ; j<i ; j++) {
	    if((gsCap[i][j].cMax > thresholdCap) || (gsCap[i][j].cMin > thresholdCap)) {
		realCapNo++;
	    }
	}
    }

    if(spiceFile != NULL) {
	if((fspi=fopen(spiceFile, "w")) == NULL) {
	    fprintf(stderr, "Couldn't open file %s\n", spiceFile);
	    gsExit(1);
	}
    }
    
    fprintf(fout, "number of nets %d\n", gsNetNum);
    fprintf(fout, "number of net pairs %d\n", realCapNo);
    fprintf(fout, "!n1\tn2\tCmax (fF)     Cmin (fF)\n");

    for(i=1 ; i<gsNetNum ; i++) {
	for(j=0 ; j<i ; j++) {
	    if((gsCap[i][j].cMax>thresholdCap) || (gsCap[i][j].cMin>thresholdCap)) {
		fprintf(fout, "%s\t%s\t%9.3f", gsNet[j].name, gsNet[i].name,
			gsCap[i][j].cMax*1.0e15);
		if(gsCap[i][j].extracted == GS_ESTIMATED_CAP)
		  fprintf(fout, " %9.3f", gsCap[i][j].cMin*1.0e15);
		if(fspi) {
		    fprintf(fspi, "C%sTO%s %s %s %sF\n",
			    gsNet[j].name, gsNet[i].name,
			    gsNet[j].name, gsNet[i].name,
			    util_pretty_print(gsCap[i][j].cMax));
		}
		fprintf(fout, "\n");
	    }
	}
    }

    fprintf(fout, "\n!capacitances vs substrate\n");
    for(i=1 ; i<gsNetNum ; i++) {
	if((gndCap[i].cMax > thresholdCap) || (gndCap[i].cMin > thresholdCap)) {
	    fprintf(fout, "%s\t\t%sF", gsNet[i].name, util_pretty_print(gndCap[i].cMax));
	    if(gndCap[i].extracted == GS_ESTIMATED_CAP)
	      fprintf(fout, " %sF", util_pretty_print(gndCap[i].cMin));
	    fprintf(fout, "\n");
	}
	if( fspi ) {
	    fprintf(fspi, "C%sTOgnd %s *Ground* %sF\n",
		    gsNet[i].name, gsNet[i].name,
		    util_pretty_print(gndCap[i].cMax));
	}
    }

    fprintf(fout, "\n");
    if(fspi)  fclose( fspi );
}

main(argc,argv)
int argc;
char **argv;
{
    int option;
    int i;
    char  *FacetName     = NULL;
    char  *localTechFile = NULL;
    char  *outputFile    = NULL;
    char  *spiceFile     = NULL;
    char  *netName1 = NULL, *netName2 = NULL;

    fprintf(stderr, "%s: cross-capacitance extraction and/or estimation.\n",argv[0]);
    fprintf(stderr, "Vers. %s  -  by Enrico Malavasi, Aug. 1991\n", VERSION_NUMBER);

    verbose_flag = 0;
    extract_estimate = GS_ESTIMATE;
    fout     = stdout;
    thresholdCap = 10.0e-18;     /*  0.01 fF */
    while ( ( option = optGetOption (argc, argv)) != EOF) {
	switch (option) {
	  case 't':
	    thresholdCap = atof(optarg) * 1.0e-15;
	    break;
	  case 'v':
	    verbose_flag = 1;
	    break;
	  case 'b':
	    netName1 = util_strsav(optarg);
	    netName2 = util_strsav(optarg);
	    for(i=0 ; i<strlen(netName1) ; i++) {
		if(netName1[i] == ',') {
		    netName1[i] = '\0';
		    break;
		}
	    }
	    netName2 = netName2+i+1;
	    break;
	  case 'e':
	    extract_estimate = GS_EXTRACT;
	    break;
	  case 's':
	    spiceFile = util_strsav( optarg );
	    break;
	  case 'u':
	    extract_estimate = GS_UNPLACED;
	    break;
	  case 'E':
	    extract_estimate = GS_EXTRACT | GS_ESTIMATE;
	    break;
	  case 'f':
	    localTechFile = util_strsav(optarg);
	    break;
	  case 'o':
	    outputFile = util_strsav( optarg );
	    break;

	  default:
	    optUsage();
	}
    }
    if(optind != (argc-1)) optUsage(argv[0]);

    VOVbegin( argc, argv );
    if(outputFile != NULL) {
	if((fout=fopen(util_file_expand( outputFile ), "w")) == NULL) {
	    fprintf(stderr, "Couldn't open file %s\n", outputFile);
	    gsExit(1);
	}
	VOVoutput( VOV_UNIX_FILE , util_file_expand(outputFile) );
    }
    FacetName = util_strsav(argv[optind]);

    gsEstimate(FacetName, localTechFile, netName1, netName2);
    gsPrintOut( spiceFile );
    gsExit(0);
}
