/*
 * contacts.c: Routines for extracting contacts
 */
#include "port.h"
#include "utility.h"
#include "fang.h"
#include "harpoon.h"
#include "oct.h"
#include "st.h"
#include "oh.h"
#include "fa.h"
#include "gn.h"
#include "list.h"
#include "kh.h"
#include "kahlua.h"

kh_term* termConstruct( x, y, id )
    int x,y;
    octId id;
{
    kh_term* termEntry = ALLOC(kh_term, 1);
    termEntry->position.x = x;
    termEntry->position.y = y;
    termEntry->termId = id;
    termEntry->equiv_next = 0;
    termEntry->equiv_head = 0;
    termEntry->gridPosition.x = -1; /* Illegal position */
    termEntry->gridPosition.y = -1; /* Illegal position */
    termEntry->overlap = 0;
    return termEntry;
}

/*
 * handleContacts:
 *    for each contact layer {
 *        frame using merge rule;
 *        break into boxes;
 *        for each box {
 *            make master if necessary;
 *            add to terminal table;
 *        }
 *    }
 */
/*ARGSUSED*/
handleContacts(inFacetPtr, outFacetPtr, geoListPtr, tsListPtr)
octObject *inFacetPtr, *outFacetPtr;
harList *geoListPtr;
khList *tsListPtr;
{
    octObject connectorBag, instance, aterm, layer, prop;
    octGenerator genLayer;
    char *layername;
    int x, y;
    lsList term1List, term2List;
    kh_term *termEntry;
    int width, length;
    char master[BUFFER_SIZE], view[BUFFER_SIZE], name[BUFFER_SIZE];
    octTransformType type;
    octObject layerBag;
    char *layer1name, *layer2name;
    int layer1overlap, layer2overlap, cutSize, cutSpace, mergeRule;
    fa_geometry *conGeo, framedGeo;
    fa_box *fangBoxGen, *fangBoxList, fangBox;

    /* print trace message */
    if (trace) {
	(void) fprintf(stderr, "%sextracting contacts ...\n", INDENT);
    }

    /* get connector bag */
    OH_ASSERT(ohGetByBagName(outFacetPtr, &connectorBag, "CONNECTORS"));

    OH_ASSERT(ohGetByBagName(&globalFacet, &layerBag, "CONTACTS"));
    OH_ASSERT(octInitGenContents(&layerBag, OCT_LAYER_MASK, &genLayer));
    while (octGenerate(&genLayer, &layer) == OCT_OK) {

	/* 
	 * grab rules from technology
	 * facet 'globalFacet'
	 */
	OH_ASSERT(ohGetByPropName(&layer, &prop, "LAYER1"));
	layer1name = prop.contents.prop.value.string;

	OH_ASSERT(ohGetByPropName(&layer, &prop, "LAYER2"));
	layer2name = prop.contents.prop.value.string;

	OH_ASSERT(ohGetByPropName(&layer, &prop, "CUT_SIZE"));
	cutSize = prop.contents.prop.value.integer;

	OH_ASSERT(ohGetByPropName(&layer, &prop, "CUT_SPACE"));
	cutSpace = prop.contents.prop.value.integer;

	OH_ASSERT(ohGetByPropName(&layer, &prop, "LAYER1_OVERLAP"));
	layer1overlap = prop.contents.prop.value.integer;

	OH_ASSERT(ohGetByPropName(&layer, &prop, "LAYER2_OVERLAP"));
	layer2overlap = prop.contents.prop.value.integer;

	OH_ASSERT(ohGetByPropName(&layer, &prop, "MERGE_RULE"));
	mergeRule = prop.contents.prop.value.integer;

	/* grab contact geo */
	layername = layer.contents.layer.name;
	conGeo = harGetNamedGeo(layername, geoListPtr);
	fa_frame(conGeo, mergeRule/2, mergeRule/2, &framedGeo);
	fa_to_boxes(&framedGeo, &fangBoxList);

	/* 
	 * for each cut region, create a master if necessary,
	 * and instantiate the cell, and add terminal information
	 * to terminal list
	 */
	fangBoxGen = fangBoxList;
	while (faGenerateBox(&fangBoxGen, &fangBox) == 1) {

	    faBoxCenter(&fangBox, &x, &y);
	    width = fangBox.right - fangBox.left;
	    length = fangBox.top - fangBox.bottom;

	    /* cannot make a contact smaller than cutSize */
	    if ((width < cutSize) || (length < cutSize)) {
		(void) fprintf(stderr, "Ignored: cannot make a contact ");
		(void) fprintf(stderr, "smaller than minimum cut size\n");
	    } else {
		/*
		 * make master if necessary,
		 * then instantiate it
		 */
		makeContactMaster(layername, layer1name, layer2name, cutSpace,
		  cutSize, layer1overlap, layer2overlap, width, length, &type,
		  master, view);
		mkInstanceName(name);
		/* keep count */
		contactCount++;
		OH_ASSERT(ohCreateInstance(outFacetPtr, &instance, name, master,
		  view, x, y, type));
		OH_ASSERT(octAttach(&connectorBag, &instance));

		/* grab actual terminal of contact */
		OH_ASSERT(octGenFirstContent(&instance, OCT_TERM_MASK, 
		  &aterm));

		/* make terminal entries */
		term1List = khGetNamedLs(layer1name, tsListPtr);
		term2List = khGetNamedLs(layer2name, tsListPtr);
		termEntry = termConstruct( x, y, aterm.objectId );
		(void) lsNewBegin(term1List, (Pointer) termEntry, LS_NH);
		(void) lsNewBegin(term2List, (Pointer) termEntry, LS_NH);
	    }
	}
    }

    /*
     * need to free memory 
     */
    fa_free_boxes(fangBoxList);
    fa_free_geo(&framedGeo);
}

