/*
 * disassemble.c -- The actual disassembly process is done by
 *    several major subroutines.  This routines in this file
 *    are used to call them.
 */

#include "port.h"
#include "utility.h"
#include "fang.h"
#include "harpoon.h"
#include "oct.h"
#include "st.h"
#include "oh.h"
#include "fa.h"
#include "gn.h"
#include "list.h"
#include "kh.h"
#include "kahlua.h"
#include "vov.h"

/*
 * disassemble: This routine hierarchically calls the appropriate
 *    sub-routines for disassembling a mask layout.  The steps
 *    are as follows --
 *        1) disassemble subcells by recursively calling itself
 *        2) initialize terminal lists and build fang geo list
 *        3) promote all instance terminals and instantiate all instances
 *        4) build formal terminals a) explicit; b) implicit
 *        5) extract contacts
 *        6) extract transistors
 *        6.5) derive implicit formal terminals
 *        7) decompose interconnects
 *        8) add standard symbolic properties
 *        9) make interface facet
 *       10) free memory allocated for fang geo list
 */
disassemble(inCell, inView, outCell, outView)
    char *inCell, *inView, *outCell, *outView;
{
    octObject inputFacet, outputFacet, ifacet, instance, bag;
    octGenerator genInst;
    char buffer[BUFFER_SIZE];
    char *master, *view;
    harList geoList;
    khList tsList;

    /* open input facet */
    OH_ASSERT(ohOpenFacet(&inputFacet, inCell, inView, "contents", "r"));
    VOVinputFacet( &inputFacet );

    if (!parentOnly) {
	/*
	 * 1) disassemble subcells by recursively calling itself
	 */
	OH_ASSERT(octInitGenContents(&inputFacet, OCT_INSTANCE_MASK, &genInst));
	while (octGenerate(&genInst, &instance) == OCT_OK) {
	    master = instance.contents.instance.master;
	    view = instance.contents.instance.view;
	    (void) sprintf(buffer, "%s:%s", master, view);
	    if (ohGetByBagName(&disBag, &bag, buffer) != OCT_OK) {
		OH_ASSERT(ohCreateBag(&disBag, &bag, buffer));
		/* keep count */
		uniqueSubcellCount++;
		disassemble(master, view, master, outView);
	    }
	}
    }

    /* print out trace message */
    if (trace) {
	(void) fprintf(stderr, "disassembling [%s:%s] to [%s:%s] ...\n",
	  inCell, inView, outCell, outView);
    }

    /*
     * 2) initialize terminal lists and build fang geo list
     */
    geoList = HAR_EMPTY_LIST;
    tsList = KH_EMPTY_LIST;

    /* print trace message */
    if (trace) {
	(void) fprintf(stderr, "%sreading input facet ...\n", INDENT);
    }
    harAddFacetToFangList(&inputFacet, &geoList, HAR_NO_TRANSFORM);

    /*
     * begin producing output --
     */

    /* open output facet */
    OH_ASSERT(ohOpenFacet(&outputFacet, outCell, outView, "contents", "w"));
    VOVoutputFacet( &outputFacet );
    addInstanceAndConnectorBags(&outputFacet);
    addSymbolicProperties(&outputFacet);

    /* handle the instances */
    if (!parentOnly) {
	handleInstances(&inputFacet, &outputFacet, &geoList, &tsList);
    }

    /* handle the formal terminals */
    if (explicitFterm) {
	handleExplicitFormals(&inputFacet, &outputFacet, &geoList, &tsList);
    }

    /* handle leaf cells */
    handleContacts(&inputFacet, &outputFacet, &geoList, &tsList);
    handleMosfets(&inputFacet, &outputFacet, &geoList, &tsList);

    /* derive the implicit formal terminals */
    if (implicitFterm) {
	handleImplicitFormals(&inputFacet, &outputFacet, &geoList, &tsList);
    }

    if (!deviceOnly) {
	/* handle net decomposition */
	handleNets(&inputFacet, &outputFacet, &geoList, &tsList);
    }

    /*
     * 8) add standard symbolic properties
     */

    
    /*
     * 9) make interface facet
     */
    OH_ASSERT(gnGenInterfaceFacet(&outputFacet,&ifacet));

    /* close output facet */
    OH_ASSERT(octCloseFacet(&outputFacet));
    OH_ASSERT(octCloseFacet(&ifacet));

    /* close input facet */
    OH_ASSERT(octCloseFacet(&inputFacet));

    /*
     * 10) free fang geo list and terminal list
     */
    harFreeList(&geoList);
    khFreeList(&tsList);
}


/*
 * addSymbolicProperties:
 *    TECHNOLOGY = <technology>
 *    VIEWTYPE = SYMBOLIC
 *    EDITSTYLE = SYMBOLIC
 *    CELLCLASS = MODULE
 */
addSymbolicProperties(facetPtr)
octObject *facetPtr;
{
    octObject prop;

    OH_ASSERT(ohCreatePropStr(facetPtr, &prop, "TECHNOLOGY", technology));
    OH_ASSERT(ohCreatePropStr(facetPtr, &prop, "VIEWTYPE", "SYMBOLIC"));
    OH_ASSERT(ohCreatePropStr(facetPtr, &prop, "EDITSTYLE", "SYMBOLIC"));
    OH_ASSERT(ohCreatePropStr(facetPtr, &prop, "CELLCLASS", "MODULE"));
}


/*
 * addInstanceAndConnectorBags:
 *    INSTANCES, and CONNECTORS
 */
addInstanceAndConnectorBags(facetPtr)
octObject *facetPtr;
{
    octObject bag;

    OH_ASSERT(ohCreateBag(facetPtr, &bag, "INSTANCES"));
    OH_ASSERT(ohCreateBag(facetPtr, &bag, "CONNECTORS"));
}
