/*
 * globals.c -- sets up technology facet for global use.
 *    This includes reading the technology file and setting
 *    up design rules for disassembly use.  All the necessary 
 *    linkings are done.  
 */

#include "port.h"
#include "utility.h"
#include "oct.h"
#include "st.h"
#include "oh.h"
#include "gn.h"
#include "list.h"
#include "kahlua.h"
#include "errtrap.h"

octObject globalFacet;		/* technology facet for global use */
octObject masBag;		/* table of leaf masters already created */
octObject disBag;		/* table of cells already disassembled */

/*
 * initGlobals: reads technology file and builds extensive
 *    technology facet for internal use.  Technology facet
 *    is layer driven.  Different layers will have associated
 *    design rules and properties attach to them.  In addition,
 *    a table will be kept of unique cell masters already created,
 *    and a table will be kept of modules already disassembled.
 */
initGlobals()
{
    octObject prop, layer;
    octObject netBag, conBag, fetBag;
    char realTechFile[BUFFER_SIZE];
    char *line, *buffer, *fstatus;
    FILE *fp;
    char *nextword();

    /* print trace message */
    if (trace) {
	(void) fprintf(stderr, "reading technology file ...\n\n");
    }

    /* open technology facet */
    OH_ASSERT(ohOpenFacet(&globalFacet, KH_GLOBAL_CELL, KH_GLOBAL_VIEW, "contents","w"));

    /* set up hash tables */
    OH_ASSERT(ohCreateBag(&globalFacet, &masBag, "MASTERS"));
    OH_ASSERT(ohCreateBag(&globalFacet, &disBag, "DISASSEMBLED"));

    /* create bag structure for keeping unique cell masters */
    OH_ASSERT(ohCreateBag(&globalFacet, &netBag, "NETS"));
    OH_ASSERT(ohCreateBag(&globalFacet, &conBag, "CONTACTS"));
    OH_ASSERT(ohCreateBag(&globalFacet, &fetBag, "MOSFETS"));

    if ((fp = fopen(util_file_expand(techFile), "r")) == NULL) {
	errRaise("kahlua",1 , "Panic: couldn't open techfile %s\n", techFile);
    }

    /* 
     * start reading technology file: four possible entries --
     *    1) scale 2) net 3) contact 4) mosfet.
     */
    line = ALLOC(char, 500);
    while ((fstatus = fgets(line, 500, fp)) !=  NIL(char)) {
	buffer = nextword(&line);

	/* skip comments and blank lines */
	if ((buffer != NULL) && (buffer[0] != '#')) {
	    /*
	     * parsing a scale entry
	     */
	    if (!strcmp(buffer, "scale")) {
		buffer = nextword(&line);
		OH_ASSERT(ohCreatePropInt(&globalFacet, &prop, "FACTOR", atoi(buffer)));
	    } 
	    
	    /*
	     * parsing a net entry
	     */
	    else if (!strcmp(buffer, "net")) {
		buffer = nextword(&line);
		OH_ASSERT(ohCreateLayer(&globalFacet, &layer, buffer));
		OH_ASSERT(octAttach(&netBag, &layer));
		buffer = nextword(&line);
		OH_ASSERT(ohCreatePropInt(&layer, &prop, "MINIMUM_WIDTH", 
		  expandFactor*atoi(buffer)));
	    } 
	    
	    /*
	     * parsing a contact entry
	     */
	    else if (!strcmp(buffer, "contact")) {
		buffer = nextword(&line);
		OH_ASSERT(ohCreateLayer(&globalFacet, &layer, buffer));
		OH_ASSERT(octAttach(&conBag, &layer));
		buffer = nextword(&line);
		OH_ASSERT(ohCreatePropStr(&layer, &prop, "LAYER1", buffer));
		buffer = nextword(&line);
		OH_ASSERT(ohCreatePropStr(&layer, &prop, "LAYER2", buffer));
		buffer = nextword(&line);
		OH_ASSERT(ohCreatePropInt(&layer, &prop, "CUT_SIZE", 
		  expandFactor*atoi(buffer)));
		buffer = nextword(&line);
		OH_ASSERT(ohCreatePropInt(&layer, &prop, "CUT_SPACE", 
		  expandFactor*atoi(buffer)));
		buffer = nextword(&line);
		OH_ASSERT(ohCreatePropInt(&layer, &prop, "LAYER1_OVERLAP", 
		  expandFactor*atoi(buffer)));
		buffer = nextword(&line);
		OH_ASSERT(ohCreatePropInt(&layer, &prop, "LAYER2_OVERLAP", 
		  expandFactor*atoi(buffer)));
		buffer = nextword(&line);
		OH_ASSERT(ohCreatePropInt(&layer, &prop, "MERGE_RULE", 
		  expandFactor*atoi(buffer)));
	    } 
	    
	    /*
	     * parsing a mosfet entry
	     */
	    else if (!strcmp(buffer, "mosfet")) {
		buffer = nextword(&line);
		OH_ASSERT(ohCreateLayer(&globalFacet, &layer, buffer));
		OH_ASSERT(octAttach(&fetBag, &layer));
		buffer = nextword(&line);
		OH_ASSERT(ohCreatePropStr(&layer, &prop, "POLY_LAYER", buffer));
		buffer = nextword(&line);
		OH_ASSERT(ohCreatePropStr(&layer, &prop, "DIFF_LAYER", buffer));
		buffer = nextword(&line);
		OH_ASSERT(ohCreatePropInt(&layer, &prop, "POLY_OVERLAP_RULE", 
		  expandFactor*atoi(buffer)));
		buffer = nextword(&line);
		OH_ASSERT(ohCreatePropInt(&layer, &prop, "DIFF_OVERLAP_RULE", 
		  expandFactor*atoi(buffer)));
	    }
	}
    }
}


/*
 * endGlobals: frees memory used for technology facet
 */
endGlobals()
{
    OH_ASSERT(octFreeFacet(&globalFacet));
}


