#include "port.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "gn.h"
#include "kahlua.h"
#include "vulcan.h"

/*
 *  gnPathDirection: Determines direction of path
 *    returns: 2 -- if path has zero length
 *             1 -- if path is vertical
 *             0 -- if path is horizontal
 *            -1 -- if there is an error
 */
gnPathDirection(t)
octObject *t;
{
    struct octPoint *p;
    octStatus status;
    octCoord p1x, p1y, p2x, p2y;
    int32 num_points = 2;

    p = ALLOC(struct octPoint, 2);
    status = octGetPoints(t, &num_points, p);
    if ((status < OCT_OK) || (num_points != 2)) return -1;

    p1x = p[0].x;
    p1y = p[0].y;
    p2x = p[1].x;
    p2y = p[1].y;

    /* free memory */
    FREE(p);

    if ((p1x != p2x) && (p1y != p2y)) return -1;
    if ((p1x == p2x) && (p1y == p2y)) return 2;
    if (p1x == p2x) return 1;
    if (p1y == p2y) return 0;

    /* shouldn't happen */
    return -1;
}


/*
 *  gnCreatePath: Creates a path
 */
octStatus gnCreatePath(c, t, p1x, p1y, p2x, p2y, width)
octObject *c, *t;
octCoord p1x, p1y, p2x, p2y;
octCoord width;
{
    struct octPoint p[2];
    octStatus status;

    t->type = OCT_PATH;
    t->contents.path.width = width;
    status = octCreate(c, t);
    if (status < OCT_OK) return status;

    p[0].x = p1x;
    p[0].y = p1y;
    p[1].x = p2x;
    p[1].y = p2y;
    status = octPutPoints(t, 2, p);

    return status;
}


/*
 * gnDetachNetConnectors: Detach connector terminals from nets in facet
 */
octStatus gnDetachNetConnectors(facet)
octObject *facet;
{
    octObject net, aterm, instance, connectorBag;
    octGenerator genTerm, genNet, genInst;
    int i, c;

    /* grab connect instances */
    OH_ASSERT(ohGetByBagName(facet, &connectorBag, "CONNECTORS"));

    /* 
     * for each connector instance,
     * detach its terminal from the
     * connecting net
     */
    OH_ASSERT(octInitGenContents(&connectorBag, OCT_INSTANCE_MASK, &genInst));
    while (octGenerate(&genInst, &instance) == OCT_OK) {
	OH_ASSERT(octInitGenContents(&instance, OCT_TERM_MASK, &genTerm));
	while (octGenerate(&genTerm, &aterm) == OCT_OK) {
	    /*
	     * count the number of containers first and
	     * use for loop instead.  this is to prevent
	     * an oct bug that causes infinite loops
	     */
	    c= ohCountContents( &aterm, OCT_NET_MASK );
	    OH_ASSERT(octInitGenContainers(&aterm, OCT_NET_MASK, &genNet));
	    for (i=0; i<c; i++) {
		octGenerate(&genNet, &net);
		OH_ASSERT(octDetach(&net, &aterm));
	    }
	    octFreeGenerator( &genNet );
	}
    }

    return OCT_OK;
}


/*
 *  gnGenInterfaceFacet: Generates an interface facet for given contents facet
 */
octStatus gnGenInterfaceFacet(Facet,IntFacet)
octObject *Facet;
octObject *IntFacet;
{
    int quick = 1;

    if ( quick ) {
	*IntFacet = *Facet;
	IntFacet->contents.facet.facet = "interface";
	IntFacet->contents.facet.mode = "w";
	OH_ASSERT( octOpenFacet( IntFacet ));
	vulcan( Facet, IntFacet );
	
    } else {
	char *cell, *view, *facet;
	char command[500];

	OH_ASSERT(octFlushFacet( Facet ));
	cell = Facet->contents.facet.cell;
	view = Facet->contents.facet.view;
	facet = Facet->contents.facet.facet;

	if (strcmp(facet, "contents")) return OCT_ERROR;

	(void) sprintf(command, "vulcan -k %s:%s", cell, view);
	if ( trace ) {
	    fprintf(stderr, "\t%s\n", command );
	}
	(void) util_csystem(command);
    }

    return OCT_OK;
}


/*
 *  gnSymbolicActualTerminalBoxAndLayer: Find the box and layer of 
 *      an actual terminal from a symbolic facet
 */
octStatus gnSymbolicActualTerminalBoxAndLayer(aterm, box, layer)
octObject *aterm;
octObject *box;
octObject *layer;
{
    octObject instance, master, fterm;
    octStatus status;
    struct octTransform *transform;

    /* Get the instance from the terminal */
    status = ohGetById(&instance, aterm->contents.term.instanceId);
    if (status != OCT_OK) return status;

    /* open the interface facet of the master of the instance */
    status = ohOpenMaster(&master, &instance, "interface", "r");
    if (status < OCT_OK) return status;

    /* Find formal terminal (which matches this actual terminal) */
    status = ohGetByTermName(&master, &fterm, aterm->contents.term.name);
    if (status != OCT_OK) return status;

    /* Get the first 'terminal frame' (which must be a box) */
    status = octGenFirstContent(&fterm, OCT_BOX_MASK, box);
    if (status != OCT_OK) return status;

    /* Get the layer associated with box */
    status = octGenFirstContainer(box, OCT_LAYER_MASK, layer);
    if (status != OCT_OK) return status;

    /* Transform the box */
    transform = &instance.contents.instance.transform;
    octTransformGeo(box, transform);

    return OCT_OK;
}


/*
 *  gnPhysicalFormalTerminalBoxAndLayer: Find the box and layer of 
 *      a formal terminal from a physical facet
 */
octStatus gnPhysicalFormalTerminalBoxAndLayer(fterm, box, layer)
octObject *fterm;
octObject *box;
octObject *layer;
{

    /* Get the first 'terminal frame' (which must be a box) */
    OH_ASSERT(  octGenFirstContent(fterm, OCT_BOX_MASK, box));
    
    /* Get the mask layer associated with the box */
    OH_ASSERT(  octGenFirstContainer(box, OCT_LAYER_MASK, layer));

    return OCT_OK;
}


/*
 * gnBoxCenter: Finds center of oct box
 */
octStatus gnBoxCenter(box, x, y)
octObject *box;
octCoord *x, *y;
{
    struct octBox bb;

    bb = box->contents.box;
    *x = (bb.upperRight.x - bb.lowerLeft.x)/2 + bb.lowerLeft.x;
    *y = (bb.upperRight.y - bb.lowerLeft.y)/2 + bb.lowerLeft.y;

    return OCT_OK;
}


/*
 * gnBoxWidthAndLength: Finds width and length of oct box
 */
octStatus gnBoxWidthAndLength(box, w, l)
octObject *box;
int *w, *l;
{
    struct octBox bb;

    bb = box->contents.box;
    *w = bb.upperRight.x - bb.lowerLeft.x;
    *l = bb.upperRight.y - bb.lowerLeft.y;

    return OCT_OK;
}


/*
 * gnMergeNetsInFacet: Merge any nets that share common terminals
 */
octStatus gnMergeNetsInFacet(facetPtr)
octObject *facetPtr;
{
    octObject instance, term;
    octGenerator genInst, genTerm;
    int i, count;

    /* get instance count */
    count = ohCountContents(facetPtr, OCT_INSTANCE_MASK);

    /* 
     * generate through all the instances 
     */
    OH_ASSERT(octInitGenContents(facetPtr, OCT_INSTANCE_MASK, &genInst));
    for (i = 0; i < count; i++) {
        OH_ASSERT(octGenerate(&genInst, &instance));

	/* 
	 * for each actual terminal on
	 * an instance, merge all its
	 * nets
	 */
	OH_ASSERT(octInitGenContents(&instance, OCT_TERM_MASK, &genTerm));
	while (octGenerate(&genTerm, &term) == OCT_OK) {
            OH_ASSERT(gnMergeNetsOnTerm(&term));
	}
    }
    octFreeGenerator( &genInst );

    /* 
     * for each formal terminal in
     * facet, merge all its nets
     */
    OH_ASSERT(octInitGenContents(facetPtr, OCT_TERM_MASK, &genTerm));
    while (octGenerate(&genTerm, &term) == OCT_OK) {
	OH_ASSERT(gnMergeNetsOnTerm(&term));
    }

    return OCT_OK;
}


/*
 * gnMergeNetsOnTerm: Merge multiple nets on a terminal
 */
octStatus gnMergeNetsOnTerm(termPtr)
octObject *termPtr;
{
    octObject net, merged_net;
    octGenerator genNet;
    int numNets;

    /* 
     * count the number of net containers to this
     * actual terminal 
     */
    numNets = ohCountContainers( termPtr, OCT_NET_MASK );

    /* 
     * merge them if there is more than one
     */
    if (numNets > 1) {
	/* 
	 * merge all nets into the first one
	 */
	OH_ASSERT(octInitGenContainers(termPtr, OCT_NET_MASK, &genNet));
	OH_ASSERT(octGenerate(&genNet, &merged_net));
	while (octGenerate(&genNet, &net) == OCT_OK) {
	    /* 
	     * first check to make sure the object id's
	     * of the nets are different.  this is because
	     * sometimes a terminal is attached to a net
	     * more than once.  shouldn't happen though,
	     * but should check.
	     */
	    if (net.objectId != merged_net.objectId) {
		/* 
		 * first copy the contents 
		 */
		/* fprintf( stderr, "Merging %s and %s\n", ohFormatName(&net), ohFormatName( &merged_net)); */
		ohAttachOnceContents(&net, &merged_net, OCT_ALL_MASK);
		/*
		 * copy containers
		 */
		ohAttachOnceContainers(&net, &merged_net, OCT_ALL_MASK &~ OCT_FACET_MASK);

		OH_ASSERT(octDelete(&net)); /* Delete the old net. */

	    } else {
		fprintf( stderr,"Detaching multiple attached net %s (%s)\n", 
			ohFormatName(&net), ohFormatName(termPtr));
		octDetach( &net, termPtr );
	    }
	}
    }

    return OCT_OK;
}


#ifndef lint
dumpFacet(facetPtr)
octObject *facetPtr;
{
    octObject instance, term, net;
    octGenerator genInst, genTerm, genNet;
    int fi, it, tn;
    char *objname;
    char *objname2;

    fi = 0;
    OH_ASSERT(octInitGenContents(facetPtr, OCT_INSTANCE_MASK, &genInst));
    while (octGenerate(&genInst, &instance) == OCT_OK) {
	objname = instance.contents.instance.master;
	objname2 = instance.contents.instance.name;
	printf("gen facet to instance (%d) master (%s) named (%s)\n", 
	  fi++, objname, objname2);

	it = 0;
	OH_ASSERT(octInitGenContents(&instance, OCT_TERM_MASK, &genTerm));
	while (octGenerate(&genTerm, &term) == OCT_OK) {
	    objname = term.contents.term.name;
	    printf("    gen instance to term (%d) named (%s)\n", it++, objname);

	    tn = 0;
	    OH_ASSERT(octInitGenContainers(&term, OCT_NET_MASK, &genNet));
	    while (octGenerate(&genNet, &net) == OCT_OK) {
		objname = net.contents.net.name;
		printf("        gen term to net (%d) named (%s)\n", tn++, objname);
	    }
	}
    }
}
#endif
