/*
 * implicits.c: Routines in this file attempt to determine
 *    the necessary formal terminals for this cell to interact
 *    externally.   these formal terminals are inferred from
 *    the boundaries
 */
#include "port.h"
#include "utility.h"
#include "fang.h"
#include "harpoon.h"
#include "oct.h"
#include "st.h"
#include "oh.h"
#include "fa.h"
#include "gn.h"
#include "list.h"
#include "kh.h"
#include "kahlua.h"

/*
 * handleImplicitFormals:
 */
handleImplicitFormals(inFacetPtr, outFacetPtr, geoListPtr, tsListPtr)
octObject *inFacetPtr, *outFacetPtr;
harList *geoListPtr;
khList *tsListPtr;
{
    octObject connectorBag, layerBag, layer;
    struct octBox facetBB;
    octGenerator genLayer;
    char *layername;
    int cx, cy;
    int size;
    octObject prop;
    int minimum;
    fa_box *fangBoxGen, *fangBoxList, fangBox, fangBB;
    fa_geometry *netGeo;

    /* print trace message */
    if (trace) {
	(void) fprintf(stderr, "%ssynthesizing implicit formal terminals ...\n", 
	  INDENT);
    }

    /* get the bounding box the facet and its center */
    OH_ASSERT(octBB(inFacetPtr, &facetBB));
    fangBB.left = facetBB.lowerLeft.x; 
    fangBB.bottom = facetBB.lowerLeft.y;
    fangBB.right = facetBB.upperRight.x; 
    fangBB.top = facetBB.upperRight.y;

    /* get connector bag */
    OH_ASSERT(ohGetByBagName(outFacetPtr, &connectorBag, "CONNECTORS"));

    /* for each net layer */
    OH_ASSERT(ohGetByBagName(&globalFacet, &layerBag, "NETS"));
    OH_ASSERT(octInitGenContents(&layerBag, OCT_LAYER_MASK, &genLayer));
    while (octGenerate(&genLayer, &layer) == OCT_OK) {
        layername = layer.contents.layer.name;

	/* grab minimum width */
	OH_ASSERT(ohGetByPropName(&layer, &prop, "MINIMUM_WIDTH"));
	minimum = prop.contents.prop.value.integer;

	/* grab net fang geo and term list */
        netGeo = harGetNamedGeo(layername, geoListPtr);

	/*
	 * now for each generated box, check if it lies
	 * on a border and which one.  if its width is
	 * between minimum width and X times the minimum
	 * width, then make it into a formal terminal.
	 * X is hard-wired to be 3 for now
	 */
	fa_to_boxes(netGeo, &fangBoxList);
	fangBoxGen = fangBoxList;
	while (faGenerateBox(&fangBoxGen, &fangBox) == 1) {
	    /* check left */
	    if (checkBoundaries(&fangBB, &fangBox, &size, &cx, &cy, minimum, 1)) {
		mkImplicit(tsListPtr, cx, cy, layername, size, &connectorBag, 
		  outFacetPtr);
	    }

	    /* check bottom */
	    if (checkBoundaries(&fangBB, &fangBox, &size, &cx, &cy, minimum, 2)) {
		mkImplicit(tsListPtr, cx, cy, layername, size, &connectorBag, 
		  outFacetPtr);
	    }

	    /* check top */
	    if (checkBoundaries(&fangBB, &fangBox, &size, &cx, &cy, minimum, 3)) {
		mkImplicit(tsListPtr, cx, cy, layername, size, &connectorBag, 
		  outFacetPtr);
	    }

	    /* check right */
	    if (checkBoundaries(&fangBB, &fangBox, &size, &cx, &cy, minimum, 4)) {
		mkImplicit(tsListPtr, cx, cy, layername, size, &connectorBag, 
		  outFacetPtr);
	    }
	}
    }

    /*
     * need to free memory
     */
    fa_free_boxes(fangBoxList);
    fa_free_boxes(&fangBB);
}


/*
 * checkBoundaries: Checks for terminals on the cell boundaries
 *    1 -- checks left
 *    2 -- checks bottom
 *    3 -- checks top
 *    4 -- checks right
 *
 * returns 1 if there is a terminal, and 0 otherwise
 */
checkBoundaries(fangBBPtr, fangBoxPtr, sizePtr, px, py, minimum, field)
fa_box *fangBBPtr, *fangBoxPtr;
int *sizePtr, *px, *py;
int minimum;
int field;
{
    int bbLeft, bbBottom, bbRight, bbTop;
    int boxLeft, boxBottom, boxRight, boxTop;
    int check = 0;
    int x, y;
    int maximum;
    int s;
    int height, width;

    /* initialize */
    maximum = 3 * minimum;

    /* bounding box coordinates */
    bbLeft = fangBBPtr->left;
    bbBottom = fangBBPtr->bottom;
    bbRight = fangBBPtr->right;
    bbTop = fangBBPtr->top;

    /* terminal box in question */
    boxLeft = fangBoxPtr->left;
    boxBottom = fangBoxPtr->bottom;
    boxRight = fangBoxPtr->right;
    boxTop = fangBoxPtr->top;


    /* box size */
    height = boxTop - boxBottom;
    width = boxRight - boxLeft;

    switch (field) {
	case 1:
	    /* check left */
	    if (boxLeft == bbLeft) {
		s = height;
		if ((height <= width) && (s <= maximum) && (s >= minimum)) {
		    check = 1;
		    x = boxLeft + (s/2);
		    y = boxTop - (s/2);
		}
	    }
	    break;
        case 2:
	    /* check bottom */
	    if (boxBottom == bbBottom) {
		s = width;
		if ((width <= height) && (s <= maximum) && (s >= minimum)) {
		    check = 1;
		    x = boxRight - (s/2);
		    y = boxBottom + (s/2);
		}
	    }
	    break;
        case 3:
	    /* check top */
	    if (boxTop == bbTop) {
		s = width;
		if ((width <= height) && (s <= maximum) && (s >= minimum)) {
		    check = 1;
		    x = boxRight - (s/2);
		    y = boxTop - (s/2);
		}
	    }
	    break;
        case 4:
	    /* check right */
	    if (boxRight == bbRight) {
		s = height;
		if ((height <= width) && (s <= maximum) && (s >= minimum)) {
		    check = 1;
		    x = boxRight - (s/2);
		    y = boxTop - (s/2);
		}
	    }
	    break;
	default:
	    (void) fprintf(stderr, "Panic: don't understand field\n");
	    OH_ASSERT(OCT_ERROR);
	    exit(-1);
	    break;
    }

    /* returns */
    *sizePtr = s;
    *px = x;
    *py = y;
    return check;
}


/*
 * checkExistingTerminal: check to see if there is already a device terminal
 *    at the location.
 *
 *    return 1: if a terminal is found with the same x,y location.
 *           0: if none exist.
 */
checkExistingTerminal(tsListPtr, x, y)
khList *tsListPtr;
int x, y;
{
    lsList list;
    lsGen gen;
    kh_term *entry;
    int ret_val = 0;

    while (*tsListPtr != KH_EMPTY_LIST) {
	list = khFirstLs(*tsListPtr);
	gen = lsStart(list);
	while (lsNext(gen, (Pointer *) &entry, LS_NH) == LS_OK) {
	    if ((entry->position.x == x) && (entry->position.y == y)) {
		ret_val = 1;
	    }
	}
    }

    /* free generator */
    lsFinish(gen);

    return ret_val;
}


mkImplicit(tsListPtr, cx, cy, layername, size, connectorBagPtr, outFacetPtr)
khList *tsListPtr;
int cx, cy;
char *layername;
int size;
octObject *connectorBagPtr, *outFacetPtr;
{
    octObject instance, net, fterm, aterm;
    lsList termList;
    kh_term *termEntry;
    char master[BUFFER_SIZE], view[BUFFER_SIZE], name[BUFFER_SIZE];
    octTransformType type;
    int check;

    check = checkExistingTerminal(tsListPtr, cx, cy);
    if (!check) {
	/*
	 * make master if necessary
	 * then instantiate it
	 */
	makeConnectorMaster(layername, size, size, &type, 
	  master, view);
	mkInstanceName(name)
	/* keep count */
	connectorCount++;
	OH_ASSERT(ohCreateInstance(outFacetPtr, &instance, name, 
	  master, view, cx, cy, type));
	OH_ASSERT(octAttach(connectorBagPtr, &instance));

	/*
	 * make formal terminal
	 */
	mkFtermName(name);
	OH_ASSERT(ohCreateTerm(outFacetPtr, &fterm, name));

	/*
	 * make a net.
	 * get actual terminal of instance.
	 * attache fterm to net, and aterm to fterm
	 * and net
	 */
	mkNetName(name);
	OH_ASSERT(octGenFirstContent(&instance, OCT_TERM_MASK, &aterm));
	OH_ASSERT(ohCreateNet(outFacetPtr, &net, name));
	OH_ASSERT(octAttach(&net, &fterm));
	/* OH_ASSERT(octAttach(&net, &aterm)); */
	OH_ASSERT(octAttach(&fterm, &aterm));

	/* make terminal entry */
	termList = khGetNamedLs(layername, tsListPtr);
	termEntry = termConstruct( cx,cy,aterm.objectId );
	(void) lsNewBegin(termList, (Pointer) termEntry, LS_NH);
    }
}
