/*
 * main.c -- basic front end for handling options and defaults
 */
#include "port.h"
#include "utility.h"
#include "oct.h"
#include "st.h"
#include "oh.h"
#include "options.h"
#include "gn.h"
#include "list.h"
#include "kahlua.h"
#include "vov.h"
#include <sys/file.h>

#define VERSION "2.0(AC)"

/* options variables */
int parentOnly = 0;
int deviceOnly = 0;
int explicitFterm = 1;
int implicitFterm = 0;
int trace = 0;
int minWidth = 0;
int expandFactor = 1;
char *leafView = "physical";
char *defLibDir = "./kahlua_prims";
char *techFile;
char *technology;
int ratio = 20;
int interactive = 1;


optionStruct optionList[] = {
    { "o",  "cell[:view]",    "Specify output cell or view" },
    { "l",  "leafview",	      "Specify output view for leaf cells" },
    { "t",  "techfile",	      "Specify technology file" },
    { "e",  "expand",	      "Expand design rules 'expand' factor" },
    { "p",  "directory",      "Specify directory for primitives" },
    { "a",  "ratio",	      "Specify maximum aspect ratio for paths" },
    /* { "d", 0,		      "Do devices only" }, */
    /* { "n", 0,		      "Do parent only" }, */
    /* { "f", 0,		      "Do implicit Formal terminals" }, */
    { "g", 0,	              "Ignore explicit formal terminals" },
    /* { "i", 0, 		      "Interactive" }, */
    { "m", 0,	              "Force minimum segment width" },
    { "v", 0,	              "Verbose information as kahlua executes" },
    { OPT_RARG,	"cell:view",  "input cell and view" },
    { 0,	0,	      0 }
};



main (argc, argv)
    int argc;
    char **argv;
{
     int c;
     extern int optind;
     extern char *optarg;
     char *inCell, *inView, *outCell, *outView;
     int techFileDefault = 1;
     extern void *setPrimitiveDirectory();
     extern char *getTechnology();

     /* set default i/o cell:view */
     inView = "physical";
     outView = "symbolic";
     parse_facet_spec(util_strsav(argv[argc-1]), &inCell, &inView);
     outCell = util_strsav(inCell);

     /* parse options */
     while ((c = optGetOption( argc, argv )) != EOF) {
	 switch (c) {
	 case 'd':
	     deviceOnly = 1;
	     implicitFterm = 0;
	     explicitFterm = 0;
	     break;
	 case 'n':
	     parentOnly = 1;
	     break;
	 case 'f':
	     implicitFterm = 1;
	     break;
	 case 'g':
	     explicitFterm = 0;
	     break;
	 case 'v':
	     trace = 1;
	     break;
	 case 'i':
	     interactive = 0;
	     break;
	 case 'm':
	     minWidth = 1;
	     break;
	 case 'o':
	     parse_facet_spec(util_strsav(optarg), &outCell, &outView);
	     break;
	 case 'l':
	     leafView = util_strsav(optarg);
	     break;
	 case 't':
	     techFile = util_strsav(optarg);
	     techFileDefault = 0;
	     break;
	 case 'e':
	     expandFactor = atoi(optarg);
	     break;
	 case 'p':
	     defLibDir = util_strsav(optarg);
	     break;
	 case 'a':
	     ratio = atoi(optarg);
	     break;
	 default:
	     optUsage();
	 }
    }

    /* finish the command line argument parsing */
    if (optind != argc-1) {
	optUsage();
    }

     VOVbegin( argc, argv );
    /* grab technology property from input facet */
    technology = getTechnology(inCell, inView);

    /* set up default directory for storing the primitives */
     setPrimitiveDirectory( defLibDir );

    /* set up default technology file for reading process rules */
    if (techFileDefault) {
	techFile = ALLOC(char, BUFFER_SIZE);
	(void) sprintf(techFile,"%s/tech/%s/design_rules", "$OCTTOOLS", technology);
    }

    /* print trace message */
    if (trace) {
	(void) fprintf(stderr, "Kahlua version %s initialized\n\n", VERSION);
	(void) fprintf(stderr, "TECHNOLOGY = %s\n", technology);
	(void) fprintf(stderr, "using technology file: %s\n\n", techFile);
    }

    /* 
     * execute the process
     */
    initGlobals();
    disassemble(inCell, inView, outCell, outView);
    endGlobals();

     VOVend(0);
}


#ifdef OLD_CODE
/*
 * usage: Prints out usage
 */
usage(command)
char *command;
{
     (void) fprintf(stderr, "\n");
     (void) fprintf(stderr, "usage: %s [options] cellname[:viewname] \n", command);
     (void) fprintf(stderr, "\n");
     (void) fprintf(stderr, "  -o cell[:view]  Specify output cell or view\n");
     (void) fprintf(stderr, "  -l leafview\t  Specify output view for leaf cells\n");
     (void) fprintf(stderr, "  -t techfile\t  Specify technology file\n");
     (void) fprintf(stderr, "  -e expand\t  Expand design rules 'expand' factor\n");
     (void) fprintf(stderr, "  -p directory\t  Specify directory for primitives\n");
     (void) fprintf(stderr, "  -a ratio\t  Specify maximum aspect ratio for paths\n");
     (void) fprintf(stderr, "  -g\t\t  Ignore explicit formal terminals\n");
     (void) fprintf(stderr, "  -m\t\t  Force minimum segment width\n");
     (void) fprintf(stderr, "  -v\t\t  Verbose information as kahlua executes\n");
/*
 * IGNORE for the time being
 *
 * (void) fprintf(stderr, "  -n\t\t  Disassemble parent level only\n");
 * (void) fprintf(stderr, "  -d\t\t  Perform device extraction only\n");
 * (void) fprintf(stderr,"  -f\t\t  Attempt to determine implicit formal terminals\n");
 * (void) fprintf(stderr, "  -i\t\t  Turn off interactive facilities\n");
*/
     (void) fprintf(stderr, "\n");
     exit(-1);
}
#endif

/*
 * setPrimitiveDirectory: Set default primitive directory name.
 *    Make the necessary sub-directory if it does not exist.
 */
void  *setPrimitiveDirectory( path )
    char *path;
{
    if ( access( path, F_OK ) == -1 ) {
	printf("Creating directory %s", path );
	mkdir( path, 0777 );
    }
}


/*
 * getTechnology: Grab TECHNOLOGY property from input facet.
 */
char *getTechnology(cell, view)
char *cell, *view;
{
    octObject facet, prop;

    OH_ASSERT(ohOpenFacet(&facet, cell, view, "contents", "r"));

    OH_ASSERT(ohGetByPropName(&facet, &prop, "TECHNOLOGY"));
    OH_ASSERT(octCloseFacet(&facet));

    return util_strsav(prop.contents.prop.value.string);
}

