/*
 * map.c
 *
 * Generic Virtual Coordinate Mapping package
 *
 * Bill Lin
 * University of California, Berkeley
 *
 * This file contains the procedural routines available with
 * the Virtual Coordinate Mapping package.  This is stand 
 * alone package.
 */

#include "port.h"
#include "utility.h"
#include "map.h"

struct sort_obj_struct {
    int obj;
    struct sort_obj_struct *next;
};

#define TYPE		struct sort_obj_struct
#define NEXT		next
#define FIELD		obj
#define SORT		linked_list_sort_helper
#define SORT1		linked_list_sort

#include "lsort.h"	/* generic sorting package */

/*
 * mapCreate: Creates a map table by allocating memory
 */
mapTable *mapCreate()
{
    return ALLOC(mapTable, 1);
}


/*
 * mapFree: Frees the memory allocated for mapTable
 */
void mapFree(tablePtr)
mapTable **tablePtr;
{
    mapTable *table;

    table = *tablePtr;
    FREE(table->table);
    FREE(table);
    return;
}


/*
 * mapSortTable: Sort the integer table in-place 
 *     using the compare function 'compare'.
 *
 * returns: MAP_OK if success, and
 *          MAP_ERROR if failed.
 */
mapStatus mapSortTable(tableMap, compare)
mapTable *tableMap;
int (*compare)();
{
    int i;
    int size;
    int *table;
    struct sort_obj_struct *linknext, *head, *link;

    size = tableMap->size;
    table = tableMap->table;

    /* Gen all the entries into a linked list */
    head = NIL(struct sort_obj_struct);
    for (i = 0; i < size; i++) {
	if ((link = ALLOC(struct sort_obj_struct, 1)) == 0) {
	    return MAP_ERROR;
	}
	link->next = head;
	head = link;
	link->obj = table[i];
    }

    /* Sort this linked list */
    head = linked_list_sort(head, compare);

    /* Re-copy the entries to the table */
    i = 0;
    for(link = head; link != 0; link = linknext) {
       if (i > size) {
	   return MAP_ERROR;
       }
       table[i] = link->obj;
       linknext = link->next;
       FREE(link);
       i++;
    }

    return MAP_OK;
}


/*
 * mapVirtualCoord: Returns in 'vxPtr' the virtual coordinate
 *    of 'rxPtr'
 *
 * returns: MAP_OK if found, and
 *          MAP_ERROR if not found
 */
mapStatus mapVirtualCoord(tableMap, rxPtr, vxPtr)
mapTable *tableMap;
int *rxPtr;
int *vxPtr;
{
    int size;
    int *table;
    int i, rx;

    size = tableMap->size;
    table = tableMap->table;
    rx = *rxPtr;

    for (i = 0; i < size; i++) {
	if (table[i] == rx) {
	    *vxPtr = i;
	    return MAP_OK;
	}
    }

    return MAP_ERROR;
}


/*
 * mapRealCoord: Returns in 'rxPtr' the actual coordinate
 *    of 'vxPtr'
 *
 * returns: MAP_OK if found, and
 *          MAP_ERROR if not found
 */
mapStatus mapRealCoord(tableMap, vxPtr, rxPtr)
mapTable *tableMap;
int *vxPtr;
int *rxPtr;
{
    int *table;
    int vx;

    table = tableMap->table;
    vx = *vxPtr;

    *rxPtr = table[vx];
    return MAP_OK;

}


/*
 * mapCompare1: Compare function for sorting table
 *    in increasing order
 * returns: 0 if list1->obj <= list2->obj
 *          1 otherwise
 */
int mapCompare1(list1, list2)
struct sort_obj_struct *list1, *list2;
{
    if (list1->obj <= list2->obj) {
	return 0;
    } else {
	return 1;
    }
}


/*
 * mapCompare2: Compare function for sorting table
 *    in decreasing order
 * returns: 0 if list1->obj >= list2->obj
 *          1 otherwise
 */
int mapCompare2(list1, list2)
struct sort_obj_struct *list1, *list2;
{
    if (list1->obj >= list2->obj) {
	return 0;
    } else {
	return 1;
    }
}

#ifndef lint
/*
 * mapDumpTable: Prints the map table in a
 *    human readable form
 */
void mapDumpTable(tableMap)
mapTable *tableMap;
{
    int i;
    int size;
    int *table;

    size = tableMap->size;
    table = tableMap->table;

    (void) fprintf(stdout, "virtual -- real\n");
    for (i = 0; i < size; i++) {
	(void) fprintf(stdout, "\t%d\t%d\n", i, table[i]);
    }
}
#endif

