/*
 * masters.c: Routines for creating cell masters
 */
#include "port.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "gn.h"
#include "list.h"
#include "masters.h"

/*
 * lower: Returns the lowercase of 't'
 */
lower(t)
char *t;
{
    char c;

    while (*t) {
	c = *t;
	*t = (isupper(c) ? tolower(c) : c);
	t++;
    }
}


/*
 * makeConnectorMaster: Creates a connector cell master if one
 *   does not exist already.  master:view name will be constructed
 *   from path/root.widthxlength:view.  where --
 *       path = libDir;
 *       root = lower(name);
 *       width = w;
 *       length = l;
 *       view = leafView;
 */
makeConnectorMaster(name, w, l, type, master, view)
    char *name;
    int w, l;
    octTransformType *type;
    char *master, *view;
{
    octObject bag;
    int width, length;
    char *root, *path;
    char entry[BUFFER_SIZE];

    /*
     * define master and view names
     */
    if (w > l) {
	length = w;
	width = l;
	*type = OCT_ROT90;
    } else {
	length = l;
	width = w;
	*type = OCT_NO_TRANSFORM;
    }

    (void) sprintf(view, "%s", leafView);
    root = util_strsav(name);
    lower(root);
    path = defLibDir;

    (void) sprintf(master, "%s/%s.%dx%d", path, root, width, length);
    (void) sprintf(entry, "%s:%s", master, view);

    /*
     * make one if it doesn't exist already
     */
    if (ohGetByBagName(&masBag, &bag, entry) != OCT_OK) {
	/* print trace message */
	if (trace) {
	    (void) fprintf(stderr, "%s%screating leaf cell [%s:%s] ...\n", 
	      INDENT, INDENT, master, view);
	}

	OH_ASSERT(ohCreateBag(&masBag, &bag, entry));
	/* keep count */
	uniqueConnectorCount++;
	makeInternalConnectorMaster(master, view, name, width, length);
    }
}


/*
 * makeInternalConnectorMaster: Make a connector master
 */
makeInternalConnectorMaster(master, view, name, w, l)
    char *master, *view, *name;
    int w, l;
{
    octObject facet, ifacet, fterm, layer, box, prop;
    int llx, lly, urx, ury;

    /* determine box coordinates */
    llx = - (w/2); urx = (w/2);
    lly = - (l/2); ury = (l/2);

    /* open master */
    OH_ASSERT(ohOpenFacet(&facet, master, view, "contents", "w"));

    OH_ASSERT(ohCreateLayer(&facet, &layer, name));
    OH_ASSERT(ohCreateBox(&layer, &box, llx, lly, urx, ury));
    OH_ASSERT(ohCreateTerm(&facet, &fterm, "t"));
    OH_ASSERT(octAttach(&fterm, &box));

    /* add properties */
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "TECHNOLOGY", technology));
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "VIEWTYPE", "PHYSICAL"));
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "EDITSTYLE", "PHYSICAL"));
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "CELLCLASS", "LEAF"));

    /* close facet and generate interface */
    OH_ASSERT(gnGenInterfaceFacet(&facet,&ifacet));
    OH_ASSERT(ohCreatePropStr(&ifacet, &prop, "CELLTYPE", "CONTACT"));
    OH_ASSERT(octFlushFacet(&facet));
    OH_ASSERT(octFlushFacet(&ifacet));
}


/*
 * makeContactMaster: Creates a contact cell master if one
 *   does not exist already.  master:view name will be constructed
 *   from path/root.widthxlength:view.  where --
 *       path = libDir;
 *       root = lower(name);
 *       width = (w + overlap);
 *       length = (l + overlap);
 *       view = leafView;
 */
makeContactMaster(name, name1, name2, space, cut,
  overlap1, overlap2, w, l, type, master, view)
char *name;
char *name1, *name2;
int space, cut, overlap1, overlap2;
int w, l;
octTransformType *type;
char *master, *view;
{
    octObject bag;
    int overlap;
    int width, length;
    char *root, *path;
    char entry[BUFFER_SIZE];

    overlap = ((overlap1 > overlap2) ? overlap1 : overlap2);

    /*
     * define master and view names
     */
    if (w > l) {
	length = w;
	width = l;
	*type = OCT_ROT90;
    } else {
	length = l;
	width = w;
	*type = OCT_NO_TRANSFORM;
    }

    (void) sprintf(view, "%s", leafView);
    root = util_strsav(name);
    lower(root);
    path = defLibDir;

    (void) sprintf(master, "%s/%s.%dx%d", path, root, 
      (width + (2 * (overlap))), (length + (2 * (overlap))));
    (void) sprintf(entry, "%s:%s", master, view);

    /*
     * make one if it doesn't exist already
     */
    if (ohGetByBagName(&masBag, &bag, entry) != OCT_OK) {
	/* print trace message */
	if (trace) {
	    (void) fprintf(stderr, "%s%screating leaf cell [%s:%s] ...\n", 
	      INDENT, INDENT, master, view);
	}

	OH_ASSERT(ohCreateBag(&masBag, &bag, entry));
	/* keep count */
	uniqueContactCount++;
	makeInternalContactMaster(master, view, name, 
	  name1, name2, space, cut, overlap1, overlap2, width, length);
    }
}


/*
 * makeInternalContactMaster: Make a contact master
 */
makeInternalContactMaster(master, view, name, name1, name2, 
  space, cut, overlap1, overlap2, w, l)
char *master, *view, *name;
char *name1, *name2;
int space, cut, overlap1, overlap2;
int w, l;
{
    octObject facet, ifacet, fterm, layer, box1, box2, prop;
    int llx, lly, urx, ury;
    int tllx, tlly, turx, tury;
    int x, y;

    /* determine box coordinates */
    llx = - (w/2); urx = (w/2);
    lly = - (l/2); ury = (l/2);
    effectiveCoord(w, l, space, cut, &tllx, &tlly, &turx, &tury);

    /* open master */
    OH_ASSERT(ohOpenFacet(&facet, master, view, "contents", "w"));

    /* create formal terminal */
    OH_ASSERT(ohCreateTerm(&facet, &fterm, "t"));

    /* create cut geo */
    OH_ASSERT(ohCreateLayer(&facet, &layer, name));
    for (x = tllx; x <= turx - cut; x = x + cut + space) {
	for (y = tlly; y <= tury - cut; y = y + cut + space) {
	    OH_ASSERT(ohCreateBox(&layer, &box1, x, y, x + cut, y + cut));
	    OH_ASSERT(octAttach(&fterm, &box1));
	}
    }

    /* create layer1 geo */
    OH_ASSERT(ohCreateLayer(&facet, &layer, name1));
    OH_ASSERT(ohCreateBox(&layer, &box1, 
      llx - overlap1, lly - overlap1, 
      urx + overlap1, ury + overlap1));
    OH_ASSERT(octAttach(&fterm, &box1));

    /* create layer2 geo */
    OH_ASSERT(ohCreateLayer(&facet, &layer, name2));
    OH_ASSERT(ohCreateBox(&layer, &box2, 
      llx - overlap2, lly - overlap2, 
      urx + overlap2, ury + overlap2));
    OH_ASSERT(octAttach(&fterm, &box2));

    /* add properties */
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "TECHNOLOGY", technology));
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "VIEWTYPE", "PHYSICAL"));
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "EDITSTYLE", "PHYSICAL"));
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "CELLCLASS", "LEAF"));

    /* close facet and generate interface */
    OH_ASSERT(gnGenInterfaceFacet(&facet,&ifacet));
    OH_ASSERT(ohCreatePropStr(&ifacet, &prop, "CELLTYPE", "CONTACT"));
    OH_ASSERT(ohCreatePropStr(&ifacet, &prop, "CONTACTTYPE", name));
    OH_ASSERT(octFlushFacet(&facet));
    OH_ASSERT(octFlushFacet(&ifacet));
}


effectiveCoord(w, l, space, cut, tllx, tlly, turx, tury)
int w, l, space, cut;
int *tllx, *tlly, *turx, *tury;
{
    int nx, ny;

    nx = w + space;
    ny = l + space;

    nx = nx / (cut + space);
    ny = ny / (cut + space);

    nx *= (cut + space);
    ny *= (cut + space);

    nx -= space;
    ny -= space;

    *tllx = -(nx/2); *turx = (nx/2);
    *tlly = -(ny/2); *tury = (ny/2);
}


/*
 * makeMosfetMaster: Creates a mosfet cell master if one
 *   does not exist already.  master:view name will be constructed
 *   from path/root.widthxlength:view.  where --
 *       path = libDir;
 *       root = lower(name);
 *       width = w;
 *       length = l;
 *       view = leafView;
 */
makeMosfetMaster(name, name1, name2, overlap1, overlap2, w, l, master, view)
char *name;
char *name1, *name2;
int overlap1, overlap2;
int w, l;
char *master, *view;
{
    octObject bag;
    int width, length;
    char *root, *path;
    char entry[BUFFER_SIZE];

    length = l;
    width = w;

    (void) sprintf(view, "%s", leafView);
    root = util_strsav(name);
    lower(root);
    path = defLibDir;

    (void) sprintf(master, "%s/%s.%dx%d", path, root, length, width);
    (void) sprintf(entry, "%s:%s", master, view);

    /*
     * make one if it doesn't exist already
     */
    if (ohGetByBagName(&masBag, &bag, entry) != OCT_OK) {
	/* print trace message */
	if (trace) {
	    (void) fprintf(stderr, "%s%screating leaf cell [%s:%s] ...\n", 
	      INDENT, INDENT, master, view);
	}

	OH_ASSERT(ohCreateBag(&masBag, &bag, entry));
	/* keep count */
	uniqueMosfetCount++;
	makeInternalMosfetMaster(master, view, name, name1, name2, 
	  overlap1, overlap2, width, length);
    }
}


/*
 * makeInternalMosfetMaster: Make a mosfet master
 */
makeInternalMosfetMaster(master, view, name, name1, name2, 
  overlap1, overlap2, w, l)
char *master, *view, *name;
char *name1, *name2;
int overlap1, overlap2;
int w, l;
{
    octObject facet, ifacet, fterm, layer1, layer2, box, prop;
    octObject jtBag, jBag;
    int llx, lly, urx, ury;

    /* determine box coordinates */
    llx = - (w/2); urx = (w/2);
    lly = - (l/2); ury = (l/2);

    /* open master */
    OH_ASSERT(ohOpenFacet(&facet, master, view, "contents", "w"));

    /* create layers */
    OH_ASSERT(ohCreateLayer(&facet, &layer1, name1));
    OH_ASSERT(ohCreateLayer(&facet, &layer2, name2));

    /* create active area first */
    OH_ASSERT(ohCreateBox(&layer1, &box, llx, lly, urx, ury));
    OH_ASSERT(ohCreateBox(&layer2, &box, llx, lly, urx, ury));

    /* create gate term */
    OH_ASSERT(ohCreateBox(&layer1, &box, llx, ury, urx, ury + overlap1));
    OH_ASSERT(ohCreateTerm(&facet, &fterm, "GATE"));
    OH_ASSERT(octAttach(&fterm, &box));

    /* create GATE1 term */
    OH_ASSERT(ohCreateBox(&layer1, &box, llx, lly - overlap1, urx, lly));
    OH_ASSERT(ohCreateTerm(&facet, &fterm, "GATE1"));
    OH_ASSERT(octAttach(&fterm, &box));

    /* create source term */
    OH_ASSERT(ohCreateBox(&layer2, &box, llx - overlap2, lly, llx, ury));
    OH_ASSERT(ohCreateTerm(&facet, &fterm, "SOURCE"));
    OH_ASSERT(octAttach(&fterm, &box));

    /* create drain term */
    OH_ASSERT(ohCreateBox(&layer2, &box, urx, lly, urx + overlap2, ury));
    OH_ASSERT(ohCreateTerm(&facet, &fterm, "DRAIN"));
    OH_ASSERT(octAttach(&fterm, &box));

    /* add properties */
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "TECHNOLOGY", technology));
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "VIEWTYPE", "PHYSICAL"));
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "EDITSTYLE", "PHYSICAL"));
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "CELLCLASS", "LEAF"));

    /* close facet and generate interface */
    OH_ASSERT(gnGenInterfaceFacet(&facet,&ifacet));

    /* addition interface information */
    OH_ASSERT(ohCreatePropStr(&ifacet, &prop, "CELLTYPE", "MOSFET"));
    OH_ASSERT(ohCreatePropStr(&ifacet, &prop, "MOSFETTYPE", name));
    OH_ASSERT(ohCreatePropReal(&ifacet, &prop, "MOSFETLENGTH", (double) w));
    OH_ASSERT(ohCreatePropReal(&ifacet, &prop, "MOSFETWIDTH", (double) l));

    /* make join terms for gate terminals */
    OH_ASSERT(ohCreateBag(&ifacet, &jtBag, "JOINEDTERMS"));
    OH_ASSERT(ohCreateBag(&jtBag, &jBag, "JOINED"));
    OH_ASSERT(ohGetByTermName(&ifacet, &fterm, "GATE"));
    OH_ASSERT(octAttach(&jBag, &fterm));
    OH_ASSERT(ohGetByTermName(&ifacet, &fterm, "GATE1"));
    OH_ASSERT(octAttach(&jBag, &fterm));
    OH_ASSERT(octCloseFacet(&facet));
    OH_ASSERT(octCloseFacet(&ifacet));
}


