/*
 * mazeroute.c
 *    generic mazerouting package.
 *
 * last modified on September 6, 1987
 */
#include "port.h"
#include "utility.h"
#include <sys/file.h>
#include "fang.h"
#include "oct.h"
#include "st.h"
#include "oh.h"
#include "fa.h"
#include "gn.h"
#include "list.h"
#include "map.h"
#include "kahlua.h"

mapStatus mpAddEntry();

/*
 * mazeroute: Routes the specified problem -- region
 *    defines the boundaries of the routing region;
 *    termList defines the list of terminals that
 *    the routing solution needs to connect.  Note
 *    that actual routing routine is in 'route'.
 *
 *    returns: 0 if unsuccessful, and 1 otherwise
 */
int mazeroute(region, termList, pathList)
fa_region *region;
lsList termList;
lsList pathList;
{
    int s;
    mapTable *xTable, *yTable;

    /* 
     * route region only if it has more
     * than one terminal
     */
    if (lsLength(termList) < 2) {
	return 0;
    }

    /*
     * set up the mapping table for use
     * in a virtual grid mode.  first
     * create the map tables.  then make
     * the entries.  finally, sort the tables
     */
    xTable = mapCreate();
    yTable = mapCreate();
    buildMapTables(region, termList, xTable, yTable);
    MAP_ASSERT(mapSortTable(xTable, mapCompare1));
    MAP_ASSERT(mapSortTable(yTable, mapCompare2));

    /* call the virtual grid router */
    s = router(region, termList, pathList, xTable, yTable);

    /* free map tables */
    mapFree(&xTable);
    mapFree(&yTable);

    /* return */
    return s;
}


/*
 * buildMapTables: Maps the real coordinates of the
 *    region boundaries and terminal locations into
 *    virtual coordinates.  tables are assumed to have
 *    memory allocated to them using mapCreate() already.
 */
buildMapTables(region, termList, xTable, yTable)
fa_region *region;
lsList termList;
mapTable *xTable, *yTable;
{
    int rcount, tcount;
    int maxSize;
    int sizeX, sizeY;
    int x, y;
    int *tableX, *tableY;
    kh_term *termEntry;
    lsGen termGen;
    fa_box *fangBoxGen, *fangBoxList, fangBox;

    rcount = region->count;
    tcount = lsLength(termList);
    maxSize = 2 * (rcount + tcount);

    /* allocate table memory space */
    tableX = ALLOC(int, maxSize);
    tableY = ALLOC(int, maxSize);

    /* init sizes */
    sizeX = 0; sizeY = 0;

    /* generate thru regions */
    fa_regions_to_boxes(region, &fangBoxList);
    fangBoxGen = fangBoxList;
    while (faGenerateBox(&fangBoxGen, &fangBox) == 1) {
	/* add four box corners */
	MAP_ASSERT(mpAddEntry(tableX, fangBox.left, maxSize, &sizeX));
	MAP_ASSERT(mpAddEntry(tableY, fangBox.bottom, maxSize, &sizeY));
	MAP_ASSERT(mpAddEntry(tableX, fangBox.right, maxSize, &sizeX));
	MAP_ASSERT(mpAddEntry(tableY, fangBox.top, maxSize, &sizeY));

	/* add box center */
	faBoxCenter(&fangBox, &x, &y);
	MAP_ASSERT(mpAddEntry(tableX, x, maxSize, &sizeX));
	MAP_ASSERT(mpAddEntry(tableY, y, maxSize, &sizeY));
    }

    /* generate thru terms */
    termGen = lsStart(termList);
    while (lsNext(termGen, (Pointer *) &termEntry, LS_NH) == LS_OK) {
	x = termEntry->position.x;
	y = termEntry->position.y;

	/* add terminal location */
	MAP_ASSERT(mpAddEntry(tableX, x, maxSize, &sizeX));
	MAP_ASSERT(mpAddEntry(tableY, y, maxSize, &sizeY));
    }

    /* assign to map tables */
    xTable->size = sizeX;
    xTable->table = tableX;
    yTable->size = sizeY;
    yTable->table = tableY;

    /*
     * need to free memory
     */
    fa_free_boxes(fangBoxList);
    lsFinish(termGen);
}


/*
 * mpAddEntry: Looks into table for an entry
 *    value of 'c'.  Creates an entry if one
 *    does not exist and update the 'sizePtr'.
 *
 * returns MAP_OK on success, and MAP_ERROR on failure
 *
 */
mapStatus mpAddEntry(table, c, maxSize, sizePtr)
int *table;
int c;
int maxSize;
int *sizePtr;
{
    int i;

    /* error check */
    if (*sizePtr > maxSize) {
	return MAP_ERROR;
    }

    /* if table empty, automatically add entry */
    if (*sizePtr == 0) {
	table[0] = c;
	*sizePtr = 1;
	return MAP_OK;
    }

    /* do nothing if exist */
    for (i = 0; i < *sizePtr; i++) {
	if (table[i] == c) {
	    return MAP_OK;
	}
    }

    /* error check */
    if (*sizePtr == maxSize) {
	return MAP_ERROR;
    }

    /* add one if does not exist */
    table[*sizePtr] = c;
    *sizePtr += 1;

    return MAP_OK;
}

