/* Multidigit.c: Operate with multidigit transistors */


#include "port.h"
#include "utility.h"
#include "fang.h"
#include "harpoon.h"
#include "oct.h"
#include "st.h"
#include "oh.h"
#include "fa.h"
#include "gn.h"
#include "list.h"
#include "kh.h"
#include "kahlua.h"
#include "da.h"
#include "errtrap.h"

extern char *defLibDir;

char* termName( name, i )
    char* name;
    int i;
{
    static char tname[64];
    if ( i ) {
	sprintf( tname, "%s%d", name, i );
    } else {
	strcpy( tname, name );
    }
    return tname;
}


void regionName( name, reg_number )
    char name[];
    int reg_number;
{
    if ( reg_number == 0 ) strcpy( name, "DRAIN" ); 
    else if ( reg_number == 1 ) strcpy( name, "SOURCE" );
    else sprintf( name, "REG_%d", reg_number-2 );
}




int createJoinedBag( ifacet, name )
    octObject *ifacet;
    char* name;
/* Returns the number of joined terms with similar name */
{
    octObject fterm;
    octObject jtBag;
    octObject jBag;
    int count = 0;

    /* make join terms for source terminals */
    if(ohGetByTermName(ifacet, &fterm, termName(name,1) ) == OCT_OK) {
	OH_ASSERT( ohGetOrCreateBag( ifacet, &jtBag, "JOINEDTERMS" ));
	OH_ASSERT( ohCreateBag(&jtBag, &jBag, "JOINED") );
	while(ohGetByTermName( ifacet, &fterm, termName(name,count)) == OCT_OK ) {
	    OH_ASSERT(octAttach(&jBag, &fterm));
	    count++;
	}
    }
    return count;			     /*  */
}



void MDgetSmallestDiffGeo( diffGeo, mosfetGeo, newGeo )
    fa_geometry *diffGeo;
    fa_geometry *mosfetGeo;
    fa_geometry *newGeo;
{
    fa_geometry g1,g2,g3;		     /* Temporary geometries. */
    
    fa_copy( mosfetGeo, &g1 );
    fa_grow( &g1, 40, &g2 );		     /* Grow poly by 2 lambda. */
    fa_and( diffGeo, &g2, &g3 );	     /* Intersect grown poly with diff. */
#ifdef OLD
    fa_grow( &g3, 20, &g1 );		     /* Smooth diff region by growth and shrink */
    fa_grow( &g1, -20, &g3 );		     /*  */
    fa_smooth( &g3, 40, &g2 );		     /* Eliminate 1 lambda features (min is 2 lambdas) */
#endif
    fa_andnot( &g3, mosfetGeo, &g1 );	     /* Subtract mosfet geo again. */
    fa_copy( &g1, newGeo );		     /* Save result.  */
    
    fa_free_geo( &g1 );
    fa_free_geo( &g2 );
    fa_free_geo( &g3 );
}

MDCreateDiffRegions( diffGeo, mosfetGeo, dlayer, facet )
    fa_geometry *diffGeo;
    fa_geometry *mosfetGeo;
    octObject *dlayer;
    octObject *facet;
{
    int   count_regions = 0;
    char  name[64];
    fa_geometry smallDiffGeo;
    fa_region *regionList;
    fa_region *regionGen, region;

    
    MDgetSmallestDiffGeo( diffGeo, mosfetGeo, &smallDiffGeo );
#ifdef TEST
    fa_copy( diffGeo, &smallDiffGeo );
#endif

    fa_to_regions(&smallDiffGeo, &regionList);

    regionGen = regionList;
    while(faGenerateRegion(&regionGen, &region) == 1) {
	
	if ( !faGeoRegionIntersect( mosfetGeo, &region ) ) {
	    continue;
	}

	if ( region.interior_regions ) {
	    /* 
	     * This routine handles regions with a hole.
	     * For complete generality, It should be rewritten in recursive style.
	     */
	    fprintf( stderr, "Found region with a hole in it\n" );
	    if (region.interior_regions->interior_regions ) {
		fa_region *internal_region = region.interior_regions->interior_regions;
		fprintf( stderr, "Found region with inside the hole\n" );
		regionName( name, count_regions );
		myCreateTerm(internal_region, dlayer, facet, name );
		count_regions++;
		region.interior_regions->interior_regions = 0; /* So the following code does not see the hole.  */
	    }
	} 

	regionName( name, count_regions );
	myCreateTerm(&region, dlayer, facet, name );
	count_regions++;
    }
}





MDmakeMyMosfetMaster(gates, gateBoxList, diffGeo,
		     mosfetRegion, name, playername, dlayername,
		     length, width, master, view)
    fa_box *gateBoxList;
    fa_geometry *diffGeo;
    fa_region *mosfetRegion;
    char *name;
    char *playername, *dlayername;
    int length, width;
    char *master,*view;
{
    char *root, *path;
    static int number=0;
    octObject facet, ifacet, player, dlayer, prop, bag, box, fterm;
    fa_geometry geo;
    fa_box *boxList, fangBox;
        
    number++;
    (void) (void) sprintf(view, "%s", leafView);
    root = util_strsav(name);
    lower(root);
    path = defLibDir;

    (void) (void) sprintf(master, "%s/%s.%dx%d.multi%d", path, root,
		   width, length, number);

    OH_ASSERT(ohGetOrCreateBag(&masBag, &bag, master));
    if (trace) {
	(void) fprintf(stderr, "%s%screating leaf cell [%s:%s] ...\n", 
		INDENT, INDENT, master, view);
    }

    /* keep count */
    uniqueMosfetCount++;
	
    /* open master */
    OH_ASSERT(ohOpenFacet(&facet, master, view, "contents", "w"));

    /* create layers */
    OH_ASSERT(ohGetOrCreateLayer(&facet, &player, playername));
    OH_ASSERT(ohGetOrCreateLayer(&facet, &dlayer, dlayername));

    /* create active area */
    faRegionGeo(mosfetRegion, &geo);
    fa_regions_to_boxes(mosfetRegion, &boxList);
    while(faGenerateBox(&boxList, &fangBox) == 1) {
	OH_ASSERT(ohCreateBox(&player, &box, fangBox.left, fangBox.bottom,
			     fangBox.right, fangBox.top));
	OH_ASSERT(ohCreateBox(&dlayer, &box, fangBox.left, fangBox.bottom,
			     fangBox.right, fangBox.top));
    }
    
    /* create source, drain an possibly other diff terminals */
    {
	fa_geometry mosfetGeo;
	faRegionGeo( mosfetRegion, &mosfetGeo );
	MDCreateDiffRegions( diffGeo, &mosfetGeo, &dlayer, &facet );
	fa_free_geo( &mosfetGeo );
    }


    /* create gates */
    
    {
	int i;
	for ( i = 0 ; i < gates ; i++ ) {
	    OH_ASSERT(ohCreateBox(&player, &box, gateBoxList[i].left,
				  gateBoxList[i].bottom, gateBoxList[i].right,
				  gateBoxList[i].top));
	    OH_ASSERT(ohCreateTerm(&facet, &fterm, termName("GATE",i)));
	    OH_ASSERT(octAttach(&fterm, &box));
	}
    }

    /* create properties */
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "TECHNOLOGY", technology));
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "VIEWTYPE", "PHYSICAL"));
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "EDITSTYLE", "PHYSICAL"));
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "CELLCLASS", "LEAF"));

    /* close facet and generate interface */
    OH_ASSERT(gnGenInterfaceFacet(&facet,&ifacet));

    /* additional interface information */
    OH_ASSERT(ohCreatePropStr(&ifacet, &prop, "CELLTYPE", "MOSFET"));
    OH_ASSERT(ohCreatePropStr(&ifacet, &prop, "MOSFETTYPE", name));
    OH_ASSERT(ohCreatePropReal(&ifacet, &prop, "MOSFETLENGTH", (double) width));
    OH_ASSERT(ohCreatePropReal(&ifacet, &prop, "MOSFETWIDTH", (double) length));

    /* make join terms for gate terminals */
    createJoinedBag( &ifacet, "GATE" );
    createJoinedBag( &ifacet, "SOURCE" );
    createJoinedBag( &ifacet, "DRAIN" );

    {
	char name[64];
	int reg_count = 0;
	octObject term;

	while ( 1 ) {
	    sprintf( name, "REG_%d", reg_count++ );
	    if ( ohGetByTermName( &ifacet, &term, name ) != OCT_OK ) break; /* Exit while loop */
	    createJoinedBag( &ifacet, name );
	}
    }
    OH_ASSERT(octFlushFacet(&ifacet));
    OH_ASSERT(octFlushFacet(&facet));
}


int MDfindTranRegions(mosfetRegion, gateBoxList, polyGeo, diffGeo, poverlap)
    fa_region *mosfetRegion;
    fa_box    **gateBoxList;		     /* Returned */
    fa_geometry *polyGeo, *diffGeo;
    int poverlap;
{
    fa_geometry geo;
    fa_vertex *faVertex;
    dary faVertexList = daAlloc( fa_vertex, 10 );
    int vcount = 0;
    int gateTerms = 0;

    faRegionGeo(mosfetRegion, &geo);
    for(faVertex = geo.head; faVertex != NULL; faVertex = faVertex->next) {
	switch(faVertex->type) {
	case UPPER_RIGHT_NEG:
	case UPPER_LEFT_NEG:
	case LOWER_RIGHT_NEG:
	case LOWER_LEFT_NEG:
	    break;
	case LEFT_DIAGONAL:
	case RIGHT_DIAGONAL:
	    (void) fprintf(stderr,"Error in find gate stuff\n");
	case UPPER_RIGHT_POS:
	case LOWER_LEFT_POS:
	    /* - + , + - */
	    if(faCoordInGeo(polyGeo, faVertex->x-1, faVertex->y+1) == 1 ||
	       faCoordInGeo(polyGeo, faVertex->x+1, faVertex->y-1) == 1) {
		*daSet( fa_vertex, faVertexList, vcount ) =  *faVertex;
		vcount++;
	    }
	    break;
	case UPPER_LEFT_POS:
	case LOWER_RIGHT_POS:
	    /* + + , - - */
	    if(faCoordInGeo(polyGeo, faVertex->x+1, faVertex->y+1) == 1 ||
	       faCoordInGeo(polyGeo, faVertex->x-1, faVertex->y-1) == 1) {
		*daSet( fa_vertex, faVertexList, vcount ) =  *faVertex;
		vcount++;
	    }
	    break;
	}
   

    }

    gateTerms = vcount / 2;		/* there are two terms per gateTerms */
    if ( gateTerms != 2 ) {
	fprintf( stderr, "Multidigit transistor: %d gate terminals\n", gateTerms );
    }
    
    if ( gateTerms ) {
	MDmakeGateRegions( faVertexList, vcount, poverlap, polyGeo, gateBoxList);
    }

    /* diffRegions = daAlloc( fa_region, 5 ); */
    /* Normally I expect 2, 5 should take care of most weird cases */
    /* MDmakeDrainAndSourceRegions( diffGeo, faVertexList, diffRegions); */
    return gateTerms;
}


int MDmakeGateRegions( vertexList, vcount, overlap, polyGeo, gateBoxList )
    dary vertexList;
    int vcount;
    int overlap;
    fa_geometry *polyGeo;
    fa_box **gateBoxList;
{
    fa_vertex tmp, *vertex;
    fa_box *boxList, *boxGen, box;
    fa_box *barray;
    int rno, i;
    int gateTerms = vcount / 2;
    fa_box *boxArray;

    barray = ALLOC( fa_box, vcount );
    boxArray = ALLOC( fa_box, gateTerms );
    
    /* Find the boxes closest to where the gate terms need to be added */
    fa_to_boxes(polyGeo, &boxList);
    rno = 0;
    while(rno < vcount) {
	boxGen = boxList;
	vertex = daGet( fa_vertex, vertexList, rno );
	while(faGenerateBox(&boxGen, &box) == 1) {
	    if(faCoordInBox(&box, vertex->x, vertex->y) == 1) {
		barray[rno++] = box;
		break;
	    }
	}
    }
    
    rno = 0;
    for ( i = 0; i < vcount && rno < gateTerms ; i++ ) {
	int j = ( i + 1 ) % vcount;
	fa_vertex *cv = daGet( fa_vertex, vertexList, i ); /* Get current vertex */
	fa_vertex *nv = daGet( fa_vertex, vertexList, j ); /* Get next vertex  */

	if ( !faCoordInBox( &barray[i], nv->x, nv->y ) ) {
	    continue;	/* If vertices do not share the same box, skip */
	}
	if( cv->x == nv->x) {
	    boxArray[rno].left   = leftOverlap( cv, polyGeo, overlap);
	    boxArray[rno].right  = rightOverlap( cv, polyGeo, overlap);
	    boxArray[rno].bottom = MIN( cv->y, nv->y);
	    boxArray[rno].top    = MAX( cv->y, nv->y);
	} else if( cv->y == nv->y ) {
	    boxArray[rno].bottom = bottomOverlap( cv, polyGeo, overlap);
	    boxArray[rno].top    = topOverlap( cv, polyGeo, overlap);
	    boxArray[rno].right  = MAX(cv->x, nv->x);
	    boxArray[rno].left   = MIN(cv->x, nv->x);
	} else {
	    errRaise( "kahlua", 1, "unaligned boxes" );
	}
	rno++;
    }

    FREE( barray );
    *gateBoxList = boxArray;
}

MDmakeDrainAndSourceRegions( diffGeo, vertexList, diffRegions)
    fa_geometry *diffGeo;
    dary  vertexList;
    dary  diffRegions;		/* Returned. */
{
    fa_region *regionList, *regionGen, region;


    printf( "Should not be used");
    fa_to_regions(diffGeo, &regionList);
    regionGen = regionList;
    while(faGenerateRegion(&regionGen, &region) == 1) {
	int i = 0;
	for ( i = 0; i < daLen( vertexList ) ; i += 2 ) {
	    int j = i / 2;
	    fa_vertex *v = daGet( fa_vertex, vertexList, i );
	    if(faCoordInRegion(&region, v->x, v->y ) == 1) {
		fa_region *reg = daSet( fa_region, diffRegions, j );
		*reg = region;
	    }
	}
    }
}















