/*
 * mosfets.c -- Routines for extracting transistors and
 *     forming cell masters.
 */
#include "port.h"
#include "utility.h"
#include "fang.h"
#include "harpoon.h"
#include "oct.h"
#include "st.h"
#include "oh.h"
#include "fa.h"
#include "gn.h"
#include "list.h"
#include "th.h"			/* Tap helper */
#include "kh.h"
#include "kahlua.h"

extern char* termName();

kh_term* registerTerm( termList, instance, name, equiv )
    lsList termList;
    octObject*  instance;
    char* name;
    kh_term* equiv;
/*
  * Add named terminal to terminal list.
  * Returns 0 if term does not exist, 1 if all works OK.
*/
{
    octObject aterm;
    int x = 0,y = 0;		/* Possibly useless variables XXX */
    int tx,ty;
    kh_term *termEntry;
    struct octBox bb;

    if ( ohGetByTermName(instance, &aterm, name ) != OCT_OK ) return NIL(kh_term);
    OH_ASSERT(ohTerminalBB(&aterm, &bb));
    determineTerminalLocation(bb, x, y, &tx, &ty);
    
    termEntry = termConstruct( tx, ty, aterm.objectId );
    if ( equiv ) {
	termEntry->equiv_next = 0;
	equiv->equiv_next = termEntry;
	if ( equiv->equiv_head == 0 ) equiv->equiv_head = equiv;
	termEntry->equiv_head = equiv->equiv_head;
    } 
    (void) lsNewBegin(termList, (Pointer) termEntry, LS_NH);
    return termEntry;
}


int registerEquivTerms( termList, instance, name )
    lsList termList;
    octObject*  instance;
    char* name;
{
    int count = 0;
    kh_term* termEntry = 0;

    do {
	termEntry = registerTerm( termList, instance, termName( name, count++ ), termEntry);
    } while (termEntry!=0);
    return count;
}

/*
 * handleMosfets:
 *    for each mosfet layer {
 *        make active area;
 *        update the two interacting layers;
 *        if mosfet is rectangular {
 *            make a master if necessary;
 *            add to terminal table;
 *        } else {
 *            interactively build master;
 *        }
 *    }
 */
/*ARGSUSED*/
handleMosfets(inFacetPtr, outFacetPtr, geoListPtr, tsListPtr)
octObject *inFacetPtr, *outFacetPtr;
harList *geoListPtr;
khList *tsListPtr;
{
    octObject instanceBag, instance, aterm, layer, prop;
    octGenerator genLayer;
    char *layername;
    int x, y;
    int tx, ty;
    lsList termList;
    kh_term *termEntry;
    int width, length;
    char master[BUFFER_SIZE], view[BUFFER_SIZE], name[BUFFER_SIZE];
    octTransformType type;
    octObject layerBag;
    char *layer1name, *layer2name;
    int layer1overlap, layer2overlap;
    fa_geometry *polyGeo, *diffGeo, *mosfetGeo;
    fa_box *fangBoxGen, *fangBoxList, fangBox;
    fa_region *fangRegionGen, *fangRegionList, fangRegion;
    fa_region drainRegion, sourceRegion;
    fa_box *gateBoxList = 0;
    struct octBox bb;
    int tmp,count;
    char myname[BUFFER_SIZE];
    
    /* print trace message */
    if (trace) {
	(void) fprintf(stderr, "%sextracting mosfets ...\n", INDENT);
    }

    /* get instance bag */
    OH_ASSERT(ohGetByBagName(outFacetPtr, &instanceBag, "INSTANCES"));

    OH_ASSERT(ohGetByBagName(&globalFacet, &layerBag, "MOSFETS"));
    OH_ASSERT(octInitGenContents(&layerBag, OCT_LAYER_MASK, &genLayer));
    while (octGenerate(&genLayer, &layer) == OCT_OK) {

	/* 
	 * grab rules from technology
	 * facet 'globalFacet'
	 */
	OH_ASSERT(ohGetByPropName(&layer, &prop, "POLY_LAYER"));
	layer1name = prop.contents.prop.value.string;

	OH_ASSERT(ohGetByPropName(&layer, &prop, "DIFF_LAYER"));
	layer2name = prop.contents.prop.value.string;

	OH_ASSERT(ohGetByPropName(&layer, &prop, "POLY_OVERLAP_RULE"));
	layer1overlap = prop.contents.prop.value.integer;

	OH_ASSERT(ohGetByPropName(&layer, &prop, "DIFF_OVERLAP_RULE"));
	layer2overlap = prop.contents.prop.value.integer;

	/* grab poly and diff fang geos */
	polyGeo = harGetNamedGeo(layer1name, geoListPtr);
	diffGeo = harGetNamedGeo(layer2name, geoListPtr);


	/* build mosfet fang geo */
	layername = layer.contents.layer.name;
	mosfetGeo = harGetNamedGeo(layername, geoListPtr); /* This initialzes mosfetGeo */

	/* Grow diff by one lambda and then shrink it by the same.
	 * This will merge diff region under 2 lamda wide poly and
	 * allow to detect FETs created by adjacent poly/diff regions
	 * 
	 */
	/* faGrowShrink( diffGeo, 20 );		 */

	fa_and(polyGeo, diffGeo, mosfetGeo);

	/* rebuild poly and diff fang geos */
	faAndNotInPlace(polyGeo, mosfetGeo);
	faAndNotInPlace(diffGeo, mosfetGeo);

	/*
	 * for each active region, create a master if necessary,
	 * and instantiate the cell, and add terminal information
	 * to terminal list
	 */
	fa_to_regions(mosfetGeo, &fangRegionList);
	fangRegionGen = fangRegionList;
	while (faGenerateRegion(&fangRegionGen, &fangRegion) == 1) {
	    fa_regions_to_boxes(&fangRegion, &fangBoxList);

	    if (faCountBoxes(fangBoxList) != 1) {
		/*
		 * my hack to get non-rectangular to work
		 */
		int useMD = 1;

		if ( useMD ) {
		    int gates = MDfindTranRegions(&fangRegion, &gateBoxList,
				      polyGeo, diffGeo,
				      layer1overlap);
		    calcLengthAndWidth(fangBoxList, &width, &length);
		    MDmakeMyMosfetMaster(gates, gateBoxList, diffGeo,
				       &fangRegion, layername,
				       layer1name, layer2name, length,
				       width, master, view);
		
		    /* keep count */
		    mosfetCount++;
		
		    /* guess at the center, doesn't work */
		    x = y = 0;
		    type = OCT_NO_TRANSFORM;
		} else {
		    findTranRegions(&fangRegion, gateBoxList, &drainRegion,
				    &sourceRegion, polyGeo, diffGeo,
				    layer1overlap);
		    calcLengthAndWidth(fangBoxList, &width, &length);
		    makeMyMosfetMaster(gateBoxList, &drainRegion, &sourceRegion, 
				       &fangRegion, layername,
				       layer1name, layer2name, length,
				       width, master, view);
		
		    /* keep count */
		    mosfetCount++;
		
		    /* guess at the center, doesn't work */
		    x = y = 0;
		    type = OCT_NO_TRANSFORM;
		}

	    } else {
		fangBoxGen = fangBoxList;
		(void) faGenerateBox(&fangBoxGen, &fangBox);
		faBoxCenter(&fangBox, &x, &y);
		width = fangBox.right - fangBox.left;
		length = fangBox.top - fangBox.bottom;

		/* test orientation */
		if (verticalOrientation(&fangBox, polyGeo, layer1overlap)) {
		    type = OCT_NO_TRANSFORM;
		} else {
		    type = OCT_ROT90;
		    tmp = width;
		    width = length;
		    length = tmp;
		}

                /*
		 * make master if necessary,
		 * then instantiate it
		 */
		makeMosfetMaster(layername, layer1name, layer2name, 
		  layer1overlap, layer2overlap, width, length, 
		  master, view);
	    }
	    /* keep count */
	    mosfetCount++;
	    (void) sprintf(name, "m.%d", mosfetCount);
	    OH_ASSERT(ohCreateInstance(outFacetPtr, &instance, name,
				      master, view, x, y, type));
	    OH_ASSERT(octAttach(&instanceBag, &instance));
	    /*
	     * add mosfet width and length
	     *
	     * booboo: the length of the active area is really
	     * the MOSFETWIDTH, and the width of the active area
	     * is really MOSFETLENGTH.
	     */
	    OH_ASSERT(ohCreatePropReal(&instance, &prop, "MOSFETWIDTH", (double) length * thCoordSize()));
	    OH_ASSERT(ohCreatePropReal(&instance, &prop, "MOSFETLENGTH", (double) width) * thCoordSize());
	    
	    /* make layer1 terminal entries */
	    termList = khGetNamedLs(layer1name, tsListPtr);
	    

	    registerEquivTerms( termList, &instance, "GATE" );


	    /* make layer2 terminal entries */
	    termList = khGetNamedLs(layer2name, tsListPtr);
	    registerEquivTerms( termList, &instance, "SOURCE" );
	    registerEquivTerms( termList, &instance, "DRAIN" );

	    {
		int other_regions = 0;
		do {
		    char rname[64];
		    count = 0;
		    sprintf( rname, "REG_%d", other_regions++ );
		    count = registerEquivTerms( termList, &instance, rname );
		} while( count > 1 );
	    }

	}
    }

    /*
     * need to free memory
     */
    fa_free_regions(fangRegionList);
    fa_free_boxes(fangBoxList);
}


/*
 * verticalOrientation: test if mosfet is rotated or not.
 *    returns: 1 if it is not rotated.
 *             0 if it is rotated.
 */
verticalOrientation(boxPtr, geoPtr, overlap)
fa_box *boxPtr;
fa_geometry *geoPtr;
int overlap;
{
    fa_geometry geo1, geo2;
    fa_box *fBoxList;
    int l, b, r, t;
    int ret_val;

    fa_init(&geo1);
    l = boxPtr->left; 
    b = boxPtr->top; 
    r = boxPtr->right; 
    t = boxPtr->top + (overlap / 2);
    fa_add_box(&geo1, l, b, r, t);

    fa_and(&geo1, geoPtr, &geo2);
    fa_to_boxes(&geo2, &fBoxList);
    if (faCountBoxes(fBoxList) >= 1) {
	ret_val = 1;
    } else {
	ret_val = 0;
    }

    /*
     * need to free memory
     */
    fa_free_geo(&geo1);
    fa_free_geo(&geo2);
    fa_free_boxes(fBoxList);

    return ret_val;
}


/*
 * determineTerminalLocation
 *
 * find the x and y location of a mosfet terminal
 * bordering the active area
 *
 * results
 *    none
 *
 * side effects
 *    the location will be returned in tx and ty
 */
determineTerminalLocation(bb, x, y, tx, ty)
struct octBox bb;
int x, y;
int *tx, *ty;
{
    int dx, dy;

    dx = (bb.upperRight.x + bb.lowerLeft.x)/2;
    dy = (bb.upperRight.y + bb.lowerLeft.y)/2;

#ifdef OLD_CODE
    if (dx > x) {
	dx = bb.lowerLeft.x;
    } else if (dx < x) {
	dx = bb.upperRight.x;
    } else if (dy > y) {
	dy = bb.lowerLeft.y;
    } else if (dy < y) {
	dy = bb.upperRight.y;
    } else {
	OH_ASSERT(OCT_ERROR);
    }
#endif
    /* return */
    *tx = dx;
    *ty = dy;
}



#ifdef INCOMPLETE_CODE    
int mosfetCheckEdges( polyGeo, diffGeo, mosfetGeo )
    fa_geometry* polyGeo;
    fa_geometry* diffGeo;
    fa_geometry* mosfetGeo;
{
    fa_edge *list[2];
    fa_edge *edge;

    fa_to_edges( polyGeo, &list[0], &list[1] );

    for ( i = 0; i < 2 ; i++ ) {
	for ( edge = list[i]; edge != 0 ; edge = edge->next ) {
	    printf( "%d %d" );
	}
    }
}
#endif    
