/*
 * nets.c -- Routines for deriving the symbolic wires
 */
#include "port.h"
#include "utility.h"
#include "fang.h"
#include "harpoon.h"
#include "oct.h"
#include "st.h"
#include "oh.h"
#include "fa.h"
#include "gn.h"
#include "list.h"
#include "kh.h"
#include "kahlua.h"
#include "symlib.h"

#define NO_MAXIMUM -1
int scale;

octStatus actualTerminalBoxOnLayer();

void myAddTermToNet( net, term )
    octObject *term;
    octObject *net;
    /*
     * Takes care of JOINED terms too.
     */
{
    if ( term->type == OCT_TERM ) {
	octAttachOnce( net, term );
	if ( ! symConnectorp( term ) ) {
	    octObject inst, bag, fterm, aterm;
	    octGenerator genTerm;
	
	    OH_ASSERT(ohFindFormal( &fterm, term ));
	    
	    OH_ASSERT(ohGetById( &inst, term->contents.term.instanceId ));	/* Get instance */
	    bag.type = OCT_BAG;				/*  			*/
	    bag.contents.bag.name = "JOINED";		/*  */
	    
	    if ( octGetContainerByName( &fterm, &bag ) == OCT_OK ) {
		if ( trace ) fprintf( stderr, "Found JOINED bag\n" );
		OH_ASSERT(octInitGenContents( &bag, OCT_TERM_MASK, &genTerm ));
		while ( octGenerate( &genTerm, &aterm ) == OCT_OK ) {
		    OH_ASSERT(octGetByName( &inst, &aterm ));
		    OH_ASSERT(octAttachOnce( net, &aterm ));
		}
	    }
	}

    }
}

void RecursiveHandleRegion( region, hole_p, netLs, minimum, octLayer, outFacetPtr )
    fa_region* region;
    int hole_p;			/* Is the region a hole? */
    lsList netLs;
    int   minimum;
    octObject* octLayer;
    octObject* outFacetPtr;
{
    char* layername = octLayer->contents.layer.name;
    octObject connectorBag;
    lsList termList, pathList;

    int routed;

    if ( hole_p ) {
	if ( region->interior_regions ) {
	    fa_region *reg = region->interior_regions;

	    fprintf( stderr, "there are regions inside the hole\n" );

	    region->interior_regions = 0;
	    RecursiveHandleRegion( reg, 0, netLs, minimum, octLayer, outFacetPtr );
	    return;
	}
    } else {
	if ( region->interior_regions ) {
	    /* Deal with the hole */
	    fa_region* hole = region->interior_regions;

	    fprintf( stderr, "there are holes inside the region\n" );


	    RecursiveHandleRegion( hole, 1, netLs, minimum, octLayer, outFacetPtr );
	    
	}	
    }

    termList = lsCreate();
    pathList = lsCreate();

    /* 
     * build termList any terminal entries
     * contained in this region
     */
    buildLocalTermList( region, netLs, termList);

    /* ignore if only one pin on net */
    if (lsLength(termList) > 1) {
	/*
	 * now route the region, ignore if unsuccessful
	 */
	routed = executeRouter(region, termList, pathList, layername);

	/* keep counter */
	routeRegionCount++;

	/* implement wires if routing successful */
	OH_ASSERT( ohGetOrCreateBag( outFacetPtr, &connectorBag, "CONNECTORS" ));
	if (routed) {
	    implementPathList( region, termList, pathList, 
			      layername, minimum, octLayer, &connectorBag, 
			      outFacetPtr);
	}
    }
    lsDestroy(pathList, free);
    lsDestroy(termList, 0);


}




/*
 * handleNets:
 *    for each net layer {
 *        break them into routing regions;
 *        find intersecting terminals;
 *        try to route region;
 *        if route is successful {
 *            determine widths of wires;
 *        } else {
 *            interactively route region;
 *        }
 *
 *        determine connectors necessary;
 *        generate wires;
 *    }
 *
 *    merge nets;
 *    detach connectors from nets;
 */
/*ARGSUSED*/
handleNets(inFacetPtr, outFacetPtr, geoListPtr, tsListPtr)
octObject *inFacetPtr, *outFacetPtr;
harList *geoListPtr;
khList *tsListPtr;
{
    octObject layer, prop;
    octObject layerBag;
    octObject octLayer, connectorBag;
    octGenerator genLayer;
    lsList termList, pathList, netLs;
    fa_geometry *netGeo;
    fa_region *fangRegionGen, *fangRegionList, fangRegion;
    char *layername;
    int minimum;
    int routed;
    int count;

    /* print trace message */
    if (trace) {
	(void) fprintf(stderr, "%sdecomposing nets ...\n", INDENT);
    }

    /* grab connector bag */
    OH_ASSERT(ohGetByBagName(outFacetPtr, &connectorBag, "CONNECTORS"));

    /* grab scale */
    OH_ASSERT(ohGetByPropName(&globalFacet, &prop, "FACTOR"));
    scale = prop.contents.prop.value.integer;

    /* for each net layer */
    OH_ASSERT(ohGetByBagName(&globalFacet, &layerBag, "NETS"));
    OH_ASSERT(octInitGenContents(&layerBag, OCT_LAYER_MASK, &genLayer));
    while (octGenerate(&genLayer, &layer) == OCT_OK) {
	/* create the layer in outFacet */
	layername = layer.contents.layer.name;
	OH_ASSERT(ohGetOrCreateLayer(outFacetPtr, &octLayer, layername));

	/* print trace message */
	if (trace) {
	    (void) fprintf(stderr, "%s%sprocessing interconnect layer [%s] ...\n", 
	      INDENT, INDENT, layername);
	}

	/*
	 * grab rule from technology
	 * facet 'globalFacet'
	 */
	OH_ASSERT(ohGetByPropName(&layer, &prop, "MINIMUM_WIDTH"));
	minimum = prop.contents.prop.value.integer;

	/* grab net fang geo */
	netGeo = harGetNamedGeo(layername, geoListPtr);
	netLs = khGetNamedLs(layername, tsListPtr);

	/*
	 * for each net region, build a list of covering terminals,
	 * then call the maze router, then determine the widths of
	 * the wires, then make the implementation with necessary
	 * connectors
	 */
	fa_to_regions(netGeo, &fangRegionList);
	fangRegionGen = fangRegionList;
	while (faGenerateRegion(&fangRegionGen, &fangRegion) == 1) {
	    RecursiveHandleRegion( &fangRegion, 0, netLs, minimum, &octLayer, outFacetPtr );
#ifdef OLD_CODE
  {
		termList = lsCreate();
		pathList = lsCreate();

		/* 
		 * build termList any terminal entries
		 * contained in this region
		 */
		buildLocalTermList(&fangRegion, netLs, termList);

		/* ignore if only one pin on net */
		if (lsLength(termList) > 1) {
		    /*
		     * now route the region, ignore if unsuccessful
		     */
		    routed = executeRouter(&fangRegion, termList, pathList, layername);

		    /* keep counter */
		    routeRegionCount++;

		    /* implement wires if routing successful */
		    if (routed) {
			implementPathList(&fangRegion, termList, pathList, 
					  layername, minimum, &octLayer, &connectorBag, 
					  outFacetPtr);
		    }
		}
	    }
#endif
	}
    }

    /* merge nets and update number of nets */
    OH_ASSERT(gnMergeNetsInFacet(outFacetPtr));
    OH_ASSERT(gnDetachNetConnectors(outFacetPtr));
    count = ohCountContents(outFacetPtr, OCT_NET_MASK);
    majorNetCount += count;

    /* 
     * check to make sure that each net 
     * contains at least two terminals
     */
    checkNets(outFacetPtr);

    /*
     * need to free memory
     */
    fa_free_regions(fangRegionList);
}


/* 
 * buildLocalTermList: build termList any terminal entries
 * contained in this region
 */
buildLocalTermList(fangRegionPtr, netLs, termList)
fa_region *fangRegionPtr;
lsList netLs, termList;
{
    lsGen netLsGen;
    kh_term *netTermEntry, *termEntry;
    int x, y;
    int count = 0;
    
    netLsGen = lsStart(netLs);
    while (lsNext(netLsGen, (Pointer *)&netTermEntry, LS_NH) == LS_OK) {
	netTermEntry->gridPosition.x = -1; /* Reset to illegal value. */
	netTermEntry->gridPosition.y = -1;
	
	x = netTermEntry->position.x;
	y = netTermEntry->position.y;

	if (faCoordInRegion(fangRegionPtr, x, y) == 1) {
	    count++;
	    (void) lsNewBegin(termList, (Pointer) netTermEntry, LS_NH);
	}
    }

    fprintf( stderr, "%d terminals in this region\n", count );
    /* free generator */
    lsFinish(netLsGen);
}


/*
 * executeRouter 
 *
 * calls a virtual grid maze routing routine to route the specified
 * problem
 *
 * returns 
 *    1 - successful 
 *    0 - otherwise
 *
 * side effects
 *    pathList contains a list of paths implementing
 *    the routed solution
 */
/*ARGSUSED*/
int
executeRouter(fangRegionPtr, termList, pathList, layername)
fa_region *fangRegionPtr;
lsList termList, pathList;
char *layername;
{
    int routed;

    /*
     * now route the region, ignore if unsuccessful
     */
    if (mazeroute(fangRegionPtr, termList, pathList) == 0) {
	/* print error message */
	if (trace) {
	    (void) fprintf(stderr, 
	      "%s%s%sunable to route a region ...\n",
	      INDENT, INDENT, INDENT);
	}
	routed = 0;
    } else {
	/* print trace message */
	if (trace) {
	    (void) fprintf(stderr, 
	      "%s%s%sautomatically routed a region ...\n",
	      INDENT, INDENT, INDENT);
	}
	routed = 1;
    }

    /* return code */
    return routed;
}


/*
 * implementPathList:  Takes a pathList, and implements
 *    the wires onto the container 'octLayerPtr'.  Creates
 *    the necessary connectors.  For now, this just
 *    implements minimum width.
 */
implementPathList(fangRegionPtr, termList, pathList, layername, minimum, 
  octLayerPtr, connectorBagPtr, outFacetPtr)
fa_region *fangRegionPtr;
lsList termList, pathList;
char *layername;
int minimum;
octObject *octLayerPtr;
octObject *connectorBagPtr;
octObject *outFacetPtr;
{
    int maximum;
    int p1x, p1y, p2x, p2y;
    int width, length;
    int count;			/* Used to avoid infinit loops. */
    char master[BUFFER_SIZE], view[BUFFER_SIZE], name[BUFFER_SIZE];
    lsGen pathLsGen;
    kh_path *pathEntry;
    octTransformType type;
    octObject oldinstance, instance, aterm1, aterm2, path, net;
    octObject tmpBag;
    octGenerator genInst;
    octObject termBag1, termBag2, obj;
    octGenerator gen;
    
    /* 
     * make a minimum size connector for temporary use,
     * and make a temporary bag on 'octLayerPtr' for
     * storing temporary connectors that will be later
     * deleted
     */
    makeConnectorMaster(layername, minimum, minimum, &type, master, view);
    OH_ASSERT(ohCreateBag(octLayerPtr, &tmpBag, "CONNECTORS"));

    {
	lsGen termLsGen = lsStart(termList);
	kh_term* termEntry;
	while( lsNext(termLsGen,(Pointer*)&termEntry, LS_NH) == LS_OK ) {
	    if ( termEntry->overlap ) {
		octObject t1,t2,path,net;
		octPoint pt1,pt2;
		ohGetById(&t1,termEntry->termId );
		ohGetById(&t2,termEntry->overlap->termId );
	    
		fprintf( stderr, "Routing overlapping terminals %s and %s\n",
			ohFormatName(&t1), ohFormatName(&t2) );
		symCreateSegment( octLayerPtr, &pt1, &t1, SYM_USE_EXISTING,
				 &pt2, &t2, SYM_USE_EXISTING, -1, SYM_VERTICAL );
		
		/*  because may replaces the end connector */

		termEntry->termId = t1.objectId;
		termEntry->overlap->termId = t2.objectId;
	    
	    }
	}
}
    /*
     * implement the paths
     */
    pathLsGen = lsStart(pathList);
    while (lsNext(pathLsGen, (Pointer *) &pathEntry, LS_NH) == LS_OK) {
	/* create terminal bags */
	OH_ASSERT(ohCreateBag(outFacetPtr, &termBag1, NIL(char)));
	OH_ASSERT(ohCreateBag(outFacetPtr, &termBag2, NIL(char)));

	/* get path coordinates */
	p1x = pathEntry->start.x; p1y = pathEntry->start.y;
	p2x = pathEntry->end.x; p2y = pathEntry->end.y;

	/* get path terminals */
	(void) getOrCreatePathTerminal(&termBag1, termList, p1x, p1y, 
	  outFacetPtr, &tmpBag, connectorBagPtr, master, view, type);
	(void) getOrCreatePathTerminal(&termBag2, termList, p2x, p2y, 
	  outFacetPtr, &tmpBag, connectorBagPtr, master, view, type);

	/* need to determine max path width */
	OH_ASSERT(octGenFirstContent(&termBag1, OCT_TERM_MASK, &aterm1));
	OH_ASSERT(octGenFirstContent(&termBag2, OCT_TERM_MASK, &aterm2));
	maximum = maxPathWidth(pathEntry, &aterm1, &aterm2, layername, &tmpBag);

	/* determine path width by examining region boundaries */
	width = determinePathWidth(fangRegionPtr, pathEntry, minimum, maximum);

	/* round off */
	width = width / scale;
	width = width * scale;

	/* create the path geometry */
	OH_ASSERT(gnCreatePath(octLayerPtr, &path, p1x, p1y, p2x, p2y, width));

	/* update counter */
	pathCount ++;

	/* create a net and attach path to it */
	mkNetName(name);
	OH_ASSERT(ohCreateNet(outFacetPtr, &net, name));
	/* fprintf( stderr, "Created %s\n", ohFormatName( &net ) ); */

	/* 
	 * connect the terminals, net, and path
	 * appropriately
	 */
	
	OH_ASSERT(octAttachOnce(&net, &path));

	OH_ASSERT(octInitGenContents(&termBag1, OCT_ALL_MASK, &gen));
	while(octGenerate(&gen, &obj) == OCT_OK) {
	    /* fprintf( stderr, "Attaching once %s to %s\n", ohFormatName(&obj), ohFormatName(&net));  */

	    myAddTermToNet( &net, &obj );

	    OH_ASSERT(octAttachOnce(&path, &obj));
	}
	OH_ASSERT(octDelete(&termBag1));
	
	OH_ASSERT(octInitGenContents(&termBag2, OCT_ALL_MASK, &gen));
	while(octGenerate(&gen, &obj) == OCT_OK) {
	    /* fprintf( stderr, "Attaching once %s to %s\n", ohFormatName(&obj), ohFormatName(&net)); */
	    myAddTermToNet( &net, &obj );
	    OH_ASSERT(octAttachOnce(&path, &obj));
	}
	OH_ASSERT(octDelete(&termBag2));
    }

    /*
     * now look into 'tmpBag' for all the
     * temporary connectors.  for each
     * connector, determine what size it
     * should be by examining the paths
     * that are connected to it.  then
     * replace that connector with a 
     * new connector
     */
    count = ohCountContents( &tmpBag, OCT_INSTANCE_MASK );
    OH_ASSERT(octInitGenContents(&tmpBag, OCT_INSTANCE_MASK, &genInst));
    while ( count-- && octGenerate(&genInst, &instance) == OCT_OK) {
        /*
	 * determine what size the connector instance
	 * should be
	 */
	determineConnectorSize(&instance, &width, &length, minimum);

	/*
	 * make a new connector, set instance transformation, and
	 * replace old instance with new master
	 */
	makeConnectorMaster(layername, width, length, &type, master, view);
	instance.contents.instance.transform.transformType = type;
	/* keep count */
	connectorCount++;
	oldinstance = instance;
	instance.contents.instance.master = master;
	instance.contents.instance.view = view;
	ohReplaceInstance(&instance, &oldinstance);
    }
    octFreeGenerator( &genInst );

    /* now delete the temporary bag */
    OH_ASSERT(octDelete(&tmpBag));

    /* free generator */
    lsFinish(pathLsGen);
}


/*
 * getOrCreatePathTerminal: gets the oct terminal that is supposed
 *    to connect to the x-y location of a path.   if a connecting
 *    terminal does not exist, then a connector terminal will be 
 *    created
 *
 * returns 0: if not found, and one is created
 *         1: otherwise
 */
getOrCreatePathTerminal(aBagPtr, termList, x, y, outFacetPtr, 
			tmpBagPtr, connectorBagPtr, master, view, type)
    octObject *aBagPtr;
    lsList termList;
    int x, y;
    octObject *outFacetPtr, *tmpBagPtr, *connectorBagPtr;
    char *master, *view;
    octTransformType type;
{
    kh_term *termEntry;
    octObject instance;
    char name[BUFFER_SIZE];
    octId termId;
    octObject aterm;
    lsGen termLsGen;
    int in_while = 0;
    
    /* 
     * find the terminal(s) and attach it 
     * to the net
     */
    termLsGen = lsStart(termList);
    while(lsGetTermByCoord(&termLsGen, &termEntry, x, y)) {
	in_while = 1;
	termId = termEntry->termId;
	OH_ASSERT(ohGetById(&aterm, termId));
	OH_ASSERT(octAttachOnce(aBagPtr, &aterm));

	/* Joined terms ? */

    }
    
    if(!in_while) {
	/* 
	 * instantiate a connector if a leaf cell
	 * does not exist at this steiner node
	 */
	mkInstanceName(name);
	OH_ASSERT(ohCreateInstance(outFacetPtr, &instance, name, master, view, x, y, type));
	OH_ASSERT(octAttach(connectorBagPtr, &instance));
	/* attach to temporary bag for later deletion */
	OH_ASSERT(octAttach(tmpBagPtr, &instance));
	OH_ASSERT(octGenFirstContent(&instance, OCT_TERM_MASK, &aterm));
	/* fprintf( stderr, "Attaching %s to %s\n", ohFormatName(&aterm), ohFormatName(aBagPtr)); */
	OH_ASSERT(octAttachOnce(aBagPtr, &aterm));

	/*
	 * add this new connector terminal to
	 * the terminal list
	 */
	
	termEntry = termConstruct( x, y, aterm.objectId );

	(void) lsNewBegin(termList, (Pointer) termEntry, LS_NH);

	return 0;
    }
    return 1;
}


/*
 * determinePathWidth: Examines the boundaries of
 *    'fangRegionPtr' to determine how wide the
 *    path can be.
 *
 * returns the width of the path
 */
int determinePathWidth(fangRegionPtr, pathEntry, minimum, maximum)
fa_region *fangRegionPtr;
kh_path *pathEntry;
int minimum;
int maximum;
{
    fa_edge *h_list, *v_list; 
    fa_edge *list;
    int direction;
    int p1x, p1y, p2x, p2y;
    int p1, p2, pc, start, end, center;
    int width, length, delta;

    /* error check */
    if ((maximum < minimum) && (maximum != NO_MAXIMUM)) {
	(void) fprintf(stderr, 
	  "Panic: maximum path width %d cannot be less than minimum path width %d\n", maximum, minimum);
	/*
	 * just return the minimum of 'maximum' and 'minimum'
	 * instead of aborting
	 */
	return MIN(maximum, minimum);
    }

    /* return minimum width if option specifies */
    if (minWidth) {
	return minimum;
    }

    faRegionEdges(fangRegionPtr, &h_list, &v_list);
    p1x = pathEntry->start.x; p1y = pathEntry->start.y;
    p2x = pathEntry->end.x; p2y = pathEntry->end.y;
    direction = faPathDirection(p1x, p1y, p2x, p2y);

    /* error check */
    if ((direction == 2) || (direction == -1)) {
	(void) fprintf(stderr, "Panic: error in determining path direction\n");
	OH_ASSERT(OCT_ERROR);
    }

    /* check direction */
    if (direction == 0) {
	/* if path is horizontal */
	list = h_list;
	p1 = p1x;
	p2 = p2x;
	pc = p1y;
    } else {
	/* else path is vertical */
	list = v_list;
	p1 = p1y;
	p2 = p2y;
	pc = p1x;
    }

    /* now check against boundaries */
    length = abs(p2 - p1);
    width = -1;
    while (list != NIL(fa_edge)) {
	start = list->start;
	end = list->end;
	center = list->center;
        if (faIsBetween(p1, start, end) || faIsBetween(p2, start, end) ||
	  faIsBetween(start, p1, p2) || faIsBetween(end, p1, p2)) {
	    delta = 2 * abs(pc - center);
	    if (width < 0) {
		width = delta;
	    }
	    width = MIN(delta, width);
	}
	list = list->next;
    }

    /*
     * need to free memory
     */
    fa_free_edges(h_list);
    fa_free_edges(v_list);

    /* test external constraints */
    width = MAX(width, minimum);
    if (maximum != NO_MAXIMUM) {
	width = MIN(width, maximum);
    }

    /* check aspect ratio */
    if ((width/length) > ratio) {
	return minimum;
    }

    /* return */
    return width;
}


/*
 * determineConnectorSize: Returns in 'wp' and 'lp'
 *    the width and length of the connector, respectively.
 *    the size is determine by examining the paths
 *    that connects this instance terminal.  the connector
 *    has to be big enough to accommodate all the intersecting
 *    paths.
 */
determineConnectorSize(instancePtr, wp, lp, minimum)
octObject *instancePtr;
int *wp, *lp;
int minimum;
{
    octObject aterm, path;
    octGenerator genPath;
    int direction;
    int size;

    /* initialize sizes */
    *wp = minimum;
    *lp = minimum;

    OH_ASSERT(octGenFirstContent(instancePtr, OCT_TERM_MASK, &aterm));
    OH_ASSERT(octInitGenContainers(&aterm, OCT_PATH_MASK, &genPath));
    while (octGenerate(&genPath, &path) == OCT_OK) {
	/* determine path direction */
	direction = gnPathDirection(&path);

	/* error check */
	if ((direction == 2) || (direction == -1)) {
	    (void) fprintf(stderr, "Panic: error in determining path direction\n");
	    OH_ASSERT(OCT_ERROR);
	}

	/* update sizes */
	size = path.contents.path.width;
	if (direction == 0) {
	    /* if path is horizontal */
	    *lp = MAX(*lp, size);
	} else {
	    /* else path is vertical */
	    *wp = MAX(*wp, size);
	}
    }
}


/*
 * maxPathWidth: returns the maximum pathEntry width constrained 
 *    by the terminal frames that it connects.  returns NO_MAXIMUM
 *    if there are no constraints.
 */
maxPathWidth(pathEntry, atermPtr1, atermPtr2, layername, connBagPtr)
kh_path *pathEntry;
octObject *atermPtr1, *atermPtr2;
char *layername;
octObject *connBagPtr;
{
    struct octBox bb1, bb2;
    int direction;
    int size1, size2;
    int p1x, p1y, p2x, p2y;
    octStatus s1, s2;
    octObject instance;

    /* test if terminal is a connector */
    OH_ASSERT(ohGetById(&instance, atermPtr1->contents.term.instanceId));
    s1 = octIsAttached(connBagPtr, &instance);
    OH_ASSERT(ohGetById(&instance, atermPtr2->contents.term.instanceId));
    s2 = octIsAttached(connBagPtr, &instance);

    /* if no constraints */
    if ((s1 == OCT_OK) && (s2 == OCT_OK)) {
	return NO_MAXIMUM;
    } 

    p1x = pathEntry->start.x; p1y = pathEntry->start.y;
    p2x = pathEntry->end.x; p2y = pathEntry->end.y;
    direction = faPathDirection(p1x, p1y, p2x, p2y);

    OH_ASSERT(actualTerminalBoxOnLayer(atermPtr1, layername, &bb1));
    OH_ASSERT(actualTerminalBoxOnLayer(atermPtr2, layername, &bb2));

    /* error check */
    if ((direction == -1) || (direction == 2)) {
	(void) fprintf(stderr, "Panic: error in determining path direction\n");
	OH_ASSERT(OCT_ERROR);
    }

    /* check direction */
    if (direction == 0) {
	/* if path is horizontal */
	size1 = bb1.upperRight.y - bb1.lowerLeft.y;
	size2 = bb2.upperRight.y - bb2.lowerLeft.y;
    } else {
	/* else path is vertical */
	size1 = bb1.upperRight.x - bb1.lowerLeft.x;
	size2 = bb2.upperRight.x - bb2.lowerLeft.x;
    }

    /* returns */
    if (s1 == OCT_OK) {
	return size2;
    } else if (s2 == OCT_OK) {
	return size1;
    } else {
	return MIN(size1, size2);
    }
}


/*
 *  actualTerminalBoxOnLayer: Find the bounding box of an actual terminal 
 *      on a layer
 */
octStatus actualTerminalBoxOnLayer(aterm, layername, bb)
octObject *aterm;
char *layername;
struct octBox *bb;
{
    octObject layer, instance, master, fterm, box;
    octGenerator gen;
    octStatus status;
    struct octTransform *transform;

    /* Get the instance from the terminal */
    OH_ASSERT(  ohGetById(&instance, aterm->contents.term.instanceId));

    /* open the interface facet of the master of the instance */
    OH_ASSERT(  ohOpenMaster(&master, &instance, "interface", "r"));

    /* Find terminal layer */
    OH_ASSERT(  ohGetByLayerName(&master, &layer, layername));

    /* Find formal terminal (which matches this actual terminal) */
    OH_ASSERT(  ohGetByTermName(&master, &fterm, aterm->contents.term.name));

    /* Get the first 'terminal frame' on specified layer */
    OH_ASSERT(  octInitGenContents(&fterm, OCT_BOX_MASK, &gen));
    while (octGenerate(&gen, &box) == OCT_OK) {
	if (octIsAttached(&layer, &box) == OCT_OK) {
	    transform = &instance.contents.instance.transform;
	    octTransformGeo(&box, transform);
	    *bb = box.contents.box;
	    octFreeGenerator( &gen );
	    return OCT_OK;
	}
    }

    return OCT_ERROR;
}


/*
 * lsGetTermByCoord: Gets the terminal entry in
 *   termList with coordinates x and y.
 *
 * returns: 1 if found, else 0
 */
int lsGetTermByCoord(termLsGen, termEntryPtr, x, y)
    lsGen *termLsGen;
    kh_term **termEntryPtr;
    int x, y;
{
    int px, py;

    while (lsNext(*termLsGen, (Pointer *) termEntryPtr, LS_NH) == LS_OK) {
	px = (**termEntryPtr).position.x;
	py = (**termEntryPtr).position.y;
	if ((px == x) && (py == y)) {
	    return 1;
	}
    }

    /* free generator */
    lsFinish(*termLsGen);
    return 0;
}


/*
 * checkNets
 *
 * checks to make sure that each net in the facet
 * contains at least two terminals
 *
 * results
 *    none
 *
 * side effects
 *    none
 */
checkNets(outFacetPtr)
octObject *outFacetPtr;
{
    int c = 0;
    octGenerator gen;
    octGenerator genTerm;
    octObject term;
    octObject net;

    OH_ASSERT(octInitGenContents(outFacetPtr, OCT_NET_MASK, &gen));
    while (octGenerate(&gen, &net) == OCT_OK) {
	c = 0;
	OH_ASSERT(octInitGenContents(&net,OCT_TERM_MASK,&genTerm));
	while ( octGenerate(&genTerm,&term ) == OCT_OK ) {
	    if ( symConnectorp( &term ) ) {
		OH_ASSERT(octDetach( &net, &term ));
	    } else {
		c++;
	    }
	}
	if (c < 2) {
	    (void) fprintf(stderr, "illegal: net (%s) has less than 2 terminals\n",
		net.contents.net.name);
	}
	
    }
}
