
/* initGlb.h - data structure for Nishizaki's part of Mercury global router
 
   Copyright (C) 1988 by Yoshihito Nishizaki.
		All rights reserved.

 */


#include "port.h"
#include "oct.h"
#include "oh.h"
#include "mercury.h"
#include "graph.h"
#include "intProg.h"
#include "fiboHeap.h"

#define RUSAGE_SELF	0
#define RUSAGE_CHILDREN	-1

#define G_UNITPITCH	160
#define OCT_PER_MICRON	20

#define PR_MODE1	1	/* placement refinement mode 1 */
#define PR_MODE2	2	/* placement refinement mode 2 */
#define NETGROUP_THRESHOLD	750	
				/* if #net >= this value, form net groups */


/* File definition */
/*-----------------*/
FILE *graphFile;
FILE *netFile;
FILE *tileFile;
FILE *lindoFile;
FILE *resultFile;

char buf[G_BSIZE];


/**********/
/* macros */
/**********/

/* Macro for the definition of the NIL pointer */
/*---------------------------------------------*/
#define	 NIL(type)	    ((type *) 0 )

/* Macro functions that return the maximun or minimum independent of type. */
/*-------------------------------------------------------------------------*/
#define  MAX(a,b)           ((a) > (b) ? (a) : (b))
#define  MIN(a,b)           ((a) < (b) ? (a) : (b))

/* Macro function that returns the absolute value of a floating point number. */
/*----------------------------------------------------------------------------*/
#define  ABS(a)             ((a) < 0.0 ? -(a) : (a))

/* Macro procedure that swaps two entities. */
/*------------------------------------------*/
#define  SWAP(type, a, b)   {type swapx; swapx = a; a = b; b = swapx;}


/*-------------------------------------------*/
/* Memory allocation and dealocation macros. */
/*-------------------------------------------*/
char *p_shmalloc();

#define MALLOC(type,number)  (type *)malloc((unsigned)(sizeof(type)*(number)))
#define CALLOC(type,number)  (type *)calloc(sizeof(type),(unsigned)(number))
#define REALLOC(ptr,type,number)  \
           ptr = (type *)realloc((char *)ptr,(unsigned)(sizeof(type)*(number)))
#define FREE(pointer)  if ((pointer) != NULL)  \
	   { free((char *)pointer); (pointer)=NULL; }

/* Debugging aid stuff */
/*---------------------*/
/*
char *bufP;
#define MALLOC(type,number)  (type *)(bufP = malloc((unsigned)(sizeof(type)*(number)))); if ((type *)bufP == (type *)0x1416cc) printf("Mallocing pointer %x in file %s on line %d\n", (type *)bufP, __FILE__, __LINE__)

#define FREE(ptr) {if ((char *)ptr == (char *)0x1416cc) printf("Freeing pointer %x in file %s on line %d\n", ptr, __FILE__, __LINE__); free((char *)ptr);}
 */


/* Macro to copy a string: it returns the pointer to the string */
/*--------------------------------------------------------------*/
#define SAVE(ptostring)  strcpy(CALLOC(char,strlen(ptostring)+1),ptostring)

/* Macro for random number generation */
/*------------------------------------*/
#define RAND(a) ( a = ( a * 1103515245 + 12345 ) & MAX_INT ))
#define NORM(n) ( (double) n / (double) MAX_INT )
#define NRAND(a) ( (double) ( a = ( a * 1103515245 + 12345 ) & MAX_INT ) / (double) MAX_INT )

#define EXP(d,t) ( (t) == 0. ? 0. : exp( -(double) (d) / (t) ) )
#define LOG_B(base,x) ( log(x)/log(base) )

#define RINT(a) ( (int) floor( (double) a + 0.5 ) )



struct parm {

    /* Cell information */
    char	*cell;			/* cell name */
    char	*view;			/* view name */
    char	*outcell1;		/* output cell name 1 */
    char	*outview1;		/* output view name 1 */
    char	*outcell2;		/* output cell name 2 */
    char	*outview2;		/* output view name 2 */

    /* Overall framework */
    int		form_netgroup;		/* if ON, form net groups */
    int		zero_compact;		/* if ON, compact channel to 0 width */
    int		placeref_only;		/* if ON, modify cell location only
						(do not create subnets) */
    int		criticality_threshold;	/* if > this value, critical net */

    /* Tile creation */
    double	chipExpand;		/* expansion factor used in tile 
					   partitioner to get a slicing 
					   structure */
    int		Zflag;			/* if ON, use Z shaped channel */

    /* Candidate selection */
    int		M;			/* # of candidate */
    int		m;			/* # of paths per pin connection */
    int		netSizeLimit;		/* use 1 candidate if pin/net >= this */
    int		FH_dijkstra;		/* if ON, use Fibonacci heap version of
					   Dijkstra algorithm */
    /* Area estimation model and IP formulation */
    int		placeRefMode;		/* mode for placement refinement */
    float	netlength_weight;	/* weight of net length in obj func */
    float	blockloc_weight;	/* weight of block location in obj */
    int		modelDiagonal;		/* if ON, model diagonal effect at +
					   junction */
    int		tile_threshold;		/* min size of bottleneck */
    float	alpha;			/* ratio of unit usage between
					   pass-thru net and terminating net */
    float	expandFactor;		/* expand the initial dimension of chip
					   by this factor */
    int		INTEGER;		/* if ON, use INTEGER declaration */
    int		BIP;			/* if ON, use BIP declaration */
};

struct statis {

    /* Nets and net groups */
    int		num_net;		/* # nets */
    int		num_crit_net;		/* # critical nets */
    int		num_trivial_net;	/* # trivial nets */
    int		num_netgroup;		/* # net groups */
    int		num_nontriv_netgroup;	/* # non-trivial net groups */
    float	ave_net_per_netgroup;	/* average # nets per net group */

    /* Blocks */
    int		num_block;		/* # blocks excluding pads */

    /* Tile planes */
    int		num_hor_bot;		/* # horizontal bottleneck tiles */
    int		num_ver_bot;		/* # vertical bottleneck tiles */
    int		num_overcell;		/* # over-the-cell routing region */

    /* Global routing graph */
    int		num_GRgraph_node;	/* # nodes */
    int		num_GRgraph_edge;	/* # edges */

    /* Polar graphs */
    int		num_polarpath[G_NHV];	/* # polar paths */

    /* Dimension and area */
    /* Critical dimensions are determined by the usage of shortest routes */
    int		critical_dim[G_NHV];	/* length of longest polar paths (u) */
    int		init_area;		/* initial area (u^2) = 
					   critical_dim[H] x critical_dim[V] */
    int		init_netlength;		/* initial total net length (u) */

		/* The following two are in oct units */
    int		LP_init_dim[G_NHV];	/* initial chip dimension for LP */
    int		LP_improv_dim[G_NHV];	/* improvement in chip dimension */

    int		final_dim[G_NHV];	/* final chip dimension (u) */
    int		final_area;		/* final area (u^2) */
    float	dim_ratio[G_NHV];	/* final dim / initial dim */
    float	area_ratio;		/* final area / initial area */

    /* Size of LP */
    int		num_variable;		/* # variables */
    int		num_constraint;		/* # constraints */

    /* LP result */
    int		num_net_cand_i[G_MAXROUTE];
					/* # nets taking candidate route i */
    float	percent_net_cand_i[G_MAXROUTE];
					/* % nets taking candidate route i */

};

struct parm     parm;
struct statis   statis;

GRegionDef	*getNeighbor();



