
/*
#ifndef lint
static char rcsid[] = "$Header: /octtools/release-3.0/src/mercury/lindo.c,v 1.1 90/02/07 17:02:43 octtools Exp $";
#endif
*/

/*
 *  lindo.c
 *  *******
 *
 *  Copyright (C) 1988  Yoshihito Nishizaki
 *	University of California, Berkeley
 *
 *  Author:			Advising professor:
 *	Yoshihito Nishizaki	    Alberto Sangiovanni-Vincentelli
 *
 *  --------------------------------------------------------------------------
 *
 *  This file contains routines for formulating a global router
 *	problem by interger programming
 *
 *  functions contained in this file:
 *	formulateLindo()
 *	formulateMode1()
 *	formulateMode2()
 *
 *
 *  Import descriptions:
 *	initGlb.h	: macros for mercury global router
 *
 */

#include <stdio.h>
#include <strings.h>
#include <math.h>
#include "initGlb.h"



/*
 *  formulateLindo()
 *
 *  DESCRIPTION
 *  -----------
 *	Formulate an integer programming in LINDO format
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	initSize<input>	: initial dimension of chip before IP optimization
 *
 *  Local variables:
 *  ---------------
 *	capacity	: capacity remaining after fixed usage of critical nets
 *	numTerm		: counter for the number of terms per line for lindo
 *	termsPerLine	: maximum number of terms per line for lindo
 *	netlength_fac	: factor for net length in objective function
 *	blockloc_fac	: factor for block location in objective function
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  NOTE
 *  ----
 *	Improvement in dimension may be negative and thus represented as
 *		mw = mw1 - mw2
 *		mh = mh1 - mh2
 *	using two >= 0 variables.
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	GNetGr		: pointer to the first net group
 *	GRegList	: pointer to the first region
 *	GNetPointer[]	: pointers to nets
 *	parm		: parameter struct
 *
 *
 */


formulateLindo(initSize)

int		initSize[G_NHV];


{


    GRegionDef	*nextRegion;
    GNetGrDef	*nextNetGr;

    int		i,j,n;
    int		M;
    int		capacity;
    int		numTerm;
    int		termsPerLine = 4;
    float		netlength_fac;
    float		blockloc_fac;



    M = parm.M;

    /* netlength_fac = weight * (W0 H0) / netlength */
    netlength_fac = (float)statis.critical_dim[G_HOR] *
      (float)statis.critical_dim[G_VER] *
	parm.netlength_weight / (float)statis.init_netlength;
    /* blockloc_fac = weight * (W0 H0) / #block(W0 + H0) */
    blockloc_fac = (float)statis.critical_dim[G_HOR] *
      (float)statis.critical_dim[G_VER] *
	parm.blockloc_weight / 
	  ((float)statis.num_block *
	   ((float)statis.critical_dim[G_HOR] +
	    (float)statis.critical_dim[G_VER]));
    lindoFile = fopen("lindo.in","w");


    /* Generate objective function */
    /*-----------------------------*/
    fprintf(lindoFile,"MAX %dmh1 - %dmh2 + %dmw1 - %dmw2\n",
	    initSize[G_HOR],initSize[G_HOR],initSize[G_VER],initSize[G_VER]);
    statis.num_variable += 4;

    /* Add the terms for the total net length (note the - sign, not +) */
    for (nextNetGr = GNetGr; nextNetGr != NULL; nextNetGr = nextNetGr->next) {

	if (netlength_fac == 0.0) {
	    break;
	}
	if ((askTrivialRoute(nextNetGr) == YES) 
	    || (nextNetGr->NoCritNet == nextNetGr->NoNet)) {
	    continue;
	}

	numTerm = 1;
	fprintf(lindoFile," - %.1fx%dv1 ", 
		netlength_fac * (float)(nextNetGr->candRoute[0]->length), 
		nextNetGr->id);
	for (j = 1; j <= G_MAXROUTE - 1; ++j) {
	    if (nextNetGr->candRoute[j] == NULL) {
		fprintf(lindoFile,"\n");
		break;
	    }
	    numTerm++;
	    if (numTerm<=termsPerLine) {
		fprintf(lindoFile,"- %.1fx%dv%d ",
			netlength_fac * (float)(nextNetGr->candRoute[j]->length),
			nextNetGr->id, j+1);
	    } else {
		numTerm = 1;
		fprintf(lindoFile,"\n - %.1fx%dv%d ",
			netlength_fac * (float)(nextNetGr->candRoute[j]->length),
			nextNetGr->id, j+1);
	    }
	}
    }

    /* Add the terms for minimum coordinate of blocks (note the - sign, not +) */
    for (nextRegion = GRegList; nextRegion; nextRegion = nextRegion->next) {
	if (parm.placeRefMode == 0 || blockloc_fac == 0.0) {
	    break;
	}
	if (nextRegion->type != G_INST) {
	    continue;
	}
	fprintf(lindoFile," - %fpx%d - %fpy%d\n", 
		blockloc_fac, nextRegion->id, blockloc_fac, nextRegion->id);
    }

    /* Generate constraints */
    /*----------------------*/
    fprintf(lindoFile,"ST\n");

    /* xiv1 + xiv2 + ..... + xivn = (# non-critical net)	(for net group i) */
    /*------------------------------------------------------------------------*/
    for (nextNetGr = GNetGr; nextNetGr != NULL; nextNetGr = nextNetGr->next) {

	if ((askTrivialRoute(nextNetGr) == YES) 
	    || (nextNetGr->NoCritNet == nextNetGr->NoNet)) {
	    continue;
	}

	numTerm = 1;
	fprintf(lindoFile,"x%dv1 ",nextNetGr->id);
	++(statis.num_variable);
	for (j = 1; j <= G_MAXROUTE - 1; ++j) {
	    if (nextNetGr->candRoute[j] == NULL) {
		break;
	    }
	    numTerm++;
	    if (numTerm<=termsPerLine) {
		fprintf(lindoFile,"+ x%dv%d ",nextNetGr->id,j+1);
	    } else {
		numTerm = 1;
		fprintf(lindoFile,"\n + x%dv%d ",nextNetGr->id,j+1);
	    }
	    ++(statis.num_variable);
	}

	fprintf(lindoFile," = %d\n", nextNetGr->NoNet - nextNetGr->NoCritNet);
	++(statis.num_constraint);
    }

    /* Formulate polar path length constraint
       Mode1 summs up tile dimension to get the length
       Mode2 uses coordinates of tiles and sets a capacity constraint
       Mode2 includes graph topology preservation constraint
       ------------------------------------------------------------------- */
    if (parm.placeRefMode == OFF) {
	formulateMode1(initSize);
    } else {
	/* We need x, y placement refinement variables for each block */
	statis.num_variable += statis.num_block * 2;
	formulateMode2(initSize);
    }


    /* Demand of over cell routing <= capacity of over cell region */
    /*-------------------------------------------------------------*/
    for (nextRegion = GRegList; nextRegion != NULL; nextRegion = nextRegion->next) {
	if (nextRegion->type != G_OVERCELL) {
	    continue;
	}

	for (nextNetGr = GNetGr; nextNetGr != NULL; nextNetGr = nextNetGr->next) {
	    for (j = 0; j < M; ++j) {

		fprintf(lindoFile,"%.1fx%dv%d",
 			WGT[nextRegion->id - 1][M * (nextNetGr->id - 1) + j],
			nextNetGr->id, j + 1);
		if ((nextNetGr->next != NULL) || (j != M - 1)) {
		    fprintf(lindoFile," + ");
		}
	    }
	}

	capacity = nextRegion->capacity[0];
	for (n = 0; n <= G_MAXREGION - 1; ++n) {
	    capacity -= FIX[G_OVERCELL][nextRegion->id - 1];
	}
	fprintf(lindoFile," <= %d\n",capacity);
	++(statis.num_constraint);
    }


    fprintf(lindoFile,"END\n");


    /* Declaration of integrality */
    /*----------------------------*/
    for (nextNetGr = GNetGr; nextNetGr != NULL; nextNetGr = nextNetGr->next) {

	if ((askTrivialRoute(nextNetGr) == YES) 
	    || (nextNetGr->NoCritNet == nextNetGr->NoNet)) {
	    continue;
	}

	/* If there are more than 1 net, skip this */

	if (nextNetGr->NoNet - nextNetGr->NoCritNet > 1 || parm.INTEGER == OFF) {
	    continue;
	}

	/* Otherwise declaration of integers */

	for (j = 0; j <= G_MAXROUTE - 1; ++j) {
	    if (nextNetGr->candRoute[j] == NULL) {
		break;
	    }
	    fprintf(lindoFile,"INTEGER x%dv%d\n",nextNetGr->id,j+1);
	}
    }

    /* Divert output to a file */

    fprintf(lindoFile,"DIVERT\n");
    fprintf(lindoFile,"lindo.out\n");
    if (parm.BIP == ON) {
	fprintf(lindoFile,"BIP 0\n");
    }
    fprintf(lindoFile,"GO\n");
    fprintf(lindoFile,"RVRT\n");
    fprintf(lindoFile,"QUIT\n");
    fclose(lindoFile);



}





/*
 *  formulateMode1()
 *
 *  DESCRIPTION
 *  -----------
 *	Formulate polar path length constraints
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	initSize<input>	: initial dimension of chip before IP optimization
 *
 *  Local variables:
 *  ---------------
 *	unitPitch	: unit pitch between tracks in OCT units
 *	coeff		: coefficient for integer programming
 *	numTerm		: counter for the number of terms per line for lindo
 *	termsPerLine	: maximum number of terms per line for lindo
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  NOTE
 *  ----
 *	Improvement in dimension may be negative and thus represented as
 *		mw = mw1 - mw2
 *		mh = mh1 - mh2
 *	using two >= 0 variables.
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	buf		: string buffer for system command
 *	GNetGr		: pointer to the first net group
 *	GID2Reg[i]	: pointers to region with id i
 *	FIX[i][j]	: fixed usage of a region j in direction i
 *	INC[i][j][k]	: k th region contained in polar path j in direction i
 *	WGT[i][j]	: weight of candidate route j to region i
 *	parm		: parameter struct
 *	statis		: statistics struct
 *
 *
 */


formulateMode1(initSize)

int		initSize[G_NHV];


{


    GNetGrDef	*nextNetGr;

    int		i,j;
    int		l,m,n;
    int		M;
    int		size;
    int		unitPitch = G_UNITPITCH;
    float		coeff;
    int		flag;
    int		flag2;
    int		numTerm;
    int		termsPerLine = 4;



    M = parm.M;

    /* Length of polar path + margin <= initial length */
    /*-------------------------------------------------*/

    /* For each direction, G_HOR and G_VER */

    for (l = 0; l <= G_NHV - 1; ++l) {

	/* For each polar path */

	for (m = 0; m <= statis.num_polarpath[l] - 1; ++m) {

	    /* For each net group */

	    for (flag = ON, nextNetGr = GNetGr; nextNetGr != NULL;
		 nextNetGr = nextNetGr->next) {

		if ((askTrivialRoute(nextNetGr) == YES) 
		    || (nextNetGr->NoCritNet == nextNetGr->NoNet)) {
		    continue;
		}

		/* For each candidate route */

		for (flag2 = OFF, j = 0; j <= M - 1; ++j) {

		    /* For each routing region */

		    for (coeff = 0.0, n = 0; n <= G_MAXREGION - 1; ++n) {
			if (INC[l][m][n] == 0) {
			    break;
			} else if (GID2Reg[INC[l][m][n]-1]->type==l) {
			    coeff += WGT[INC[l][m][n] - 1][M*(nextNetGr->id-1)+j];
			}
			else if (WGT[INC[l][m][n]-1][M*(nextNetGr->id-1)+j]!=0.0) {
			    printf("Looks like an error at l=%d n=%d INC=%d\n",
				   l, n, INC[l][m][n]);
			}
		    }

		    /* Multiply unitPitch to get the actual width in OCT units */
		    /*---------------------------------------------------------*/
		    coeff *= (float)unitPitch;

		    if (coeff != 0.0) {
			if (flag == ON) {
			    fprintf(lindoFile,"   %dx%dv%d",(int)coeff,
				    nextNetGr->id,j+1);
			    flag = OFF;
			    numTerm = 1;
			} else if (numTerm<termsPerLine) {
			    numTerm++;
			    fprintf(lindoFile," + %dx%dv%d",(int)coeff,
				    nextNetGr->id,j+1);
			} else {
			    numTerm = 1;
			    fprintf(lindoFile,"\n + %dx%dv%d",(int)coeff,
				    nextNetGr->id,j+1);
			}
			flag2 = ON;
		    }

		}
		/* Only if this net group contributes this path, change the line */
		if (flag2 == ON) {
		    numTerm = 0;
		    fprintf(lindoFile,"\n");
		}
	    }
	    (void) sprintf(buf, (l == G_HOR) ? "mw1 - mw2" : "mh1 - mh2");
	    fprintf(lindoFile," + %s",buf);

	    /* Take care of fixed usage */
	    /*--------------------------*/
	    size = initSize[l];
	    for (n = 0; n <= G_MAXREGION - 1; ++n) {
		if (INC[l][m][n] == 0) {
		    break;
		} else {
		    size -= FIX[l][INC[l][m][n] - 1];
		}
	    }
	    fprintf(lindoFile," <= %d\n",size);
	    ++(statis.num_constraint);
	}
    }


}




/*
 *  formulateMode2()
 *
 *  DESCRIPTION
 *  -----------
 *	Formulate polar path length constraints using coordinate.
 *	It finds a capapcity for each bottleneck and sets up a capacity
 *	constraint for each bottleneck.
 *	Positions of blocks are controlled to preserve the current topology
 *	at the specified level.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	initSize<input>	: initial dimension of chip before IP optimization
 *
 *  Local variables:
 *  ---------------
 *	coeff		: coefficient for integer programming
 *	numTerm		: counter for the number of terms per line for lindo
 *	termsPerLine	: maximum number of terms per line for lindo
 *	reg1		: block region adjacent to the bottleneck (left or down)
 *	reg2		: block region adjacent to the bottleneck (right or up)
 *	regUL		: block region on upper left of XSEC region
 *	regUR		: block region on upper right of XSEC region
 *	regLL		: block region on lower left of XSEC region
 *	regLR		: block region on lower right of XSEC region
 *	rhs		: right hand side constant
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  NOTE
 *  ----
 *	Improvement in dimension may be negative and thus represented as
 *		mw = mw1 - mw2
 *		mh = mh1 - mh2
 *	using two >= 0 variables.
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	GNetGr		: pointer to the first net group
 *	GRegList	: pointer to the first region
 *	FIX[i][j]	: fixed usage of a region j in direction i
 *	adjList[i]	: pointer to the first adjacent graph node to region i
 *	parm		: parameter struct
 *
 *
 */


formulateMode2(initSize)

int		initSize[G_NHV];


{


GRegionDef	*nextReg;
GNetGrDef	*nextNetGr;
int		i,j;
int		M;
int		threshold;
float		coeff;
int		unitPitch = G_UNITPITCH;
int		flag;
int		flag2;
int		numTerm;
int		termsPerLine = 4;

GRegionDef	*nextBot;
GRegionDef	*reg1,*reg2;
GRegionDef	*regUL, *regUR, *regLL, *regLR;
adjNodeType	*nextAdj;
int		rhs;



M = parm.M;
threshold = parm.tile_threshold;

for (nextReg = GRegList; nextReg != NULL; nextReg = nextReg->next) {

    rhs = 0;

    /* Estimated width of bottleneck <= given space for bottleneck */
    /*-------------------------------------------------------------*/
    if ((nextReg->type == G_HOR_BN) || (nextReg->type == G_VER_BN)) {

	/* Obtain area contribution of each route candidate */
	/*--------------------------------------------------*/
	for (flag = ON, nextNetGr = GNetGr; nextNetGr != NULL;
			nextNetGr = nextNetGr->next) {

	    if ((askTrivialRoute(nextNetGr) == YES) 
		|| (nextNetGr->NoCritNet == nextNetGr->NoNet)) {
		continue;
	    }

	    for (j = 0, flag2 = OFF; j <= M - 1; ++j) {

 		coeff = WGT[nextReg->id - 1][M*(nextNetGr->id-1)+j];

		/* Multiply unitPitch to get the actual width in OCT units */
		/*---------------------------------------------------------*/
		coeff *= (float)unitPitch;

		if (coeff != 0.0) {
		    if (flag == ON) {
			fprintf(lindoFile,"   %dx%dv%d",(int)coeff,
				nextNetGr->id,j+1);
			flag = OFF;
			numTerm = 1;
		    } else if (numTerm<termsPerLine) {
			numTerm++;
			fprintf(lindoFile," + %dx%dv%d",(int)coeff,
				nextNetGr->id,j+1);
		    } else {
			numTerm = 1;
			fprintf(lindoFile,"\n + %dx%dv%d",(int)coeff,
				nextNetGr->id,j+1);
		    }
		    flag2 = ON;
		}
	    }
	    /* Only if this net group contributes this path, change the line */
	    if (flag2 == ON) {
		numTerm = 0;
		fprintf(lindoFile,"\n");
	    }
	}

	/* Take care of fixed usage */
	rhs -= FIX[nextReg->type][nextReg->id - 1];

	/* Calculate the space given between blocks */
	/*------------------------------------------*/
	if (nextReg->type == G_HOR_BN) {

	    /* Special cases where the region is adjacent to source or sink */

	    if (nextReg->adjReg[G_RIGHT] == NULL) {
	
		/* X..... <= (initWidth - mw) - (PXreg1 + dimension) */

		reg1 = nextReg->adjReg[G_LEFT];
		reg2 = NULL;
		fprintf(lindoFile," + mw1 - mw2 + px%d\n", reg1->id);
		rhs += initSize[G_HOR];
		rhs -= reg1->dimension[G_HOR];
	    } else if (nextReg->adjReg[G_LEFT] == NULL) {

		/* X..... <= PXreg2 - 0 */

		reg1 = NULL;
		reg2 = nextReg->adjReg[G_RIGHT];
		fprintf(lindoFile," - px%d\n", reg2->id);
	    } else {

		/* X..... <= PXreg2 - (PXreg1 + dimension) */

		reg1 = nextReg->adjReg[G_LEFT];
		reg2 = nextReg->adjReg[G_RIGHT];
		fprintf(lindoFile," - px%d + px%d\n", reg2->id, reg1->id);
		rhs -= reg1->dimension[G_HOR];
	    }
	} else {
	    if (nextReg->adjReg[G_UP] == NULL) {

		/* X..... <= (initHeight - mh) - (PYreg1 + dimension) */

		reg1 = nextReg->adjReg[G_DOWN];
		reg2 = NULL;
		fprintf(lindoFile," + mh1 - mh2 + py%d\n", reg1->id);
		rhs += initSize[G_VER];
		rhs -= reg1->dimension[G_VER];
	    } else if (nextReg->adjReg[G_DOWN] == NULL) {

		/* X..... <= PYreg2 - 0 */

		reg1 = NULL;
		reg2 = nextReg->adjReg[G_UP];
		fprintf(lindoFile," - py%d\n", reg2->id);
	    } else {

		/* X..... <= PYreg2 - (PYreg1 + dimension) */

		reg1 = nextReg->adjReg[G_DOWN];
		reg2 = nextReg->adjReg[G_UP];
		fprintf(lindoFile," - py%d + py%d\n", reg2->id, reg1->id);
		rhs -= reg1->dimension[G_VER];
	    }
	}

	fprintf(lindoFile," <= %d\n", rhs);
	++(statis.num_constraint);

	/* Constraint to maintain bottlenecks */
	/*------------------------------------*/
	if (parm.placeRefMode != PR_MODE1) {
	    continue;
	}
	if ((nextReg->type == G_HOR_BN) && (reg1 != NULL) && (reg2 != NULL)) {

	    /* (PYreg2 + dimension) - PYreg1 >= threshold */ 

	    fprintf(lindoFile,"py%d - py%d >= %d\n",
		reg2->id, reg1->id, threshold - reg2->dimension[G_VER]);
	    ++(statis.num_constraint);

	    /* (PYreg1 + dimension) - PYreg2 >= threshold */ 

	    fprintf(lindoFile,"py%d - py%d >= %d\n",
		reg1->id, reg2->id, threshold - reg1->dimension[G_VER]);
	    ++(statis.num_constraint);
	}
	if ((nextReg->type == G_VER_BN) && (reg1 != NULL) && (reg2 != NULL)) {

	    /* (PXreg2 + dimension) - PXreg1 >= threshold */ 

	    fprintf(lindoFile,"px%d - px%d >= %d\n",
		reg2->id, reg1->id, threshold - reg2->dimension[G_HOR]);
	    ++(statis.num_constraint);

	    /* (PXreg1 + dimension) - PXreg2 >= threshold */ 

	    fprintf(lindoFile,"px%d - px%d >= %d\n",
		reg1->id, reg2->id, threshold - reg1->dimension[G_HOR]);
	    ++(statis.num_constraint);
	}

    /* Constraint to maintain topology at the XSEC */
    /*---------------------------------------------*/
    } else if (nextReg->type == G_DOM_XSEC) {

	/* Check if this region is a + type juction */

	for (i = 0, nextAdj = adjList[nextReg->id - 1]; nextAdj != NULL;
		nextAdj = nextAdj->next) {
	    ++i;
	}
	if (i != G_NDIR) {
	    continue;
	} else {
	    printf("Found a + junction at region %d\n", nextReg->id);
	}

	/* Obtain the four blocks surrounding this XSEC region */

	for (nextBot = GRegList; nextBot != NULL; nextBot = nextBot->next) {
	    if ((nextBot->type != G_HOR_BN) && 
			(nextBot->type != G_VER_BN)) {
		continue;
	    }
	    if (nextBot->adjReg[G_DOWN] == nextReg) {
		regUL = nextBot->adjReg[G_LEFT];
		regUR = nextBot->adjReg[G_RIGHT];
	    }
	    if (nextBot->adjReg[G_UP] == nextReg) {
		regLL = nextBot->adjReg[G_LEFT];
		regLR = nextBot->adjReg[G_RIGHT];
	    }
	}

	/* PXregLR - (PXregUL + dimension) >= threshold */ 

	fprintf(lindoFile,"px%d - px%d >= %d\n",
		regLR->id,regUL->id, threshold + regUL->dimension[G_HOR]);
	++(statis.num_constraint);

	/* PYregUL - (PYregLR + dimension) >= threshold */ 

	fprintf(lindoFile,"py%d - py%d >= %d\n",
		regUL->id,regLR->id, threshold + regLR->dimension[G_VER]);
	++(statis.num_constraint);

	/* PXregUR - (PXregLL + dimension) >= threshold */ 

	fprintf(lindoFile,"px%d - px%d >= %d\n",
		regUR->id,regLL->id, threshold + regLL->dimension[G_HOR]);
	++(statis.num_constraint);

	/* PYregUR - (PYregLL + dimension) >= threshold */ 

	fprintf(lindoFile,"py%d - py%d >= %d\n",
		regUR->id,regLL->id, threshold + regLL->dimension[G_VER]);
	++(statis.num_constraint);

    /* Otherwise continue */
    /*--------------------*/
    } else {
	continue;
    }
}
printf("\n");


}



