/* mercury.h - common data structure for MERCURY global router
 
   Copyright (C) 1988 by Mitch Igusa and Yoshihito Nishizaki.
		All rights reserved.

 */

#ifndef MAXINT
#define MAXINT		(2147483647)
#endif /* MAXINT */
#define MAXLINE		80

#define RUSAGE_SELF	0
#define RUSAGE_CHILDREN	-1

#define FOUND 1
#define NOT_FOUND 0
#define ON 1
#define OFF 0
#define YES 1
#define NO 0
#define TRUE 1
#define FALSE 0

typedef char	Boolean;

/* Globals for macro cell global routing
-----------------------------------------------------------------------------*/
typedef	struct	GChipDef	GChipDef;
typedef	struct	GRegionDef	GRegionDef;
typedef	struct	GPolarDef	GPolarDef;

typedef	struct	GChanTermDef	GChanTermDef;
typedef	struct	GTermDef	GTermDef;
typedef	struct	GPinDef		GPinDef;

typedef	struct	GNodeDef	GNodeDef;
typedef	struct	GPathDef	GPathDef;
typedef	struct	GRouteDef	GRouteDef;
typedef	struct	GNetGrDef	GNetGrDef;
typedef	struct	GNetDef		GNetDef;

typedef	struct	route		routeType;
typedef	struct	path		pathType;
typedef struct	nodeGroup	nodeGroupType;

#define G_BSIZE		512

#define	G_NLAY		2	/* Number of routing layers */
#define	G_NDIR		4	/* Up, down, left or right */
#define G_NHV		2	/* Horizontal or vertical */
#define	G_NLR		2	/* Left or right, up or down */

#define G_HOR		0	/* horizontal */
#define G_VER		1	/* vertical */

#define	HORIZONTAL  	1

#define G_HOR_BN	0	/* horizontal bottleneck */
#define G_VER_BN	1	/* vertical bottleneck */
#define G_OVERCELL	2	/* over the cell routing region */
#define G_INST		3	/* instance (solid tile) */
#define G_SWBOX		4	/* switch box */
#define G_CHANNEL	5	/* channel */
#define G_DOM_XSEC	6	/* crossing section of two dominant tiles */

#define G_UP		0	/* Different directions */
#define G_LEFT		1
#define G_DOWN		2
#define G_RIGHT		3

#define G_REVDIR(dir) ((dir+2)%4)

#define G_MAXNET	1000
#define G_MAXNETGR	G_MAXNET
#define G_MAXREGION	500
#define G_MAXNODE	G_MAXREGION
#define G_MAXROUTE	20
#define G_MAXPATH	G_MAXROUTE
#define G_MAXPOLAR	100		/* # polar paths for each direction */
#define G_MAXPIN_PER_NET	100

#define G_REGBOX_LL_X(r) ((r)->centerLocX - (r)->dimension[G_HOR]/2)
#define G_REGBOX_LL_Y(r) ((r)->centerLocY - (r)->dimension[G_VER]/2)
#define G_REGBOX_UR_X(r) ((r)->centerLocX + (r)->dimension[G_HOR]/2)
#define G_REGBOX_UR_Y(r) ((r)->centerLocY + (r)->dimension[G_VER]/2)

struct GChipDef {
    char	*chipName;
    char	doNotUseExtChannelsFlg;
};

struct GRegionDef {
    int		type;			/* type of this region */
    int		id;			/* region ID > 0 */
    char	*dbRef;			/* Data base reference */
    Boolean	expandableFlg;		/* Channel is expandable? */
    Boolean	externFlg;		/* External channel flag */
    int		orient;			/* orientation of bottleneck,
					/* G_HOR or G_VER */
    int		overcellDir;		/* possible direction
					   for over cell routing */

/* used when bottleneck is split intentionally because of its big size */
    GRegionDef	*parent;		/* NULL if master */
    GRegionDef	*sibling;		/* Go through all subregions
					   under parent */
    GRegionDef	*child;			/* First child under parent */

#ifdef UNUSED
/* used when bottleneck is split by orthogonal tiles crossing it */
    GRegionDef	*parentBottleneck;	/* NULL if master */
    GRegionDef	*siblingBottleneck;	/* Go through all subregions
					   under parent */
    GRegionDef	*childBottleneck;	/* First child under parent */
#endif

    GRegionDef	*adjReg[G_NDIR];	/* Adjacent regions:
					/*   ... like corner stitching
					/*   adjReg[G_UP] = to right most region
					/*       above this region
					/*   adjReg[G_DOWN] = to left most
					/*       region below this region
					/*   adjReg[G_LEFT] = to bottom most
					/*       region to left of this region
					/*   adjReg[G_RIGHT] = to top most
					/*	 region to right of this region
					*/

    int		capacity[G_NLAY];	/* Capacity of channel */
    int		usage[G_NLAY];		/* Used portion of channel */
    int		estUsage;		/* estimation of usage */
    int		dimension[G_NHV];	/* H/V size of instance */
					/* represents smallest dimension
					/* in case of rectilinear instance */

    int		centerLocX;		/* (Xmin+Xmax)/2 for this region */
    int		centerLocY;		/* (Ymin+Ymax)/2 for this region */

    int		newLocX;		/* New lower left location for region */
    int		newLocY;

    GPinDef	*pin;			/* first pin belonging to this region */
					/* created when make net group */
    GRegionDef	*next;			/* pointer to next region
					/* for sequential link */
};

struct GPolarDef {
    GRegionDef	*region;		/* Points to corresponding GRegion */
					/* Either an instance or bottleneck */
    GPolarDef	*pGraph;		/* Next upper/right polar graph 
					/*   node starting from left/bottom */
    GPolarDef	*nextPGraph;		/* Goes through the rest of edges of 
					/*   polar graph starting from pGraph */

/* This may have a non-zero value even for a bottleneck in the existence of 
	rectilinear cells. Otherwise should be 0 for bottlenecks */
/* If it has a non-zero value for a bottleneck, 
	the adjacent cell must have several different width/height */
    int		length;			/* Dimensions of instance/bottleneck */
};

struct GNetDef {
    int		critical;		/* this net is critical? */
    int		criticality;		/* index of net criticality */
    int		maxLength;		/* maximum allowable net length */
    int		trivial;		/* all pins located in one region? */
    int		routeNo;		/* index to candidate route chosen */
    char	*dbRef;			/* Data base reference */
    GNetDef	*next;			/* next net */
    GNetDef	*nextGrMem;		/* next member of the same net group */
    GNetGrDef	*netGr;			/* the net group this net belongs to */
    GPinDef	*pin;			/* list of physical pins of this net */
    GTermDef	*term;			/* Terminals */

/* Imported from cds.h */
    octId		id;		/* octId */
    char		*name;		/* net name */
    int 		number;		/* Sequence number of the net name */
    int			length;		/* net length */
    struct subnetList	*subnet;	/* pointer to first subnet */
};

struct GPinDef {
    char	*name;			/* pin name */
    char	*dbRef;			/* Data base reference */
    GRegionDef	*region;		/* region containing this pin */
    int		offset;			/* Offset of pin from bottom or left */
    int		netStat;		/* net status
					/* if ON, net starts at this pin
					/* if OFF, net ends at this pin
					/* created when make net group */
					/* used for max density calculation */
    Boolean	mutExFlg;		/* equivalent pins should be used
					/*	mutually exclusively ? */
					/*	not handled for present */
    GPinDef	*equiv;			/* Next equivalent pin of pin group */
    GPinDef	*next;			/* Next pin of net or region */
};

struct GNodeDef {
    GRegionDef	*region;		/* region corresponding to this node */
    GNodeDef	*nextEq;		/* next equivalent node */
    GNodeDef	*next;			/* next node of the path */
};

#ifdef UNUSED
struct GPathDef {
    int length;				/* length of the path */
    GNodeDef	*node;			/* first node included in this path */
    GPathDef	*next;			/* next path of the route */
};

struct GRouteDef {
    GPathDef	*path;			/* first path included in this route */
    GRouteDef	*next;			/* next route */
    int length;				/* length of the route */
};
#endif

struct route {
    int 	length;
					/* if ON, region i included in this route */
    int 	regionFlg[G_MAXREGION];		
    pathType 	*path;			/* List of 2 end point paths */
    routeType	*next;			/* Linked list of routes (no meaning) */
};

struct path {				/* 2 end point path */
    int		length;
    nodeGroupType *node;		/* Ordered list of nodes along path */
    pathType	 *next;			/* To next 2 end point path */
};

struct nodeGroup {
    int		id;			/* Region id */
    nodeGroupType *next;		/* Linked list of nodes */
};

struct GNetGrDef {
    int		id;			/* id of this net group */
    int		NoNet;			/* # nets belonging to this net group */
    int		NoCritNet;		/* # critical nets among the above */
    int		NoRegion;		/* # regions defining net topology */
    GNetDef	*net;			/* first net belonging to this group */
    GNodeDef	*node;			/* first node belonging to this group */
    GNetGrDef	*next;			/* next net group */
    routeType	*candRoute[G_MAXROUTE];	/* candidate routes for this group */
					/*   if null then no more candidate */
					/*   routes */
    int		IPresult[G_MAXROUTE];	/* result of IP,
					/* # nets taking each candidate */
};

GPolarDef	*GPGraph[G_NHV];	/* source of polar graph */
GNetDef		*GNet;			/* first net */
GNetGrDef	*GNetGr;		/* first net group */
GNetGrDef	*GNetGrTail;		/* last net group */
GRegionDef	*GRegList;		/* first GRegion
					/* it should be at lower-left corner */

GNetDef		*GNetPointer[G_MAXNET];	/* i-th net */
GRegionDef	*GID2Reg[G_MAXREGION];	/* region id i --> GRegionDef */

Boolean		DFSflag[G_MAXREGION];	/* flag, if ON, threaded */
