#include "port.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "symlib.h"
#include "th.h"

#define R_HORIZONTAL 0
#define R_VERTICAL   1




/*

Routing done using a single axis Steiner tree.

  |                       |
  |        |              | 
  *--*-----*--------------*----------*
     |                               |

*/

typedef struct route_point {
    struct octBox forbiddenBox;
    octCoord t[2];						 /* Terminal coords */
    octCoord c[2];						 /* Contact coords */
    octId    termId;
    octId    contactId;
} route_point;

typedef struct route_tree {
    struct octBox bb;
    int axis_dir;
    octCoord axis_coord;
    octCoord start;
    octCoord end;

    int      size;		/* Size of array. */
    route_point *p;		/* Term coord array. */
    int      count;		/* Number of terms in array. */
    
    octObject stubLayer;
    octObject axisLayer;
} route_tree;


route_tree* initTree( net )
    octObject *net;
{
    int size  = ohCountContents( net, OCT_TERM_MASK );
    route_tree *t = ALLOC( route_tree, 1 );

    t->size = size;
    t->p    = ALLOC( route_point , size );
    t->count = 0;
    ohNetBB( net, &t->bb );
    
    return t;
}

findRoutingLayers( net, t )
    octObject *net;
    route_tree *t;
{
    octObject facet;

    octGetFacet( net, &facet );

    ohGetOrCreateLayer( &facet, &t->axisLayer, "WIRING" );
    ohGetOrCreateLayer( &facet, &t->stubLayer, "WIRING" );
}



void deleteTree( t )
    route_tree *t;
{
    FREE(t->p);
    FREE( t );
}

#define PROPEQ(p,s) ( !strcmp((p)->contents.prop.value.string, (s)))

void placeAxis( net,  t )
    octObject *net;
    route_tree *t;
{
    octPoint center;
    octCoord w = ohBoxWidth( &t->bb );
    octCoord h = ohBoxHeight( &t->bb );
    octCoord correction =  2 * thLambda( );
    octObject prop;


    ohBoxCenter( &t->bb, &center );


    if ( ohGetByPropName(net,&prop,"NETTYPE") ==OCT_OK ) {
	if ( PROPEQ( &prop,"SUPPLY" ) ) {
	    t->axis_dir = R_HORIZONTAL;
	    t->axis_coord = t->bb.upperRight.y - correction;;
	    return;
	} else if ( PROPEQ( &prop,"GROUND" ) ) {
	    t->axis_dir = R_HORIZONTAL;
	    t->axis_coord = t->bb.lowerLeft.y + correction;;
	    return;
	} 
    }
    
    if ( h > w ) {
	t->axis_dir = R_VERTICAL;
	t->axis_coord = center.x;
	if ( w <= thLambda() ) t->axis_coord += correction;
    } else {
	t->axis_dir = R_HORIZONTAL;
	t->axis_coord = center.y;
	if ( h < thLambda() ) t->axis_coord += correction;
    }
}

int sortPins( net, t )
    octObject *net;
    route_tree *t;
{
    octGenerator genTerm;
    octObject term, instance;
    int dir = t->axis_dir;
    int other_dir = 1-dir;
    int count = 0;
    int i, spot;
    struct octBox term_bb;
    struct octPoint term_center;
    route_point point;

    
    octInitGenContents( net, OCT_TERM_MASK, &genTerm );
    while ( octGenerate( &genTerm, &term ) == OCT_OK ) {
	if ( octIdIsNull( term.contents.term.instanceId ) ) {
	    continue;		/* Ignore formals. */
	}
	if ( ohTerminalBB( &term, &term_bb ) != OCT_OK ) {
	    continue;		/* Ignore terminals without a BB. */
	}
	ohBoxCenter( &term_bb, &term_center );
	point.t[R_HORIZONTAL] = term_center.x;
	point.t[R_VERTICAL] = term_center.y;
	point.c[dir] = point.t[dir];
	point.c[other_dir] = t->axis_coord;
	point.termId = term.objectId;
	point.contactId = oct_null_id;

	OH_ASSERT(ohGetById( &instance, term.contents.term.instanceId ));
	OH_ASSERT(octBB( &instance, &point.forbiddenBox ));

	spot = 0;
	while ( spot < count && point.t[dir] > t->p[spot].t[dir] ) spot++; /* Find spot */
	for ( i = count ; i > spot ; i-- ) {			 /* Shift points forward.  */
	    t->p[i] = t->p[i-1];
	}
	t->p[spot] = point;					 /* Add new point at spot. */
	count++;
    }
    t->count = count;
    t->start = t->p[0].t[dir];
    t->end = t->p[count-1].t[dir];
}


static int correction_flag = 0;

int axisInBox( t, b )
    route_tree *t;
    struct octBox *b;
{
    octCoord c = t->axis_coord;
    int pad = thLambda();
    octCoord correction_minus;
    octCoord correction_plus;

    if ( t->axis_dir == R_VERTICAL ) {
	if ( t->end < b->lowerLeft.y || t->start > b->upperRight.y  ) return 0;
	correction_minus = b->lowerLeft.x - pad - c;
	correction_plus = b->upperRight.x +pad - c;
	if ( correction_minus > 0 || correction_plus < 0 ) {
	    return 0;
	}
    } else {
	if ( t->end < b->lowerLeft.x || t->start > b->upperRight.x  ) return 0;
	correction_minus = b->lowerLeft.y - pad - c;
	correction_plus = b->upperRight.y +pad - c;
	if ( correction_minus > 0 || correction_plus < 0 ) {
	    return 0;
	}
    }
    
    if ( correction_flag == 1 || ABS( correction_minus ) > ABS( correction_plus ) ) {
	correction_flag = 1;
	return correction_plus;
    } else {
	correction_flag = -1;
	return correction_minus;
    }
}
    
int axisIsBad( t )
    route_tree *t;
{
    int i = 0;
    int correction ;
    for (i = 0; i < t->count ;i++) {
	correction =  axisInBox( t, &t->p[i].forbiddenBox );
	if ( correction ) {
	    return correction;
	}
    }
    return 0;
}

realignAxis( t )
    route_tree *t;
{
    octCoord correction;
    octCoord totalCorrection = 0;
    correction_flag = 0;	/* To make avoid infinite loops, make all corrections with same sign. */
    
    while (  correction = axisIsBad( t ) ) {
	t->axis_coord += correction;
	totalCorrection += correction;
	/* printf( "\tCorrect by %d\n", correction ); */
    }

}


routeStubs( t )
    route_tree *t;
{
    int i;
    int stub_dir = ( t->axis_dir == R_VERTICAL ) ? SYM_HORIZONTAL: SYM_VERTICAL;
    octObject t1,t2;
    struct octPoint pt1,pt2;
    octCoord pathWidth = 2;

    
    for ( i = 0 ; i < t->count ; i++ ) {
	t->p[i].c[1-t->axis_dir] = t->axis_coord;
	OH_ASSERT(ohGetById( &t1, t->p[i].termId ));
	pt2.x = t->p[i].c[R_HORIZONTAL];
	pt2.y = t->p[i].c[R_VERTICAL];
	t2.objectId = oct_null_id;
	symCreateSegment( &t->stubLayer, &pt1, &t1, SYM_USE_EXISTING, 
			 &pt2, &t2, SYM_CREATE, 0, stub_dir );
	t->p[i].contactId = t2.objectId;
	
	if ( i > 0 ) {
	    OH_ASSERT(ohGetById( &t1, t->p[i-1].contactId ));
	    symCreateSegment( &t->axisLayer, &pt2, &t2, SYM_USE_EXISTING, &pt1, &t1, 
			     SYM_USE_EXISTING, 0, stub_dir );
	    
	}
    }
}

route( net )
    octObject *net;
{
    route_tree *t = initTree( net );
    findRoutingLayers( net, t );
    placeAxis( net, t );
    sortPins( net,  t );
    realignAxis( t );
    routeStubs( t );
    deleteTree( t );
}


implementNets( facet )
    octObject *facet;
{
    octGenerator genNet, genTerm;
    octObject t[2];
    octObject layer, path, net;
    int simplestTree = 0;

    ohGetOrCreateLayer( facet, &layer, "WIRING" );


    octInitGenContents( facet, OCT_NET_MASK, &genNet );
    while ( octGenerate( &genNet, &net ) == OCT_OK ) {
	if ( !simplestTree ) {
	    route( &net );
	} else {
	    int count = 0, i;
	    octInitGenContents( &net, OCT_TERM_MASK, &genTerm );
	    while ( octGenerate( &genTerm, &t[1] ) == OCT_OK ) {
		if ( octIdIsNull( t[1].contents.term.instanceId ) ) {
		    continue; /* Ignore formals. */
		}
		if ( count++ > 0 ) {
		    struct octBox b[2];
		    struct octPoint p[2];
		
		    for ( i = 0; i < 2 ; i++ ) {
			ohTerminalBB( &t[i], &b[i] );
			ohBoxCenter( &b[i], &p[i] );
		    }
		    path.type = OCT_PATH;
		    path.contents.path.width = 0;
		    octCreate( &layer, &path );
		    octAttach( &net, &path );
		    octAttach( &path, &t[0] );
		    octAttach( &path, &t[1] );
		    octPutPoints( &path, 2, p );

		} 
		t[0] = t[1];
	    }
	}
    }
}
