#include "octopus.h"

extern int	Rows, AndCols, OrCols, OddEvenBreak;
extern char	**Core;
extern s_arrays	Arrays;
extern s_cells	Cells;
extern s_output	**Outputs;
extern s_ruleNames RuleNames;
extern s_options Options;

octObject	*getInst();

/*
 * doGround
 *
 * For each plane, instantiate all the ground contacts (diffussion pads)
 * and ground plane contacts (metal to diffusion contacts), then wire
 * them together.
 * The OR plane is more compicated than the AND plane because of the
 * OddEvenBreak.  In the OR plane, on the left side of the OddEvenBreak,
 * rows in between poly lines that have even numbered transistor rows
 * below and odd numbered rows above are devoted to output signal contacts,
 * and the rows above odd transistor rows but below even transistor rows
 * are used for ground contacts.  The opposite is true on the right side
 * of the OddEvenBreak.
 */
doGround()
{
    int	row, col, gndSpace;

    /* place ground contacts (diffusion pads) next to OR plane trans	*/
    for ( row = 0; row < Rows; row += 2) {
	for ( col = OddEvenBreak; col < OrCols; ++col) {
	    if (( IS_PRODUCT( row, col )) ||
		(( row < Rows - 1 ) && (IS_PRODUCT( row + 1, col )))){
		mkAttach( Arrays.orEvenGnd, mkPoint( row, col ),
			getInst( Cells.orGnd.name, OCT_NO_TRANSFORM ));
	    }
	}
	for ( col = OddEvenBreak - 1; col >= 0; --col) {
	    if (( IS_PRODUCT( row, col )) ||
		(( row > 0 ) && (IS_PRODUCT( row - 1, col )))){
		mkAttach( Arrays.orOddGnd, mkPoint( row, col ),
			getInst( Cells.orGnd.name, OCT_NO_TRANSFORM ));
	    }
	}
    }
    /* handle top row of Odd side if Rows is even			*/
    if( Rows % 2 == 0) {
	row = Rows - 1;
	for ( col = OddEvenBreak - 1; col >= 0; --col) {
	    if (( IS_PRODUCT( row, col )) ||
		(( row > 0 ) && (IS_PRODUCT( row - 1, col )))){
		mkAttach( Arrays.orOddGnd, mkPoint( row, col ),
			getInst( Cells.orGnd.name, OCT_NO_TRANSFORM ));
	    }
	}
    }

    /* place ground contacts (diffusion pads) next to AND plane trans	*/
    for ( col = 0; col < AndCols; col += 2 ) {
	for ( row = 0; row < Rows; row++ ) {
	    if (( IS_LITERAL( row, col )) ||
		(( col > 0 ) && (IS_LITERAL( row, col - 1 )))){
		mkAttach( Arrays.andGnd, mkPoint( row, col ),
			getInst( Cells.andGnd.name, OCT_NO_TRANSFORM ));
	    }
	}
    }
    /* if AndCols even, last and col hasn't been looked at yet		*/
    if ( (AndCols % 2) == 0)
	{
	col = AndCols - 1;
	for ( row = 0; row < Rows; row++ ) {
	    if ( IS_LITERAL( row, col )) {
		mkAttach( Arrays.andGnd, mkPoint( row, col ),
			getInst( Cells.andGnd.name, OCT_NO_TRANSFORM ));
	    }
	}
    }

    /* place and plane ground contacts (metal-diffusion contacts)	*/
    gndSpace = Options.andGndLineSep / 2;
    for ( row = gndSpace; row < Rows; row += Options.andGndLineSep) {
	for (col = 0; col < AndCols; col += 2) {
	    mkAttach( Arrays.andPlaneGnd, mkPoint( row, col),
			getInst( Cells.andPlaneGnd.name, OCT_NO_TRANSFORM));
	}
	mkAttach( Arrays.andPlaneGnd, mkPoint( row, AndCols ),
		    getInst( Cells.vss_via.name, OCT_NO_TRANSFORM));
    }
    /* if AndCols is even, last col hasn't been considered yet		*/
    if( (AndCols % 2) == 0) {
	col = AndCols - 1;
	for ( row = gndSpace; row < Rows; row += Options.andGndLineSep) {
	    mkAttach( Arrays.andPlaneGnd, mkPoint( row, col),
			getInst( Cells.andPlaneGnd.name, OCT_NO_TRANSFORM));
	}
    }

    /* if the last placed was too far from top, place another		*/
    if( row - Rows < gndSpace) {
	row = Rows;
	for (col = 0; col < AndCols; col += 2) {
	    mkAttach( Arrays.andPlaneGnd, mkPoint( row, col),
			getInst( Cells.andPlaneGnd.name, OCT_NO_TRANSFORM));
	}
	/* if AndCols even, do last column				*/
	if( (AndCols % 2) == 0) {
	    col = AndCols - 1;
	    mkAttach( Arrays.andPlaneGnd, mkPoint( row, col),
			getInst( Cells.andPlaneGnd.name, OCT_NO_TRANSFORM));
	}
	mkAttach( Arrays.andPlaneGnd, mkPoint( row, AndCols ),
		    getInst( Cells.vss_via.name, OCT_NO_TRANSFORM));
    }

    /* place or plane ground contacts					*/
    /* first do from 0 to OddEvenBreak					*/

    gndSpace = Options.orGndLineSep / 2;
    if( OddEvenBreak >= 0 ){
	for( col = OddEvenBreak; col >= 0; col -= Options.orGndLineSep) {
	    for( row = 0; row < Rows; row += 2) {
		mkAttach( Arrays.orOddPlaneGnd, mkPoint( row, col),
			    getInst( Cells.orPlaneGnd.name, OCT_NO_TRANSFORM));
	    }
	    if( Rows % 2 == 0) { 
		mkAttach( Arrays.orOddPlaneGnd, 
				     mkPoint( Rows - 1, col),
				     getInst( Cells.orPlaneGnd.name, 
					      OCT_NO_TRANSFORM));
	    }
	    mkAttach( Arrays.orPower, mkPoint( 1, col),
			getInst( Cells.vss_via.name, OCT_NO_TRANSFORM));
	    mkAttach( Arrays.orPower, mkPoint( 0, col),
			getInst( Cells.vss_via.name, OCT_NO_TRANSFORM));
	}
	if( col + Options.orGndLineSep > gndSpace ) {
	    /* bad OddEvenBreak, add one more column of contacts	*/
	    col = gndSpace - 1;
	    for( row = 0; row < Rows; row += 2) {
		mkAttach( Arrays.orOddPlaneGnd, mkPoint( row, col),
			    getInst( Cells.orPlaneGnd.name, OCT_NO_TRANSFORM));
	    }
	    if( Rows % 2 == 0) { 
		mkAttach( Arrays.orOddPlaneGnd, 
				     mkPoint( Rows - 1, col),
			             getInst( Cells.orPlaneGnd.name, 
					      OCT_NO_TRANSFORM));
	    }
	    mkAttach( Arrays.orPower, mkPoint( 1, col),
			getInst( Cells.vss_via.name, OCT_NO_TRANSFORM));
	    mkAttach( Arrays.orPower, mkPoint( 0, col),
			getInst( Cells.vss_via.name, OCT_NO_TRANSFORM));
	}
    }
    if( OddEvenBreak < OrCols ){

	/* now do from OddEvenBreak to OrCols - 1			*/
	/* note: the OddEvenBreak itself has ground contacts on both	*/
	/* the Odd rows and the Even rows				*/

	for( col = OddEvenBreak; col < OrCols; col += Options.orGndLineSep) {
	    for( row = 0; row < Rows; row += 2) {
		mkAttach( Arrays.orEvenPlaneGnd, mkPoint( row, col),
			    getInst( Cells.orPlaneGnd.name, OCT_NO_TRANSFORM));
	    }
	    /*
	    if( Rows % 2 ) { 
		mkAttach( Arrays.orEvenPlaneGnd, 
				     mkPoint( Rows - 1, col),
			             getInst( Cells.orPlaneGnd.name, 
					      OCT_NO_TRANSFORM));
	    }
	    */
	    if( col != OddEvenBreak && OddEvenBreak >= 0 ){
		mkAttach( Arrays.orPower, mkPoint( 1, col),
			    getInst( Cells.vss_via.name, OCT_NO_TRANSFORM));
		mkAttach( Arrays.orPower, mkPoint( 0, col),
			    getInst( Cells.vss_via.name, OCT_NO_TRANSFORM));
	    }
	}
	if(( OrCols - col + Options.orGndLineSep) > gndSpace ){
	    /* bad OddEvenBreak, add one more column of contacts	*/
	    if((col = OrCols - gndSpace + 1 ) > 0 ){
		for( row = 0; row < Rows; row += 2) {
		    mkAttach( Arrays.orEvenPlaneGnd, 
					 mkPoint( row, col),
					 getInst( Cells.orPlaneGnd.name, 
						  OCT_NO_TRANSFORM));
		}
	    }
	    /*
	    if( Rows % 2 ) { 
		mkAttach( Arrays.orEvenPlaneGnd, 
				     mkPoint( Rows - 1, col),
			             getInst( Cells.orPlaneGnd.name, 
					      OCT_NO_TRANSFORM));
	    }
	    */
	    mkAttach( Arrays.orPower, mkPoint( 1, col),
			getInst( Cells.vss_via.name, OCT_NO_TRANSFORM));
	    mkAttach( Arrays.orPower, mkPoint( 0, col),
			getInst( Cells.vss_via.name, OCT_NO_TRANSFORM));
	}
    }

    if ( !Options.wire ) return;

    /* AND Plane diffusion and ground					*/
    /* wire together diffusion in each and col				*/
    for( col = 0; col < AndCols; col++ ){
	mkWireArea( Arrays.andPlaneGnd, mkCol( col ),
			       RuleNames.andGndDiffusion );
    }

    /* wire transistors to ground and signal contacts,			*/
    /* and connect plane ground contacts				*/
    for( row = 0; row < Rows; row++ ){
	mkWireArea( Arrays.andTran, mkRow ( row ), RuleNames.andTranDiffusion );
	mkWireArea( Arrays.andPlaneGnd, mkRow( row ), RuleNames.andGround );
    }
    mkWireArea( Arrays.andPlaneGnd, mkRow( row ), RuleNames.andGround );

    /* OR Plane diffusion and ground					*/
    for( col = 0; col < OrCols; col++ ){
	mkWireArea( Arrays.orTran, mkCol( col ), RuleNames.orTranDiffusion );
	mkWireArea( Arrays.orOddPlaneGnd, mkCol( col ), RuleNames.orGround );
	mkWireArea( Arrays.orEvenPlaneGnd, mkCol( col ), RuleNames.orGround );
    }
    mkWireArea( Arrays.orOddPlaneGnd, mkCol( OrCols ), RuleNames.orGround );

    /* wire the power for the or plane					*/
    mkWireArea( Arrays.orPower, mkRow( 0 ), RuleNames.orGround );
    mkWireArea( Arrays.orPower, mkRow( 1 ), RuleNames.orGround );

    /* wire the or plane ground lines					*/
    for( row = 0; row < Rows; row++ ) {
	mkWireArea( Arrays.orOddPlaneGnd,
			       mkArea( row, row, 0, OddEvenBreak),
			       RuleNames.orGndDiffusion );
	mkWireArea( Arrays.orEvenPlaneGnd,
			       mkArea( row, row, OddEvenBreak, OrCols ),
			       RuleNames.orGndDiffusion );
    }
}
