#include "octopus.h"

/*
 * sanitizeMatrix.c
 *
 * This file is, with one exception, devoted to fixing the input
 * personality matrix created by Genie so it is possible to create
 * a PLA with the desired logic behavior.  The major fixes are:
 *
 * 1) Mapping 'c' cells to '.' cells and getting rid of empty rows.
 *    This is largely for convenience.
 *
 * 2) Moving complementary inputs to adjacent rows in the core.  Genie
 *    occasionally allows an input to end up folded to the opposite side
 *    of the core as its compement.  This is fixed by adding a new row
 *    in the AND plane and copying the signal's column over to the new
 *    location next to the complement.  This function is purely a Genie
 *    bug fix.
 *
 * 3) Shifting around the output column order.  This is an area-saving
 *    optomization and should remain with the program no matter what
 *    the error-status of the input.
 *
 * 4) Flipping the X coordinate system to a first quadrant representation.
 *    This is mostly a convenience to easy the interfacing with mkarray.
 *    This should change if the array folder output is changed.
 */

extern s_options Options;

/* The data as given by the input personality matrix file		*/
extern char	**GivenCore;
extern int	GivenRows, GivenAndCols, GivenOrCols;
extern s_input	**GivenInputs;
extern s_output	**GivenOutputs;

/* The data that we actually use to build the PLA			*/
extern char	**Core;
extern int	Rows, AndCols, OrCols;
extern s_input	**Inputs;
extern s_output	**Outputs;
extern int	OddEvenBreak;

/* 
 * sanitizeMatrix
 *
 * 1) remove all empty rows and map all non-transistor cells in the core
 *	to the empty cell (ei change connect cells to dot cells).
 *
 * 2) trim all outputs to only cover the minimum necessary number of rows.
 *
 * 3) If complemetary signals are not adjacent, add new columns and force
 *      them next to each other so a PLA can be created.
 *
 * 4) Arrange the output order by fold type so that ground and signal rows
 *      in the OR plane can be shared.
 *
 * 5) Flip the X-coords in the core and IO structures so the core is
 *      referenced as a first quadrent array [(0,0) to (row-1, col-1)].
 *
 * 6) Free the memory used by the GivenCore.
 */
sanitizeMatrix()
{
    mapEmptyCellsAndRmEmptyRows();

    trimGivenCoreSignals();

    sanitizeAndPlane();

    rearrangeOrPlane();

    flipCoreXCoords();

    freeGivenCore(GivenRows);
}

/*
 * mapEmptyCellsAndRmEmptyRows
 *
 * map all 'c' cells to '.' cells.
 * any rows that contain only '.' cells, remove.
 *
 * This procedure writes the the global GivenRows
 */
mapEmptyCellsAndRmEmptyRows()
{
    int		i, j, emptyRow;
    int		currentRows; /* number of rows currently ( <= GivenRows ) */

    currentRows = GivenRows;
    i = 0;
    while( i < currentRows){
	emptyRow = TRUE;
	for( j = 0; j < (GivenAndCols + GivenOrCols); ++j) {
	    if (GivenCore[i][j] == CONNECTION) {
		GivenCore[i][j] = DOT;
	    } else if (GivenCore[i][j] != DOT) {
		emptyRow = FALSE;
	    }
	}
	if (emptyRow) {
	    removeGivenCoreRow( i, &currentRows );
	    /* don't incr i since rows just shifted			*/
	} else {
	    ++i;
	}
    }
    GivenRows = currentRows;
}

/*
 * removeGivenCoreRow
 *
 * remove row row from the GivenCore, and decrement the number of rows.
 * Since rows are malloced individually, its just a matter of saving a
 * pointer to row row, shifting all rows <= row by one, freeing row row,
 * and adjusting all input and output structure values to reflect the
 * removal of the row.
 */
removeGivenCoreRow( row, numRows )
int	row;
int	*numRows;
{
    char	*theRow;
    int		i;
    s_input	*tIn;
    s_output	*tOut;

    /* remove row from core						*/
    theRow = GivenCore[row];
    for ( i = row; i < (*numRows - 1); ++i) {
	GivenCore[i] = GivenCore[i + 1];
    }
    GivenCore[*numRows - 1] = NIL(char);
    FREE(theRow);
    (*numRows)--;

    /* update inputs structure						*/
    /* psuedo code of what's going on...
     * if (bottom == row )
     *		-- bottom
     * else if (bottom > row)
     *		-- bottom
     *
     * if (top == row)
     *		do nothing
     * else if (top > row)
     *		-- top
     */
    for( i = 0; i < GivenAndCols; ++i ) {
	for ( tIn = GivenInputs[i]; tIn != NIL(s_input); tIn = tIn->next) {
	    if( tIn->col == -1 )continue;
	    if(( tIn->bottom == row )&&( tIn->top == row )){
		/* eliminate this input */
		tIn->col = tIn->top = tIn->bottom = -1;
	    } else {
		if( tIn->bottom >= row ) {
		    (tIn->bottom)--;
		}
		if( tIn->top > row ) {
		    (tIn->top)--;
		}
	    }
	}
    }

    /* update outputs structure						*/
    for( i = 0; i < GivenOrCols; ++i ) {
	for ( tOut = GivenOutputs[i]; tOut != NIL(s_output); tOut = tOut->next){
	    if(( tOut->bottom == row )&&( tOut->top == row ))
		/* eliminate this output */
		tOut->col = -1;
	    if( tOut->bottom >= row ) {
		tOut->bottom --;
	    }
	    if( tOut->top > row ) {
		tOut->top --;
	    }
	}
    }
}

/*
 * trimGivenCoreSignals
 *
 * Shrink signals across the core so that all endpoints of signals
 * fall on non-empty cells.  Its assumed that CONNECTION cells have
 * already been replaced by DOT cells.
 */
trimGivenCoreSignals()
{
    int		i;
    s_input	*tIn;
    s_output	*tOut;

    /* trim inputs							*/
    for( i = 0; i < GivenAndCols; ++i){
	if(( tIn = GivenInputs[i]) == NIL(s_input) ||
	   ( tIn->next == NIL(s_input))){
	    /* empty col or not folded					*/
	    continue;
	}
	if( tIn->col != -1){
	    while(( tIn->bottom >= 0) &&
		  ( GivenCore[ tIn->bottom ][ i ] == DOT)){
		--tIn->bottom;
	    }
	}
	tIn = tIn->next;
	if( tIn->col != -1){
	    while(( tIn->top < GivenRows - 1) &&
		  ( GivenCore[ tIn->top ][ i ] == DOT)){
		++tIn->top;
	    }
	}
#ifdef UNDEFINED
	for( tIn = GivenInputs[i]; tIn != NIL( s_input ); tIn = tIn->next) {
	    /* Take care of special cases.
	     * Due to quirks in Genie, there can be inputs attached to no
	     * transistors, which are now completely unattached,
	     * and after row removal, their starting row can end up being
	     * larger than the ending row (ie top < bottom ).
	     */
	    /*** handle special cases					*/
	    /*** note to me: might be better to handle special cases	*/
	    /*** after this, because this trimming process may (will) uncover*/
	    /*** some signals that aren't attached to anything (eg sigs on */
	    /*** completely empty cols, etc).				*/

	    /*if top isn't at top of core, incr it til its on a non-empty cell*/
	    if( tIn->top != 0) {
		while((GivenCore[tIn->top][i] == DOT)
					&& (tIn->top < GivenRows)) {
		    (tIn->top) ++;
		}
	    }
	    /*if bot isn't at bot of core, decr it til its on a non-empty cell*/
	    if( tIn->bottom != GivenRows-1) {
		while(( tIn->bottom >= 0 )&&
		      ( GivenCore[tIn->bottom][i] == DOT)){
		    (tIn->bottom) --;
		}
	    }
	}

#endif /* UNDEFINDED */
    }
    /* trim outputs							*/
    for( i = 0; i < GivenOrCols; ++i){
	if(( tOut = GivenOutputs[i]) == NIL(s_output) ||
	   ( tOut->next == NIL(s_output))){
	    /* empty col or not folded					*/
	    continue;
	}
	if( tOut->col != -1){
	    while(( tOut->bottom >= 0) &&
		  ( GivenCore[ tOut->bottom ][ i + GivenAndCols ] == DOT)){
		--tOut->bottom;
	    }
	}
	tOut = tOut->next;
	if( tOut->col != -1){
	    while(( tOut->top < GivenRows - 1) &&
		  ( GivenCore[ tOut->top ][ i + GivenAndCols ] == DOT)){
		++tOut->top;
	    }
	}
    }
}

/*
 * sanitizeAndPlane
 *
 * Expects to receive the GivenCore with all empty cells mapped to
 * the DOT cell, all empty rows removed.
 * When this procedure is done, the Core will have been created, and the
 * GivenCore will be FREEd.
 */
sanitizeAndPlane()
{
    int		newCols;

    /* switch to Inputs[] and Outputs[] (as opposed to Given...)	*/
    Inputs = GivenInputs;
    GivenInputs = NIL(s_input *);

    /* if illegal inputs, move them to legal locations			*/
    /* newCols recives the number of cols created			*/
    newCols = rearrangeInputs();

    /* set Core params and malloc memory for Core			*/
    createCore(GivenRows, GivenAndCols + newCols, GivenOrCols);

    /* copy GivenCore to Core with Inputs[] shifts, and shift Outputs	*/
    copyToCore(newCols);

    /* no longer need GivenCore						*/
}


/*
 * rearrangeInputs
 *
 * Step through Inputs[], and rearrange them if they are in illegal
 * configurations.  Illegal configurations are cases where an input
 * and its complement are not in adjacent columns and/or * do not enter
 * the core from the same side of the core.
 *
 * Assumptions:
 *    1) simple column folding (ie at most 2 inputs per column)
 *    2) if one input's complement is out of place, the complement of the other
 *       input on that column will not need to be fixed.  This is an implication
 *       of the first assumption.
 */
int
rearrangeInputs()
{
    s_input	*in, *inComp;
    int		newCols, i;

    /* set up								*/
    newCols = 0;
    for ( i = 0; i < GivenAndCols - 1; ++i) {
	in = Inputs[i + newCols];
	if(in == NIL(s_input)) {
	    continue; /* no inputs at this col, goto next column	*/
	}
	/* check top input of current col				*/
	if(( inComp = in->comp) == NIL(s_input)); /* no comp, go to bot inpt */
	else if( in->col == -1);         /* bogus input - proceed */
	else if( inComp->col < in->col); /* comp behind input: OK. goto bot in*/
	else if(( inComp->col > in->col+1) /*more than one col ahead	*/
	    || (inComp != Inputs[i+newCols+1])) { 
		   /* or its one col ahead but not top input		*/
	    ++newCols;
	    moveComplement(in, newCols);
	    continue; /* bottom input will be OK			*/
	}

	/* check bottom input at this col				*/
	if((in = in->next) == NIL(s_input)) {
	    continue; /* no bottom input				*/
	}
	if((inComp = in->comp) == NIL(s_input)) ; /* no complement	*/
	else if( in->col == -1);         /* bogus input - proceed */
	else if(inComp->col < in->col) ; /* comp behind input: OK */
	else if( (inComp->col > in->col+1) /*more than one col ahead	*/
		|| ((inComp == Inputs[i+newCols+1])
		   && (inComp->next != NIL(s_input))) ) { 
			/* or its one col ahead but its the top input	*/
			/* and not the only input at the column		*/
	    ++newCols;
	    moveComplement(in, newCols);
	}
    }

    return ( newCols );
}

/*
 * moveComplement
 *
 * In the Inputs array, move the complement of in next to in 
 * taking in to acount the newCols offset.
 */
moveComplement(in, newCols)
s_input		*in;
int		newCols;
{
    int		i;
    s_input	*inComp, *tIn;

    /* make Inputs[] big enough						*/
    Inputs = REALLOC(s_input *, Inputs, GivenAndCols + newCols);

    /* shift Inputs[] over from max to in->col + newCols 		*/
    /* Inputs[] size is GAC+nC, so max index is GAC+nC-1		*/
    for( i = GivenAndCols + newCols - 2; i >= in->col + newCols; --i){
	Inputs[i+1] = Inputs[i];
    }

    /* take in->comp out of it's list and put it in new, empty col	*/
    inComp = in->comp;
    if(( tIn = Inputs[inComp->col + newCols]) == inComp){
	Inputs[inComp->col + newCols] = inComp->next;
    } else if( tIn->next == inComp){
	tIn->next = NIL(s_input);
    } else {
	errRaise( optProgName, -1,
		"More than 2 inputs at col %d.", inComp->col);
    }

    /* Assign the comp to its new row and make sure its the only one there. */
    Inputs[in->col+newCols] = inComp;
    inComp->next = NIL(s_input);
}

/*
 * copyToCore
 *
 * Copy the core data from the GivenCore the Core.
 * This is made a little complex because some new columns have been
 * created.  The inputs in the Inputs[] have (possibly) been shifted, but
 * they still reference their old sections of the GivenCore.
 */
copyToCore(newCols)
int	newCols; /* number of new colums (for shifting Outputs[])	*/
{
    int		coreCol, givenCol, row, endRow;
    s_input	*tIn;

    /* init everything to DOT (just to be safe)				*/
    for( row = 0; row < Rows; ++row) {
	for( coreCol = 0; coreCol < AndCols+OrCols; ++coreCol) {
	    Core[row][coreCol] = DOT;
	}
    }
    /* copy over And plane by inputs					*/
    for( coreCol = 0; coreCol < AndCols ; ++coreCol) {
	for( tIn = Inputs[coreCol]; tIn != NIL(s_input); tIn = tIn->next) {
	    if( tIn->col == -1) continue;
	    givenCol = tIn->col;
	    endRow = tIn->bottom;
	    for (row = tIn->top; row <= endRow; ++row) {
		Core[row][coreCol] = GivenCore[row][givenCol];
	    }
	    tIn->col = coreCol;
	}
    }
    /* make sure cols with only one input have top and bottom 		*/
    /* covering full column						*/
    for ( coreCol = 0; coreCol < AndCols; ++coreCol) {
	tIn = Inputs[ coreCol ];
	if(( tIn != NIL(s_input)) || (tIn->col == -1)) continue;
	if( tIn->next == NIL(s_input)) {
	    tIn->top = 0;
	    tIn->bottom = Rows - 1;
	}
    }


    /* copy over Or plane directly (with offset)			*/
    for( coreCol = AndCols; coreCol < AndCols+OrCols; ++coreCol) {
	if(( givenCol = coreCol - newCols) >= (GivenAndCols + GivenOrCols)) {
	    errRaise( optProgName, -1,
		"Bad giveCol calculation, %d.", givenCol);
	}
	for ( row = 0; row < Rows; ++row) {
	    Core[row][coreCol] = GivenCore[row][givenCol];
	}
    }
}

/*
 * createCore
 *
 * set the Core paramaters, malloc Core
 */
createCore(rows, andCols, orCols)
int	rows, andCols, orCols;
{

    Rows = rows;
    AndCols = andCols;
    OrCols = orCols;

    allocCore(rows, andCols + orCols);
}

/* 
 * rearrangeOrPlane
 *
 * Arrange the or columns by fold type.  There are three types:
 *  1) Don't Care: This is when there are 0 or 1 outputs in the column,
 *       or there are 2 outputs, but they are separated (top of one to bottom
 *       of the other) by 1 or more empty rows (ie not adjacent). 
 *  2) Odd: This is when there are 2 outputs in the column that are
 *       adjacent (close folding), and the upper output ends on an odd row.
 *       All output columns of this type go on the left side of the OR plane.
 *  3) Even: 2 outputs in the columns closely folded with the upper one
 *       ending on an even column.  These go on the right.
 *
 * The Don't Cares go in between the Odds and Evens.
 * Note:
 *  The Odd and Even Cases are defined in terms of the Core, not the
 *    GivenCore.  Therefore, the fact that the core is going to be flipped
 *    later on must be taken in to account.  This is the reason for the
 *    switchType variable.  This is also why we compare the bottom of the first
 *    output to the top of the second output.  The sence of the outputs is
 *    (here) from the top of the core down.  Once the core is flipped, these
 *    outputs will be reversed and there coordiates changed so that the
 *    first output in the list is at the bottom of the core, and its TOP
 *    is then closest to the BOTTOM of the second output (it's opposite).
 */
rearrangeOrPlane()
{
    enum outputType	*outType;
    int			numDc, numEven, numOdd;
    int			dcIndex, oddIndex, evenIndex;
    int			switchType; /* switch ODD/EVEN type if even GivenRows*/
    int			row, col;
    s_output		*output, *output2;
    int			slop, lookahead;

    /* classify the folded outputs as ODD, EVEN, or DC			*/
    outType = ALLOC( enum outputType, OrCols );
    numDc = numEven = numOdd = 0;
    /* if the number of rows (GivenRows - 1) is even, when the core is	*/
    /* flipped (mirrored in x),	the sence of odd and even will flip,	*/
    /* so assign the outputs the complemetary type here.		*/
    switchType =  GivenRows % 2;
    for( col = 0; col < GivenOrCols; col++ ){
	/* if ( not folded || not closely folded )			*/
	if(((output = GivenOutputs[col]) == NIL(s_output))
		|| ((output2 = output->next) == NIL(s_output))
		|| (output->bottom < output2->top - 1)) {
	    outType[ col ] = DONT_CARE;
	    numDc++;
	} else if ( output->bottom % 2) {
	    if( switchType ) {
		outType[ col ] = EVEN;
		numEven++;
	    } else {
		outType[ col ] = ODD;
		numOdd++;
	    }
	} else {
	    if( switchType ) {
		outType[ col ] = ODD;
		numOdd++;
	    } else {
		outType[ col ] = EVEN;
		numEven++;
	    }
	}
    }
    /* reorder the output columns					*/
    dcIndex = evenIndex = oddIndex = 0;
    Outputs = ALLOC( s_output *, OrCols );
    for( col = 0; col < OrCols; col++ ){
	switch( outType[ col ] ){
	    case ODD:
		Outputs[oddIndex++] = GivenOutputs[col];
		break;
	    case DONT_CARE:
		Outputs[numOdd + dcIndex++] = GivenOutputs[col];
		break;
	    case EVEN:
		Outputs[numOdd + numDc + evenIndex++] = GivenOutputs[col];
		break;
	}
    }
    /* correct the personality matrix by replacing GivenCore[row][col]	*/
    /* with Core[ row ][ output->col ] for each output, fix output->col	*/
    for( col = 0; col < OrCols; ++col ){
	if((output = Outputs[col]) == NIL( s_output)) {
	    continue;
	}
	for( row = 0; row < Rows; ++row ){
	    Core[ row ][ col + AndCols ] = GivenCore[ row ][ output->col ];
	}
	for( ; output != NIL ( s_output ); output = output->next ){
	    output->col = col + AndCols;
	}
    }

    /*** this may not be correct */
    /* find a good place to break between odd and even outputs		*/
    OddEvenBreak = -1;
    slop = ( OrCols - 2 * ( Options.orGndLineSep/2 ))%Options.orGndLineSep;
    if( Options.debug > 0 )printf("od=%d,dc=%d,ev=%d,sl=%d\n",numOdd,numDc,numEven,slop);
    for( col = Options.orGndLineSep/2 ; col < numOdd + numDc; 
	 col += Options.orGndLineSep ){
	for( lookahead = 0; lookahead <= slop; lookahead++ ){
	    if( col + lookahead >= numOdd ) {
		OddEvenBreak = col + lookahead;
		break;
	    }
	}
	if( OddEvenBreak >= 0 )break;
    }
    /* not much thought given to this (ie, could be improved, possibly)	*/
    if( OddEvenBreak == -1 ) {
	if ( numOdd > 0) {
	    OddEvenBreak = numOdd; /* between Odd and Dc		*/
	} else if( numDc > 0) {
	    OddEvenBreak = numDc; /* between Dc and Even		*/
	} else {
	    /* do all OR ground on even rows				*/ 
	    OddEvenBreak = 0;
	}
    }

    FREE( outType);
}

/*
 * flipCoreXCoords
 *
 * Flip the x-coordinates of the Core.  This is necessary because
 * Genie defines the the 0 row to be the top row, and the mkarray package
 * (and most everyone else) defines the bottom row as 0.
 */
flipCoreXCoords()
{
    int		i;
    s_input	*input, *reverseInputs();
    s_output	*output, *reverseOutputs();

    /* flip row references in Inputs[]					*/
    for( i = 0; i < AndCols; ++i) {
	for( input = Inputs[i]; input != NIL(s_input); input = input->next) {
	    input->top = Rows - input->top - 1;
	    input->bottom = Rows - input->bottom - 1;
	}
	Inputs[i] = reverseInputs( Inputs[i] );
    }

    /* flip row references in Outputs[] 				*/
    for( i = 0; i < OrCols; ++i) {
	for(output = Outputs[i]; output != NIL(s_output);output = output->next){
	    output->top = Rows - output->top - 1;
	    output->bottom = Rows - output->bottom - 1;
	}
	Outputs[i] = reverseOutputs( Outputs[i] );
    }

    /* flip the Core (recursively)					*/
    flipCore( 0 );
}

/*
 * reverseInputs
 *
 * reverse an input list of any length
 */
s_input *
reverseInputs( head )
s_input		*head;
{
    s_input	*input, *list;

    if( head == NIL(s_input)) {
	return ( NIL(s_input));
    } else {
	list = head->next;
	head->next = NIL(s_input); /* new tail of list			*/
	while( list != NIL(s_input)) {
	    input = list;
	    list = list->next;
	    input->next = head;
	    head = input;
	}
	return( head );
    }
}

/*
 * reverseOutputs
 *
 * reverse an output list of any length
 */
s_output *
reverseOutputs( head )
s_output	*head;
{
    s_output	*output, *list;

    if( head == NIL(s_output)) {
	return ( NIL(s_output));
    } else {
	list = head->next;
	head->next = NIL(s_output); /* new tail of list			*/
	while( list != NIL(s_output)) {
	    output = list;
	    list = list->next;
	    output->next = head;
	    head = output;
	}
	return( head );
    }
}

/*
 * flipCore
 *
 * switch one row of the core and call recFlipCore to switch the rest
 */
flipCore( row )
int	row;
{
    char	*head;

    if( row < Rows ) {
	head = Core[ row ];
	flipCore( row + 1);
	Core[ Rows - row - 1 ] = head;
    }
}
