/*
 * acs.c:  Adaptive Cooling Schedule.
 *   By Fabio Romeo
 */


#include "defs.h"
#include "global.h"
#include <X11/Xlib.h>
#include "xg.h"

#define  DELTA(a,b)     ((a) != 0.0 ? ( (double) ((a) - (b)) / (double) (a) ) : MAX_DOUBLE)
#define  A_DELTA(a,b)   ( ABS(DELTA(a,b)) )


#define TOO_MANY		1
#define MODIFY_REFERENCE 	1

#define MAX_DOUBLE              1.7014118346046923171e+38


struct annPar{
	int		minNum;
	int		maxNum;
	int		finAttCell;
	double		minCost;
	double		maxTdecr;
	double		alpha;
	double 		p;
	double		delta;
	double		lambda;
	double		finTestTol;
	double		K;
};

typedef struct procInfo{
    int		exit;
    int		numMoves;
    double		totalCost;
    double		average;
    double		average2;
    double		maxDeltaCost;
    double		maxCost;
    double		minCost;
} PROCINFO, *PTOPROC;



static double lastCost;
static double refAve, refSigma, refT;
static struct annPar *acsParam;
static struct procInfo **procInfo;
extern SHARED double SM_T;
extern SHARED double SM_startT;
static int numProcessors;
static double prevAve, prevSig;
static double extAve, extSigma;

static XGraph xgACS;		/* To plot some parameters */
static XGraph xgACS1;		/* To plot some parameters */


int acs_exitInnerLoop(newstatus)
    int newstatus;
{
    static int status = FALSE;
    
    if (newstatus != -1) status = newstatus;
    return(status);
}

int acs_exitOuterLoop(newstatus)
    int newstatus;
{
    static int status = FALSE;
    
    if (newstatus != -1) status = newstatus;
    return(status);
}



int acs_innerLoop(deltaCost, accepted, procIndex)
    double deltaCost;
    int accepted;
    int procIndex;
{
    double x, s;
    int status;

    if( accepted ) procInfo[procIndex]->totalCost += deltaCost;

    procInfo[procIndex]->numMoves++;

    procInfo[procIndex]->average += procInfo[procIndex]->totalCost;
    procInfo[procIndex]->average2 += procInfo[procIndex]->totalCost*procInfo[procIndex]->totalCost;
    procInfo[procIndex]->maxDeltaCost = MAX( procInfo[procIndex]->maxDeltaCost, ABS(deltaCost) );
    procInfo[procIndex]->minCost = MIN(procInfo[procIndex]->minCost,
				       procInfo[procIndex]->totalCost );
    procInfo[procIndex]->maxCost = MAX(procInfo[procIndex]->maxCost,
				       procInfo[procIndex]->totalCost );


    if( SM_T == 0.0 ){
	status = (procInfo[procIndex]->numMoves > acsParam->finAttCell);
	return( acs_exitInnerLoop(status) );
    } 

    x = procInfo[procIndex]->average / procInfo[procIndex]->numMoves;
    s = ( procInfo[procIndex]->average2 - x * x * procInfo[procIndex]->numMoves ) 
		    / ( procInfo[procIndex]->numMoves );

    if( SM_T == INFINITE_TEMPERATURE ){
	if( A_DELTA(x,prevAve) < .05 && A_DELTA(s,prevSig) < .05 ){
	    status = TRUE;
	} else {
	    prevAve = x;
	    prevSig = s;
	    status = FALSE;
	}
	return( acs_exitInnerLoop(status) );
    }

    if( procInfo[procIndex]->numMoves < acsParam->minNum ) {
	return( acs_exitInnerLoop( FALSE ) );
    }
    if( procInfo[procIndex]->numMoves > acsParam->maxNum){
	procInfo[procIndex]->exit = TOO_MANY;
	return( acs_exitInnerLoop( TRUE ) );
    }
    if( x > ( extAve + acsParam->delta * extSigma) || x < ( extAve - acsParam->delta * extSigma) ){
	return( acs_exitInnerLoop( FALSE ) );
    }
    return( acs_exitInnerLoop( TRUE ) );
} 



/*ARGSUSED*/
acs_newTemperature(totalCost)
    double totalCost;
{
    double expon;
    double Tdecr;
    double sigma, sigma2;
    double average = 0;
    double average2 = 0;
    int numMoves = 0;
    int exit = OK;
    int i;

    for( i = 0 ; i < numProcessors ; i++ ) {
	numMoves += procInfo[i]->numMoves;
	average += procInfo[i]->average * procInfo[i]->numMoves;
	average2 += procInfo[i]->average2 * procInfo[i]->numMoves * procInfo[i]->numMoves ;
	if( procInfo[i]->exit == TOO_MANY ) exit = MODIFY_REFERENCE;
	if( procInfo[i]->minCost < acsParam->minCost ) {
	    exit = MODIFY_REFERENCE;
	    acsParam->minCost = procInfo[i]->minCost;
	}

    }
    if (numMoves <= 0) {
	Warning("ACS:Too few moves to determine new temperature");
	return ;
    }
    average /= (double) numMoves;
    average2 /= (double) numMoves;
    sigma2 = ( average2 - average * average );
    sigma = sqrt(sigma2);

    if( sigma2 == 0. ) {
	Tdecr = acsParam->maxTdecr;
    } else {
	expon = - ( acsParam->lambda * SM_T ) / sigma;
	Tdecr = MAX( acsParam->maxTdecr, exp(expon) );
    }
    if( exit == MODIFY_REFERENCE || SM_T == INFINITE_TEMPERATURE ){
	if (sigma == 0.0) {
	    Warning("ACS: anomalous condition: sigma == 0.0");
	} else {
	    refAve = average;
	    refSigma = sigma;
	    refT = SM_T;
	    acsParam->alpha = ( average - acsParam->minCost ) / (sigma2);
	    acsParam->p = acsParam->alpha * ( average - acsParam->minCost );
	}
    }


    xgSendPoint(xgACS, SM_T, Tdecr , "Eff. Alpha");
    xgSendPoint(xgACS1, SM_T, (double) numMoves , "moves");

    if( SM_T == INFINITE_TEMPERATURE ){
	SM_startT = SM_T = sigma * acsParam->K ;
	acs_setScheduleDefaults(average, sigma, sigma2, numMoves);
    } else {
	SM_T *= Tdecr;
    }
    extParGamma();
    

    /* Now determine if it is time to exit */

    {
      double maxDeltaCost = 0.;
      double minCost = 0.;
      double maxCost = 0.;
      double totalCost = 0.;

      for( i = 0 ; i < numProcessors ; i++ ){
	maxDeltaCost += (double) procInfo[i]->maxDeltaCost;
	maxCost += (double) procInfo[i]->maxCost;
	minCost += (double) procInfo[i]->minCost;
	totalCost += (double) procInfo[i]->totalCost;
	procInfo[i]->totalCost = totalCost;
	procInfo[i]->average = 0.0;
	procInfo[i]->average2 = 0.0;
	procInfo[i]->numMoves = 0;
	procInfo[i]->exit = OK;
	procInfo[i]->maxDeltaCost = 0.0;
	procInfo[i]->maxCost = 0.0;
	procInfo[i]->minCost = MAX_DOUBLE;
      }

      maxDeltaCost /= (double) numProcessors;
      maxCost /= (double) numProcessors;
      minCost /= (double) numProcessors;
      totalCost /= (double) numProcessors; /* ??? */

      if( SM_T == 0.0 ){
	if( lastCost == totalCost ) {
	  acs_exitOuterLoop(TRUE);
	} else {
	  lastCost = totalCost;
	  acs_exitOuterLoop(FALSE);
	}
      } else if (ABS((double) maxDeltaCost) >= acsParam->finTestTol * ( maxCost - minCost ) ){
	SM_T = 0.0;
	acs_exitOuterLoop(TRUE);
      } else {
	acs_exitOuterLoop(FALSE);
      }

    }
    return;
}


acs_setScheduleDefaults(average, sigma, sigma2, numMoves)
    double average;
    double sigma;
    double sigma2;
    int numMoves;
{
    numMoves /= numProcessors;

    refAve = average;
    refSigma = sigma;
    refT = SM_T ;
    acsParam->minCost = MAX(average - 4.0 * sigma , 0.0) ;
    acsParam->minNum = numMoves / 5;
    acsParam->maxNum = numMoves * 5;
    acsParam->finAttCell = numMoves / 5;
    acsParam->alpha = ( average - acsParam->minCost ) / (sigma2);
    acsParam->p = acsParam->alpha * ( average - acsParam->minCost );
}



acs_globalInit(num,totalCost)
    int num;			/* number of processors */
    double totalCost;
{
    static int initialized = FALSE;
    int i;

    if (!initialized) {
	numProcessors = num;
	acsParam = SH_MALLOC(struct annPar, 1);
	procInfo = SH_MALLOC(PTOPROC, numProcessors);
	for( i = 0 ; i < numProcessors ; i++ ){
	    procInfo[i] = SH_MALLOC( PROCINFO, 1);
	}
	xgACS = xgNewWindow(display, "puppy", "A.C.S. parameters", "Y","T",220,220,750,0);
	xgACS1 = xgNewWindow(display, "puppy", "A.C.S. parameters", "Y","T",220,220,750,220);
	initialized = TRUE;
    } else if (numProcessors != num) {
	FatalError("ACS: Cannot change number of processors on the fly");
    }

    acsParam->maxTdecr = .75;
    acsParam->delta = .5;
    acsParam->lambda = .5;
    acsParam->finTestTol = 1.0;
    acsParam->K = 5;
    for( i = 0 ; i < numProcessors ; i++ ) {
	procInfo[i]->exit = OK ;
	procInfo[i]->average = 0.0;
	procInfo[i]->average2 = 0.0;
	procInfo[i]->numMoves = 0;
	procInfo[i]->maxDeltaCost = 0.0;
	procInfo[i]->maxCost = 0.0;
	procInfo[i]->minCost = MAX_DOUBLE;
	procInfo[i]->totalCost = totalCost;
    }
    SM_T = SM_startT = INFINITE_TEMPERATURE;
}



extParGamma()
{
    double a;
    double coeff1, coeff2;

    a = acsParam->alpha;
    if( refT == -1) {
	coeff1 = a / ( SM_T * a + 1 );
	coeff2 = SM_T * coeff1;
    } else if( refT == 0.) {
	return;
    } else {
	coeff2 = ( refT * a + 1 ) * SM_T / ( SM_T * a + 1 ) / refT ;
	coeff1 = ( 1 - SM_T / refT ) * coeff2 / SM_T;
    }
    extAve = refAve - refSigma * refSigma * coeff1;
    extSigma = refSigma * coeff2;
    return;
}
