/*
 *  FILE: analogOctIO.c 
 *
 *  DESCRIPTION: Interface with OCT data-base.
 *
 *
 *  CREATION DATE: Oct.21.1990          AUTHOR: Edoardo Charbon
 *
 *  FUNCTIONS:
 *  int         readDataFromOct()
 *  int         readSymmetry()
 *  int         readMatching()
 *  int         readBias()
 *          
 *
 *  EXPORTED    FUNCTIONS:
 *  int         readAnalogDataFromOct() 
 *  
 *  IMPORTED    FUNCTIONS:
 *  int         linkSymmetricCellsToCell() <-- analogLink.c
 *
 */


#include "defs.h"
#include "global.h"
#include "sa.h"
#include "analogVariables.h"
#include "octPlacement.h"
#include "st.h"
#include "tr.h"
#include "fang.h"
#include "harpoon.h"
#include "oh.h"
#include "oct.h"
#include "vov.h"

/* IMPORTED GLOBAL VARIABLES AND STRUCTURES */


/* GLOBAL VARIABLES IN THIS MODULE ("analogOctIO.c") */
int cellAccumulator;
int clusterAccumulator;

/* MAX_NUMBER_OF_SYMMETRIES = 16 <-- defined in "analogVariable.h" */


/* IMPORTED FUNCTIONS */
extern int
linkSymmetricCellsToCell();
extern int
linkMatchedCellsToCell();

/* FUNCTION DECLARATION */
DiffusionType
globalmoduleWellType();
DiffusionType
moduleWellType();
DiffusionType
searchLeafsInModule();


/* FUNCTIONS */
int
readAnalogDataFromOct()
{
  if (!considerSymmetry)  {
    (void) sprintf(errmsg,"Symmetry constraints not considered");
    Message (errmsg);
  } else {
    /* read the symmetries from the Oct data-base and put them into an internal 
     * structure called gxSymCell for horizontal symmetries and gySymCell for
     * for vertical symmetries */
    readSymmetry();
 
    /* link the structure just created (gSymCell), containing the analog
     * circuit symmetry constraints with the existing macroCells */
    linkSymmetricCellsToCell();
  }

  if (!considerMatching)  {
    (void) sprintf(errmsg,"Matching constraints not considered");
    Message (errmsg);
  } else { 
    /* read the matched cells from the Oct data-base and put them into an internal 
     * structure called gMatchCell */ 
    readMatching();
 
    /* link the structure just created (gMatchCell), containing the analog
     * circuit matching constraints with the existing macroCells */
    linkMatchedCellsToCell();
  }

  if (debug) Message("Analog Constraints reading: done"); 
  return(OK);
}


/* this routine reads the data concerning vertical and horizontal symmetry 
 * and put it in a datastructure which is called gxSymCell (horizontal 
 * symmetry) and gxSymCell (vertical symmetry) */

int 
readSymmetry()

{
  octObject           Symmetry;
  int                 noxSymmetry = FALSE;
  int                 noySymmetry = FALSE;

  gySymCell = SH_MALLOC (SymmetricCell, 1);
  gySymCell->axisType = VERTICAL;
  gySymCell->numberOfAxis = 0;
  gxSymCell = SH_MALLOC (SymmetricCell, 1);
  gxSymCell->axisType = HORIZONTAL;  
  gxSymCell->numberOfAxis = 0;
 
  Message("Looking for Symmetry Constraints");
  Symmetry.type = OCT_BAG;
  Symmetry.contents.bag.name = "YSYMMETRIES";
  noySymmetry = readSymmetricInstance(&Symmetry, gySymCell);
  Symmetry.contents.bag.name = "XSYMMETRIES";
  noxSymmetry = readSymmetricInstance(&Symmetry, gxSymCell);
 
  if (noySymmetry && noxSymmetry)  {
    (void) sprintf(errmsg,"neither horizontal nor vertical symmetry axes found");
    Message (errmsg);
    considerSymmetry = FALSE;
  } else {
    (void) sprintf(errmsg,"%d vertical and %d horizontal symmetric cells found",
		   gySymCell->numberOfSymCell, gxSymCell->numberOfSymCell);
    Message (errmsg);
    /* considerSymmetry = TRUE; */
  }
  return OK;
}


int
readSymmetricInstance(SymmetryBag, SymCell)
      octObject           *SymmetryBag;
      SymmetricCell       *SymCell;
    
{
  octObject           SymBag, CoupleBag, PropBag;
  octObject           SymmetricInstance;
  octStatus           symStatus;
  octGenerator        genBag, genInstance, genSym;
  octId               idCell2;
  int                 i, l, m, symCellAccumulator;
  int                 maxNumberOfCouples, maxNumberOfAxes;
  octId               pt[3];
  MacroCell           macro1, macro2, macro;
  SymmetricCell       Cell;
  int                 *numberOfBags;
  int                 noSymmetry = FALSE, singleFlag = FALSE;
  int                 noCoupleFlag = FALSE;  

  symCellAccumulator = 0;
  maxNumberOfCouples = MAX_NUMBER_OF_COUPLES; /* defined in analogVariables.h */
  maxNumberOfAxes = MAX_NUMBER_OF_SYMMETRIES; /* defined in analogVariables.h */

  if (octGetByName(&Chip,SymmetryBag) == OCT_NOT_FOUND)  {
    noSymmetry = TRUE;
    SymCell->numberOfSymCell = 0;
    return noSymmetry;
  } 
  OH_ASSERT( octInitGenContents(SymmetryBag,OCT_BAG_MASK,&genSym) );
  m = 0;
  /* search bags "SYM" inside  the SYMMETRIES bag */
  while ((symStatus = octGenerate(&genSym,&SymBag)) == OCT_OK)  {
     /* search the position of the axis which is contained in a prop */
     symStatus = ohGetByPropName(&SymBag, &PropBag, "POSITION");
     if (symStatus == OCT_NOT_FOUND)   
       FatalOctError("position of one or more Axes missing");
     if (symStatus != OCT_OK)
       FatalOctError("the property of the axis has a wrong name"); 
     SymCell->Axis[m].axisPosition = PropBag.contents.prop.value.integer;
     OH_ASSERT( octInitGenContents(&SymBag,OCT_BAG_MASK,&genBag) );
     l = 0;
     /* search couple bags "COUPLE" inside a bag SYM */
     while ((symStatus = octGenerate(&genBag,&CoupleBag)) == OCT_OK)  { 
       if (debug)
         Message("A couple of symmetric objects found");
       if (octInitGenContents(&CoupleBag,OCT_INSTANCE_MASK,&genInstance) != OCT_OK)
         Warning("The couple Bag is corrupted");
       i = 0;
       /* search instances inside a couple */
       while ((symStatus = octGenerate(&genInstance,&SymmetricInstance)) == OCT_OK)  {   
         if (debug)  {
           sprintf(errmsg,"Found cell %s in symmetry data-base", SymmetricInstance.contents.instance.name);
           Message(errmsg);
         }
         i++;
         pt[i]=SymmetricInstance.objectId;
         if (i > 2)
           FatalError("Illegal number of instances per Couple");
       }
       octFreeGenerator(&genInstance);
       symCellAccumulator += i;
       SymCell->Axis[m].couple[l].cell1 = SH_MALLOC (MacroCell, 1);  
       SymCell->Axis[m].couple[l].cell1->cellId = pt[1];
       SymCell->Axis[m].couple[l].axisPosition = SymCell->Axis[m].axisPosition;
       SymCell->Axis[m].couple[l].axisType = SymCell->axisType;
         
       switch(i) {  
       case 1 : 
         Warning("There is an instance on the symmetry axis,");
         SymCell->Axis[m].couple[l].cell2 = NULL;
         singleFlag = TRUE;
         l++;
         break;
       case 2 :
         SymCell->Axis[m].couple[l].cell2 = SH_MALLOC (MacroCell, 1);
         SymCell->Axis[m].couple[l].cell2->cellId = pt[2];
         l++;
         break;
       default :  
         SymCell->Axis[m].couple[l].cell1 = NULL;
	 SymCell->Axis[m].couple[l].cell2 = NULL;
	 Warning("Couple without cells found. Ignored");
         noCoupleFlag = TRUE;
       }
       if (l > maxNumberOfCouples)  {
         sprintf(errmsg, "more than %d couples of symmetric cells found", 
		 maxNumberOfCouples); 
         FatalError(errmsg);
       }

       singleFlag = FALSE;
       noCoupleFlag = FALSE;
       SymCell->Axis[m].numberOfCouple = l;
     }

     octFreeGenerator(&genBag);
     m++;
     if (m > maxNumberOfAxes)  {
       sprintf(errmsg,"more than %d symmetry axes found",maxNumberOfAxes);
       FatalError(errmsg);
     } 
  }

  octFreeGenerator(&genSym);  
  SymCell->numberOfSymCell = symCellAccumulator;
  SymCell->numberOfAxis = m;
  sprintf(errmsg, "%d Axes, %d Cells found for one direction", SymCell->numberOfAxis, SymCell->numberOfSymCell);
  Message(errmsg); 
  return noSymmetry;
}


/* this routine reads the data concerning matching constraints 
 * and put it in a data-structure which is called gMatchCell */

int
readMatching()

{
  octObject           Matching;
  int                 noMatching = FALSE;

  gMatchCell = SH_MALLOC (MatchedCell, 1);
  
  Message("Looking for Matching Constraints");
  Matching.type = OCT_BAG;
  Matching.contents.bag.name = "MATCHING";
  noMatching = readMatchedInstance(&Matching, gMatchCell);
     
  if (noMatching)  {
    (void) sprintf(errmsg,"no matching constraints found");
    Message (errmsg);
    considerMatching = FALSE;
  } else {
    (void) sprintf(errmsg,"%d matched cells found (with repetition)", gMatchCell->numberOfMatchCell);
    Message (errmsg);
    /* considerMatching = TRUE; */
  }
  if (debug)
    scanMatching(gMatchCell->matchFirstCluster);
  return OK;
}


/* procedure that scans the database gMatchCell */
/* notice the recursion inside it */
 
scanMatching(MatchCell) 
    struct MatchingCluster *MatchCell;

{
  struct MatchingCluster *Cell;
  int clusterFirstIndex = 0;

  Cell = MatchCell;
  scanInnerLoop(Cell, clusterFirstIndex);
}

 
int
scanInnerLoop(cluster, clusterIndex)
   struct MatchingCluster *cluster;
   int                clusterIndex;

{
  int  cellIndex;                         /* index */
  struct MatchingCluster  *child;   /* visited branches */ 
  MacroCell                *cell;

  while (cluster != NULL)  {
    cellIndex = 0;
    cell = cluster->MatchingFirstCell; 
    if (cell != NULL) 
      clusterIndex++;
    while (cell != NULL)  {
      sprintf(errmsg,"Cluster %d, index %d, matched cell %d, linked cell %d (X) %d, distance %d", clusterIndex, cellIndex, cell->cellId, cell->matchLink->cellId, cell->matchLink->curCenter.x, cluster->ClusterLimitDistance);
      Message(errmsg);
      cell = cell->matchNextCell;
      cellIndex++;
    }
    child = cluster->Cluster;
    clusterIndex = scanInnerLoop(child, clusterIndex);
    cluster = cluster->matchNextCluster;
  }
  return clusterIndex;
}


int
readMatchedInstance(MatchingBag, Cell)
      octObject    *MatchingBag;
      MatchedCell         *Cell;

{
  int noMatching = FALSE; 

  if (octGetByName(&Chip,MatchingBag) == OCT_NOT_FOUND)  {
    noMatching = TRUE;
    Cell->numberOfMatchCell = 0;
    Cell->matchFirstCluster = NULL;
    return(noMatching);
  } else {
    Cell->matchFirstCluster = SH_MALLOC (struct MatchingCluster, 1); 
    cellAccumulator = 0;      /* global variable of this module */
    searchHierarchicalMatching(MatchingBag, Cell->matchFirstCluster);
    Cell->numberOfMatchCell = cellAccumulator;
    return(noMatching);
  } 
}


searchHierarchicalMatching(Bag, MatchCell)
      octObject                     *Bag;
      struct MatchingCluster  *MatchCell;
      
{
  octObject           MatchBag, ClusterBag, PropBag;
  octObject           MatchedInstance;
  octStatus           matchStatus;
  octGenerator        genInstance;
  int                 attachedCellAccumulator; /* Cells directly attached to a "MATCH" bag */
  octGenerator        genMatch;
  MacroCell           *attachedCell;
  int                 noMatching = FALSE;
  int                 noClusterFlag = FALSE;  
   
  OH_ASSERT( octInitGenContents(Bag,OCT_BAG_MASK,&genMatch) );
  /* search bags "MATCH" inside a MATCH bag */
  while ((matchStatus = octGenerate(&genMatch,&MatchBag)) == OCT_OK)  {
    clusterAccumulator += 1;   /* every new cluster increments clusterAccumulator */ 
    /* read the cluster limit distance which is contained in a prop */
    matchStatus = ohGetByPropName(&MatchBag, &PropBag, "DISTANCE");
    if (matchStatus == OCT_NOT_FOUND)       
      FatalOctError("Limit distance of one cluster is missing");
    if (matchStatus != OCT_OK)
      FatalOctError("the property of the cluster has a wrong name");
    MatchCell->ClusterLimitDistance = PropBag.contents.prop.value.integer;
    attachedCell = NULL;
    attachedCellAccumulator = 0;
    OH_ASSERT( octInitGenContents(&MatchBag,OCT_INSTANCE_MASK,&genInstance) );
    /* search instances directly attached to "MATCH" bag. */
    while ((matchStatus = octGenerate(&genInstance,&MatchedInstance)) == OCT_OK)  {
      MatchCell->MatchingFirstCell = SH_MALLOC(MacroCell,1); 
      MatchCell->MatchingFirstCell->cellId = MatchedInstance.objectId;
      MatchCell->MatchingFirstCell->matchNextCell = attachedCell;
      attachedCell = MatchCell->MatchingFirstCell;
      cellAccumulator += 1;
      attachedCellAccumulator += 1;
#ifdef OBSOLETE_TO_BE_DELETED
sprintf(errmsg, "cell %d found; name %s, distance %d", MatchCell->MatchingFirstCell->cellId, MatchedInstance.contents.instance.name, MatchCell->ClusterLimitDistance);
Message(errmsg);
#endif
    }
    octFreeGenerator(&genInstance);
    MatchCell->Cluster = SH_MALLOC (struct MatchingCluster, 1);
    MatchCell->numberOfMatchingCell = attachedCellAccumulator;
    searchHierarchicalMatching(&MatchBag, MatchCell->Cluster);
    MatchCell->matchNextCluster = SH_MALLOC(struct MatchingCluster,1);
    MatchCell = MatchCell->matchNextCluster;
  }
  octFreeGenerator(&genMatch);
  return OK; 
}


/* this procedure assignes the right biasing status to the macro cell 
 * according to the information contained in the OCT data-base.
 * A cell can be described as self-biased or independent (if it is a
 * complex device made of p- and n-channel or pnp and npn transistors).
 * If no description is present, the program looks for the diffusions of
 * the device: if a p-diffusion is found, the device is considered an n-type
 * (npn) transistor and therefore Vss-biased, otherwise it is assumed of p-type
 * (pnp) and therefore Vdd-biased. */

readBias(instance)
      octObject *instance;

{
  octObject           BiasBag;
  BiasStatus          biasStatus;
  octObject           SelfBag;
  octObject           SelfBiasedInstance;
  octStatus           selfBiasStatus;
  octGenerator        genBag, genInstance;
  octGenerator        genBias;

  BiasBag.type = OCT_BAG;
  BiasBag.contents.bag.name = "SELFBIAS";
  
  /* search of the bag "SELFBIAS" */
  if (octGetByName(&Chip, &BiasBag) == OCT_NOT_FOUND) {
    biasStatus = moduleWellType(instance);
    return biasStatus;
  } else {
    OH_ASSERT( octInitGenContents(&BiasBag,OCT_BAG_MASK,&genBias) );
    /* search bags "SELF" attached to "SELFBIAS" */
    while ((selfBiasStatus = octGenerate(&genBias,&SelfBag)) == OCT_OK)  {
      OH_ASSERT( octInitGenContents(&SelfBag,OCT_INSTANCE_MASK,&genInstance) );
      /* search instances attached to "SELF" bag. */
      while ((selfBiasStatus = octGenerate(&genInstance,&SelfBiasedInstance)) == OCT_OK)  {

#ifdef OBSOLETE_TO_BE_DELETED
sprintf(errmsg,"sefbiased instansce name: %s",SelfBiasedInstance.contents.instance.name);
Message(errmsg);
#endif

        if (SelfBiasedInstance.objectId == instance->objectId) {
          sprintf(errmsg, "self biased cell found; name %s", SelfBiasedInstance.contents.instance.name);
          Message(errmsg);
          biasStatus = SELF;
          return biasStatus;
        }
      }
      octFreeGenerator(&genInstance);
    }
    octFreeGenerator(&genBias);
    biasStatus = moduleWellType(instance);
    return biasStatus;
  } 
}
  

#define  NDIF_LAYER    "NDIF"
#define  PDIF_LAYER    "PDIF"

/* 
 *  This function returns the N/P type of a LEAF or MODULE cell
 */

DiffusionType
moduleWellType(inst)
    octObject *inst;       /*  Instance to analyze  */
{
  octObject       interf;  /* Interface instance */
  octObject         item;  /* Property or geometry */
  int nDiffusion = FALSE;  /* flag that indicates if an n-diffusion cell already appeared */
  int pDiffusion = FALSE;  /* flag that indicates if a p-diffusion cell already appeared */
  DiffusionType diffusionType; 

/*   Open the interface facet  */
  
  VOVinputFacet( &interf );
  VOVoutputFacet( &interf );
  ohOpenMaster(&interf, inst, "interface", "r");

  /*   Look for the CELLTYPE property. It must be MOS  */
  if(ohGetByPropName(&interf, &item, "CELLTYPE") == OCT_NOT_FOUND) {
    sprintf(errmsg, "No CELLTYPE property in cell %s:%s. Cell is assumed independently biased", 
	    interf.contents.facet.cell, interf.contents.facet.view);
    Warning(errmsg);
    octCloseFacet(&interf);
    return NEITHER_TYPE;
  }

  if(strcmp(item.contents.prop.value.string, "MOS") != 0) {
    octCloseFacet(&interf);
    return NEITHER_TYPE;
  }

  diffusionType = searchLeafsInModule(&nDiffusion, &pDiffusion, &interf, inst);
  return diffusionType; 
}



/* recursive function to search a module's structure */

DiffusionType
searchLeafsInModule(nDiff, pDiff, interface, instance)
    int           *nDiff;
    int           *pDiff;
    octObject *interface;
    octObject  *instance;  
  
{
    octObject        layer, item;
    octObject     *Interf, *Inst;
    octGenerator         genInst;
    octObject            InstBag;
    octStatus         instStatus;
  
    /* Look for the CELLCLASS property. It must be LEAF or MODULE  */
    if(ohGetByPropName(interface, &item, "CELLCLASS") == OCT_NOT_FOUND) {
	sprintf(errmsg, "No CELLCLASS property in cell %s:%s. Cell is assumed independently biased", interface->contents.facet.cell, interface->contents.facet.view);
	Warning(errmsg);
	return NEITHER_TYPE;
    }

    if (strcmp(item.contents.prop.value.string, "LEAF") == 0) { 

	/*   NDIF and no PDIF ---> NDIF type  */
	if((ohGetByLayerName(interface, &layer, NDIF_LAYER) != OCT_NOT_FOUND) &&
	   (octGenFirstContent(&layer, OCT_POLYGON_MASK, &item) == OCT_OK)) {
	    if((ohGetByLayerName(interface, &layer, PDIF_LAYER) == OCT_NOT_FOUND) ||
	       (octGenFirstContent(&layer, OCT_POLYGON_MASK, &item) != OCT_OK)) {
		octCloseFacet(interface);
		if (*pDiff) {
		    return NEITHER_TYPE;
		} else {
		    *nDiff = TRUE;
		    return NDIF_TYPE;
		} 
	    }
	}
	/*   PDIF and no NDIF ---> PDIF type  */
	if((ohGetByLayerName(interface, &layer, PDIF_LAYER) != OCT_NOT_FOUND) &&
	   (octGenFirstContent(&layer,OCT_POLYGON_MASK, &item) == OCT_OK)) {
	    if((ohGetByLayerName(interface, &layer, NDIF_LAYER) == OCT_NOT_FOUND) ||
	       (octGenFirstContent(&layer,OCT_POLYGON_MASK, &item) != OCT_OK)) {
		octCloseFacet(interface);
		if (*nDiff) {
		    return NEITHER_TYPE;
		} else {
		    *pDiff = TRUE;
		    return PDIF_TYPE;
		} 
	    }
	}
    }
  
    if( strcmp(item.contents.prop.value.string, "MODULE") == 0) {
	if (ohGetOrCreateBag(instance, &InstBag, "INSTANCES") == OCT_NOT_FOUND) {
	    FatalOctError("the module does not contain instances");
	} 
	OH_ASSERT( octInitGenContents(&InstBag,OCT_INSTANCE_MASK,&genInst) );
	/* search instances inside  the INSTANCES bag */
	while ((instStatus = octGenerate(&genInst,Inst)) == OCT_OK)  {
	    /* Open the interface facet  */
	    ohOpenMaster(Interf, Inst, "interface", "r");
	    /* search the leafs inside the instance */
	    searchLeafsInModule(nDiff, pDiff, Interf, Inst);
	}  
	octFreeGenerator(&genInst);
	*nDiff = TRUE;
	*pDiff = TRUE;
	return NEITHER_TYPE;
    }
}




















