/*
 *  FILE: compCost.c
 *
 *  DESCRIPTION: computation of cost functions.
 *
 *
 *  CREATION DATE: Aug.21.1987          AUTHOR: Andrea Casotto 
 *
 *  FUNCTIONS:
 *  int         configurationCost()
 *  int         printCost()
 *  int         moveCost()
 *  int         halfPerimeter(box)
 *  int         curChiparea()
 *  int         newChipArea(moveP)
 *
 *
 *  EXPORTED    FUNCTIONS:
 *  int         configurationCost()
 *
 *  IMPORTED    FUNCTIONS:
 *  int         curSymmetryCost() <- analogCost.c
 *  int         curMatchingCost() <- analogCost.c
 *  int         MoveSymmetryCost() <- analogCost.c
 *  int         MoveMatchingCost() <- analogCost.c
 *
 */

#include "defs.h"
#include "global.h"
#include "sa.h"
#include "analogVariables.h" 
#include "constraintDrivenVariables.h"
#include "parall.h"

/* IMPORTED GLOBAL VARIABLES AND STRUCTURES */


/* IMPORTED FUNCTIONS */
extern int
CurSymmetryCost();
extern int
CurMatchingCost();
extern int
CurWellCost();
extern int
MoveSymmetryCost();
extern int
MoveMatchingCost();
extern int
MoveWellCost();


/*
 *	configurationCost()
 *	compute the cost of a configuration
 *      To be used with parsimony, very slow.
 *      Affects the followiong variables:
 *      totOverlap, wireLength.
 *      RETURN cost (double).
 *              confArea
 *              confOverlap
 *              confWireLength
 *              confClusterCost
 */

int
configurationCost( costP )
     Cost *costP;

{
  double TotalCost();

  if (debug) Message("Computing placement cost");
  CurWireLengthCost(costP);     /* Wire Length should come before Overlap */
  CurOverlapCost(costP);
  CurAreaCost(costP);
  CurClusterCost(costP);
  if (analogMode) {
    CurSymmetryCost(costP);
    CurMatchingCost(costP);
    wellBoundary = (int) ((chipBbox->lowerLeft.y + chipBbox->upperRight.y)/2);
    CurWellCost(costP);
    CurConstraintCost(costP);
  } else {
    costP->symmetry = 0.0;
    costP->matching = 0.0;
    costP->well = 0.0;
    costP->constraint = 0.0;
  }
  TotalCost(costP);
  return (OK);
}

int
printCost(costP,s)
    Cost *costP;                /*  */
    char *s;                    /* Comment string */
{
  printf("\n *** %s ***\n", s);
  printf("Overlap      %11.2f\n" , costP->overlap);
  printf("Area         %11.2f\n" , costP->area);
  printf("Wire Length  %11.2f\n" , costP->wireLength);
  if (analogMode)  {
    printf("Symmetry     %11.2f\n" , costP->symmetry);
    printf("Matching     %11.2f\n" , costP->matching);
    printf("Well         %11.2f\n" , costP->well);
    printf("Constraints  %11.2f\n" , costP->constraint);
  }
  if (enableSubnets == TRUE) {
    printf("(Subnets)\n");
  } else {
    printf("(Bounding boxes)\n");
  }
  printf("Violations   %11d\n",    costP->conViol);
  printf("*** TOTAL %11.2f\n", costP->total);
}

double
TotalCost(costP)
    Cost        *costP;
{
  /*
   * the total cost is the (weighted) sum of
   * the extimated wirelength , the total
   * overlapping area and the chip area
   */
  costP->total = (costP->overlap    * SM_ovWeight +
                  costP->area       * SM_areaWeight +
                  costP->wireLength * SM_netWeight +
                  costP->conViol    * SM_conViolWeight +
                  costP->matching   * SM_matchWeight +
                  costP->symmetry   * SM_symWeight +
                  costP->well       * SM_wellWeight +
		  costP->constraint * SM_constraintWeight);

  return costP->total;
}


/*
 *	moveCost( moveP)
 *	it returns the current cost that might be affected by this cell
 *	if it is moved according to move. If move = CC_INTERCHANGE or BB_INTERCHANGE then
 *	cell2 points to the other cell interchanged 
 *       if move = CHANGE_OWNERSHIP procNumber contains the number 
 *       of the processor that gets tha cell.
 *	the total cost affected by the two cells
 *	that is the estimated wirelength of the nets with
 *	terminals on this cell and the overlapping of this cell
 *	and the chip area.
 *	>>>> if a reference is done to a cell that doesn't belong
 *	to myself, increment the refCounter of the cell (NYI)
 */
double 
moveCost(moveP)
    Move *moveP;		/* descriptor of the current move */
{
  double TotalCost();
  
  zeroesCost(&moveP->cost);
  newMoment(moveP);		/* compute the new momenta */
  if (moveP->type != CHANGE_OWNERSHIP) {
    MoveWireLengthCost(moveP);
    if (analogMode) {
      MoveConstraintCost(moveP);
    } else {
      moveP->cost.constraint = 0.0;
    }
    MoveOverlapCost(moveP);
    MoveAreaCost(moveP);
    if (analogMode) {
      MoveSymmetryCost(moveP);
      MoveMatchingCost(moveP);
      MoveWellCost(moveP);
      
    } else {
      moveP->cost.symmetry = 0.0;
      moveP->cost.matching = 0.0;
      moveP->cost.well = 0.0;
    }
  } else {
    MoveClusterCost(moveP);
  }

  if (moveP->type == CHANGE_OWNERSHIP) {
    moveP->cost.total = moveP->cost.cluster * SM_clusterWeight;
  } else {
    TotalCost(&(moveP->cost));
  }
  return moveP->cost.total;
}


/*
 *	halfPerimeter(box)
 *	returns the halfPerimeter of box
 */
int
halfPerimeter(boxP)
     octBox *boxP;
{
    return((boxP->upperRight.x-boxP->lowerLeft.x) + (boxP->upperRight.y-boxP->lowerLeft.y));
}

/*
 *	curChipArea()
 */
double
curChipArea()
{
  octCoord left,right,top,bottom;

  left  = gFirstCell->lyoNext[PUPPY_L]->curBbox.lowerLeft.x;
  right = gFirstCell->lyoNext[PUPPY_R]->curBbox.upperRight.x;
  top   = gFirstCell->lyoNext[PUPPY_T]->curBbox.upperRight.y;
  bottom= gFirstCell->lyoNext[PUPPY_B]->curBbox.lowerLeft.y;

  return ((double) (right-left) * (double)(top-bottom));
}

printChipBBox()
{
  octCoord left,right,top,bottom;

  left  = gFirstCell->lyoNext[PUPPY_L]->curBbox.lowerLeft.x;
  right = gFirstCell->lyoNext[PUPPY_R]->curBbox.upperRight.x;
  top   = gFirstCell->lyoNext[PUPPY_T]->curBbox.upperRight.y;
  bottom= gFirstCell->lyoNext[PUPPY_B]->curBbox.lowerLeft.y;
  
  printf("(%d,%d) (%d,%d) : %dx%d\n", left,bottom,right,top, right-left, top-bottom); 
}

/*
 * newChipArea()
 *	after cell has been moved, this routine checks if the area
 *	of the chip has changed
 */
double
newChipArea(moveP)
     Move *moveP;
{
  MacroCell *cell;
  MacroCell *cell2;
  octCoord left,right,top,bottom;
  struct   MacroCell	*edgeCell;

  cell = moveP->cell;
  cell2 = moveP->cell2;

  /* initialize the temporary variables */
  left = cell->newBbox.lowerLeft.x;
  right= cell->newBbox.upperRight.x;
  bottom = cell->newBbox.lowerLeft.y;
  top = cell->newBbox.upperRight.y;

  switch (moveP->type) {
  case TRANSLATION:
  case ALIGNMENT:
  case RL_TRANSLATION:
  case TRANSLATE_AND_TRANSFORM:
  case TRANSFORMATION:
    /****** PUPPY_L   ****/
    if (cell != gFirstCell->lyoNext[PUPPY_L]) {
      edgeCell = gFirstCell->lyoNext[PUPPY_L];
    } else {
      edgeCell = cell->lyoNext[PUPPY_L]; /* take next cell in list */
    }
    left  = MIN (cell->newBbox.lowerLeft.x, edgeCell->curBbox.lowerLeft.x);

    /****** PUPPY_R   ****/
    if (cell != gFirstCell->lyoNext[PUPPY_R]) {
      edgeCell = gFirstCell->lyoNext[PUPPY_R];
    } else {
      edgeCell = cell->lyoNext[PUPPY_R];
    }
    right  = MAX (cell->newBbox.upperRight.x, edgeCell->curBbox.upperRight.x);

    /****** PUPPY_B   ****/
    if (cell != gFirstCell->lyoNext[PUPPY_B]) {
      edgeCell = gFirstCell->lyoNext[PUPPY_B];
    } else {
      edgeCell = cell->lyoNext[PUPPY_B];
    }
    bottom  = MIN (cell->newBbox.lowerLeft.y, edgeCell->curBbox.lowerLeft.y);

    /****** PUPPY_T   ****/
    if (cell != gFirstCell->lyoNext[PUPPY_T]) {
      edgeCell = gFirstCell->lyoNext[PUPPY_T];
    } else {
      edgeCell = cell->lyoNext[PUPPY_T];
    }
    top  = MAX (cell->newBbox.upperRight.y, edgeCell->curBbox.upperRight.y);
    break;
  case CC_INTERCHANGE:
  case BB_INTERCHANGE:
    /* It must work also when there are only two cells to be placed */
    /****** PUPPY_L   ****/
    edgeCell = gFirstCell->lyoNext[PUPPY_L];
    while (edgeCell == cell OR edgeCell == cell2) {
      edgeCell=edgeCell->lyoNext[PUPPY_L];
    }
    if (edgeCell != NULL) {
      left  = MIN (cell->newBbox.lowerLeft.x, edgeCell->curBbox.lowerLeft.x);
      left  = MIN (cell2->newBbox.lowerLeft.x, left );
    } else {
      left =  MIN (cell->newBbox.lowerLeft.x,cell2->newBbox.lowerLeft.x);
    }

    /****** PUPPY_R   ****/
    edgeCell = gFirstCell->lyoNext[PUPPY_R];
    while (edgeCell == cell OR edgeCell == cell2) {
      edgeCell=edgeCell->lyoNext[PUPPY_R];
    }
    if (edgeCell != NULL) {
      right  = MAX (cell->newBbox.upperRight.x, edgeCell->curBbox.upperRight.x);
      right  = MAX (cell2->newBbox.upperRight.x, right);
    } else {
      right  = MAX (cell2->newBbox.upperRight.x, right);
    }

    /****** PUPPY_B   ****/
    edgeCell = gFirstCell->lyoNext[PUPPY_B];
    while (edgeCell == cell OR edgeCell == cell2) {
      edgeCell=edgeCell->lyoNext[PUPPY_B];
    }
    if (edgeCell != NULL) {
      bottom  = MIN (cell->newBbox.lowerLeft.y, edgeCell->curBbox.lowerLeft.y);
      bottom  = MIN (cell2->newBbox.lowerLeft.y, bottom);
    } else {
      bottom  = MIN (cell2->newBbox.lowerLeft.y, bottom);
    }

    /****** PUPPY_T   ****/
    edgeCell = gFirstCell->lyoNext[PUPPY_T];
    while (edgeCell == cell OR edgeCell == cell2) {
      edgeCell=edgeCell->lyoNext[PUPPY_T];
    }
    if (edgeCell!=NULL) {
      top  = MAX (cell->newBbox.upperRight.y, edgeCell->curBbox.upperRight.y);
      top  = MAX (cell2->newBbox.upperRight.y, top);
    } else {
      top  = MAX (cell2->newBbox.upperRight.y, top);
    }
    break;
  default:
    break;
  }
  return ((double) (right-left)* (double)(top-bottom));
}


zeroesCost(costP)
    Cost *costP;
{
  costP->overlap = 0.0;
  costP->wireLength = 0.0;
  costP->area = 0.0;
  costP->cluster = 0.0;
  costP->conViol = 0;
 
  costP->matching = 0.0;
  costP->symmetry = 0.0;
  costP->well = 0.0;
  costP->constraint = 0.0;

  costP->total = 0.0;

}


accumulateCosts(acc, c)
    Cost *acc;
    Cost *c;
{
  acc->overlap  += c->overlap;
  acc->wireLength  += c->wireLength;
  acc->area  += c->area;
  acc->cluster  += c->cluster;
  acc->conViol += c->conViol;
  if (analogMode) {  
    acc->matching += c->matching;
    acc->symmetry += c->symmetry;
    acc->well += c->well;
    acc->constraint += c->constraint;
  }
  acc->total  += c->total;
}


accumulateAbsCosts(acc, c)
    Cost *acc;
    Cost *c;
{
  acc->overlap  += ABS(c->overlap);
  acc->wireLength  += ABS(c->wireLength);
  acc->area  += ABS(c->area);
  acc->cluster  += ABS(c->cluster);
  acc->conViol += ABS(c->conViol);
  if (analogMode)  {
    acc->matching += ABS(c->matching);
    acc->symmetry += ABS(c->symmetry);
    acc->well += ABS(c->well);
    acc->constraint += ABS(c->constraint);
  }
  acc->total  += ABS(c->total);
}


#define MAX_SQUARABLE_NUMBER 1E19
/* TIME CRITICAL ROUTINE !! */
double
squareWithCare(f)
    double f;
{
  if (f < 0.0) f = -f;
  if (f < MAX_SQUARABLE_NUMBER)  return( f * f);

  Warning("Too big to be squared");
  return(0.0);
}

/* There are two ways to square a number
 * One is fast, the other is prudent 
 */
#define SPEED			/* Choose the fast one */
#ifdef SPEED
#define SQUARE(x) (x) * (x)
#else
#define SQUARE(x) squareWithCare(x)
#endif


/* TIME CRITICAL ROUTINE !! */
accumulateCosts2(acc, c)
    Cost *acc;
    Cost *c;
{
  double squareWithCare();

  acc->overlap  += SQUARE (c->overlap);
  acc->wireLength  += SQUARE( c->wireLength);
  acc->area  += SQUARE( c->area);
  acc->cluster  += SQUARE( c->cluster);
  acc->conViol += SQUARE( (double) c->conViol);
  if (analogMode)  {
    acc->matching += SQUARE( c->matching);
    acc->symmetry += SQUARE( c->symmetry);
    acc->well += SQUARE( c->well);
    acc->constraint += SQUARE( c->constraint);
  } 
  acc->total  += SQUARE( c->total);
}



















