/*
 *  FILE: constraintDriven.c
 *
 *  DESCRIPTION:
 *
 *
 *  CREATION DATE: Mar.24.1991          AUTHOR: Edoardo Charbon
 *
 *
 *  FUNCTIONS: CurConstraintCost()
 *             MoveConstraintCost() 
 *             readConstraintDrivenFile()
 *             assignConstraintsToNets()
 *
 *  IMPORTED    FUNCTIONS:
 *
 *
 */


#include "defs.h" 
#include "global.h"
#include "sa.h"
#include "constraintDrivenVariables.h"
#include "vov.h"


/* IMPORTED VARIABLES */
extern double floatingNetWeight;   /* <-- "costWL.c */ 

/* GLOBAL VARIABLES OF THIS MODULE ("constraintDriven.c") */
double specificTopCap;
double specificBottomCap;
double capRatio = 1.0;
double slopeRatio; 
int numberOfNet;
int numberOfPerformance;

/* DECLARED ROUTINES USED IN THIS FILE THAT DO NOT RETURN INTEGERS */
double
computeTotalConstraintCost();
int
computeLength();
char
*readConstraintDrivenFile();
void  
transferArrayToStructure();
NetConstraint
*netHasConstraint();
void
linkConstraintToNet();
void
linkNetCouple();
void
addConstraintToNet();
int
assignConstraintsToNets();
Net
*searchSymmetricNet();
Net
*findNet();
int
lengthOfNetWithMovingTerminal();
int
lengthOfNetAttachedToMovingCells();
double
computeCapToGND();
Sensitivity
computeSumOfDeltaWmax();
void
scanCD();

/* GLOBAL CONSTANTS FOR "constraintDriven.c" */
static int       curFlag = FALSE;
static int       newFlag = TRUE;

/* IMPORTED FUNCTIONS */
extern int distance();   /* <-- "costWL.c */ 
extern MacroCell *searchOtherSymmetricCell();   /* <-- "analogCost.c" */
extern int halfPerimeter(); /* <-- "compCost.c" */
extern void curNetBBox(); /* <-- "netBbox.c" */
extern void newNetBBox(); /* <-- "netBbox.c" */

/* IMPORTED MACROS
 * Macro function that returns the rectified value of an integer point number.
 * #define  RECT(a)             ((a) < 0 ? 0 : (a))
 * defined in "defs.h"
 */


/* This procedure scans the data-base gConstraintDriven and computes the current
 * cost of critical nets according to their sensitivity */

int
CurConstraintCost(cost)
     Cost *cost;
{

  Move *move;

  move = NULL;
  cost->constraint = 0.0;
  if (!considerConstraintDriven || SM_constraintWeight <=  0.0) return OK;

  cost->constraint = computeTotalConstraintCost(move, gConstraintDriven, curFlag);

}


double
computeTotalConstraintCost(move, constraintDriven, moveFlag)
     Move *move;
     ConstraintDriven *constraintDriven;
     int moveFlag;
{

  ConstraintDriven      *constr;
  Net                   *net, *net2;
  int			estimatedLength = 0, e = 0;
  octBox		BBox;
  int			i, k;	         /* indices */
  Sensitivity           *SP, *S;
  double                cost = 0.0, temporaryCost;
  double                length;
  Sensitivity           *deltaWmax;

  constr = constraintDriven;
  deltaWmax = constr->deltaWmax;

  SP = MALLOC (Sensitivity, numberOfPerformance);
  for (k = 0; k < numberOfPerformance; k++) 
    SP[k].plus = SP[k].minus = 0.0;

  for (i = 0; i < numberOfNet; i++) {
    net = constr->netConstraint[i].net;
    S = constr->netConstraint[i].S; 

    if (moveFlag) {
      estimatedLength = computeLength(net, move);
    } else { 
      curNetBBox(net,&BBox);
      estimatedLength = halfPerimeter(&BBox);
#ifdef OBSOLETE
      if (strcmp(net->name, "net12") == 0) {
	sprintf(errmsg, "-------------estim of %s = %d", net->name, estimatedLength); 
	Message(errmsg);
	sprintf(errmsg,"capacita' = %f", specificBottomCap * estimatedLength);
	Warning(errmsg);
      }
#endif
    }
   
    net2 = net->matchedNet;
    if (net2 != NULL) {
      if (moveFlag) {
	e = computeLength(net2, move);
	estimatedLength += e;
      } else { 
	curNetBBox(net2,&BBox);
	estimatedLength += halfPerimeter(&BBox);
      }
      length = (double)(estimatedLength) / 2;
    } else {
      length = (double)(estimatedLength);
    }

    for (k = 0; k < numberOfPerformance; k++) {
      if (deltaWmax[k].plus >= 0.0)
	SP[k].plus  += length * specificBottomCap * S[k].plus;
      if (deltaWmax[k].minus >= 0.0)
	SP[k].minus += length * specificBottomCap * S[k].minus;
    }
  }
  for (k = 0; k < numberOfPerformance; k++) {
    if (deltaWmax[k].plus >= 0.0) {
      temporaryCost = 
	RECT (SP[k].plus - deltaWmax[k].plus) / capRatio * deltaWmax[k].plus;
      if (temporaryCost < 1) {
	cost += temporaryCost;
      } else {
	cost += temporaryCost * slopeRatio;
      }
    }
    if (deltaWmax[k].minus >= 0.0) {
      temporaryCost = 
	RECT (SP[k].minus - deltaWmax[k].minus) / capRatio * deltaWmax[k].minus;
      if (temporaryCost < 1) {
	cost += temporaryCost;
      } else {
	cost += temporaryCost * slopeRatio;
      }
    }
  } 
  FREE (SP);
  return cost;

}


int
computeLength(net, move)
Net *net;
Move *move;
{

  int estimatedLength = 0;
  octBox BBox;

  switch (move->type) {
  case NO_MOVE:
  case SUBNET_MOVE:
    curNetBBox(net,&BBox);
    estimatedLength = halfPerimeter(&BBox);
    break;
  case TERMINAL_MOVE:
    if (move->term->net == net) {
      estimatedLength = lengthOfNetWithMovingTerminal(move->term);
    } else {
      curNetBBox(net,&BBox);
      estimatedLength = halfPerimeter(&BBox);
    }
    break;  
  case TRANSLATION:
  case RL_TRANSLATION:
  case TRANSFORMATION:
  case TRANSLATE_AND_TRANSFORM:
  case ALIGNMENT:
  case CC_INTERCHANGE:
  case BB_INTERCHANGE:
  case MULTIPLE_CELL:
  case CHANGE_OWNERSHIP:
  case NUMBER_OF_MOVE_TYPES:	
    estimatedLength = lengthOfNetAttachedToMovingCells(move, net);
    break;
  }
  return estimatedLength;
  
}
   

int 
MoveConstraintCost(moveP)
     Move *moveP;
{

  Net *net;

  moveP->cost.constraint = 0.0;

  if (!considerConstraintDriven || SM_constraintWeight == 0.0) return OK;

  moveP->cost.constraint = ( computeTotalConstraintCost(moveP, gConstraintDriven, newFlag) - 
                             computeTotalConstraintCost(moveP, gConstraintDriven, curFlag) );

}


/* scan every net attached to the cells under movemnt. If a net is equal to the given net
 * "netP", its estimated wire-length is returned, if no nets are found returns 0. 
 */ 

int
lengthOfNetAttachedToMovingCells(moveP, netP)
     Move *moveP;
     Net  *netP;
{
  
  struct netList *n1, *n2;
  Net *net;
  octBox BBox;
  int  nel = 0;

  n1 = moveP->cell->net0.next;
  if (moveP->cell2) {
    n2 = moveP->cell2->net0.next;
  } else {
    n2 = NULL;
  }

  while (n1 != NULL || n2 != NULL) {
    
    if (n1 == NULL) {		/* list 1 is finished */
      net = n2->net; n2 = n2->next; /* take net in list 2 */
    } else if (n2 == NULL) {	/* list 2 is finished */
      net = n1->net; n1 = n1->next; /* take net in list 1 */
    } else if (n1->net->number < n2->net->number) { /* net in list 1 less than net in list 2 */
      net = n1->net; n1 = n1->next; /* take net in list 1 */
    } else if (n2->net->number < n1->net->number) { /* net in list 1 less than net in list 2 */
      net = n2->net; n2 = n2->next; /* take net in list 2 */
    } else {			/* Same net. Advance both pointers */
      net = n1->net; n1 = n1->next; n2 = n2->next;
    }

    if (net == netP) {
      newNetBBox(net,&BBox);

#ifdef OBSOLETE
fprintf(stderr, "******************(%d,%d,%d,%d)************************TROVATO-----\n",BBox.lowerLeft.x,BBox.lowerLeft.y,BBox.upperRight.x,BBox.upperRight.y);
#endif

      nel = halfPerimeter(&BBox);
      return nel;
    } 
  }   
  net = netP;
  curNetBBox(net,&BBox);

#ifdef OBSOLETE
fprintf(stderr, "-e-e-e-e-e-e-e-e-e(%d,%d,%d,%d)-e-e-e-e-e-e: non TROVATO -0----------SCEMO\n",BBox.lowerLeft.x,BBox.lowerLeft.y,BBox.upperRight.x,BBox.upperRight.y);
#endif

  nel = halfPerimeter(&BBox);
  return nel;
}


/* Find semi-perimeter difference betw. cur and new BBox of the net to which "termP" belongs, 
 * assuming that "termP" was the only displaced terminal inside the net.
 * Find BBox without the moving term and compare it with new term position */

int
lengthOfNetWithMovingTerminal(termP)
    Term *termP;
{
  
  Net *net;
  struct termList *termPtr;
  octCoord left, bottom, right, top;
  octBox curBBox, newBBox;
  int nel = 0.0;
  octPoint	*newAbsCoordP;	/* of a floating terminal */

  net = termP->net;
  if (net->size > 1) {
    termPtr = net->term0.next;
    /* Skip the first if this is the moving terminal */
    if (termPtr->term == termP) {
      
      termPtr = termPtr->next;
      if (termPtr == NULL) {
	(void) sprintf(errmsg, "Inconsistent data for net %s:\n 1 term found instaed of 2",
		       net->name);
	FatalError(errmsg);
      }
    }
    left   =   termPtr->term->curAbsCoordP->x;
    bottom =   termPtr->term->curAbsCoordP->y;
    right  =   termPtr->term->curAbsCoordP->x;
    top    =   termPtr->term->curAbsCoordP->y;
  
    while ((termPtr=termPtr->next) != NULL) {
      if (termPtr->term != termP ) {
	left   = MIN(left,   termPtr->term->curAbsCoordP->x);
	bottom = MIN(bottom, termPtr->term->curAbsCoordP->y);
	right  = MAX(right,  termPtr->term->curAbsCoordP->x);
	top    = MAX(top,    termPtr->term->curAbsCoordP->y);
      }
    }
    newAbsCoordP =  &termP->cell->floatingSites[termP->newSite].curAbsCoord; 
    newBBox.lowerLeft.x =  MIN(left,   newAbsCoordP->x);
    newBBox.lowerLeft.y =  MIN(bottom, newAbsCoordP->y);
    newBBox.upperRight.x = MAX(right,  newAbsCoordP->x);
    newBBox.upperRight.y = MAX(top,    newAbsCoordP->y);
    
    nel = halfPerimeter(&newBBox);
  }
  return nel;

}


Sensitivity
computeSumOfDeltaWmax(constr)
ConstraintDriven *constr;
{

  Sensitivity sum;
  int k;
 
  sum.plus = sum.minus = 0.0;
  if (constr == NULL) return sum;

  for( k = 0; k < constr->numberOfPerformance; k++) {
    sum.plus += constr->deltaWmax[k].plus;
    sum.minus += constr->deltaWmax[k].minus;
  }
  return sum;

}


double
computeCapToGND(metalLevel)
     MetalType metalLevel;
{

  double cap;

  if (metalLevel == FIRST_MET) {
    cap = SM_bottomCapMod0 + SM_minNetWidth * SM_bottomCapMod1;
  } else {
    cap = SM_topCapMod0 + SM_minNetWidth * SM_topCapMod1;
  }
  return cap;

}

    
char
*readConstraintDrivenFile(cellName)
     char     *cellName;
{

  FILE 			*fp;
  char 			input[40];
  char			parPuppy[32];
  BOOLEAN               first;
  Stringa               *array;
  int                   k; 


  /* file existence check */

  strncpy(parPuppy, cellName, 32 - strlen(" .CDpuppy "));
  *strchr(parPuppy, ':') = '\0';
  strcat(parPuppy, ".CDpuppy");
  if (( fp=fopen(parPuppy , "r")) == (FILE *) NULL) {
    (void) sprintf(errmsg,"Cannot find %s",parPuppy);
    Message(errmsg);
    strcpy(parPuppy,"CDpuppy");
    if (( fp=fopen(parPuppy , "r")) == (FILE *) NULL) {
      (void) sprintf(errmsg,"Cannot even find %s",parPuppy);
      Message(errmsg);
      strcpy(parPuppy, "no");
      considerConstraintDriven = FALSE;
      return parPuppy;
    }
  }
    
  /* scan file and put its values in an array */

  k = 0;
  first = TRUE;
  VOVinput( VOV_UNIX_FILE, parPuppy);
  while (fscanf(fp , " %s " , input) ==1) {
    /* The "!" sign at the beginning of a line indicates a comment in the parPuppy file */
    if  (strcmp( input , "!" ) == 0) { 
      fscanf(fp , " %s " , input);
      if ( strcmp (input, "number") == 0 ) {
	fscanf(fp , " %s " , input);
	if ( strcmp (input, "of") == 0 ) {
	  fscanf(fp , " %s " , input);
	  if ( strcmp (input, "matched") == 0 ) {
	    fscanf(fp , " %s " , input);
	    if ( strcmp (input, "pairs") == 0 ) {
	      REALLOC (array, Stringa, k + 1);
	      strcpy(array[k].str, input);
	      k++;
	    }
	  }
	}
      } else {
	while ( getc(fp) != '\n') ;	/*   COMMENT: go on to the next line */
      }
    } else {
      if (first) {
	array = MALLOC (Stringa, 1);
	first = FALSE;
      } else {
	REALLOC (array, Stringa, k + 1);
      }
      strcpy(array[k].str, input);
      k++;
    }
  }     

  if (k < 1) {
    strcpy(parPuppy, "no");
    considerConstraintDriven = FALSE;
    return parPuppy;
  }
  transferArrayToStructure(array);
  FREE (array);
  return parPuppy;
}


/* transfer the contents of "array" to a data structure "constraintDriven" */

void  
transferArrayToStructure(array)
  Stringa *array;
{
 
  int			integer; /*  */
  int			k, i;
  double 		real;	/*  */
  BOOLEAN               netFlag, performanceFlag;
  BOOLEAN               deltaWFlag, sensitivityFlag;
  BOOLEAN               groundFlag, matchedNets = FALSE;
  int                   performanceNumber = 0;
  int                   netNumber = 0, bias = 0, step = 0;
  int                   pairBias = 0;
  int                   pairNumber = 0;
  char                  *net1, *net2;

  gConstraintDriven = SH_MALLOC (ConstraintDriven, 1);
  gConstraintDriven->numberOfNet = netNumber = atoi (array[0].str);
  gConstraintDriven->numberOfPerformance = performanceNumber = atoi (array[1].str);
  gConstraintDriven->numberOfMatchedPair = 0;
  gConstraintDriven->deltaWmax = SH_MALLOC (Sensitivity, performanceNumber);

  for (k = 0; k < performanceNumber; k++ ) {
    gConstraintDriven->deltaWmax[k].plus = 
      atof (array[DATA_CONST + COUPLE_CONST * k].str);
    gConstraintDriven->deltaWmax[k].minus = 
      atof (array[DATA_CONST + COUPLE_CONST * k + 1].str);
  }

  bias = DATA_CONST + COUPLE_CONST * performanceNumber;
  step = 2 * NET_CONST + SKIP_CONST + COUPLE_CONST * performanceNumber;
  gConstraintDriven->netConstraint = SH_MALLOC (NetConstraint, netNumber);

  for (k = 0; k < netNumber; k++) {
    net1 = strsav( array[bias + GROUND_CONST + k * step].str);
    net2 = strsav( array[bias + NET_CONST + k * step].str);
    if (strcmp(net1, net2) == 0) {
      sprintf(errmsg, "Constraint on nets with the same name (%s). Ignored", net1);
      Message(errmsg);
      strcpy(gConstraintDriven->netConstraint[k].groundNetName, "");
      strcpy(gConstraintDriven->netConstraint[k].netName, "");
      continue;
    }  
    if (strcmp(net1, "0") == 0) {
      if (checkNetExistence(net1, net2) == OK) {
	strcpy(gConstraintDriven->netConstraint[k].groundNetName, net1);
	strcpy(gConstraintDriven->netConstraint[k].netName, net2);
      } else {
	sprintf(errmsg, "Net %s or %s does not exist. Constraint ignored", net1, net2);
	Warning(errmsg);
	strcpy(gConstraintDriven->netConstraint[k].groundNetName, "");
	strcpy(gConstraintDriven->netConstraint[k].netName, "");
	continue;
      }
    } else if (strcmp(net2, "0") == 0) {
	if (checkNetExistence(net1, net2) == OK) {
	    strcpy(gConstraintDriven->netConstraint[k].netName, net1);
	    strcpy(gConstraintDriven->netConstraint[k].groundNetName, net2);
	} else {
	    sprintf(errmsg, "Net %s or %s does not exist. Constraint ignored", net1, net2);
	    Warning(errmsg);
	    strcpy(gConstraintDriven->netConstraint[k].groundNetName, "");
	    strcpy(gConstraintDriven->netConstraint[k].netName, "");
	    continue;
	}
    } else {
	sprintf(errmsg, "Constraint with no ground net. Ignored", 0);
	Warning(errmsg);
	strcpy(gConstraintDriven->netConstraint[k].groundNetName, "");
	strcpy(gConstraintDriven->netConstraint[k].netName, "");
	continue;
    }
    gConstraintDriven->netConstraint[k].S = SH_MALLOC (Sensitivity, performanceNumber);
    for (i = 0; i < performanceNumber; i++) {
      gConstraintDriven->netConstraint[k].S[i].plus = 
	atof (array[bias + NET_CONST + SKIP_CONST + FIRST_CONST + k * step + COUPLE_CONST * i].str); 
      gConstraintDriven->netConstraint[k].S[i].minus = 
	atof (array[bias + NET_CONST + SKIP_CONST + SECOND_CONST + k * step + COUPLE_CONST * i].str);
    }
  }
  pairBias = bias + step * netNumber;
  if (strcmp(array[pairBias].str, "pairs") == 0) {
    pairBias++;
    pairNumber = atoi(array[pairBias].str);
    gConstraintDriven->matchedNets = MALLOC (MatchedNets, pairNumber);
    gConstraintDriven->numberOfMatchedPair = pairNumber; 
 
    for (k = 0; k < pairNumber; k++) {
      strcpy (gConstraintDriven->matchedNets[k].net1Name,
	      array[pairBias + COUPLE_CONST * k + 1].str);
      strcpy (gConstraintDriven->matchedNets[k].net2Name,
	      array[pairBias + COUPLE_CONST * k + 2].str);
    }

  } else {
    gConstraintDriven->matchedNets = NULL;
    gConstraintDriven->numberOfMatchedPair = 0;
  }
  if (debug) scanCD(gConstraintDriven);

}


int
checkNetExistence(net1, net2)
     char *net1, *net2;
{

  Net *net;
  int flag1 = FALSE;
  int flag2 = FALSE;

  if (strcmp(net1, "0") == 0) flag1 = TRUE;
  if (strcmp(net2, "0") == 0) flag2 = TRUE;
  
  net = netSet->globalNext;
  while(net != NULL) {
      if (strcmp(net->name, net1) == 0) flag1 = TRUE;
      if (strcmp(net->name, net2) == 0) flag2 = TRUE;
      if (flag1 && flag2) return OK;
      net = net->globalNext;
  }
  return NOT_OK;

}


void
scanCD(constraintDriven)
     ConstraintDriven *constraintDriven;
{

  int i, k;                             /* indices */
  int perfNumber = 0, netNumber = 0, pairNumber = 0;
  ConstraintDriven *constr;
  NetConstraint *netConstraint;
  MatchedNets *matchedNets;

  constr = constraintDriven;

  if (constr == NULL) {
    fprintf(stderr, "The variable constraintDriven is empty\n");
    return;
  }

  perfNumber = constr->numberOfPerformance;
  netNumber = constr->numberOfNet;
  pairNumber = constr->numberOfMatchedPair;

  fprintf(stderr, "Performances = %d, Nets %d\n", perfNumber, netNumber);

  for (i = 0; i < perfNumber; i++) {
    fprintf (stderr, "Performance %d: deltaW[%d]+ = %3.10f, deltaW[%d]- = %3.10f\n", 
	     i, i, constr->deltaWmax[i].plus, i, constr->deltaWmax[i].minus);
  }

  netConstraint = constr->netConstraint;
  for (k = 0; k < netNumber; k++) {
    if (strcmp(netConstraint[k].netName, "") != 0) {
      fprintf(stderr, "Net %s: ", netConstraint[k].netName);
      for (i = 0; i < perfNumber; i++) {
	fprintf (stderr, "S[%d]+ = %3.10f, S[%d]- = %3.10f ", 
		 i, netConstraint[k].S[i].plus,
		 i, netConstraint[k].S[i].minus);
      }
      fprintf (stderr, "\n");
    }
  }
  if (pairNumber > 0) {
    fprintf(stderr, "Matched nets: \n");
    matchedNets = constr->matchedNets;
    for (k = 0; k < pairNumber; k++) {
      fprintf(stderr, "%s and ", matchedNets[k].net1Name);
      fprintf(stderr, "%s\n", matchedNets[k].net2Name);
    }
  }	

}


/* This routine assignes the corresponding sensitivities to the constrained nets,
 * i.e. to those nets, whose names appear in the constraint list "constraintDriven". 
 * The same sensitivities are asigned to the corresponding symmetric nets wrt one 
 * symmetry axis. Only VERTICAL symmetry axes are considered so far.
 */  

assignConstraintsToNets()
{

  Net	        *net, *net2;
  NetConstraint *netConstr;
  ConstraintDriven *constr;
  int		i, constraintCounter = 0;	     /* index */
  SymmetryType  symmetry;

  if (SM_netWeight <=  0.0) return;

  Message ("Constraint-driven placement of analog circuits");
  
  constr = gConstraintDriven;
  symmetry = VERTICAL;
  specificTopCap = computeCapToGND(SECOND_MET);
  specificBottomCap = computeCapToGND(FIRST_MET);
  capRatio = specificBottomCap / specificTopCap;
  numberOfNet = constr->numberOfNet;
  numberOfPerformance = constr->numberOfPerformance;

  net = netSet->globalNext;
  if (debug) scanCD(constr);
  if (debug) scanNets();

  while (net != NULL) {
    netConstr = netHasConstraint(constr, net);
    if (debug) fprintf(stderr, "orig. net %s ",net->name);
    if (netConstr != NULL) {
      linkConstraintToNet(netConstr, net);
      constraintCounter++;
      net2 = searchSymmetricNet(constr, net, symmetry);
      if (net2 != NULL) {
	linkNetCouple(net, net2);
	if (debug) fprintf(stderr, "***** net %s matched to net %s\n", 
			   net2->matchedNet->name, net->matchedNet->name);
      } else {
	if (debug) fprintf(stderr, "no matched net\n");
      }
      addConstraintToNet(net, net2, OK);
    } else {
      net2 = searchSymmetricNet(constr, net, symmetry);
      netConstr = netHasConstraint(constr, net2);
      if (net2 != NULL && netConstr == NULL) {
	if (net2->number > net->number) {
	  sprintf(errmsg, "%s and %s are matched nets, but don't have constraints", 
		  net->name, net2->name);
	  Warning(errmsg); 
	}
      }
      addConstraintToNet(net, net2, NOT_OK);
      if (debug) fprintf(stderr, "\n");
    }
    net = net->globalNext;
  }
  sprintf(errmsg, "%d constrained nets found", constraintCounter);
  Message(errmsg);

}


NetConstraint
*netHasConstraint(constraintDriven, net)
     ConstraintDriven *constraintDriven;
     Net              *net;
{

  int k, constraintAccumulator = 0;
  ConstraintDriven *constr;
  NetConstraint *netConstr;

  if (net == NULL) return;
  constr = constraintDriven;

  for (k = 0; k < constr->numberOfNet; k++) {
    if (strcmp(constr->netConstraint[k].netName, net->name) == 0) {
      netConstr = &(constr->netConstraint[k]); 
      constraintAccumulator++;
    }
  }
  if (constraintAccumulator > 1) {
    sprintf(errmsg, "More than one constraint on net %s", net->name);
    FatalError (errmsg);
  }
  if (constraintAccumulator == 0) 
    netConstr = NULL;

  return netConstr;

}


void
linkConstraintToNet(netConstr, net)
     NetConstraint *netConstr;
     Net           *net;
{

  if (netConstr == NULL || net == NULL) return;
  netConstr->net = net;

}


void
linkNetCouple(net, net2)
     Net *net;
     Net *net2;
{

  if (net == NULL || net2 == NULL) return;
  net->matchedNet = net2;
  net2->matchedNet = net;

}

void
addConstraintToNet(net, net2, okFlag)
     Net *net;
     Net *net2;
     int okFlag;
{

  int i; /* index */

  if (!okFlag) {
    net->constraintStatus = NON_CONSTRAINED;
    return;
  }
  if (net == NULL) return;

  net->constraintStatus = CONSTRAINED;
  if (net2 != NULL) {
    net2->constraintStatus = CONSTRAINED;
  }

}


/* This routine searches all symmetric nets to "net" wrt every symmetry axis.
 * If info about matched nets is not given the following algorithm is run.
 * Algorithm: - generate a list of all cells attaced to "net". 
 *            - serch the remaining and not yet visited nets and do the same thing
 *            - compare every couple of cellLists and pick the net which has a list
 *              containing all the cells which are symmetric to those attached to "net".
 */
Net
*searchSymmetricNet(constraintDriven, netP, symType)
     ConstraintDriven  *constraintDriven;
     Net               *netP;
     SymmetryType      symType;
{

  MacroCell       *cell1, *cell2;
  Net             *net1, *net2;
  struct termList *listOfTerm1, *listOfTerm2;
  octCoord        symAxis;
  int             singleFlag;
  int             k;
  MatchedNets     *matchedNets;
  char            *string, *string1, *string2;
  int             matchedPairNumber = 0;
  char            *net;

  matchedNets = constraintDriven->matchedNets;
  matchedPairNumber = constraintDriven->numberOfMatchedPair;
  string = netP->name;

  if (matchedPairNumber > 0) {
    for (k = 0; k < matchedPairNumber; k++) {
      string1 = matchedNets[k].net1Name;
      string2 = matchedNets[k].net2Name;
      if (strcmp(string1, string) == 0) {
	net = string2;
	return findNet(netSet, net);
      } else {
	if (strcmp(string2, string) == 0) { 
	  net = string1;
	  return findNet(netSet, net);
 	}
      }
    }
    return NULL;
  }

  net1 = netP;
  listOfTerm1 = &(net1->term0);
  
  net2 = netSet;
  while (net2 != NULL) {
    if (net2 == net1) {
      net2 = net2->globalNext;
      continue;
    }
    listOfTerm2 = &(net2->term0);
    if (listOfTerm2 == NULL) {
      net2 = net2->globalNext;
      continue;
    }
    if (equalNumberOfTerminals(listOfTerm1, listOfTerm2) == OK &&
	cellListCompare(listOfTerm1, listOfTerm2, symType) == OK) {
      return net2;
    }
    net2 = net2->globalNext;
  }
  net2 = NULL;
  return net2;

}


Net
*findNet(setNet, net)
    Net *setNet;
    char *net;
{

  Net *set;

  set = setNet->globalNext;
  while (set != NULL) {
    if (strcmp(set->name, net) == 0) return set;
    set = set->globalNext;
  }

}


int
equalNumberOfTerminals(termList1, termList2)
     struct termList *termList1;
     struct termList *termList2;
{

  struct termList *list1, *list2;
  int termNumber1 = 0, termNumber2 = 0;

  list1 = termList1;
  list2 = termList2;
  
  while (list1 != NULL) {
    if (list1->term != NULL) termNumber1++;
    list1 = list1->next;
  }
  while (list2 != NULL) {
    if (list2->term != NULL) termNumber2++;
    list2 = list2->next;
  }
  
  if (termNumber1 == termNumber2) {
    return OK;
  } else { 
    return NOT_OK;
  }
  
}


int
cellListCompare(termList1, termList2, symType)
     struct termList *termList1;
     struct termList *termList2;
     SymmetryType  symType;
{

  struct termList *list1, *list2;
  MacroCell *cell1, *cell2;
  Term *term1, *term2;
  
  MacroCell *otherCell;
  octCoord symAxis;
  int singleFlag = 0;
  int status;

  list1 = termList1;
  list2 = termList2;
  while (list1 != NULL) {
    status = 0;
    term1 = list1->term;
    if (term1 == NULL) {
      list1 = list1->next;
      continue;
    }
    cell1 = term1->cell;
 
    while (list2 != NULL && !status) {
      term2 = list2->term;
      if (term2 == NULL) {
	list2 = list2->next;
	continue;
      }
      cell2 = term2->cell;
      otherCell = searchOtherSymmetricCell(cell1, &symAxis, &symType, &singleFlag);
      if ((!singleFlag && otherCell->cellId == cell2->cellId) ||
	  (singleFlag && cell1->cellId == cell2->cellId)) {
	status = 1;
      } else {
	list2 = list2->next;
      }
    }
    if (status) {
      list1 = list1->next;
    } else {
      return status;
    }
  }
  return status;
}







