#include "defs.h"
#include "global.h"
#include "parall.h"

int coolingSchedule = EMPIRICAL_COOLING_SCHEDULE; /* temporarily */

static int iteration;
static int shortcutThreshold;

static char *CS_names[] = {
    "CONSTANT ALPHA", "ADAPTIVE", "EMPIRICAL" };

int 
CS_globalInit(schedule)
    int    schedule;
{
    coolingSchedule = schedule;
    (void) sprintf(errmsg,"Restarting cooling schedule: %s",
			  CS_names[coolingSchedule]);
    Message(errmsg);

    switch( coolingSchedule ) {
    case FIXED_COOLING_SCHEDULE:
    case EMPIRICAL_COOLING_SCHEDULE:
	SM_T = SM_startT;
	initRangeLimiters();
	break;
    case ADAPTIVE_COOLING_SCHEDULE:
	SM_T = SM_startT;
	{ 
	    Cost curCost;
	    configurationCost(&curCost);
	    acs_globalInit(P,curCost.total);
	}
	initRangeLimiters();
	break;
    default:
	FatalError("Cooling schedule not yet implemented");
	break;
    }
}

int
CS_initInnerLoop()
{
    switch( coolingSchedule ) {
    case FIXED_COOLING_SCHEDULE:
	iteration = movesPerAnnVar;
	break;
    case EMPIRICAL_COOLING_SCHEDULE:
	iteration = movesPerAnnVar;
	shortcutThreshold = movesPerAnnVar * myProc->curNumberOfAnnealVars * 0.25;
	break;
    case ADAPTIVE_COOLING_SCHEDULE:
	/* acs_initInnerLoop(); */
	break;
    default:
	FatalError("Cooling schedule not yet implemented");
	break;
    }
}

int
CS_innerLoop(deltaCost,accepted)
    double deltaCost;
    int    accepted;
{
    switch( coolingSchedule ) {
    case FIXED_COOLING_SCHEDULE:
    case EMPIRICAL_COOLING_SCHEDULE:
	return FALSE;
    case ADAPTIVE_COOLING_SCHEDULE:
	return acs_innerLoop(deltaCost,accepted,myself);
    default:
	break;
    }
    return FALSE;
}

int
CS_exitInnerLoop()
{
    switch( coolingSchedule ) {
    case FIXED_COOLING_SCHEDULE:
	if ( iteration-- < 0) return(TRUE);
	break;
    case EMPIRICAL_COOLING_SCHEDULE:
	if ( iteration-- < 0 || myProc->moveAccepted > shortcutThreshold) return(TRUE);
	break;
    case ADAPTIVE_COOLING_SCHEDULE:
	return(acs_exitInnerLoop(-1));
	break;
    default:
	FatalError("Cooling schedule not yet implemented");
	break;
    }
    return(FALSE);
}



CS_determineInitialTemperature()
{
  double	tmp;		/* temporary value of initial temperature */
  double	om;		/* order of magnitude */
  int		array[4];	/* rounding thresholds */
  int		i = 0;
  double	pow();

  /* init rounding thresholds */
  array[0] = 1;
  array[1] = 2;
  array[2] = 5;
  array[3] = 0;			/* sentinel */
  /* Set the initial temperature proportional to the st.dev of the cost of the moves */
  /* Round the initial temperature up to a 'nice' number */
  tmp = sqrt(myProc->varMoveCost.total);
  om = floor(log10(tmp));
  
  while (1) {
    SM_startT = pow(10.0,om) * (double) array[i];
    if (SM_startT >= tmp) break; /* found the initial temperature */
    if (array[++i] == 0) {	/* check the array bounds */
      i = 0;			/* reset  */
      om += 1.0;		/* next order of magnitude */
    }
  }
  (void) sprintf(errmsg,"Setting initial temperature to %g", SM_startT);
  Message(errmsg);
}



/*
 *	updateTemperature()
 *	update the temperature in the simulated annealing
 *	algorithm
 */
CS_updateTemperature(cost)
    double cost;		/* Total configuration cost */
{
    switch(coolingSchedule) {
    case FIXED_COOLING_SCHEDULE:
    case EMPIRICAL_COOLING_SCHEDULE:
	if (SM_T == INFINITE_TEMPERATURE) {
	    CS_determineInitialTemperature();
	    SM_T = SM_startT;
	} else {
	    SM_T *= SM_alpha;
	}
	break;
    case ADAPTIVE_COOLING_SCHEDULE:
	acs_newTemperature(cost);
	break;
    default:
	break;
    }
}

#define STOP_CRITERION_THRESHOLD 0.005
/*
 *	saStopCrit()
 *	stopping criterium for sim. Anne.
 *	This is like in TimberWolf: the criterium is satisfied
 *	when the cost function's value has not changed for 4
 *	consecutive stages. The cost is considered not changed if
 *	it differs from the previous one by less then 0.1%
 *	RETURNS OK or NO
 */
CS_exitOuterLoop(cost)
    double cost;
{
    static double previousCost = 1.0; /* init to a value != 0 */
    static int counter=0;

    switch(coolingSchedule) {
    case FIXED_COOLING_SCHEDULE:
    case EMPIRICAL_COOLING_SCHEDULE:
	if ((previousCost == 0.0 AND cost > 0.001) ||
	    fabs(cost-previousCost)/previousCost > STOP_CRITERION_THRESHOLD) {
	    previousCost=cost;
	    counter=0;
	    return(FALSE);
	} else { 
	    return ((++counter < 4 OR SM_T > SM_minimumTemperature)? FALSE : TRUE);
	}
	break;
    case ADAPTIVE_COOLING_SCHEDULE:
	return(acs_exitOuterLoop(-1));
    default:
	FatalError("Cooling schedule not yet implemented");
	break;
    }
    return(FALSE);
}




/** Subroutine called by the interactive menu **/
int
PM_constAlpha()		/* cooling schedule */
{
    if (Confirm("Do you really want to change cooling schedule ?")==OK) {
	CS_globalInit(FIXED_COOLING_SCHEDULE);
    }
    return(OK);
}

int
PM_empiricalCoolSched()		/* cooling schedule */
{
    if (Confirm("Do you really want to change cooling schedule ?")==OK) {
	CS_globalInit(EMPIRICAL_COOLING_SCHEDULE);
    }
    return(OK);
}

int
PM_adaptiveCoolSched()		/* cooling schedule */
{
    if (Confirm("Do you really want to change cooling schedule ?")==OK) {
	CS_globalInit(ADAPTIVE_COOLING_SCHEDULE);
    }
    return(OK);
}


int
PM_zeroT()
{
    if ( SM_T != 0.0 &&
	Confirm("Do you really want to set the temperature to 0.0 ?")==OK) {
	SM_T = 0.0;
	Message("Temperature set to 0.0");
    }
    return(OK);
}

int 
PM_infT()
{
    if (SM_T != INFINITE_TEMPERATURE &&
	Confirm("Do you really want to set the temperature to INFINITY ?")==OK) {
	SM_T = INFINITE_TEMPERATURE;
	Message("Temperature set to INFINITY");
    }
    return(OK);
}
