/*
 *  FILE: dataFile.c
 *
 *  DESCRIPTION:
 *
 *
 *  CREATION DATE: 09/15/87		AUTHOR: Andrea Casotto
 *
 *  FUNCTIONS:
 */
#include "defs.h"
#include "global.h"
#include "sa.h"
#include "analogVariables.h"
#include "constraintDrivenVariables.h"
#include "parall.h"
#include "vov.h"

extern double computeCapToGND();

static FILE 	*DFfile;
static ListElt 	*DFfirstField = NULL;
static int	fileIsOpen = FALSE;

typedef  enum  {DF_INTEGER, DF_REAL, DF_BOOLEAN, DF_STRING} DFtype;


struct DFfield {
  char *name;
  DFtype  type;
  union {
    int *integerP;
    double *realP;
  } value;
  char *comment;
};


DFdescribe()
{
  extern Cost curCost;

  DFaddField("area", DF_REAL , &curCost.area, "Some complicated function of the chip area");
  DFaddField("WL", DF_REAL, &curCost.wireLength, "Estimated wire length");
  DFaddField("OV", DF_REAL, &curCost.overlap, "Overlap");
  DFaddField("MA", DF_REAL, &curCost.matching, "Matching");
  DFaddField("SY", DF_REAL, &curCost.symmetry, "Symmetry");
  DFaddField("WE", DF_REAL, &curCost.well, "Well");
  DFaddField("CO", DF_REAL, &curCost.constraint, "Constraints");
  DFaddField("TotalCost", DF_REAL, &curCost.total, "Total cost of the placement");
  DFaddField("T"        , DF_REAL, &SM_T, "Temperature");
}

DFopen()
{
  extern	SHARED int newDataFile;
  char		dataFileName[60]; /* Output data file */
  int		versionNumber = 0;

  /* Open a new file for the data */
  do {
    (void) sprintf(dataFileName,"%s.data", output_cell);
    /* open the file for reading */
    if ( newDataFile && ( DFfile = fopen(dataFileName, "r")) != (FILE *) NULL) {
      /* The file exists */
      (void) fclose(DFfile);
      versionNumber++;
      continue;
    }
    if (( DFfile = fopen(dataFileName, "w")) == (FILE *) NULL) {
      (void) sprintf(errmsg, "Can't open data file %s", dataFileName );
      Warning(errmsg);
      versionNumber++;
      if (versionNumber>200 && !newDataFile) {
	Warning("Cannot open data file");
	break;			/* exit from the do loop without opening the file */
      }
    } else {
      VOVoutput( VOV_UNIX_FILE, dataFileName );
      (void) sprintf(errmsg,"Data file is %s",dataFileName);
      Message(errmsg);
      fileIsOpen = TRUE;
    }
  } while (fileIsOpen == FALSE) ;
  return(0);
}

DFclose()
{
  if (fileIsOpen) {
    (void) fclose(DFfile);
  }
}


DFwriteComment(s)
    char *s;
{
  if (fileIsOpen) {
    (void) fprintf(DFfile, "# %s\n", s);
    (void) fflush(DFfile);
  }
}

DFaddField(name, type, address, comment)
    char 		*name;
    DFtype  		type;
    double 		*address;
    char 		*comment;
{
  struct DFfield 	*newP;
  ListElt		*LE;

  newP = SH_MALLOC(struct DFfield, 1);
  LE = SH_MALLOC(ListElt, 1);
  LE->pointer.generic = (char *) newP;
  newP->name = strsav(name);
  newP->type = type;
  newP->comment = strsav(comment);
  switch (type) {
 case DF_INTEGER:
 case DF_BOOLEAN:
    newP->value.integerP = (int *) address;
    break;
 case DF_REAL:
    newP->value.realP = (double *) address;
    break;
 default:
    FatalError("Wrong DFfield type");
    break;
  }
  linkEltToListHandle(&DFfirstField,LE);
}


/*
 *  OK if everything went ok, NOT_OK otherwise
 */
int
DFwrite()
{
  ListElt 		*parLE;		/* Used to run through the list of all parameters */
  struct DFfield 	*parP;		/*  */

  if (fileIsOpen) {
    DFwriteCostFunction();
    for (parLE = DFfirstField; parLE!=NULL; parLE = parLE->next) {
      parP = (struct DFfield *) parLE->pointer.generic;
      switch (parP->type) {
    case DF_INTEGER:
	fprintf(DFfile, " %d ", *parP->value.integerP);
	break;
    case DF_BOOLEAN:
	(void) fprintf(DFfile, " %d ", *parP->value.integerP);
	break;
    case DF_REAL:
	(void) fprintf(DFfile, " %.4g ", *parP->value.realP);
	break;
    default:
	FatalError("Wrong type for a DFfield");
      }
    }
    (void) fprintf(DFfile,"\n");
    (void) fflush(DFfile);
  }
}

DFwriteHeader()
{
  ListElt 		*parLE;		/* Used to run through the list of all parameters */
  struct DFfield 	*parP;		/*  */

  if (fileIsOpen) {
    printHeader(DFfile);
    
    fprintf(DFfile,"\n# ***** DESCRIPTION OF DATA FIELDS\n");
    for (parLE = DFfirstField; parLE!=NULL; parLE = parLE->next) {
      parP = (struct DFfield *) parLE->pointer.generic;
      switch (parP->type) {
    case DF_INTEGER:
	fprintf(DFfile, "# KEY: %-20s TYPE: BOOLEAN COMMENT:  %s\n", parP->name, parP->comment);
	break;
    case DF_BOOLEAN:
	fprintf(DFfile, "# KEY: %-20s TYPE: INTEGER COMMENT:  %s\n", parP->name, parP->comment);
	break;
    case DF_REAL:
	(void) fprintf(DFfile, "# KEY: %-20s TYPE: REAL    COMMENT:  %s\n", parP->name, parP->comment);
	break;
    default:
	FatalError("Wrong type for a DFfield");
      }
    }
    (void) fprintf(DFfile,"\n");
    (void) fflush(DFfile);
  }
}

DFwriteCostFunction()
{
  int	npa = 5;		/* number of elements in the pa array */
  static struct {
    double *var;
    double lastValue;
    char   *name;
  } pa[] = {
    { &SM_ovWeight , -1.0 , "overlap weight" },
    { &SM_netWeight , -1.0 , "net weight" },
    { &SM_areaWeight , -1.0 , "area weight" },
    { &SM_conViolWeight , -1.0 , "constraints violation weight" } ,
    { &SM_clusterWeight , -1.0 , "cluster weight" } ,
    { &SM_matchWeight , -1.0 , "matching weight" } , 
    { &SM_symWeight , -1.0 , "symmetry weight" } ,
    { &SM_wellWeight , -1.0 , "well weight" } ,
    { &SM_constraintWeight , -1.0 , "constraints weight" }
  };
  int	i;


  if (!fileIsOpen) return(OK);
  for (i = 0 ; i < npa ; i++) {
    if (*pa[i].var != pa[i].lastValue) {
      fprintf(DFfile,"# COST FUNCTION: %s %.4g\n", pa[i].name , *pa[i].var);
      pa[i].lastValue = *pa[i].var;
    }
  }
  return(OK);
}


/*
 * Cmoputes and prints the capacitance of critical nets.
 */

void
DFwriteCriticalNetsCap()
{
    struct net	 	*NetP;
    int			estimatedLength;
    int			subnetLength;
    octBox		curBBox;
    Term		*term1,*term2;
    int			i;	/* index for subnets */
    double specificTopCap, specificBottomCap;

    specificTopCap = computeCapToGND(SECOND_MET);
    specificBottomCap = computeCapToGND(FIRST_MET);
    fprintf(DFfile, "# *****PARASITICS ESTIMATION \n");
    fprintf(DFfile, "# Specific Capacitances: %fnF(2MET), %fnF(1MET), \n",
	    specificTopCap , specificBottomCap); 
    if (enableSubnets == TRUE) {
	NetP= netSet;
	while ((NetP=NetP->globalNext) != NULL) {
	    if (NetP->constraintStatus != CONSTRAINED) continue;
	    estimatedLength = 0;
	    for (i=0 ; i< NetP->subnets ; i++) {
		term1 = NetP->subnetArray[i].curTerm[0];
		term2 = NetP->subnetArray[i].curTerm[1];
		subnetLength = distance(term1->curAbsCoordP->x, term1->curAbsCoordP->y,
					term2->curAbsCoordP->x, term2->curAbsCoordP->y);
	
		estimatedLength += subnetLength;
		
	    }
	    fprintf(DFfile,"# Crit. net %s, estim. WL= %d, worst cap.= %fnF, best cap.= %fnF\n", 
		    NetP->name, estimatedLength, 
		    estimatedLength * specificBottomCap,
		    estimatedLength * specificTopCap );

	}

    } else {
	NetP= netSet;
	while ((NetP=NetP->globalNext) != NULL) {
	  if (NetP->constraintStatus != CONSTRAINED) continue;
	    curNetBBox(NetP,&curBBox);
	    estimatedLength = halfPerimeter(&curBBox);

	    fprintf(DFfile,"Crit. net %s, estim. WL= %d, worst cap.= %fnF, best cap.= %fnF\n", 
		    NetP->name, estimatedLength, 
		    estimatedLength * specificBottomCap,
		    estimatedLength * specificTopCap );

	}
    }
}












