/*
 *     debugging routines 
 * >>>>  ROUTINES
 *  checkData()
 *  doubleCheck()
 */
#include "defs.h"
#include "global.h"


/*	checkData()
 *	Checks the input data, looking for inconsistencies
 *	or redundancies
 */
checkData()
{
  short		warnFlag = TRUE; 	/* turn off warnings if too many */
  int		warnCounter = 0; 	/* to count the warnings */
  int		maxWarn = 20;		/* maximum number of warnings */
  MacroCell 	*cell;
  struct termList *termPtr;
  struct netList  *netPtr;
  struct net	*net;
  int		nt;		/* count terminals in a net */

  /* check for unused terminals */

  for (cell=gFirstCell->globalNext;cell!=NULL;cell=cell->globalNext) {
    for (termPtr=cell->term0.next; termPtr!=NULL; termPtr=termPtr->next) {
      if (termPtr->term->status == UNUSED && warnFlag==TRUE) {
	(void) sprintf(errmsg,"Term %s in cell %s is not used",termPtr->term->name, cell->name);
	Warning(errmsg);
	if (warnCounter++ >= maxWarn) {
	  Warning("Suppressing warnings");
	  warnFlag = FALSE;
	}
	/*
	 *  Unlink the terminal from the list
	 *  the terminal is lost, but the memory
	 *  doesn't get free. The loss of the terminal
	 *  doesn't affect the final output.
	 *  All this is done to speed up the updating
	 *  of the terminals; the unused terminals
	 *  are not updated, because they are here
	 *  eliminated
	 */
	unlinkTerm(termPtr);
      }
    }
  }
  /*
   *   check the net list of all cells to make sure they are in the right order
   */
  for (cell=gFirstCell->globalNext;cell!=NULL;cell=cell->globalNext) {
    for (netPtr=cell->net0.next; netPtr!=NULL; netPtr=netPtr->next) {
      if (netPtr->next != NULL && netPtr->next->net->number <= netPtr->net->number) {
	(void) sprintf(errmsg,"Cell %s, net list not in order between nets %s (%d) and %s (%d)",
		cell->name, netPtr->net->name, netPtr->net->number,
		netPtr->next->net->name , netPtr->next->net->number);
	FatalError(errmsg);
      }
    }
  }

  /* 
   * check that all the nets have at least two terminals
   */
  net = netSet->globalNext;
  while (net != NULL) {
    if (net->size == 1) {
      (void) sprintf(errmsg,"Net %s has only 1 term", net->name);
      Warning(errmsg);
    }
    nt = 0;
    termPtr = net->term0.next;
    while (termPtr != NULL) {
      nt++;
      termPtr = termPtr->next;
    }
    if (nt != net->size) {
      (void) sprintf(errmsg,"Inconsistent term count for net %s: %d vs %d", net->name, nt, net->size);
      Warning(errmsg);
    }
    net = net->globalNext;
  }
}



/*
 *   doubleCheck()
 *	checks some critical data structures
 *	lyo
 *	bins
 *      clusters
 */
doubleCheck()
{
#ifdef PARALLEL
  int I;			/* Process counter */
  char str[100];		/* A temporary string */
#endif
  MacroCell 	*cell;

  checkLyo();
  checkBins();
  checkSubnets();
  for (cell=gFirstCell->globalNext;cell!=NULL;cell=cell->globalNext) {
    checkFloatTerms(cell);
  }

#ifdef PARALLEL
  checkClusters();
  for (I = 0 ; I< P ; I++) {
    (void) sprintf(str, "MAILBOX of Proc%d", I);
    checkLocks( &process[I].mailbox.lock , str);
  }
#endif
}

int
checkLyo()
{
  MacroCell 	*cell, *next, *prec;
  octCoord	left,right, bottom, top;

#ifdef PARALLEL
  if (s_clock(&lyoLock) == L_FAILED) { /* should never happen */
    (void) sprintf(errmsg,"Who is still playing with my lyo data structure?");
    Warning(errmsg);
    s_lock(&lyoLock);		/* must wait  */
    s_unlock(&lyoLock);		/* no need to keep it locked, however */
  } else {
    s_unlock(&lyoLock);		/*  */
  }
#endif
  /*  check layout data structure */
  cell = gFirstCell;
  while ((cell = cell->lyoNext[PUPPY_L]) != NULL) {
    next = cell->lyoNext[PUPPY_L];
    prec = cell->lyoPrec[PUPPY_L];
    left   = cell->curBbox.lowerLeft.x;
    if (next != NULL AND next->curBbox.lowerLeft.x < left) {
      FatalError("error in lyo PUPPY_L\n");
    }
    if (prec != gFirstCell AND prec->curBbox.lowerLeft.x > left) {
      FatalError("error in lyo PUPPY_L\n");
    }
  }
  cell = gFirstCell;
  while ((cell = cell->lyoNext[PUPPY_B]) != NULL) {
    next = cell->lyoNext[PUPPY_B];
    prec = cell->lyoPrec[PUPPY_B];
    bottom = cell->curBbox.lowerLeft.y;
    if (next != NULL AND next->curBbox.lowerLeft.y < bottom) {
      FatalError("error in lyo PUPPY_B\n");
    }
    if (prec != gFirstCell AND prec->curBbox.lowerLeft.y > bottom) {
      FatalError("error in lyo PUPPY_B\n");
    }
  }
  cell = gFirstCell;
  while ((cell = cell->lyoNext[PUPPY_R]) != NULL) {
    next = cell->lyoNext[PUPPY_R];
    prec = cell->lyoPrec[PUPPY_R];
    right  = cell->curBbox.upperRight.x;
    if (next != NULL AND next->curBbox.upperRight.x > right) {
      FatalError("error in lyo PUPPY_R\n");
    }
    if (prec != gFirstCell AND prec->curBbox.upperRight.x < right) {
      FatalError("error in lyo PUPPY_R\n");
    }
  }
  cell = gFirstCell;
  while ((cell = cell->lyoNext[PUPPY_T]) != NULL) {
    next = cell->lyoNext[PUPPY_T];
    prec = cell->lyoPrec[PUPPY_T];
    top    = cell->curBbox.upperRight.y;
    if (next != NULL AND next->curBbox.upperRight.y > top) {
      FatalError("error in lyo PUPPY_T\n");
    }
    if (prec != gFirstCell AND prec->curBbox.upperRight.y < top) {
      FatalError("error in lyo PUPPY_T\n");
    }
  }
  return(OK);
}


int
checkBins()
{
  MacroCell 	*cell;
  int		I, J, K;	/* counters for the bins */
  int		*bin;

  /* check bins */
  for ( I = 1; I <= numBinX; I++ ) {
    for (J = 1; J <= numBinY; J++) {
      bin = blockArray [I][J];
      for (K=1 ; K<= bin[0]; K++) {
	cell = allCellsArray[bin[K]];
	if (cell->curBin != bin) {
	  FatalError("error in bins\n");
	}
      }				/* next K */
    }				/* next J */
  }				/* next I */

  bin = bucket;
  for (K=1 ; K<= bin[0]; K++) {
    cell = allCellsArray[bin[K]];
    if (cell->curBin != bin) {
      FatalError("error in bins\n");
    }
  }				/* next K */
  return(OK);
}

#ifdef PARALLEL
checkLocks(lp,s)
    slock_t *lp;
    char *s;
{
  if (s_clock(lp) == L_FAILED) {
    (void) sprintf(errmsg,"The lock %s was LOCKED", s);
    Warning(errmsg);
  }
  s_unlock(lp);
}

#endif


checkSubnets()
{
  Net	  	*NetP;
  SubNet 	*subnetP;
  int 		i;

  NetP= netSet;
  while ((NetP=NetP->globalNext) != NULL) {
    for (i=0 ; i< NetP->subnets ; i++) {
      subnetP = &NetP->subnetArray[i];
      if ( subnetP->curTerm[0] == subnetP->curTerm[1] ) {
	  (void) sprintf( errmsg, "Subnet in net %s is attached twice to the same terminal.",
		  subnetP->curTerm[0]->net->name );
	  FatalError( errmsg );
      }
      checkAttachmentTermSubnet(subnetP->curTerm[0],subnetP);
      checkAttachmentTermSubnet(subnetP->curTerm[1],subnetP);
    }
  }
}

checkAttachmentTermSubnet(termP,subnetP)
    Term 	*termP;
    SubNet 	*subnetP;
{
  int		subnetCounter = 0;
  int		found = FALSE;
  ListElt	*subnetLE;

  for (subnetLE = termP->firstSubnet; subnetLE != NULL ; subnetLE=subnetLE->next) {
    subnetCounter++;
    if (subnetLE->pointer.subnet == subnetP) {
      found = TRUE;
    }
  }
  if (found==FALSE)  FatalError("Term does not know that subnet is attached");
  if (termP->subnets < 1 || termP->subnets != subnetCounter)
    FatalError("subnets: term->subnets corrupted");
}


checkFloatTerms(cellP)
    MacroCell *cellP;
{
  int i;
  struct termList *termLE;
  Term		*ft;		/*  Floating terminal */
  int		*tc;		/* Temporary array to count terms in each site */

  
  tc = MALLOC( int , cellP->numberOfSites);
  
  for (i=0;i<cellP->numberOfSites;i++) {
    tc[i] = 0;
  }
  /* check all floating terminals */
  for (termLE = cellP->floatTerm0.next; termLE!=NULL; termLE=termLE->next) {
    ft = termLE->term;
    /* check the site of ft */
    if (ft->curSite >= cellP->numberOfSites || ft->curSite < 0 )  {
      (void) sprintf(errmsg,"Term %s of cell %s in illegal site %d", ft->name,cellP->name, ft->curSite);
      Warning(errmsg);
    }
    tc[ft->curSite]++;		/* count terminals in each site */
    /* Check the coordinate of ft */
    if (ft->curAbsCoordP->x != cellP->floatingSites[ft->curSite].curAbsCoord.x || 
	ft->curAbsCoordP->y != cellP->floatingSites[ft->curSite].curAbsCoord.y) {
      (void) sprintf(errmsg,"Term %s of cell %s has wrong CURRENT coordinates", ft->name,cellP->name);
      Warning(errmsg);
    }      


  }

  /* Check all sites */
  for (i=0;i<cellP->numberOfSites;i++) {
    if (cellP->floatingSites[i].terms > cellP->floatingSites[i].capacity ||
	cellP->floatingSites[i].terms < 0 ) {
      (void) sprintf(errmsg,"Site %d of cell %s exceeds capacity of %d with %d terminals",
	      i, cellP->name ,cellP->floatingSites[i].capacity,cellP->floatingSites[i].terms);
      Warning(errmsg);
    }
    if (cellP->floatingSites[i].terms != tc[i]) {
      (void) sprintf(errmsg,"Inconsistent count of floating terms in site %d: should be %d , not %d",
	      i, tc[i], cellP->floatingSites[i].terms);
      Warning(errmsg);
    }
  }
  FREE(tc);
}
