/*
 *	PUPPY
 *	placement of macrocells using simulated annealing
 *	and a multiprocessor architecture
 *
 *  This program can also be compiled on a sequential machine
 *
 *  Advising professor:
 *      Alberto Sangiovanni-Vincentelli
 *
 *  Parallel version for the Balance 8000: 
 *	Andrea Casotto
 *
 *
 */

#ifndef DEFS
#define DEFS

#define ABUTMENT

/*
 *  The Sequent uses ns32000 processors and the C-compiler
 *  defines `ns32000'. 
 */
#ifdef sequent
#define PARALLEL
#endif

#include "copyright.h"
#include "port.h"
#include <errno.h>
#include <sys/time.h>
#include <sys/resource.h>
#include "oct.h"

#ifdef hpux
#include <signal.h>
#endif

#ifdef PARALLEL
#include <parallel/parallel.h>
#endif


#ifdef PARALLEL
#define SHARED  shared
#define PRIVATE private
#else
#define SHARED  
#define PRIVATE 
#endif



/*
 *   MACRO DEFINITIONS
 *
 *   Macros are distinguished by using solely capital letters in their
 *   identifiers.  This contrasts with C defined identifiers which are strickly 
 *   lower case, and program variable and procedure names which use both upper 
 *   and lower case.
 */

/* Begin macros. */

/* Boolean data type */
#define  BOOLEAN	int
#define  FALSE		0
#define  TRUE		1
#define  NOT		!
#define  AND		&&
#define  OR		||

#define RAND	((randVar = randVar * 1103515245 + 12345) & 0x7fffffff)
#define RANDOM	((double) RAND / (double) 0x7fffffff)
#ifndef NULL
#define NULL	0
#endif
#define OK	1
#define NOT_OK	0
#define INFINITY 1.0e+10

/* number edges counterclockwise starting from the bottom */
#define PUPPY_B	0		/* bottom */
#define PUPPY_R	1		/* right */
#define PUPPY_T	2		/* top */
#define PUPPY_L	3		/* left */
#define FIRST	1		/* used in owner link/unlink routine */
#define LAST	9		/* used in owner link/unlink routine */
#define NEW	1		/*  */
#define CURRENT	0

#define OCT_LAMBDA		20
#ifdef PARALLEL
#define MAX_CPU 	12
#else
#define MAX_CPU 	1
#endif

/* Macro commands */
/* Macro functions that return the maximun or minimum independent of type. */
#define	NIL(type)	((type *) 0)

#ifndef MAX
#define MAX(a,b)	((a) > (b) ? (a) : (b))
#endif
#ifndef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))
#endif

/* Macro function that returns the absolute value of a floating point number. */
#define  ABS(a)             ((a) < 0.0 ? -(a) : (a))

/* Macro procedure that swaps two entities. */
#define  SWAP(type, a, b)   {type swapx; swapx = a; a = b; b = swapx;}
 
/* Macro function that returns the rectified value of an integer point number. */
#define  RECT(a)             ((a) < 0 ? 0 : (a))


/* Memory allocation and dealocation macros. */
char    *CheckPtr();
#define MALLOC(type,number)  (type *)CheckPtr(malloc((unsigned)(sizeof(type)*(number))))
#define CALLOC(type,number)  (type *)CheckPtr(calloc(sizeof(type),(unsigned)(number)))
#define REALLOC(ptr,type,number)  \
           ptr = (type *)realloc((char *)ptr,(unsigned)(sizeof(type)*(number)))
#define FREE(pointer)  if ((pointer) != NULL)  \
	   { free((char *)pointer); (pointer)=NULL; }

#ifdef PARALLEL
char  *shmalloc();
#define SH_MALLOC(type,number) \
  (type *)CheckPtr(shmalloc((unsigned)(sizeof(type)*(number))))
#define SH_FREE(pointer)  if ((pointer) != NULL)  \
	   { shfree((char *)pointer); (pointer)=NULL; }
#else
#define SH_MALLOC(type,number) \
  (type *)CheckPtr(malloc((unsigned)(sizeof(type)*(number))))
#define SH_FREE(pointer)  if ((pointer) != NULL)  \
	   { free((char *)pointer); (pointer)=NULL; }
#endif

#define strsav(s)	strcpy(MALLOC(char,strlen(s)+1), s)

typedef int (*PFI)();		/* pointer to a function returning integer */

enum  ChipBoundaryType {
  CB_AUTOMATIC ,		/* large enough to fit all the cells */
  CB_INITIAL_POSITION ,		/* preserve initial area  */
  CB_FIXED_CELLS ,		/* area containing all fixed cells */
  CB_SPECIFIED_BY_OCT,		/* given by OCT */
  CB_NUMBER_OF_CB_TYPES		/* a marker, not a type */
  };

#endif















