/*
 *  FILE: float.c
 *
 *  DESCRIPTION: Routines to handle floating terminals
 *
 *
 *  CREATION DATE: 08/17/87		AUTHOR: Andrea Casotto
 *
 *  FUNCTIONS: computeFloatingSites(cell)
 *	       assignFloatingSites(cell)
 */


#include "defs.h"
#include "global.h"
#include "sa.h"


#define MIN_SITE_CAPACITY	2 	/*  */

int terminalSpacing = 10;		/* lambdas between two floating terminals */
int maxSiteCapacity = 20;		/*  */
/*
 * Compute the number of sites for floating terminals and then 
 * compute the positions of these sites. There must always be a site at the center of the cell
 * The floating sites can be defined only for rectangular cells
 */
int
computeFloatingSites(cell)
    MacroCell *cell;
{
  int 		i;
  octPoint	*edgeStart;		/*  */
  octPoint	*edgeEnd;
  int		lx,ly;
  int		edgeCounter;
  int		edge;
  int		edgeLength;
  int		siteIndex;		/* index of the srray of sites */
  int		siteLengthX;		/* length ideally covered by the site */
  int		siteLengthY;		/* length ideally covered by the site */
  int		sitesOnEdge;
  char		*edgeName();

  cell->numberOfSites = 0;
  if (cell->edgesInPerim != 4) {
    (void) sprintf(errmsg,"Cell %s has floating terminals but it is not rectangular", cell->name);
    FatalError(errmsg);
  }

  for (edgeCounter = PUPPY_B ; edgeCounter <= PUPPY_L ; edgeCounter++) {
    edge = (cell->lowestLeftCorner + edgeCounter) % 4;
    edgeStart = &(cell->perimeter[edge]);
    edgeEnd   = &(cell->perimeter[(edge+1) % cell->edgesInPerim]);
    lx = ABS(edgeStart->x - edgeEnd->x);
    ly = ABS(edgeStart->y - edgeEnd->y);
    edgeLength = MAX(lx,ly);
    /* 
     * Total number of floating terms that can fit in this edge, minus the
     *  number of fixed pins already on this edge, divided by the maximum number of
     * floating terms in a site
     */
    sitesOnEdge = SitesOnEdge(edgeLength,cell->termsInEdge[edge], edge);
    cell->numberOfSites += sitesOnEdge;
  }

  cell->numberOfSites += 1;	/* account for site at the center of the cell */

  /* Allocate the sites and compute their position */
  cell->floatingSites = SH_MALLOC(struct s_site, cell->numberOfSites );

  cell->floatingSites[0].cellCoord.x = 0;
  cell->floatingSites[0].cellCoord.y = 0;
  cell->floatingSites[0].relPosition = 0;
  cell->floatingSites[0].capacity = cell->floatingTerms;
  cell->floatingSites[0].edge = 0;
  cell->floatingSites[0].terms = 0;	/* no float terms  */
  
  siteIndex = 1;
  for (edgeCounter = PUPPY_B ; edgeCounter <= PUPPY_L ; edgeCounter++) {
    edge = (cell->lowestLeftCorner + edgeCounter) % 4;
    edgeStart = &(cell->perimeter[edge]);
    edgeEnd   = &(cell->perimeter[(edge+1) % cell->edgesInPerim]);
    lx = ABS(edgeStart->x - edgeEnd->x);
    ly = ABS(edgeStart->y - edgeEnd->y);
    edgeLength = MAX(lx,ly);
    sitesOnEdge = SitesOnEdge(edgeLength,cell->termsInEdge[edge], edge);
    siteLengthX = (edgeEnd->x - edgeStart->x)  / sitesOnEdge;
    siteLengthY = (edgeEnd->y - edgeStart->y)  / sitesOnEdge;
    for (i = 0; i< sitesOnEdge ; i++) {
      cell->floatingSites[siteIndex].cellCoord.x =edgeStart->x+siteLengthX*i+(siteLengthX >> 1);
      cell->floatingSites[siteIndex].cellCoord.y =edgeStart->y+siteLengthY*i+(siteLengthY >> 1);
      
      /* Assign a relative position in the sense of growing coordinates */
      if (siteLengthX + siteLengthY > 0) {
	/* Growing coords */
	cell->floatingSites[siteIndex].relPosition =
	  (double) (2*i+1) / (double) (2*sitesOnEdge);
      } else {
	/* Decreasing coords */
	cell->floatingSites[siteIndex].relPosition =
	  (double) (2*(sitesOnEdge - i ) - 1) / (double) (2*sitesOnEdge);
      }

      cell->floatingSites[siteIndex].capacity = 
	(edgeLength/terminalSpacing-cell->termsInEdge[edge]) / sitesOnEdge +1;
      if (cell->floatingSites[siteIndex].capacity < MIN_SITE_CAPACITY) {
	(void) sprintf(errmsg,"Capacity for site %d of cell %s is %d. Set to %d", i, cell->name,
		cell->floatingSites[siteIndex].capacity,MIN_SITE_CAPACITY);
	Warning(errmsg);
      }
      cell->floatingSites[siteIndex].edge = edgeCounter;
      cell->floatingSites[siteIndex].terms = 0;	/* no float terms  */
      if (debug) printf("Site %4d (%4d,%4d) %s\n", siteIndex, 
			cell->floatingSites[siteIndex].cellCoord.x,
			cell->floatingSites[siteIndex].cellCoord.y,
			edgeName(edgeCounter));
      siteIndex++;
    }
  }
#ifdef PARALLEL
  s_init_lock(&cell->floatingSitesLock);
#endif
  return(OK);
}


/* If the sites are perfectly symmetric around the chip,
 * then the algorithm will not be able to choose one of 2 equally good
 * symmetric solutions. Break the symmetry of sites, by throwing some 
 * extra sited on some edges
 */
int
SitesOnEdge(edgeLength,fixedTermsInEdge,edge)
    int		edgeLength;
    int		fixedTermsInEdge;
    int		edge;
{
  int	sitesOnEdge;
  int 	breakSymmetry = TRUE;
  

  sitesOnEdge = ((edgeLength / terminalSpacing) - fixedTermsInEdge) / maxSiteCapacity; 
  sitesOnEdge = MAX ( 1 , sitesOnEdge);
  if (breakSymmetry == TRUE) {
    if (edge == PUPPY_L || edge == PUPPY_T) {
      sitesOnEdge += 1;		/* Slight breaking of symmetry */
    }
  }
  return(sitesOnEdge);
}



/* Assign each floating terminal to one floating site */
assignFloatingSites(cell)
    MacroCell *cell;
{
  struct termList *termPtr;
  int		site = 0;

  termPtr= &(cell->floatTerm0);
  while ((termPtr=termPtr->next) != NULL) {
    termPtr->term->curSite = site;
    termPtr->term->newSite = site;
    termPtr->term->curAbsCoordP = &cell->floatingSites[site].curAbsCoord;
    termPtr->term->newAbsCoordP = &cell->floatingSites[site].newAbsCoord;

    /* the next fields are not used by floating terminals */
    termPtr->term->cellCoord.x = 0; 	/* clear */
    termPtr->term->cellCoord.y = 0; 	/* clear */
    termPtr->term->edge = 0;		/* clear */

    /* Increment term count and , possibly, the site */
    if ( ++cell->floatingSites[site].terms >= cell->floatingSites[site].capacity ) {
      if (++site>=cell->numberOfSites) FatalError("Not enough space for all the floating terms");
    }
  }
}


int
parseFloatingEdges(str)
    char *str;
{
  char edge[15];
  int  mask = 0xff;		/* All edges legal */

  if (strcmp(str,"") != 0)  {
    /* Find all legal edges */
    mask = 0;
    while (sscanf(str,"%14s", edge) == 1) {
      if (strcmp("LEFT",edge)==0) {
	mask |= 1 << PUPPY_L;
      } else if (strcmp("TOP",edge)==0) {
	mask |= 1 << PUPPY_T;
      } else if (strcmp("RIGHT",edge)==0) {
	mask |= 1 << PUPPY_R;
      } else if (strcmp("BOTTOM",edge)==0) {
	mask |= 1 << PUPPY_B;
      } else {
	(void) sprintf(errmsg,"Unparsable string %s", edge);
      }
      
      if (strchr(str,' ') != 0) {
	str = strchr(str,' ') + 1;
      } else {
	break;
      }
    }
  }
  return(mask);
}
