/*
 *
 *    lyoRoutines
 *    This file contains all the routines that deal with the
 *	layout data structure, composed by a double
 *	linked list for each edge of the bounding box of
 *	the cells.
 * >>>>  ROUTINES
 *   lyoLink(cell, cell_a, side)
 *   lyoUnlink(cell, side)
 *   lyosort()
 *   lyoReplace()
 */


#include "defs.h"
#include "sa.h"
#include "global.h"


int	lyoCoreOnly = TRUE;	/* Only the core cells are considered for area computation */

/*
 *	lyoLink(cell,cell_a,side)
 */
lyoLink(cell,cell_a,side)
    int side;
    MacroCell *cell,*cell_a;
{
  cell->lyoNext[side] = cell_a->lyoNext[side];
  cell->lyoPrec[side] = cell_a;
  cell_a->lyoNext[side] = cell;
  if (cell->lyoNext[side] != NULL) cell->lyoNext[side]->lyoPrec[side]=cell;
}

/*
 *	lyoUnlinkCell(cell,side)
 */
lyoUnlink(cell,side)
    int side;
    MacroCell *cell;
{
  /* unlink from old list **** */
  (cell->lyoPrec[side])->lyoNext[side]=cell->lyoNext[side];
  if (cell->lyoNext[side]!=NULL) {
    (cell->lyoNext[side])->lyoPrec[side]=cell->lyoPrec[side];
  }
}


lyoZeroPointers(cellPtr)
    MacroCell *cellPtr;
{
  cellPtr->lyoNext[PUPPY_L] = NULL;
  cellPtr->lyoNext[PUPPY_R] = NULL;
  cellPtr->lyoNext[PUPPY_T] = NULL;
  cellPtr->lyoNext[PUPPY_B] = NULL;
  cellPtr->lyoPrec[PUPPY_L] = NULL;
  cellPtr->lyoPrec[PUPPY_R] = NULL;
  cellPtr->lyoPrec[PUPPY_T] = NULL;
  cellPtr->lyoPrec[PUPPY_B] = NULL;
}


int
lyoActiveCells()
{
  int i;
  int count = 0;

  for (i = 1 ; i <= numberOfCells ; i++ ) {
    if (lyoSkipCell(allCellsArray[i]) == FALSE) count++;
  }
  return(count);
}

int
lyoSkipCell(cell)
    MacroCell *cell;
{
  return((lyoCoreOnly &&
	  cell->isPad &&
	  cell->mobility == FIXED) ?
	 TRUE : FALSE);
}
/*
 *	lyosort()
 *	sorts the layout in a quadruple bidirectional list
 *	according to the position of each of the 4 sides
 *	of the bounding box of each cell
 *	The sorting is by insertion;
 */
int
lyosort()
{
  octCoord 	left,bottom,right,top;
  MacroCell	*lyoPtr;
  MacroCell	*cellPtr;
  int		activeCells;

  activeCells = lyoActiveCells();
  if (activeCells <= 1 ) {
    (void) sprintf(errmsg,"Only %d cell(s) active for computation of chip area", activeCells );
    Warning(errmsg);
    Message("Zeroing cost of area");
    SM_areaWeight = 0.0;
  }

  Message("Ordering the layout data structure");
  /* zeroes all gFirstCell pointers */
  
  lyoZeroPointers(gFirstCell);

  /* **** for each cell in the gFirstCell **** */
  cellPtr= gFirstCell ;
  do {
    cellPtr=cellPtr->globalNext;
  
    /* Skip fixed pads if only the core area should be minimized */
    if (lyoSkipCell(cellPtr)) {
      lyoZeroPointers(cellPtr);
      continue;
    }
  
    left  =cellPtr->curBbox.lowerLeft.x;
    bottom=cellPtr->curBbox.lowerLeft.y;
    right =cellPtr->curBbox.upperRight.x;
    top   =cellPtr->curBbox.upperRight.y;


    /* **** PUPPY_L **** */
    lyoPtr= gFirstCell;
    while (lyoPtr->lyoNext[PUPPY_L]!=NULL &&
	   lyoPtr->lyoNext[PUPPY_L]->curBbox.lowerLeft.x < left) {
      lyoPtr=lyoPtr->lyoNext[PUPPY_L];
    }
    lyoLink(cellPtr,lyoPtr,PUPPY_L);

    /* **** PUPPY_R **** */
    lyoPtr=  gFirstCell ;
    while (lyoPtr->lyoNext[PUPPY_R]!=NULL &&
	   lyoPtr->lyoNext[PUPPY_R]->curBbox.upperRight.x > right) {
	   lyoPtr=lyoPtr->lyoNext[PUPPY_R];
	 }
    lyoLink(cellPtr,lyoPtr,PUPPY_R);

    /* **** PUPPY_T **** */
    lyoPtr= gFirstCell ;
    while (lyoPtr->lyoNext[PUPPY_T]!=NULL &&
	   lyoPtr->lyoNext[PUPPY_T]->curBbox.upperRight.y > top) {
	   lyoPtr=lyoPtr->lyoNext[PUPPY_T];
	 }
    lyoLink(cellPtr,lyoPtr,PUPPY_T);

    /* **** PUPPY_B **** */
    lyoPtr= gFirstCell ;
    while (lyoPtr->lyoNext[PUPPY_B]!=NULL &&
	   lyoPtr->lyoNext[PUPPY_B]->curBbox.lowerLeft.y < bottom) {
	   lyoPtr=lyoPtr->lyoNext[PUPPY_B];
	 }
    lyoLink(cellPtr,lyoPtr,PUPPY_B);
  } while (cellPtr->globalNext != NULL);
  return(OK);
}





/* 
 *  lyoReplace(cell)
 *	Cell has got a new position: copy new position
 *	into current and assure consistency with
 *	lyo data structure
 *	This routine takes cell and finds its new position in 
 *	the layout list
 *	The lyo data structure is used to compute the total area
 *	of the chip.
 *	this is a linked list shared by all the processes
 *	a lock is therefore necessary
 *  New  version, probably faster, because it doesn't search
 *  throughout the whole list to find the new position, but starts
 *  from the old position, and moves locally
 */
int
lyoReplace(cell)
    MacroCell *cell;
{
  register MacroCell 		*lyoPtr;
  octCoord 			left,right,top,bottom;

  /*
   *   Only ONE process at a time can be in this routine
   *   A lock is used to guarantee this.
   */
#ifdef PARALLEL
  s_lock(&lyoLock);
#endif

  cell->curBbox.lowerLeft.x = cell->newBbox.lowerLeft.x; 
  cell->curBbox.lowerLeft.y = cell->newBbox.lowerLeft.y; 
  cell->curBbox.upperRight.x = cell->newBbox.upperRight.x; 
  cell->curBbox.upperRight.y = cell->newBbox.upperRight.y; 

  if (lyoSkipCell(cell)) {
#ifdef PARALLEL
    s_unlock(&lyoLock);
#endif
    return;
  }


  left  =cell->curBbox.lowerLeft.x;
  bottom=cell->curBbox.lowerLeft.y;
  right =cell->curBbox.upperRight.x;
  top   =cell->curBbox.upperRight.y;

  /* **** LEFT **** */
  if ((cell->lyoPrec[PUPPY_L] != gFirstCell AND 
       cell->lyoPrec[PUPPY_L]->curBbox.lowerLeft.x > left) OR
      (cell->lyoNext[PUPPY_L] != NULL AND 
       cell->lyoNext[PUPPY_L]->curBbox.lowerLeft.x < left)  ) {
    /* cell must be moved */
    /* - Get a valid address in the lyo data structure */
    if (cell->lyoNext[PUPPY_L] != NULL) {
      lyoPtr = cell->lyoNext[PUPPY_L];
    } else {
      lyoPtr = cell->lyoPrec[PUPPY_L];
    }
    lyoUnlink(cell,PUPPY_L);
    /* - Look for the new position of cell */
    while ( lyoPtr != gFirstCell && lyoPtr->curBbox.lowerLeft.x > left ) {
      lyoPtr = lyoPtr->lyoPrec[PUPPY_L];
    }
    while (lyoPtr->lyoNext[PUPPY_L]!=NULL && lyoPtr->lyoNext[PUPPY_L]->curBbox.lowerLeft.x<left){
      lyoPtr = lyoPtr->lyoNext[PUPPY_L];
    } 
    /* - Re-link cell to the list */
    lyoLink(cell,lyoPtr,PUPPY_L);
  }



  /* **** RIGHT **** */
  if ((cell->lyoPrec[PUPPY_R] != gFirstCell AND
       cell->lyoPrec[PUPPY_R]->curBbox.upperRight.x < right) OR
      (cell->lyoNext[PUPPY_R] != NULL AND
       cell->lyoNext[PUPPY_R]->curBbox.upperRight.x > right)  ) {
    /* cell must be moved */
    /* - Get a valid address in the lyo data structure */
    if (cell->lyoNext[PUPPY_R] != NULL) {
      lyoPtr = cell->lyoNext[PUPPY_R];
    } else {
      lyoPtr = cell->lyoPrec[PUPPY_R];
    }
    lyoUnlink(cell,PUPPY_R);
    /* - Look for the new position of cell */
    while ( lyoPtr != gFirstCell && lyoPtr->curBbox.upperRight.x < right ) {
      lyoPtr = lyoPtr->lyoPrec[PUPPY_R];
    }
    while ( lyoPtr->lyoNext[PUPPY_R] != NULL && lyoPtr->lyoNext[PUPPY_R]->curBbox.upperRight.x > right ) {
      lyoPtr = lyoPtr->lyoNext[PUPPY_R];
    } 
    /* - Re-link cell to the list */
    lyoLink(cell,lyoPtr,PUPPY_R);
  }

  /* **** PUPPY_T **** */
  if ((cell->lyoPrec[PUPPY_T] != gFirstCell AND
       cell->lyoPrec[PUPPY_T]->curBbox.upperRight.y < top) OR
      (cell->lyoNext[PUPPY_T] != NULL AND
       cell->lyoNext[PUPPY_T]->curBbox.upperRight.y > top)  ) {
    /* cell must be moved */
    /* - Get a valid address in the lyo data structure */
    if (cell->lyoNext[PUPPY_T] != NULL) {
      lyoPtr = cell->lyoNext[PUPPY_T];
    } else {
      lyoPtr = cell->lyoPrec[PUPPY_T];
    }
    lyoUnlink(cell,PUPPY_T);
    /* - Look for the new position of cell */
    while ( lyoPtr != gFirstCell && lyoPtr->curBbox.upperRight.y < top ) {
      lyoPtr = lyoPtr->lyoPrec[PUPPY_T];
    }
    while ( lyoPtr->lyoNext[PUPPY_T] != NULL && lyoPtr->lyoNext[PUPPY_T]->curBbox.upperRight.y > top ) {
      lyoPtr = lyoPtr->lyoNext[PUPPY_T];
    } 
    /* - Re-link cell to the list */
    lyoLink(cell,lyoPtr,PUPPY_T);
  }

  /* **** PUPPY_B **** */
  if ((cell->lyoPrec[PUPPY_B] != gFirstCell AND
       cell->lyoPrec[PUPPY_B]->curBbox.lowerLeft.y > bottom) OR
      (cell->lyoNext[PUPPY_B] != NULL AND 
       cell->lyoNext[PUPPY_B]->curBbox.lowerLeft.y < bottom)  ) {
    /* cell must be moved */
    /* - Get a valid address in the lyo data structure */
    if (cell->lyoNext[PUPPY_B] != NULL) {
      lyoPtr = cell->lyoNext[PUPPY_B];
    } else {
      lyoPtr = cell->lyoPrec[PUPPY_B];
    }
    lyoUnlink(cell,PUPPY_B);
    /* - Look for the new position of cell */
    while ( lyoPtr != gFirstCell && lyoPtr->curBbox.lowerLeft.y >= bottom ) {
      lyoPtr = lyoPtr->lyoPrec[PUPPY_B];
    }
    while ( lyoPtr->lyoNext[PUPPY_B] != NULL && lyoPtr->lyoNext[PUPPY_B]->curBbox.lowerLeft.y < bottom ) {
      lyoPtr = lyoPtr->lyoNext[PUPPY_B];
    } 
    /* - Re-link cell to the list */
    lyoLink(cell,lyoPtr,PUPPY_B);
  }

#ifdef PARALLEL
  s_unlock (&lyoLock);
#endif
}

MacroCell *
lyoPrecCell(cell,side)
    MacroCell	*cell;
    int		side;
{
  MacroCell *nc;		/* Next cell */
  
  /* If the preceding cell in direction 'side' is NULL or the 'handle', return NULL */
  nc = cell->lyoPrec[side];
  if (nc == (MacroCell *) NULL || nc == gFirstCell)  
    return((MacroCell *) NULL);
  else 
    return(nc);
}

MacroCell *
lyoAdjacentCell(cell,side)
    MacroCell	*cell;
    int		side;
{
  MacroCell *nc;		/* Next cell */
  MacroCell *lyoPrecCell();
  
  /* If the preceding cell in direction 'side' is NULL or the 'handle', return NULL */
  nc = lyoPrecCell(cell,side);
  if (nc == NULL)  return((MacroCell *) NULL);

  switch (side) {
  case PUPPY_L:
    while ( nc->curBbox.upperRight.x - cell->curBbox.lowerLeft.x >= 0)
      if ((nc = lyoPrecCell(nc,side)) == (MacroCell *) NULL)
	return((MacroCell *) NULL);
    break;
  case PUPPY_R:
    while ( nc->curBbox.lowerLeft.x - cell->curBbox.upperRight.x <= 0) 
      if ((nc = lyoPrecCell(nc,side)) == (MacroCell *) NULL)
	return((MacroCell *) NULL);
    break;
  case PUPPY_T:
    while ( nc->curBbox.lowerLeft.y - cell->curBbox.upperRight.y <= 0)
      if ((nc = lyoPrecCell(nc,side)) == (MacroCell *) NULL)
	return((MacroCell *) NULL);
    break;
  case PUPPY_B:
    while ( nc->curBbox.upperRight.y - cell->curBbox.lowerLeft.y >= 0) 
      if ((nc = lyoPrecCell(nc,side)) == (MacroCell *) NULL)
	return((MacroCell *) NULL);
    break;
  default:
    FatalError("Inconsistent enumeration of edges in lyoNextCell. Should never happen");
    break;
  }
  return(nc);
}
