/*
 *  FILE: octIO.c
 *
 *  DESCRIPTION: Interface with OCT data base
 *
 *
 *  CREATION DATE: Aug.21.1987          AUTHOR: Andrea Casotto
 *
 *  FUNCTIONS:
 *  int         readDataFromOct()
 *  int         addNetToCell(netptr, cell)
 *  int         saveResultsToOct(finalFlag)
 *              printTermList(cell)
 *
 *  IMPORTED FUNCTIONS:
 *  int         readAnalogDataFromOct() <-- analogOctIO.c
 *  BiasStatus  readWell(instanceP) <-- analogOctIO.c
 *  PowerDissip *readPowerDissip(instanceP)
 */
 
#include "defs.h"
#include "global.h"
#include "sa.h"
#include "analogVariables.h"
#ifndef ABUTMENT 
#include "constraintDrivenVariables.h"
#endif
#include "octPlacement.h"
#include "st.h"
#include "tr.h"
#include "fang.h"
#include "harpoon.h"
#include "oh.h"
#include "vov.h"

/* IMPORTED GLOBAL VARIABLES AND STRUCTURES */


/* DEFINITION OF GLOBAL CONSTANTS FOR THIS MODULE */

#define ANALOG_MAX_NET_SIZE 50
#define DIGITAL_MAX_NET_SIZE 16
  
st_table	*st_cells;	/* hash table to get MacroCells from Instances ids */
st_table	*st_masters;	/* hash table to get cells from master ids */
st_table	*st_terminals;	/* hash table to get terminals from Term ids */
octObject	InstanceBag, ConnectorBag;


/* IMPORTED FUNCTIONS */

extern BiasStatus readWell();
extern int readAnalogDataFromOct();
extern int saveAxis2oct();
extern double readPeakCurrent();
extern octCoord computeTermWidth();

/* FUNCTIONS */

/*
 *  Read all nets in Chip
 *  For each net, find attached terminals	
 *  For each term find the cell and do all appropriate linkings		
 */
void readAllNets()
{
    int                 k;       /* index */
    MacroCell		*cell;
    struct term		*term;
    struct net 		*net;
    struct netList  	*netptr;
    struct termList 	*termptr;
    octObject		InstTerm; /* terminal of an instance */
    octObject		NetTerm; /* terminal on a net */
    octObject		ChipNet;
    octStatus		status;	/* to record the return codes */
    octGenerator	genTerm, genNet; /* generators */
    int	 		nTerm ;	/* counter */
    int			maxNetSize;  
                        /* This limit has been set to ANALOG_MAX_NET_SIZE 
			 * for analog circuits (option -r) and to
			 * DIGITAL_MAX_NET_SIZE for digital circuits.
                         * ANALOG_MAX_NET_SIZE and DIGITAL_MAX_NET_SIZE 
			 * have been defined in "octIO.c" (top) */

    if (analogMode)  {
      maxNetSize = ANALOG_MAX_NET_SIZE;
    } else {
      maxNetSize = DIGITAL_MAX_NET_SIZE;
    }
    
    if (octInitGenContents(&Chip,OCT_NET_MASK, &genNet) != OCT_OK) 
      FatalOctError("generating nets");
    netSet = SH_MALLOC ( Net ,1);
    bzero( netSet, sizeof(Net) );
    netSet->globalNext = netSet->globalPrec =NULL;
    while ((status = octGenerate(&genNet,&ChipNet)) == OCT_OK) {
	/* check that there are at least two terminals attached to this net */
	octInitGenContents(&ChipNet,OCT_TERM_MASK,&genTerm);

	/* **** count terminals **** */
	nTerm=0;	
	while ( octGenerate(&genTerm,&NetTerm) == OCT_OK ) nTerm++;
	if (nTerm == 0) {
	    (void) sprintf(errmsg,"Net %s ignored because has NO terminals",
			   ChipNet.contents.net.name);
	    Warning(errmsg);
	    continue;		/* go on to the next net */
	} else if (nTerm == 1) {
	    (void) sprintf(errmsg,"Net %s has only ONE terminal: ignored",
			   ChipNet.contents.net.name);
	    Warning(errmsg);
	    continue;		/* go on to the next net */
	} else if (nTerm > maxNetSize ) {
	    (void) sprintf(errmsg, "%s has %d pins (max is %d): ignored",
			   ohFormatName( &ChipNet ), nTerm, maxNetSize );
	    Warning( errmsg );
	    continue;
	}

    
	net = SH_MALLOC ( Net, 1);
	net->number = ++numberOfNets;
	net->term0.next = net->term0.prec = NULL;

	/* **** add this net to netSet **** */
	netptr = SH_MALLOC( struct netList ,1);
	netptr->net=net;

	/* put this net in the global list of the nets */
	net->globalNext = netSet->globalNext;
	net->globalPrec = netSet;
	netSet->globalNext = net;
	if (net->globalNext != NULL) {
	    net->globalNext->globalPrec=net;
	}

	net->name=ChipNet.contents.net.name;
	net->weight = 1;	/* default */
	net->maxLength = 0;	/* means "NO max length" */
	net->type = SIGNAL;     /* default type for a net */
	net->status = NON_ABUTTABLE;        /* net by default is non abuttable */
	net->constraintStatus = NON_CONSTRAINED;

	if (octGetByName(&ChipNet,&Criticality)==OCT_OK) {
	    net->weight = (int) exp((double) Criticality.contents.prop.value.integer / 1000);
	}
	if (octGetByName(&ChipNet,&NetMaxLength)==OCT_OK) {
	    net->maxLength = NetMaxLength.contents.prop.value.integer;
	}
	/* for each net generate all the terminals */
	if (octInitGenContents(&ChipNet,OCT_TERM_MASK,&genTerm) != OCT_OK) {
	    (void) sprintf(errmsg,"Generating terms of net %s", ChipNet.contents.net.name);
	    FatalOctError(errmsg);
	}
	nTerm=0;		/* **** count terminals **** */
	while (octGenerate(&genTerm,&NetTerm) == OCT_OK) {
	    if (NetTerm.contents.term.instanceId == 0) {
	      /* This is a formal terminal attached to a net: 
		 we must get the terminal implementing it*/
	      if (octGenFirstContent(&NetTerm,OCT_TERM_MASK,&InstTerm) 
		  != OCT_OK) {
		sprintf(errmsg,"Formal term %s doesn't contain actual terms",
			NetTerm.contents.term.name);
		if (debug) Warning(errmsg);
		continue;	/* skip this formal terminal */
	      }
	      if (st_lookup(st_cells, 
			    (char *) InstTerm.contents.term.instanceId, 
			    (char **) &cell)== 0) {
		sprintf(errmsg,"Net %s connected to non-existent cell", 
			net->name);
		FatalError(errmsg);
	      }
	      /* if the cell that we found is a not connector, 
		 than we must skip it, because we will find the 
		 same terminal again */
	      if (cell->isConnector == FALSE) {
		continue;	/* skip this cell */
	      } else {
		(NetTerm = InstTerm);
	      }
	    } else {
	      if (debug) printf("Debug: instId for this term %d\n",
				NetTerm.contents.term.instanceId);
	      /* Find cell that contains the terminal from my internal 
		 data structure */
	      if (st_lookup(st_cells,(char *)NetTerm.contents.term.instanceId,
			    (char **) &cell)== 0) {
		sprintf(errmsg,"Net %s connected to non-existent cell",
			net->name);
		FatalError(errmsg);
	      }
	    }
	    if (debug) printf("Looking for term %s in ",
			      NetTerm.contents.term.name);
	    if (debug) printTermList(cell);

	    nTerm++;
	    termptr = SH_MALLOC (struct termList , 1);
	    netptr  = SH_MALLOC (struct netList , 1 );
	    netptr->net=net;

	    if (st_lookup(st_terminals, (char *) NetTerm.objectId , 
			  (char **) &term) == 0) {
	      sprintf(errmsg,"Net %s connected to non-existent term %s, of cell %s",
		      netptr->net->name,
		      NetTerm.contents.term.name,
		      cell->name);
	      FatalError(errmsg);
	    }
      
	    if (term->status == USED) {
		sprintf(errmsg, "Term %s of %s is attached to net %s and to net %s",
			term->name, term->cell->name, term->net->name , net->name);
		FatalError(errmsg);
	      }
	    term->abutStatus = NON_ABUTTABLE; /* term by default is non abuttable */
	    term->status  = USED;
	    term->type = SIGNAL; /* default status for a term. in "power.c" 
			terms can also be characterized as SUPPLY or CLOCK */
	    term->net     = net;
	    termptr->term = term;

	    /* **** link terminal to net **** */
	    linkTerm(termptr,&(net->term0));
	    /* **** link net to cell **** */
	    status = addNetToCell(netptr,cell);
	    if (debug) {
		printf("##### Terminal %s cell %s net %s\n", 
		       term->name, cell->name, net->name);
		if (status==1) {
		  printf("Net %s added to cell %s\n", net->name, cell->name);
		} else {
		  printf("Net %s already attached to cell %s\n", 
			 net->name , cell->name);
		}
	    }
	}
	net->size = nTerm;

	/* recognizes power nets */
	handlePowerNet(net);
	
	if ( net->size > 2) createAnnealVar(AV_NET, (char *) net);
	initSubnets(net);
	numberOfSubnets += net->subnets;
	if (debug) printf("Found %d terminals in net %s\n",nTerm,net->name);
	net->age = 0;		/* initialize the age (even if it is not used yet) */
    }
   
    sprintf(errmsg,"****** Found %d Nets",numberOfNets);
    Message(errmsg);
    if (status!=OCT_GEN_DONE) FatalOctError("Generating nets");

    /* Generates a structure like netSet called powerNetSet for the 
       computation of power net length */
    genPowerNetSet();

}


/*
 *    readDataFromOct()
 *	reads form the OCT data base
 *	Updates the following external variables:
 */
int
readDataFromOct()
{
  MacroCell		*cell;
  struct netList  	*netptr;
  octObject		Instance;
  octObject		Connector;
  octObject		InstTerm; 	/* terminal of an instance */
  octObject		FormalTerm;	/* formal terminal of a master*/
  octStatus		status;		/* to record the return codes */
  octGenerator	        genInst; 	/* generators */
  int	 		nNet ;		/* counter */

  st_cells =  st_init_table(st_numcmp, st_numhash);
  st_masters =  st_init_table(st_numcmp, st_numhash);
  st_terminals = st_init_table(st_numcmp, st_numhash);

  (void) sprintf(errmsg,"Reading %s:%s", Chip.contents.facet.cell, Chip.contents.facet.view);
  Message(errmsg);
  numberOfCells = 0;
  numberOfMovingCells = 0;
  numberOfFloatingPins = 0;
  numberOfNets = 0;
  numberOfSubnets = 0;
  numberOfAnnealVars = 0;
  totalArea = 0.0;
  maxArea   = 0.0;
  maxEdgeX = maxEdgeY = 0;

  /* reading data concerning instances, terminals, nets and connectors  
   * from Oct. This operation is performed for digital circuits
   * and for analog circuits as well. */

  ConnectorBag.type = OCT_BAG;
  ConnectorBag.contents.bag.name = "CONNECTORS";
  InstanceBag.type = OCT_BAG;
  InstanceBag.contents.bag.name = "INSTANCES";

  if (octGetByName(&Chip,&InstanceBag)==OCT_NOT_FOUND) FatalOctError("Missing instance bag");
  if (octGetOrCreate(&Chip,&ConnectorBag)!=OCT_OK) FatalOctError("Problem with connector bag");
   
  /*  For each  instance take all the relevant information */
  gFirstCell->globalNext = gFirstCell->globalPrec = NULL;
  if (octInitGenContents(&InstanceBag,OCT_INSTANCE_MASK,&genInst)!=OCT_OK)
    FatalOctError("generating instances");
  while ((status = octGenerate(&genInst,&Instance)) == OCT_OK) {
    readInstance(&Instance);
  }
  if (status!=OCT_GEN_DONE) FatalOctError("generating instances");

  /*  Read in all the connectors that implement formal terminals also */
  if (octInitGenContents(&ConnectorBag,OCT_INSTANCE_MASK,&genInst)!=OCT_OK)
    FatalOctError("generating connectors");
  while ((status = octGenerate(&genInst,&Connector)) == OCT_OK) {
    if (octGenFirstContent(&Connector,OCT_TERM_MASK,&InstTerm)!=OCT_OK) {
      (void) sprintf(errmsg,"No actual terminal in Connector %s", Connector.contents.instance.name);
      FatalOctError(errmsg);
    }
    if (octGenFirstContainer(&InstTerm,OCT_TERM_MASK,&FormalTerm) == OCT_OK) {
      if (FormalTerm.contents.term.instanceId == 0) {
	readInstance(&Connector);
      } else {
	(void) sprintf(errmsg,
		"Term %s on connector %s is contained by term %s which is not a formal terminal",
		InstTerm.contents.term.name,
		Connector.contents.instance.name,
		FormalTerm.contents.term.name);
	FatalOctError(errmsg);
      }
    }
  }
  if (status!=OCT_GEN_DONE) FatalOctError("generating instances");


  if (numberOfCells == 0) {
    FatalError("No cells to be placed");
  }
  if (numberOfCells == 1) {
    FatalError("Only one cell in this placement problem");
  }

  (void) sprintf(errmsg,"****** Found %d Instances",numberOfCells);
  Message(errmsg);

  readAllNets();
  
  /*******************************************************/
  /* **** now count the nets connected to each cell **** */
  /* **** and find absolute positions of terminals  **** */
  /*******************************************************/
  cell= gFirstCell;
  maxNets = 0;
  while ((cell=cell->globalNext) != NULL) {
    nNet=0;			/* count the nets in this cell */
    netptr= &cell->net0;
    while ((netptr=netptr->next) != NULL)  nNet++;
    cell->totNets=nNet;
    maxNets = MAX (maxNets, nNet);
  }
  setupCellArrays();

  /* reading data concerning the analog circuit general constraints,
   * the circuit topology and circuit matching constraints
   * only if the flag "analogMode" is set. The Flag is set in main.c */

  if (analogMode)  {
    (void) sprintf(errmsg,"Reading analog constraints from %s:%s", 
		   Chip.contents.facet.cell, Chip.contents.facet.view);
    Message(errmsg);
    readAnalogDataFromOct();
  }
  return(OK);
}

/*
 *   Set up the array allCellsArray[] for quick
 *   reference to the cells
 */
int setupCellArrays()
{
    int		I;		/* cell index */
    int		varI = 0;	/* index for moving cells */
    MacroCell 	*cell;		/* loop pointer */
  
    allCellsArray = SH_MALLOC( MacroCellPointer, numberOfCells + 1 );
    varCellsArray = SH_MALLOC( MacroCellPointer, numberOfMovingCells );
    for (I=0;I<=numberOfCells;I++) {
	allCellsArray[I]=NULL;	/* reset the array */
    }
    for (I=0;I<numberOfMovingCells;I++) {
	varCellsArray[I]=NULL;	/* reset the array */
    }
    cell = gFirstCell;
    while ( (cell = cell->globalNext) != NULL) {
	if (cell->mobility == MOVABLE) {
	    varCellsArray[varI++] = cell;
	}
	I = cell->number;
	if (allCellsArray[I] != NULL)  FatalError("Numbering the cells");
	allCellsArray[I] = cell;
    }
}

int
readInstance(InstanceP)
    octObject *InstanceP;
{
  MacroCell		*cell;		/*  */
  MacroCell		*lookAlikeCell;
  octObject		Master;
  octCoord		correctiveTranslationX;
  octCoord		correctiveTranslationY;
  char			*encodeTransformConstraints();

  showActivity();
  cell = SH_MALLOC ( MacroCell, 1);	/* any instance is one cell */
  cell->number = numberOfCells + 1;     /* cells are numbered 1,2,3,... */
  cell->cellId = InstanceP->objectId;
  cell->name   = InstanceP->contents.instance.name;
  cell->owner  = 0;		/* initialize to allow colored display */
  cell->status = OWNED;		/* initialize the status */
  cell->moveId = 0;		/* not  moving */
  cell->age    = 0;      	/* reset age */
  cell->xSymPointer = NULL;     /* reset pointer to hor. symmetry data-base */
  cell->ySymPointer = NULL;     /* reset pointer to vert. symmetry data-base */ 
  cell->matchPointer = NULL;    /* reset pointer to the matching data-base */
  cell->analogStatus = NONE;    /* cell by default hasn't analog constraints */
  cell->biasStatus = INDEPENDENT; /* cell by default is independently biased */ 
  cell->abutStatus = NON_ABUTTABLE; /* cell by default is non abuttable */   
  
  /* in case the circuit to be placed is analog the bias flag shell assume
   * another value */ 
  if (analogMode && considerWell) {
    cell->biasStatus = readBias(InstanceP);
  }  

  /* add cell at the beginning of the list of all the cells */
  cell->globalNext = gFirstCell->globalNext;
  cell->globalPrec = gFirstCell;
  gFirstCell->globalNext = cell;
  if (cell->globalNext != NULL) {
    cell->globalNext->globalPrec=cell;
  }

  if (debug) printf("MASTER %s\n",InstanceP->contents.instance.master);
  st_insert(st_cells, (char *) InstanceP->objectId, (char *) cell);

  /* get the Master */
  Master.type=OCT_FACET;
  Master.contents.facet.cell = InstanceP->contents.instance.master;
  Master.contents.facet.view = InstanceP->contents.instance.view;
  Master.contents.facet.version = InstanceP->contents.instance.version;
  Master.contents.facet.facet="interface";
  Master.contents.facet.mode="r";
  VOVinputFacet( &Master );
  if (octOpenFacet(&Master) < OCT_OK) {
    (void) sprintf(errmsg,"Opening master %s:%s",
			  Master.contents.facet.cell,Master.contents.facet.view);
    FatalOctError(errmsg);
  }

  if (st_lookup(st_masters, (char *) Master.objectId , 
		(char **) &lookAlikeCell)!= 0) {

    /* Grab all information for cell from lookAlikeCell */
    copyMasterInfo(lookAlikeCell,cell);

  } else {

    st_insert(st_masters, (char *) Master.objectId , (char *) cell);
    cell->masterId = Master.objectId;
    readCellGeometry(cell);
    readCellTerminals(cell);

  }

  readCellProperties(cell);

  if (cell->mobility == MOVABLE) {
    createAnnealVar(AV_CELL, (char *) cell);
    numberOfMovingCells++;
  }
  numberOfCells++;

  /* Update maxEdgeX and maxEdgeY */
  /* this is done here in the simplest way */
  maxEdgeX= MAX (maxEdgeX,cell->W);
  maxEdgeX= MAX (maxEdgeX,cell->H);
  maxEdgeY=maxEdgeX;
  maxArea = MAX (maxArea, cell->area);
  totalArea += cell->area;
  /*
   *   transform and translation. In PUPPY the transforms don't 
   *   move the center of the cell as they do in OCT
   */
  cellTransform(cell,InstanceP->contents.instance.transform.transformType);
  ExecuteCellMovement(cell);		/* actually execute the movement */
  /* This must be done explicitly here.  Normally it is done in lyoReplace, and not in
   * ExecuteCellMovement, because the bbox is used by the lyo datastructure which can
   * only be modified synchronously, using lyoLock. But since we are reading in the data
   * now, we can do it without using the lock.
   */
  cell->curBbox = cell->newBbox; 

  /* At this point the cell has the right orientation, but is in the wrong position */
  /* Compute the corrective translation to match the centers */
  switch (InstanceP->contents.instance.transform.transformType) {
  case OCT_NO_TRANSFORM:
      correctiveTranslationX = 0;
      correctiveTranslationY = 0;
      break;
  case OCT_MIRROR_X:
      correctiveTranslationX = - 2 * cell->curCenter.x;
      correctiveTranslationY = 0;
      break;
  case OCT_MIRROR_Y:
      correctiveTranslationX = 0;
      correctiveTranslationY = - 2 * cell->curCenter.y;
      break;
  case OCT_ROT90:
      correctiveTranslationX = -cell->curCenter.x - cell->curCenter.y;
      correctiveTranslationY = -cell->curCenter.y + cell->curCenter.x;
      break;
  case OCT_ROT180:
      correctiveTranslationX = - 2 * cell->curCenter.x;
      correctiveTranslationY = - 2 * cell->curCenter.y;
      break;
  case OCT_ROT270:
      correctiveTranslationX = -cell->curCenter.x +cell->curCenter.y;
      correctiveTranslationY = -cell->curCenter.y -cell->curCenter.x;
      break;
  case OCT_MX_ROT90:
      correctiveTranslationX = -cell->curCenter.x -cell->curCenter.y;
      correctiveTranslationY = -cell->curCenter.y -cell->curCenter.x;
      break;
  case OCT_MY_ROT90:
      correctiveTranslationX = -cell->curCenter.x +cell->curCenter.y;
      correctiveTranslationY = -cell->curCenter.y +cell->curCenter.x;
      break;
  default:
      break;
  }

  cellTranslate(cell,
		correctiveTranslationX +
		InstanceP->contents.instance.transform.translation.x / OCT_LAMBDA,
		correctiveTranslationY +
		InstanceP->contents.instance.transform.translation.y / OCT_LAMBDA);
  ExecuteCellMovement(cell);

  numberOfFloatingPins += cell->floatingTerms;


  if (debug) {
    printf("Cell %s, Id#%d  ",cell->name,cell->cellId);
    printf("Bounding Box : %d,%d   %d,%d\n", 
		    cell->newBbox.lowerLeft.x,
		    cell->newBbox.lowerLeft.y,
		    cell->newBbox.upperRight.x,
		    cell->newBbox.upperRight.y);
    printf("lowerLeft in the master %d,%d\n", cell->masterLLC.x, cell->masterLLC.y );
  }
  return OK;
}


int
readCellProperties(cell)
    MacroCell *cell;
{
  octObject		ForbiddenEdges;
  char			*encodeTransformConstraints();
  octObject Instance;
  octObject Master;

  Instance.objectId = cell->cellId;
  Master.objectId = cell->masterId;

  if (octGetById(&Instance)!=OCT_OK) FatalOctError("Corrupted cell instance id");
  if (octGetById(&Master)!=OCT_OK) FatalOctError("Corrupted cell master id");

  /* Update the properties */
  if (octGetByName(&Instance,&InstanceClass)==OCT_OK) {
    (void) sprintf(errmsg,"Obsolete property %s found", InstanceClass.contents.prop.name);
    Warning(errmsg);
    PlacementClass.contents.prop.value.string = InstanceClass.contents.prop.value.string;
    if (octDetach(&Instance,&InstanceClass) !=OCT_OK) FatalError("d2");
    octGetOrCreate(&Instance,&PlacementClass);
  }
  if (octGetByName(&Instance,&OLDInstanceClass)==OCT_OK) {
    (void) sprintf(errmsg,"Obsolete property %s found", OLDInstanceClass.contents.prop.name);
    Warning(errmsg);
    PlacementClass.contents.prop.value.string = OLDInstanceClass.contents.prop.value.string;
    if (octDetach(&Instance,&OLDInstanceClass) !=OCT_OK) FatalError("d2");
    octGetOrCreate(&Instance,&PlacementClass);
  }

  /* is the cell a pad ? */
  if (octGetByName(&Instance,&PlacementClass)==OCT_OK) {
    if (strcmp(PlacementClass.contents.prop.value.string,"PAD")==0)  {
      cell->isPad = TRUE;
    } else {
      cell->isPad = FALSE;
    }
  } else if (octGetByName(&Master,&PlacementClass)==OCT_OK &&
    strcmp(PlacementClass.contents.prop.value.string,"PAD")==0) {
    cell->isPad = TRUE;
  } else {
    cell->isPad = FALSE;
  }

  /* Is the cell a connector ? */
  if (octIsAttached(&ConnectorBag,&Instance) == OCT_OK) {
    cell->isConnector = TRUE;		/* If it is a connector */
    cell->isPad = TRUE;			/* then assume it is also a pad */
  } else {
    cell->isConnector = FALSE;
  }

  cell->forbiddenEdgeMask = 0;
  if (octGetByName(&Master,&FloorplanBag) == OCT_OK) {
    ForbiddenEdges.type = OCT_PROP;
    ForbiddenEdges.contents.prop.name = "FORBIDDEN_EDGES";
    if (octGetByName(&FloorplanBag,&ForbiddenEdges)==OCT_OK) {
      cell->forbiddenEdgeMask = parseFloatingEdges(ForbiddenEdges.contents.prop.value.string) ;
    }
  }

  if (octGetByName(&Instance ,&Mobility)==OCT_OK &&
      strcmp(Mobility.contents.prop.value.string,"FIXED")==0) {
    cell->mobility = FIXED;
  } else {
    cell->mobility = MOVABLE;
  }
  if (cell->isConnector==TRUE) cell->mobility = FIXED;
  
  if (octGetByName(&Instance ,&TranCons)==OCT_OK) {
    (void) sprintf(errmsg,"Obsolete property %s", TranCons.contents.prop.name);
    Warning(errmsg);
    /* If there are constraints, convert to new policy */
    if (TranCons.contents.prop.value.integer != 0) {
      LegalTransform.contents.prop.value.string = 
	encodeTransformConstraints(TranCons.contents.prop.value.integer);
      octCreate(&Instance, &LegalTransform);
    }
    octDetach(&Instance, &TranCons);
  }
  processLegalTransforms(cell);
  return(OK);
}


readCellTerminals(cell)
    MacroCell *cell;
{
  struct term		*term;
  struct termList 	*termptr;
  octObject		Box;
  octObject		MasterTerm;	/* terminal on a master */	
  octObject		InstTerm; 	/* terminal of an instance */
  octGenerator		genTerm;
  octCoord		x1,x2,y1,y2; /* used to get term.  coordinates */
  int	 		nTerm = 0 ;	/* counter */
  int			masterModified = FALSE;
  int                   I;              /* index */
  octObject 		Instance;
  octObject 		Master;


  Instance.objectId = cell->cellId;
  Master.objectId = cell->masterId;

  if (octGetById(&Instance)!=OCT_OK) FatalOctError("Corrupted cell instance id");
  if (octGetById(&Master)!=OCT_OK) FatalOctError("Corrupted cell master id");

  /* **** find all the terminals in Master **** */
  cell->floatingTerms = 0;
  cell->numberOfSites = 0;
  cell->term0.next = NULL;		/* clear the handle for the terminal list */
  cell->term0.prec = NULL;		/* clear the handle for the terminal list */
  cell->term0.term = NULL;		/* clear the handle for the terminal list */
  cell->floatTerm0.next = NULL;		/* clear the handle for the terminal list */
  cell->floatTerm0.prec = NULL;		/* clear the handle for the terminal list */
  cell->floatTerm0.term = NULL;		/* clear the handle for the terminal list */
  cell->lastMovedTermLE = NULL;
  cell->net0.next = NULL;
  cell->net0.prec = NULL;
  cell->net0.net  = NULL;


  if (octInitGenContents(&Master,OCT_TERM_MASK,&genTerm) != OCT_OK)
    FatalOctError("generating  Terminals in the master");

  while (octGenerate(&genTerm,&MasterTerm)==OCT_OK) {
    nTerm++;
    InstTerm.type = OCT_TERM;		/* find the corresponding term on the instance */
    InstTerm.contents.term.name = MasterTerm.contents.term.name;
    if (octGetByName(&Instance,&InstTerm)!=OCT_OK) {
      sprintf(errmsg,"Term %s in master %s:%s but not on instance %s",
	      InstTerm.contents.term.name, Master.contents.facet.cell,
	      Master.contents.facet.view, Instance.contents.instance.name);
      FatalOctError(errmsg);
    }
    term = SH_MALLOC ( Term, 1);
    st_insert(st_terminals, (char *) InstTerm.objectId , (char *) term);

    termptr = SH_MALLOC (struct termList, 1);
    termptr->term = term;
    term->status = UNUSED;
    term->name = MasterTerm.contents.term.name;
    term->cell = cell;
    term->subnets = 0;
    term->firstSubnet = NULL;
    
    /* look for the property PEAKCURRENT in term */
    term->peakCurrent = readPeakCurrent(&InstTerm, &Instance); 

    /* compute minimum width for the erminal */
    term->minWidth = computeTermWidth(term);

    if (debug) {
      sprintf(errmsg,"term %s, connection width = %f ", term->name,
	      (double)(term->minWidth/lambda));
      Message(errmsg);
    }

    /* term->net is to be filled later */

    /* Is this terminal floating ? */
    if (octGetByName(&MasterTerm,&FloatingEdge)==OCT_OK) {
      term->floating = TRUE;

      term->illegalEdges  = ~parseFloatingEdges(FloatingEdge.contents.prop.value.string);
      term->illegalEdges |= cell->forbiddenEdgeMask;	/* Logic OR */
      term->illegalEdges &= 0x0f;		/* Keep only the first 4 bits */

      /* Not yet parsing the value of FloatingEdge */
      cell->floatingTerms++;
      linkTerm (termptr,&cell->floatTerm0);
      if (debug) printf("Terminal %s is floating (%d,%d)\n",
		      term->name, term->cellCoord.x, term->cellCoord.y);

      createAnnealVar(AV_FLOAT_TERM, (char *) term);

    } else if  (octGetByName(&MasterTerm,&OldFloatingEdge)==OCT_OK) {
      (void) sprintf(errmsg,"Obsolete property found: %s: property being replaced by %s",
		      OldFloatingEdge.contents.prop.name,
		      FloatingEdge.contents.prop.name);
      Warning(errmsg);
      Master.contents.facet.mode = "a";
      VOVinputFacet( &Master );
      VOVoutputFacet( &Master );
      if (octOpenFacet(&Master) < OCT_OK) {
	Warning("Cannot modify master");
      } else {
	FloatingEdge.contents.prop.value.string = OldFloatingEdge.contents.prop.value.string;
	if (octDetach(&MasterTerm,&OldFloatingEdge)!=OCT_OK) FatalOctError("Deletion failed");
	if (octCreate(&MasterTerm,&FloatingEdge)!=OCT_OK) FatalOctError("Attachement failed");
	masterModified = TRUE;
	term->floating = TRUE;
	cell->floatingTerms++;
	linkTerm (termptr,&cell->floatTerm0);
	if (debug) printf("Terminal %s is floating (%d,%d)\n",
			  term->name, term->cellCoord.x, term->cellCoord.y);
      }
    } else {
      term->floating = FALSE;
      term->curAbsCoordP = SH_MALLOC(octPoint , 1);
      term->newAbsCoordP = SH_MALLOC(octPoint , 1);

      /* get the implementation of the terminal */
      if (octGenFirstContent(&MasterTerm,OCT_BOX_MASK,&Box)!=OCT_OK) {
	(void) sprintf(errmsg, "Term %s of %s:%s has no implementation",
		MasterTerm.contents.term.name,
		Master.contents.facet.cell,
		Master.contents.facet.view);
	Warning(errmsg);
	term->cellCoord.x = 0;
	term->cellCoord.y = 0;
	term->edge = 0;		/* not really on this edge */
      } else {
	x1 = Box.contents.box.lowerLeft.x / OCT_LAMBDA;
	y1 = Box.contents.box.lowerLeft.y / OCT_LAMBDA;
	x2 = Box.contents.box.upperRight.x / OCT_LAMBDA;
	y2 = Box.contents.box.upperRight.y / OCT_LAMBDA;
	/* We are only interested in the center of the box   */
	/* Find coord with respect to the CENTER of the cell */
	term->cellCoord.x = ((x1+x2) >> 1) - cell->curCenter.x;
	term->cellCoord.y = ((y1+y2) >> 1) - cell->curCenter.y;
	term->edge = findTermEdge(cell,term);
	cell->termsInEdge[term->edge]++; /* update number of terms on edge */
	if (considerPower) {
	  cell->widthOfEdge[term->edge] += 
	    (int)((term->minWidth - minWireWidth) / minWireWidth);
	  /* compute rounded number of additional equiv. wires of 
	     min-geometry introduced by power buses */
	} else {
	  cell->widthOfEdge[term->edge] = 0;
	}
      }
      linkTerm (termptr,&cell->term0);
      if (debug) printf("Terminal %s in point %d,%d (edge %d)\n",
		      term->name, term->cellCoord.x, term->cellCoord.y, term->edge);
    }
  }
  if (masterModified) {
    if (octFlushFacet(&Master)!= OCT_OK) FatalOctError("flush master");
  }

  if (debug) printf("\tFound %d terms in cell %s, %d of which are floating",
		    nTerm,cell->name,cell->floatingTerms);

  if (debug) printTermList(cell);

  return(OK);
}


/* link the net in order of growing id's */
/* should be smarter to avoid repetitions */
/* returns (0) if the net was already in the list, 1 if it has been added */
int
addNetToCell(newNet, cell)
    struct netList *newNet;
    MacroCell	*cell;
{
  struct netList *netP;		/* pointer in the existing list of nets */

  netP = &(cell->net0);		/* get the first net in the list */
  
  while (netP->next != NULL && netP->next->net->number < newNet->net->number) netP = netP->next;
  /* Add the net if it is not already in the list */
  if ( netP->next == NULL || netP->next->net->number != newNet->net->number) {
    linkNet(newNet,netP);	/* link newNet after netP */
    return(1);
  } else {
    return(0);
  }
}


/*
 *    saveResultsToOct(finalFlag)
 */
int
saveResultsToOct(finalFlag)
    int	finalFlag;
{
  octObject  		Instance;	/* */
  octObject		MasterInterface;
  octObject		MasterTerm;
  octGenerator		genInst;	/* generator of all instances */
  octCoord 		tx,ty;		/* translation vector */
  MacroCell		*cell;		/* cell corresponding to the Instance */
  struct octTransform	transform;	/* Used to compute the transformation of masterLLC */
  int			llc;		/* lowestLeftCorner of the cell */
  octPoint		masterLLC;	/* */
  char			*edgeName();
  struct termList	*termPtr;

  (void) sprintf(errmsg,"Saving %s placement in %s:%s",
	 (finalFlag)? "FINAL" : "INTERMEDIATE" ,
	 Chip.contents.facet.cell, Chip.contents.facet.view);
  Message(errmsg);

  /* Generate each instance of the output facet, and adjust */
  if (octInitGenContents(&Chip, OCT_INSTANCE_MASK, &genInst) != OCT_OK)
    FatalOctError("generating instances for output");

  /* for each  instance fill the position and transform */
  while ( octGenerate(&genInst,&Instance) == OCT_OK) {
    if (st_lookup(st_cells, (char *) Instance.objectId , (char **) &cell)== 0) {
      FatalError("Corrupted hash table");
    }
    if (cell->mobility != FIXED) {
      Instance.contents.instance.transform.transformType = cell->curTransform;
      transform.transformType = cell->curTransform;
      transform.translation.x = 0;
      transform.translation.y = 0;
      (masterLLC = cell->masterLLC);
      tr_oct_transform(&transform , &masterLLC.x, &masterLLC.y);
      llc = cell->lowestLeftCorner; /* reference point to compute the translation */
      tx = cell->curPerimeter[llc].x - masterLLC.x;
      ty = cell->curPerimeter[llc].y - masterLLC.y;
      Instance.contents.instance.transform.translation.x = tx * OCT_LAMBDA;
      Instance.contents.instance.transform.translation.y = ty * OCT_LAMBDA;
      if (octModify(&Instance) != OCT_OK) FatalOctError("Modifying instance");
      if (debug) printf("Cell %s: transformation = %d , (%d,%d)\n",
			cell->name,
			Instance.contents.instance.transform.transformType,
			Instance.contents.instance.transform.translation.x,
			Instance.contents.instance.transform.translation.y);
    }
    if (cell->floatingTerms > 0) {
      MasterInterface.type = OCT_FACET;
      MasterInterface.contents.facet.cell = Instance.contents.instance.master;
      MasterInterface.contents.facet.view = Instance.contents.instance.view;
      MasterInterface.contents.facet.facet = "interface";
      MasterInterface.contents.facet.version = Instance.contents.instance.version;
      MasterInterface.contents.facet.mode = "a";
      
      VOVinputFacet( &MasterInterface );
      VOVoutputFacet( &MasterInterface );
      if (octOpenFacet(&MasterInterface)< OCT_OK) {
	(void) sprintf(errmsg,"Cannot modify %s:%s:%s", MasterInterface.contents.facet.cell,
		MasterInterface.contents.facet.view, MasterInterface.contents.facet.facet);
	Warning(errmsg);
      } else {
	for (termPtr = cell->floatTerm0.next; termPtr != NULL;termPtr = termPtr->next) {
	  MasterTerm.type = OCT_TERM;
	  MasterTerm.contents.term.name = termPtr->term->name;
	  if (octGetByName(&MasterInterface,&MasterTerm)!=OCT_OK) FatalOctError("missing term");
	  if (octGetOrCreate(&MasterTerm,&TermEdge)!=OCT_OK) FatalOctError("get or creat prop");
	  TermEdge.contents.prop.value.string =
	           edgeName(cell->floatingSites[termPtr->term->curSite].edge);
	  if (octModify(&TermEdge)!=OCT_OK) FatalOctError("Modif termEdge");

	  if (octGetOrCreate(&MasterTerm,&RelTermPosition)!=OCT_OK) FatalOctError("get/creat prop");
	  RelTermPosition.contents.prop.value.real =
	    cell->floatingSites[termPtr->term->curSite].relPosition;
	  if (octModify(&RelTermPosition)!=OCT_OK) FatalOctError("Modif RelTermPosition");
	  if (debug) printf("\tFloat Term %20s : %s , %6.2f\n",
			    termPtr->term->name,
			    TermEdge.contents.prop.value.string,
			    RelTermPosition.contents.prop.value.real);
			    
	  
	}
	if (octFlushFacet(&MasterInterface)!=OCT_OK) FatalOctError("flushing interface");
      }
    }
  }
  savePuppyStatus(finalFlag);
  if (octFlushFacet(&Chip) != OCT_OK) FatalOctError("Flushing");
  (void) sprintf(errmsg,"%s placement saved in %s:%s",
		  (finalFlag) ? "FINAL" : "INTERMEDIATE",
		  Chip.contents.facet.cell, Chip.contents.facet.view);
  Message(errmsg);

  if (analogMode && considerVirtual) {
      saveAxis2oct(&Chip, gySymCell);
      saveAxis2oct(&Chip, gxSymCell);
    }

  return(OK);
}


savePuppyStatus(finalFlag)
    int	finalFlag;			/* intermediate */
{
  octObject	PuppyBag;
  octObject	Temperature;

  PuppyBag.type = OCT_BAG;
  PuppyBag.contents.bag.name = "PUPPY_BAG";
  Temperature.type = OCT_PROP;
  Temperature.contents.prop.type = OCT_REAL;
  Temperature.contents.prop.name = "TEMPERATURE";
  Temperature.contents.prop.value.real = SM_T;

  if (finalFlag == FALSE) {
    if (octGetOrCreate(&Chip, &PuppyBag)!=OCT_OK) FatalOctError("puppy_bag");
    if (octCreateOrModify(&PuppyBag,&Temperature)!=OCT_OK) FatalOctError("temperature");
  } else {
    if (octGetByName(&Chip, &PuppyBag) == OCT_OK) {
      if (octDelete(&PuppyBag)!=OCT_OK) FatalOctError("deleting puppy_bag");
    }
  }
}

printTermList(cell)
    MacroCell *cell;
{
  struct termList *termptr;
  int i;			/* counter */

  printf(" Term list for cell %s\n",cell->name);
  for (termptr=cell->term0.next; termptr!=NULL ; termptr=termptr->next)  {
    printf(" %s  ",termptr->term->name);
  }
  printf("\n");
  printf("Term distribution : ");
  for (i=0;i<cell->edgesInPerim;i++) printf(" %d->%d, width %d ",
					    i, cell->termsInEdge[i], 
					    cell->widthOfEdge[i]);
  printf("\n");
}


char *edgeNames[] = { "BOTTOM" , "RIGHT", "TOP", "LEFT"};

char *edgeName(i)
    int i;
{
  if (i < 0) FatalError("negative value to edgename()");
  if (i < 4)  return edgeNames[i];
  return "***";		/*   */
}

/* find the edge with minimum distance from the center of the terminal */
int
findTermEdge(cell,term)
    MacroCell *cell;
    struct term *term;
{
  int 	distance;
  int	minDist;
  int 	i;
  int	nexti;
  int	edge;

  minDist = MAX(cell->W, cell->H); /* initialize to a large value */
  for (i=0; i<cell->edgesInPerim ; i++) {
    nexti = (i + 1) % cell->edgesInPerim;
    if (cell->perimeter[i].x == cell->perimeter[nexti].x) {
      distance = ABS( term->cellCoord.y - cell->perimeter[i].y);
    } else {
      distance = ABS( term->cellCoord.x - cell->perimeter[i].x);
    }
    
    if (distance < minDist) {
      edge = i ;
      minDist = distance;
    }
  }
  /* make sure the edge returned is correctly numbered */
  return((edge - cell->lowestLeftCorner + cell->edgesInPerim) % cell->edgesInPerim);
}




























































































