/*
 *  FILE: sa.h
 *
 *  DESCRIPTION:
 * 
 *
 *  CREATION DATE: xx/xx/xx		AUTHOR: Andrea Casotto
 * 
 *  
 *  LAST MODIFICATION: 3/22/91          AUTHOR: Edoardo Charbon
 *
 *
 *  FUNCTIONS:
 */


#ifndef SA_HEADER
#define SA_HEADER
#include "oct.h" 
#include "defs.h"

typedef enum {AV_ANY , AV_NONE, AV_CELL, AV_FLOAT_TERM ,  AV_NET, AV_SUBNET } annealVarType;
typedef enum StatusOfCell {FIXED, MOVABLE, TEMP_BLOCKED} CellStatus;
/*
 * if a cell is OWNED the owner can move it
 * if it is BORROWED the owner cannot move it
 * if it is RETURNED the owner must get it again
 */
typedef enum OwnStatus  {OWNED, BORROWED, RETURNED} OwnStatus;
typedef enum PowerStatus  {SIGNAL, CLOCK, SUPPLY} PowerStatus;
typedef enum AnalogStatus  {SYMMETRIC, MATCHED, MULTIPLE_MATCHED, BOTH, BOTH_MULTIPLE, NONE} AnalogStatus;
typedef enum AbutStatus  {NON_ABUTTABLE, ABUTTABLE, CANDIDATE, ABUTTED} AbutStatus;
typedef enum ConstraintStatus  {NON_CONSTRAINED, CONSTRAINED} ConstraintStatus;
typedef enum SymmetryType  {NO_SYMMETRY, NO_THIS_SYMMETRY, VERTICAL, HORIZONTAL} SymmetryType;
typedef enum BiasStatus  {INDEPENDENT,VDD,VSS,SELF} BiasStatus;
typedef enum ReadType {READ, RE_READ} ReadType;
typedef enum SpecStatus {MET, NOT_MET} SpecStatus;

typedef struct listElt		ListElt;
typedef struct subnet 		SubNet;
typedef struct MacroCell 	MacroCell;
typedef struct MacroCell 	*MacroCellPointer;
typedef struct net		Net;
typedef struct term 		Term;
typedef struct cost_struct 	Cost;
typedef struct move_struct 	Move;
typedef struct powerDissip      PowerDissip;

#define FIXED_COOLING_SCHEDULE		0
#define ADAPTIVE_COOLING_SCHEDULE	1
#define EMPIRICAL_COOLING_SCHEDULE	2

#define INFINITE_TEMPERATURE		-1.0


/* general structures for the annealing algorithm */
typedef enum {  
  NO_MOVE,
  TRANSLATION,
  RL_TRANSLATION,		/* Range limited translation */
  TRANSFORMATION,
  TRANSLATE_AND_TRANSFORM,
  ALIGNMENT,
  CC_INTERCHANGE,
  BB_INTERCHANGE,
  MULTIPLE_CELL,
  TERMINAL_MOVE,
  SUBNET_MOVE,
  CHANGE_OWNERSHIP,		/* only for parallel applications */
  NUMBER_OF_MOVE_TYPES		/* not a move, just a label */
  } MoveType;



/* Annealing variables */
typedef struct AnnealVar_struct {
  annealVarType type;
  union {
    MacroCell	*cell;
    Term	*term;
    Net		*net;
    SubNet	*subnet;
  } pointer;
  int	owner;
  struct AnnealVar_struct	*ownerNext;
  struct AnnealVar_struct	*ownerPrec;
} AnnealVar;



struct cost_struct {
  double	overlap;	/* overlap */
  double	area;		/* areaCost / areaWeight */
  double	wireLength;	/*  */
  double	cluster;	/*  */
  int   	conViol;        /* counter of all contstraints violation */

  double        symmetry;       /* cost due to asymmetric placement of analog cells */
  double        matching;       /* cost due to analog cells mismatching */ 
  double        well;           /* cost due to well discontinuity */ 
  double        constraint;     /* cost due to constraint violations */

  double	total;		/* this field must be computed from the other fields */
};

struct moveStat {
  MoveType 	type;
  char		*name;
  int		enabled;		/* Enable flag */
  double 	relativeFrequency;	/*  */
  int		attempted;		/* number of attempted moves of this type */
  int		accepted;		/* number of accepted  moves of this type */
  double 	acceptanceRatio;	/* accepted over attempted */
  Cost  	sumCost;		/* accumulated sum of accepted moves of this type */
  Cost  	sumAbsCost;		/* ... */
  double 	qf;			/* quality factor */
  double	*movingAverageArray;
};

/* This sstructure is used for efficient selection of a move */
struct moveSelect {
  MoveType type;		/* type of move */
  int		index;		/* in the corresponding moveStatArray */
  double	freq;		/* frequency at which this type of move should be selected */
  double	thresh;		/* threshold, used by the selectMove() routine */
};


typedef struct {
  int	size;			/* size of the move set (no. of moves in the set) */
  struct moveStat   globalStat;
  struct moveStat   *moveStatArray;
  struct moveSelect *moveSelArray; /* for quick selection */
} MoveSet;

/* define all structures used to represent the placement */
/* nets, terminals, cells */

/* Generic list element */

struct listElt {
  union {
    char	*generic;	/* A generic pointer to something */
    MacroCell	*cell;
    Term	*term;
    Net		*net;
    SubNet	*subnet;
    AnnealVar	*annealVar;
  } pointer;
  struct listElt	*next;
  struct listElt	*prec;
}; 


/* soon to be replaced by listElt */
struct cellList{		/* represent sets of cells*/
    struct MacroCell *cell;
    struct cellList *prec;
    struct cellList *next;
};

/* soon to be replaced by listElt */

struct termList {
    Term 	*term;
    struct termList 	*prec;
    struct termList 	*next;
};
/* soon to be replaced by listElt */
struct netList{			/* repr. sets of nets */
    Net 		*net;
    struct netList 	*prec;
    struct netList 	*next;
};



struct subnet {
  Term 	*curTerm[2];		/* two terminals attached to the subnet */
  Term 	*newTerm[2];		/* two terminals attached to the subnet */
  double curAngle[2];		/* See note below  */
  double newAngle[2];		/* See note below  */
#ifdef PARALLEL
  int  	softLock;		/* Locked if the subnet is being moved */
#endif
};

struct net  {
  int 		number;		/* unique id for the net */
  char 		*name;		/* name of the net */
  int 		size;		/* number terminals on the net */
  int 		weight;		/* */
  int		maxLength;	/* maximum estimated length */
  struct cellList cell0;	/* handle for list of all cells */
  struct termList term0;	/* handle for list of all terminals */
  Term          *padTerm;       /* pointer to the terminal of a pad */
  double        peakCurrent;    /* peak current tolerated by net */
  double        normPeakCurrent;/* normalized peak curr. tolerated by net */
  int		subnets;	/* number of subnets */
  SubNet	*subnetArray;
  ListElt	*subnetLEArray;
  Net 		*globalNext;	/* pointer for list of all nets */
  Net 		*globalPrec;	/* pointer for list of all nets */
  int		age;
  PowerStatus   type;           /* signal, supply, clock */
  AbutStatus    status;         /* abuttability status:
				 * non_abuttable, abuttable, candidate */ 
  Net           *matchedNet;    /* net to which this net is matched by a symmetry constraint */
  struct Sensitivity *sensitivity; /* this structure contains the sensitivity info
				    * about the net */  
  SpecStatus    specStatus;     /* specs MET, NOT_MET */
  ConstraintStatus constraintStatus; /* NON_CONSTRAINED, CONSTRAINED */
};

struct Sensitivity {
  char                 *performanceName;
  int                  performanceId;
  double               plus;
  double               minus;
};


struct s_site {
  octPoint	cellCoord;		/* coordinate of this site wrt center of the cell */
  octPoint 	curAbsCoord;		/* layout coordinates */
  octPoint 	newAbsCoord;		/* layout coordinates */
  int		edge;			/* edge of the site */
  double	relPosition;		/* relative position in the edge in [0.0 , 1.0] */
  int		terms;			/* terminals currently attached to this site */
  int		capacity;		/* Max number of floating terminals on this site */
};

enum StatusOfTerm {UNUSED, USED};

struct term {
    char 		*name;		/* of the terminal */
    MacroCell	 	*cell;		/* to which the terminal belongs */
    int			floating; 	/* TRUE for a floating terminal */
    int			edge;		/* edge of the cell on which the term appears */
    double              peakCurrent;    /* max. current absorbed by term */
    double              normPeakCurrent;/* norm. current absorbed by term */
    octCoord            minWidth;       /* min width of wiring connected to it*/
    int			illegalEdges; 	/*  */
    int			curSite; 	/* current site of a floating terminal */
    int			newSite; 	/* new site of a floating terminal */
    octPoint 		cellCoord;	/* coordinates with respect to the center of the cell */
    octPoint 		*curAbsCoordP;	/* layout coordinates */
    octPoint 		*newAbsCoordP;	/* layout coordinates */
    enum StatusOfTerm	status;		/* used or not */
    AbutStatus          abutStatus;     /* non-abuttable, abuttable, candidate */
    PowerStatus         type;           /* SIGNAL, SUPPLY, CLOCK */
    Term		*equivNext; 	/* Implement a double linked list of equivalent terms */
    Term		*equivPrec; 	/* Implement a double linked list of equivalent terms */
    Net  		*net;		/* attached to the terminal */
    int			subnets; 	/* current number of subnets attached to this term */
    ListElt		*firstSubnet;   /* the subnets attached to this term */
    int			age;		/* age of a floating terminal in a site */
};

/*
 *  NOTE ON THE FIELD "ANGLE" IN THE TERM STRUCTURE: curAngle and newAngle
 *  "angle" is the angle of the subnet attached to the terminal, with respect
 *  to the normal to the edge the terminal belongs to.
 *  Thus an angle 0.0 means that the subnet is exiting out perpendicular to
 *  the edge. A angle 180.0 means that the subnet is going over the cell.
 *  The angle is measured counterclockwise.
 *  The angle will be used to improve the routing area estimation
 */



struct MacroCell {
  int		number;			/* unique local id for the cell */
  octId 	cellId;			/* octId of the instance represented by this MacroCell */
  octId		masterId;
  char		*name;			/* name of the instance */
  int		isPad;			/* TRUE if this is a pad */
  int		isConnector;		/* TRUE if this is a connector cell */
  int		edgesInPerim;		/* edges in perimeter */
  int		*termsInEdge;		/* (array) number of terminals coming 
					   out of each edge */
  int           *widthOfEdge;           /* (array) edge width: number of equiv.
					   min-geometry wires coming out of each
					   edge */
  int		lowestLeftCorner; 	/* in perimeter */
  octPoint	masterLLC;		/* lowestLeftCorner in master, to store data back to OCT */
  octPoint	*perimeter;		/* with respect to the center of the BBox */
  octPoint	*curPerimeter;		/* absolute coordinates */
  octPoint	*newPerimeter;		/* absolute coordinates */
  int		numberOfSites;
  struct s_site	*floatingSites;		/* sites for floating terminals */
#ifdef PARALLEL
  slock_t	floatingSitesLock; 	/*  */
#endif
  octBox 	newBbox;		/* new bounding box */
  octBox 	curBbox;		/* current bounding box */
  double 	area;			/* actual area of the cell */
  octCoord	W;			/* unexpanded width */
  octCoord	H;			/* unexpanded height */
  octCoord 	rW;			/* expanded width  */
  octCoord 	rH;			/* expanded height */
  octCoord	rHalfW;			/* half expanded width */
  octCoord	rHalfH;			/* half expanded height */
  int		totNets;		/* number of nets connected */
  enum OwnStatus status;		/* owned or borrowed etc.*/
  CellStatus	mobility;		/* movable or fixed */
  struct moveStat   *moveStatArray;
  struct moveSelect *moveSelArray;
  octCoord	rangeLimiter;
  int		legalTransforms; 	/* this is a bit mask */
  int		numberOfLegalTransforms;
  int		*legalTransformsArray;
  int		forbiddenEdgeMask;	/* 	 */
  int		owner;			/* Parallel version only */
  int		age;			/* not yet used */
  int		moveId;			/* each move has its own id */
  octTransformType curTransform;	/* current transform */
  octTransformType newTransform;	/* new transform */
  octPoint		centerOffset;
  octPoint 		curCenter;	/* current center of the cell*/
  octPoint		newCenter;	/* new center of the cell */
  int			*curBin;	/* current bin for the cell */
  struct netList 	net0;		/* handle for list of all nets */
  struct termList 	term0;		/* handle for list of terminals */
  struct termList	*lastMovedTermLE;	/*  */
  int			floatingTerms; 	/* Number of floating terminals */
  struct termList	floatTerm0; 	/* handle for list of floating terminals */
  struct MacroCell	*globalNext; 	/* linked list of all cells */
  struct MacroCell	*globalPrec;  	/* linked list of all cells */
#ifdef OBSOLETE
  struct MacroCell	*ownerNext;  	/* linked list of cells owned by the same processor*/
  struct MacroCell	*ownerPrec;  	/* linked list of cells owned by the same processor*/
#endif
  struct MacroCell	*moveNext; 	/* for complex moves */
  struct MacroCell	*movePrec; 	/* for complex moves */
  struct MacroCell	*lyoPrec[4];	/* pointers to implement the LYO data structure */
  struct MacroCell	*lyoNext[4];	/* pointers to implement the LYO data structure */
  struct SymmetricCouple *ySymPointer;  /* pointer to the vert. symmetry data-base */
  struct SymmetricCouple *xSymPointer;  /* pointer to the hor. symmetry data-base */
  struct MacroCell     *matchNextCell;  /* pointer to the next matched cell in the cluster */
  struct MacroCell         *matchLink;  /* pointer from the MatchCell data-base to the gFirstCell data-base */
  struct MatchingCluster *matchPointer;  /* pointer to the matching data-base */
  enum AnalogStatus     analogStatus;    /* flag containing info about the analog
                                          * costraints of the cell */ 
  enum BiasStatus       biasStatus;      /* biasing of the cell:
					  * Vdd,Vss,self,independent */
  enum AbutStatus       abutStatus;      /* candidate or not to abutment:
					  * non_abuttable, abuttable, candidate */ 
  PowerDissip           *powerDissip;    /* peakCurrent, effectiveCurrent */

};

struct SymmetricCouple  {
  octCoord          axisPosition;      
  enum SymmetryType     axisType;
  MacroCell               *cell1;
  MacroCell               *cell2;
};

struct MatchingCluster  {
  int                  ClusterLimitDistance;
  double                   ClusterInnerCost;  /* cost of placement of cluster */
  int               numberOfMatchingCluster;
  struct MatchingCluster  *matchNextCluster;  
  int                  numberOfMatchingCell;
  MacroCell              *MatchingFirstCell;
  struct MatchingCluster           *Cluster;  
};

struct PowerDissip  {
  double                        peakCurrent;
  double                   effectiveCurrent;
  Term                             *termVDD;
  Term                             *termVSS;
};
  



struct move_struct {
  MoveType 	type;
  int		index;
  int		complexity;
  AnnealVar	*annealVar;
  MacroCell 	*cell;		/* list of all the moving cells */
  MacroCell	*cell2;		/* second cell in pairwise interchanges */
  Term		*term;		/* list of terms changing status (floating, equivalent) */
  Net		*net;		/* list of all nets affected in the move */
  SubNet	*subnet;	/*  */
  SubNet	*fixedSubnet;
  ListElt	*subnetLE;
  int		subnetIndex;
  Cost		cost;		/* cost of the move */
  int		procNumber;	/* local Id of the other process interacting for this move */
};




#define SZF_CELLLIST sizeof(struct cellList)
#define SZF_NETLIST  sizeof(struct netList)
#define SZF_TERMLIST sizeof(struct termList)



/* structure to compute the clustering of cells owned by one processor */
struct cluster_t {
#ifdef PARALLEL
  slock_t lock;
#endif
  double totalArea;
  double sumXArea;
  double sumYArea;
  double sumX2Area;
  double sumY2Area;
  double barX;			/* baricenter X coord */
  double barY;
  double momentum;
};

#endif























