#include "copyright.h"
#include "sparcs.h"
#include "oct.h"
#include "oh.h"
#include "facet.h"
#include "vov.h"
#include "th.h"

/*
 *  This routine performs the oct calls necessary to prepare the appropriate
 *  facet for reading.
 *
 *  The steps are:
 *	1.  open the input view (named InView)
 *	2.  copy it to the output view (named OutView)
 *	3.  open the output view and read it via readOct()
 *
 *  Calls:  checkPolicy(), readOct()
 *
 *  1/9/86
 *  last modified 9/22/88
 */

void
initOct()
{
    octObject inFacet;
    octObject techProp;
    octObject dataBag;
    octObject cpBag;
    octStatus status;
    struct object *po;
    struct wire *pw;
    void checkPolicy();
    char *techName = NIL(char);
    int debug = 0;

    inFacet.type = OCT_FACET;
    techProp.type = OCT_PROP;
    dataBag.type = OCT_BAG;
    cpBag.type = OCT_BAG;

    if (debug)  {
	fprintf(pout, "\nReading \"%s\" facet of cell \"%s\", with\n",
								Facet, InCell);
	fprintf(pout, "InView \"%s\" and OutView \"%s\"\n\n", InView, OutView);
    }

    /*
     *  Open the working facet according to whether the input facet is being
     *  overwritten or not.  If the input and output facets differ, then first
     *  copy the input facet to the output facet.  If they are the same, then
     *  remove any old critical-path data that might be present.
     */
    if (strcmp(InCell, OutCell)==0 && strcmp(InView, OutView)==0)  {
	WorkingFacet.type = OCT_FACET;
	WorkingFacet.contents.facet.cell = InCell;
	WorkingFacet.contents.facet.view = InView;
	WorkingFacet.contents.facet.facet = Facet;
	WorkingFacet.contents.facet.version = OCT_CURRENT_VERSION;
	WorkingFacet.contents.facet.mode = "a";

	VOVinputFacet( &WorkingFacet );
	VOVoutputFacet( &WorkingFacet );


	if (TraceFlag)  {
	    fprintf(pout, "\nOpening working facet %s...  ", ohFormatName( &WorkingFacet ));
	    fflush(pout);
	}

	if (octOpenFacet(&WorkingFacet) != OCT_OLD_FACET)  {
	    octError("initOct:  can't open the WorkingFacet.");
	    VOVend(-1);
	}
	if (TraceFlag)  {
	    fprintf(pout, "\n");
	    fflush(pout);
	}
	
	/*
	 *  The old critical-path data is invalid so remove it.  Check for the
	 *  SPARCS-DATA bag.  If present, look for old critical-path bags,
	 *  which are in turn deleted if present.
	 */
	dataBag.contents.bag.name = "SPARCS-DATA";
	if ((status=octGetByName(&WorkingFacet, &dataBag)) == OCT_OK)  {
	    cpBag.contents.bag.name = "X-CRITICAL-PATH";
	    if ((status=octGetByName(&dataBag, &cpBag)) == OCT_OK)  {
		if (octDelete(&cpBag) != OCT_OK)  {
		    printf("initOct:  can't delete X-CRITICAL-PATH bag.\n");
		    octError("");
		    exit(-1);
		}
	    } else if (status != OCT_NOT_FOUND)  {
		printf("initOct:  can't getByName X-CRITICAL-PATH bag.\n");
		octError("");
		exit(-1);
	    }
	    cpBag.contents.bag.name = "Y-CRITICAL-PATH";
	    if ((status=octGetByName(&dataBag, &cpBag)) == OCT_OK)  {
		if (octDelete(&cpBag) != OCT_OK)  {
		    printf("initOct:  can't delete Y-CRITICAL-PATH bag.\n");
		    octError("");
		    exit(-1);
		}
	    } else if (status != OCT_NOT_FOUND)  {
		printf("initOct:  can't getByName Y-CRITICAL-PATH bag.\n");
		octError("");
		exit(-1);
	    }
	} else if (status != OCT_NOT_FOUND)  {
	    printf("initOct:  can't getByName SPARCS-DATA bag.\n");
	    octError("");
	    exit(-1);
	}
    } else  {
	/*
	 *  Since they differ do a copy first.
	 */
	inFacet.contents.facet.cell = InCell;
	inFacet.contents.facet.view = InView;
	inFacet.contents.facet.facet = Facet;
	inFacet.contents.facet.version = OCT_CURRENT_VERSION;
	inFacet.contents.facet.mode = "r";

	WorkingFacet.type = OCT_FACET;
	WorkingFacet.contents.facet.cell = OutCell;
	WorkingFacet.contents.facet.view = OutView;
	WorkingFacet.contents.facet.facet = Facet;
	WorkingFacet.contents.facet.version = OCT_CURRENT_VERSION;
	WorkingFacet.contents.facet.mode = "a";

	if (octCopyFacet(&WorkingFacet, &inFacet) != OCT_OK)  {
	    octError("initOct:  can't copy into the WorkingFacet.");
	    exit(-1);
	}
	if (TraceFlag)  {
	    fprintf(pout, "\nOpening working facet %s...  ", ohFormatName( &WorkingFacet ));
	    fflush(pout);
	}
	VOVinputFacet(  &inFacet );
	VOVoutputFacet( &WorkingFacet );
	if (octOpenFacet(&WorkingFacet) != OCT_OLD_FACET)  {
	    octError("initOct:  can't open the WorkingFacet.");
	    exit(-1);
	}
	if (TraceFlag)  {
	    fprintf(pout, "\n");
	    fflush(pout);
	}
    }


    thTechnology( &WorkingFacet );
    UNITSPERLAMBDA = thLambda();
    /* fprintf( stderr,  "Lambda is %d oct units\n", UNITSPERLAMBDA ); */



    /*
     *  Read the TECHNOLOGY property so that the rules file can be found.
     */
    if ( ohGetByPropName( &WorkingFacet, &techProp, "TECHNOLOGY") != OCT_OK)  {
	errRaise("initOct", 1, "no TECHNOLOGY property present.\n");
    } 

    techName = STRSAVE(techProp.contents.prop.value.string);

    /*
     *  Read the technology data (spacing rules and resistance info).
     */
    openTech(&WorkingFacet, techName);

    /*
     *  Now invoke readOct() to run the generators.
     */
    /*********
    if (RicksReaderFlag)  {
        ricksReadOct();
	printf("ricksReadOct temporarily removed...\n");
	exit(0);
    } else  {
	checkPolicy(&WorkingFacet);
        readOct();
    }
    *********/
    checkPolicy(&WorkingFacet);

    readOct();
    
    ElementByKey = ALLOC(struct element, NumElements+1);
    for (po=pObjectList; po!=NIL(struct object); po=po->pnext)  {
	ElementByKey[po->key].ptr.po = po;
    }
    for (pw=pWireList; pw!=NIL(struct wire); pw=pw->pnext)  {
	ElementByKey[pw->key].ptr.pw = pw;
    }

    return;
}
