#include "copyright.h"
#include "sparcs.h"
#include "oct.h"
#include "facet.h"
#include "oh.h"
#include "tr.h"

#define NONE	1
#define BOTH	2
#define FIRST	3
#define SECOND	4

#define PRINT_SEGMENT(sparcsLayer, points) \
    printf("    segment on %s, endpoints (%d, %d) (%d, %d),", \
	   LayerMap[sparcsLayer].name, points[0].x, \
	   points[0].y, points[1].x, points[1].y); \
    printf(" width %d.\n", segment.contents.path.width)

/*
 *  Reads the inputted symbolic layout into the internal structs of SPARCS.
 *  The comments below describe the algo.
 *
 *  Modified 5/16/87 to read the terminal frames from the f-terms, instead of
 *  from the layers.
 *
 *  Called by:  stageOct
 *  Calls:	procPFrames(), readTFrames()
 *
 *  8/2/86
 *  last modified 4/14/88
 */

char buffer[256];

void
readOct()
{
    octObject inst;
    octObject master;
    octObject layer;
    octObject segment;
    octObject term;
    octObject aterm1, aterm2;
    octObject net;
    octGenerator instGenId,
		 layerGenId,
		 termGenId,
		 segGenId,
	         netGenId;
    octStatus instGenStatus,
	      layerGenStatus,
	      termGenStatus,
	      segGenStatus,
	      netGenStatus,
	      status;
    int netId;

    tr_stack *pstack;			/*  for tr package  */
    struct octTransform *pxform;	/*  for the instance's transform  */

    int netCount;
    int end1Intersect, end2Intersect;
    int findTFrame();
    directionType wireDir;
    
    int numInstances = 0;
    int numMasters = 0;
    int numSegments = 0;

    int dupSeg();
    
    int debug = 0;

    /***
    int cnt;
    ***/
    struct edgePtr *peP;

    struct object *pobj, *makeObject();
    struct object *po1, *po2;
    struct wire *makeWire(), *pw;
    struct terminal *pt1, *pt2;

    int32 numPoints = 2;
    struct octPoint points[2];

    void SpoctMakeFrame();
    void SPmakeFramesRelative();
    /* int mkEdgeList(); */

    int sparcsLayer;			/*  internal layer number  */
    int i;
    /***
    int thinLayer;
    int fatLayer;
    ***/
    int numTerms, numFrames;
    int fatalErrorFlag = FALSE;

    inst.type = OCT_INSTANCE;
    master.type = OCT_FACET;
    layer.type = OCT_LAYER;
    segment.type = OCT_PATH;
    term.type = OCT_TERM;
    aterm1.type = aterm2.type = OCT_TERM;
    net.type = OCT_NET;
    
    pstack = tr_create();

    if (TraceFlag)  {
	fprintf(pout, "Reading facet/performing setup...  ");
	fflush(pout);
    }

    if (ReadDB)  {
	fprintf(pout, "\nReading facet \"%s\" ", Facet);
	fprintf(pout, "of view \"%s\" of cell \"%s\"\n\n", OutView, InCell);
    }

    /*
     *  The first part of this routine reads all information associated with
     *  the instances in "WorkingFacet" (which is a global).  First, a gener-
     *  ator is executed until all instances are returned.  As each instance
     *  is found its corresponding master is opened.  The layers of the master
     *  are generated.  For each layer, all geometries on the layer are gen-
     *  erated.  Protection frames are contained by the layer only, whereas
     *  terminal frames are contained by the layer and by the terminal defini-
     *  tion.  So, the containing terminals of each frame are generated to dis-
     *  tinguish the terminal frames from the protection frames.
     */
    if (octInitGenContents(&WorkingFacet, OCT_INSTANCE_MASK, &instGenId) !=
								    OCT_OK)  {
	octError("readOct:  can't begin the instance generator.");
	exit(-50);
    }
    while ((instGenStatus=octGenerate(&instGenId, &inst)) == OCT_OK)  {

	++numInstances;
	numFrames = 0;
	numTerms  = 0;

	if (ReadDB)  {
	    fprintf(pout, "    instance %d, id %s, name \"%s\", ",
					    numInstances,
					    octFormatId(inst.objectId),
					    inst.contents.instance.name);
	    fprintf(pout, "master cell \"%s\"\n",
					    inst.contents.instance.master);
	}

	/*
	 *  Open the master for this instance
	 */
	master.contents.facet.facet = "interface";
	master.contents.facet.version = OCT_CURRENT_VERSION;
	master.contents.facet.mode = "r";
	status = octOpenMaster(&inst, &master);
	VOVinputFacet( &master );
	if (status != OCT_OLD_FACET && status != OCT_ALREADY_OPEN)  {
	    sprintf(buffer, "readOct: can't open master %s/%s/%s.",
		    inst.contents.instance.master,
		    inst.contents.instance.view,
		    master.contents.facet.facet);
	    octError(buffer);
	    exit(-50);
	}

	/*
	 *  Make a new sparcs object and increment the object count.
	 */
	pobj = makeObject(&inst);
	NumObjects++;
	NumElements++;

	/*
	 *  Set-up for the coordinate transforms.  Reset pstack, then compute
	 *  whatever transform is present.  Then, can pass pstack to the frame
	 *  routines and each of them can find absolute coordinates by using
	 *  pstack and tr_transform.
	 */
	tr_identity(pstack);		/*  reset the transform  */

        pxform = &(inst.contents.instance.transform);
	tr_translate(pstack,pxform->translation.x,pxform->translation.y,ZERO);
	if (pxform->transformType == OCT_FULL_TRANSFORM)  {
	    tr_rotate_matrix(pstack, pxform->generalTransform, ZERO);
	} else {
	    tr_do_op(pstack, pxform->transformType, ZERO);
	}

	/*
	 *  Generate each layer of the master, and read its p-frames.
	 */
	if (octInitGenContents(&master, OCT_LAYER_MASK, &layerGenId) !=
								    OCT_OK)  {
	    octError("readOct:  can't begin layer generator.");
	    exit(-50);
	}
	while ((layerGenStatus=octGenerate(&layerGenId, &layer))==OCT_OK)  {
	    procPFrames(&layer, pobj, pstack, &numFrames);
	}
	if (layerGenStatus != OCT_GEN_DONE)  {
	    octError("readOct:  Bad exit status from layer generator.");
	    exit(-50);
	}
	/*
	 *  Need to know whether the instance has p-frames, since determines
	 *  whether the t-frames get middle edgePtrs; do the check here.
	 */
	if (pobj->phorizontalEdges != NIL(struct edge))  {
	    pobj->hasPFrames = 'T';
	}

	/*
	 *  Generate each f-term of the master, and read its t-frames.
	 */
	if (octInitGenContents(&master, OCT_TERM_MASK, &termGenId) != OCT_OK) {
	    octError("readOct:  can't begin f-term generator.");
	    exit(-1);
	}
	while ((termGenStatus=octGenerate(&termGenId, &term))==OCT_OK)  {
	    numTerms++;
	    readTFrames(&term, pobj, &inst, pstack);
	}
	if (termGenStatus != OCT_GEN_DONE)  {
	    octError("readOct:  Bad exit status from f-term generator.");
	    exit(-1);
	}

	if (numTerms == 0  &&  numFrames == 0)  {
	    printf("readOct:  inst with no terminal or protection frames.\n");
	    fatalErrorFlag = TRUE;
	    continue;
	}
	if (numTerms == 0)  {
	    printf("readOct:  WARNING - inst w/out terminals.\n");
	}
	/*
	if (numFrames == 0)  {
	    printf("readOct:  inst w/out protection frames.\n");
	}
	*/
	/*
	 *  Done creating the current sparcs object, and its bounding box has
	 *  been found.  All coordinates of the object's frames are absolute
	 *  at this point, so make them relative to the object's origin.
	 *  Before making them relative, create the edge and edgePtr structs
	 *  used in the scanner; they must be in terms of absolute coordinates.
	 */
	/* cnt = mkEdgeList(pobj, NIL(struct wire), OBJECT, UNDEF); */

	/*  Count the edgePtrs in the list...  */
	/**********************
	cnt = 0;
	for (peP=pobj->pminHEP; peP!=NIL(struct edgePtr); peP=peP->plist)  {
	    cnt++;
	}
	printf("\ninstance has %d minHEPs.\n", cnt);
	cnt = 0;
	for (peP=pobj->pminVEP; peP!=NIL(struct edgePtr); peP=peP->plist)  {
	    cnt++;
	}
	printf("instance has %d minVEPs.\n\n", cnt);

	cnt = 0;
	for (peP=pobj->pmidHEP; peP!=NIL(struct edgePtr); peP=peP->plist)  {
	    cnt++;
	}
	printf("\ninstance has %d midHEPs.\n", cnt);
	cnt = 0;
	for (peP=pobj->pmidVEP; peP!=NIL(struct edgePtr); peP=peP->plist)  {
	    cnt++;
	}
	printf("instance has %d midVEPs.\n\n", cnt);

	cnt = 0;
	for (peP=pobj->pmaxHEP; peP!=NIL(struct edgePtr); peP=peP->plist)  {
	    cnt++;
	}
	printf("\ninstance has %d maxHEPs.\n", cnt);
	cnt = 0;
	for (peP=pobj->pmaxVEP; peP!=NIL(struct edgePtr); peP=peP->plist)  {
	    cnt++;
	}
	printf("instance has %d maxVEPs.\n\n", cnt);
	**********************/

	/**
	pobj->origin.x = pobj->relBB.lowerL.x;
	pobj->origin.y = pobj->relBB.lowerL.y;
	pobj->initOrigin.x = pobj->relBB.lowerL.x;
	pobj->initOrigin.y = pobj->relBB.lowerL.y;
	**/
	pobj->initOrigin.x = pobj->origin.x;
	pobj->initOrigin.y = pobj->origin.y;

	SPmakeFramesRelative(pobj);

	/*  update MaxBBX, MaxBBY  */
	if ((pobj->bbUpperR.x + pobj->origin.x) > MaxBBX)  {
	    MaxBBX = pobj->bbUpperR.x + pobj->origin.x;
	}
	if ((pobj->bbUpperR.y + pobj->origin.y) > MaxBBY)  {
	    MaxBBY = pobj->bbUpperR.y + pobj->origin.y;
	}
	if (pobj->origin.x < MinBBX)  {
	    MinBBX = pobj->origin.x;
	}
	if (pobj->origin.y < MinBBY)  {
	    MinBBY = pobj->origin.y;
	}

	/*  link it into the ObjectList  */
	pobj->pnext = pObjectList;
	pObjectList = pobj;
    }
    if (instGenStatus != OCT_GEN_DONE)  {
	octError("readOct:  Bad exit status from instance generator.");
	exit(-50);
    }
    /*
     *  All the instances of WorkingFacet have been read in.
     */
    if (ReadDB)  {
	fprintf(pout, "\nnumber of masters present is %d\n", numMasters);
	fprintf(pout, "number of instances present is %d\n\n", numInstances);
    }

    /*
     *  If there are no instances then quit.
     */
    if (numInstances <= 0)  {
      	fprintf(pout, "readOct:  Cell has no instances!  exiting.\n");
	exit(-50);
    }

    tr_free(pstack);

    /*
     *  In this second major part the wire segments and connectivity is read.
     *  A generator is invoked on WorkingFacet (again) to return all layers.
     *  On each layer all octPath records are read.  For each path a contents
     *  generator returns the two actual terminals (definitions) that the path
     *  is connected to.
     */
    if (octInitGenContents(&WorkingFacet, OCT_LAYER_MASK, &layerGenId) !=
								    OCT_OK)  {
	octError("readOct:  can't begin layer generator");
	exit(-50);
    }
    while ((layerGenStatus = octGenerate(&layerGenId, &layer)) == OCT_OK)  {

	/*
	 *  Get the internal layer number out of LayerMap.
	 */
	for (i=0; i<NumLayers; i++)  {
	    if (strcmp(LayerMap[i].name, layer.contents.layer.name) == 0)  {
		sparcsLayer = i;
		break;
	    }
	}
	if (i >= NumLayers)  {
	    /*
	     *  Layer isn't present in the rules file.  Ignore it iff there are
	     *  no paths on it, otherwise fatal error.
	     */
	    if ((segGenStatus=octGenFirstContent(&layer, OCT_PATH_MASK,
						 &segment)) == OCT_GEN_DONE)  {
		;
	    } else if (segGenStatus == OCT_OK)  {
		printf("readOct:  layer %s ", layer.contents.layer.name);
		printf("contains paths but is not in rules file.\n");
		fatalErrorFlag = TRUE;
	    } else  {
		sprintf(buffer, "readOct:  trying to genFirst path on %s",
			layer.contents.layer.name);
		octError(buffer);
		fatalErrorFlag = TRUE;
	    }
	    continue;
	}

	/*
	 *  Generate the segments on the current layer.
	 */
	if (octInitGenContents(&layer,OCT_PATH_MASK,&segGenId) != OCT_OK)  {
	    octError("readOct:  can't begin segment generator.\n");
	    exit(-50);
	}
	while ((segGenStatus=octGenerate(&segGenId,&segment))==OCT_OK)  {

	    NumWires++;
	    NumElements++;

	    if (ReadDB)  printf("found segment %d\n", ++numSegments);

	    /*
	     *  Grab its points, check for nonzero length and non-Manhattan.
	     */
	    if (octGetPoints(&segment, &numPoints, points) != OCT_OK)  {
		printf("readOct:  can't get points for segment on %s;",
		       layer.contents.layer.name);
		printf(" numPoints = %d\n", numPoints);
		numPoints = 2;
		fatalErrorFlag = TRUE;
		continue;
	    } else  {
		if (points[0].x == points[1].x)  {
		    if (points[0].y == points[1].y)  {
			octObject bag;
			printf("readOct:  zero-length wire segment found.\n");
			ohGetOrCreateBag( &WorkingFacet, &bag, "SPARCS ZERO LENGTH PATHS" );
			octAttach( &bag, &segment );
			fatalErrorFlag = TRUE;
			continue;
		    } else  {
			wireDir = VERTICAL;
		    }
		} else if (points[0].y == points[1].y)  {
		    wireDir = HORIZONTAL;
		} else   {
		    octObject bag;
		    printf("readOct:  non-Manhattan wire segment found.\n");
		    ohGetOrCreateBag( &WorkingFacet, &bag, "SPARCS NON-MANHATTAN PATHS" );
		    octAttach( &bag, &segment );
		    fatalErrorFlag = TRUE;
		    continue;
		}
	    }

	    /*
	     *  Segment must be in exactly one net.
	     */
	    if (octInitGenContainers(&segment, OCT_NET_MASK, &netGenId) !=
								    OCT_OK)  {
		octError("readOct:  can't begin net generator.\n");
		exit(-50);
	    }
	    netCount = 0;
	    while ((netGenStatus=octGenerate(&netGenId, &net))==OCT_OK)  {
		netId = net.objectId;
		netCount++;
	    }
	    if (netGenStatus != OCT_GEN_DONE)  {
		octError("readOct:  Bad exit from net generator.");
		exit(-1);
	    }
	    if (netCount == 0)  {
		printf("readOct:  segment isn't contained by a net;\n");
		PRINT_SEGMENT(sparcsLayer, points);
		fatalErrorFlag = TRUE;
	        continue;
	    } else if (netCount > 1)  {
		printf("readOct:  segment contained by > 1 net;\n");
		PRINT_SEGMENT(sparcsLayer, points);
		fatalErrorFlag = TRUE;
		continue;
	    }

	    /*
	     *  Find the aterms contained by this segment (subnet).  If there
	     *  are not exactly two, then complain.
	     */
	    if (octInitGenContents(&segment, OCT_TERM_MASK, &termGenId) !=
								    OCT_OK)  {
		octError("readOct:  can't begin segment term'l generator");
		PRINT_SEGMENT(sparcsLayer, points);
		fatalErrorFlag = TRUE;
		continue;
	    }
	    if ((status=octGenerate(&termGenId, &aterm1)) != OCT_OK)  {
		if (status == OCT_GEN_DONE)  {
		    printf("readOct:  no aterms contained by segment;\n");
		} else  {
		    octError("readOct:  can't get aterms of segment");
		}
		PRINT_SEGMENT(sparcsLayer, points);
		fatalErrorFlag = TRUE;
		continue;
	    }
	    if ((status=octGenerate(&termGenId, &aterm2)) != OCT_OK)  {
		if (status == OCT_GEN_DONE)  {
		    printf("readOct:  only one aterm contained by segment;\n");
		} else  {
		    octError("readOct:  can't get second aterm of segment");
		}
		PRINT_SEGMENT(sparcsLayer, points);
		fatalErrorFlag = TRUE;
		continue;
	    }
	    if (octGenerate(&termGenId, &aterm2) != OCT_GEN_DONE)  {
		printf("readOct:  more than 2 aterms for segment\n");
		PRINT_SEGMENT(sparcsLayer, points);
		fatalErrorFlag = TRUE;
		continue;
	    }

	    /*
	     *  Find the corresponding sparcs terminals that intersect the
	     *  ends of the segment.
	     */
	    end1Intersect = findTFrame(points, sparcsLayer, netId,
				       aterm1.objectId,
				       aterm1.contents.term.instanceId,
				       &po1, &pt1);
	    end2Intersect = findTFrame(points, sparcsLayer, netId,
				       aterm2.objectId,
				       aterm2.contents.term.instanceId,
				       &po2, &pt2);

	    /*
	     *  This should not happen - a segment with both ends attached to
	     *  the same aterm.  But since it is possible, check for it.
	     */
	    if (pt1 == pt2)  {
		printf("readOct:  wire connected twice to the same aterm\n");
		PRINT_SEGMENT(sparcsLayer, points);
		fatalErrorFlag = TRUE;
		continue;
	    }
	     
	    /*
	     *  end*Intersect can be NONE, FIRST, SECOND, or BOTH.  Of the
	     *  16 possible combinations,  4 are illegal:
	     *	    end1Intersect == NONE
	     *      end2Intersect == NONE
	     *	    end1Intersect == FIRST   &&  end2Intersect == FIRST
	     *	    end1Intersect == SECOND  &&  end2Intersect == SECOND
	     *  All others are ok.
	     */
	    if (end1Intersect == NONE)  {
		printf("readOct:  no connection for first end of wire\n");
		PRINT_SEGMENT(sparcsLayer, points);
		fatalErrorFlag = TRUE;
		continue;
	    }
	    if (end2Intersect == NONE)  {
		printf("readOct:  no connection for second end of wire\n");
		PRINT_SEGMENT(sparcsLayer, points);
		fatalErrorFlag = TRUE;
		continue;
	    }
	    if (end1Intersect == FIRST  &&  end2Intersect == FIRST)  {
		/*
		 *  Both intersections are FIRST, so error.
		 */
		printf("readOct:  both ends of segment intersect first");
		printf(" term, neither intersects second\n");
		PRINT_SEGMENT(sparcsLayer, points);
		fatalErrorFlag = TRUE;
		continue;
	    } else if (end1Intersect == SECOND  &&  end2Intersect == SECOND)  {
		/*
		 *  Both intersections are SECOND, so error.
		 */
		printf("readOct:  both ends of segment intersect second");
		printf(" term, neither intersects second\n");
		PRINT_SEGMENT(sparcsLayer, points);
		fatalErrorFlag = TRUE;
		continue;
	    }

	    /*
	     *  See if there are duplicate segments between these aterms on
	     *  this layer.
	     */
	    if (dupSeg(&aterm1, &aterm2, &segment, layer.contents.layer.name))  {
		printf("\n\tERROR: several %s segments connect the same two terminals",
		       LayerMap[sparcsLayer].name);
		{
		    octObject bag;
		    ohGetOrCreateBag( &WorkingFacet, &bag, "SPARCS ILLEGAL PATHS" );
		    octAttach( &bag, &segment );
		}
		fatalErrorFlag = TRUE;
		continue;
	    }
	    
	    /*
	     *  All tests passed, so create the wire, trim its length to fit
	     *  the terminals, and create its edges.
	     */
	    pw = makeWire(&segment, points, wireDir, sparcsLayer);
	    trimWire(&segment, pw, pt1, pt2, po1, po2);
	    makeWireEdges(pw, netId);

	    /*
	     *  Update MaxBBX, MaxBBY.  (This shouldn't be necessary...  BB
	     *  determined by instances)
	     */
	    if (pw->horizontal == TRUE)  {
		if (pw->p2.x > pw->p1.x)  {
		    if (pw->p1.x < MinBBX)  MinBBX = pw->p1.x;
		    if (pw->p2.x > MaxBBX)  MaxBBX = pw->p2.x;
		} else  {
		    if (pw->p2.x < MinBBX)  MinBBX = pw->p2.x;
		    if (pw->p1.x > MaxBBX)  MaxBBX = pw->p1.x;
		}
		if ((pw->p1.y + pw->halfWidth) > MaxBBY)
		    MaxBBY = pw->p1.y + pw->halfWidth;
		if (pw->p1.y - pw->halfWidth < MinBBY)
		    MinBBY = pw->p2.y - pw->halfWidth;
	    } else  {
		if (pw->p2.y > pw->p1.y)  {
		    if (pw->p1.y < MinBBY)  MinBBY = pw->p1.y;
		    if (pw->p2.y > MaxBBY)  MaxBBY = pw->p2.y;
		} else  {
		    if (pw->p2.y < MinBBY)  MinBBY = pw->p2.y;
		    if (pw->p1.y > MaxBBY)  MaxBBY = pw->p1.y;
		}
		if ((pw->p1.x + pw->halfWidth) > MaxBBX)
		    MaxBBX = pw->p1.x + pw->halfWidth;
		if (pw->p1.x - pw->halfWidth < MinBBX)
		    MinBBX = pw->p2.x - pw->halfWidth;
	    }
	}
	if (segGenStatus != OCT_GEN_DONE)  {
	    octError("readOct:  Bad exit from segment generator.");
	    exit(-50);
	}

	/*
	 *  XXX  Stick in the hack for the wide metal wires.
	 */
	/******
	if (strcmp(LayerMap[sparcsLayer].name, "MET1") == 0)  {
	    thinLayer = sparcsLayer;
	    for (i=0; i<NumLayers; i++)  {
		if (strcmp(LayerMap[i].name, "MET11") == 0)  {
		    fatLayer = i;
		    break;
		}
	    }
	    if (i >= NumLayers)  {
		printf("readOct:  can't find layer MET11 in LayerMap.\n");
		exit(-10);
	    }
	    for (pw=pWireList; pw!=NIL(struct wire); pw=pw->pnext)  {
		if (pw->maskNumber == thinLayer)  {
		    if (pw->width > 120)  {
			printf("readOct:  found a fat wire.\n");
			pw->maskNumber = fatLayer;
			pw->pfirstMaxEP->pedge->maskNumber = fatLayer;
			pw->plastMaxEP->pedge->maskNumber = fatLayer;
			pw->pfirstMinEP->pedge->maskNumber = fatLayer;
			pw->plastMinEP->pedge->maskNumber = fatLayer;
		    }
		}
	    }
	}
	*******/

    }
    if (layerGenStatus != OCT_GEN_DONE)  {
	octError("readOct:  Bad exit status from layer generator");
	exit(-50);
    }

    if (debug)  printf("\nreadOct:  saving new wire coords...\n");
    if (octFlushFacet(&WorkingFacet) != OCT_OK)  {
	octError("updateOct:  can't save the OutView.");
	exit(-1);
    }
    

    /*
     *  Done getting the nets/interconnect.
     */

    if (NumElements != NumObjects+NumWires)  {
	errRaise( "SPARCS", SPARCS_BUG_CODE, "readOct:  NumElements is incorrect!\n");
    }

    if (fatalErrorFlag)  {
	errRaise( "SPARCS", SPARCS_BUG_CODE, "Some serious errors are present.\n");
    }

    if (TraceFlag)  {
	fprintf(pout, "%d instances, %d segments.\n", NumObjects, NumWires);
	fflush(pout);
    }


    /*
     *  Read the user-constraints.
     */
    if (NoUserConstraintsFlag == FALSE)  {
	/*printf("\nReading user-constraints...\n");*/
	readUserConsts();
    }
    return;
}
