#include "copyright.h"
#include "sparcs.h"
#include "oct.h"
#include "facet.h"
#include "grid.h"
#include "oh.h"
#define GRIDOVERLAP 120

/*
 *  Routine to update the locations of the objects in oct.  After the
 *  octInstances and octPaths are updated, the OutView is updated via
 *  octFlushFacet().
 *
 *  Instance updating is controlled by "direction", ShowIterationsFlag, and
 *  SaveIterationsFlag.  If either ShowIterationsFlag or SaveIterationsFlag
 *  is TRUE, the update is selective according to direction and the trans-
 *  formation is calculated from pobj->lastOrigin and pobj->origin.
 *  If it is false then both transformation coordinates of the instances are
 *  updated using pobj->initOrigin and pobj->origin.
 *
 *  1/28/86
 *  last modified 1/21/88
 */

void
updateOct(direction)
    directionType direction;	/* HORIZONTAL or VERTICAL */
{
    struct object *pobj;
    struct wire *pw;
    /* struct gridLine *pgL; */
    octObject inst;
    octObject segment;
    int deltaX, deltaY;
    int moved;
    struct octPoint path[2];
    int debug = 0;

    inst.type = OCT_INSTANCE;
    inst.contents.instance.name = NIL(char);
    inst.contents.instance.master = NIL(char);
    inst.contents.instance.view = NIL(char);
    inst.contents.instance.version = OCT_CURRENT_VERSION;
    segment.type = OCT_PATH;

    if (TraceFlag)  {
        fprintf(pout, "\nUpdating instances...  ");
        fflush(pout);
    }

    if (direction!=HORIZONTAL && direction!=VERTICAL)  {
	fprintf(pout, "updateOct:  bad direction.\n");
	return;
    }

    /*
     *  Update the instances.
     */

    pobj = pObjectList;
    while (pobj!=NIL(struct object))  {

	moved = FALSE;

	if (debug)
	    fprintf(pout, "\nupdateOct:  updating object %d\n", pobj->key);

	/*
	 *  Need to update the transform if the object moved.  First
	 *  calculate the deltas from the old and new positions of the
	 *  bounding box.
	 */
	if (ShowIterationsFlag == FALSE  &&  SaveIterationsFlag == FALSE)  {
	    /*
	     *  No incremental updating of oct has been done, so calculate
	     *  change in transformation from pobj->initOrigin.
	     */
	    deltaX = pobj->initOrigin.x - pobj->origin.x;
	    deltaY = pobj->initOrigin.y - pobj->origin.y;
	    moved = TRUE;
	} else  {
	    /*
	     *  Since have been incrementally updating oct, the new transform-
	     *  ation is found from pobj->lastOrigin.  Also, only one coordi-
	     *  nate needs to be changed.
	     */
	    if (direction == HORIZONTAL)  {
		if (pobj->lastOrigin.x != pobj->origin.x)  {
		    deltaX = pobj->lastOrigin.x - pobj->origin.x;
		    deltaY = 0;
		    moved = TRUE;
		}
	    } else if (direction == VERTICAL)  {
		if (pobj->lastOrigin.y != pobj->origin.y)  {
		    deltaY = pobj->lastOrigin.y - pobj->origin.y;
		    deltaX = 0;
		    moved = TRUE;
		}
	    }
	}

	if (moved)  {
	    inst.objectId = pobj->instanceId;
#ifdef RPC
	    /*
	     *  In the remote version the transform is cached to avoid the
	     *  octGetById().  The transform has already been updated, in
	     *  updateX() and updateY().
	     */
	    inst.contents.instance.name = pobj->name;
	    inst.contents.instance.transform.translation.x =
		pobj->transform.translation.x;
	    inst.contents.instance.transform.translation.y =
		pobj->transform.translation.y;
	    inst.contents.instance.transform.transformType =
		pobj->transform.transformType;
	    inst.contents.instance.transform.generalTransform[0][0] =
		pobj->transform.generalTransform[0][0];
	    inst.contents.instance.transform.generalTransform[0][1] =
		pobj->transform.generalTransform[0][1];
	    inst.contents.instance.transform.generalTransform[1][0] =
		pobj->transform.generalTransform[1][0];
	    inst.contents.instance.transform.generalTransform[1][1] =
		pobj->transform.generalTransform[1][1];
#else
	    /*
	     *  Get the instance struct from oct...
	     */
	    if (octGetById(&inst) != OCT_OK)  {
		fprintf(pout, "updateOct:  can't get oct instance for object");
		fprintf(pout, " %d (octIc %d)\n", pobj->key, pobj->instanceId);
	    }
	    /*
	     *  update the transformation...
	     */
	    inst.contents.instance.transform.translation.x -= deltaX;
	    inst.contents.instance.transform.translation.y -= deltaY;
#endif

	    /*
	     *  and finally octModify.
	     */
	    OH_ASSERT(octModify(&inst));
	}
	pobj=pobj->pnext;
    }
    if (TraceFlag)  {
	fprintf(pout, "wires...");
	fflush(stdout);
    }

    /*  Now do the wires.  */
    pw = pWireList;
    while (pw!=NIL(struct wire))  {

	if (debug)
	    fprintf(pout, "\nupdateOct:  updating wire %d\n", pw->key);

	/*
	 *  Update if the segment moved or changed length.
	 */
	if (pw->changed)  {

	    /* printf("updating wire %d\n", pw->key;  fflush(stdout); */
	    segment.objectId = pw->pathId;

	    path[0].x = pw->p1.x;
	    path[0].y = pw->p1.y;
	    path[1].x = pw->p2.x;
	    path[1].y = pw->p2.y;

	    /*
	     *  Twiddle the endpoints if the path is zero-length.
	     */
	    if (pw->horizontal == TRUE)  {
		if (pw->p1.x == pw->p2.x)  {
		    path[0].x -= 1;
		    path[1].x += 1;
		}
	    } else  {
		if (pw->p1.y == pw->p2.y)  {
		    path[0].y -= 1;
		    path[1].y += 1;
		}
	    }

	    if (octPutPoints(&segment, 2, path) != OCT_OK)  {
		octError("updateOct:  can't update path's points");
		exit(100);
	    }

	    pw->changed = FALSE;
	}
	pw = pw->pnext;
    }
    
    
    /*  update the grid if there is one  */
    /********************************************
    if (VGridFlag)  {
	if ((pgL = pHorizontalGrid) == NIL(struct gridLine))  {
	    fprintf(pout, "updateOct:  NIL pointer for horiz. VGrid.\n");
	    exit(200);
	} else  {
	    while (pgL != NIL(struct gridLine))  {
		geo.geoID = pgL->squidID;
		if ((status=SQ(sqGet,sqGeo,&geo,path,pathLength)) != OCT_OK)  {
		    fprintf(pout, "updateOct:  can't sqGet geo.\n");
		    exit(100);
		}
		geo.def.line.path[0].x = pgL->loc;
		geo.def.line.path[0].y = MinBBY - GRIDOVERLAP;
		geo.def.line.path[1].x = pgL->loc;
		geo.def.line.path[1].y = MaxBBY + GRIDOVERLAP;

		if ((status = SQ(sqUpdate, sqGeo, geo)) != OCT_OK)  {
		    fprintf(pout, "updateOct:  can't sqUpdate geo.\n");
		    exit(100);
		}

		pgL = pgL->pnext;
	    }
	}
	if ((pgL = pVerticalGrid) == NIL(struct gridLine))  {
	    fprintf(pout, "updateOct:  NIL pointer for vert. VGrid.\n");
	    exit(200);
	} else  {
	    while (pgL != NIL(struct gridLine))  {
		geo.geoID = pgL->squidID;
		if ((status=SQ(sqGet,sqGeo,&geo,path,pathLength)) != OCT_OK)  {
		    fprintf(pout, "updateOct:  can't sqGet geo.\n");
		    exit(100);
		}
		geo.def.line.path[0].x = MinBBX - GRIDOVERLAP;
		geo.def.line.path[0].y = pgL->loc;
		geo.def.line.path[1].x = MaxBBX + GRIDOVERLAP;
		geo.def.line.path[1].y = pgL->loc;

		if ((status = SQ(sqUpdate, sqGeo, geo)) != OCT_OK)  {
		    fprintf(pout, "updateOct:  can't sqUpdate geo.\n");
		    exit(100);
		}

		pgL = pgL->pnext;
	    }
	}
    }
    ********************************************/
    
    /*
     *  Flush the changes just made, regardless of whether incremental or not.
     */
    /*
     *  9/3/87  Now done in saveOct().
     */
    /********
      OH_ASSERT( octFlushFacet(&WorkingFacet) );
    ********/
    if (TraceFlag)  {
        fprintf(pout, "\n");
        fflush(pout);
    }
}

