#include "copyright.h"
#include "sparcs.h"

/*
 *  Routine to add arcs to the graph to keep wires on their terminals.
 *
 *  The algo is:  (for HORIZONTAL)
 *		for (each wire segment pw)
 *		    if (pw->pt1 is non-NIL)
 *			add an arc from pw->po1 to pw;
 *		    if (pw->pt2 is non-NIL)
 *			add an arc from pw->po2 to pw;
 *
 *  Called By:	buildGraph
 *  Calls:
 *
 *  8/27/85
 *  last modified 3/23/88
 */

void
constraintsForTerminals(direction)
directionType direction;		/*  HORIZONTAL or VERTICAL  */
{
    struct wire     *pw;
    struct terminal *pt;
    int lowerBound, upperBound;
    int fatalErrorFlag = FALSE;

    switch (direction)  {

	case HORIZONTAL:
	    for (pw=pWireList; pw!=NIL(struct wire); pw=pw->pnext)  {

		/*
		 *  Consider only vertical wires.
		 */
		if (pw->horizontal == TRUE)  continue;

		/*
		 *  Process the first end of the wire.
		 */
		if ((pt = pw->pt1) == NIL(struct terminal))  {
		    printf("constraintsForTerminals:  NIL terminal for ");
		    printf("first end of wire %d\n", pw->key);
		    fatalErrorFlag = TRUE;
		    continue;
		}

		/*
		 *  Since the wire is vertical, the lower bound on the wire
		 *  with respect to the object that the terminal belongs to
		 *  is just the location of the lower-left corner of the 
		 *  terminal frame.
		 */
		lowerBound = pt->bb.lowerL.x;

		/*
		 *  The upper bound is the distance from the object's origin
		 *  to the far side (i.e., upper-right x coordinate) of the
		 *  terminal frame LESS the width of the wire.
		 */
		upperBound = pt->bb.upperR.x - pw->width;

		if (upperBound < lowerBound)  {
		    fprintf(pout, "%s%s\n",
			    "Wire is wider than terminal it connects",
			    " to; this is a fatal error.\n");
		    fprintf(pout, "\tterminal (%d, %d) (%d, %d) on %s\n",
				pw->po1->origin.x + pt->bb.lowerL.x,
				pw->po1->origin.y + pt->bb.lowerL.y,
				pw->po1->origin.x + pt->bb.upperR.x,
				pw->po1->origin.y + pt->bb.upperR.y,
				LayerMap[pt->maskNumber].name);
		    printf("\twire (%d, %d) (%d, %d), width %d, on %s\n",
			    pw->p1.x, pw->p1.y, pw->p2.x, pw->p2.y,
			    pw->width, LayerMap[pw->maskNumber].name);
		    fatalErrorFlag = TRUE;
		    continue;
		}

		/*
		 *  If nodes not activated then error.
		 */
		if (NodeArray[pw->po1->key]->key == UNDEF)  {
		    printf("constraintsForTerminals:  node for %d UNDEF\n",
			    pw->po1->key);
		    fatalErrorFlag = TRUE;
		    continue;
		}
		if (NodeArray[pw->key]->key == UNDEF)  {
		    printf("constraintsForTerminals:  node for %d UNDEF\n",
			    pw->key);
		    fatalErrorFlag = TRUE;
		    continue;
		}
		
		link(NodeArray[pw->po1->key], NodeArray[pw->key],
				    lowerBound, upperBound, 0);


		/*
		 *  Now repeat for the second end of the wire.
		 */
		if ((pt = pw->pt2) == NIL(struct terminal))  {
		    printf("constraintsForTerminals:  NIL terminal for ");
		    printf("second end of wire %d\n", pw->key);
		    fatalErrorFlag = TRUE;
		    continue;
		}

		lowerBound = pt->bb.lowerL.x;
		upperBound = pt->bb.upperR.x - pw->width;

		if (upperBound < lowerBound)  {
		    fprintf(pout, "%s%s\n",
			    "Wire is wider than terminal it connects",
			    " to; this is a fatal error.\n");
		    fprintf(pout, "\tterminal (%d, %d) (%d, %d) on %s\n",
				pw->po2->origin.x + pt->bb.lowerL.x,
				pw->po2->origin.y + pt->bb.lowerL.y,
				pw->po2->origin.x + pt->bb.upperR.x,
				pw->po2->origin.y + pt->bb.upperR.y,
				LayerMap[pt->maskNumber].name);
		    printf("\twire (%d, %d) (%d, %d), width %d, on %s\n",
			    pw->p1.x, pw->p1.y, pw->p2.x, pw->p2.y,
			    pw->width, LayerMap[pw->maskNumber].name);
		    fatalErrorFlag = TRUE;
		    continue;
		}

		if (NodeArray[pw->po2->key]->key == UNDEF)  {
		    printf("constraintsForTerminals:  node for %d UNDEF\n",
			    pw->po2->key);
		    fatalErrorFlag = TRUE;
		    continue;
		}
		
		link(NodeArray[pw->po2->key], NodeArray[pw->key],
				    lowerBound, upperBound, 0);
	    }

	    break;

	case VERTICAL:
	    for (pw=pWireList; pw!=NIL(struct wire); pw=pw->pnext)  {

		if (pw->horizontal == FALSE)  continue;

		if ((pt = pw->pt1) == NIL(struct terminal))  {
		    printf("constraintsForTerminals:  NIL terminal for ");
		    printf("first end of wire %d\n", pw->key);
		    fatalErrorFlag = TRUE;
		    continue;
		}

		/*
		 *  Wire is horizontal ==> lower bound on wire with respect
		 *  to the object that the terminal belongs to is just the
		 *  location of the lower-left corner of the terminal frame.
		 */
		lowerBound = pt->bb.lowerL.y;

		/*
		 *  Upper bound is the distance from the object's origin to
		 *  the far side (i.e., upper-right y coordinate) of the
		 *  terminal frame LESS the width of the wire.
		 */
		upperBound = pt->bb.upperR.y - pw->width;

		if (upperBound < lowerBound)  {
		    fprintf(pout, "%s%s\n",
			    "Wire is wider than terminal it connects",
			    " to; this is a fatal error.\n");
		    fprintf(pout, "\tterminal (%d, %d) (%d, %d) on %s\n",
				pw->po1->origin.x + pt->bb.lowerL.x,
				pw->po1->origin.y + pt->bb.lowerL.y,
				pw->po1->origin.x + pt->bb.upperR.x,
				pw->po1->origin.y + pt->bb.upperR.y,
				LayerMap[pt->maskNumber].name);
		    printf("\twire (%d, %d) (%d, %d), width %d, on %s\n",
			    pw->p1.x, pw->p1.y, pw->p2.x, pw->p2.y,
			    pw->width, LayerMap[pw->maskNumber].name);
		    fatalErrorFlag = TRUE;
		    continue;
		}

		/*
		 *  If nodes not activated then error.
		 */
		if (NodeArray[pw->po1->key]->key == UNDEF)  {
		    printf("constraintsForTerminals:  node for %d UNDEF\n",
			    pw->po1->key);
		    fatalErrorFlag = TRUE;
		    continue;
		}
		if (NodeArray[pw->key]->key == UNDEF)  {
		    printf("constraintsForTerminals:  node for %d UNDEF\n",
			    pw->key);
		    fatalErrorFlag = TRUE;
		    continue;
		}
		
		link(NodeArray[pw->po1->key], NodeArray[pw->key],
				    lowerBound, upperBound, 0);


		if ((pt = pw->pt2) == NIL(struct terminal))  {
		    printf("constraintsForTerminals:  NIL terminal for ");
		    printf("second end of wire %d\n", pw->key);
		    fatalErrorFlag = TRUE;
		    continue;
		}

		lowerBound = pt->bb.lowerL.y;
		upperBound = pt->bb.upperR.y - pw->width;

		if (upperBound < lowerBound)  {
		    fprintf(pout, "%s%s\n",
			    "Wire is wider than terminal it connects",
			    " to; this is a fatal error.\n");
		    fprintf(pout, "\tterminal (%d, %d) (%d, %d) on %s\n",
				pw->po2->origin.x + pt->bb.lowerL.x,
				pw->po2->origin.y + pt->bb.lowerL.y,
				pw->po2->origin.x + pt->bb.upperR.x,
				pw->po2->origin.y + pt->bb.upperR.y,
				LayerMap[pt->maskNumber].name);
		    printf("\twire (%d, %d) (%d, %d), width %d, on %s\n",
			    pw->p1.x, pw->p1.y, pw->p2.x, pw->p2.y,
			    pw->width, LayerMap[pw->maskNumber].name);
		    fatalErrorFlag = TRUE;
		    continue;
		}

		if (NodeArray[pw->po2->key]->key == UNDEF)  {
		    printf("constraintsForTerminals:  node for %d UNDEF\n",
			    pw->po2->key);
		    fatalErrorFlag = TRUE;
		    continue;
		}
		
		link(NodeArray[pw->po2->key], NodeArray[pw->key],
				    lowerBound, upperBound, 0);
	    }
	    break;

	default:
	    fprintf(pout, "constraintsForTerminals:  illegal direction.\n");
	    exit(-1);
    }
    if (fatalErrorFlag)  {
	printf("\nfatal error(s) in constraintsForTerminals.\n");
	exit(-1);
    }
}
