#include "copyright.h"
#include "sparcs.h"
#include "rpc.h"
#include "facet.h"

/*
 *  Similar in purpose to initOct(), this routine performs the oct calls
 *  necessary to prepare for running the oct generators that extract the
 *  instances, frames, terminals, and wires for the layout to be spaced.
 *
 *  The WorkingFacet, which is also the result facet, is initialized here.
 *  That is, the input facet is copied into the WorkingFacet and a window
 *  is opened onto the WorkingFacet.  An octOpenFacet() is executed initially
 *  on WorkingFacet to determine the state, and the actions are:
 *
 *	OCT_ALREADY_OPEN:  WorkingFacet and window(s) left from prior session.
 *		Hard close all windows on WorkingFacet.
 *		Copy input facet to WorkingFacet and open window on it.
 *
 *	OCT_OLD_FACET:  Facet exists but is not open, no windows open.
 *		Hard close WorkingFacet.
 *		Copy input facet to WorkingFacet and open window on it.
 *			WASTES A FACET READ!
 *
 *	OCT_NEW_FACET:  Facet does not exist
 *		Hard close WorkingFacet.
 *		Copy and open window.
 *
 *  NOTE:  Could do an immediate copy of the input facet to WorkingFacet
 *	   followed by an open of WorkingFacet for append, but must copy
 *	   twice if the open returns OCT_ALREADY_OPEN.
 *
 *  There is one special case - if the input facet and the output facet are
 *  the same, the copy is not done and no new windows are created.  The
 *  window where sparcs was invoked is later redrawn to show the results.
 *
 *  Modified 10/8/87 for vem_5/rpc_5
 *
 *  Calls:  readOct()
 *
 *  Renamed 2/24/87 from SPinitRpcOct().
 *
 *  7/22/86
 *  last modified 9/22/88
 */

void
initRpcOct(pinSpot, poutSpot)
struct RPCSpot *pinSpot;	/*  Where rpc_sparcs was invoked  */
struct RPCSpot *poutSpot;	/*  New window created here; filled in  */
{
    octObject inFacet;
    octObject techProp;
    octStatus status;
    struct RPCSpot tmpSpot;
    struct object *po;
    struct wire *pw;
    char *techName = NIL(char);
    void readOct();
    int debug = 0;

    inFacet.type = OCT_FACET;
    techProp.type = OCT_PROP;

    if (debug)  {
	fprintf(pout, "\nReading \"%s\" facet of cell \"%s\", with\n",
								Facet, InCell);
	fprintf(pout, "InView \"%s\" and OutView \"%s\"\n\n", InView, OutView);
    }

    if (TraceFlag)  {
	fprintf(pout, "\nOpening working facet...  ");
	fflush(pout);
    }

    /*
     *  Read the TECHNOLOGY property so that the rules file can be found.
     */
    inFacet.objectId = pinSpot->facet;
    techProp.contents.prop.name = "TECHNOLOGY";
    techProp.contents.prop.type = OCT_STRING;
    if ((status=octGetByName(&inFacet, &techProp)) != OCT_OK)  {
	if (status == OCT_NOT_FOUND)  {
	    printf("initRpcOct:  WARNING - no TECHNOLOGY property present.\n");
	} else  {
	    octError("initRpcOct:  trying to get TECHNOLOGY property");
	    exit(-1);
	}
    } else  {
	techName = STRSAVE(techProp.contents.prop.value.string);
	/* printf("initRpcOct:  TECHNOLOGY is \"%s\"\n", techName); */
    }

    /*
     *  And now the technology data (spacing rules and resistance info).
     */
    openTech(&inFacet, techName);

    WorkingFacet.type = OCT_FACET;
    WorkingFacet.contents.facet.cell = OutCell;
    WorkingFacet.contents.facet.view = OutView;
    WorkingFacet.contents.facet.facet = Facet;
    WorkingFacet.contents.facet.version = OCT_CURRENT_VERSION;
    WorkingFacet.contents.facet.mode = "a";

    /*
     *  First deal with the special case noted above.
     */
    if (strcmp(InCell, OutCell)==0 && strcmp(InView, OutView)==0)  {
	WorkingFacet.objectId = pinSpot->facet;
	poutSpot->theWin = pinSpot->theWin;
	poutSpot->facet = pinSpot->facet;
	poutSpot->thePoint.x = pinSpot->thePoint.x;
	poutSpot->thePoint.y = pinSpot->thePoint.y;
    } else  {
	/*
	 *  Open the WorkingFacet to determine its status.
	 */
	if ((status=octOpenFacet(&WorkingFacet)) == OCT_ALREADY_OPEN)  {
	    /*
	     *  Facet exists and there are windows open onto it.  Hard-close
	     *  the windows and thereby WorkingFacet.
	     */
	    tmpSpot.facet = WorkingFacet.objectId;
	    vemCommand("kill-buffer", &tmpSpot, lsCreate(), 0);
	} else if (status==OCT_OLD_FACET || status==OCT_NEW_FACET)  {
	    /*
	     *  Close it.
	     */
	    if (octCloseFacet(&WorkingFacet) != OCT_OK)  {
		octError("initRpcOct:  can't close the WorkingFacet.\n");
		exit(-1);
	    }
	} else  {
	    octError("initRpcOct:  can't open WorkingFacet to check status.");
	    exit(-1);
	}
	
	/*
	 *  Copy into the WorkingFacet.
	 */
	inFacet.contents.facet.cell = InCell;
	inFacet.contents.facet.view = InView;
	inFacet.contents.facet.facet = Facet;
	inFacet.contents.facet.version = OCT_CURRENT_VERSION;
	inFacet.contents.facet.mode = "r";
	if ((status=octCopyFacet(&WorkingFacet, &inFacet)) != OCT_OK)  {
	    octError("initRpcOct:  can't copy into the WorkingFacet.");
	    exit(-1);
	}
	
	/*
	 *  Now re-open it and create a window on it.
	 */
	if (octOpenFacet(&WorkingFacet) != OCT_OLD_FACET)  {
	    octError("initRpcOct:  can't re-open facet.");
	    exit(-1);
	}
	poutSpot->theWin = vemOpenWindow(&WorkingFacet);
	poutSpot->facet = WorkingFacet.objectId;
	poutSpot->thePoint.x = 0;
	poutSpot->thePoint.y = 0;
    }
    
    if (TraceFlag)  {
	fprintf(pout, "\n");
	fflush(pout);
    }
    
    /*
     *  Finally, invoke readOct() to run the generators.
     */
    checkPolicy(&WorkingFacet);
    ricksReadOct();


    ElementByKey = ALLOC(struct element, NumElements+1);
    for (po=pObjectList; po!=NIL(struct object); po=po->pnext)  {
	ElementByKey[po->key].ptr.po = po;
    }
    for (pw=pWireList; pw!=NIL(struct wire); pw=pw->pnext)  {
	ElementByKey[pw->key].ptr.pw = pw;
    }
}
