#include "copyright.h"
#include "sparcs.h"

/*
 *  Scanline routine.  The state is determined from two lists, one of the MIN
 *  corner of each edge and the other of the MAX corner from each edge.  The
 *  next state is the smallest coord value from the lists.
 *
 *  Algorithm:
 *	while (have states)  {
 *	    get new state;
 *	    if (state is from MIN list OR from MID list)  {
 *		if (state is from MIN list)  {
 *		    add edge(s) to edge structure;
 *		    process edges just added to edge struct;  (find const's)
 *		} else  {
 *		    un-mask edges in edge struct;
 *		}
 *	    } else  {     ...state is from MAX list
 *		delete edge(s) from edge structure;
 *	    }
 *	}
 *
 *  NOTE that the first state always comes from the MIN list, and the last
 *  state always comes from the MAX list.  This is reflected in the state
 *  calculation below.
 *
 *  2/16/87
 *  last modified 9/22/88
 */

/* XXX */
struct edge *pEdgeHead;
struct edge *pEdgeTail;
struct elist *pTFStack = NIL(struct elist);    /* tos ptr, newly-visible tfs */
struct elist *pPFStack = NIL(struct elist);    /* tos ptr, newly visible pfs */
struct hlist *phardTFStack = NIL(struct hlist);    /* tos ptr, newly-visible,
						      hard-masked tfs */
int numHardMasked;
int numSoftMasked;
int numPFMasked;

void
doScan(direction)
directionType direction;
{
    struct edgePtr *pmin;	/* scanner for MIN list (pMinEdgeListHead) */
    struct edgePtr *pmid;	/* scanner for MID list (pMidEdgeListHead) */
    struct edgePtr *pmax;	/* scanner for MAX list (pMaxEdgeListHead) */
    struct edgePtr *pfirstCur;	/* 1st eP in list with coord == curCoord */
    struct edge *pedge;
    struct elist *pedgeList;
    struct hlist *phardEdgeList;
    int curCoord;
    void addToEdgeStruct();
    void deleteFromEdgeStruct();
    void newDeleteFromEdgeStruct();
    void unmaskEdges();
    void printEdgeStruct();
    void procEdge();
    int debug = 0;

    pEdgeHead = NIL(struct edge);
    pEdgeTail = NIL(struct edge);

    pmin = pMinEdgeListHead;
    pmid = pMidEdgeListHead;
    pmax = pMaxEdgeListHead;

    numHardMasked = 0;
    numSoftMasked = 0;
    numPFMasked = 0;

    /*****
    printf("\nBeginning of doScan:  numSoftMasked = %d, numPFMasked = %d\n\n",
	    numSoftMasked, numPFMasked);
    *****/
    /*
     *  While there are states...
     */
    while (pmin != NIL(struct edgePtr)  ||  pmax != NIL(struct edgePtr))  {
	/*
	 *  Get the new state.  Note that if pmin->coord == pmax->coord or if
	 *  pmid->coord == pmax->coord, then the corresponding operations are
	 *  performed before the deletions.  pmax is not checked for NIL here,
	 *  since pMaxEdgeList must be finished last.
	 */
	if ((pmin != NIL(struct edgePtr)  &&  pmin->coord <= pmax->coord)  ||
	    (pmid != NIL(struct edgePtr)  &&  pmid->coord <= pmax->coord))  {
	    /*
	     *  If neither list is empty, then the state is from the list
	     *  containing the lesser coord.  If the coords are equal, then
	     *  the state is from the MIN list.  If either list is empty, the
	     *  state is from the other list.
	     */
	    if ((pmin!=NIL(struct edgePtr) && pmid!=NIL(struct edgePtr) &&
		pmin->coord<=pmid->coord)  ||  pmid == NIL(struct edgePtr))  {
		/*
		 *  MIN list has the lesser (or equal) coord, or the MID list
		 *  is empty.  So process from the MIN list.
		 */
		if (debug)  {
		    printf("\nMIN list sets state;  coord:  %d\n",pmin->coord);
		}

		pfirstCur = pmin;
		curCoord = pmin->coord;
		
		/*
		 *  Add the new edges...
		 */
		while (pmin!=NIL(struct edgePtr) && pmin->coord==curCoord)  {
		    if (debug)  {
			printf("    adding %s edge from inst %d, ",
				(pmin->pedge->type == MIN) ? "MIN" : "MAX",
				pmin->pedge->key);
			printf("on layer %s, ",
				LayerMap[pmin->pedge->maskNumber].name);
			if (pmin->pedge->netId == UNDEF)  {
			    printf("netId UNDEF");
			} else  {
			    printf("netId %d", pmin->pedge->netId);
			}
			printf(" masked %c\n", pmin->pedge->masked);
		    }
		    addToEdgeStruct(curCoord, pmin->pedge);
		    pmin = pmin->pnext;
		}
		if (debug)  {
		    printf("\nAfter adding:\n");
		    printEdgeStruct(pEdgeHead, TRUE);
		}

		/*
		 *  Process the edges just added.  Masked pfs aren't processed.
		 *  Hard (tf-tf) masked tfs are not processed.  Non-masked pfs
		 *  are processed by walkPF().  Connected, non-masked tfs are
		 *  processed by walkTF().  Connected, soft-masked tfs are
		 *  processed by walkSoftTF().
		 *
		 *  Since hard-masked edges should not be seen by pfs, all con-
		 *  nected tfs are processed before any calls to walkPF().
		 */
		pmin = pfirstCur;
		while (pmin!=NIL(struct edgePtr) && pmin->coord==curCoord)  {
		    /*
		     *  Connected tfs; process if not masked (neither H nor S).
		     */
		    if (pmin->pedge->netId!=UNDEF &&
			pmin->pedge->masked=='N') {
			if (debug)  {
			    printf("    %s edge at %d from %d, ",
				    (pmin->pedge->type == MIN) ? "MIN" : "MAX",
				    pmin->pedge->loc, pmin->pedge->key);
			    printf("on layer %s, ",
				    LayerMap[pmin->pedge->maskNumber].name);
			    if (pmin->pedge->netId == UNDEF)  {
				printf("netId UNDEF, ");
			    } else  {
				printf("netId %d, ", pmin->pedge->netId);
			    }
			    printf("masked = %c: ", pmin->pedge->masked);
			    printf("walkTF\n");
			}
			walkTF(curCoord, pmin->pedge, direction);
		    }
		    pmin = pmin->pnext;
		}

		/******
		if (debug)  {
		    printf("\nAfter processing tfs:\n");
		    printEdgeStruct(pEdgeHead, TRUE);
		    printf("\n");
		}
		******/

		pmin = pfirstCur;
		while ( pmin!=NIL(struct edgePtr) && pmin->coord==curCoord )  {
		    /*
		     *  Process connected, soft-masked tfs.
		     */
		    if (pmin->pedge->netId!=UNDEF &&
			pmin->pedge->masked=='S')  {
			if (debug)  {
			    printf("    %s edge at %d from %d, ",
				    (pmin->pedge->type == MIN) ? "MIN" : "MAX",
				    pmin->pedge->loc, pmin->pedge->key);
			    printf("on layer %s, ",
				    LayerMap[pmin->pedge->maskNumber].name);
			    if (pmin->pedge->netId == UNDEF)  {
				printf("netId UNDEF, ");
			    } else  {
				printf("netId %d, ", pmin->pedge->netId);
			    }
			    printf("masked = %c: ", pmin->pedge->masked);
			    printf("walkSoftTF\n");
			}
			walkSoftTF(curCoord, pmin->pedge, direction);
		    }
		    pmin = pmin->pnext;
		}
		/***
		if (debug)  {
		    printf("\nAfter processing stfs:\n");
		    printEdgeStruct(pEdgeHead, TRUE);
		    printf("\n");
		}
		***/


		/*
		 *  Now the pf edges just added.
		 */
		pmin = pfirstCur;
		while ( pmin!=NIL(struct edgePtr) && pmin->coord==curCoord )  {
		    if (pmin->pedge->netId == UNDEF)  {
			if (debug)  {
			    printf("    %s edge at %d from %d, ",
				    (pmin->pedge->type == MIN) ? "MIN" : "MAX",
				    pmin->pedge->loc, pmin->pedge->key);
			    printf("on layer %s, ",
				    LayerMap[pmin->pedge->maskNumber].name);
			    if (pmin->pedge->netId == UNDEF)  {
				printf("netId UNDEF, ");
			    } else  {
				printf("netId %d, ", pmin->pedge->netId);
			    }
			    printf("masked = %c: ", pmin->pedge->masked);
			}

			if (pmin->pedge->masked == 'N')  {
			    if (debug)  {
				printf("walkPF\n");
			    }
			    walkPF(curCoord, pmin->pedge, direction);
			} else  {
			    if (debug)  {
				printf("skipped\n");
			    }
			}
		    }
		    pmin = pmin->pnext;
		}
	    } else  {
		/*
		 *  MID list has the lesser coord, or the MIN list is empty.
		 *  Process from the MID list.
		 *
		 *  unmaskEdges(pmid->pedge) processes all edges that are
		 *  masked by pmin->pedges, regardless of whether pmin->pedge
		 *  itself is masked or not.
		 */
		if (debug)  {
		    printf("\nMID list sets state;  coord:  %d\n",pmid->coord);
		}

		curCoord = pmid->coord;
		
		while ( pmid!=NIL(struct edgePtr) && pmid->coord==curCoord)  {
		    if (debug)  {
			printf("    possible unmasking wrt %s edge, at loc %d",
				(pmid->pedge->type == MIN) ? "MIN" : "MAX",
				pmid->pedge->loc);
			printf(", inst %d, ", pmid->pedge->key);
			printf("layer %s, ",
				LayerMap[pmid->pedge->maskNumber].name);
			if (pmid->pedge->netId == UNDEF)  {
			    printf("netId UNDEF\n");
			} else  {
			    printf("netId %d\n", pmid->pedge->netId);
			}

			if (pmid->pedge->pmasks != NIL(struct edge))  {
			    printf("\tpf-pf or soft-masks another edge\n");
			}
			if (pmid->pedge->phardMasks != NIL(struct elist))  {
			    printf("\thard-masks other edges\n");
			}
		    }
		    if (pmid->pedge->pmasks != NIL(struct edge)  ||
		        pmid->pedge->phardMasks != NIL(struct elist))  {
			if (debug)  {
		    	    if (pmid->pedge->pmasks != NIL(struct edge))  {
				printf("\tpf-pf or soft-masks another edge\n");
			    }
			    if (pmid->pedge->phardMasks != NIL(struct elist))  {
				printf("\thard-masks other edges\n");
			    }
			}
		        unmaskEdges(pmid->pedge);
		    } else  {
			if (debug)  {
			    printf("\tskipping; doesn't mask any edges.\n");
			}
		    }
		    pmid = pmid->pnext;
		}

		/*
		 *  Some masked edges may have become visible due to the
		 *  unmasking just performed.  If there are any of these
		 *  newly-visible edges, process them according to whether
		 *  they are tfs or pfs.
		 *
		 *  See notes of 6/8/87, 6/21/87, 6/26/87, 6/30/87.
		 */
		if (phardTFStack != NIL(struct hlist))  {
		    phardEdgeList = phardTFStack;
		    phardTFStack = phardTFStack->pnext;
		    while (phardEdgeList != NIL(struct hlist))  {
			pedge = phardEdgeList->pedge;
			if (pedge->active == 'Y')  {
			    /*
			     *  pedge is still in the edge struct.
			     */
			    if (pedge->netId == UNDEF)  {
				printf("doScan:  newly-visible edge must ");
				printf("be tf\n");
				exit(-1);
			    }

			    if (pedge->masked == 'S' || pedge->masked == 'P') {
				printf("doScan:  newly-visible hard tf edge ");
				printf("is illegally masked (%c)",
					pedge->masked);
				printf("; must be H or N.\n");
				exit(-1);
			    }
			    if (debug)  {
				printf("\nNewly-visible hard tf edge:\n");
				printf("    %s edge at %d from %d, ",
					(pedge->type == MIN) ? "MIN" : "MAX",
					pedge->loc, pedge->key);
				printf("on layer %s, ",
					LayerMap[pedge->maskNumber].name);
				if (pedge->netId == UNDEF)  {
				    printf("netId UNDEF, ");
				} else  {
				    printf("netId %d, ", pedge->netId);
				}
				printf("masked = %c: ", pedge->masked);
				printf("rewalkHardTF\n");
			    }
			    rewalkHardTF(curCoord, phardEdgeList, direction);
			} else  {
			    printf("doScan:  phardTFStack edges must ");
			    printf("be active.\n");
			    exit(-1);
			}

			/*
			 *  Pop the stack.
			 */
			phardEdgeList = phardTFStack;
			/* XXX */
			if (phardTFStack) phardTFStack  = phardTFStack->pnext;
		    }
		    /*
		     *  Reset phardTFStack.
		     */
		    phardTFStack = NIL(struct hlist);
		}
		if (pTFStack != NIL(struct elist))  {
		    pedgeList = pTFStack;
		    pTFStack = pTFStack->pnext;
		    while (pedgeList != NIL(struct elist))  {
			pedge = pedgeList->pedge;
			if (pedge->active == 'Y')  {
			    /*
			     *  pedge is still in the edge struct.
			     */
			    if (pedge->netId == UNDEF)  {
				printf("doScan:  newly-visible edge must ");
				printf("be tf\n");
				exit(-1);
			    }

			    if (pedge->masked == 'S'  ||
				pedge->masked == 'P'  ||
				pedge->masked == 'H')  {
				printf("doScan:  newly-visible tf edge is ");
				printf("illegally masked (%c)", pedge->masked);
				printf("; must be N.\n");
				exit(-1);
			    }
			    if (debug)  {
				printf("\nNewly-visible tf edge:\n");
				printf("    %s edge at %d from %d, ",
					(pedge->type == MIN) ? "MIN" : "MAX",
					pedge->loc, pedge->key);
				printf("on layer %s, ",
					LayerMap[pedge->maskNumber].name);
				if (pedge->netId == UNDEF)  {
				    printf("netId UNDEF, ");
				} else  {
				    printf("netId %d, ", pedge->netId);
				}
				printf("masked = %c: ", pedge->masked);
				printf("walkTF\n");
			    }
			    rewalkTF(curCoord, pedge, direction);
			} else  {
			    printf("doScan:  pTFStack edges must be active.\n");
			    exit(-1);
			}

			/*
			 *  Pop the stack.
			 */
			pedgeList = pTFStack;
			/* XXX */
			if (pTFStack) pTFStack  = pTFStack->pnext;
		    }
		    /*
		     *  Reset pTFStack.
		     */
		    pTFStack = NIL(struct elist);
		}
		while (pPFStack != 0)  {
		    pedgeList = pPFStack;
		    pPFStack  = pPFStack->pnext;
		    pedge = pedgeList->pedge;
		    if (pedge->active == 'Y')  {
			/*
			 *  pedge is still in the edge struct.
			 */
			if (pedge->netId != UNDEF)  {
			    printf("doScan: edge in pPFStack must be pf\n");
			    exit(-1);
			}

			if (pedge->masked != 'N')  {
			    printf("doScan:  newly-visible pf edge is ");
			    printf("illegally  masked (%c); must be N.\n",
				    pedge->masked);
			    exit(-1);
			}
			if (debug)  {
			    printf("\nNewly-visible pf edge:\n");
			    printf("    %s edge at %d from %d, ",
				    (pedge->type == MIN) ? "MIN" : "MAX",
				    pedge->loc, pedge->key);
			    printf("on layer %s, ",
				    LayerMap[pedge->maskNumber].name);
			    if (pedge->netId == UNDEF)  {
				printf("netId UNDEF, ");
			    } else  {
				printf("netId %d, ", pedge->netId);
			    }
			    printf("masked = %c: ", pedge->masked);
			    printf("walkPF\n");
			}
			rewalkPF(curCoord, pedge, direction);
		    } else  {
			printf("doScan:  pPFStack edges must be active.\n");
			exit(-1);
		    }
		}
		/*
		 *  Reset pPFStack.
		 */
		pPFStack = NIL(struct elist);
	    }

	} else  {
	    /*
	     *  State is from MAX list.  Delete edges from the edge struct.
	     */
	    if (debug)  {
	        printf("\nMAX list sets state;  coord:  %d\n", pmax->coord);
	    }

	    curCoord = pmax->coord;


	    /***
	    printf("\nBefore deleting:\n");
	    printEdgeStruct(pEdgeHead, TRUE);
	    printf("\n");
	    ***/

	    /*
	     *  Remove any edges from the edge structure that end at the
	     *  current coordinate.
	     */
	    while ( pmax != NIL(struct edgePtr) && pmax->coord == curCoord )  {
		if (debug)  {
		    printf("    deleting %s edge from inst %d, ",
			    (pmax->pedge->type == MIN) ? "MIN" : "MAX",
			    pmax->pedge->key);
		    printf("on layer %s, ",
			    LayerMap[pmax->pedge->maskNumber].name);
		    if (pmax->pedge->netId == UNDEF)  {
		        printf("netId UNDEF\n");
		    } else  {
		        printf("netId %d\n", pmax->pedge->netId);
		    }
		}
		newDeleteFromEdgeStruct(pmax->pedge);
		pmax = pmax->pnext;
	    }
	    if (debug)  {
		printf("\nAfter deleting:\n");
		printEdgeStruct(pEdgeHead, TRUE);
	    }
	}
    }
    /***
    printf("\nEnd of doScan():  numSoftMasked = %d, numPFMasked = %d\n\n",
	    numSoftMasked, numPFMasked);
    ***/
}


/*
 *  Add pedge to the edge structure.  There may be than one edge with the same
 *  loc.  MAX edges are added to the structure to the left of all other edges
 *  with the same loc; MIN edges are added to the right of all other edges with
 *  the same loc.  This results in all MAX edges being to the left of all MIN
 *  edges, when multiple edges have the same loc value.
 *
 *  Also, do any pf-pf and pf-tf masking that is appropriate.
 *
 *  Modified 6/30/87 to mask only if the edges actually overlap (as opposed to
 *  pedge overlapping the extension of pe)
 *
 *  2/16/87
 *  last modified 6/30/87
 */

void
addToEdgeStruct(stateCoord, pedge)
int stateCoord;
struct edge *pedge;
{
    struct edge *pe;

    /*
     *  Sanity check...
     */
    if (pedge->masked != 'N')  {
	printf("addToEdgeStruct:  input edge is masked!\n");
	exit(-1);
    }

    if (pedge->type == MAX)  {
	/*
	 *  pedge is MAX; insert to the LEFT of any other edges at pedge->loc.
	 */
	for (pe=pEdgeHead; pe!=NIL(struct edge); pe=pe->pnext)  {
	    if (pe->loc < pedge->loc)  {
		continue;
	    } else  {
		break;
	    }
	}
	pedge->active = 'Y';
	dllPreAdd(&pEdgeHead, &pEdgeTail, pe, pedge);

	/*
	 *  See if this edge is masked by, or masks, an opposite-polarity
	 *  edge at this loc, from the same element, and on the same layer.
	 */
	pe = pedge->pnext;
	for (; pe!=NIL(struct edge) && pe->loc==pedge->loc; pe=pe->pnext)  {
	    /*
	     *  Skip masked edges, edges of the same polarity, edges from other
	     *  elements, and edges on other layers.
	     *
	     *  Also skip edges whose extensions only overlap pedge. 
	     */
	    if (pe->masked == 'P' || pe->masked == 'S' ||
	    	pe->type == MAX  ||  pe->key != pedge->key  ||
		pe->maskNumber != pedge->maskNumber || pe->max <= stateCoord) {
		continue;
	    } else  {
		/*
		 *  Found a match.  Make the appropriate edges invisible.
		 *  i.e.,
		 *	both pfs, (or unconn tfs):  mask both 'P'
		 *	one pf, one conn. tf:       mask both 'S'
		 *	both conn. tfs:	            ERROR ???????
		 */
		if (pe->netId == UNDEF)  {
		    if (pedge->netId == UNDEF)  {
			/*
			 *  Both pfs ==> both pf-pf masked.  Note no multiple
			 *  masking is allowed, unlike tf-tf masking.
			 */
			pedge->masked = 'P';
			pe->masked = 'P';
			pedge->pmasks = pe;
			pe->pmasks = pedge;
			numPFMasked += 2;
		    } else  {
			/*
			 *  pe is pf, pedge is connected tf ==> both soft-mask.
			 *  Again note no multiple masking.
			 */
			pedge->masked = 'S';
			pe->masked = 'S';
			pedge->pmasks = pe;
			pe->pmasks = pedge;
			numSoftMasked += 2;
		    }
		} else  {
		    if (pedge->netId == UNDEF)  {
			/*
			 *  pe is conected tf, pedge is pf ==> both soft-mask.
			 */
			pedge->masked = 'S';
			pe->masked = 'S';
			pedge->pmasks = pe;
			pe->pmasks = pedge;
			numSoftMasked += 2;
		    }
		}
	    }
	}
    } else  {
	/*
	 *  pedge is MIN; insert to the RIGHT of any other edges at pedge->loc.
	 */
	for (pe=pEdgeHead; pe!=NIL(struct edge); pe=pe->pnext)  {
	    if (pe->loc <= pedge->loc)  {
		continue;
	    } else  {
		break;
	    }
	}
	pedge->active = 'Y';
	dllPreAdd(&pEdgeHead, &pEdgeTail, pe, pedge);

	pe = pedge->pprev;
	for (; pe!=NIL(struct edge) && pe->loc==pedge->loc; pe=pe->pprev)  {
	    /*
	     *  Skip masked edges, edges of the same polarity, edges from other
	     *  elements, and edges on other layers.
	     */
	    if (pe->masked == 'P' || pe->masked == 'S' ||
	    	pe->type == MIN  ||  pe->key != pedge->key  ||
		pe->maskNumber != pedge->maskNumber || pe->max <= stateCoord) {
		continue;
	    } else  {
		/*
		 *  Found a match, so make invisible.
		 */
		if (pe->netId == UNDEF)  {
		    if (pedge->netId == UNDEF)  {
			/*
			 *  Both pfs.
			 */
			pedge->masked = 'P';
			pe->masked = 'P';
			pedge->pmasks = pe;
			pe->pmasks = pedge;
			numPFMasked += 2;
		    } else  {
			/*
			 *  pe is pf, pedge is connected tf.
			 */
			pedge->masked = 'S';
			pe->masked = 'S';
			pedge->pmasks = pe;
			pe->pmasks = pedge;
			numSoftMasked += 2;
		    }
		} else  {
		    if (pedge->netId == UNDEF)  {
			/*
			 *  pe is connected tf, pedge is pf.
			 */
			pedge->masked = 'S';
			pe->masked = 'S';
			pedge->pmasks = pe;
			pe->pmasks = pedge;
			numSoftMasked += 2;
		    }
		}
	    }
	}
    }
}


/*
 *  Unmask the edge(s) that are masked by pedge.  If pedge performs 'S' or 'P'
 *  masking, then it must be masked the same way as well.  If it performs 'H'
 *  masking, it may or may not be 'S'-masked, depending on the order of the
 *  unmaskings.
 *
 *  If edges that are unmasked are still in the edge structure following the
 *  current state, then they are saved on the appropriate stack for re-walking.
 *  This can be determined from the max fields of the two edges.
 *
 *  6/26/87
 *  last modified 2/16/88
 */

void
unmaskEdges(pedge)
struct edge *pedge;
{
    struct edge *pmaskedEdge;		/* edge masked by pedge */
    struct elist *pHMList, *ptmpList;
    struct klist *pHMB, *plastHMB;
    struct hlist *phardStack;
    int found;
    int debug = 0;

    if (pedge->active == 'N')  {
	printf("unmaskEdges:  pedge isn't in edge struct\n");
	printf("\t %s edge from inst %d, ",
		(pedge->type == MIN) ? "MIN" : "MAX", pedge->key);
	printf("on layer %s, ", LayerMap[pedge->maskNumber].name);
	if (pedge->netId == UNDEF)  {
	    printf("netId UNDEF");
	} else  {
	    printf("netId %d", pedge->netId);
	}
	printf(" masked %c\n", pedge->masked);
	exit(-1);
    }

    if (pedge->pmasks != NIL(struct edge))  {
	pmaskedEdge = pedge->pmasks;
	if (pmaskedEdge->masked == 'P')  {
	    /*
	     *  pedge does pf-pf masking; it masks only one edge.
	     */
	    
	    /*
	     *  XXX 2/16/88
	     *  Is this the only place where this situation applies???
	     */
	    if (pmaskedEdge->max == pedge->max)  {
		/*
		 *  Don't unmask!  The edges leave the edge struct at the
		 *  same time.  So, since the edges masked, their extensions
		 *  should mask as well!
		 */
		if (debug)  {
		    printf("%s%s\n", "unmaskEdges:  pedge NOT resetting 'P' ",
			   "edge; BOTH edges leave edge struct at same time.");
		}
		return;
	    }

	    if (debug)  {
	        printf("unmaskEdges:  pedge resetting 'P' edge..\n");
	    }
	    /*
	     *  pedge->masked must be 'P', and pmaskedEdge->pmasks must be
	     *  pedge.
	     */
	    if (pedge->masked == 'P'  &&  pmaskedEdge->pmasks == pedge)  {
		pmaskedEdge->masked = 'N';
		pmaskedEdge->pmasks = NIL(struct edge);
		pedge->masked = 'N';
		pedge->pmasks = NIL(struct edge);
		numPFMasked -= 2;
		/*
		 *  If pmaskedEdge extends beyond pedge, then a portion
		 *  becomes newly visible due to this un-masking and it must
		 *  be re-processed.
		 */
		if (pmaskedEdge->max > pedge->max)  {
		    ptmpList = ALLOC(struct elist, 1);
		    ptmpList->pedge = pmaskedEdge;
		    ptmpList->pnext = pPFStack;
		    pPFStack = ptmpList;
		}
	    } else  {
		printf("unmaskEdges:  pedge should be 'P', and ");
		printf("pedge->pmasks->pmasks should be pedge.\n");
		exit(-1);
	    }
	} else if (pmaskedEdge->masked == 'S')  {
	    /*
	     *  pedge does pf-tf masking; it masks only one edge.
	     */
	    if (debug)  {
	        printf("unmaskEdges:  pedge resetting 'S' edge..\n");
	    }
	    /*
	     *  pedge->masked must be 'S', and pmaskedEdge->pmasks must be
	     *  pedge.
	     */
	    if (pedge->masked == 'S'  &&  pmaskedEdge->pmasks == pedge)  {
		pmaskedEdge->masked = 'N';
		pmaskedEdge->pmasks = NIL(struct edge);
		pedge->masked = 'N';
		pedge->pmasks = NIL(struct edge);
		numSoftMasked -= 2;
		/*
		 *  As above, if pmaskedEdge extends beyond pedge then it must
		 *  be re-processed.  The type of re-processing (walkTF() or
		 *  walkPF()) varies here, though.
		 */
		if (pmaskedEdge->max > pedge->max)  {
		    ptmpList = ALLOC(struct elist, 1);
		    ptmpList->pedge = pmaskedEdge;
		    if (pmaskedEdge->netId == UNDEF)  {
			ptmpList->pnext = pPFStack;
			pPFStack = ptmpList;
		    } else  {
			ptmpList->pnext = pTFStack;
			pTFStack = ptmpList;
		    }
		}
	    } else  {
		printf("unmaskEdges:  pedge should be 'S', and ");
		printf("pedge->pmasks->pmasks should be pedge.\n");
		exit(-1);
	    }
	} else  {
	    printf("\nunmaskEdges:  inconsistent masking data;\n");
	    printf("\tpedge->pmasks->masked = %c; must be S or P.\n\n",
		    pedge->pmasks->masked);
	    exit(-1);
	}
    }
    
    /*
     *  If the edge being removed from the edge struct 'H'-masks one or more
     *  other edges, then pedge->phardMasks is a list of pointers to them and
     *  pedge->pmasks is NIL.  To do hard-masking, pedge must have been 'S',
     *  so it must still be 'S' or it must be 'N'.
     */
    if (pedge->phardMasks != NIL(struct elist))  {
        if (pedge->masked != 'S'  &&  pedge->masked != 'N')  {
	    printf("\nunmaskEdges:  inconsistent masking data;\n");
	    printf("\tpedge hard-masks others, but pedge->masked = %c.\n\n",
		    pedge->masked);
	    exit(-1);
	}

	pHMList = pedge->phardMasks;
	while (pHMList != NIL(struct elist))  {
	    if (debug)  {
	        printf("unmaskEdges: pedge ==> resetting 'H' edge...");
	    }
	    pmaskedEdge = pHMList->pedge;

	    if (pmaskedEdge->masked == 'H')  {
		/*
		 *  pedge->key appears in the phardMaskedBy list of
		 *  pmaskedEdge; delete it.
		 */
		pHMB = pmaskedEdge->phardMaskedBy;
		plastHMB = NIL(struct klist);
		found = FALSE;
		while (pHMB != NIL(struct klist))  {
		    if (pHMB->key == pedge->key)  {
			found = TRUE;
			if (plastHMB == NIL(struct klist))  {
			    pmaskedEdge->phardMaskedBy = pHMB->pnext;
			} else  {
			    plastHMB->pnext = pHMB->pnext;
			}
			break;
		    } else  {
			plastHMB = pHMB;
			pHMB = pHMB->pnext;
		    }
		}
		if (found == FALSE)  {
		    printf("unmaskEdges:  no inst %d", pedge->key);
		    printf(" in phardMaskedBy list of pmaskedEdge (inst %d)\n",
			    pmaskedEdge->key);
		    exit(-1);
		}
		/*
		 *  Set pmaskedEdge->masked to 'N' IFF no other edges still
		 *  mask it.
		 */
		if (pmaskedEdge->phardMaskedBy == NIL(struct klist))  {
		    pmaskedEdge->masked = 'N';
		    numHardMasked--;
		    if (debug)  printf("  setting to 'N'\n");
		} else  {
		    if (debug)  printf("  leaving; others mask\n");
		}
	    } else  {
		printf("unmaskEdges:  bad tf-tf masking data;");
		printf(" pmaskedEdge->masked = %c, should be H.\n",
			pmaskedEdge->masked);
		exit(-1);
	    }

	    ptmpList = pHMList;
	    pHMList = pHMList->pnext;

	    /*
	     *  Since pmaskedEdge is being un-masked, push it onto the stack
	     *  of newly-visible edges if it extends past pedge.
	     */
	    if (pmaskedEdge->max > pedge->max)  {
		phardStack = ALLOC(struct hlist, 1);
		phardStack->pedge = pmaskedEdge;
		phardStack->hardMaskedBy = pedge->key;
		phardStack->pnext = phardTFStack;
		phardTFStack = phardStack;
	    }
	}
	/*
	 * XXX BAD!!  Should re-claim the elist memory...
	 *
	 */
	pedge->phardMasks = NIL(struct elist);
    }
}


/*
 *  Delete pedge from the edge structure.  If pedge masks any other edges, then
 *  something is wrong!
 *
 *  2/16/87
 *  last modified 6/26/87
 */

void
deleteFromEdgeStruct(pedge)
struct edge *pedge;
{
    if (pedge->active == 'N')  {
	printf("deleteFromEdgeStruct:  pedge isn't in edge struct\n");
	printf("\t %s edge from inst %d, ",
		(pedge->type == MIN) ? "MIN" : "MAX", pedge->key);
	printf("on layer %s, ", LayerMap[pedge->maskNumber].name);
	if (pedge->netId == UNDEF)  {
	    printf("netId UNDEF");
	} else  {
	    printf("netId %d", pedge->netId);
	}
	printf(" masked %c\n", pedge->masked);
	exit(-1);
    }

    if (pedge->pmasks != NIL(struct edge))  {
	printf("deleteFromEdgeStruct:  pedge '%c'-masks another edge.\n",
		pedge->pmasks->masked);
	printf("  pedge: %s from inst %d, ",
		(pedge->type == MIN) ? "MIN" : "MAX", pedge->key);
	printf("on layer %s, ", LayerMap[pedge->maskNumber].name);
	if (pedge->netId == UNDEF)  {
	    printf("netId UNDEF\n");
	} else  {
	    printf("netId %d\n", pedge->netId);
	}
	printf("  pedge->pmasks: %s from inst %d, ",
		(pedge->pmasks->type == MIN) ? "MIN" : "MAX",
		pedge->pmasks->key);
	printf("on layer %s, ", LayerMap[pedge->pmasks->maskNumber].name);
	if (pedge->pmasks->netId == UNDEF)  {
	    printf("netId UNDEF\n");
	} else  {
	    printf("netId %d\n", pedge->pmasks->netId);
	}
	/* exit(-1); */
    }
    
    if (pedge->phardMasks != NIL(struct elist))  {
	printf("deleteFromEdgeStruct:  pedge hard-masks other edges.\n");
	exit(-1);
    }

    pedge->active = 'N';
    dllDelete(&pEdgeHead, &pEdgeTail, pedge);

    return;
}




/*
 *  New routine to delete pedge from the edge structure.  Allows for edge
 *  pairs to be 'P' masked at this stage, IFF they have the same max coord.
 *  See notes of 2/16/88 for description.
 *
 *  from deleteFromEdgeStruct 2/16/88
 *  last modified 2/16/88
 */

void
newDeleteFromEdgeStruct(pedge)
struct edge *pedge;
{
    struct edge *pmaskedEdge;		/* edge masked by pedge */
    int debug = 0;

    if (pedge->active == 'N')  {
	printf("newDeleteFromEdgeStruct:  pedge isn't in edge struct\n");
	printf("\t %s edge from inst %d, ",
		(pedge->type == MIN) ? "MIN" : "MAX", pedge->key);
	printf("on layer %s, ", LayerMap[pedge->maskNumber].name);
	if (pedge->netId == UNDEF)  {
	    printf("netId UNDEF");
	} else  {
	    printf("netId %d", pedge->netId);
	}
	printf(" masked %c\n", pedge->masked);
	exit(-1);
    }

    if (pedge->pmasks != NIL(struct edge))  {
	pmaskedEdge = pedge->pmasks;
	if (pedge->masked == 'P'  &&  pmaskedEdge->masked == 'P'  &&
	    pmaskedEdge->pmasks == pedge)  {
	    /*
	     *  pedge does pf-pf masking; it masks only one edge.  This case
	     *  is OK IFF both edges end at the same coordinate.  All other
	     *  masking combinations are illegal at this point, as before.
	     *  (see deleteFromEdgeStruct).
	     */
	    if (pmaskedEdge->max != pedge->max)  {
		printf("%s %s",
		       "newDeleteFromEdgeStruct:  pedge 'P'-masks another",
		       "edge, but edges have different max coords.\n");
		printf("  pedge: %s from inst %d, ",
		       (pedge->type == MIN) ? "MIN" : "MAX", pedge->key);
		printf("on layer %s, ", LayerMap[pedge->maskNumber].name);
		if (pedge->netId == UNDEF)  {
		    printf("netId UNDEF\n");
		} else  {
		    printf("netId %d\n", pedge->netId);
		}
		printf("  pedge->pmasks: %s from inst %d, ",
		       (pedge->pmasks->type == MIN) ? "MIN" : "MAX",
		       pedge->pmasks->key);
		printf("on layer %s, ",
		       LayerMap[pedge->pmasks->maskNumber].name);
		if (pedge->pmasks->netId == UNDEF)  {
		    printf("netId UNDEF\n");
		} else  {
		    printf("netId %d\n", pedge->pmasks->netId);
		}
		exit(-1);
	    }

	    if (debug)  {
	        printf("newDeleteFromEdgeStruct:  pedge resetting 'P' edge\n");
	    }
	    pmaskedEdge->masked = 'N';
	    pmaskedEdge->pmasks = NIL(struct edge);
	    pedge->masked = 'N';
	    pedge->pmasks = NIL(struct edge);
	    numPFMasked -= 2;
	} else  {
	    printf("newDeleteFromEdgeStruct: pedge '%c'-masks another edge.\n",
		   pedge->pmasks->masked);
	    printf("  pedge: %s from inst %d, ",
		   (pedge->type == MIN) ? "MIN" : "MAX", pedge->key);
	    printf("on layer %s, ", LayerMap[pedge->maskNumber].name);
	    if (pedge->netId == UNDEF)  {
		printf("netId UNDEF\n");
	    } else  {
		printf("netId %d\n", pedge->netId);
	    }
	    printf("  pedge->pmasks: %s from inst %d, ",
		   (pedge->pmasks->type == MIN) ? "MIN" : "MAX",
		   pedge->pmasks->key);
	    printf("on layer %s, ", LayerMap[pedge->pmasks->maskNumber].name);
	    if (pedge->pmasks->netId == UNDEF)  {
		printf("netId UNDEF\n");
	    } else  {
		printf("netId %d\n", pedge->pmasks->netId);
	    }
	    /* exit(-1); */
	}
    }
    if (pedge->phardMasks != NIL(struct elist))  {
	printf("newDeleteFromEdgeStruct:  pedge hard-masks other edges.\n");
	exit(-1);
    }

    pedge->active = 'N';
    dllDelete(&pEdgeHead, &pEdgeTail, pedge);
}
