#include "copyright.h"
#include "sparcs.h"

/*
 *  Process the soft-masked terminal frame edges.  Note that such edges are,
 *  by definition, connected.
 *
 *  Each soft tf must perform 2 functions.  It must:
 *	1.  Provide the delta-terminal rule (to a same-polarity, same net, same
 *	    layer, nonmasked tf edge).
 *	2.  Do the job of the pf that is soft-masked by it.  That is, generate
 *	    rules w.r.t. same-polarity (opposite-polarity w.r.t. corresponding
 *	    soft-masked pf) tfs of the same net on DIFFERENT layers, and w.r.t.
 *	    same-polarity pfs, tfs of other nets, on all layers.
 *
 *  Note that the stopping criteria is:
 *	1.  opposite-polarity pf, or tf of another net; since this stops the
 *	    walkTF() scan.
 *
 *  This routine treats 'H' edges the same as 'N' edges.
 *
 *  Note that if pedge is MAX, the walk direction is down (to the left); if it
 *  is MIN, the walk is up (to the right).  This is the opposite of the other
 *  walk routines!
 *
 *  6/3/87
 *  last modified 9/22/88
 */

void
walkSoftTF(stateCoord, pedge, direction)
int stateCoord;
struct edge *pedge;	/* current edge */
directionType direction;
{
    struct edge *pscan;	    /* scanner to find edges related to cur. edge */
    struct elist *phM;
    struct klist *pHMB;
    int curNet;
    struct node *pfrom;
    struct node *pto;
    struct object *pedgeObj;
    struct object *pscanObj;
    struct wire *pscanWire;
    char mergeable;
    int edgeOrigin;
    int scanOrigin;
    int rule;
    int spacing;
    int debug = 0;

    if (pedge->masked != 'S')  {
	printf("walkSoftTF:  input edge must be soft-masked.\n");
	exit(-1);
    }
    if (pedge->netId == UNDEF)  {
	printf("walkSoftTF:  input edge must be a tf.\n");
	exit(-1);
    }

    if (pedge->key <= NumObjects)  {
        pedgeObj = ElementByKey[pedge->key].ptr.po;
	if (direction == HORIZONTAL)  {
	    edgeOrigin = pedgeObj->origin.x;
	} else  {
	    edgeOrigin = pedgeObj->origin.y;
	}
    } else  {
	printf("walkSoftTF:  input edge must be from an instance.\n");
	exit(-1);
    }

    curNet = pedge->netId;

    if (pedge->type == MIN)  {
	/*
	 *  Scan to the RIGHT, looking for MIN edges.  The search uncondition-
	 *  ally stops when a MAX pf edge on the same edge as pedge is reached.
	 *
	 *  Must process all edges with the same loc as pedge.  So, back up to
	 *  the left to the first edge in the edge structure with the same loc
	 *  before starting the scan to the right.
	 */
	if (pedge->pprev==NIL(struct edge) || pedge->pprev->loc<pedge->loc)  {
	    pscan = pedge->pnext;
	    /* if (debug)  printf("    (didn't back up)\n"); */
	} else  {
	    pscan = pedge->pprev;
	    for (; pscan != NIL(struct edge) &&
		   pscan->pprev != NIL(struct edge); pscan = pscan->pprev)  {
		if (pscan->pprev->loc != pedge->loc)  break;
	    }
	    /*
	    if (debug)  {
		printf("    (backed up to %s edge, inst %d, layer %s, ",
			(pscan->type == MIN) ? "MIN" : "MAX",
			pscan->key,
			LayerMap[pscan->maskNumber].name);
		if (pscan->netId == UNDEF)  {
		    printf("netId UNDEF)\n");
		} else  {
		    printf("netId %d)\n", pscan->netId);
		}
	    }
	    */
	}
	for (; pscan != NIL(struct edge); pscan = pscan->pnext)  {
	    /*
	     *  Don't compare two edges from the same element.  If pscan is
	     *  pf-tf (soft) masked ('S'), then it is ignored; if it is pf-pf
	     *  masked ('P'), then it is ignored.
	     */
	    if (pscan->key == pedge->key)  continue;
	    if (pscan->masked == 'S')  {
		if (debug)  {
		    printf("walkSoftTF:  skipping soft-masked edge...\n");
		}
		continue;
	    } else if (pscan->masked == 'P')  {
		if (debug)  {
		    printf("walkSoftTF:  skipping pf-pf-masked edge...\n");
		}
		continue;
	    }

	    if (pscan->netId == curNet)  {
		/*
		 *  pscan is a tf from the same net.  May need to apply the
		 *  delta-term'l rule if it is on the same layer and is not
		 *  masked.  Note that opposite-polarity (i.e., MAX) edges
		 *  don't matter here, since same net.
		 */
	        if (pscan->type == MIN)  {
		    if (pscan->maskNumber == pedge->maskNumber)  {
			rule = getRule(pedge->maskNumber, pscan->maskNumber,
					&mergeable);
			/*
			 *  See if pedge and pscan overlap sufficiently.
			 */
			if (stateCoord - pscan->max >= rule)  {
			    /*
			     *  Don't overlap enough; skip.
			     */
			    continue;
			}

			/*
			 *  Invoke the delta-term'l rule.  pscan->masked MUST
			 *  BE 'H' or 'N'.  It is 'H' if it is already hard-
			 *  masked by a different edge.
			 */
			if (pscan->masked != 'N'  &&  pscan->masked != 'H')  {
			    printf("walkSoftTF:  incorrectly masked (%c)",
				    pscan->masked);
			    printf(" edge reached.\n");
			    exit(-1);
			}

			if (debug)
			    printf("walkSoftTF:  adding delta-term'l rule\n");

			/*
			 *  Get the two origins, to determine the
			 *  direction of the branch.
			 */
			if (pscan->key <= NumObjects)  {
			    pscanObj = ElementByKey[pscan->key].ptr.po;
			    if (direction == HORIZONTAL)  {
				scanOrigin = pscanObj->origin.x;
			    } else  {
				scanOrigin = pscanObj->origin.y;
			    }
			} else  {
			    pscanWire = ElementByKey[pscan->key].ptr.pw;
			    if (direction == HORIZONTAL)  {
				scanOrigin = pscanWire->p1.x -
						(pscanWire->width)/2;
			    } else  {
				scanOrigin = pscanWire->p1.y -
						(pscanWire->width)/2;
			    }
			}

			/*
			 *  XXX This mess will go away if treat wires and
			 *      instances more uniformly.
			 */
			if (pscan->key <= NumObjects)  {
			    spacing = (pedge->loc - edgeOrigin) -
					(pscan->loc - scanOrigin);
			} else  {
			    /*
			     *  XXX FIX!!
			     *
			     *  Reject if object and wire are connected.
			     */
			    if (pedgeObj == pscanWire->po1  ||
				    pedgeObj == pscanWire->po2)  {
				continue;
			    } else  {
				spacing = pedge->loc - edgeOrigin;
			    }
			}

			if (edgeOrigin < scanOrigin)  {
			    /*
			     *  This should be the usual case, due to the
			     *  direction of the scan.  pedge belongs to
			     *  pfrom, pscan belongs to pto.
			     */
			    if (debug)  {
				printf("l.b. delta-term'l spacing constraint,");
				printf(" %d -> %d:  %d,\n",
					pedge->key, pscan->key, spacing);
				printf("\tpedge: MIN, at %d, on %s;",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
				printf("  pscan: MIN, at %d, on %s\n",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
			    }
			    pfrom = NodeArray[pedge->key];
			    pto   = NodeArray[pscan->key];
			    link(pfrom, pto, spacing, INFINITY, 0);
			} else if (scanOrigin < edgeOrigin)  {
			    /*
			     *  pscan belongs to pfrom, pedge belongs to
			     *  pto.  Must put spacing in as an upper.
			     */
			    if (debug)  {
				printf("u.b. delta-term'l spacing constraint,");
				printf(" %d -> %d:  %d,\n",
					pscan->key, pedge->key, -spacing);
				printf("\tpedge: MIN, at %d, on %s;",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
				printf("  pscan: MIN, at %d, on %s\n",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
			    }
			    pfrom = NodeArray[pscan->key];
			    pto   = NodeArray[pedge->key];
			    link(pfrom, pto, -INFINITY, -spacing, 0);
			} else  {
			    /*
			     *  tie; let pfrom be the one of lesser key.
			     */
			    if (pedge->key < pscan->key)  {
				/*
				 *  pedge belongs to pfrom, ...
				 */
				if (debug)  {
				    printf("l.b. delta-term'l spacing constraint,");
				    printf(" %d -> %d:  %d,\n",
					    pedge->key, pscan->key, spacing);
				    printf("\tpedge: MIN, at %d, on %s;",
					    pedge->loc,
					    LayerMap[pedge->maskNumber].name);
				    printf("  pscan: MIN, at %d, on %s\n",
					    pscan->loc,
					    LayerMap[pscan->maskNumber].name);
				}
				pfrom = NodeArray[pedge->key];
				pto   = NodeArray[pscan->key];
				link(pfrom, pto, spacing, INFINITY, 0);
			    } else  {
				/*
				 *  pscan belongs to pfrom, put in spacing
				 *  as an upper.
				 */
				if (debug)  {
				    printf("u.b. delta-term'l spacing constraint,");
				    printf(" %d -> %d:  %d,\n",
					    pscan->key, pedge->key, -spacing);
				    printf("\tpedge: MIN, at %d, on %s;",
					    pedge->loc,
					    LayerMap[pedge->maskNumber].name);
				    printf("  pscan: MIN, at %d, on %s\n",
					    pscan->loc,
					    LayerMap[pscan->maskNumber].name);
				}
				pfrom = NodeArray[pscan->key];
				pto   = NodeArray[pedge->key];
				link(pfrom, pto, -INFINITY,-spacing,0);
			    }
			}

			/*
			 *  Mark pscan as tf-tf masked by pedge; i.e., add
			 *  pscan to list of edges masked by pedge, and note
			 *  with pscan that it is masked by instance
			 *  pedge->key.
			 */
			pscan->masked = 'H';

			phM = ALLOC(struct elist, 1);
			phM->pedge = pscan;
			phM->pnext = pedge->phardMasks;
			pedge->phardMasks = phM;

			pHMB = ALLOC(struct klist, 1);
			pHMB->key = pedge->key;
			pHMB->pnext = pscan->phardMaskedBy;
			pscan->phardMaskedBy = pHMB;
		    } else  {
			/*
			 *  pedge is on a different layer, but is in the same
			 *  net as pedge.  Rule value is conditional on the
			 *  fact that the edges are in the same net.
			 */
			if (debug)  {
			    printf("\nwalkSoftTF:  NEED SAME-NET, CROSS LAYER");
			    printf(" RULES!!  ANSWER IS WRONG!!\n\n");
			}
		    }
		}
	    } else  {
		/*
		 *  pscan is from a different net, or it is a pf.  If it is a
		 *  a MIN edge, then add the appropriate constraint.  If it is
		 *  a MAX, and is on the same layer as pedge, then terminate
		 *  the walk.  Note that pedge is a protection frame in this
		 *  context!
		 */
		if (pscan->type == MIN)  {
		    /*
		     *  Potential edge from pedge's object to pscan's object.
		     */
		    rule = getRule(pedge->maskNumber, pscan->maskNumber,
				    &mergeable);
		    if (rule != NO_RULE)  {
			/*
			 *  Rule only if pedge and pscan overlap sufficiently.
			 */
			if (stateCoord - pscan->max >= rule)  {
			    /*
			     *  Don't overlap enough, so skip this edge.
			     */
			    continue;
			}
			/*
			 *  Get the two origins, to determine the direction of
			 *  the branch.
			 */
			if (pscan->key <= NumObjects)  {
			    pscanObj = ElementByKey[pscan->key].ptr.po;
			    if (direction == HORIZONTAL)  {
				scanOrigin = pscanObj->origin.x;
			    } else  {
				scanOrigin = pscanObj->origin.y;
			    }
			} else  {
			    pscanWire = ElementByKey[pscan->key].ptr.pw;
			    if (direction == HORIZONTAL)  {
				scanOrigin = pscanWire->p1.x -
						(pscanWire->width)/2;
			    } else  {
				scanOrigin = pscanWire->p1.y -
						(pscanWire->width)/2;
			    }
			}

			/*
			 *  XXX This mess will go away if treat wires and
			 *      instances more uniformly.
			 */
			if (pscan->key <= NumObjects)  {
			    spacing = (pedge->loc - edgeOrigin) + rule -
					(pscan->loc - scanOrigin);
			} else  {
			    /*
			     *  XXX FIX!!
			     *
			     *  Reject if object and wire are connected.
			     */
			    if (pedgeObj == pscanWire->po1  ||
				    pedgeObj == pscanWire->po2)  {
				continue;
			    } else  {
				spacing = (pedge->loc - edgeOrigin) + rule;
			    }
			}

			if (edgeOrigin < scanOrigin)  {
			    /*
			     *  This should be the usual case, due to the
			     *  direction of the scan.  pedge belongs to pfrom,
			     *  pscan belongs to pto.
			     */
			    if (debug)  {
				printf("l.b. spacing constraint, ");
				printf("%d -> %d:  %d,\n",
					pedge->key, pscan->key, spacing);
				printf("\tpedge: MIN, at %d, on %s;",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
				printf("  pscan: MIN, at %d, on %s\n",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
			    }
			    pfrom = NodeArray[pedge->key];
			    pto   = NodeArray[pscan->key];
			    link(pfrom, pto, spacing, INFINITY, 0);
			} else if (scanOrigin < edgeOrigin)  {
			    /*
			     *  pscan belongs to pfrom, pedge belongs to pto.
			     *  Must put spacing in as an upper.
			     */
			    if (debug)  {
				printf("u.b. spacing constraint, ");
				printf("%d -> %d:  %d,\n",
					pscan->key, pedge->key, -spacing);
				printf("\tpedge: MIN, at %d, on %s;",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
				printf("  pscan: MIN, at %d, on %s\n",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
			    }
			    pfrom = NodeArray[pscan->key];
			    pto   = NodeArray[pedge->key];
			    link(pfrom, pto, -INFINITY, -spacing, 0);
			} else  {
			    /*
			     *  tie; let pfrom be the one of lesser key.
			     */
			    if (pedge->key < pscan->key)  {
				/*
				 *  pedge belongs to pfrom, ...
				 */
				if (debug)  {
				    printf("l.b. spacing constraint, ");
				    printf("%d -> %d:  %d,\n",
					    pedge->key, pscan->key, spacing);
				    printf("\tpedge: MIN, at %d, on %s;",
					    pedge->loc,
					    LayerMap[pedge->maskNumber].name);
				    printf("  pscan: MIN, at %d, on %s\n",
					    pscan->loc,
					    LayerMap[pscan->maskNumber].name);
				}
				pfrom = NodeArray[pedge->key];
				pto   = NodeArray[pscan->key];
				link(pfrom, pto, spacing, INFINITY, 0);
			    } else  {
				/*
				 *  pscan belongs to pfrom, put in spacing as
				 *  an upper.
				 */
				if (debug)  {
				    printf("u.b. spacing constraint, ");
				    printf("%d -> %d:  %d,\n",
					    pscan->key, pedge->key, -spacing);
				    printf("\tpedge: MIN, at %d, on %s;",
					    pedge->loc,
					    LayerMap[pedge->maskNumber].name);
				    printf("  pscan: MIN, at %d, on %s\n",
					    pscan->loc,
					    LayerMap[pscan->maskNumber].name);
				}
				pfrom = NodeArray[pscan->key];
				pto   = NodeArray[pedge->key];
				link(pfrom, pto, -INFINITY, -spacing, 0);
			    }
			}
		    }
		} else  {
		    if (pscan->maskNumber == pedge->maskNumber)  {
			if (pscan->max > stateCoord)  {
			    if (debug)  {
				printf("Terminating scan; found edge w/same ");
				printf("polarity, same mask.\n");
			    }
			    return;
			}
		    }
		}
	    }
	}
    } else  {
	/*
	 *  pedge is a MAX edge; scan LEFT, looking for other MAX edges.  Stop
	 *  when a MIN pf on the same layer is reached.
	 *
	 *  Since must process all edges with the same loc as pedge, back up to
	 *  the right to the first edge in the edge structure with the same loc
	 *  before starting the scan to the left.
	 */
	if (pedge->pnext==NIL(struct edge) || pedge->pnext->loc>pedge->loc)  {
	    pscan = pedge->pprev;
	    /* if (debug)  printf("    (didn't back up)\n"); */
	} else  {
	    pscan = pedge->pnext;
	    for (; pscan != NIL(struct edge) &&
		   pscan->pnext != NIL(struct edge); pscan = pscan->pnext)  {
		if (pscan->pnext->loc != pedge->loc)  break;
	    }
	    /*
	    if (debug)  {
		printf("    (backed up to %s edge, inst %d, layer %s, ",
			(pscan->type == MIN) ? "MIN" : "MAX",
			pscan->key,
			LayerMap[pscan->maskNumber].name);
		if (pscan->netId == UNDEF)  {
		    printf("netId UNDEF)\n");
		} else  {
		    printf("netId %d)\n", pscan->netId);
		}
	    }
	    */
	}
	for (; pscan != NIL(struct edge); pscan = pscan->pprev)  {
	    if (pscan->key == pedge->key)  continue;
	    if (pscan->masked == 'S')  {
		if (debug)  {
		    printf("walkSoftTF:  skipping soft-masked edge...\n");
		}
		continue;
	    } else if (pscan->masked == 'P')  {
		if (debug)  {
		    printf("walkSoftTF:  skipping pf-pf-masked edge...\n");
		}
		continue;
	    }
	    if (pscan->netId == curNet)  {
		/*
		 *  pscan is a tf from the same net.  May need to apply the
		 *  delta-term'l rule if it is on the same layer and is not
		 *  masked.  Note that opposite-polarity edges don't matter
		 *  here, since same net.
		 */
	        if (pscan->type == MAX)  {
		    if (pscan->maskNumber == pedge->maskNumber)  {

			rule = getRule(pedge->maskNumber, pscan->maskNumber,
					&mergeable);
			if (stateCoord - pscan->max >= rule)  {
			    continue;
			}

			/*
			 *  Invoke the delta-term'l rule.  pscan->masked MUST
			 *  BE 'H' or 'N'.
			 */
			if (pscan->masked != 'N'  &&  pscan->masked != 'H')  {
			    printf("walkSoftTF:  incorrectly masked (%c)",
				    pscan->masked);
			    printf(" edge reached.\n");
			    exit(-1);
			}

			if (debug)
			    printf("walkSoftTF:  adding delta-term'l rule\n");

			if (pscan->key <= NumObjects)  {
			    pscanObj = ElementByKey[pscan->key].ptr.po;
			    if (direction == HORIZONTAL)  {
				scanOrigin = pscanObj->origin.x;
			    } else  {
				scanOrigin = pscanObj->origin.y;
			    }
			} else  {
			    pscanWire = ElementByKey[pscan->key].ptr.pw;
			    if (direction == HORIZONTAL)  {
				scanOrigin = pscanWire->p1.x -
						(pscanWire->width)/2;
			    } else  {
				scanOrigin = pscanWire->p1.y -
						(pscanWire->width)/2;
			    }
			}

			/*
			 *  scan element is left, edge element right.
			 */
			if (pscan->key <= NumObjects)  {
			    spacing = (pscan->loc - scanOrigin) -
					(pedge->loc - edgeOrigin);
			} else  {
			    /*
			     *  XXX FIX!!
			     *
			     *  Reject if object and wire are connected.
			     */
			    if (pscanWire->po1 == pedgeObj  ||
				    pscanWire->po2 == pedgeObj)  {
				continue;
			    } else  {
				spacing = pscanWire->width -
					    (pedge->loc - edgeOrigin);
			    }
			}

			if (scanOrigin < edgeOrigin)  {
			    /*
			     *  This should be the usual case, due to the
			     *  direction of the scan.  pedge belongs to
			     *  pto, pscan belongs to pfrom.
			     */
			    if (debug)  {
				printf("l.b. delta-term'l spacing constraint,");
				printf(" %d -> %d:  %d,\n",
					pscan->key, pedge->key, spacing);
				printf("\tpscan: MAX, at %d, on %s;n",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
				printf("  pedge: MAX, at %d, on %s\n",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
			    }
			    pfrom = NodeArray[pscan->key];
			    pto   = NodeArray[pedge->key];
			    link(pfrom, pto, spacing, INFINITY, 0);
			} else if (edgeOrigin < scanOrigin)  {
			    /*
			     *  pscan belongs to pto, pedge belongs to
			     *  pfrom.  Must put spacing in as an upper.
			     */
			    if (debug)  {
				printf("u.b. delta-term'l spacing constraint,");
				printf(" %d -> %d:  %d,\n",
					pedge->key, pscan->key, -spacing);
				printf("\tpscan: MAX, at %d, on %s;n",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
				printf("  pedge: MAX, at %d, on %s\n",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
			    }
			    pfrom = NodeArray[pedge->key];
			    pto   = NodeArray[pscan->key];
			    link(pfrom, pto, -INFINITY, -spacing, 0);
			} else  {
			    /*
			     *  tie; let pfrom be the one of lesser key.
			     */
			    if (pscan->key < pedge->key)  {
				/*
				 *  pedge belongs to pto, ...
				 */
				if (debug)  {
				    printf("l.b. delta-term'l spacing constraint,");
				    printf(" %d -> %d:  %d,\n",
					    pscan->key, pedge->key, spacing);
				    printf("\tpscan: MAX, at %d, on %s;n",
					    pscan->loc,
					    LayerMap[pscan->maskNumber].name);
				    printf("  pedge: MAX, at %d, on %s\n",
					    pedge->loc,
					    LayerMap[pedge->maskNumber].name);
				}
				pfrom = NodeArray[pscan->key];
				pto   = NodeArray[pedge->key];
				link(pfrom, pto, spacing, INFINITY, 0);
			    } else  {
				/*
				 *  pscan belongs to pto, put in spacing
				 *  as an upper.
				 */
				if (debug)  {
				    printf("u.b. delta-term'l spacing constraint,");
				    printf(" %d -> %d:  %d,\n",
					    pedge->key, pscan->key, -spacing);
				    printf("\tpscan: MAX, at %d, on %s;n",
					    pscan->loc,
					    LayerMap[pscan->maskNumber].name);
				    printf("  pedge: MAX, at %d, on %s\n",
					    pedge->loc,
					    LayerMap[pedge->maskNumber].name);
				}
				pfrom = NodeArray[pedge->key];
				pto   = NodeArray[pscan->key];
				link(pfrom, pto, -INFINITY,-spacing,0);
			    }
			}

			/*
			 *  Mark pscan as tf-tf masked by pedge; i.e., add
			 *  pscan to list of edges masked by pedge, and note
			 *  with pscan that it is masked by instance
			 *  pedge->key.
			 */
			pscan->masked = 'H';

			phM = ALLOC(struct elist, 1);
			phM->pedge = pscan;
			phM->pnext = pedge->phardMasks;
			pedge->phardMasks = phM;

			pHMB = ALLOC(struct klist, 1);
			pHMB->key = pedge->key;
			pHMB->pnext = pscan->phardMaskedBy;
			pscan->phardMaskedBy = pHMB;
		    } else  {
			/*
			 *  pedge is on a different layer, but is in the same
			 *  net as pedge.  Rule value is conditional on the
			 *  fact that the edges are in the same net.
			 */
			if (debug)  {
			    printf("\nwalkSoftTF:  NEED SAME-NET, CROSS LAYER");
			    printf(" RULES!!  ANSWER IS WRONG!!\n\n");
			}
		    }
		}
	    } else  {
		/*
		 *  pscan is from a different net, or it is a pf.  If it is a
		 *  a MAX edge, then add the appropriate constraint.  If it is
		 *  a MIN, and is on the same layer as pedge, then terminate
		 *  the walk.  Note that pedge is a protection frame in this
		 *  context!
		 */
		if (pscan->type == MAX)  {
		    /*
		     *  Potential edge from pscan's object to pedge's object.
		     */
		    rule = getRule(pedge->maskNumber, pscan->maskNumber,
				    &mergeable);
		    if (rule != NO_RULE)  {
			if (stateCoord - pscan->max >= rule)  {
			    continue;
			}
			if (pscan->key <= NumObjects)  {
			    pscanObj = ElementByKey[pscan->key].ptr.po;
			    if (direction == HORIZONTAL)  {
				scanOrigin = pscanObj->origin.x;
			    } else  {
				scanOrigin = pscanObj->origin.y;
			    }
			} else  {
			    pscanWire = ElementByKey[pscan->key].ptr.pw;
			    if (direction == HORIZONTAL)  {
				scanOrigin = pscanWire->p1.x -
						(pscanWire->width)/2;
			    } else  {
				scanOrigin = pscanWire->p1.y -
						(pscanWire->width)/2;
			    }
			}

			/*
			 *  scan element is left, edge element right.
			 */
			if (pscan->key <= NumObjects)  {
			    spacing = (pscan->loc - scanOrigin) + rule -
					(pedge->loc - edgeOrigin);
			} else  {
			    /*
			     *  XXX FIX!!
			     *
			     *  Reject if object and wire are connected.
			     */
			    if (pscanWire->po1 == pedgeObj  ||
				    pscanWire->po2 == pedgeObj)  {
				continue;
			    } else  {
				spacing = pscanWire->width + rule -
					    (pedge->loc - edgeOrigin);
			    }
			}

			if (scanOrigin < edgeOrigin)  {
			    /*
			     *  This should be the usual case, due to the
			     *  direction of the scan.  pedge belongs to pto,
			     *  pscan belongs to pfrom.
			     */
			    if (debug)  {
				printf("l.b. spacing constraint, ");
				printf("%d -> %d:  %d,\n",
					pscan->key, pedge->key, spacing);
				printf("\tpscan: MAX, at %d, on %s;",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
				printf("  pedge: MAX, at %d, on %s\n",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
			    }
			    pfrom = NodeArray[pscan->key];
			    pto   = NodeArray[pedge->key];
			    link(pfrom, pto, spacing, INFINITY, 0);
			} else if (edgeOrigin < scanOrigin)  {
			    /*
			     *  pscan belongs to pto, pedge belongs to pfrom.
			     *  Must put spacing in as an upper.
			     */
			    if (debug)  {
				printf("u.b. spacing constraint, ");
				printf("%d -> %d:  %d,\n",
					pedge->key, pscan->key, -spacing);
				printf("\tpscan: MAX, at %d, on %s;",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
				printf("  pedge: MAX, at %d, on %s\n",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
			    }
			    pfrom = NodeArray[pedge->key];
			    pto   = NodeArray[pscan->key];
			    link(pfrom, pto, -INFINITY, -spacing, 0);
			} else  {
			    /*
			     *  tie; let pfrom be the one of lesser key.
			     */
			    if (pscan->key < pedge->key)  {
				/*
				 *  pedge belongs to pto, ...
				 */
				if (debug)  {
				    printf("l.b. spacing constraint, ");
				    printf("%d -> %d:  %d,\n",
					    pscan->key, pedge->key, spacing);
				    printf("\tpscan: MAX, at %d, on %s;",
					    pscan->loc,
					    LayerMap[pscan->maskNumber].name);
				    printf("  pedge: MAX, at %d, on %s\n",
					    pedge->loc,
					    LayerMap[pedge->maskNumber].name);
				}
				pfrom = NodeArray[pscan->key];
				pto   = NodeArray[pedge->key];
				link(pfrom, pto, spacing, INFINITY, 0);
			    } else  {
				/*
				 *  pscan belongs to pto, put in spacing as an
				 *  upper.
				 */
				if (debug)  {
				    printf("u.b. spacing constraint, ");
				    printf("%d -> %d:  %d,\n",
					    pedge->key, pscan->key, -spacing);
				    printf("\tpscan: MAX, at %d, on %s;",
					    pscan->loc,
					    LayerMap[pscan->maskNumber].name);
				    printf("  pedge: MAX, at %d, on %s\n",
					    pedge->loc,
					    LayerMap[pedge->maskNumber].name);
				}
				pfrom = NodeArray[pedge->key];
				pto   = NodeArray[pscan->key];
				link(pfrom, pto, -INFINITY, -spacing, 0);
			    }
			}
		    }
		} else  {
		    if (pscan->maskNumber == pedge->maskNumber)  {
			if (pscan->max > stateCoord)  {
			    if (debug)  {
				printf("Terminating scan; found edge w/same ");
				printf("polarity, same mask.\n");
			    }
			    return;
			}
		    }
		}
	    }
	}
    }
}
