#include "copyright.h"
#include "sparcs.h"

/*
 *  Find any constraints due to pedge, which has just been added to the edge
 *  structure.  All input edges processed by this routine must be connected
 *  tfs which ARE NOT soft-masked or hard-masked.
 *
 *  MERGEABLE tf-tf ('H') masked edges that are found while walking the edge
 *  struct are treated as 'N' edges here.
 *
 *  Nodes are linked such that the node with the lesser coordinate is the
 *  fromNode and the node with the greater coordinate is the toNode.  In a tie,
 *  the node with the lesser key is the fromNode.
 *
 *  6/2/87
 *  last modified 9/22/88
 */

void
walkTF(stateCoord, pedge, direction)
int stateCoord;		/* current state coord; i.e., MIN coord of pedge */
struct edge *pedge;	/* current edge */
directionType direction;
{
    struct edge *pscan;	    	/* scanner to search edge struct */
    struct klist *pHMB;
    struct elist *phM;
    struct edge *pfirstLoc;	/* first edge at current loc in edge struct */
    int curNet;		    	/* netId of pedge */
    int curLoc;
    char mergeable;
    struct node *pfrom;
    struct node *pto;
    struct object *pedgeObj;
    struct wire *pedgeWire;
    struct object *pscanObj;
    struct wire *pscanWire;
    int edgeOrigin;
    int scanOrigin;
    int rule;
    int spacing;
    int debug = 0;

    if (pedge->masked != 'N')  {
	printf("walkTF:  input edge as masked = %c; must be 'N'\n",
		pedge->masked);
	exit(-1);
    }

    if (pedge->key <= NumObjects)  {
        pedgeObj = ElementByKey[pedge->key].ptr.po;
	if (direction == HORIZONTAL)  {
	    edgeOrigin = pedgeObj->origin.x;
	} else  {
	    edgeOrigin = pedgeObj->origin.y;
	}
    } else  {
        pedgeWire = ElementByKey[pedge->key].ptr.pw;
	if (direction == HORIZONTAL)  {
	    edgeOrigin = pedgeWire->p1.x - pedgeWire->halfWidth;
	} else  {
	    edgeOrigin = pedgeWire->p1.y - pedgeWire->halfWidth;
	}
    }

    curNet = pedge->netId;

    if (pedge->type == MAX)  {
	/*
	 *  Must process all edges with the same loc as pedge.  So, back up to
	 *  the left to the first edge in the edge structure with the same loc
	 *  before starting the scan to the right.
	 */
	if (pedge->pprev==NIL(struct edge) || pedge->pprev->loc<pedge->loc)  {
	    pscan = pedge->pnext;
	    /* if (debug)  printf("    (didn't back up)\n"); */
	} else  {
	    pscan = pedge->pprev;
	    for (; pscan != NIL(struct edge) &&
		   pscan->pprev != NIL(struct edge); pscan = pscan->pprev)  {
		if (pscan->pprev->loc != pedge->loc)  break;
	    }
	    /*
	    if (debug)  {
		printf("    (backed up to %s edge, inst %d, layer %s, ",
			(pscan->type == MIN) ? "MIN" : "MAX",
			pscan->key,
			LayerMap[pscan->maskNumber].name);
		if (pscan->netId == UNDEF)  {
		    printf("netId UNDEF)\n");
		} else  {
		    printf("netId %d)\n", pscan->netId);
		}
	    }
	    */
	}

	/*
	 *  Scan to the right, looking for .... edges.  The search uncondition-
	 *  ally stops when a MAX edge on the same edge as pedge is reached.
	 */
	while (pscan != NIL(struct edge))  {
	    /*
	     *  First, look at all edges at this loc for a soft-masked tf,
	     *  on the same layer, in the same net, and of the same polarity.
	     *  If found, compute the delta-term'l rule and return.  Otherwise
	     *  do the usual.
	     */
	    curLoc = pscan->loc;
	    pfirstLoc = pscan;
	    for (; pscan != NIL(struct edge) && pscan->loc == curLoc;
		    pscan = pscan->pnext)  {
		/*
		 *  This loop is only concerned with finding a same-polarity,
		 *  same-net, same-layer, soft-masked tf; so ignore all others.
		 */
		if (pscan->key == pedge->key)  continue;
		if (pscan->type == MAX  &&  pscan->netId == curNet  &&
			pscan->maskNumber == pedge->maskNumber  && 
			pscan->masked == 'S')  {

		    /*
		     *  There is a relationship only if pedge and pscan
		     *  overlap sufficiently.
		     */
		    rule = getRule(pedge->maskNumber, pscan->maskNumber,
					&mergeable);
		    if (stateCoord - pscan->max >= rule)  {
			/*
			 *  Don't overlap enough, so skip this edge.
			 */
			continue;
		    }

		    if (debug)  {
			printf("walkTF:  found delta-term'l rule edge.\n");
		    }

		    /*
		     *  Get the two origins, to determine the direction of
		     *  the branch.
		     */
		    if (pscan->key <= NumObjects)  {
			pscanObj = ElementByKey[pscan->key].ptr.po;
			if (direction == HORIZONTAL)  {
			    scanOrigin = pscanObj->origin.x;
			} else  {
			    scanOrigin = pscanObj->origin.y;
			}
		    } else  {
			pscanWire = ElementByKey[pscan->key].ptr.pw;
			if (direction == HORIZONTAL)  {
			    scanOrigin = pscanWire->p1.x -
					    (pscanWire->width)/2;
			} else  {
			    scanOrigin = pscanWire->p1.y -
					    (pscanWire->width)/2;
			}
		    }

		    /*
		     *  XXX This mess will go away if treat wires and
		     *      instances more uniformly.
		     */
		    if (pscan->key <= NumObjects)  {
			spacing = (pedge->loc - edgeOrigin) -
				    (pscan->loc - scanOrigin);
		    } else  {
			/*
			 *  XXX FIX!!
			 *
			 *  Reject if object and wire are connected.
			 */
			if (pedgeObj == pscanWire->po1  ||
				pedgeObj == pscanWire->po2)  {
			    continue;
			} else  {
			    spacing = pedge->loc - edgeOrigin;
			}
		    }

		    if (edgeOrigin < scanOrigin)  {
			/*
			 *  This should be the usual case, due to the
			 *  direction of the scan.  pedge belongs to pfrom,
			 *  pscan belongs to pto.
			 */
			if (debug)  {
			    printf("l.b. delta-term'l spacing constraint, ");
			    printf("%d -> %d:  %d,\n", pedge->key, pscan->key,
				    spacing);
			    printf("\tpedge: MAX, at %d, on %s;", pedge->loc,
				    LayerMap[pedge->maskNumber].name);
			    printf("  pscan: MAX, at %d, on %s\n", pscan->loc,
				    LayerMap[pscan->maskNumber].name);
			}
			pfrom = NodeArray[pedge->key];
			pto   = NodeArray[pscan->key];
			link(pfrom, pto, spacing, INFINITY, 0);
		    } else if (scanOrigin < edgeOrigin)  {
			/*
			 *  pscan belongs to pfrom, pedge belongs to pto.
			 *  Must put spacing in as an upper.
			 */
			if (debug)  {
			    printf("u.b. delta-term'l spacing constraint,");
			    printf(" %d -> %d:  %d,\n",
				    pscan->key, pedge->key, -spacing);
			    printf("\tpedge: MAX, at %d, on %s;",
				    pedge->loc,
				    LayerMap[pedge->maskNumber].name);
			    printf("  pscan: MAX, at %d, on %s\n",
				    pscan->loc,
				    LayerMap[pscan->maskNumber].name);
			}
			pfrom = NodeArray[pscan->key];
			pto   = NodeArray[pedge->key];
			link(pfrom, pto, -INFINITY, -spacing, 0);
		    } else  {
			/*
			 *  tie; let pfrom be the one of lesser key.
			 */
			if (pedge->key < pscan->key)  {
			    /*
			     *  pedge belongs to pfrom, ...
			     */
			    if (debug)  {
				printf("l.b. delta-term'l spacing constraint,");
				printf(" %d -> %d:  %d,\n",
					pedge->key, pscan->key, spacing);
				printf("\tpedge: MAX, at %d, on %s;",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
				printf("  pscan: MAX, at %d, on %s\n",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
			    }
			    pfrom = NodeArray[pedge->key];
			    pto   = NodeArray[pscan->key];
			    link(pfrom, pto, spacing, INFINITY, 0);
			} else  {
			    /*
			     *  pscan belongs to pfrom, put in spacing as
			     *  an upper.
			     */
			    if (debug)  {
				printf("u.b. delta-term'l spacing constraint,");
				printf(" %d -> %d:  %d,\n",
					pscan->key, pedge->key, -spacing);
				printf("\tpedge: MAX, at %d, on %s;",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
				printf("  pscan: MAX, at %d, on %s\n",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
			    }
			    pfrom = NodeArray[pscan->key];
			    pto   = NodeArray[pedge->key];
			    link(pfrom, pto, -INFINITY, -spacing, 0);
			}
		    }

		    /*
		     *  Mask pedge and add pedge to list of edges masked by
		     *  pscan, and note with pedge that it is masked by inst.
		     *  pscan->key.
		     */
		    pedge->masked = 'H';

		    pHMB = ALLOC(struct klist, 1);
		    pHMB->key = pscan->key;
		    pHMB->pnext = pedge->phardMaskedBy;
		    pedge->phardMaskedBy = pHMB;

		    phM = ALLOC(struct elist, 1);
		    phM->pedge = pedge;
		    phM->pnext = pscan->phardMasks;
		    pscan->phardMasks = phM;

		    /*
		     *  This edge is finished!
		     */
		    return;
		}
	    }

	    /*
	     *  Since this point has been reached, there is no delta-term'l
	     *  relationship at the current loc.  Go thru the edges at this
	     *  loc again, putting in any relevant rules.
	     */
	    pscan = pfirstLoc;
	    for (; pscan != NIL(struct edge) && pscan->loc == curLoc ;
		    pscan = pscan->pnext)  {
		/*
		 *  Skip edges from the same element, edges from the same
		 *  net on MERGEABLE layers, and masked edges.
		 *
		 *  First weed-out the obvious ones.
		 */
		if (pscan->key == pedge->key)  continue;
		if (pscan->masked != 'N')  {
		    if (debug)  {
			if (pscan->type == MIN)  {
			    printf("walkTF:  skipping masked edge...\n");
			}
		    }
		    continue;
		}
		if (pscan->netId == curNet  &&
			pscan->maskNumber == pedge->maskNumber)  {
		    if (debug)  {
			if (pscan->type == MIN)  {
			    printf("walkTF:  skipping edge from same net,");
			    printf(" same layer.\n");
			}
		    }
		    continue;
		}
		/*
		 *  If the edge is a MIN edge, there may be a rule.
		 */
		if (pscan->type == MIN)  {

		    rule = getRule(pedge->maskNumber, pscan->maskNumber,
					&mergeable);
		    if (rule != NO_RULE)  {
			/*
			 *  Skip the case of mergeable layers in the same net.
			 */
			if (pscan->netId == curNet)  {
			    if (mergeable == 'Y')  {
				if (debug)  {
				    printf("walkTF:  skipping MERGEABLE edge ");
				    printf("from same net, different layer.\n");
				}
				continue;
			    } else  {
				if (debug)  {
				    printf("walkTF:  not skipping edge from ");
				    printf("same net, different layer;");
				    printf(" NOT mergeable.\n");
				}
			    }
			}

			/*
			 *  There is a rule, because the layers are not merge-
			 *  able or because the edges are from different nets,
			 *  but IFF pedge and pscan overlap sufficiently.
			 */
			if (stateCoord - pscan->max >= rule)  {
			    /*
			     *  Don't overlap enough, so skip this edge.
			     */
			    continue;
			}

			/*
			 *  Get the two origins, to determine the direction of
			 *  the branch.
			 */
			if (pscan->key <= NumObjects)  {
			    pscanObj = ElementByKey[pscan->key].ptr.po;
			    if (direction == HORIZONTAL)  {
				scanOrigin = pscanObj->origin.x;
			    } else  {
				scanOrigin = pscanObj->origin.y;
			    }
			} else  {
			    pscanWire = ElementByKey[pscan->key].ptr.pw;
			    if (direction == HORIZONTAL)  {
				scanOrigin = pscanWire->p1.x -
						pscanWire->halfWidth;
			    } else  {
				scanOrigin = pscanWire->p1.y -
						pscanWire->halfWidth;
			    }
			}

			/*
			 *  XXX This mess will go away if treat wires and
			 *      instances more uniformly.
			 */
			if (pedge->key <= NumObjects)  {
			    if (pscan->key <= NumObjects)  {
				spacing = (pedge->loc - edgeOrigin) + rule -
					    (pscan->loc - scanOrigin);
			    } else  {
				/*
				 *  XXX FIX!!
				 *
				 *  Reject if object and wire are connected.
				 */
				if (pedgeObj == pscanWire->po1  ||
					pedgeObj == pscanWire->po2)  {
				    continue;
				} else  {
				    spacing = (pedge->loc - edgeOrigin) + rule;
				}
			    }
			} else  {
			    if (pscan->key <= NumObjects)  {
				/*
				 *  XXX FIX!!
				 *
				 *  Reject if object and wire are connected.
				 */
				if (pedgeWire->po1 == pscanObj  ||
					pedgeWire->po2 == pscanObj)  {
				    continue;
				} else  {
				    spacing = pedgeWire->width + rule -
						(pscan->loc - scanOrigin);
				}
			    } else  {
				spacing = pedgeWire->width + rule;
			    }
			}

			if (edgeOrigin < scanOrigin)  {
			    /*
			     *  This should be the usual case, due to the
			     *  direction of the scan.  pedge belongs to pfrom,
			     *  pscan belongs to pto.
			     */
			    if (debug)  {
				printf("l.b. spacing constraint, ");
				printf("%d -> %d:  %d,\n",
					pedge->key, pscan->key, spacing);
				printf("\tpedge: MAX, at %d, on %s;",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
				printf("  pscan: MIN, at %d, on %s\n",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
			    }
			    pfrom = NodeArray[pedge->key];
			    pto   = NodeArray[pscan->key];
			    link(pfrom, pto, spacing, INFINITY, 0);
			} else if (scanOrigin < edgeOrigin)  {
			    /*
			     *  pscan belongs to pfrom, pedge belongs to pto.
			     *  Must put spacing in as an upper.
			     */
			    if (debug)  {
				printf("u.b. spacing constraint, ");
				printf("%d -> %d:  %d,\n",
					pscan->key, pedge->key, -spacing);
				printf("\tpedge: MAX, at %d, on %s;",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
				printf("  pscan: MIN, at %d, on %s\n",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
			    }
			    pfrom = NodeArray[pscan->key];
			    pto   = NodeArray[pedge->key];
			    link(pfrom, pto, -INFINITY, -spacing, 0);
			} else  {
			    /*
			     *  tie; let pfrom be the one of lesser key.
			     */
			    if (pedge->key < pscan->key)  {
				/*
				 *  pedge belongs to pfrom, ...
				 */
				if (debug)  {
				    printf("l.b. spacing constraint, ");
				    printf("%d -> %d:  %d,\n",
					    pedge->key, pscan->key, spacing);
				    printf("\tpedge: MAX, at %d, on %s;",
					    pedge->loc,
					    LayerMap[pedge->maskNumber].name);
				    printf("  pscan: MIN, at %d, on %s\n",
					    pscan->loc,
					    LayerMap[pscan->maskNumber].name);
				}
				pfrom = NodeArray[pedge->key];
				pto   = NodeArray[pscan->key];
				link(pfrom, pto, spacing, INFINITY, 0);
			    } else  {
				/*
				 *  pscan belongs to pfrom, put in spacing as
				 *  an upper.
				 */
				if (debug)  {
				    printf("u.b. spacing constraint, ");
				    printf("%d -> %d:  %d,\n",
					    pscan->key, pedge->key, -spacing);
				    printf("\tpedge: MAX, at %d, on %s;",
					    pedge->loc,
					    LayerMap[pedge->maskNumber].name);
				    printf("  pscan: MIN, at %d, on %s\n",
					    pscan->loc,
					    LayerMap[pscan->maskNumber].name);
				}
				pfrom = NodeArray[pscan->key];
				pto   = NodeArray[pedge->key];
				link(pfrom, pto, -INFINITY, -spacing, 0);
			    }
			}
		    }
		} else if (pscan->maskNumber == pedge->maskNumber)  {
		    if (pscan->max > stateCoord)  {
			if (debug)  {
			    printf("Terminating scan; found edge w/same ");
			    printf("polarity, same mask.\n");
			}
			return;
		    }
		}
	    }
	}
    } else  {
	/*
	 *  Since must process all edges with the same loc as pedge, back up to
	 *  the right to the first edge in the edge structure with the same loc
	 *  before starting the scan to the left.
	 */
	if (pedge->pnext==NIL(struct edge) || pedge->pnext->loc>pedge->loc)  {
	    pscan = pedge->pprev;
	    /* if (debug)  printf("    (didn't back up)\n"); */
	} else  {
	    pscan = pedge->pnext;
	    for (; pscan != NIL(struct edge) &&
		   pscan->pnext != NIL(struct edge); pscan = pscan->pnext)  {
		if (pscan->pnext->loc != pedge->loc)  break;
	    }
	    /*
	    if (debug)  {
		printf("    (backed up to %s edge, inst %d, layer %s, ",
			(pscan->type == MIN) ? "MIN" : "MAX",
			pscan->key,
			LayerMap[pscan->maskNumber].name);
		if (pscan->netId == UNDEF)  {
		    printf("netId UNDEF)\n");
		} else  {
		    printf("netId %d)\n", pscan->netId);
		}
	    }
	    */
	}
	/*
	 *  Scan to the left, looking for MAX edges.
	 */
	while (pscan != NIL(struct edge))  {
	    /*
	     *  First, look at all edges at this loc for a soft-masked tf,
	     *  on the same layer, in the same net, and of the same polarity.
	     *  If found, compute the delta-term'l rule and return.  Otherwise
	     *  do the usual.
	     */
	    curLoc = pscan->loc;
	    pfirstLoc = pscan;
	    for (; pscan != NIL(struct edge) && pscan->loc == curLoc;
		    pscan = pscan->pprev)  {
		/*
		 *  This loop is only concerned with finding a soft-masked,
		 *  same-polarity, same-net tf.  So ignore all others.
		 */
		if (pscan->key == pedge->key)  continue;
		if (pscan->type == MIN  &&  pscan->netId == curNet  &&
			pscan->maskNumber == pedge->maskNumber  &&
			pscan->masked == 'S')  {
	
		    /*
		     *  There is a relationship only if pedge and pscan
		     *  overlap sufficiently.
		     */
		    rule = getRule(pedge->maskNumber, pscan->maskNumber,
					&mergeable);
		    if (stateCoord - pscan->max >= rule)  {
			/*
			 *  Don't overlap enough, so skip this edge.
			 */
			continue;
		    }

		    if (debug)  {
			printf("walkTF:  found delta-term'l rule edge.\n");
		    }

		    if (pscan->key <= NumObjects)  {
			pscanObj = ElementByKey[pscan->key].ptr.po;
			if (direction == HORIZONTAL)  {
			    scanOrigin = pscanObj->origin.x;
			} else  {
			    scanOrigin = pscanObj->origin.y;
			}
		    } else  {
			pscanWire = ElementByKey[pscan->key].ptr.pw;
			if (direction == HORIZONTAL)  {
			    scanOrigin = pscanWire->p1.x -
					    (pscanWire->width)/2;
			} else  {
			    scanOrigin = pscanWire->p1.y -
					    (pscanWire->width)/2;
			}
		    }

		    /*
		     *  scan element is left, edge element right.
		     */
		    if (pscan->key <= NumObjects)  {
			spacing = (pscan->loc - scanOrigin) -
				    (pedge->loc - edgeOrigin);
		    } else  {
			/*
			 *  XXX FIX!!
			 *
			 *  Reject if object and wire are connected.
			 */
			if (pscanWire->po1 == pedgeObj  ||
				pscanWire->po2 == pedgeObj)  {
			    continue;
			} else  {
			    spacing = pscanWire->width -
					(pedge->loc - edgeOrigin);
			}
		    }

		    if (scanOrigin < edgeOrigin)  {
			/*
			 *  This should be the usual case, due to the
			 *  direction of the scan.  pedge belongs to pto,
			 *  pscan belongs to pfrom.
			 */
			if (debug)  {
			    printf("l.b. delta-term'l spacing constraint, ");
			    printf("%d -> %d:  %d,\n", pscan->key, pedge->key,
				    spacing);
			    printf("  pscan: MIN, at %d, on %s\n", pscan->loc,
				    LayerMap[pscan->maskNumber].name);
			    printf("\tpedge: MIN, at %d, on %s\n", pedge->loc,
				    LayerMap[pedge->maskNumber].name);
			}
			pfrom = NodeArray[pscan->key];
			pto   = NodeArray[pedge->key];
			link(pfrom, pto, spacing, INFINITY, 0);
		    } else if (edgeOrigin < scanOrigin)  {
			/*
			 *  pscan belongs to pto, pedge belongs to pfrom.
			 *  Must put spacing in as an upper.
			 */
			if (debug)  {
			    printf("u.b. delta-term'l spacing constraint,");
			    printf(" %d -> %d:  %d,\n",
				    pedge->key, pscan->key, -spacing);
			    printf("  pscan: MIN, at %d, on %s\n",
				    pscan->loc,
				    LayerMap[pscan->maskNumber].name);
			    printf("\tpedge: MIN, at %d, on %s\n",
				    pedge->loc,
				    LayerMap[pedge->maskNumber].name);
			}
			pfrom = NodeArray[pedge->key];
			pto   = NodeArray[pscan->key];
			link(pfrom, pto, -INFINITY, -spacing, 0);
		    } else  {
			/*
			 *  tie; let pfrom be the one of lesser key.
			 */
			if (pscan->key < pedge->key)  {
			    /*
			     *  pedge belongs to pto, ...
			     */
			    if (debug)  {
				printf("l.b. delta-term'l spacing constraint");
				printf(" %d -> %d:  %d,\n",
					pscan->key, pedge->key, spacing);
				printf(",  pscan: MIN, at %d, on %s\n",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
				printf("\tpedge: MIN, at %d, on %s\n",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
			    }
			    pfrom = NodeArray[pscan->key];
			    pto   = NodeArray[pedge->key];
			    link(pfrom, pto, spacing, INFINITY, 0);
			} else  {
			    /*
			     *  pscan belongs to pto, put in spacing as an
			     *  upper.
			     */
			    if (debug)  {
				printf("u.b. delta-term'l spacing constraint");
				printf(", %d -> %d:  %d,\n",
					pedge->key, pscan->key, -spacing);
				printf("  pscan: MIN, at %d, on %s\n",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
				printf("\tpedge: MIN, at %d, on %s\n",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
			    }
			    pfrom = NodeArray[pedge->key];
			    pto   = NodeArray[pscan->key];
			    link(pfrom, pto, -INFINITY, -spacing, 0);
			}
		    }

		    /*
		     *  Add pedge to list of edges masked by pscan, and
		     *  note with pedge that it is masked by instance
		     *  pscan->key.
		     */
		    pedge->masked = 'H';

		    pHMB = ALLOC(struct klist, 1);
		    pHMB->key = pscan->key;
		    pHMB->pnext = pedge->phardMaskedBy;
		    pedge->phardMaskedBy = pHMB;

		    phM = ALLOC(struct elist, 1);
		    phM->pedge = pedge;
		    phM->pnext = pscan->phardMasks;
		    pscan->phardMasks = phM;
		 
		    return;
		}
	    }

	    /*
	     *  Since this point has been reached, there is no delta-term'l
	     *  relationship at the current loc.  Go thru the edges at this
	     *  loc again, putting in any relevant rules.
	     */
	    pscan = pfirstLoc;
	    for (; pscan != NIL(struct edge) && pscan->loc == curLoc ;
		    pscan = pscan->pprev)  {
		/*
		 *  Skip edges from the same element, edges from the same
		 *  net on MERGEABLE layers, and masked edges.
		 *
		 *  First weed-out the obvious ones.
		 */
		if (pscan->key == pedge->key)  continue;
		if (pscan->masked != 'N')  {
		    if (debug)  {
			if (pscan->type == MAX)  {
			    printf("walkTF:  skipping masked edge...\n");
			}
		    }
		    continue;
		}
		if (pscan->netId == curNet  &&
			pscan->maskNumber == pedge->maskNumber)  {
		    if (debug)  {
			if (pscan->type == MAX)  {
			    printf("walkTF:  skipping edge from same net,");
			    printf(" same layer.\n");
			}
		    }
		    continue;
		}

		/*
		 *  If the edge is a MAX edge, there may be a rule.
		 */
		if (pscan->type == MAX)  {

		    rule = getRule(pedge->maskNumber, pscan->maskNumber,
					&mergeable);
		    if (rule != NO_RULE)  {
			/*
			 *  Skip the case of mergeable layers in the same net.
			 */
			if (pscan->netId == curNet)  {
			    if (mergeable == 'Y')  {
				if (debug)  {
				    printf("walkTF:  skipping MERGEABLE edge ");
				    printf("from same net, different layer.\n");
				}
				continue;
			    } else  {
				if (debug)  {
				    printf("walkTF:  not skipping edge from ");
				    printf("same net, different layer;");
				    printf(" NOT mergeable.\n");
				}
			    }
			}

			/*
			 *  There is a rule, because the layers are not merge-
			 *  able or because the edges are from different nets,
			 *  but IFF pedge and pscan overlap sufficiently.
			 */
			if (stateCoord - pscan->max >= rule)  {
			    /*
			     *  Don't overlap enough, so skip this edge.
			     */
			    continue;
			}

			if (pscan->key <= NumObjects)  {
			    pscanObj = ElementByKey[pscan->key].ptr.po;
			    if (direction == HORIZONTAL)  {
				scanOrigin = pscanObj->origin.x;
			    } else  {
				scanOrigin = pscanObj->origin.y;
			    }
			} else  {
			    pscanWire = ElementByKey[pscan->key].ptr.pw;
			    if (direction == HORIZONTAL)  {
				scanOrigin = pscanWire->p1.x -
						pscanWire->halfWidth;
			    } else  {
				scanOrigin = pscanWire->p1.y -
						pscanWire->halfWidth;
			    }
			}

			/*
			 *  scan element is left, edge element right.
			 */
			if (pscan->key <= NumObjects)  {
			    if (pedge->key <= NumObjects)  {
				spacing = (pscan->loc - scanOrigin) + rule -
					    (pedge->loc - edgeOrigin);
			    } else  {
				/*
				 *  XXX FIX!!
				 *
				 *  Reject if object and wire are connected.
				 */
				if (pedgeWire->po1 == pscanObj  ||
					pedgeWire->po2 == pscanObj)  {
				    continue;
				} else  {
				    spacing = (pscan->loc - scanOrigin) + rule;
				}
			    }
			} else  {
			    if (pedge->key <= NumObjects)  {
				/*
				 *  XXX FIX!!
				 *
				 *  Reject if object and wire are connected.
				 */
				if (pscanWire->po1 == pedgeObj  ||
					pscanWire->po2 == pedgeObj)  {
				    continue;
				} else  {
				    spacing = pscanWire->width + rule -
						(pedge->loc - edgeOrigin);
				}
			    } else  {
				spacing = pscanWire->width + rule;
			    }
			}

			if (scanOrigin < edgeOrigin)  {
			    /*
			     *  This should be the usual case, due to the
			     *  direction of the scan.  pedge belongs to pto,
			     *  pscan belongs to pfrom.
			     */
			    if (debug)  {
				printf("l.b. spacing constraint, ");
				printf("%d -> %d:  %d,\n",
					pscan->key, pedge->key, spacing);
				printf("\tpscan: MAX, at %d, on %s;",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
				printf("  pedge: MIN, at %d, on %s\n",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
			    }
			    pfrom = NodeArray[pscan->key];
			    pto   = NodeArray[pedge->key];
			    link(pfrom, pto, spacing, INFINITY, 0);
			} else if (edgeOrigin < scanOrigin)  {
			    /*
			     *  pscan belongs to pto, pedge belongs to pfrom.
			     *  Must put spacing in as an upper.
			     */
			    if (debug)  {
				printf("u.b. spacing constraint, ");
				printf("%d -> %d:  %d,\n",
					pedge->key, pscan->key, -spacing);
				printf("\tpscan: MAX, at %d, on %s;",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
				printf(" pedge: MIN, at %d, on %s\n",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
			    }
			    pfrom = NodeArray[pedge->key];
			    pto   = NodeArray[pscan->key];
			    link(pfrom, pto, -INFINITY, -spacing, 0);
			} else  {
			    /*
			     *  tie; let pfrom be the one of lesser key.
			     */
			    if (pscan->key < pedge->key)  {
				/*
				 *  pedge belongs to pto, ...
				 */
				if (debug)  {
				    printf("l.b. spacing constraint, ");
				    printf("%d -> %d:  %d,\n",
					    pscan->key, pedge->key, spacing);
				    printf("\tpscan: MAX, at %d, on %s;",
					    pscan->loc,
					    LayerMap[pscan->maskNumber].name);
				    printf(" pedge: MIN, at %d, on %s\n",
					    pedge->loc,
					    LayerMap[pedge->maskNumber].name);
				}
				pfrom = NodeArray[pscan->key];
				pto   = NodeArray[pedge->key];
				link(pfrom, pto, spacing, INFINITY, 0);
			    } else  {
				/*
				 *  pscan belongs to pto, put in spacing as an
				 *  upper.
				 */
				if (debug)  {
				    printf("u.b. spacing constraint, ");
				    printf("%d -> %d:  %d,\n",
					    pedge->key, pscan->key, -spacing);
				    printf("\tpscan: MAX, at %d, on %s;",
					    pscan->loc,
					    LayerMap[pscan->maskNumber].name);
				    printf(" pedge: MIN, at %d, on %s\n",
					    pedge->loc,
					    LayerMap[pedge->maskNumber].name);
				}
				pfrom = NodeArray[pedge->key];
				pto   = NodeArray[pscan->key];
				link(pfrom, pto, -INFINITY, -spacing, 0);
			    }
			}
		    }
		} else if (pscan->type == MIN  &&
			    pscan->maskNumber == pedge->maskNumber)  {
		    if (pscan->max > stateCoord)  {
			if (debug)  {
			    printf("Terminating scan; found edge w/same");
			    printf(" polarity, same mask.\n");
			}
			return;
		    }
		}
	    }
	}
    }
}
