#include "copyright.h"
#include "sparcs.h"
#include "oct.h"
#include "tap.h"
#include "errtrap.h"
#include "th.h"
#include "oh.h"

int SparcsPrintRules = 0;
/*
 *  Read the spacing rules from a file and put them into Rules[][].  Also add
 *  the layers to LayerMap.
 *
 *  The first line in the file must be
 *
 *	    "technology techname"
 *
 *  where techname is the same name as that specified in the TECHNOLOGY
 *  property.
 *
 *  All other lines are for rules, which are given in the forms
 *
 *	    "layer1  layer2  spacing  mergeable"
 *	    "layer1  layer1  spacing  maxrule  resistance"
 *
 *  layer1 (2):	layer names
 *  spacing:	integer spacing rule, in lambda; scaled here by UNITSPERLAMBDA
 *  mergeable:	integer, 1 or 0
 *  maxrule:	largest spacing rule w.r.t. layer1, in lambda
 *  resistance:	integer, for force assignment in slack distribution phase.
 *		If == -1, then set to UNDEF.
 *
 *  Comments in the rules file are denoted by beginning the line with "#".
 *
 *  COMPUTE_SIGNATURE no longer used, because not guaranteed unique and,
 *  for short strings, strcmp() is faster!
 *
 *  Known as readRules() prior to 4/13/87.
 *
 *  1/29/86
 *  last modified 8/24/87
 */

void
readTech(pf, techName)
FILE *pf;
char *techName;
{
    int spacing, mergeable, maxRule, resistance;
    int tmpValue;
    int args;
    int foundLayer1, foundLayer2;
    int i;
    int sparcsLayer1, sparcsLayer2;
    char buffer[LINELENGTH];
    char layer1[80], layer2[80];	/* buffers for mask layer names */
    char keywd1[80], keywd2[80];	/* for 'width', 'ops' keywords */
    char *pc;
    int line = 0;
    int lastLayer = -1;			/* the index of the last FILLED entry
					    in LayerMap */
    
    if (RulesDB)  {
	if (RuleExp != 1)  {
	    printf("readTech:  rule expansion factor is %d.\n", RuleExp);
	}
	printf("readTech:  rules:\n");
    }

    /*
     *  Read the first (real) line for the technology name, which must match
     *  that found in the TECHNOLOGY property.
     */
    while ((pc=fgets(buffer, LINELENGTH, pf)) != NULL)  {
	args = sscanf(buffer, "%s %s", keywd1, keywd2);
	line++;
	if (keywd1[0] == '#')  {
	    continue;
	} else  {
	    break;
	}
    }
    if (pc == NULL)  {
	printf("readTech:  rules file is empty.\n");
	exit(-1);
    }
    
    if (args != 2  ||  strcmp(keywd1, "technology") != 0)  {
	printf("readTech:  first data line of rules file must be");
	printf(" \"`technology' techname\"\n");
	exit(-1);
    } else  {
	/*
	 *  Compare the names (unless TECHNOLOGY was omitted) and quit if there
	 *  is a mismatch.
	 */
	if (techName != NIL(char))  {
	    if (strcmp(techName, keywd2) != 0)  {
		printf("readTech:  TECHNOLOGY property has value \"%s\";\n",
		       techName);
		printf("\t   technology specified in rules file is \"%s\".\n",
		       keywd2);
		printf("\n\t   The two specifications must match.\n");
		exit(-1);
	    }
	}
    }
    
    /*
     *  Read each rule and fill in the rules matrix appropriately.  Each layer
     *  name is assigned a number, and each name has a unique signature assoc-
     *  iated with it.
     */
    while (fgets(buffer, LINELENGTH, pf) != NULL)  {
        args = sscanf(buffer, "%s %s %d %d %d", layer1, layer2, &spacing,
			&tmpValue, &resistance);
	line++;

	if (RulesDB)  (void) fputs(buffer, stdout);

	if (layer1[0] == '#')  continue;
	
	if (args == 4)  {
	    /*
	     *  Layers must be different.
	     */
	    if (strcmp(layer1, layer2) == 0)  {
		printf("readTech:  incorrect format, line %d;", line);
		printf("  legal formats:\n");
		printf("\tlayerName1 layerName2 value value\n");
		printf("\tlayerName1 layerName1 value value value\n");
		exit(-1);
	    }
	    mergeable = tmpValue;
	} else if (args == 5)  {
	    /*
	     *  Layers must be the same.
	     */
	    if (strcmp(layer1, layer2) != 0)  {
		printf("readTech:  incorrect format, line %d;", line);
		printf("  legal formats:\n");
		printf("\tlayerName1 layerName2 value value\n");
		printf("\tlayerName1 layerName1 value value value\n");
		exit(-1);
	    }
	    maxRule = tmpValue;
	} else  {
	    printf("readTech:  wrong number of entries, line %d;", line);
	    printf("  legal formats:\n");
	    printf("\tlayerName1 layerName2 value value\n");
	    printf("\tlayerName1 layerName1 value value value\n");
	    exit(-1);
	}

	spacing *= UNITSPERLAMBDA;
	if (RuleExp > 1)  spacing *= RuleExp;
	
	/*
	 *  See if the layers are new or not; add entries for any new ones to
	 *  LayerMap.  lastLayer is the number (and index of) the last internal
	 *  layer that has been assigned.
	 */
	foundLayer1 = foundLayer2 = FALSE;
	sparcsLayer1 = sparcsLayer2 = 0;
	for (i=0; i<=lastLayer; i++)  {
	    if (strcmp(LayerMap[i].name, layer1) == 0)  {
		sparcsLayer1 = i;
		foundLayer1 = TRUE;
		break;
	    }
	}
	if (args == 5)  {
	    sparcsLayer2 = sparcsLayer1;
	    foundLayer2 = foundLayer1;
	} else  {
	    for (i=0; i<=lastLayer; i++)  {
		if (strcmp(LayerMap[i].name, layer2) == 0)  {
		    sparcsLayer2 = i;
		    foundLayer2 = TRUE;
		    break;
		}
	    }
	}
	
	/*
	 *  There are 2 cases; when layer1 and layer2 are the same, and
	 *  when they are not.
	 */
	if (sparcsLayer1 == sparcsLayer2)  {
	    if (foundLayer1 == FALSE)  {
		/*
		 *  The layer has not been installed in LayerMap, so do so.
		 */
		if (lastLayer == MAXLAYERS-1)  {
		    printf("readTech:  too many mask layers; up to ");
		    printf("MAXLAYERS (%d) are allowed.\n", MAXLAYERS);
		    exit(-1);
		}
		lastLayer++;
		NumLayers++;
		sparcsLayer1 = sparcsLayer2 = lastLayer;
		LayerMap[sparcsLayer1].name = STRSAVE(layer1);
		if (resistance == -1)  {
		    LayerMap[sparcsLayer1].ohmsPerSquare = UNDEF;
		} else  {
		    LayerMap[sparcsLayer1].ohmsPerSquare = resistance;
		}
		LayerMap[sparcsLayer1].maxRule = UNITSPERLAMBDA * maxRule;
	    }
	    /*
	     *  Add the rule to the matrix, after checking that there isn't
	     *  already one there.
	     */
	    if (Rules[sparcsLayer1][sparcsLayer1].spacing != NO_RULE)  {
		printf("readTech:  error on line %d; %s-%s rule", line,
							    layer1, layer2);
		printf(" is already present.\n");
		exit(-1);
	    } else  {
		Rules[sparcsLayer1][sparcsLayer1].spacing = spacing;
		Rules[sparcsLayer1][sparcsLayer1].mergeable = 'Y';
	    }
	} else  {
	    /*
	     *  For now, require that the layer1-layer1 rule and the layer2-
	     *  layer2 rule appear before the layer1-layer2 rule.
	     */
	    if (foundLayer1 == FALSE)  {
		printf("readTech:  error on line %d;\n\t%s-%s rule",
			line, layer1, layer1);
		printf(" must appear before %s-%s rule.\n",
			line, layer1, layer2);
		exit(-1);
	    }
	    if (foundLayer2 == FALSE)  {
		printf("readTech:  error on line %d;\n\t%s-%s rule",
			line, layer2, layer2);
		printf(" must appear before %s-%s rule.\n",
			line, layer1, layer2);
		exit(-1);
	    }

	    if (Rules[sparcsLayer1][sparcsLayer2].spacing != NO_RULE)  {
		printf("readTech:  error on line %d; %s-%s rule",
			line, layer1, layer2);
		printf(" is already present.\n");
		exit(-1);
	    } else  {
		Rules[sparcsLayer1][sparcsLayer2].spacing = spacing;
		if (mergeable == 1)  {
	 	    Rules[sparcsLayer1][sparcsLayer2].mergeable = 'Y';
		} else if (mergeable == 0)  {
	 	    Rules[sparcsLayer1][sparcsLayer2].mergeable = 'N';
		} else  {
		    printf("readTech:  error on line %d; 4th ", line);
		    printf("value must be 1 or 0.\n");
		    exit(-1);
		}
	    }
	    if (Rules[sparcsLayer2][sparcsLayer1].spacing != NO_RULE)  {
		printf("readTech:  error on line %d; %s-%s rule",
			line, layer2, layer1);
		printf(" is already present.\n");
		exit(-1);
	    } else  {
		Rules[sparcsLayer2][sparcsLayer1].spacing = spacing;
		if (mergeable == 1)  {
	 	    Rules[sparcsLayer2][sparcsLayer1].mergeable = 'Y';
		} else if (mergeable == 0)  {
	 	    Rules[sparcsLayer2][sparcsLayer1].mergeable = 'N';
		} else  {
		    printf("readTech:  error on line %d; 4th ", line);
		    printf("value must be 1 or 0.\n");
		    exit(-1);
		}
	    }
	}
    }
    if (line == 0)  {
	printf("readTech:  rules file is empty.\n");
	exit(-1);
    }

    if ( SparcsPrintRules ) printRules();

    if (fclose(pf) == EOF)  {
	printf("readTech:  trouble closing rules file.\n");
    }
}


printRules()
{
    int i,j;

    printf("\nreadTech:  %d layers read from ", NumLayers);
    printf("rules file (max. allowed = %d).\n", MAXLAYERS);
    for (i=0; i < NumLayers; i++)  {
	printf("    sparcsLayer %d:  name \"%s\",", i, LayerMap[i].name);
	printf(" ohmsPerSquare %d, maxRule %d.\n",
	       LayerMap[i].ohmsPerSquare, LayerMap[i].maxRule);
	
	for ( j = 0; j < NumLayers; j++ ) {
	    if ( Rules[i][j].spacing != NO_RULE && Rules[i][j].spacing != UNDEF ) {
		printf( "\t%s %s spacing = %d %c\n", LayerMap[i].name, LayerMap[j].name, Rules[i][j].spacing ,
		       Rules[i][j].mergeable );
	    }
	}
    }
}

static int skipLayer( layer )
    octObject *layer;
{
    char* class =  thGetLayerClass( layer );
    if ( class && !strcmp( class, "ANNOTATION" ) ) {
	/* printf( "Skipping layer %s because it is of class ANNOTATION\n", layer->contents.layer.name ); */
	return 1;
    } 
    return 0; /* Do not skip this layer. */
}



static int checkRules()
{
    int i;
    char buf[128];

    for ( i = 0; i < NumLayers; i++ ) {
	if ( LayerMap[i].maxRule == UNDEF ) {
	    fprintf( stderr, "No rules for layer %s, I will ignore it.\n", LayerMap[i].name );
	    sprintf( buf, "%s-ignored", LayerMap[i].name );
	    LayerMap[i].name = util_strsav( buf );
	    noWarningAdd( LayerMap[i].name );
	}
    }
}



readTechFromTap( chip )
    octObject *chip;
{
    octObject tapFacet, layer1, layer2;
    octGenerator gen1, gen2;
    int i1,i2;
    char tapFacetName[256];
    double res;
    int ignoreContactRules = 0;

    /* printf( "Reading tech info from TAP ..." ); */

    thTechnology( chip );
    sprintf( tapFacetName, "%s/tap.views:symbolic:contents", tapGetDirectory( chip ) );
    
    ohUnpackDefaults( &tapFacet, "r", tapFacetName  );
    ohUnpackFacetName( &tapFacet, tapFacetName );

    if ( octOpenFacet( &tapFacet ) < OCT_OK ) {
	errRaise( "SPARCS", 1, "Cannot open tech facet %s", ohFormatName( &tapFacet ) );
    }
    
    NumLayers = ohCountContents( &tapFacet, OCT_LAYER_MASK );

    if ( NumLayers >= MAXLAYERS ) {
	errRaise( "SPARCS", SPARCS_BUG_CODE, "Too many layers %d (max is %d)", NumLayers, MAXLAYERS );
    }

    octInitGenContents( &tapFacet, OCT_LAYER_MASK, &gen1 );
    for ( i1 = 0; octGenerate( &gen1, &layer1 ) == OCT_OK; i1++ ) {

	/* A few thigs to do always */
	LayerMap[i1].name = STRSAVE(layer1.contents.layer.name);

	if ( skipLayer( &layer1 ) ) {
	    LayerMap[i1].ohmsPerSquare = 0;
	    LayerMap[i1].maxRule = 0;
	    Rules[i1][i1].spacing = 0;
	    Rules[i1][i1].mergeable = 'Y';
	    continue;
	}

	res = thGetSheetResistance( &layer1 );
	if ( res >= 0.0 ) {
	    /* If not zero, make it at least 1 */
	    LayerMap[i1].ohmsPerSquare = ( res == 0.0 ) ? 0 : SPMAX(1,(int)res);	
	}


	if (isContact(&layer1) && ignoreContactRules ) {
	    LayerMap[i1].maxRule = 0;
	    Rules[i1][i1].spacing = 0;
	    Rules[i1][i1].mergeable = 'Y';
	} else {
	    octInitGenContents( &tapFacet, OCT_LAYER_MASK, &gen2 );
	    for ( i2 = 0; octGenerate( &gen2, &layer2 ) == OCT_OK; i2++ ) {
		octCoord s;
		if ( skipLayer( &layer2 ) ) continue;

		s = tapGetMinSpacing( &layer1, &layer2 );
		if ( s != TAP_NO_RULE ) {
		    Rules[i1][i2].spacing = s;
		    LayerMap[i1].maxRule = SPMAX( LayerMap[i1].maxRule, s );
		}
		
		Rules[i1][i2].mergeable = (i1 == i2 ) ? 'Y' : 'N'; /* XXX trivial initialization */
		    
		ERR_IGNORE( Rules[i1][i2].mergeable = thGetLayerLayerPropInt( &layer1, &layer2, "MERGEABLE" ) ? 'Y' : 'N');
	    }
	}
    }

    checkRules();

    if ( SparcsPrintRules ) printRules();

}


int isContact( layer )
    octObject *layer;
{
    return !strcmp( thGetLayerClass( layer ), "CONTACT" );
}
