#include "port.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "wolfe.h"

int feed_count = 0;		/* available to main.c for stat purposes */

static int place_cell();
static void sort_rows();
static void shift_rows();
static void dump_rows();


/*
 *  read_place_file: create the block_list structure from the TimberWolf
 *  'pl1' placement file.
 *
 *  The TimberWolf file format is:
 *	1) instance name
 *	2) x-coord lower left
 *	3) y-coord lower left (ignored)
 *	4) x-coord upper right
 *	5) y-coord upper right (ignored)
 *	6) Orientation (0-7)
 *	7) which 'block' number the instance was placed
 *
 *  The instance is given a transform if necessary.  Then, the x-coord is
 *  used to place the cell.  (The y-position is calulcated later when
 *  the routing is actually performed).
 *  
 *  The minimum and maximum x-coords are computed, and all of the cells are
 *  shifted so that the leftmost cell has its left edge on the y-axis.
 */

int read_place_file(prefix, facet)
char *prefix;			/* IN: prefix for the filename */
octObject *facet;		/* IN: the facet */
{
    octObject bag, instance;
    octCoord xl, yl, xu, yu;
    octStatus status;
    block_record_t *block_record;
    FILE *fp;
    char filename[1024], name[1024];
    int i, orient, block, linenum;

    linenum = 0;
    min_xposition = 9999999;
    max_xposition = -9999999;

    /* Open the placement file */
    (void) sprintf(filename, "%s.pl1", prefix);
    if ((fp = fopen(filename, "r")) == NULL) {
	perror(filename);
	return 1;
    }

    /* initialize the linked list of blocks -- GLOBAL Structure */
    block_list = ALLOC(block_record_t *, num_rows+1);
    for(i = 1; i <= num_rows; i++) {
	block_list[i] = NIL(block_record_t);
    }

    /* Read TimberWolf placement file line by line */
    for(;;) {
	linenum++;
	i = fscanf(fp, "%s %d %d %d %d %d %d\n", 
	      name, &xl, &yl, &xu, &yu, &orient, &block);
	/* Check for scanf() failure */
	if (i == EOF) {
	    break;			/* natural termination of loop */
	} else if (i != 7) {
	    (void) fprintf(stderr, "\"%s\", line %d: bad count from scanf\n",
		filename, linenum);
	    return 1;
	}

	/* Check the block number */
	if (block == 0) {
	    continue;			/* ignore PADS */

	/* Make sure the block number is in the range we expect */
	} else if (block < 1 || block > num_rows) {
	    (void) fprintf(stderr, 
		"\"%s\", line %d: block number %d out of range for %s\n",
		filename, linenum, block, name);
	    return 1;
	}
	    

	/* 
	 *  Do we have an instance or a feedthru cell ?
	 */
	if (strncmp(name, "twfeed", 6) == 0) {
	    /* Create a feedthru instance */
	    feed_count++;
	    OH_ASSERT(ohCreateInst(facet, &instance, name, 
				feedthru_cell, feedthru_view));

	    /* Attach to the WOLFE-FEED bag */
	    OH_ASSERT(ohGetOrCreateBag(facet, &bag, "WOLFE-FEED"));
	    OH_ASSERT(octAttach(&bag, &instance));

	    /* attache instance to the connectors bag as well */
	    OH_ASSERT(ohGetOrCreateBag(facet, &bag, "CONNECTORS"));
	    OH_ASSERT(octAttach(&bag, &instance));


	} else {
	    /* A real instance ... */
	    status = ohGetByInstName(facet, &instance, name);
	    if (status != OCT_OK) {
		(void) fprintf(stderr, 
		    "\"%s\", line %d: cannot find %s by name\n",
		    filename, linenum, name);
		OH_ASSERT(status);
	    }
	}

	/* Map the orientation given by TimberWolf into the instance */
	if (place_cell(orient, xl, &instance) != 0) {
	    (void) fprintf(stderr, 
		"\"%s\", line %d: unimplemented transform %d for %s\n",
		filename, linenum, orient, name);
	    return 1;
	}

	/* Add this instance to the linked list for this block */
	block_record = ALLOC(block_record_t, 1);
	block_record->objectId = instance.objectId;
	block_record->next = block_list[block];
	block_list[block] = block_record;

	if (xl < min_xposition) min_xposition = xl;
	if (xu > max_xposition) max_xposition = xu;
    }
    (void) fclose(fp);

    /* Shift the rows towards the y-axis */
    min_xposition = (min_xposition - vlayer_pitch) * units_per_lambda;
    max_xposition = (max_xposition + vlayer_pitch) * units_per_lambda;
    max_xposition -= min_xposition;
    shift_rows();

    /* Sort the rows and (optionally) dump the info */
    sort_rows();
    if (debug > 2) dump_rows();
    return 0;
}

/*
 *  place_cell: map a TimberWolf orientation into an Oct transformation 
 *
 *  The left edge of the cell is placed at 'xl';
 *  The x-axis within the cell is placed on the x-axis
 */

static int place_cell(orient, xl, instance)
int orient;
int xl;
octObject *instance;
{
    struct octBox bb;
    struct octTransform *transform;

    transform = &(instance->contents.instance.transform);

    switch (orient) {
	case 0: 
	    transform->transformType = OCT_NO_TRANSFORM;
	    break;
	case 2: 
	    transform->transformType = OCT_MIRROR_X;
	    break;
	default:
	    return 1;
    }

    /* Apply the transformation */
    OH_ASSERT(octModify(instance));

    /* Now get bounding box and move to proper location */
    OH_ASSERT(octTB(instance, &bb));
    transform->translation.x += xl*units_per_lambda - bb.lowerLeft.x;

    /* The y-coord is assigned in route.c; place it on the x-axis for now */
    transform->translation.y = 0;

    /* Apply the translation */
    OH_ASSERT(octModify(instance));
    return 0;
}

/*
 *  Use the 'generic' linked-list sorter to sort the cells in a row
 */

#define SORT1	sort_block_list
#define SORT	sort_block_list_helper
#define TYPE 	block_record_t
#define NEXT	next
#include "lsort.h"


static int block_compare(pi, pj)
block_record_t *pi, *pj;
{
    octObject object;
    struct octBox bb;
    int lli, llj;

    OH_ASSERT(ohGetById(&object, pi->objectId));
    OH_ASSERT(octBB(&object, &bb));
    lli = bb.lowerLeft.x;
    OH_ASSERT(ohGetById(&object, pj->objectId));
    OH_ASSERT(octTB(&object, &bb));
    llj = bb.lowerLeft.y;
    return lli - llj;
}


static void sort_rows()
{
    int i;

    for(i = 1; i <= num_rows; i++) {
	block_list[i] = sort_block_list(block_list[i], block_compare);
    }
}
	

static void shift_rows()
{
    octObject inst;
    block_record_t *p;
    int i;

    for(i = 1; i <= num_rows; i++) {
	for(p = block_list[i]; p != NIL(block_record_t); p = p->next) {
	    OH_ASSERT(ohGetById(&inst, p->objectId));
	    inst.contents.instance.transform.translation.x -= min_xposition;
	    OH_ASSERT(octModify(&inst));
	}
    }
}


static void dump_rows()
{
    octObject object;
    struct octBox bb;
    block_record_t *p;
    int i;

    (void) printf("\n\nCELL LIST\n");
    for(i = 1; i <= num_rows; i++) {
	(void) printf("\nrow %d:\n", i);
	for(p = block_list[i]; p != NIL(block_record_t); p = p->next) {
	    OH_ASSERT(ohGetById(&object, p->objectId));
	    OH_ASSERT(octTB(&object, &bb));
	    (void) printf("   %-20s ll=(%5d,%5d) ur=(%5d,%5d)\n", 
		object.contents.instance.name, bb.lowerLeft.x, bb.lowerLeft.y,
		bb.upperRight.x, bb.upperRight.y);
	}
    }
}


int discard_block_table()
{
    block_record_t *p, *pnext;
    int i;

    for(i = 1; i <= num_rows; i++) {
	for(p = block_list[i]; p != NIL(block_record_t); p = pnext) {
	    pnext = p->next;
	    FREE(p);
	}
    }
}
