#include "port.h"
#include "utility.h"
#include <sys/wait.h>
#include "oct.h"
#include "wolfe.h"

static int write_yacr_file();
static int read_yacr_file();
static int longest, hcount, vcount, maximum_density, cycle_count;

/*
 *  yacr: perform channel routing
 */

int yacr(top, bottom, left, right, ncol, horz, vert, 
    nrows, ncols, density, cycles)
int *top;		/* IN: list of top pins */
int *bottom;		/* IN: list of bottom pins */
int *left;		/* IN: list of left pins (first is count) */
int *right;		/* IN: list of right pins (first is count) */
int ncol;		/* IN: number of columns in pin lists */
int ***horz;		/* OUT: horizontal layer routing */
int ***vert;		/* OUT: vertical layer routing */
int *nrows;		/* OUT: number of rows in routing */
int *ncols;		/* OUT: number of columns in routing */
int *density;		/* OUT: maximum channel density */
int *cycles;		/* OUT: number of cycles */
{
    FILE *fpin, *fpout;
    static char *argv[] = {"csh", "-f", "-c", "yacr -H", 0};
    static char *debug_argv[] = {
	"csh", "-f", "-c", "tee -a input | yacr -H | tee -a output", 0
    };
    int i, j;
#ifdef _IBMR2
    int status;
#else
    union wait status;
#endif

    if (util_pipefork(debug > 2 ? debug_argv : argv, &fpout, &fpin, 0) == 0) {
	(void) fprintf(stderr, "Unable to pipefork() yacr\n");
	return 1;
    }

    if (debug > 1) {
	(void) printf("YACR Input File\n");
	(void) write_yacr_file(stdout, top, bottom, left, right, ncol);
    }

    if (write_yacr_file(fpout, top, bottom, left, right, ncol) != 0) {
	(void) fprintf(stderr, "Problem writing yacr file\n");
	return 1;
    }
    if (read_yacr_file(fpin,top,bottom,left,right,horz,vert,nrows,ncols) != 0) {
	(void) fprintf(stderr, "Problem reading yacr file\n");
	return 1;
    }
    *density = maximum_density;
    *cycles = cycle_count;

    /* dump the routing if debug is set */
    if (debug > 1) {
	(void) printf("Horizontal layer (sideways)\n");
	for(i = 0; i <= *ncols + 1; i++) {
	    for(j = 0; j <= *nrows + 1; j++) {
		(void) printf(" %3d", (*horz)[j][i]);
	    }
	    (void) printf("\n");
	}
	(void) printf("Vertical layer (sideways)\n");
	for(i = 0; i <= *ncols + 1; i++) {
	    for(j = 0; j <= *nrows + 1; j++) {
		(void) printf(" %3d", (*vert)[j][i]);
	    }
	    (void) printf("\n");
	}
    }

    (void) fclose(fpin);
    (void) wait(&status);
/*    return status.w_status;*/
    return 0;
}

static int int_compare(pi, pj)
int *pi, *pj;
{
    return *pi - *pj;
}


static int write_yacr_file(fp, top, bottom, left, right, ncol)
FILE *fp;
int *top;
int *bottom;
int *left;
int *right;
int ncol;
{
    int *temp, i, j, cnt, last, unique;

    /* How gross!!  Must count # unique net numbers */
    temp = ALLOC(int, ncol*2 + left[0] + right[0] + 10);
    cnt = 0;
    for(i = 0; i < ncol; i++) if (top[i] != 0) temp[cnt++] = top[i];
    for(i = 0; i < ncol; i++) if (bottom[i] != 0) temp[cnt++] = bottom[i];
    for(i = 1; i <= left[0]; i++) temp[cnt++] = left[i];
    for(i = 1; i <= right[0]; i++) temp[cnt++] = right[i];
    qsort((char *) temp, cnt, sizeof(int), int_compare);

    unique = 0;
    last = 0;
    for(i = 1; i <= cnt; i++) {
	if (i == cnt || temp[last] != temp[i]) {
	    if (i - last < 2) {
		(void) fprintf(stderr, 
		    "Single connection in channel group %d\n", temp[last]);
	    }
	    unique++;
	    last = i;
	}
    }
    FREE(temp);

    (void) fprintf(fp, "%d\n", unique);
    (void) fprintf(fp, "%d\n", ncol);
    for(j = 0; j < ncol; j++) (void) fprintf(fp, " %2d", top[j]);
    (void) fprintf(fp, "\n");
    for(j = 0; j < ncol; j++) (void) fprintf(fp, " %2d", bottom[j]);
    (void) fprintf(fp, "\n");
    for(j = 0; j <= left[0]; j++) (void) fprintf(fp, " %2d", left[j]);
    (void) fprintf(fp, "\n");
    for(j = 0; j <= right[0]; j++) (void) fprintf(fp, " %2d", right[j]);
    (void) fprintf(fp, "\n");

    /* return TRUE if any error occured */
    if (fp == stdout) return 0;
    return ferror(fp) || (fclose(fp) == EOF);
}

/*
 *  read_yacr_file: read the output from yacr into 2 2-d arrays
 *
 *  The rows are 1 ... *nrows, and columns are 1 ... *ncols.  
 *
 *  Row 0 of the vertical layer is a copy of the top pin list. 
 *  Row *nrows+1 is a copy of the bottom pin list.
 *  Column 0 of the horizontal layer is a copy of the left pin list.
 *  Column *ncols+1 of the horizontal layer is a copy of the right pin list.
 */

static int read_yacr_file(fp, top, bottom, left, right, horz, vert, nrows, ncols)
FILE *fp;			/* IN: file pointer to read from */
int *top;			/* IN: top pin connections */
int *bottom;			/* IN: bottom pin connections */
int *left;			/* IN: left pin connections */
int *right;			/* IN: right pin connections */
int ***horz;			/* OUT: matrix of horizontal layer */
int ***vert;			/* OUT: matrix of vertical layer */
int *nrows;			/* OUT: number of rows */
int *ncols;			/* OUT: number of columns */
{
    int i, j;

    if (fscanf(fp, "%d\n%d\n", nrows, ncols) != 2) {
	return 1;
    }

    *horz = ALLOC(int *, *nrows+4);
    for(i = 0; i <= *nrows + 1; i++) {
	(*horz)[i] = ALLOC(int, *ncols+4);
	for(j = 0; j <= *ncols + 1; j++) {
	    if (i == 0 || i == *nrows+1 || j == 0 || j == *ncols+1) {
		(*horz)[i][j] = 0;
	    } else if (fscanf(fp, "%d", (*horz)[i] + j) != 1) {
		return 1;
	    }
	}
    }

    *vert = ALLOC(int *, *nrows+4);
    for(i = 0; i <= *nrows + 1; i++) {
	(*vert)[i] = ALLOC(int, *ncols+4);
	for(j = 0; j <= *ncols + 1; j++) {
	    if (j == 0 || j == *ncols+1) {
		(*vert)[i][j] = 0;
	    } else if (i == 0) {
		(*vert)[i][j] = top[j-1];
	    } else if (i == *nrows+1) {
		(*vert)[i][j] = bottom[j-1];
	    } else if (fscanf(fp, "%d", (*vert)[i] + j) != 1) {
		return 1;
	    }
	}
    } 

    /* Patch the left pin list, check horz layer leftmost column */
    for(i = 1; i <= *nrows; i++) {
	if ((*horz)[i][1] != 0) {
	    for(j = 1; j <= left[0]; j++) {
		if ((*horz)[i][1] == left[j]) {
		    (*horz)[i][0] = left[j];
		    break;
		}
	    }
	}
    }

    /* Patch the right pin list, check horz layer rightmost column */
    for(i = 1; i <= *nrows; i++) {
	if ((*horz)[i][*ncols] != 0) {
	    for(j = 1; j <= right[0]; j++) {
		if ((*horz)[i][*ncols] == right[j]) {
		    (*horz)[i][*ncols+1] = right[j];
		    break;
		}
	    }
	}
    }

    /* read special info at end */
    if (fscanf(fp, "%d %d %d %d %d", &longest, &hcount, 
	&vcount, &maximum_density, &cycle_count) != 5)
	return 1;

    /* return TRUE if any error occured */
    return ferror(fp) || (fclose(fp) == EOF);
}
